/************************************************************************
 *
 * Copyright (c) 2004
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/*
 * i2c-algo-inca.c
 *
 * Inca-IP I2C algorithm driver
 *
 * Andre Messerschmidt 
 *
 * For questions about this file please contact your local Infineon 
 * representative.
 *
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>

#include <asm/incaip/inca-ip.h>
#include <asm/incaip/irq.h>
#include <linux/i2c.h>
#include "i2c-algo-inca.h"
#if 1
#define TT printk("TRACE #%d\n",__LINE__)
#else
#define TT 
#endif

static int master = 1;
MODULE_PARM(master, "i");
MODULE_PARM_DESC(master, "Master/Slave Select (1=Master (default), 0=Slave)");

static int slave_addr = IIC_DEFAULT_SLAVE_ADDRESS;
MODULE_PARM(slave_addr, "i");
MODULE_PARM_DESC(slave_addr, "Slave I2C address (default=0x66)");

static int slave_addr_bits = 7;
MODULE_PARM(slave_addr_bits, "i");
MODULE_PARM_DESC(slave_addr_bits, "Slave address bits (7 (default) or 10)");

static int baudrate = IIC_DEFAULT_BAUDRATE;
MODULE_PARM(baudrate, "i");
MODULE_PARM_DESC(baudrate, "Baudrate (100000 (default) or 400000)");

extern unsigned long volatile jiffies;
extern unsigned int incaip_get_cpu_hz(void);

static char local_read_buffer[] = "This is INCA-IP speaking";
#define I2C_SLAVE_WRITE_BUFFER_SIZE 30

static int delay_100[3][2] = {{800,400},{1000,500},{800,400}};
static int delay_400[3][2] = {{200,100},{260,130},{200,100}};
static int* delays;

static struct i2c_device_data{
	char* read_buffer;
	int read_buffer_size;
	int read_buffer_pos;
	int read_bytes;
	char write_buffer[I2C_SLAVE_WRITE_BUFFER_SIZE+1];
	int write_buffer_size;
	int write_buffer_pos;
	int written_bytes;
	char transmitting;
	char receiving;
} i2c_inca_device ;

/* Poll as long as bit is set */
#define POLL_TIMEOUT_SET(bit) \
	{ \
		u32 pollreg=0; \
		unsigned long volatile timeout_value; \
		timeout_value = jiffies+(2*HZ); \
		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,pollreg); \
		while (pollreg & bit) { \
			if (timeout_value < jiffies) { \
				printk("i2c-algo-inca: POLL Timeout: line %d\n",__LINE__); \
				break;\
			} \
			SYNC; \
	 		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,pollreg); \
			NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS; \
		} \
	} \

/* Poll as long as bit is cleared */
#define POLL_TIMEOUT_CLEARED(bit) \
	{ \
		u32 pollreg=0; \
		unsigned long volatile timeout_value; \
		timeout_value = jiffies+(2*HZ); \
		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,pollreg); \
		while (!(pollreg & bit)) { \
			if (timeout_value < jiffies) { \
				printk("i2c-algo-inca: POLL Timeout: line %d\n",__LINE__); \
				break;\
			} \
			SYNC; \
	 		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,pollreg); \
			NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS;NOPS; \
		} \
	} \
		
		
unsigned int incaip_get_fpi_hz(void);


void inca_i2c_data_int(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 reg = 0;
	struct i2c_device_data* dev;
	unsigned long flags;
	
	if (master == 1) return;
	save_and_cli(flags);
	
//	printk("DATA-INT\n");
//	dev = (struct i2c_device_data*) dev_id;
	dev = &i2c_inca_device;
	
	if (dev->transmitting == 1) {
		
		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
		if ((reg & INCA_IP_IIC_IIC_SYSCON_0_IRQE)==0) {
			
			// Write byte to transmit buffer
			I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,
					dev->read_buffer[dev->read_buffer_pos]);
			// Advance buffer position
			dev->read_buffer_pos++;
			if (dev->read_buffer_pos == dev->read_buffer_size) {
				dev->read_buffer_pos = 0;
			}
			dev->read_bytes++;
		}
		else {
			// Write dummy byte to transmit buffer
			I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,0xff);
			dev->transmitting=0;
		}
	}
	if (dev->receiving == 1) {
		// Read byte from receive buffer
		I2C_READ_REG(INCA_IP_IIC_IIC_RTB,
				dev->write_buffer[dev->write_buffer_pos]);
				
//	printk("DATA-INT receiving val=0x%02x\n",dev->write_buffer[dev->write_buffer_pos]&0xff);
		// Advance buffer position
		dev->write_buffer_pos++;
		if (dev->write_buffer_pos == dev->write_buffer_size) {
			dev->write_buffer_pos = 0;
		}
		dev->written_bytes++;
		
	}
	// Clear Data interrupt
	I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
			INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
			0);
	I2C_IRQD_CLEAR();
	restore_flags(flags);
}

void inca_i2c_prot_int(int irq, void *dev_id, struct pt_regs *regs)
{
	u32 reg = 0;
	struct i2c_device_data* dev;
	unsigned long flags;
	
	if (master == 1) return;
	save_and_cli(flags);
	
//	printk("PROT-INT\n");
//	dev = (struct i2c_device_data*) dev_id;
	dev = &i2c_inca_device;
	
	I2C_READ_REG(INCA_IP_IIC_IIC_RTB,reg);
//	printk("PROT-INT reg=0x%02x\n",reg&0xff);

	if ((reg & 0x01) == 1) {
		// READ from master
//		printk("PROT-INT transmit\n");
		
		dev->transmitting = 1;
		dev->receiving = 0;
		// Write byte to transmit buffer
		I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,
				dev->read_buffer[dev->read_buffer_pos]);
		// Advance buffer position
		dev->read_buffer_pos++;
		if (dev->read_buffer_pos == dev->read_buffer_size) {
			dev->read_buffer_pos = 0;
		}
		dev->read_bytes++;

		// Clear Protocol interrupt
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
				INCA_IP_IIC_IIC_WHBSYSCON_SETACKDIS |
				INCA_IP_IIC_IIC_WHBSYSCON_SETTRX |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
				0);
	}
	else {
		// WRITE from master
//		printk("PROT-INT receiving\n");
		
		dev->receiving = 1;
		dev->transmitting = 0;

		// Clear Protocol interrupt
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
				INCA_IP_IIC_IIC_WHBSYSCON_CLRTRX |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
				0);
		
	}
	I2C_IRQP_CLEAR();
	restore_flags(flags);
}

void inca_i2c_end_int(int irq, void *dev_id, struct pt_regs *regs)
{
	u32	reg;
	struct i2c_device_data* dev;
	unsigned long flags;
	
	if (master == 1) return;
	save_and_cli(flags);

	I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
//	printk("END-INT SYSCON=0x%08x\n",reg);
//	dev = (struct i2c_device_data*) dev_id;
	dev = &i2c_inca_device;

	if ((reg & INCA_IP_IIC_IIC_SYSCON_0_BB)==0) {
		if (dev->transmitting==1) {
			if (dev->read_buffer_pos == 0) {
				dev->read_buffer_pos = dev->read_buffer_size;
			}
			dev->read_buffer_pos--;
			dev->read_bytes--;
		}
		dev->transmitting = 0;
		dev->receiving = 0;
		// Clear all interrupts
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
				INCA_IP_IIC_IIC_WHBSYSCON_CLRTRX |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
				0);
		I2C_IRQD_CLEAR();
		I2C_IRQP_CLEAR();
	}
	else {
		// Clear End interrupt
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
			INCA_IP_IIC_IIC_WHBSYSCON_CLRTRX |
			INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS |
			INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE |
			INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
			INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
			0);
		I2C_IRQD_CLEAR();
		I2C_IRQP_CLEAR();
	}
	I2C_IRQE_CLEAR();
	restore_flags(flags);
}

int inca_i2c_set_baudrate(u32 freq)
{
#if 1
	u32 prev, preval, fpifreq, brp=0;
	u32 prediv;
	u32 buscon, reg;

//	fpi_freq = incaip_get_fpi_hz() / 1024;
	fpifreq = incaip_get_fpi_hz();

	
	prediv=2;
	for (preval=64;preval!=0;preval = preval / 8) {
		brp=(1024*freq)/(fpifreq / preval);
		if (brp <= 0xFF) {
			/* this combination fits into the register */
			break;
		}
		prediv--;
	}
	
#if 0
	for (prev = 2, preval = 64; prev > 0; prev--, preval = preval / 8)
	{
		/* calculate for each possible pre devider to get the best value  */
		/* add a half of the value to round up the value                  */
		brp = ((freq * preval) + (fpifreq >> 1)) / fpifreq;

		if (brp <= 0xFF) {
			/* this combination fits into the register */
			break;
		}
	}
	/* calculate the specific register value */
	buscon = (1 << 31) | (prev << 29) | (brp << 8);
#endif 
 
	/* calculate the specific register value */
	buscon = (1 << 31) | (prediv << 29) | (brp << 8);

	printk("freq=%d pre_div=%d brp=%d pre_val=%d\n",fpifreq,prediv,brp,preval);

	I2C_WRITE_REG(INCA_IP_IIC_IIC_BUSCON_0,buscon|INIT_IIP_BUSCON);
#else


	/* update the hardware register */
	if (baudrate == 100000) {
		I2C_WRITE_REG(INCA_IP_IIC_IIC_BUSCON_0,INIT_IIP_BUSCON_100);
	}
	else {
		if (baudrate == 400000){
			I2C_WRITE_REG(INCA_IP_IIC_IIC_BUSCON_0,INIT_IIP_BUSCON_400);
		}
	}
#endif
	return(0);
}

static int inca_i2c_do_start_addr(unsigned char addr,
							unsigned char upper_addr,int restart)
{
	u32 reg=0;
	int t;

	if (master == 0) return 0;
/*
	if (upper_addr != 0x00) {
		POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,upper_addr);
	}
*/
	
	/* Send Repeated Start Condition if already started */ 
	if (restart != 0) {
		POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		
//		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
//		reg |= INCA_IP_IIC_IIC_SYSCON_0_INT;
//		I2C_WRITE_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
		
		/* write device address to the interface */
//		I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,addr);

//		for (t=0;t<200;t++) NOPS;
//		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
//		reg &= ~(INCA_IP_IIC_IIC_SYSCON_0_INT);
//		I2C_WRITE_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);

		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
		if (reg & INCA_IP_IIC_IIC_SYSCON_0_BB) {
//			printk("Restart\n");
			I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
							INCA_IP_IIC_IIC_WHBSYSCON_SETACKDIS |
							INCA_IP_IIC_IIC_WHBSYSCON_SETRSC |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE |
							0);
			/* write device address to the interface */
			I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,addr);
			POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);

// removed for testing
//			for (t=0;t<200;t++) NOPS;

//			POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_RSC);
//			POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_RSC);
//			for (t=0;t<1200;t++) NOPS;
		}
		else {
			I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON, 
							INCA_IP_IIC_IIC_WHBSYSCON_SETBUM |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRSTP |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE |
							0);
							
			/* write device address to the interface */
			I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,addr);
//			for (t=0;t<150;t++) NOPS;
//			for (t=0;t<100;t++) NOPS;
//			POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
			POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_BB);
		
		}
	} 
	else {
//		printk("Fresh start\n");
		/* write device address to the interface */
		I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,addr);
		/* clear all events and send start*/
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON, 
						INCA_IP_IIC_IIC_WHBSYSCON_SETBUM |
  						INCA_IP_IIC_IIC_WHBSYSCON_SETACKDIS |
//  						INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS |
  						INCA_IP_IIC_IIC_WHBSYSCON_CLRSTP |
						INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
  						INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
						INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE | 
						0);
						
//		for (t=0;t<120;t++) NOPS;
//		for (t=0;t<100;t++) NOPS;
		POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_BB);
	
	}
	
	return 0;
}


static int inca_i2c_stop(void)
{
	u32 reg=0;
	int t;
	
//	I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
//	printk("reg=0x%08x\n",reg);
//	for (t=0;t<120;t++) NOPS;
	/* set the stop condition */
	I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
 				INCA_IP_IIC_IIC_WHBSYSCON_CLRBUM |
				INCA_IP_IIC_IIC_WHBSYSCON_SETSTP |
				INCA_IP_IIC_IIC_WHBSYSCON_SETACKDIS |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE);
	
	I2C_IRQD_CLEAR();
	I2C_IRQP_CLEAR();
	I2C_IRQE_CLEAR();
//	POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQE);
//	for (t=0;t<200;t++) NOPS;
	return 0;
}
 
static int inca_i2c_write(unsigned char addr, char buf[], int len)
{
	int count = 0;
	u32 reg = 0;
	int t,i;

	/* enter hardware critical area */
//   semTake(pCon->Semaphore, WAIT_FOREVER);

//	POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
	if (len == 0) return 0;
// removed for testing
//	for (t=0;t<200;t++) NOPS;
 
//printk("write len=%d\n",len);
	while (count < len)
	{
//		i=delays[0];

// MARK		
// 		POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		//for (t=0;t<i;t++) NOPS;


		/* wait till the byte is transmitted */
		/* SYSCON_0 = Write and Read mirror disabled */
 		POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);

		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
		if (reg & INCA_IP_IIC_IIC_SYSCON_0_LRB) {
			printk("i2c_inca_algo.o: Write Missing acknowledgement SYSCON=0x%08x\n",reg);
		}
		if (reg & INCA_IP_IIC_IIC_SYSCON_0_AL) {
			printk("i2c_inca_algo.o: Arbitration lost SYSCON=0x%08x\n",reg);
			I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
 				INCA_IP_IIC_IIC_WHBSYSCON_CLRAL);
		}
		
//		POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		/* write the byte to the interface */
		I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,buf[count]);

		count++; 
 		POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		
	}

	/* SYSCON_0 = Write and Read mirror disabled */
//	POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
	I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
//	printk("reg=0x%08x\n",reg);
	if (reg & INCA_IP_IIC_IIC_SYSCON_0_LRB) {
		printk("i2c_inca_algo.o: Write Missing acknowledgement SYSCON=0x%08x\n",reg);
	}
	if (reg & INCA_IP_IIC_IIC_SYSCON_0_AL) {
		printk("i2c_inca_algo.o: Arbitration lost SYSCON=0x%08x\n",reg);
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
 				INCA_IP_IIC_IIC_WHBSYSCON_CLRAL);
	}
	/* clear all events */
//	for (t=0;t<120;t++) NOPS;
//	printk("write SYSCON=0x%08x count=%d\n",reg,count);
//	I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,(
//						  INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD | 
//						  INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
//						  INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE));


//	for (t=0;t<580;t++) NOPS;
	/* leave hardware critical area */
//   semGive(pCon->Semaphore);
	return count;
}
 
static int inca_i2c_read(unsigned char addr, char buf[],int len)
{
	u32 tmp;
	u32 reg = 0;
	int count = 0;
	int t,i;
	
	if (len == 0)
		return -EINVAL;

	/* enter hardware critical area */
//   semTake(pCon->Semaphore, WAIT_FOREVER);

	/* wait till the byte is transmitted */
	POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);

	I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);
	if (reg & INCA_IP_IIC_IIC_SYSCON_0_LRB) {
		printk("i2c_inca_algo.o: Read Addr Missing acknowledgement SYSCON=0x%08x\n",reg);
	}
//	printk("reg=0x%08x\n",reg);

// removed for testing
//	for (t=0;t<200;t++) NOPS;
	if (len == 1) {
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
				INCA_IP_IIC_IIC_WHBSYSCON_CLRTRX |
				INCA_IP_IIC_IIC_WHBSYSCON_SETSTP |
				INCA_IP_IIC_IIC_WHBSYSCON_SETACKDIS |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE |
				0 );
//		i=delays[1];
//		for (t=0;t<i;t++) NOPS;
// MARK		
// 		POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
	}
	else {
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
				INCA_IP_IIC_IIC_WHBSYSCON_CLRTRX |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP |
				INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE |
				0 );
//		i=delays[1];
//		for (t=0;t<i;t++) NOPS;
// MARK		
// 		POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
	}
	
	/* read the dummy byte back from the interface */
	I2C_READ_REG(INCA_IP_IIC_IIC_RTB,tmp);
	POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);

	while (count < len) {
//		i=delays[1];
//		for (t=0;t<i;t++) NOPS;
// MARK		
// 		POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		POLL_TIMEOUT_CLEARED(INCA_IP_IIC_IIC_SYSCON_0_IRQD);

		if (count == (len-2)) {
			I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,
					INCA_IP_IIC_IIC_WHBSYSCON_SETSTP |
					0 );
		}
		
		
		/* get a byte from the interface */
		I2C_READ_REG(INCA_IP_IIC_IIC_RTB,reg);
	 	buf[count] = reg & 0xff;
		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,reg);

		count++;
 		POLL_TIMEOUT_SET(INCA_IP_IIC_IIC_SYSCON_0_IRQD);
		
	}
	
	/* leave hardware critical area */
//   semGive(pCon->Semaphore);

	return count;
}

 
/*
 * Create address for I2C device (lower part)
 */
static unsigned char inca_i2c_get_addr(struct i2c_msg* pmsg)
{
	unsigned char addr;

	/* create address */
	addr = ( pmsg->addr << 1 );
	if (pmsg->flags & I2C_M_RD )
		addr |= 1;
	if (pmsg->flags & I2C_M_REV_DIR_ADDR )
		addr ^= 1;
	//	printk("i2c-algo-inca.o: address=0x%02x\n",addr&0xff);
	return addr;
}

/*
 * Create upper part of 10bit addresses
 */
static unsigned char inca_i2c_get_upper_addr(struct i2c_msg* pmsg)
{
	unsigned char addr;

	/* create address */
	if (pmsg->flags & I2C_M_TEN) {
		addr = (pmsg->addr >> 8)&0x03;
		addr = addr | IIC_10BIT_PREFIX;
	}
	else {
		return 0;
	}
	return addr;
}

static int inca_xfer(struct i2c_adapter *i2c_adap,
		    struct i2c_msg msgs[], int num)
{
	struct i2c_msg *pmsg;
	unsigned long flags=0;
	
//	struct i2c_algo_inca_data *adap = i2c_adap->algo_data;
	
	int i=0,ret;
	unsigned char addr;
	unsigned char upper_addr=0;
	u32 reg;
	int t;
	
//	save_and_cli(flags);
//	for (t=0;t<1700;t++) NOPS;
	
	//	printk("i2c-algo-inca.o: inca_xfer entered. %d messages\n",num);
	for (i=0;i<num;i++) {
//		printk("i2c-algo-inca.o: inca_xfer message #%d\n",i);
		pmsg = &msgs[i];

		addr = inca_i2c_get_addr(pmsg);
/*
		upper_addr = inca_i2c_get_upper_addr(pmsg);
*/		
		if (!(pmsg->flags & I2C_M_NOSTART)) {
			inca_i2c_do_start_addr(addr,upper_addr,i);
		}

		if (pmsg->flags & I2C_M_RD ) {
			//printk("i2c-algo-inca.o: reading\n");
			/* read bytes into buffer*/
			//printk("i2c-algo-inca.o: read %d bytes.\n",ret);
			ret = inca_i2c_read(addr,pmsg->buf,pmsg->len);
			if (ret < pmsg->len ) {
				return (ret<0)? ret : -EREMOTEIO;
			}
		} else {
			//printk("i2c-algo-inca.o: writing\n");
			/* write bytes from buffer */
			//printk("i2c-algo-inca.o: writing #%d\n",pmsg->len);
			ret = inca_i2c_write(addr,pmsg->buf,pmsg->len);
			if (ret < pmsg->len ) {
				return (ret<0) ? ret : -EREMOTEIO;
			}
		}
	}
	inca_i2c_stop();
	
//	restore_flags(flags);
	return num;
}

static int inca_control(struct i2c_adapter *adapter, 
	unsigned int cmd, unsigned long arg)
{
	return 0;
}

static u32 inca_func(struct i2c_adapter *adap)
{
//	return I2C_FUNC_SMBUS_EMUL | I2C_FUNC_10BIT_ADDR | 
//	       I2C_FUNC_PROTOCOL_MANGLING;
	return I2C_FUNC_SMBUS_EMUL | I2C_FUNC_PROTOCOL_MANGLING;
}


static struct i2c_algorithm i2c_inca_algo = {
	"Inca-IP algorithm",
	I2C_ALGO_INCA,
	inca_xfer,
	NULL,
	NULL,				/* slave_xmit		*/
	NULL,				/* slave_recv		*/
	inca_control,			/* ioctl		*/
	inca_func,			/* functionality	*/
};


int i2c_inca_add_bus(struct i2c_adapter *adap)
{
	u32  tmp;
	u32	addr=0;
	u32	upper_addr=0;
	u32	reg;

	printk("i2c-algo-inca.o: hw routines for %s registered.\n",
	            adap->name);

	/* Enable the IIC Global Bit in the 'PM Global Enable Register' */
	I2C_READ_REG(INCA_IP_PMU_PM_GEN,reg);
	reg |= INCA_IP_PMU_PM_GEN_EN15;
	I2C_WRITE_REG(INCA_IP_PMU_PM_GEN,reg);

	/* set the basic initialization of the IIC interface */
	I2C_WRITE_REG(INCA_IP_IIC_IIC_CLC,INIT_IIP_CLC);
//	I2C_WRITE_REG(INCA_IP_IIC_IIC_PISEL,INIT_IIP_PISEL);
	I2C_WRITE_REG(INCA_IP_IIC_IIC_SYSCON_0,INIT_IIP_SYSCON);

	inca_i2c_set_baudrate(baudrate);

	I2C_WRITE_REG(INCA_IP_IIC_IIC_RTB,INIT_IIP_RTB);
	I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,INIT_IIP_WHBSYSCON);

 	if (master == 1) {
		/* set TRX, BUM, Transmit 1 Byte Buffer, Master Mode */
		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,tmp);
		tmp = (tmp & ~(T4BYTE | MMMODE | M10)) | (BUM | T1BYTE | MSMODE);
		I2C_WRITE_REG(INCA_IP_IIC_IIC_SYSCON_0,tmp);

		/* set the stop condition */
		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON,(INCA_IP_IIC_IIC_WHBSYSCON_CLRBUM | 
							INCA_IP_IIC_IIC_WHBSYSCON_SETACKDIS | 
							INCA_IP_IIC_IIC_WHBSYSCON_SETSTP));
	} 
	else {
		/* set TRX, BUM, Transmit 1 Byte Buffer, Master Mode */
		I2C_READ_REG(INCA_IP_IIC_IIC_SYSCON_0,tmp);
		tmp = (tmp & ~(T4BYTE | MMMODE | M10 | LRB | IGE)) | 
					(T1BYTE | SLMODE );
/*
		if (slave_addr_bits == 10) tmp = tmp | M10;
*/
		I2C_WRITE_REG(INCA_IP_IIC_IIC_SYSCON_0,tmp);

		I2C_READ_REG(INCA_IP_IIC_IIC_BUSCON_0,tmp);
		addr=slave_addr;
/*
		if (slave_addr_bits == 10) {
			upper_addr= (0x300 & slave_addr) << 16;
			tmp = (tmp & ~((0x3) << 24));
			tmp = (tmp | upper_addr);
		}
*/
		tmp = (tmp & ~((0x7f) << 17));
		tmp = (tmp | ((addr&0xff) << 17));
 		I2C_WRITE_REG(INCA_IP_IIC_IIC_BUSCON_0,tmp);

		I2C_WRITE_REG(INCA_IP_IIC_IIC_WHBSYSCON, 
							INCA_IP_IIC_IIC_WHBSYSCON_CLRACKDIS | 
							INCA_IP_IIC_IIC_WHBSYSCON_CLRTRX | 
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQD | 
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQP | 
							INCA_IP_IIC_IIC_WHBSYSCON_CLRIRQE | 
							0);
		I2C_IRQD_CLEAR();
		I2C_IRQP_CLEAR();
		I2C_IRQE_CLEAR();
		I2C_IRQD_ENABLE();
		I2C_IRQP_ENABLE();
		I2C_IRQE_ENABLE();
	}
	
//	I2C_WRITE_REG(INCA_IP_IIC_IIC_SYSCON_0,tmp);
//	printk("syscon=0x%08x\n",tmp);
	/* register new adapter to i2c module... */

	adap->id |= i2c_inca_algo.id;
	adap->algo = &i2c_inca_algo;

	adap->timeout = 100;	/* default values, should	*/
	adap->retries = 3;	/* be replaced by defines	*/

#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	i2c_add_adapter(adap);

	return 0;
}

int i2c_inca_del_bus(struct i2c_adapter *adap)
{
	int res;
	u32	reg;

	/* Enable the IIC Global Bit in the 'PM Global Enable Register' */
	I2C_READ_REG(INCA_IP_PMU_PM_GEN,reg);
	reg &= ~(INCA_IP_PMU_PM_GEN_EN15);
	I2C_WRITE_REG(INCA_IP_PMU_PM_GEN,reg);

	printk("i2c-algo-inca.o: adapter unregistered: %s\n",adap->name);

	if ((res = i2c_del_adapter(adap)) < 0) {
		printk("i2c-algo-inca.o: error while unregistering: %d\n",res);
		return res;
	}
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}

int i2c_algo_read_procmem(char *buf, char **start, off_t offset, 
					int count, int *eof, void *data)
{
	long	len = 0;
	
	len = sprintf(buf,"\nInca-IP I2C Settings\n");
	len=len+sprintf(buf+len,"Baudrate=%d\n",baudrate);
//	len=len+sprintf(buf+len,"Delays %d %d\n",delays[0],delays[1]);
	if (master == 1) len=len+sprintf(buf+len,"Master\n");	
	else {
		len=len+sprintf(buf+len,"Slave (Addr=0x%02x)\n",slave_addr&0xff);
		len=len+sprintf(buf+len,"Write Buffer: %s\n",i2c_inca_device.write_buffer);
		len=len+sprintf(buf+len,"Write Buffer Pos: %d\n",i2c_inca_device.write_buffer_pos);
		len=len+sprintf(buf+len,"Write Buffer Size: %d\n",i2c_inca_device.write_buffer_size);
		len=len+sprintf(buf+len,"Written Bytes: %d\n",i2c_inca_device.written_bytes);
		len=len+sprintf(buf+len,"Read Buffer: %s\n",i2c_inca_device.read_buffer);
		len=len+sprintf(buf+len,"Read Buffer Pos: %d\n",i2c_inca_device.read_buffer_pos);
		len=len+sprintf(buf+len,"Read Buffer Size: %d\n",i2c_inca_device.read_buffer_size);
		len=len+sprintf(buf+len,"Read Bytes: %d\n",i2c_inca_device.read_bytes);
		
	}
		
	*eof = 1;	// No more data available
	return len;
	
}

int __init i2c_algo_inca_init (void)
{
	int err =0;
	int t;
	unsigned int cpu_freq;
	int cpu_index=0;
	
	cpu_freq=incaip_get_cpu_hz();
	switch (cpu_freq) {
		case 100000000: cpu_index=0;
					 break;
		case 133333333: cpu_index=1;
					 break;
		case 150000000: cpu_index=2;
					 break;
		default:		 cpu_index=0;
					 break;
	}
/*
	if (baudrate == 100000) {
		delays = &delay_100[cpu_index][0];
	}
	else {
		if (baudrate == 400000){
			delays = &delay_400[cpu_index][0];
		}
		else return -EINVAL;
	}
*/	
	for (t=0;t<I2C_SLAVE_WRITE_BUFFER_SIZE+1;t++) i2c_inca_device.write_buffer[t]=0x00;
	
	i2c_inca_device.read_buffer = local_read_buffer;
	i2c_inca_device.read_buffer_size = strlen(i2c_inca_device.read_buffer);
	i2c_inca_device.read_buffer_pos = 0;
	i2c_inca_device.read_bytes = 0;
 	i2c_inca_device.write_buffer_size = I2C_SLAVE_WRITE_BUFFER_SIZE;
 	i2c_inca_device.write_buffer_pos = 0;
	i2c_inca_device.written_bytes = 0;	
 	i2c_inca_device.transmitting = 0;
 	i2c_inca_device.receiving = 0;
	
	if (master == 0) {
		err = request_irq(INT_NUM_IM2_IRL22, inca_i2c_data_int, SA_INTERRUPT,
							"i2c_data", &i2c_inca_device);
		err = request_irq(INT_NUM_IM2_IRL23, inca_i2c_prot_int, SA_INTERRUPT,
							"i2c_prot", &i2c_inca_device);
		err = request_irq(INT_NUM_IM2_IRL24, inca_i2c_end_int, SA_INTERRUPT,
							"i2c_end", &i2c_inca_device);
	}

	create_proc_read_entry("driver/inca_i2c", 0, NULL, i2c_algo_read_procmem, NULL);
 
	printk("i2c-algo-inca.o: i2c INCA-IP algorithm module\n");
	return 0;
}

module_init(i2c_algo_inca_init);

#ifdef MODULE
MODULE_AUTHOR("Andre Messerschmidt");
MODULE_DESCRIPTION("I2C-Bus algorithm for INCA-IP");
MODULE_LICENSE("GPL");

/*
int init_module(void) 
{
	return i2c_algo_inca_init();
}
*/

void cleanup_module(void) 
{
	if (master == 0) {
		I2C_IRQD_DISABLE();
		I2C_IRQP_DISABLE();
		I2C_IRQE_DISABLE();
		free_irq(INT_NUM_IM2_IRL22, NULL);
		free_irq(INT_NUM_IM2_IRL23, NULL);
		free_irq(INT_NUM_IM2_IRL24, NULL);
	}
	remove_proc_entry("driver/inca_i2c",NULL);
}
#endif
