/************************************************************************
 *
 * Copyright (c) 2004
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
 
#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/slab.h> // former "malloc.h"
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <linux/bitops.h> /* ffs */

#include <asm/incaip/inca-ip.h>
#include <asm/incaip/mux.h>
//#  define inca_mux_key_pins() 8 /* to be removed when mux module exists! */
#include <asm/incaip/irq.h>

#include <asm/incaip/keypad.h>
#include "keypad_defs.h"


#undef KEYPAD_DEBUG


/*>>>>>>>>>>>>>>>>>>>>>>>>> Module Specific Stuff <<<<<<<<<<<<<<<<<<<<<<<<<<<*/
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andreas Schmidt");
MODULE_DESCRIPTION("Inca-IP Keypad Driver");
/*
 * Parameter Definitions for Module
 */
static int key_pins = -1;	// Number of KEY pins (2..14) 
MODULE_PARM(key_pins, "i");
MODULE_PARM_DESC(key_pins, "Number of KEY pins (2..14)");

static int major = 0;	// Major ID of keypad device (default: 0=dynamic)
MODULE_PARM(major, "i");
MODULE_PARM_DESC(major, "Major ID of keypad device (default: 0=dynamic)");

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>> Global Variables <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
keypad_callback_t keypad_callback = NULL;

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Local Variables <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static int opened_by_user = 0;	/* evtl. replace with spin_lock */
static int opened_by_kernel = 0;

static u16 scan_lines_buf0[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
static u16 scan_lines_buf1[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
static u16 *last_scan_lines = scan_lines_buf0;
static u16 *curr_scan_lines = scan_lines_buf1;

static key_event_t key_fifo[KEY_FIFO_DEPTH];
static key_event_t *rd_ptr = key_fifo;
static key_event_t *wr_ptr = key_fifo;

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct proc_dir_entry *inca_keypad_proc;

static int inca_keypad_read_procmem(char *buf, char **start, off_t offset, 
					int count, int *eof, void *data)
{
	long	len = 0;
	int i;
	char *prefix;

	for (i = 0; i < KEY_FIFO_DEPTH; i++) {
		if (&key_fifo[i] == rd_ptr && &key_fifo[i] == wr_ptr) 
			prefix = "rw->";
		else if (&key_fifo[i] == rd_ptr) prefix = "rd->";
		else if (&key_fifo[i] == wr_ptr) prefix = "wr->";
		else prefix = "    ";
		len += sprintf(buf+len,"%skeycode[%02d]=%02d, state=%d, "
					"timestamp=%u\n", prefix, i,
					key_fifo[i].key_code,
					key_fifo[i].key_state,
					(u32) key_fifo[i].time_stamp);
	}
	*eof = 1;	// No more data available
	return len;
}

static int inca_keypad_write_procmem(struct file *file, const char *buffer,
					unsigned long count, void *data)
{
	return count;
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Device Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct file_operations keypad_fops = {
	open    : inca_keypad_open,
	release : inca_keypad_release,
	ioctl   : inca_keypad_ioctl,
	poll    : inca_keypad_poll
};

/**
 *     inca_keypad_open - Open keypad device
 *
 *     Returns -EBUSY if the keypad is already reserved by kernel module.
 *     Returns 0 if OK.
 */
static int inca_keypad_open(struct inode *inode, struct file *filp)
{
	keypad_dev_t *dev;
	int err =0;
	
	opened_by_user = 1;
	if (!opened_by_kernel) {
		filp->private_data = kmalloc(sizeof(keypad_dev_t), GFP_KERNEL);
		if (filp->private_data == NULL) {
			opened_by_user = 0;
			return -ENOMEM;
		}
		dev = (keypad_dev_t *) filp->private_data;
		init_waitqueue_head(&dev->waitq);
		
		/* Interrupt */
		err = request_irq(INT_NUM_IM2_IRL6, inca_keypad_isr, 0,
							"keypad", dev);
		if (err) {
			printk(KERN_ERR "Could not request_irq().\n");
			kfree(filp->private_data);
			opened_by_user = 0;
			return err;
		}
		KEYPAD_INTERRUPTS_CLEAR();
		KEYPAD_INTERRUPTS_ENABLE();
		printk(KERN_INFO "Keypad interrupt enabled.\n");
		return 0;
	}
	else {
		opened_by_user = 0;
		printk(KERN_ERR "Keypad driver already registered by "
						"other kernel module.\n");
		return -EBUSY;
	}
}

/**
 *     inca_keypad_release - Release keypad device
 *
 *     Returns 0 (=OK).
 */
static int inca_keypad_release(struct inode *inode, struct file *filp)
{
	keypad_dev_t *dev;
	
	KEYPAD_INTERRUPTS_DISABLE();
	dev = (keypad_dev_t *) filp->private_data;
	free_irq(INT_NUM_IM2_IRL6, dev);

	kfree(filp->private_data);
	opened_by_user = 0;

	return 0;
}

/**
 *     inca_keypad_ioctl - Ioctl call entry
 *     @cmd: KEYPAD_GET - Get key event via Ioctl call
 *           (only this command supported!)
 *
 *     Returns -EAGAIN if no key event stored in FIFO.
 *     Returns -EINVAL if wrong command was issued.
 *     Returns positive integer if OK:
 *               <timestamp[30:8]> + <key_state[7:7]> + <keycode[6:0]>.
 */
static int inca_keypad_ioctl(struct inode *inode, struct file *filp, 
					unsigned int cmd, unsigned long arg) 
{
	int res = 0;
	
	switch (cmd) {
	case KEYPAD_GET:
		if (rd_ptr != wr_ptr) {
			if (copy_to_user((void *) arg, (void *) rd_ptr,
					sizeof(rd_ptr[0])))
				res = -EFAULT;
#ifdef KEYPAD_DEBUG					
			printk(KERN_INFO "keycode=%02d, state=%d, "
				"timestamp=%d, res=%08X\n",
				rd_ptr->key_code, rd_ptr->key_state,
				(u32) rd_ptr->time_stamp, res);
#endif			
			PTR_INCR(rd_ptr);
		}
		else {
			res = -EAGAIN; /* no key event stored */
		}
		break;
	default:
		res = -ENOIOCTLCMD;
	}
	return res;
}

/**
 *     inca_keypad_poll - entry for select() call from user application.
 *                        
 *     Returns 0 if the key FIFO is empty.
 *     Returns POLLPRI otherwise.
 */
static unsigned int inca_keypad_poll(struct file *filp, poll_table *wait)
{
	unsigned int mask = 0;
	keypad_dev_t *dev = filp->private_data;
	
	poll_wait(filp, &dev->waitq, wait);
	if (rd_ptr != wr_ptr) mask |= POLLPRI; /* FIFO is not empty */
	
	return mask;
	
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>> Public Functions <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
/**
 *     inca_key_register_callback - Called by other kernel module to register 
 *                                  a callback function for key events.
 *     @fp: function pointer
 *
 *     Returns -EBUSY if keypad device is already opened by user application.
 *     Returns 0 if OK.
 */
int inca_key_register_callback(keypad_callback_t fp) {
	int err =0;
	
	opened_by_kernel = 1;
	if (!opened_by_user) {
		keypad_callback = fp;
		
		/* Interrupt */
		err = request_irq(INT_NUM_IM2_IRL6, inca_keypad_isr, 0,
							"keypad", NULL);
		if (err) {
			printk(KERN_ERR "Could not request_irq().\n");
			return -EBUSY;
		}
		KEYPAD_INTERRUPTS_CLEAR();
		KEYPAD_INTERRUPTS_ENABLE();
		printk(KERN_INFO "Keypad interrupt enabled.\n");
		return 0;
	}
	else {
		opened_by_kernel = 0;
		printk(KERN_ERR "Keypad driver already opened by user "
							"application.\n");
		return -EBUSY;
	}
}

/**
 *     inca_key_unregister_callback - Called by other kernel module to 
 *                                    unregister its callback function. 
 *     @scan_lines: pointer to the scan lines array
 *
 *     Returns 0 (=OK).
 */
int inca_key_unregister_callback() {
	KEYPAD_INTERRUPTS_DISABLE();
	free_irq(INT_NUM_IM2_IRL6, NULL);
	
	keypad_callback = NULL;
	opened_by_kernel = 0;
	return 0;
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Local Functions <<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
/**
 *     get_scanlines - Read keypad scan results from registers and prepare 
 *                     for further processing.
 */
static void get_scanlines(u16 *scan_lines) 
{
	scan_lines[ 0] = (u16) (~(*INCA_IP_TSF_SREG01)        ) & 0x1FFF;
	scan_lines[ 1] = (u16) (~(*INCA_IP_TSF_SREG01)    >>15) & 0x1FFE;
	scan_lines[ 2] = (u16) (~(*INCA_IP_TSF_SREG23)    << 2) & 0x1FFC;
	scan_lines[ 3] = (u16) (~(*INCA_IP_TSF_SREG23)    >>13) & 0x1FF8;
	scan_lines[ 4] = (u16) (~(*INCA_IP_TSF_SREG456)   << 4) & 0x1FF0;
	scan_lines[ 5] = (u16) (~(*INCA_IP_TSF_SREG456)   >>11) & 0x1FE0;
	scan_lines[ 6] = (u16) (~(*INCA_IP_TSF_SREG456)   >>18) & 0x1FC0;
	scan_lines[ 7] = (u16) (~(*INCA_IP_TSF_SREG7to12) << 7) & 0x1F80;
	scan_lines[ 8] = (u16) (~(*INCA_IP_TSF_SREG7to12)     ) & 0x1F00;
	scan_lines[ 9] = (u16) (~(*INCA_IP_TSF_SREG7to12) >> 7) & 0x1E00;
	scan_lines[10] = (u16) (~(*INCA_IP_TSF_SREG7to12) >>10) & 0x1C00;
	scan_lines[11] = (u16) (~(*INCA_IP_TSF_SREG7to12) >>13) & 0x1800;
	scan_lines[12] = (u16) (~(*INCA_IP_TSF_SREG7to12) >>16) & 0x1000;
};								     

/**
 *     validate_scanlines - Check keypad scan results wrt. row/column rules
 *                          (refer to User's Manual for details). On error,
 *                          reset all resulting scan lines.
 *     @scan_lines: pointer to the scan lines array
 */
static int validate_scanlines(u16 *scan_lines) 
{
	int i, first_set, err = 0;
	u16 dead_clm = 0;
	
	for (i = 0; i < key_pins -1; i++) {
		if (scan_lines[i]) {
			first_set = ffs(scan_lines[i]) -1;
//			printk(KERN_INFO "ffs(0x%08X)=%d\n", scan_lines[i], first_set); 
			/* only one bit in the row? 
			   and no conflict with other column yet? */
			if ((1 << first_set) == scan_lines[i] && 
					!(scan_lines[i] & dead_clm)) {

				dead_clm |= scan_lines[i];
			}
			else {
				printk(KERN_ERR "row/column conflict!\n"); 
				err = -1; 
				break;
			}
		}
	}
	
	return err;	
};

#ifdef KEYPAD_DEBUG
static u32 keymap = 0;
#endif
/**
 *     add_keycodes - Compare the last scan with the current scan and derive
 *                    key events. Add these key events to FIFO.
 *     @last_scan: results from last keypad scan.
 *     @curr_scan: results from current keypad scan.
 *
 *     Returns count of added key events.
 */
static int add_keycodes(u16 *last_scan, u16 *curr_scan) 
{
	int i, first_set, keybase = 0, key_state, keycount = 0;
	u16 delta;
	char keycode;
	unsigned long time_stamp;
	
	for (i = 0; i < key_pins -1; i++) {
		delta = last_scan[i] ^ curr_scan[i];
		while (delta) {
			first_set = ffs(delta) -1;
			keycode = keybase + first_set;
			key_state = (  (curr_scan[i] >> first_set) & 1
					? KEY_PRESSED
					: KEY_RELEASED);
			time_stamp = jiffies;
			
			if (!FIFO_FULL(rd_ptr, wr_ptr)) {
				/* add keycodes to FIFO */
				wr_ptr->key_code = keycode;
				wr_ptr->key_state = key_state;
				wr_ptr->time_stamp = time_stamp;
				PTR_INCR(wr_ptr);
				keycount++;
			} 
			else {
				printk(KERN_ERR "Key buffer full. "
					"Ignoring key event!\n");
				break;
			}
			
#ifdef KEYPAD_DEBUG
			if (key_state) {
				keymap |= (1 << keycode);
			}
			else {
				keymap &= ~(1 << keycode);
			}
#endif			
			delta &= ~(1 << first_set);
		}
		keybase += (key_pins-i-2);
	}
#ifdef KEYPAD_DEBUG
	printk("\r");
	for (i=0; i<32;i++) {
		if ((keymap >> i) & 1) printk("%2d", i);
		else printk("  ");
	}
	printk("           ");
#endif	
	return keycount;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> Interrupt <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
 *     inca_keypad_isr - Keypad interrupt handler. 
 */
void inca_keypad_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	int count;
	keypad_dev_t *dev = (keypad_dev_t *) dev_id;
	
//	printk(KERN_INFO "Keypad Interrupt @time %d\n", (int) jiffies);
	
	get_scanlines(curr_scan_lines);
	
	if (validate_scanlines(curr_scan_lines) == 0) { /* i.e. valid! */
	
		count = add_keycodes(last_scan_lines, curr_scan_lines);
		if (count == -1) return; /* leave scan codes unchanged! */
		
		/* issue callbacks to other kernel module, if available */
		if (opened_by_kernel) {
			while (rd_ptr != wr_ptr) {
				(*keypad_callback) (rd_ptr->key_code,
							rd_ptr->key_state,
							rd_ptr->time_stamp);
				PTR_INCR(rd_ptr);
			}
		}

		/* wake up user process, if available */
		if (opened_by_user) {
			wake_up_interruptible(&dev->waitq);
		}
/*	
		for (i=0; i<13; i++) {
			printk(KERN_INFO 
				"validated curr_scan_lines[%2d]=%04X\n", 
				i, curr_scan_lines[i]);
		}	
*/
		/* concept: buffer switching */
		if (last_scan_lines == scan_lines_buf1) {
			last_scan_lines = scan_lines_buf0;
			curr_scan_lines = scan_lines_buf1;
		}
		else {
			last_scan_lines = scan_lines_buf1;
			curr_scan_lines = scan_lines_buf0;
		}
	}
	else return; /* leave scan codes unchanged! */
	
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>> Module Init/Exit <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/****************************************************************************
Description : Module initialization
Arguments   :
Return      :  0 - OK
              <0 - ERROR
Remarks     :
****************************************************************************/
static int __init inca_keypad_init(void)
{
	int err = 0, res;
	unsigned long flags;
	
	if (key_pins == -1) key_pins = inca_mux_key_pins();
	
	if (key_pins < 2) {	/* too few KEY pins requested by user! */
		printk(KERN_ERR "too few KEY pins requested by user "
							"(min: 2)!\n");
		return -EINVAL;
	}
	
	/* allocate Linux resources */
	if (request_region((u32) INCA_IP_TSF_KEY_CONF, 0x14, 
					"INCA-IP Keypad") == NULL) {
		printk(KERN_ERR "request_region() failed!\n");
		err = -EBUSY;
		goto FAIL_REQUEST_REGION;
	}
	
#ifndef DONE_ELSEWHERE	
	/* enable clock of the TSF (LED, Key, PWM) in Power Management Unit */
 	save_flags(flags); cli();
	*INCA_IP_PMU_PM_GEN |= INCA_IP_PMU_PM_GEN_EN9;   
 	restore_flags(flags);
#endif	 


	/* enable the keypad */
 	save_flags(flags); cli();
	*INCA_IP_TSF_TSF_CONF |= INCA_IP_TSF_TSF_CONF_KEYEN;
//	*INCA_IP_Ports_P2_PUDSEL |= 0x6000; /* pull-up characteristic */
//	*INCA_IP_Ports_P2_PUDEN  |= 0x6000; /* enable pull(-up) transistor */
 	restore_flags(flags);
	*INCA_IP_TSF_KEY_CONF = key_pins -1; /* scan KEY0-7 */
	
   	/* register keypad device */
   	res = register_chrdev(major, "keypad", &keypad_fops);
	if (res < 0) {
		printk(KERN_ERR "Could not register keypad device!\n");
		goto FAIL_REGISTER_DEVICE;
	}
	if (major == 0) major = res;
	
	SET_MODULE_OWNER(&keypad_fops);
	
	printk(KERN_NOTICE "Keypad device registered. major: %d.\n", major);
	
	/* create PROC entry */
	inca_keypad_proc = create_proc_entry( "driver/keypad", 0, 0);
	if (!inca_keypad_proc) {
		printk(KERN_ERR "could not create /proc/driver/keypad!\n");
		err = -ENOENT;
		goto FAIL_CREATE_PROC;
	}
	inca_keypad_proc->read_proc  = inca_keypad_read_procmem;
	inca_keypad_proc->write_proc = inca_keypad_write_procmem;

	printk(KERN_NOTICE "/proc/driver/keypad created!\n");

	
	printk(KERN_NOTICE "Inca-IP keypad initialized with pins KEY[%d:0].\n", 
								key_pins-1);
//	printk(KERN_NOTICE "Register P2_PUDESL=%08X, P2_PUDEN=%08X.\n", 
//			*INCA_IP_Ports_P2_PUDSEL, *INCA_IP_Ports_P2_PUDEN);

	return 0; /* success */

	/* if initialization fails, release all resources aquired so far! */
//	remove_proc_entry("driver/keypad", 0);
FAIL_CREATE_PROC:
	unregister_chrdev(major, "keypad");
FAIL_REGISTER_DEVICE:
	release_region((u32) INCA_IP_TSF_KEY_CONF, 0x14);
FAIL_REQUEST_REGION:
	return err;
}

/****************************************************************************
Description : Module exit
Arguments   :
Return      : 
Remarks     :
****************************************************************************/
static void __exit inca_keypad_exit(void)
{
	unsigned long flags;
	
	/* disable the keypad module */
 	save_flags(flags); cli();
	*INCA_IP_TSF_KEY_CONF    = 0x0; /* disable key scan */
	*INCA_IP_TSF_TSF_CONF    &= ~INCA_IP_TSF_TSF_CONF_KEYEN;
 	restore_flags(flags);

	/* Release Linux Resources (in inverse order) */
	remove_proc_entry("driver/keypad", 0);

	unregister_chrdev(major, "keypad");

	release_region((u32) INCA_IP_TSF_KEY_CONF, 0x14);
	
	printk(KERN_NOTICE "/proc/driver/keypad removed!\n");

}

#ifdef MODULE
module_init(inca_keypad_init);
module_exit(inca_keypad_exit);
#endif
