#ifndef _SYS_DRV_CUSTOMER_H
#define _SYS_DRV_CUSTOMER_H
/****************************************************************************
       Copyright (c) 2003, Infineon Technologies.  All rights reserved.

                               No Warranty
   Because the program is licensed free of charge, there is no warranty for
   the program, to the extent permitted by applicable law.  Except when
   otherwise stated in writing the copyright holders and/or other parties
   provide the program "as is" without warranty of any kind, either
   expressed or implied, including, but not limited to, the implied
   warranties of merchantability and fitness for a particular purpose. The
   entire risk as to the quality and performance of the program is with
   you.  should the program prove defective, you assume the cost of all
   necessary servicing, repair or correction.

   In no event unless required by applicable law or agreed to in writing
   will any copyright holder, or any other party who may modify and/or
   redistribute the program as permitted above, be liable to you for
   damages, including any general, special, incidental or consequential
   damages arising out of the use or inability to use the program
   (including but not limited to loss of data or data being rendered
   inaccurate or losses sustained by you or third parties or a failure of
   the program to operate with any other programs), even if such holder or
   other party has been advised of the possibility of such damages.
 ****************************************************************************
   Module      : sys_drv_customer.h
   Date        : 2003-12-16
   Description : This file contains the includes and the defines
                 specific to the OS
   Remarks     :
      The driver makes use of the following system resources:
      - malloc and free of memory
      - interrupt functions
      - timers
      - mutex semaphores
      - semaphores or events for internal communication
      - queues/semaphores for external waiting tasks in case of select
        or polling call
      - wait (with scheduling) and delay (without scheduling) functions
      - system time function to retrieve the system time
      - debug print out
      Some features are implemented in a project specific C module, as for
      example timers and interrupt related functions.
*******************************************************************************/

/* ============================= */
/* Global Includes               */
/* ============================= */

#include <stdio.h>
#include <stdlib.h>
/* include operating system specific files */


/******************************************************************************
                                Global macros
******************************************************************************/

/** Inline functions */
#define INLINE
/** For local functions */
#define LOCAL  static
/* return values */
#define SUCCESS 0
#define ERR  -1

#ifdef DEBUG
#define ASSERT(expr) \
	if(!(expr)) { \
		printk ( "\n\r" __FILE__ ":%d: Assertion " #expr " failed!\n\r",__LINE__); \
	}
#else
#define ASSERT(cond)
#endif

/******************************************************************************
                                Hardware related
******************************************************************************/
/** Gets the processors base address */
#define OS_UC_BASE

/** Define byte order */
#ifndef __LITTLE_ENDIAN
#error define little endian
/* #define __LITTLE_ENDIAN ??? */
#endif
#ifndef __BIG_ENDIAN
#error define big endian
/* #define __BIG_ENDIAN    ???  */
#endif

#ifndef __BYTE_ORDER
#error Define byte order
/* either
#define __BYTE_ORDER __LITTLE_ENDIAN
or
#define __BYTE_ORDER __BIG_ENDIAN */
#endif

/** Returns the processor time in milliseconds */
extern DWORD OS_GetTime();
/** enables interrupt. If enabled 0 is returned */
extern int Enable_Irq(int irq);
/** disables interrupt */
extern void Disable_Irq(int irq);
/* Install interrupt service routine   */
extern void OS_InstallIRQHandler (void* pDevice, int nIrq);
/* Uninstall interrupt service routine   */
extern void OS_RemoveIRQHandler(void* pDevice);

/******************************************************************************
                                 Time related
******************************************************************************/
/** Delay excution with task schedule */
#define OS_Wait(time)                     /* schedule_timeout */
/** Short active delay in milli seconds without schedule */
#define OS_DelayMs(time)                  /* mdelay(time) */
/** Short active delay in micro seconds without schedule */
#define OS_DelayUs(time)                  /* udelay(time) */

/******************************************************************************
                                 Basics
******************************************************************************/
#define OS_MALLOC(size)                   malloc (size)
#define OS_FREE  (ptr)                    free ((void*)ptr)

/******************************************************************************
                     Events for internal driver communication
******************************************************************************/

/** Type for event
VxWorks:
   SEM_ID
Linux:
   wait_queue_head_t */
#define SYS_EVENT

/** Initialize an event. No return
VxWorks:
   (queue) = semBCreate(SEM_Q_FIFO, SEM_EMPTY)
Linux:
   init_waitqueue_head (&(event))
*/
#define OS_InitEvent(queue)

/** Signal an event. Return TRUE if success, otherwise FALSE
VxWorks:
   (semGive ((queue))==OK ? TRUE : FALSE)
Linux:
   TRUE;wake_up_interruptible(&(event))
 */
#define OS_WakeUpEvent(queue)

/** Reset event to initial state
VxWorks:
   semFlush (*(queue)); \
   semTake (*(queue), NO_WAIT);
Linux:
   init_waitqueue_head (&(event))
 */
#define OS_ClearEvent(queue)

/** Waits for a specified event to occur or timeout.
   \param wq         - wait queue
   \param condition  - condition to occur
   \param timeout    - time out in ms
   \return  1 if event occured, otherwise 0
VxWorks:
   (semTake((sem), ((timeout) == (WAIT_FOREVER) ? (WAIT_FOREVER) :
   ((timeout)==(NO_WAIT)?(NO_WAIT):(((timeout) * sysClkRateGet() / 1000) + 1))))==(OK)?(TRUE):(FALSE))
Linux:
   (interruptible_sleep_on_timeout(&(event),(HZ * timeout / 1000)) != 0 ? TRUE : FALSE)
Windows:
   (WaitForSingleObject(wq, timeout) == WAIT_OBJECT_0)
*/
#define OS_WaitEvent(sem, timeout)

/******************************************************************************
               Mutural exclusion for internal driver communication
******************************************************************************/
/** Type of mutex
VxWorks:
   SEM_ID
Linux:
   struct semaphore
*/
#define OS_MUTEX                          HANDLE

/** initialze mutex
VxWorks:
   sem = CreateMutex(NULL, FALSE, NULL)
Linux:
   (sem)=kmalloc(sizeof(struct semaphore), GFP_KERNEL); sema_init((sem), 1);
*/
#define OS_MutexInit(sem)

/** lock/take the mutex
VxWorks:
   semTake(sem, WAIT_FOREVER)
Linux:
   down(sem)*/
#define OS_MutexLock(sem)
/** unlock/give the mutex
VxWorks:
   semGive(sem)
Linux:
   up(sem)*/
#define OS_MutexUnlock(sem)

/** Delete a mutex element
VxWorks:
   semDelete(sem)
Linux:
   free(sem) */
#define Sem_MutexDelete(sem)


/******************************************************************************
                  external communication for select/polling
******************************************************************************/
#define SYS_POLL;

/* typedefs */
#define SEL_WAKEUP_LIST                   HANDLE
/* Initialize a queue. */
#define Init_WakeupQueue(queue)           queue = CreateEvent(NULL, TRUE, TRUE, NULL)
#define Sleep_On_Queue(queue, timeout)    WaitForSingleObject (queue, timeout)
#define Wake_Up_Queue(queue)              ResetEvent (queue)

#define SYS_EVENT                         HANDLE
#define SYS_POLL                          HANDLE

#define  SYS_NOSELECT   0x00000000
#define  SYS_DATAIN     0x00000001
#define  SYS_EXCEPT     0x00000002
#define  SYS_DATAOUT    0x00000004

/** copy from user to kernel space
   \note Different return codes! */
#define copy_from_user(to, from, n)       memcpy(to, from, n)
/** copy from kernel to user space
   \note Different return codes! */
#define copy_to_user(to, from, n)         memcpy(to, from, n)

#define getTime Get_Tick
#define TIMESTAMP int
#define TIMER_ELEMENT   int



/* List macros, using linux/list.h */
typedef struct {
	struct LIST *next, *prev;
}LIST;

/****************************************************************************
Description:
   Initializes the list
Arguments:
  *head: list head to add it after, type *LIST
****************************************************************************/
#define LIST_INIT(head) do { \
   	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
   } while (0)

/****************************************************************************
Description:
   Insert a new entry after the specified head
Arguments:
  *element: new entry to be added, type *LIST
  *head: list head to add it after, type *LIST
Remarks:
 This is good for implementing stacks.
Example:
   typedef {
      // ...
      // list handling struct
      LIST listHead;
      // ...
   }DEVICE;

   typedef {
      // list handling struct
      LIST list;
      WORD pData[250];
      // ...
   }PACKET;
   // ...
   List_Add ((&pData->list), &(pDev->listHead));
****************************************************************************/
#define List_Add(element, head)     \
	head->next->prev = element;      \
	element->next = head->next;      \
  	element->prev = head;            \
	head->next = element



/****************************************************************************
Description:
   Insert a new entry before the specified head
Arguments:
  *element: new entry to be added, type *LIST
  *head: list head to add it after, type *LIST
Remarks:
   This is useful for implementing queues.
****************************************************************************/
#define List_Insert(element, head)  list_add_tail(element, head)

/****************************************************************************
Description:
   Delete a list entry
Arguments:
  *element: new entry to be deleted, type *LIST
****************************************************************************/
#define List_Delete(element)              \
	element->next->prev = element->prev;   \
	element->prev->next = element->next

/****************************************************************************
Description:
   tests whether a list is empty
Arguments:
  *head: the list head to test, , type *LIST
Code:
	return head->next == head;
****************************************************************************/
#define List_Empty(head)  (head->next == head)


/****************************************************************************
Description:
   get the struct for this entry
Arguments:
   ptr - the &list_head pointer.
   type - the type of the struct this is embedded in.
   member - the name of the list_struct within the struct.
Example:
   PACKET* pEl;
   pEl = list_entry (pDev->listHead, PACKET, list);
   pEl->pData[0] = 1;
****************************************************************************/
#define List_Entry(ptr, type, member) \
	   ((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))


/* ============================= */
/* Global Structures             */
/* ============================= */

/* ============================= */
/* Global function declaration   */
/* ============================= */

/* Device Init. Called form Board_DevOpen when the device has been opened for
   the first time. The device structure will be cached */
extern int OS_InitDevice (int nDevn);
/* Device is stopped. Called on device close */
extern void OS_StopDevice (void* pDevice);


#endif /* _SYS_DRV_CUSTOMER_H */

