#ifndef _SYS_DRV_WINDOWS_H
#define _SYS_DRV_WINDOWS_H
/****************************************************************************
       Copyright (c) 2001, Infineon Technologies.  All rights reserved.

                               No Warranty
   Because the program is licensed free of charge, there is no warranty for
   the program, to the extent permitted by applicable law.  Except when
   otherwise stated in writing the copyright holders and/or other parties
   provide the program "as is" without warranty of any kind, either
   expressed or implied, including, but not limited to, the implied
   warranties of merchantability and fitness for a particular purpose. The
   entire risk as to the quality and performance of the program is with
   you.  should the program prove defective, you assume the cost of all
   necessary servicing, repair or correction.

   In no event unless required by applicable law or agreed to in writing
   will any copyright holder, or any other party who may modify and/or
   redistribute the program as permitted above, be liable to you for
   damages, including any general, special, incidental or consequential
   damages arising out of the use or inability to use the program
   (including but not limited to loss of data or data being rendered
   inaccurate or losses sustained by you or third parties or a failure of
   the program to operate with any other programs), even if such holder or
   other party has been advised of the possibility of such damages.
 ****************************************************************************
   Module      : drv_VINETICLINUX.c
   Date        : 21. 11. 2001
   Description : This file contains the includes and the defines
                 specific to the linux OS
   Remarks     : Please use the compiler switches here if you have
                 more than one OS.
*******************************************************************************/

/* ============================= */
/* Global Includes               */
/* ============================= */

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <windows.h>
#include <winbase.h>
#include <WinSock.h>


#define GetImmr() 0xfff00000
/* Debug defines */

#define DBG_LEVEL_OFF 		4
#define DBG_LEVEL_HIGH		3  /* Trace from Driver entry */
#define DBG_LEVEL_NORMAL	2  /* Trace without chip access */
#define DBG_LEVEL_LOW 		1  /* All Traces */

/** Do not inline functions. */
#define INLINE
/** Define functions only local. */
#define LOCAL  static


/* functions macros */
#define Wait(time)                        Sleep (time)
#define Wait_Milli_Sec(time)              Sleep (time)
#define Wait_Micro_Sec(time)              Sleep (time / 1000)
#define Alloc_Mem(size)                   malloc (size)
#define Free_Mem(ptr)                     free ((void*)ptr)

/* Events for internal driver communication */
/** Initialize an event. */
#define Init_Event(queue)                 queue = CreateEvent(NULL, TRUE, TRUE, NULL)
/** Signal an event. */
#define Wake_Up_Event(queue)              SetEvent(queue)
/** Reset event to initial state. */
#define Clear_Event(queue)                ResetEvent(queue)

/** Waits for a specified event to occur or timeout.
   \ingroup SYS_DRV_VXWORKS_EVENT

   \param wq         - wait queue
   \param condition  - condition to occur
   \param timeout    - time out in ms
   \return  1 if event occured and condition is true, otherwise 0
*/
#define WaitEvent(wq, condition, timeout) \
      (WaitForSingleObject(wq, timeout) == WAIT_OBJECT_0)

/* mutural exclusion */
#define SEM_ID                            HANDLE
#define Sem_MutexInit(sem)                sem = CreateMutex(NULL, FALSE, NULL)
#define Sem_Lock(sem)                     WaitForSingleObject (sem,INFINITE)
#define Sem_Unlock(sem)                   ReleaseMutex (sem)

/* external communication for select/polling */
/* typedefs */
#define SEL_WAKEUP_LIST                   HANDLE
/* Initialize a queue. */
#define Init_WakeupQueue(queue)           queue = CreateEvent(NULL, TRUE, TRUE, NULL)
#define Sleep_On_Queue(queue, timeout)    WaitForSingleObject (queue, timeout)
#define Wake_Up_Queue(queue)              ResetEvent (queue)
#define CPM_Dpalloc(size)                 malloc (size)
#define Host_Alloc(size)                  malloc (size)

#define SYS_EVENT                         HANDLE
#define SYS_POLL                          HANDLE

#define  SYS_NOSELECT   0x00000000
#define  SYS_DATAIN     0x00000001
#define  SYS_EXCEPT     0x00000002
#define  SYS_DATAOUT    0x00000004

/** copy from user to kernel space
   \note Different return codes! */
#define copy_from_user(to, from, n)       memcpy(to, from, n)
/** copy from kernel to user space
   \note Different return codes! */
#define copy_to_user(to, from, n)         memcpy(to, from, n)

#ifdef DEBUG
#define ASSERT(expr) \
	if(!(expr)) { \
		printk ( "\n\r" __FILE__ ":%d: Assertion " #expr " failed!\n\r",__LINE__); \
	}
#else
#define ASSERT(cond)
#endif

#define getTime Get_Tick
#define TIMESTAMP int
#define TIMER_ELEMENT   int

/* return values */
#define SUCCESS 0
#define ERR  -1


/* List macros, using linux/list.h */
typedef struct {
	struct LIST *next, *prev;
}LIST;

/****************************************************************************
Description:
   Initializes the list
Arguments:
  *head: list head to add it after, type *LIST
****************************************************************************/
#define LIST_INIT(head) do { \
   	(ptr)->next = (ptr); (ptr)->prev = (ptr); \
   } while (0)

/****************************************************************************
Description:
   Insert a new entry after the specified head
Arguments:
  *element: new entry to be added, type *LIST
  *head: list head to add it after, type *LIST
Remarks:
 This is good for implementing stacks.
Example:
   typedef {
      // ...
      // list handling struct
      LIST listHead;
      // ...
   }DEVICE;

   typedef {
      // list handling struct
      LIST list;
      WORD pData[250];
      // ...
   }PACKET;
   // ...
   List_Add ((&pData->list), &(pDev->listHead));
****************************************************************************/
#define List_Add(element, head)     \
	head->next->prev = element;      \
	element->next = head->next;      \
  	element->prev = head;            \
	head->next = element



/****************************************************************************
Description:
   Insert a new entry before the specified head
Arguments:
  *element: new entry to be added, type *LIST
  *head: list head to add it after, type *LIST
Remarks:
   This is useful for implementing queues.
****************************************************************************/
#define List_Insert(element, head)  list_add_tail(element, head)

/****************************************************************************
Description:
   Delete a list entry
Arguments:
  *element: new entry to be deleted, type *LIST
****************************************************************************/
#define List_Delete(element)              \
	element->next->prev = element->prev;   \
	element->prev->next = element->next

/****************************************************************************
Description:
   tests whether a list is empty
Arguments:
  *head: the list head to test, , type *LIST
Code:
	return head->next == head;
****************************************************************************/
#define List_Empty(head)  (head->next == head)


/****************************************************************************
Description:
   get the struct for this entry
Arguments:
   ptr - the &list_head pointer.
   type - the type of the struct this is embedded in.
   member - the name of the list_struct within the struct.
Example:
   PACKET* pEl;
   pEl = list_entry (pDev->listHead, PACKET, list);
   pEl->pData[0] = 1;
****************************************************************************/
#define List_Entry(ptr, type, member) \
	   ((type *)((char *)(ptr)-(unsigned long)(&((type *)0)->member)))


/* ============================= */
/* Global Structures             */
/* ============================= */

/* ============================= */
/* Global function declaration   */
/* ============================= */

extern DWORD Get_Tick();
extern int Enable_Irq(int irq);
extern void Disable_Irq(int irq);

extern void (*print) (LPCSTR pszFormat, ...);

/*
extern void cpm_install_handler(int vec, void (*handler)(void *), void *dev_id);
extern void cpm_free_handler(int vec);
*/
/* unsigned long GetImmr(); */
/* Device Init. Called form Board_DevOpen when the device has been opened for
   the first time. The device structure will be cached */
extern int OS_InitDevice (int nDevn);
/* Device is stopped. Called on device close */
extern void OS_StopDevice (void* pDevice);
/* Install VINETIC ISR   */
extern void OS_Install_IRQ_Handler(void* pDevice, int nIrq);
/* Uninstall VINETIC ISR */
extern void OS_UnInstall_IRQ_Handler(void* pDevice);


#endif //_SYS_DRV_WINDOWS_H

