/************************************************************************
 *
 * Copyright (c) 2004
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
 
/******************************************************************************
   File        : $RCSfile: drv_inca_oak_linux.c,v $
   Revision    : $Revision: 1.3 $
   Date        : $Date: 2004/06/01 11:43:39 $
   Description : Linux driver for the INCA-IP OAK DSP 
   
******************************************************************************/
#ifdef LINUX
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/version.h>
#include <asm/irq.h>
#include <asm/incaip/irq.h>

/* traces */
#ifndef ENABLE_TRACE
/*#define ENABLE_TRACE*/
#endif
#ifndef ENABLE_LOG
/*#define ENABLE_LOG*/
#endif

#include "drv_api.h"
#include "sys_drv_linux.h"
#include "RegIncaIP.h"
#include "drv_inca_oak_api.h"
#include "drv_inca_oak_interface.h"
#include "INCAOAK_DeviceDriver.h"
#include "INCAOAK_DeviceDriver_Ext.h"
#include "inca_oak_version.h"

/* external function declaration */
extern void oak_downstream_voice(int irq, void *pDev, struct pt_regs *regs);
extern void oak_upstream_cmd_isr(int irq, void *pDev, struct pt_regs *regs);
extern void oak_upstream_voice_isr(int irq, void *pDev, struct pt_regs *regs);
extern void oak_downstream_voice_isr(int irq, void *pDev, struct pt_regs *regs);

/* local function declaration */
int inca_oak_open(struct inode *inode, struct file *filp);
int inca_oak_close(struct inode *inode, struct file *filp);
static ssize_t inca_oak_write(struct file *file_p, const char *buf, size_t count, loff_t * ppos);
static ssize_t inca_oak_read(struct file *file_p, char *buf, size_t length, loff_t * ppos);
static unsigned int inca_oak_poll(struct file *file_p, poll_table *wait);
int inca_oak_ioctl(struct inode *inode, struct file *file_p, unsigned int nCmd, unsigned long arg);

/* Declarations for debug interface */
CREATE_TRACE_GROUP(INCA_OAK_DRV)
CREATE_LOG_GROUP(INCA_OAK_DRV)

/* what string support, driver version string */
const char INCA_OAK_WHATVERSION[] = DRV_INCA_OAK_WHAT_STR;

static BYTE inca_oak_major_number = 0;
MODULE_PARM(inca_oak_major_number, "b");
MODULE_PARM_DESC(inca_oak_major_number, "to override automatic major number");

/* the driver callbacks */
static struct file_operations inca_oak_fops =
{
   owner:   THIS_MODULE,
   read:    inca_oak_read,
   write:   inca_oak_write,
   poll:    inca_oak_poll,
   ioctl:   inca_oak_ioctl,
   open:    inca_oak_open,
   release: inca_oak_close
};

/* device structure for the OAK DSP */
OAK_DEV inca_oak_dev_struct = {};

#if CONFIG_PROC_FS 
static struct proc_dir_entry *inca_oak_proc_dir;
#endif

/****************************************************************************
Description:
  Open the device from user mode (e.g. application.
Arguments:
  inode  - pointer to the inode
  file_p - pointer to the file descriptor
Return Value:
 0       - if no error
 else    - error code
Remarks:
 None.
****************************************************************************/
int inca_oak_open(struct inode *inode, struct file *file_p)
{
   OAK_DEV           *pDev = &inca_oak_dev_struct;
   OAK_MB_DEV_STR    *pMBDev;
   BOOL              bcommand = FALSE;
   int               from_kernel = 0;
   DEVTYPE           num;

   /* Check whether called from user or kernel mode */
   if ( (inode == (struct inode *)1) || (inode == (struct inode *)2) 
     || (inode == (struct inode *)3)) {
      from_kernel = 1;
      num = (int)inode;
   } else {
      num = (DEVTYPE)MINOR(inode->i_rdev);  /* the real device */
   }
         
   /* check the device number */
   switch(num) {
   case command:
      /* command type */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,\
        ("INCA_OAK_DRV: open command mailbox\n"));
      pMBDev = &pDev->CommandMB;
      bcommand = TRUE;
      break;
   case voice0:
      /* voice channel 0 */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,\
       ("INCA_OAK_DRV: open voice0 mailbox\n"));
      pMBDev = &pDev->VoiceMB[0];
      break;
   case voice1:
      /* voice channel 1 */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,\
       ("INCA_OAK_DRV: open voice1 mailbox\n"));
      pMBDev = &pDev->VoiceMB[1];
      break;
   default:
      TRACE(INCA_OAK_DRV, DBG_LEVEL_HIGH,\
      ("INCA_OAK_DRV ERROR: max. device number exceed!\n"));
      return -ENOENT;
   }
   
   if ((OK) == oak_common_open(pDev, pMBDev, bcommand)) {

      if (!from_kernel) {
         /* installation was successfull */
         /* and use filp->private_data to point to the device data */
         file_p->private_data = pMBDev;
      
#ifdef MODULE
         /* increment module use counter */
         MOD_INC_USE_COUNT;
#endif
      }
	
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, \
         ("INCA_OAK_DRV: Device open successful\n"));
      return 0;
   } else {
      /* installation failed */
       TRACE(INCA_OAK_DRV, DBG_LEVEL_HIGH,\
         ("INCA_OAK_DRV ERROR: Device is already open!\n"));
       return -EMFILE;
   }
}


/****************************************************************************
Description:
   Release the device.
Arguments:
   inode -
   filp -
Return Value:
  0 - on success
Remarks:
  None.
****************************************************************************/
int inca_oak_close(struct inode *inode, struct file *filp)
{
   OAK_MB_DEV_STR *pMBDev;
   int            from_kernel = 0;
    
   /* Check whether called from user or kernel mode */
   if ((inode == (struct inode *)1) || (inode == (struct inode *)2) || (inode == (struct inode *)3)) {
      from_kernel = 1;
      switch((int)inode) {
      case command:
         pMBDev = &inca_oak_dev_struct.CommandMB;
         break;
      case voice0:
         pMBDev = &inca_oak_dev_struct.VoiceMB[0];
         break;
      case voice1:
         pMBDev = &inca_oak_dev_struct.VoiceMB[1];
         break;
      default:
         return (-EINVAL);
      }
   } else {
      pMBDev = filp->private_data;
   }

   TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: close\n"));  
   
   if (NULL != pMBDev)
   {
      /* device is still available */
      oak_common_close(pMBDev);

#ifdef MODULE
      if (!from_kernel) {
         /* increment module use counter */
         MOD_DEC_USE_COUNT;
      }
#endif
      
      return 0;      
   }
   else
   {
      /* something went totally wrong */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_HIGH,\
         ("INCA_OAK_DRV ERROR: pMBDev pointer is NULL!\n"));
      return -ENODEV;
   }
}


/****************************************************************************
Description:
   Writing data to the dsp device. This function is a dummy function. It
   returns the given size. 
Arguments:
   inode -
   filp -
   buf - source buffer
   count - data length
Return Value:
   always returns the given size.
Remarks:
   None.
****************************************************************************/
static ssize_t inca_oak_write(struct file *file_p, const char *buf, 
   size_t count, loff_t * ppos)
{
   TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: write\n"));

   return count;
}


/****************************************************************************
Description:
   Reads data from the dsp device. This function is a dummy function. It
   does not read any data from the device. It always returns zero.
Arguments:
   inode  -
   filp   -
   buf    - destination buffer
   length - max size of data to read
Return Value:
   always zero.
Remarks:
   None.
****************************************************************************/
static ssize_t inca_oak_read(struct file * file_p, char *buf, 
   size_t count, loff_t * ppos)
{
   TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: read\n"));

   return 0;
}


/****************************************************************************
Description:
   The select function of the driver. A user space program may sleep until
   the driver wakes it up.
Arguments:
   file_p -
   wait -
Return Value:
   POLLIN - data available
   0 - no data
   POLLERR - device pointer is zero
Remarks:
   None.
****************************************************************************/
static unsigned int inca_oak_poll(struct file *file_p, poll_table *wait)
{
   OAK_MB_DEV_STR    *pMBDev = file_p->private_data;
   unsigned int      mask;
   
   TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: oak_poll\n"));  

   /* add to poll queue */
   poll_wait(file_p, &(pMBDev->oak_wklist), wait);

   mask = 0;

   /* upstream queue */
   if(pMBDev->upstrm_fifo.pWrite != pMBDev->upstrm_fifo.pRead)
   {
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,\
         ("INCA_OAK_DRV: oak_poll: upstream queue filled\n"));
      
      /* queue is empty */
      mask = POLLIN | POLLRDNORM;
   }
   
   /* downstream queue */
   if(oak_fifo_write_space(&pMBDev->dwstrm_fifo) != 0)
   {
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,\
         ("INCA_OAK_DRV: oak_poll: downstream queue space\n"));  
      
      /* queue is not full */
      mask |= POLLOUT | POLLWRNORM;
   }
   return mask;
}


/****************************************************************************
Description:
   Configuration / Control for the device.
Arguments:
   inode    -
   file_p   - 
   nCmd     - function id's
   arg      - additional optional argument
Return Value:
   negative value - ioctl failed
Remarks:
   None.
****************************************************************************/
int inca_oak_ioctl(struct inode *inode, struct file *file_p, unsigned int nCmd, unsigned long arg)
{
   void              *pTemp;
   int               retvalue;
   unsigned long     flags;
   OAK_GLOBAL_REG    *pBaseGlobal;
   UINT32            reg;
   OAK_READWRITE     rw_struct;
   OAK_FWDWNLD       dwnld_struct;
   OAK_MB_DEV_STR    *pMBDev;
   int               from_kernel = 0;

   if ((inode == (struct inode *)1) || (inode == (struct inode *)2) || (inode == (struct inode *)3)) {
      from_kernel = 1;
      switch((int)inode) {
      case command:
         pMBDev = &inca_oak_dev_struct.CommandMB;
         break;
      case voice0:
         pMBDev = &inca_oak_dev_struct.VoiceMB[0];
         break;
      case voice1:
         pMBDev = &inca_oak_dev_struct.VoiceMB[1];
         break;
      default:
         return (-EINVAL);
      }
   } else {
      pMBDev = file_p->private_data;
   }

   switch(nCmd) {
   case FIO_INCAIP_OAK_EVENT_REG:
      pMBDev->event_mask = arg;
      retvalue = OK;
      return( retvalue );
   case FIO_INCAIP_OAK_EVENT_UNREG:
      pMBDev->event_mask = 0xFFFFFFFF;
      retvalue = OK;
      return( retvalue );
   case FIO_INCAIP_OAK_MB_READ:
      /* Read the data from mailbox stored in local FIFO */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: oak_ioctl MB read\n"));  
      
      if (from_kernel) {
         retvalue = oak_read_fifo(pMBDev,(OAK_READWRITE*)arg, 0);
      } else {
         UINT32 *pUserBuf;
         /* Copy OAK_READWRITE from usermode */
         copy_from_user((char*)&rw_struct, (char*)arg, sizeof(OAK_READWRITE));
         /* Remember usermode buffer */
         pUserBuf = rw_struct.pRead;
         /* Allocate kernelmode buffer for fetching data */
         rw_struct.pRead = kmalloc(rw_struct.nRead, GFP_KERNEL);
         if (rw_struct.pRead == NULL) {
            return -ENOMEM; 
         }
         retvalue = oak_read_fifo(pMBDev,&rw_struct, 0);
         /* Copy data to usermode buffer... */
         copy_to_user((char*)pUserBuf, (char*)rw_struct.pRead, rw_struct.nRead);
         kfree(rw_struct.pRead); 
         /* ... and finally restore the buffer pointer and copy OAK_READWRITE back! */
         rw_struct.pRead = pUserBuf;
         copy_to_user((char*)arg, (char*)&rw_struct, sizeof(OAK_READWRITE));
      }
      return retvalue;
   
   case FIO_INCAIP_OAK_MB_WRITE:  
      /* Write data to send to the mailbox into the local FIFO */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: oak_ioctl MB write\n"));  
      
      if (from_kernel) {
         if(pMBDev->devID == command) {
            retvalue = oak_write_cmd_mailbox(pMBDev, (OAK_READWRITE*)arg);
         } else {
            retvalue = oak_write_fifo(pMBDev, (OAK_READWRITE*)arg);
         }   
      } else {
         UINT32 *pUserBuf;
         copy_from_user((char*)&rw_struct, (char*)arg, sizeof(OAK_READWRITE));

         /* Remember usermode buffer */
         pUserBuf = rw_struct.pWrite;
         /* Allocate kernelmode buffer for fetching data */
         /* TODO: data needs to be copied from user directly to FIFO! */
         rw_struct.pWrite = kmalloc(rw_struct.nWrite, GFP_KERNEL);
         if (rw_struct.pWrite == NULL) {
            return -ENOMEM; 
         }

         /* Now copy data to kernelmode buffer */
         copy_from_user((char*)rw_struct.pWrite, (char*)pUserBuf, rw_struct.nWrite);
         if(pMBDev->devID == command) {
            retvalue = oak_write_cmd_mailbox(pMBDev, &rw_struct);
         } else {
            retvalue = oak_write_fifo(pMBDev, &rw_struct);
         }   
         /* ... and finally restore the buffer pointer and copy OAK_READWRITE back! */
         kfree(rw_struct.pWrite);
         rw_struct.pWrite = pUserBuf;
         copy_to_user((char*)arg, (char*)&rw_struct, sizeof(OAK_READWRITE));
      }
      return retvalue;
   
   case FIO_INCAIP_OAK_MB_WR_ACCESS:
      /* Write commands to the voice mailbox directly */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl MB access\n"));  

      pBaseGlobal = pMBDev->pBaseGlobal;
      reg         = pMBDev->MBC_Mask01;

      if (!from_kernel) {
         copy_from_user((char*)&rw_struct, (char*)arg, sizeof(OAK_READWRITE));
      }
      
      save_and_cli(flags); {
         /* disable the interrupts of the mailbox during write access */
         pBaseGlobal->MBC_MSK01 |= reg;
         pBaseGlobal->MBC_MSK10 |= reg;
      } restore_flags(flags);

      if (from_kernel) {
         if(pMBDev->devID == command) {
            retvalue = oak_write_cmd_mailbox(pMBDev, (OAK_READWRITE*)arg);
         } else {
            retvalue = oak_write_voice_mailbox(pMBDev, (OAK_READWRITE*)arg);
         }   
      } else {
         if(pMBDev->devID == command) {
            retvalue = oak_write_cmd_mailbox(pMBDev, &rw_struct);
         } else {
            retvalue = oak_write_voice_mailbox(pMBDev, &rw_struct);
         }   
      }

      save_and_cli(flags); {
        /* enable the interrupts of the mailbox after write access */
        pBaseGlobal->MBC_MSK01 &= ~reg;
        pBaseGlobal->MBC_MSK10 &= ~reg;
      } restore_flags(flags);
      
      if (!from_kernel) {
         copy_to_user((char*)arg, (char*)&rw_struct, sizeof(OAK_READWRITE));
      }
      
      return retvalue;

   case FIO_INCAIP_OAK_DWNLOAD:
      /* Download of OAK DSP firmware file */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW,("INCA_OAK_DRV: oak_ioctl download\n"));  

      if(pMBDev->devID == command) {
         if (from_kernel) {
            retvalue = oak_download(pMBDev,(OAK_FWDWNLD*)arg);
         } else {
            UINT32 *pUserBuffer;
            copy_from_user((char*)&dwnld_struct, (char*)arg, sizeof(OAK_FWDWNLD));
            printk("INCA_OAK_DEV: Download firmware (size %li bytes)... ", dwnld_struct.Length);
            pUserBuffer = dwnld_struct.pData;
            dwnld_struct.pData = vmalloc(dwnld_struct.Length);
            retvalue = copy_from_user((char*)dwnld_struct.pData,(char*)pUserBuffer, dwnld_struct.Length);
            retvalue = oak_download(pMBDev, &dwnld_struct);
            if (retvalue != 0) {
               printk(" error (%i)!\n", retvalue);
            } else {
               printk(" ok!\n");
            }
            vfree(dwnld_struct.pData);
         }
         return retvalue;
      } else {
         return (-EINVAL);
      }

   case FIO_INCAIP_OAK_GETVERSION:
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl get version\n"));  

      if (from_kernel) {
         memcpy((char*)arg, (char*)string_oak_version, strlen(string_oak_version));
      } else {
         copy_to_user((char*)arg, (char*)string_oak_version, strlen(string_oak_version));
      }
      return (OK);

   case FIO_INCAIP_RST_QUEUE :
      /* Reset the QUEUE Upstream/downstream Note : The interrupts and the
      tasks should be not be running if the queues have to be disabled */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl get version\n"));  

      if(arg == UPSTREAM) {
         pTemp = &(pMBDev->upstrm_fifo);
      } else {
         pTemp = &(pMBDev->dwstrm_fifo);
      }
      oak_fifo_clear((OAK_FIFO_MB *)pTemp);
      return( OK );

   case  FIO_INCAIP_OAK_RESET:
      /* Reset of the DSP */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl dsp reset\n"));  

      return oak_reset_dsp(pMBDev->pBaseGlobal);

   case  FIO_INCAIP_OAK_RESTART:
      /* Restart of the DSP */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl dsp restart\n"));  

      return oak_restart_dsp(pMBDev->pBaseGlobal);

#ifdef OAK_FIFO_BLOCKING_WRITE
   case  FIO_INCAIP_TXFIFO_SET:
      /* Set the mailbox TX FIFO to blocking */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl tx fifo set\n"));  

      if(pMBDev->devID == command) {
         return (-EINVAL); /* not supported for this command MB */
      } else {
         if (arg > 0) {
            pMBDev->bBlockWriteMB = TRUE;
         } else {
            pMBDev->bBlockWriteMB = FALSE;
            Sem_Unlock(pMBDev->sem_write_fifo);
         }
         return (OK);
      }

   case  FIO_INCAIP_TXFIFO_GET:
      /* Get the mailbox TX FIFO to blocking */
      TRACE(INCA_OAK_DRV, DBG_LEVEL_LOW, ("INCA_OAK_DRV: oak_ioctl tx fifo get\n"));  

      if(pMBDev->devID == command) {
         return (-EINVAL); /* not supported for this command MB */
      } else {
         if (!from_kernel) {
            copy_to_user((char*)arg, (char*)&pMBDev->bBlockWriteMB, sizeof(BOOL));
         }
         return (OK);
      }
#endif /* OAK_FIFO_BLOCKING_WRITE */

   default:
      break ;
   }
   return (-EINVAL);
}

/****************************************************************************
Description:
   Allows the upper layer to register a callback function either for
   downstream (tranmsit mailbox space available) or for upstream (read data 
   available)
Arguments:
   type     - DSP device entity ( 1 - command, 2 - voice0, 3 - voice1)
   dir      - Direction (1 - upstream, 2 - downstream)
   callback - Callback function to register
Return Value:
   OK    - Callback registered successfully
   ENXIO - Wrong DSP device entity (only 1-3 supported)
   EBUSY - Callback already registered
Remarks:
   None.
****************************************************************************/
int inca_oak_data_callback_register(DEVTYPE type, UINT dir, void (*callback)(DEVTYPE type))
{
   OAK_MB_DEV_STR    *pMBDev;
   /* Get corresponding mailbox device structure */
   switch(type) {
   case command:
      pMBDev = &inca_oak_dev_struct.CommandMB;
      break;
   case voice0:
      pMBDev = &inca_oak_dev_struct.VoiceMB[0];
      break;
   case voice1:
      pMBDev = &inca_oak_dev_struct.VoiceMB[1];
      break;
   default:
      return (-ENXIO);
   }
   if (callback == NULL) {
      return (-EINVAL);
   }
   /* Enter the desired callback function */
   switch(dir) {
   case 1:
      if (pMBDev->up_callback != NULL) {
         return (-EBUSY);
      } else {
         pMBDev->up_callback = callback;
      }
      break;
   case 2:
       if (pMBDev->down_callback != NULL) {
         return (-EBUSY);
      } else {
         pMBDev->down_callback = callback;
      }
      break;
   default:
      break;
   }
   return(OK);
}

/****************************************************************************
Description:
   Allows the upper layer to un-register the callback function previously
   registered using inca_oak_data_callback_register
Arguments:
   type     - DSP device entity ( 1 - command, 2 - voice0, 3 - voice1)
   dir      - Direction (1 - upstream, 2 - downstream)
Return Value:
   OK    - Callback registered successfully
   ENXIO - Wrong DSP device entity (only 1-3 supported)
   EINVAL- Callback not registered
Remarks:
   None.
****************************************************************************/
int inca_oak_data_callback_unregister(DEVTYPE type, UINT dir)
{
   OAK_MB_DEV_STR    *pMBDev;
   /* Get corresponding mailbox device structure */
   switch(type) {
   case command:
      pMBDev = &inca_oak_dev_struct.CommandMB;
      break;
   case voice0:
      pMBDev = &inca_oak_dev_struct.VoiceMB[0];
      break;
   case voice1:
      pMBDev = &inca_oak_dev_struct.VoiceMB[1];
      break;
   default:
      return (-ENXIO);
   }
   /* Delete the desired callback function */
   switch(dir) {
   case 1:
      if (pMBDev->up_callback == NULL) {
         return (-EINVAL);
      } else {
         pMBDev->up_callback = NULL;
      }
      break;
   case 2:
       if (pMBDev->down_callback == NULL) {
         return (-EINVAL);
      } else {
         pMBDev->down_callback = NULL;
      }
      break;
   default:
      break;
   }
   return(OK);
}

/****************************************************************************
Description:
   Allows the upper layer to register a callback function for event
   notification. 
Arguments:
   type     - DSP device entity ( 1 - command, 2 - voice0, 3 - voice1)
   mask     - mask according to MBC_ISR content
   callback - Callback function to register
Return Value:
   OK    - Callback registered successfully
   ENXIO - Wrong DSP device entity (only 1-3 supported)
   EBUSY - Callback already registered
Remarks:
   None.
****************************************************************************/
int inca_oak_event_callback_register(DEVTYPE type, UINT mask, void (*callback)(UINT mask))
{
   OAK_MB_DEV_STR    *pMBDev;
   /* Get corresponding mailbox device structure */
   switch(type) {
   case command:
      pMBDev = &inca_oak_dev_struct.CommandMB;
      break;
   case voice0:
      pMBDev = &inca_oak_dev_struct.VoiceMB[0];
      break;
   case voice1:
      pMBDev = &inca_oak_dev_struct.VoiceMB[1];
      break;
   default:
      return (-ENXIO);
   }
   if (callback == NULL) {
      return (-EINVAL);
   }
   /* Enter the desired callback function */
   if (pMBDev->event_callback != NULL) {
      return (-EBUSY);
   } else {
	 pMBDev->event_mask = mask;
      pMBDev->event_callback = callback;
   }
   return(OK);
}

/****************************************************************************
Description:
   Allows the upper layer to un-register the callback function previously
   registered using inca_oak_event_callback_register
Arguments:
   type     - DSP device entity ( 1 - command, 2 - voice0, 3 - voice1)
   mask     - mask according to MBC_ISR content
Return Value:
   OK    - Callback registered successfully
   ENXIO - Wrong DSP device entity (only 1-3 supported)
   EINVAL- Callback not registered
Remarks:
   None.
****************************************************************************/
int inca_oak_event_callback_unregister(DEVTYPE type, UINT mask)
{
   OAK_MB_DEV_STR    *pMBDev;
   /* Get corresponding mailbox device structure */
   switch(type) {
   case command:
      pMBDev = &inca_oak_dev_struct.CommandMB;
      break;
   case voice0:
      pMBDev = &inca_oak_dev_struct.VoiceMB[0];
      break;
   case voice1:
      pMBDev = &inca_oak_dev_struct.VoiceMB[1];
      break;
   default:
      return (-ENXIO);
   }
   /* Delete the desired callback function */
    if (pMBDev->event_callback == NULL) {
      return (-EINVAL);
   } else {
      pMBDev->event_callback = NULL;
   }
   return(OK);
}

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None.
****************************************************************************/
int inca_oak_mailbox_read(DEVTYPE type, OAK_READWRITE *rw)
{
   int ret;

   switch(type) {
   case command:
      ret = oak_read_fifo(&inca_oak_dev_struct.CommandMB, rw, 0);
      break;
   case voice0:
      ret = oak_read_fifo(&inca_oak_dev_struct.VoiceMB[0], rw, 0);
      break;
   case voice1:
      ret = oak_read_fifo(&inca_oak_dev_struct.VoiceMB[1], rw, 0);
      break;
   default:
      ret = -ENXIO;
   }
   return(ret);
}

/****************************************************************************
Description:
Arguments:
Return Value:
Remarks:
   None.
****************************************************************************/
int inca_oak_mailbox_write(DEVTYPE type, OAK_READWRITE *rw)
{
   int ret;

   switch(type) {
   case command:
      ret = oak_write_cmd_mailbox(&inca_oak_dev_struct.CommandMB, rw);
      break;
   case voice0:
      ret = oak_write_voice_mailbox(&inca_oak_dev_struct.VoiceMB[0], rw);
      break;
   case voice1:
      ret = oak_write_voice_mailbox(&inca_oak_dev_struct.VoiceMB[1], rw);
      break;
   default:
      ret = -ENXIO;
   }
   return(ret);
}

#if CONFIG_PROC_FS 
/****************************************************************************
Description:
   Read the version information from the driver.
Arguments:
   buf - destination buffer
Return Value:
   length
Remarks:
   None.
****************************************************************************/
static int inca_oak_get_version_proc(char *buf)
{
   int len;

   len = sprintf(buf, "%s\n", &INCA_OAK_WHATVERSION[4]);

   len += sprintf(buf + len, "Compiled on %s, %s for Linux kernel %s\n", 
      __DATE__, __TIME__, UTS_RELEASE);

   return len;
}

/****************************************************************************
Description:
   Read the status information from the driver.
Arguments:
   buf - destination buffer
Return Value:
   length
Remarks:
   None.
****************************************************************************/
static int inca_oak_get_status_proc(char *buf)
{
   int len;

   len = sprintf(buf, "%d connections of the driver are used\n", MOD_IN_USE);
   
   /* Print internals of the command mailbox fifo */
   len += sprintf(buf+len, 
      "Fifo size CMD fifo:\t\tread %4ld bytes\n", 
      inca_oak_dev_struct.CommandMB.upstrm_fifo.FifoSize);

   len += sprintf(buf+len, 
      "Fill level CMD fifo:\t\tread %4ld bytes\n", 
      inca_oak_dev_struct.CommandMB.upstrm_fifo.FifoSize - 
      oak_fifo_write_space(&inca_oak_dev_struct.CommandMB.upstrm_fifo));

   len += sprintf(buf+len, 
      "Free level CMD fifo:\t\tread %4ld bytes\n", 
      oak_fifo_write_space(&inca_oak_dev_struct.CommandMB.upstrm_fifo));

   /* Print internals of the voice0  fifo */
   len += sprintf(buf+len, 
      "Fifo size voice0 fifo:\t\twrite %4ld bytes,\tread %4ld bytes\n", 
      inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo.FifoSize, 
      inca_oak_dev_struct.VoiceMB[0].upstrm_fifo.FifoSize);

   len += sprintf(buf+len, 
      "Fill level voice0 fifo\t\twrite %4ld bytes,\tread %4ld bytes\n", 
      inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo.FifoSize - 
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo),
      inca_oak_dev_struct.VoiceMB[0].upstrm_fifo.FifoSize - 
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[0].upstrm_fifo));

   len += sprintf(buf+len, 
      "Free level voice0 fifo:\t\twrite %4ld bytes,\tread %4ld bytes\n", 
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo),
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[0].upstrm_fifo));

   /* Print internals of the voice1  fifo */
   len += sprintf(buf+len, 
      "Fifo size voice1 fifo:\t\twrite %4ld bytes,\tread %4ld bytes\n", 
      inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo.FifoSize, 
      inca_oak_dev_struct.VoiceMB[1].upstrm_fifo.FifoSize);

   len += sprintf(buf+len, 
      "Fill level voice1 fifo\t\twrite %4ld bytes,\tread %4ld bytes\n", 
      inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo.FifoSize - 
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo),
      inca_oak_dev_struct.VoiceMB[1].upstrm_fifo.FifoSize - 
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[1].upstrm_fifo));

   len += sprintf(buf+len, 
      "Free level voice1 fifo:\t\twrite %4ld bytes,\tread %4ld bytes\n", 
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo),
      oak_fifo_write_space(&inca_oak_dev_struct.VoiceMB[1].upstrm_fifo));

#if 0
   /* get the fifo pointer */
   len += sprintf(buf+len, 
      "CMD read fifo pointer:\t\twrite 0x%08lX,\tread 0x%08lX,"
      "\n\t\t\t\tstart 0x%08lX,\tend  0x%08lX\n", 
      (UINT32)inca_oak_dev_struct.CommandMB.upstrm_fifo.pWrite,
      (UINT32)inca_oak_dev_struct.CommandMB.upstrm_fifo.pRead,
      (UINT32)inca_oak_dev_struct.CommandMB.upstrm_fifo.pStart,
      (UINT32)inca_oak_dev_struct.CommandMB.upstrm_fifo.pEnd);

   len += sprintf(buf+len, 
      "Voice0 write fifo pointer:\twrite 0x%08lX,\tread 0x%08lX,"
      "\n\t\t\t\tstart 0x%08lX,\tend  0x%08lX\n", 
      (UINT32)inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo.pWrite,
      (UINT32)inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo.pRead,
      (UINT32)inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo.pStart,
      (UINT32)inca_oak_dev_struct.VoiceMB[0].dwstrm_fifo.pEnd);

   len += sprintf(buf+len, 
      "Voice0 read fifo pointer:\twrite 0x%08lX,\tread 0x%08lX,"
      "\n\t\t\t\tstart 0x%08lX,\tend  0x%08lX\n", 
      (UINT32)inca_oak_dev_struct.VoiceMB[0].upstrm_fifo.pWrite,
      (UINT32)inca_oak_dev_struct.VoiceMB[0].upstrm_fifo.pRead,
      (UINT32)inca_oak_dev_struct.VoiceMB[0].upstrm_fifo.pStart,
      (UINT32)inca_oak_dev_struct.VoiceMB[0].upstrm_fifo.pEnd);

   len += sprintf(buf+len, 
      "Voice1 write fifo pointer:\twrite 0x%08lX,\tread 0x%08lX,"
      "\n\t\t\t\tstart 0x%08lX,\tend  0x%08lX\n", 
      (UINT32)inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo.pWrite,
      (UINT32)inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo.pRead,
      (UINT32)inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo.pStart,
      (UINT32)inca_oak_dev_struct.VoiceMB[1].dwstrm_fifo.pEnd);

   len += sprintf(buf+len, 
      "Voice1 read fifo pointer:\twrite 0x%08lX,\tread 0x%08lX,"
      "\n\t\t\t\tstart 0x%08lX,\tend  0x%08lX\n", 
      (UINT32)inca_oak_dev_struct.VoiceMB[1].upstrm_fifo.pWrite,
      (UINT32)inca_oak_dev_struct.VoiceMB[1].upstrm_fifo.pRead,
      (UINT32)inca_oak_dev_struct.VoiceMB[1].upstrm_fifo.pStart,
      (UINT32)inca_oak_dev_struct.VoiceMB[1].upstrm_fifo.pEnd);
#endif


   /* Printout the number of interrupts and fifo misses */
   len += sprintf(buf+len, 
      "Num Rx interrupts:\t\tCMD %4ld,\tVoice0 %4ld,\tVoice1 %4ld\n", 
      inca_oak_dev_struct.CommandMB.RxnumIRQs, 
      inca_oak_dev_struct.VoiceMB[0].RxnumIRQs, 
      inca_oak_dev_struct.VoiceMB[1].RxnumIRQs);

   len += sprintf(buf+len, 
      "Num Tx interrupts:\t\tCMD %4ld,\tVoice0 %4ld,\tVoice1 %4ld\n", 
      inca_oak_dev_struct.CommandMB.TxnumIRQs, 
      inca_oak_dev_struct.VoiceMB[0].TxnumIRQs, 
      inca_oak_dev_struct.VoiceMB[1].TxnumIRQs);

   len += sprintf(buf+len, 
      "Num Rx FIFO miss:\t\tCMD %4ld,\tVoice0 %4ld,\tVoice1 %4ld\n", 
      inca_oak_dev_struct.CommandMB.RxnumMiss, 
      inca_oak_dev_struct.VoiceMB[0].RxnumMiss, 
      inca_oak_dev_struct.VoiceMB[1].RxnumMiss);

   len += sprintf(buf+len, 
      "Num Tx FIFO miss:\t\tCMD %4ld,\tVoice0 %4ld,\tVoice1 %4ld\n", 
      inca_oak_dev_struct.CommandMB.TxnumMiss, 
      inca_oak_dev_struct.VoiceMB[0].TxnumMiss, 
      inca_oak_dev_struct.VoiceMB[1].TxnumMiss);

   len += sprintf(buf+len, 
      "Num Rx bytes:\t\tCMD %4ld,\tVoice0 %4ld,\tVoice1 %4ld\n", 
      inca_oak_dev_struct.CommandMB.RxnumBytes, 
      inca_oak_dev_struct.VoiceMB[0].RxnumBytes, 
      inca_oak_dev_struct.VoiceMB[1].RxnumBytes);

   len += sprintf(buf+len, 
      "Num Tx bytes:\t\tCMD %4ld,\tVoice0 %4ld,\tVoice1 %4ld\n", 
      inca_oak_dev_struct.CommandMB.TxnumBytes, 
      inca_oak_dev_struct.VoiceMB[0].TxnumBytes, 
      inca_oak_dev_struct.VoiceMB[1].TxnumBytes);

   len += sprintf(buf+len, "Events:\n");
   len += sprintf(buf+len, "\t\tCommand: 0x%08x (mask 0x%08x)\n", inca_oak_dev_struct.CommandMB.event, inca_oak_dev_struct.CommandMB.event_mask);
   len += sprintf(buf+len, "\t\tVoice 0: 0x%08x (mask 0x%08x)\n", inca_oak_dev_struct.VoiceMB[0].event, inca_oak_dev_struct.VoiceMB[0].event_mask);
   len += sprintf(buf+len, "\t\tVoice 1: 0x%08x (mask 0x%08x)\n", inca_oak_dev_struct.VoiceMB[1].event, inca_oak_dev_struct.VoiceMB[1].event_mask);

   /* printout the content of the ICU*/
   len += sprintf(buf+len, "Content of the ICU modules:\n");
   len += sprintf(buf+len, "\t\tICU1_ISR Content:\t0x%08lX\n", 
      *INCA_IP_ICU_IM1_ISR);
   len += sprintf(buf+len, "\t\tICU1_IER Content:\t0x%08lX\n", 
      *INCA_IP_ICU_IM1_IER);
   len += sprintf(buf+len, "\t\tICU1_IOSR Content:\t0x%08lX\n", 
      *INCA_IP_ICU_IM1_IOSR);

   /* printout the context of the mailbox interrupts */
   len += sprintf(buf+len, "Content of the mailbox interrupt module:\n");
   len += sprintf(buf+len, "\t\tMBC_ISR Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_ISR);
   len += sprintf(buf+len, "\t\tMBC_MSK Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_MSK);
   len += sprintf(buf+len, "\t\tMBC_MSK01 Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_MSK01);
   len += sprintf(buf+len, "\t\tMBC_MSK10 Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_MSK10);

   /* printout the space of the mailboxes */
   len += sprintf(buf+len, "Fill/Free level of the mailboxes:\n");
   len += sprintf(buf+len, "\tPackage\tMBC_FS0 Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_FS0);
   len += sprintf(buf+len, "\tCommand\tMBC_FS1 Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_FS1);
   len += sprintf(buf+len, "\tPackage\tMBC_DA2 Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_DA2);
   len += sprintf(buf+len, "\tCommand\tMBC_DA3 Content:\t0x%08lX\n", 
      *INCA_IP_MBC_MBC_DA3);

   return len;
}


/****************************************************************************
Description:
   The proc filesystem: function to read and entry.
Arguments:
   page - 
   start -
   off -
   count -
   eof -
   data -
Return Value:
   length
Remarks:
   None.
*******************************************************************************/
static INT inca_oak_read_proc(char *page, char **start, off_t off,
   int count, int *eof, void *data)
{
   int len;

   int (*fn)(char *buf);

   if ( data != NULL )
   {
      fn = data;
      len = fn(page);
   }
   else
      return 0;

   if ( len <= off+count ) *eof = 1;
   *start = page + off;
   len -= off;
   if ( len>count ) len = count;
   if ( len<0 ) len = 0;
   return len;
}
#endif


/****************************************************************************
Description:
   Initialize the module
Arguments:
   None
Return Value:
   Error code or 0 on success
Remarks:
   Called by the kernel.
****************************************************************************/
static int __init inca_oak_init_module(void)
{
   int result;

   printk(KERN_INFO "%s\n", &INCA_OAK_WHATVERSION[4]);
   printk(KERN_INFO "(c) Copyright 2003, Infineon Technologies AG\n");
   
   SetTraceLevel(INCA_OAK_DRV, DBG_LEVEL_LOW);
   SetLogLevel(INCA_OAK_DRV, DBG_LEVEL_LOW);

   result = register_chrdev(inca_oak_major_number , DRV_INCA_OAK_NAME , &inca_oak_fops);

   if ( result < 0 )
   {
      TRACE(INCA_OAK_DRV,DBG_LEVEL_HIGH,\
         ("INCA_OAK_DRV: can't get major %d\n", inca_oak_major_number));
      return result;
   }
   
   /* dynamic major                       */
   if ( inca_oak_major_number == 0 )
      inca_oak_major_number = result;

   /* reset the device before initializing the device driver */
   if ((ERROR) == oak_reset_dsp((OAK_GLOBAL_REG*)OAK_BASEADDRESS))
   {
      TRACE(INCA_OAK_DRV,DBG_LEVEL_HIGH,\
         ("INCA_OAK_DRV: Hardware reset failed!\n"));
      return -EPERM;   
   }
   
   /* init the device driver structure*/
   if (0 != oak_init_structures(&inca_oak_dev_struct))
      return -ENOMEM;

/* TODO: Free memory before returning on error */

	result = request_irq(INT_NUM_IM1_IRL20, oak_upstream_cmd_isr, 
      0, "oak_read_cmd", &inca_oak_dev_struct);
	if (result)
		return result;

	result = request_irq(INT_NUM_IM1_IRL21, oak_upstream_voice_isr, 
      0, "oak_read_voice", &inca_oak_dev_struct);
	if (result)
		return result;

	result = request_irq(INT_NUM_IM1_IRL25, oak_downstream_voice_isr, 
      0, "oak_write_voice", &inca_oak_dev_struct);
	if (result)
		return result;

#if CONFIG_PROC_FS 
   /* install the proc entry */
   TRACE(INCA_OAK_DRV,DBG_LEVEL_LOW,("INCA_OAK_DRV: using proc fs\n"));

   inca_oak_proc_dir = proc_mkdir("driver/" DRV_INCA_OAK_NAME, NULL);
   if ( inca_oak_proc_dir != NULL )
   {
      inca_oak_proc_dir->owner = THIS_MODULE;
      create_proc_read_entry("version" , S_IFREG|S_IRUGO,
         inca_oak_proc_dir, inca_oak_read_proc, (void *)inca_oak_get_version_proc);
      create_proc_read_entry("status" , S_IFREG|S_IRUGO,
         inca_oak_proc_dir, inca_oak_read_proc, (void *)inca_oak_get_status_proc);
   }
   else
   {
      TRACE(INCA_OAK_DRV,DBG_LEVEL_HIGH,\
         ("INCA_OAK_DRV: cannotrun flash_nf create proc entry\n"));
   }
#endif
   return 0;
}


/****************************************************************************
Description:
   Clean up the module if unloaded.
Arguments:
   None.
Return Value:
   None.
Remarks:
   Called by the kernel.
****************************************************************************/
static void __exit inca_oak_cleanup_module(void)
{
   OAK_GLOBAL_REG *pBaseGlobal = inca_oak_dev_struct.pBaseGlobal;

   /* disable all interrupts in the MBC */
   pBaseGlobal->MBC_MSK   = 0xFFFFFFFF;
   pBaseGlobal->MBC_MSK01 = 0xFFFFFFFF;
   pBaseGlobal->MBC_MSK10 = 0xFFFFFFFF;
   
   /* disable all MBC interrupts in the ICU */
   *((volatile UINT32*)INCA_IP_ICU_IM1_IER) &=
      ~(EXC_INCAIP_DSP_B2DAIR | EXC_INCAIP_DSP_B0EIR |
       EXC_INCAIP_DSP_SCIR0 | EXC_INCAIP_DSP_B3DAIR  | 
       EXC_INCAIP_DSP_B1EIR | EXC_INCAIP_DSP_WDTIR   | 
       EXC_INCAIP_DSP_SCIR8);
   
   unregister_chrdev(inca_oak_major_number , DRV_INCA_OAK_NAME);
   
   /* release the memory usage of the device driver structure */
   oak_release_structures(&inca_oak_dev_struct);

   /* release all interrupts at the system */
   free_irq(INT_NUM_IM1_IRL20, &inca_oak_dev_struct);
   free_irq(INT_NUM_IM1_IRL21, &inca_oak_dev_struct);
   free_irq(INT_NUM_IM1_IRL25, &inca_oak_dev_struct);
   
#if CONFIG_PROC_FS 
   remove_proc_entry("version", inca_oak_proc_dir);
   remove_proc_entry("status", inca_oak_proc_dir);
   remove_proc_entry("driver/" DRV_INCA_OAK_NAME, NULL);
#endif

   TRACE(INCA_OAK_DRV,DBG_LEVEL_NORMAL,("INCA_OAK_DRV: cleanup successful\n"));
}


/****************************************************************************/

module_init(inca_oak_init_module);

#ifndef DEBUG
EXPORT_SYMBOL(inca_oak_data_callback_register);
EXPORT_SYMBOL(inca_oak_data_callback_unregister);
EXPORT_SYMBOL(inca_oak_event_callback_register);
EXPORT_SYMBOL(inca_oak_event_callback_unregister);
EXPORT_SYMBOL(inca_oak_mailbox_read);
EXPORT_SYMBOL(inca_oak_mailbox_write);
EXPORT_SYMBOL(inca_oak_ioctl);
EXPORT_SYMBOL(inca_oak_open);
EXPORT_SYMBOL(inca_oak_close);
#endif

#ifdef MODULE

MODULE_AUTHOR("Infineon Technologies AG");
MODULE_DESCRIPTION("OAK DSP driver for INCA-IP");
MODULE_SUPPORTED_DEVICE("INCA-IP OAK DSP");
MODULE_LICENSE("GPL");

module_exit(inca_oak_cleanup_module);
#endif

#endif /* LINUX */


