/************************************************************************
 *
 * Copyright (c) 2004
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
 
/******************************************************************************
   Module      : boardAutoMdix.c
   Date        : 2/24/2003
   Description : INCA IP auto MDIX Infinon-Board specific configuration 


   Provide the callback functions used by the Auto MDIX driver

   AutoMDIX feature:
   =================
   Because of the auto MDIX configuration is board specific the INCA-IP don't have 
   knowledge about the default mode or how to set MDI or MDIX. 
   Therefore to set the board specific settings some board-specific callback functions
   are required.
   While initialization the driver gets this functions and call them to control or 
   if changes have to be done.

   For use the autoMDIX feature the used board at first must be able to support 
   this feature and the feature must be switched on. 
   See vxWorks autoMDIX project settings
   "SUPPORT_AUTOMDIX_xxx_PORT" for LAN and PC port
   "AUTOMDIX_xxx_PORT_ON"      for LAN and PC port
   
   Basically the following callback routines are required.
   - provide auto MDIX settings like 
      auto MDIX is ON or OFF, Power over LAN (PoL) is ON or OFF
      setup and config the used board specific GPIO pins
      status delay between switch between MDI and MDIX
      number of attempts to switch between MDI and MDIX
   - Set MDI or MDIX mode
   - Set RESET (DETECT) or STARTUP mode
   - Get the current mode, check the current port configuration

   These function must be provided for both PHY ports (LAN and PC)

   This file contains the basic callback functions based on the Infineon test board.
   Configurations can be changed via the vxWorks project settings and via the 
   appropriate header file.

   Within the header file the used GPIO pins can be defined.

*/
/*
modification history
--------------------


*/

#if 0
/* includes */
#include "vxWorks.h"
#include "tickLib.h"
#include "stdlib.h"
#include "sysLib.h"
#include <logLib.h>


#endif

#include <linux/sched.h>
#include <linux/spinlock.h>
#include "vxWorks_type.h"

#include "inca_board_automdix.h"

/* DEBUG MACROS */
#ifdef SW_AMDIX_DEBUG_ENABLE
   /* debug print out masks */
   #define SW_AMDIX_DEBUG_OFF          0x00000000
   #define SW_AMDIX_DEBUG_PC_PHY       0x00000001
   #define SW_AMDIX_DEBUG_LAN_PHY      0x00000002
   #define SW_AMDIX_DEBUG_FULL         0xFFFFFFFF

/*   #define SW_DEBUG_MASK         (SW_DEBUG_PHY | SW_DEBUG_INT) */
   #define SW_AMDIX_DEBUG_MASK   (SW_AMDIX_DEBUG_OFF)
/*   #define SW_AMDIX_DEBUG_MASK   (SW_AMDIX_DEBUG_LAN_PHY) */
/*   #define SW_AMDIX_DEBUG_MASK   (SW_AMDIX_DEBUG_PC_PHY) */

   #define SW_AMDIX_DEBUG(type,X)         if((UINT32)type & SW_AMDIX_DEBUG_MASK) printf X
   #define SW_AMDIX_LOG(type,S)           if((UINT32)type & SW_AMDIX_DEBUG_MASK) logMsg(S,0,0,0,0,0,0)
   #define SW_AMDIX_LOG1(type,S,P1)       if((UINT32)type & SW_AMDIX_DEBUG_MASK) logMsg(S,P1,0,0,0,0,0)
   #define SW_AMDIX_LOG2(type,S,P1,P2)    if((UINT32)type & SW_AMDIX_DEBUG_MASK) logMsg(S,P1,P2,0,0,0,0)
   #define SW_AMDIX_LOG3(type,S,P1,P2,P3) if((UINT32)type & SW_AMDIX_DEBUG_MASK) logMsg(S,P1,P2,P3,0,0,0)
   #define SW_AMDIX_LOG4(type,S,P1,P2,P3,P4)    if((UINT32)type & SW_AMDIX_DEBUG_MASK) logMsg(S,P1,P2,P3,P4,0,0)
   #define SW_AMDIX_LOG5(type,S,P1,P2,P3,P4,P5) if((UINT32)type & SW_AMDIX_DEBUG_MASK) logMsg(S,P1,P2,P3,P4,P5,0)
#else
   #define SW_AMDIX_DEBUG(type,X)          
   #define SW_AMDIX_LOG(type,S)            
   #define SW_AMDIX_LOG1(type,S,P1)        
   #define SW_AMDIX_LOG2(type,S,P1,P2)     
   #define SW_AMDIX_LOG3(type,S,P1,P2,P3)
   #define SW_AMDIX_LOG4(type,S,P1,P2,P3,P4)
   #define SW_AMDIX_LOG5(type,S,P1,P2,P3,P4,P5)

#endif /* SW_AMDIX_DEBUG_ENABLE */


/*******************************************************************************
*
* sysGetAutoMDIX_PHY_ResetDelay - auto MDIX initialization
*
* delivers the delay time the PHY module needs after RESET
*
* RETURNS: delay time of the PHY module after RESET [ms]
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_PHY_ResetDelay(void)
{
   return(INCA_IP_AUTOMDIX_PHY_RESET_DELAY_ms);
}

/*******************************************************************************
*
* sysGetAutoMDIX_PC_ON_OFF - Auto MDIX initialization                          
*
* Switch ON / OFF the auto MDIX feature                      
*
* RETURNS: 0: auto MDIX is OFF, 1: auto MDIX is ON
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_PC_ON_OFF(void)
{
   #if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_PC == TRUE)
      SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC: autoMDIX = %d\n", INCA_IP_AUTO_MDIX_PC_AUTOMDIX_ON_OFF ) ;
      return(INCA_IP_AUTO_MDIX_PC_AUTOMDIX_ON_OFF) ;
   #else
      SW_AMDIX_LOG(SW_AMDIX_DEBUG_PC_PHY, "PC: autoMDIX disabled\n") ;
      return(INCA_IP_AUTO_MDIX_PC_AUTOMDIX_OFF) ;
   #endif
}

/*******************************************************************************
*
* sysGetAutoMDIX_PC_Support_POL - Auto MDIX with Power Over LAN (POL)
*
* For PC port normally not required - but return a "1" force the task to
* skip over the PHY reset at the beginning.
*
* If Power Over LAN is supported and used, the driver task should not do
* an PHY reset at the beginning of the initialization.
* If Power Over LAN is supported but not used, return 0 to force a PHY reset
* at the task initialization.
*
* RETURNS: 0: no POL, 1: support POL
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_PC_Support_POL(void)
{
      SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "PC: POWER over LAN supported = %d\n", INCA_IP_AUTO_MDIX_PC_SUPPORT_POL) ;
      return(INCA_IP_AUTO_MDIX_PC_SUPPORT_POL) ;
}

/*******************************************************************************
*
* sysGetAutoMDIX_PC_StatusDelay - auto MDIX initialization
*
* delivers the delay time the PHY module needs until the new    
* state is valid in the PHY status registers after the the state 
* has changed.
*
* RETURNS: delay time of the PHY module
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_PC_StatusDelay(void)  
{
   SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC: autoMDIX Status Delay = %d ms\n", 
                 INCA_IP_AUTOMDIX_STATUS_DELAY_ms ) ;
   return(INCA_IP_AUTOMDIX_STATUS_DELAY_ms);
}

/*******************************************************************************
*
* sysGetAutoMDIX_PC_LinkTries - auto MDIX initialization
*
* delivers a counter for attempts of link detection used by the 
* driver internal auto MDIX task. 
* After EnergyOn is detected the task search for an valid link.
* Therefore the relays will be switched "LinkTries" times until 
* the task will go to RESET state (listen on both lines again)
*
* RETURNS: number of attempts for detecting a link
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_PC_LinkTries(void)  
{
   SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC: autoMDIX Link Tries = %d\n", 
                 INCA_IP_AUTOMDIX_PC_TRY_NUM ) ;
   return(INCA_IP_AUTOMDIX_PC_TRY_NUM);
}

/*******************************************************************************
*
* sysSetupAutoMDIX_PC_GPIO - Hardware initialization for the AutoMDIX feature
*
* Initialize the GPIO pins to control the external relais.
*
* RETURNS: 
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
void sysSetupAutoMDIX_PC_GPIO(void)
{
   if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_PC)
   {
      /* PC-port: setup 2 auto MDIX GPIO pins */
      *INCA_IP_AUTO_MDIX_PC_PORTS_DIR_RX |= (INCA_IP_AUTO_MDIX_PC_GPIO_PIN_MASK_RX);
      *INCA_IP_AUTO_MDIX_PC_PORTS_ALTSEL_RX |= (INCA_IP_AUTO_MDIX_PC_GPIO_PIN_MASK_RX);
/*      *INCA_IP_AUTO_MDIX_PC_PORTS_DIR_TX |= (INCA_IP_AUTO_MDIX_PC_GPIO_PIN_MASK_TX); */
/*      *INCA_IP_AUTO_MDIX_PC_PORTS_ALTSEL_TX |= (INCA_IP_AUTO_MDIX_PC_GPIO_PIN_MASK_TX); */
      /* configure EBU config register --> latch corresponding chip select */


      /* 
         INCA_IP_EBU_EBU_BUSCON1_CMULT_SET(1) |
         INCA_IP_EBU_EBU_BUSCON1_HOLDC_SET(3) |
         INCA_IP_EBU_EBU_BUSCON1_WAITRDC_SET(2) |
         INCA_IP_EBU_EBU_BUSCON1_WAITWRC_SET(2) |
         INCA_IP_EBU_EBU_BUSCON1_ALEC_GET(value) |
         INCA_IP_EBU_EBU_BUSCON1_PORTW_SET(2) |
         INCA_IP_EBU_EBU_BUSCON1_SETUP(1) |

      */
      *INCA_IP_EBU_EBU_BUSCON1 = 0x004242B1 ;

   }
}

/*******************************************************************************
*
* sysSetAutoMDIX_PC_MDIMode - change auto MDIX settings to MDI mode
*
* establish a straight connection via the external relays
* GPIO pins:
* - incoming TX --> RX INCA IP port 
* - incoming RX --> TX INCA IP port
*
* RETURNS: 1: INCA_IP_AUTOMDIX_UMODE_MDI   (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX  (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_PC_MDIMode(void)
{
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: set MDI mode\n") ;
   INCA_IP_AUTO_MDIX_PC_TX_RELAIS_TX_to_TX(*INCA_IP_AUTO_MDIX_PC_PORTS_OUT_TX);
   INCA_IP_AUTO_MDIX_PC_RX_RELAIS_RX_to_RX(*INCA_IP_AUTO_MDIX_PC_PORTS_OUT_RX);

   return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
}

/*******************************************************************************
*
* sysSetAutoMDIX_PC_MDIXMode - change auto MDIX settings to MDIX mode
*
* establish a crossed connection via the external relays
* GPIO pins:
* - incoming TX --> TX INCA IP port
* - incoming RX --> RX INCA IP port
*
* RETURNS: 1: INCA_IP_AUTOMDIX_UMODE_MDI   (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX  (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_PC_MDIXMode(void)
{
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: set MDIX mode\n") ;

   INCA_IP_AUTO_MDIX_PC_RX_RELAIS_RX_to_TX(*INCA_IP_AUTO_MDIX_PC_PORTS_OUT_RX) ;
   INCA_IP_AUTO_MDIX_PC_TX_RELAIS_TX_to_RX(*INCA_IP_AUTO_MDIX_PC_PORTS_OUT_TX) ;


   return(INCA_IP_AUTOMDIX_UMODE_MDIX) ;
}

/*******************************************************************************
*
* sysSetAutoMDIX_PC_RESETMode - change auto MDIX settings to RESET mode
*
* both incoming lines (TX and RX) are connected to the INCA IP RX port 
* any traffic will be detected independed of straight or crossed connection 
*    --> wait for "EnergyOn"
* GPIO pins:
* - incoming TX --> RX INCA IP port 
* - incoming RX --> RX INCA IP port
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_PC_RESETMode(void)
{
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: set RESET (DETECT) mode\n") ;
   if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_PC)
   {
      INCA_IP_AUTO_MDIX_PC_RX_RELAIS_RX_to_RX(*INCA_IP_AUTO_MDIX_PC_PORTS_OUT_RX) ;
      INCA_IP_AUTO_MDIX_PC_TX_RELAIS_TX_to_RX(*INCA_IP_AUTO_MDIX_PC_PORTS_OUT_TX) ;
   }
   /* DETECT-mode supported --> return RESET-mode */
   return (INCA_IP_AUTOMDIX_UMODE_RESET) ;
}

/*******************************************************************************
*
* sysSetAutoMDIX_PC_StartupMode - change auto MDIX settings to Startup mode
*
* in RESET mode both incoming lines (TX and RX) are connected to the 
* INCA IP RX port --> any traffic will be detected independed of 
* straight or crossed connection --> wait for "EnergyOn"
* after ENERGYON --> go the the user defined mode after reset.
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_PC_StartupMode(void)
{
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: set STARTUP mode\n") ;
   return( sysSetAutoMDIX_PC_MDIXMode() ) ;
}

/*******************************************************************************
*
* sysGetAutoMDIX_PC_Mode - get and return the current auto MDIX mode
*
* checks the current setting of the PC GPIO port 
* In the current implementation, there are 2 GPIO-pins necessary
*
* RESET (DETECT) mode:
*   incoming TX --> INCA-IP RX:  TX-Relay - GPIO pin = 1
*   incoming RX --> INCA-IP RX:  RX-Relay - GPIO pin = 0
*
* MDI mode:
*   incoming TX --> INCA-IP RX:  TX-Relay - GPIO pin = 1
*   incoming RX --> INCA-IP TX:  RX-Relay - GPIO pin = 1
*
* MDIX mode:
*   incoming TX --> INCA-IP TX:  TX-Relay - GPIO pin = 0
*   incoming RX --> INCA-IP RX:  RX-Relay - GPIO pin = 0
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*         -1: ERROR - no valid mode
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
int sysGetAutoMDIX_PC_Mode(void)
{
   UINT32   currentOutReg = 0 ;

   /* read GPIO pin */
   INCA_IP_AUTO_MDIX_PC_GET_MDIX_GPIO_REG( currentOutReg, 
                                           *INCA_IP_AUTO_MDIX_PC_PORTS_OUT_RX,
                                           *INCA_IP_AUTO_MDIX_PC_PORTS_OUT_TX ) ;

   switch(currentOutReg)
   {
      case INCA_IP_AUTO_MDIX_PC_GPIO_PIN_RESET_MODE:
         /* DETECT mode */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: current mode (%x) is DETECT\n",
                                              currentOutReg) ;
         return(INCA_IP_AUTOMDIX_UMODE_RESET) ;
         break ;

      case INCA_IP_AUTO_MDIX_PC_GPIO_PIN_MDIX_MODE:
         /* MDIX mode - default after power on */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: current mode (%x) is MDIX\n",
                                              currentOutReg) ;
         return(INCA_IP_AUTOMDIX_UMODE_MDIX) ;
         break ;

      case INCA_IP_AUTO_MDIX_PC_GPIO_PIN_MDI_MODE:
         /* MDI mode */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: current mode (%x) is MDI\n",
                                              currentOutReg) ;
         return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
         break ;
      default :
         /* MDI mode */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: ERROR no valid mode = %x\n",
                                              currentOutReg ) ;
         return(-1) ;
         break ;
   }
   return (-1) ;
}

/*******************************************************************************
*
* sysCheckAutoMDIX_PC_PortSetup - get and return the current LAN auto MDIX port setup
*
* check the current setting of the LAN GPIO port 
* check DIR register
* check ALTSEL register
*
* RETURNS: 0:  port setup ok
*          -1: port configuration not OK
*
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
int sysCheckAutoMDIX_PC_PortSetup(void)
{
   UINT32   currentReg = 0 ;

   /* read DIR register pin */
   INCA_IP_AUTO_MDIX_PC_GET_MDIX_GPIO_REG( currentReg,
                                           *INCA_IP_AUTO_MDIX_PC_PORTS_DIR_RX ,
                                           *INCA_IP_AUTO_MDIX_PC_PORTS_DIR_TX ) ;

   if (INCA_IP_AUTO_MDIX_PC_GPIO_PIN_RXTX_STATE != currentReg)
   {
      /* GPIO DIR pins not set --> config error */
      SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: DIR register (%d) not correct\n",
                                            currentReg) ;
      return(-1) ;
   }

   currentReg = 0 ;
   /* read ALTSEL register pin */
   INCA_IP_AUTO_MDIX_PC_GET_MDIX_GPIO_REG( currentReg,
                                           *INCA_IP_AUTO_MDIX_PC_PORTS_ALTSEL_RX ,
                                           *INCA_IP_AUTO_MDIX_PC_PORTS_ALTSEL_TX ) ;
   if (INCA_IP_AUTO_MDIX_PC_GPIO_PIN_RXTX_STATE != currentReg)
   {
      /* GPIO ALTSEL pins not set --> config error */
      SW_AMDIX_LOG1(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: ALTSEL register (%x) not correct\n",
                                           currentReg) ;
      return(-1) ;
   }

   SW_AMDIX_LOG(SW_AMDIX_DEBUG_PC_PHY, "PC autoMDIX: port setup OK\n") ;
   return(0) ;
}

/*** REMARK ********************************************************************
* The LAN port in the Siemens solution does not support an special DETECT-mode
* DETECT-mode (both incoming lines to the RX INCA-IP port):
*  - incoming LAN lines TX --> RX INCA IP port 
*  - incoming LAN lines RX --> RX INCA IP port
*
* So to detect incoming signals the INCA-IP has to toggle between MDI and MDIX mode
*
* A permanent toggling of the external relays will be accepted in case of 
* unplugged LAN cable                                   
*
*******************************************************************************/

/*******************************************************************************
*
* sysGetAutoMDIX_LAN_ON_OFF - Auto MDIX initialization                          
*
* Switch ON / OFF the auto MDIX feature                      
*
* RETURNS: 0: auto MDIX is OFF, 1: auto MDIX is ON
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_LAN_ON_OFF(void)
{
/*
   if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_LAN == 1)
      return(INCA_IP_AUTO_MDIX_LAN_AUTOMDIX_ON_OFF) ;
   else
      return(INCA_IP_AUTO_MDIX_LAN_AUTOMDIX_OFF) ;
*/
   #if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_LAN == 1)
      SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN: autoMDIX = %d\n", INCA_IP_AUTO_MDIX_LAN_AUTOMDIX_ON_OFF ) ;
      return(INCA_IP_AUTO_MDIX_LAN_AUTOMDIX_ON_OFF) ;
   #else
      SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN: autoMDIX disabled\n") ;
      return(INCA_IP_AUTO_MDIX_LAN_AUTOMDIX_OFF) ;
   #endif

}


/*******************************************************************************
*
* sysGetAutoMDIX_LAN_Support_POL - Auto MDIX with Power Over LAN (POL)
*
* If Power Over LAN is supported and used, the driver task should not do
* an PHY reset at the beginning of the initialization.
* If Power Over LAN is supported but not used, return 0 to force a PHY reset
* at the task initialization.
*
* RETURNS: 0: no POL, 1: support POL
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_LAN_Support_POL(void)
{
   #if (INCA_IP_BOARD_SUPPORT_POL_LAN == 1)
      SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN: POWER over LAN support = %d\n", INCA_IP_AUTO_MDIX_LAN_SUPPORT_POL) ;
      return(INCA_IP_AUTO_MDIX_LAN_SUPPORT_POL) ;
   #else
      SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN: POWER over LAN not supported\n") ;
      return(INCA_IP_AUTO_MDIX_LAN_SUPPORT_POL_NO) ;
   #endif
}

/*******************************************************************************
*
* sysGetAutoMDIX_LAN_StatusDelay - auto MDIX initialization
*
* delivers the delay time the PHY module needs until the new    
* state is valid in the PHY status registers after the the state 
* has changed.
*
* RETURNS: delay time of the PHY module
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysGetAutoMDIX_LAN_StatusDelay(void)  
{
   return(INCA_IP_AUTOMDIX_STATUS_DELAY_ms);
}

/*******************************************************************************
*
* sysGetAutoMDIX_PC_LinkTries - auto MDIX initialization
*/
UINT32 sysGetAutoMDIX_LAN_LinkTries(void)  
{
   return(INCA_IP_AUTOMDIX_LAN_TRY_NUM);
}

/*******************************************************************************
*
* sysSetupAutoMDIX_LAN_GPIO - Hardware initialization for the AutoMDIX feature
*
* Initialize the GPIO pins to control the external relais.
*
* RETURNS: 
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
void sysSetupAutoMDIX_LAN_GPIO(void)
{
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: setup LAN GPIO\n") ;
   if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_LAN)
   {
      /* LAN-port: setup 1 auto MDIX GPIO pin */
      *INCA_IP_AUTO_MDIX_LAN_PORTS_DIR    |= (INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_MASK_RXTX);
      *INCA_IP_AUTO_MDIX_LAN_PORTS_ALTSEL |= (INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_MASK_RXTX);
   }
}

/*******************************************************************************
*
* sysSetAutoMDIX_LAN_MDIMode - change auto MDIX settings to MDI mode
*
* establish a straight connection via the external relays
* GPIO pins:
* - incoming TX --> RX INCA IP port 
* - incoming RX --> TX INCA IP port
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_LAN_MDIMode(void)
{

   SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: set MDI mode\n") ;
   #ifdef INCA_IP_AUTOMDIX_ONE_GPIO_PIN
      INCA_IP_AUTO_MDIX_LAN_SET_MDI(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
   #else
      INCA_IP_AUTO_MDIX_LAN_TX_RELAIS_TX_to_RX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
      INCA_IP_AUTO_MDIX_LAN_RX_RELAIS_RX_to_TX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
   #endif

   return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
}

/*******************************************************************************
*
* sysSetAutoMDIX_LAN_MDIXMode - change auto MDIX settings to MDIX mode
*
* establish a crossed connection via the external relays
* establish MDI (via one GPIO pin):
* - incoming TX --> TX INCA IP port 
* - incoming RX --> RX INCA IP port
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_LAN_MDIXMode(void)
{
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: set MDIX mode\n") ;

   #ifdef INCA_IP_AUTOMDIX_ONE_GPIO_PIN
      INCA_IP_AUTO_MDIX_LAN_SET_MDIX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
   #else
      INCA_IP_AUTO_MDIX_LAN_RX_RELAIS_RX_to_RX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
      INCA_IP_AUTO_MDIX_LAN_TX_RELAIS_TX_to_TX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
   #endif

   return(INCA_IP_AUTOMDIX_UMODE_MDIX) ;
}

/*******************************************************************************
*
* sysSetAutoMDIX_LAN_RESETMode - change auto MDIX settings to RESET mode
*
* The LAN port in the Siemens solution does not support DETECT-mode  
* A permanent toggling of the external relays will be accepted in case of 
* unplugged LAN cable                                   
* RESET-mode (DETECT):
*    for the Siemens board, no special RESET-mode (DETECT) is supported.
*    The used mode is MDIX
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_LAN_RESETMode(void)
{
   if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_LAN)
   {
      #ifdef INCA_IP_AUTOMDIX_ONE_GPIO_PIN
         /* no DETECT-mode for LAN supported - default after power up is MDI */
         SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: set RESET mode --> No Detect mode supported\n") ;
         return (INCA_IP_AUTOMDIX_UMODE_MDI) ;
      #else
         /* set detect mode */
         SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: set RESET mode --> MDIy\n") ;

         INCA_IP_AUTO_MDIX_LAN_RX_RELAIS_RX_to_RX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
         INCA_IP_AUTO_MDIX_LAN_TX_RELAIS_TX_to_RX(*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;

         return( sysSetAutoMDIX_LAN_MDIMode() ) ;
      #endif
   }
   return (INCA_IP_AUTOMDIX_UMODE_RESET) ;
}

/*******************************************************************************
*
* sysSetAutoMDIX_LAN_StartupMode - change auto MDIX settings to Startup mode
*
* After the RESET-mode (DETECT) this is the first mode the auto MDIX driver 
* chooses after detection of a valid signal for the external relays to 
* set a prefered and expected configuration
*
* The LAN port in the Siemens solution does not support DETECT-mode  
* A permanent toggling of the external relays will be accepted in case of 
* unplugged LAN cable                                   
* In this solution we toggle the relays between MDI and MDIX until a link
* is found.
* 
* IMPORTANT:
* Because there is no DETECT-mode in this solution, the reset mode was already
* the correct mode - so toggling makes no sence
* --> set the same mode like used in the DETECT (RESET) mode
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
UINT32 sysSetAutoMDIX_LAN_StartupMode(void) 
{
   if (INCA_IP_BOARD_SUPPORT_AUTOMDIX_LAN)
   {
      SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: set STARTUP mode --> MDIy\n") ;
      return( sysSetAutoMDIX_LAN_MDIXMode() ) ;
   }
   else
   {
      /* no DETECT-mode for LAN supported - default after power up is MDI */
      return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
   }
}

/*******************************************************************************
*
* sysGetAutoMDIX_LAN_Mode - get and return the current auto MDIX mode
*
* checks the current setting of the LAN GPIO port 
* GPIO pin = 0 : not set - MDI-mode (default mode for the LAN port after power on)
* GPIO pin = 1 : set - MDIX-mode
*
* RETURNS: 0: INCA_IP_AUTOMDIX_UMODE_RESET   (drv: INCA_IP_AUTOMDIX_MODE_RESET)
*          1: INCA_IP_AUTOMDIX_UMODE_MDI     (drv: INCA_IP_AUTOMDIX_MODE_MDI) 
*          2: INCA_IP_AUTOMDIX_UMODE_MDIX    (drv: INCA_IP_AUTOMDIX_MODE_MDIX)
*
* Important: must match with driver definitions
*
* REMARK: this function will be called by the switch driver 
*         and must be provided by the BSP
*/
int sysGetAutoMDIX_LAN_Mode(void)
{
   UINT32   currentOutReg = 0 ;

   #ifdef INCA_IP_AUTOMDIX_ONE_GPIO_PIN
      /* read GPIO pin */
      INCA_IP_AUTO_MDIX_LAN_GET_MDIX_GPIO_REG(currentOutReg, *INCA_IP_AUTO_MDIX_LAN_PORTS_OUT) ;
      
      if (0 == currentOutReg)
      {
         /* GPIO pin not set --> MDI (default after power up) */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: current mode (%x) is MDI\n",
                                               currentOutReg) ;
         return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
      }
      else
      {
         /* GPIO pin set --> MDIX (default after power up is MDI) */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: current mode (%x) is MDIX\n",
                                               currentOutReg) ;
         return(INCA_IP_AUTOMDIX_UMODE_MDIX) ;
      }
   #else
      switch(currentOutReg)
      {
         case INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_DETECT_MASK:
            /* DETECT mode */
            SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: current mode (%x) is DETECT\n",
                                                 currentOutReg) ;
            return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
            break ;

         case INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_MDIX_MASK:
            /* MDIX mode - default after power on */
            SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: current mode (%x) is MDIX\n",
                                                 currentOutReg) ;
            return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
            break ;

         case INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_MDI_MASK:
            /* MDI mode */
            SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: current mode (%x) is MDI\n",
                                                 currentOutReg) ;
            return(INCA_IP_AUTOMDIX_UMODE_MDI) ;
            break ;
         default :
            /* MDI mode */
            SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: ERROR no valid mode = %x\n",
                                                 currentOutReg ) ;
            return(-1) ;
            break ;
      }
   #endif
   return (-1) ;
}

/*******************************************************************************
*
* sysCheckAutoMDIX_LAN_PortSetup - get and return the current LAN auto MDIX port setup
*
* check the current setting of the LAN GPIO port 
* check DIR register
* check ALTSEL register
*
* RETURNS: 0:  port setup ok
*          -1: port configuration not OK
*
*
* REMARK: this function will be called by the switch driver
*         and must be provided by the BSP
*/
int sysCheckAutoMDIX_LAN_PortSetup(void)
{
   UINT32   currentReg = 0 ;

   #ifdef INCA_IP_AUTOMDIX_ONE_GPIO_PIN
      /* read DIR register pin */
      INCA_IP_AUTO_MDIX_LAN_GET_MDIX_GPIO_REG(currentReg, *INCA_IP_AUTO_MDIX_LAN_PORTS_DIR) ;
      if (0 == currentReg)
      {
         /* GPIO pin not set --> MDI (default after power up) */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: DIR register (%x) not correct\n",
                                               currentReg) ;
         return(-1) ;
      }

      /* read DIR register pin */
      INCA_IP_AUTO_MDIX_LAN_GET_MDIX_GPIO_REG(currentReg, *INCA_IP_AUTO_MDIX_LAN_PORTS_ALTSEL) ;
      if (0 == currentReg)
      {
         /* GPIO pin not set --> MDI (default after power up) */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: ALTSEL register (%x) not correct\n",
                                                currentReg) ;
         return(-1) ;
      }
   #else
      /* read DIR register pin */
      INCA_IP_AUTO_MDIX_LAN_GET_MDIX_GPIO_REG(currentReg, *INCA_IP_AUTO_MDIX_LAN_PORTS_DIR) ;
      if (INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_MASK_RXTX == currentReg)
      {
         /* GPIO pin not set --> MDI (default after power up) */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: DIR register (%d) not correct\n",
                                               currentReg) ;
         return(-1) ;
      }

      /* read DIR register pin */
      INCA_IP_AUTO_MDIX_LAN_GET_MDIX_GPIO_REG(currentReg, *INCA_IP_AUTO_MDIX_LAN_PORTS_DIR) ;
      if (INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_MASK_RXTX == currentReg)
      {
         /* GPIO pin not set --> MDI (default after power up) */
         SW_AMDIX_LOG1(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: ALTSEL register (%x) not correct\n",
                                               currentReg) ;
         return(-1) ;
      }
   #endif

   /* GPIO pin set --> MDIX (default after power up is MDI) */
   SW_AMDIX_LOG(SW_AMDIX_DEBUG_LAN_PHY, "LAN autoMDIX: port setup OK\n") ;
   return(0) ;
}


