/************************************************************************
 *
 * Copyright (c) 2004
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
 
/******************************************************************************
   Module      : INCAIPSwitchHash.c
   Date        : 24.04.2003
   Description : provide fuctions to calculate the hash algorithm used within
                 the INCA-IP switch module to calculate hash table index for
                 a MAC address.
                 While learning new a MAC address the switch module calculates 
                 the MAC table index via a hash algorithm. To setup a MAC address
                 manual the index has to be calculated also manual.
                 Via a modulation algorithm the MAC address for hash calculation
                 can be changed. So at least there are 4 possibilities for calculate
                 the index:
                     hash( original MAC )
                     hash( modulated MAC - code 0 )
                     hash( modulated MAC - code 1 )
                     hash( modulated MAC - code 2 )
                 To setup a MAC entry manual calculate one hash and check if the 
                 calculated location is free. If not calculate the next hash.


   $Log: inca_switch_hash.c,v $
   Revision 1.1  2003/07/14 08:57:09  rjanse_a
   Initial revision of ported vxWorks switch driver (inca_switch_hash, inca_switch_access) and a simple control interface (inca_switch_api, api.sh)

   Revision 1.2  2003/05/20 12:36:33  mangele
   shift test function to the PHY and SWITCH module (IncaSwitchApi.c)

   Revision 1.1  2003/04/25 19:14:20  mangele
   These HASH functions are used for calculate the MAC table index within the INCA-IP switch module
   They are used first time in the "IncaSwitchEnd.c" module version 1.17.


   
******************************************************************************/

/*  #include "vxWorks.h" */
/*  #include "stdio.h" */
/*  #include "string.h" */

/*  #include "IncaSwitchHash.h" */

#define EXPORT_SYMTAB

#include <asm/incaip/inca-ip.h>
#include <asm/io.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/init.h>

#include "vxWorks_type.h"
#include "inca_switch_hash.h"


/* 
   codes for modulate the MAC address, used for calculate the MAC table index 
*/
LOCAL IncaIpHashCodes   IncaIpMacModulateCodes = 
{ 
   {0xAAAA, 0xCCCC, 0x9999} ,
   {0xA5A5, 0x9696, 0xAA55} ,
   {0x9966, 0xA55A, 0x9669} ,
} ;

/* 
   polynome used for hash calculation 
*/
LOCAL UINT IncaIpHashGenPolyb[8] = { 1, 1, 0, 0, 0, 0, 0, 0 };

/*******************************************************************************
*
* sw21553_IncaIpInitHashGenPoly - init the polynom for hash calculation
*
* REMARK:
*     The polynom has to be initialzed before first use
*     (changed to static initialization during compile time)
*/
void sw21553_IncaIpInitHashGenPoly()
{
/*     memset(IncaIpHashGenPolyb, 0x00, sizeof(IncaIpHashGenPolyb) ) ; */

/*  	IncaIpHashGenPolyb[1] = 1; */
/*  	IncaIpHashGenPolyb[0] = 1; */
    
   return ;
}


/*******************************************************************************
*
* IncaIpSwitchModulateMac - Modulate the address based on the given modulate code
*     The result is a modulated MAC address (based on the orignial address) which
*     can be used for further hash calculation.
*
* PARAMETERS:
*     modulateCode:  3x 16 bit stream used for modulation
*     inAddrPart:    16 bit part of the original MAC address. The address part can be 
*                    a combination of the original MAC address.
*     retAddr:       retrun parameter, give back the modulated MAC address
*
*/
LOCAL void IncaIpSwitchModulateMac(UINT16 modulateCode[3], UCHAR inAddrPart[2], UCHAR *retAddr)
{
	int inputb[16];
	int codeb[48], not_codeb[48];
	int signalb[48];
	
	int i, j, n;

   /* init inputb table */
	for (i = 0; i < 8; i++)
	{	
		inputb[8+i]  = ((inAddrPart[0] >> i) & 0x0001);
      inputb[i]    = ((inAddrPart[1] >> i) & 0x0001);   
   }        /* for (i = 0; i < 8; i++) {...} */

   /* init codeb and not_codeb array */
   for (i = 0; i < 16; i++)
	{	
		codeb[32+i] = (modulateCode[0] >> i) & 0x0001 ;
      not_codeb[32+i] = (codeb[32+i] == 1) ? 0:1 ;

		codeb[16+i] = (modulateCode[1] >> i) & 0x0001 ;
      not_codeb[16+i] = (codeb[16+i] == 1) ? 0:1 ;

		codeb[i] =    (modulateCode[2] >> i) & 0x0001 ;
      not_codeb[i] = (codeb[i] == 1) ? 0:1 ;
	}        /* for (i = 0; i < 16; i++) {...} */

   /* reset signalb array */
   memset(signalb, 0x00, sizeof(signalb) ) ;

	for (n = i = 0; i < 16; i++)
	{
		if (inputb[i] == 1)
		{
			for (j = 0; j < 3; j++, n++)
			{
				signalb[n] = codeb[n];
			}
		}
		else     /* if (inputb[i] == 1) {...} */
		{
			for (j = 0; j < 3; j++, n++)
			{
				signalb[n] = not_codeb[n];
			}
		}        /* if (inputb[i] == 1) {...} */
	}        /* for (n = i = 0; i < 16; i++) {...} */

   /* set return address */
   memset(retAddr, 0x00, sizeof(retAddr) ) ;

	for (i = 8; i >=0; i--)
	{
      retAddr[5] = retAddr[5] << 1 ;
      retAddr[4] = retAddr[4] << 1 ;
      retAddr[3] = retAddr[3] << 1 ;
      retAddr[2] = retAddr[2] << 1 ;
      retAddr[1] = retAddr[1] << 1 ;
      retAddr[0] = retAddr[0] << 1 ;

		if (signalb[i] == 1)
			retAddr[5] |= 0x01;
      if (signalb[i+8] == 1)
			retAddr[4] |= 0x01;

      if (signalb[i+16] == 1)
			retAddr[3] |= 0x01;
      if (signalb[i+24] == 1)
			retAddr[2] |= 0x01;

      if (signalb[i+32] == 1)
			retAddr[1] |= 0x01;
      if (signalb[i+40] == 1)
			retAddr[0] |= 0x01;
	}        /* for (i = 8; i >=0; i--) {...} */
}        /* LOCAL void IncaIpSwitchModulateMac(...) {...} */


/*******************************************************************************
*
* IncaIpSwitchCalcMacHash - Calculate a hash value form the given MAC address
*
* PARAMETERS:
*     inMacAddr:     MAC address used for hash calculation
*     len:           currently fixed 8
*
*  RETURN:
*     calculated hash value
*
*  REMARK:
*     This value is used as MAC table index for locate a MAC address within the
*     INCA-IP MAC table (switch module)
*/
LOCAL UCHAR IncaIpSwitchCalcMacHash(UCHAR inMacAddr[6], UINT32 len)
{
	int   i, j ;
	int   inputb[48] ;
	int   crc_reg[8], crc[8] ;
	UCHAR crc_reg_final ;

   /* init inputb table */
	for (i = 0; i < 8; i++)
	{	
		inputb[40+i] = ((inMacAddr[0] >> i) & 0x0001);
		inputb[32+i] = ((inMacAddr[1] >> i) & 0x0001);
		inputb[24+i] = ((inMacAddr[2] >> i) & 0x0001);
		inputb[16+i] = ((inMacAddr[3] >> i) & 0x0001);
		inputb[8+i]  = ((inMacAddr[4] >> i) & 0x0001);
      inputb[i]    = ((inMacAddr[5] >> i) & 0x0001);
	}        /* for (i = 0; i < 8; i++) {...} */

   /* init CRC table */
	for(i = 0; i < 8; i++)
	{
		crc_reg[i] = 0;
		crc[i] = 0;
	}

 	for (i = 47; i >= 0; i--)
	{
		if ((crc_reg[len-1] ^ inputb[i]) == 1)
		{			
			crc_reg[len-1] = 0;
			for (j = 0; j < 8; j++)
			{
				crc[7-j] = crc_reg[7-j] ^ IncaIpHashGenPolyb[7-j];
			}
			
			for(j = 0; j < 7; j++)
			{
				crc[7-j] = crc[7-j-1];
			}
			crc[0] = 1;
		}
		else     /* if ((crc_reg[len-1] ^ inputb[i]) == 1) {...} else ... */
		{
			for(j = 0; j < 7; j++)
			{
				crc[7-j] = crc[7-j-1];
			}
			crc[0] = 0;
		}        /* if ((crc_reg[len-1] ^ inputb[i]) == 1) {...} else {...} */
		
		for (j = 0; j < 8; j++)
			crc_reg[j] = crc[j];
	}        /* for (i = 47; i >= 0; i--) {...} */

 	crc_reg_final = 0x00;
	for (i = 7; i >=0; i--)
	{
		crc_reg_final = crc_reg_final << 1;
		
		if (crc_reg[i] == 1)
			crc_reg_final = crc_reg_final | 0x01;
			
	}        /* for (i = 7; i >=0; i--) {...} */

   return (crc_reg_final) ;
}

/*******************************************************************************
*
* sw21553_IncaIpSwitchCalcHash - choose a modulation code and the MAC address part
*     used for calculation of the modulated MAC address
*
* Parameters:
*     modulateNum:   range 0..2 - chose the used modulation code
*     inMacAddr:     original MAC address From this address a modulation depended part 
*                    will be used for calculation
*
*  Return:
*     hash value: returns the calculated hash value
*     -1:         ERROR, invalid modulation number
*
*  Remark:
*     The MAC address is expected a simple byte stream (no endian)
*     example: 00:50:04:cb:18:8c 
*              --> UCHAR inMacAddr[6] = {0x00, 0x50, 0x04, 0xcb, 0x18, 0x8c}
*/
UCHAR sw21553_IncaIpSwitchCalcHash(UINT modulateNum, UCHAR inMacAddr[6])
{
   UCHAR  modulateMacAddrPart[2]  ; 
   UCHAR  modulateMacAddr[6] ;

   switch(modulateNum)
   {
      case 0:
         /* 2. HASH */
         /* calculate next modulated hash */
         modulateMacAddrPart[0] = inMacAddr[0] ^ inMacAddr[4] ;
         modulateMacAddrPart[1] = inMacAddr[1] ^ inMacAddr[5] ;
         memset(modulateMacAddr, 0x00, sizeof(modulateMacAddr) ) ;
         IncaIpSwitchModulateMac( IncaIpMacModulateCodes.Code_2, 
                                  &modulateMacAddrPart[0], 
                                  modulateMacAddr) ;
         return( IncaIpSwitchCalcMacHash(modulateMacAddr, 8) ) ;
         break ;
      case 1:
         /* calculate next modulated hash */
         modulateMacAddrPart[0] = inMacAddr[2] ^ inMacAddr[4] ;
         modulateMacAddrPart[1] = inMacAddr[3] ^ inMacAddr[5] ;
         memset(modulateMacAddr, 0x00, sizeof(modulateMacAddr) ) ;
         IncaIpSwitchModulateMac( IncaIpMacModulateCodes.Code_1, 
                                  &modulateMacAddrPart[0], 
                                  modulateMacAddr) ;
         return( IncaIpSwitchCalcMacHash(modulateMacAddr, 8) ) ;
         break ;
      case 2:
         /* calculate next modulated hash */
         modulateMacAddrPart[0] = inMacAddr[4] ;
         modulateMacAddrPart[1] = inMacAddr[5] ;
         memset(modulateMacAddr, 0x00, sizeof(modulateMacAddr) ) ;
         IncaIpSwitchModulateMac( IncaIpMacModulateCodes.Code_0, 
                                  &modulateMacAddrPart[0], 
                                  modulateMacAddr) ;
         return( IncaIpSwitchCalcMacHash(modulateMacAddr, 8) ) ;
         break ;
      case 3:
         /* calculate hash based on original MAC address */
         return( IncaIpSwitchCalcMacHash(inMacAddr, 8) ) ;
         break ;
      default:
         return(ERROR) ;
         break ;
   }        /* switch(modulateNum) {...} */
}        /* UCHAR sw21553_IncaIpSwitchCalcHash(...) {...} */




/* Linux specific enhancements */

MODULE_LICENSE("Infineon");

int inca_switch_hash_init(void)
{
    return 0;
}

void inca_switch_hash_exit(void)
{
}

module_init(inca_switch_hash_init);
module_exit(inca_switch_hash_exit);

EXPORT_SYMBOL(sw21553_IncaIpInitHashGenPoly);
EXPORT_SYMBOL(sw21553_IncaIpSwitchCalcHash);
