   /*
   ** Copyright (C) 2005 Wu Qi Ming <Qi-Ming.Wu@infineon.com>
   **  
   ** This program is free software; you can redistribute it and/or modify
   ** it under the terms of the GNU General Public License as published by
   ** the Free Software Foundation; either version 2 of the License, or
   ** (at your option) any later version.
   ** 
   ** This program is distributed in the hope that it will be useful,
   ** but WITHOUT ANY WARRANTY; without even the implied warranty of
   ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   ** GNU General Public License for more details.
   ** 
   ** You should have received a copy of the GNU General Public License
   ** along with this program; if not, write to the Free Software 
   ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
   */
   /*
    * Description:
    * Driver for Infineon IFX DMA
    * Author:	Wu Qi Ming[Qi-Ming.Wu@infineon.com]
    * Created:	26-September-2005
    */

#include <linux/config.h> /* retrieve the CONFIG_* macros */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#if defined(MODVERSIONS) && !defined(__GENKSYMS__)
   //#include <linux/modversions.h>
#endif

#ifndef EXPORT_SYMTAB
#define EXPORT_SYMTAB /* need this one 'cause we export symbols */
#endif


#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/selection.h>
#include <linux/kmod.h>
#include <linux/vmalloc.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/semaphore.h>
#include <asm/incaip2/ifx_types.h>
#include <asm/incaip2/irq.h>
#ifdef CONFIG_INCAIP2
#include "ifx_reg_dma.h"
#else
#include <asm/ifx/ifx_reg_base.h>
#include <asm/ifx/ifx_reg_mps.h>
#include <asm/ifx/ifx_reg_dma.h>
#endif
#include <asm/incaip2/ifx_dma.h>

#ifdef CONFIG_SVIP
#include <asm/ifx/ifx_reg_sys1.h>
#endif
#include "ifx_dma_core.h"


/*****************definitions for the macros used in dma-core.c***************/
/*
 * Some peripherals rely on the transmit-complete indication before transmitting
 * the next packet. It is recommended to #define USE_TX_CPT_INT.
 */
#define USE_TX_CPT_INT
/*
 * If DMA Polling is enabled, the chip internal buses are loaded with descriptor
 * transfers even if channels have no data to transfer. To avoid this load it is 
 * recommended to #undef DMA_POLLING_MODE.
 */
#undef DMA_POLLING_MODE

#define IFX_SUCCESS 1
#define IFX_ERROR   0

#define IFX_DMA_EMSG(fmt, args...) printk( KERN_ERR  "%s: " fmt,__FUNCTION__, ## args)

/* Multiprocessor Semaphore for channel- and port-specific register accesses */
#ifdef IFX_MBS_BASE
#define IFX_MBSR_DMA        IFX_MPS_MBSR0
#define IFX_MBS_GRAB(x)     do {} while(*(x) != 0)
#define IFX_MBS_RELEASE(x)  *(x) = 0; asm("sync")
#else
#define IFX_MBS_GRAB(x)
#define IFX_MBS_RELEASE(x)
#endif

/*25 descriptors for each dma channel,4096/8/20=25.xx*/
#define IFX_DMA_DESCRIPTOR_OFFSET 25 

#define DMA_INT_BUDGET      100  /*budget for interrupt handling*/
#define DMA_POLL_COUNTER    32   /*set the correct counter value here!*/ 
/******************************************************************************/

/******************Global variables ******************************************/

struct proc_dir_entry* g_ifx_dma_dir;
u64* g_desc_list;

struct dma_device_info  dma_devs[MAX_DMA_DEVICE_NUM] = IFX_DMA_DEVICE_INFO;
struct dma_channel_info dma_chan[MAX_DMA_CHANNEL_NUM] = IFX_DMA_CHANNEL_INFO;

u32 g_ifx_dma_int_status=0;
/******************************************************************************/

/***********function definitions***********************************************/
int ifx_dma_init(void);
void do_dma_tasklet(unsigned long);
DECLARE_TASKLET(dma_tasklet,do_dma_tasklet,0);
irqreturn_t dma_interrupt(int irq, void *dev_id, struct pt_regs *regs);


/******************************************************************************/
u8* common_buffer_alloc(int len,int* byte_offset,void** opt)
{
   u8* buffer=(u8*)kmalloc(len*sizeof(u8),GFP_KERNEL);
   *byte_offset=0;
   return buffer;

}

int common_buffer_free(u8* dataptr,void* opt)
{
   if(dataptr)
      kfree(dataptr);
   return 0;
}

int enable_ch_irq(struct dma_channel_info* pCh)
{
   int chan_no=(int)(pCh-dma_chan);
   int flag;
   u32 val;

   if(pCh->dir==DIR_RX)
#ifndef DMA_POLLING_MODE
      val = IFX_DMA_CIE_DESCPT | IFX_DMA_CIE_DUR;
#else
      val = IFX_DMA_CIE_DESCPT;
#endif
   else
#ifdef USE_TX_CPT_INT
      val = IFX_DMA_CIE_DESCPT;
#else
      val = 0;
#endif

   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   *IFX_DMA_CIE=val;
   *IFX_DMA_IRNEN |=1<<chan_no;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   local_irq_restore(flag);
   enable_irq(pCh->irq);

   return IFX_SUCCESS;
}

int disable_ch_irq(struct dma_channel_info* pCh)
{
   int flag;
   int chan_no=(int)(pCh-dma_chan);

   local_irq_save(flag);
   g_ifx_dma_int_status &=~(1<<chan_no);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   *IFX_DMA_CIE=0;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   *IFX_DMA_IRNEN &=~(1<<chan_no);
   local_irq_restore(flag);
   disable_irq(pCh->irq);

   return IFX_SUCCESS;
}



int open_chan(struct dma_channel_info* pCh)
{
   int flag;
   int j;
   int chan_no=(int)(pCh-dma_chan);
   u8* buffer;
   int byte_offset;
   struct rx_desc* rx_desc_p;
   struct tx_desc* tx_desc_p;

   if (pCh->control==IFX_DMA_CH_ON)
      return -1;
      
   if (pCh->dir == DIR_RX)
   {
      for(j=0;j<pCh->desc_len;j++)
      {
         rx_desc_p=(struct rx_desc*)pCh->desc_base+j;
         buffer=pCh->dma_dev->buffer_alloc(pCh->packet_size,&byte_offset,(void*)&(pCh->opt[j]));
         if(!buffer) return -ENOBUFS;

         rx_desc_p->Data_Pointer=(u32)CPHYSADDR((u32)buffer);
         rx_desc_p->status.word = 0;
         rx_desc_p->status.field.byte_offset=byte_offset;
         rx_desc_p->status.field.data_length=pCh->packet_size;
         rx_desc_p->status.field.OWN=DMA_OWN;
      }
   }
   else /* DIR_TX */
   {
      for(j=0;j<pCh->desc_len;j++)
      {
         tx_desc_p=(struct tx_desc*)pCh->desc_base+j;
         tx_desc_p->Data_Pointer= 0;
         tx_desc_p->status.word = 0;
      }
   }
   pCh->xfer_cnt = 0;

   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   *IFX_DMA_CDLEN=pCh->desc_len;
   *IFX_DMA_CIS=0x7e;
   *IFX_DMA_CCTRL = IFX_DMA_CCTRL_TXWGT_VAL(pCh->tx_weight)
                  | IFX_DMA_CCTRL_CLASS_VAL(pCh->pri)
                  | (pCh->dir == DIR_RX ? IFX_DMA_CCTRL_ON_OFF : 0);
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   pCh->control=IFX_DMA_CH_ON;
   local_irq_restore(flag);

   if (request_irq(pCh->irq,dma_interrupt,SA_INTERRUPT, "dma-core",(void*)pCh) != 0)
   {
      printk("error, cannot get dma_irq!\n");
      free_irq(pCh->irq,(void*)&dma_interrupt);
      return -EFAULT;
   }
   enable_ch_irq(pCh);
   return 0;
}


int close_chan(struct dma_channel_info* pCh)
{
   int flag;
#if 0
   int val;
#endif
   int j;
   int chan_no=(int)(pCh-dma_chan);
   struct rx_desc* desc_p;
   
   if (pCh->control==IFX_DMA_CH_OFF)
      return -1;
      
   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   *IFX_DMA_CCTRL&=~IFX_DMA_CCTRL_ON_OFF;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   disable_ch_irq(pCh);
   free_irq(pCh->irq,(void*)&dma_interrupt);
   pCh->control=IFX_DMA_CH_OFF;
   local_irq_restore(flag);

#if 0
   /* can waiting for channel-off be forever? */
   do {
      local_irq_save(flag);
      IFX_MBS_GRAB(IFX_MBSR_DMA);
      *IFX_DMA_CS=chan_no;
      val=*IFX_DMA_CCTRL;
      IFX_MBS_RELEASE(IFX_MBSR_DMA);
      local_irq_restore(flag);
   } while(val & IFX_DMA_CCTRL_ON_OFF);
#endif

   /* free descriptors in use */
   for(j=0;j<pCh->desc_len;j++)
   {
      desc_p=(struct rx_desc*)pCh->desc_base+j;
      if((desc_p->status.field.OWN==CPU_OWN && desc_p->status.field.C) ||
         (desc_p->status.field.OWN==DMA_OWN))
      {
         pCh->dma_dev->buffer_free((u8*)__va(desc_p->Data_Pointer),(void*)pCh->opt[j]);
      }
   }

   return 0;
}


int reset_chan(struct dma_channel_info* pCh)
{
   int flag;
   int val;
   int chan_no=(int)(pCh-dma_chan);

   close_chan(pCh);
   
   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   *IFX_DMA_CCTRL|=IFX_DMA_CCTRL_RST;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   local_irq_restore(flag);

   do {
      local_irq_save(flag);
      IFX_MBS_GRAB(IFX_MBSR_DMA);
      *IFX_DMA_CS=chan_no;
      val=*IFX_DMA_CCTRL;
      IFX_MBS_RELEASE(IFX_MBSR_DMA);
      local_irq_restore(flag);
   } while(val & IFX_DMA_CCTRL_RST);
   
   return 0;
}


static inline void rx_chan_intr_handler(int chan_no)
{
   struct dma_device_info* pDev=(struct dma_device_info*)dma_chan[chan_no].dma_dev;
   struct dma_channel_info* pCh=&dma_chan[chan_no];
   struct rx_desc* rx_desc_p;
   int flag;
   u32 val;

   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   val=*IFX_DMA_CIS;
   *IFX_DMA_CIS=IFX_DMA_CIS_DESCPT;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);

#ifndef DMA_POLLING_MODE
   if (val & IFX_DMA_CIS_DUR)
      pCh->dur = 1;
#endif

   /*handle command complete interrupt*/
   rx_desc_p=(struct rx_desc*)pCh->desc_base+pCh->curr_desc;
   if((rx_desc_p->status.word & (DMA_DESC_OWN_DMA|DMA_DESC_CPT_SET)) == DMA_DESC_CPT_SET)
   {
      local_irq_restore(flag);
      /*Every thing is correct, then we inform the upper layer*/
      pDev->current_rx_chan=pCh->rel_chan_no;
      if(pDev->intr_handler) pDev->intr_handler(pDev, RCV_INT);
      pCh->weight--;
   }
   else
   {
      g_ifx_dma_int_status &=~(1<<chan_no);
      local_irq_restore(flag);
      enable_irq(dma_chan[chan_no].irq);
   }
}


static inline void tx_chan_intr_handler(int chan_no)
{
   struct dma_device_info* pDev=(struct dma_device_info*)dma_chan[chan_no].dma_dev;
   struct dma_channel_info* pCh=&dma_chan[chan_no];
   struct tx_desc* tx_desc_p;
   int flag;

   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   *IFX_DMA_CIS=IFX_DMA_CIS_DESCPT;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);

#ifdef USE_TX_CPT_INT
   tx_desc_p=(struct tx_desc*)pCh->desc_base+pCh->prev_desc;
   if((tx_desc_p->status.word & (DMA_DESC_OWN_DMA|DMA_DESC_CPT_SET)) == DMA_DESC_CPT_SET)
   {
      local_irq_restore(flag);

      pDev->buffer_free((u8*)__va(tx_desc_p->Data_Pointer),pCh->opt[pCh->prev_desc]);
      memset(tx_desc_p,0,sizeof(struct tx_desc));
      pDev->current_tx_chan=pCh->rel_chan_no;
      if(pDev->intr_handler) pDev->intr_handler(pDev, TRANSMIT_CPT_INT);
      pCh->weight--;

      pCh->prev_desc=(pCh->prev_desc+1)%(pCh->desc_len);
   }
   else
#endif
   {
      g_ifx_dma_int_status &=~(1<<chan_no);
      local_irq_restore(flag);
      enable_irq(dma_chan[chan_no].irq);
   }
}


void do_dma_tasklet(unsigned long unused)
{
   int i;
   int chan_no=0;
   int budget=DMA_INT_BUDGET;
   int weight=0;

   while(g_ifx_dma_int_status)
   {
      if(budget--<0)
      {
         tasklet_schedule(&dma_tasklet);
         return;
      }
      chan_no=-1;
      weight=0;
      /*WFQ algorithm to select the channel*/
      for(i=0;i<MAX_DMA_CHANNEL_NUM;i++)
      {
         if(g_ifx_dma_int_status&(1<<i) && dma_chan[i].weight>0)
         {
            if(dma_chan[i].weight>weight)
            {
               chan_no=i;
               weight=dma_chan[chan_no].weight;
            }
         }
      }
      if(chan_no>=0){
         if(dma_chan[chan_no].dir==DIR_RX)
            rx_chan_intr_handler(chan_no);
         else tx_chan_intr_handler(chan_no);
      }
      else/*reset all the channels*/
      {
         for(i=0;i<MAX_DMA_CHANNEL_NUM;i++)
         {
            dma_chan[i].weight=dma_chan[i].default_weight;
         }
      }
   }
}

irqreturn_t dma_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
   struct dma_channel_info* pCh;
   int chan_no=0;

   pCh=(struct dma_channel_info*)dev_id;
   chan_no=(int)(pCh-dma_chan);
   if((unsigned)chan_no >= MAX_DMA_CHANNEL_NUM)
   {
      printk("dma_interrupt irq=%d chan_no=%d\n",irq,chan_no);
   }
   g_ifx_dma_int_status|=1<<chan_no;
   disable_irq(irq);

   tasklet_schedule(&dma_tasklet);

   return IRQ_RETVAL(1);
}


struct dma_device_info* dma_device_reserve(char* dev_name)
{
   int i;

   ifx_dma_init();
   for(i=0;i<MAX_DMA_DEVICE_NUM;i++)/*may put some hash function here in the future:)*/
   {
      if(strcmp(dev_name,dma_devs[i].device_name)==0)
      {
         if(dma_devs[i].reserved) return NULL;
         dma_devs[i].reserved=1;
         break;
      }
   }
   if (i == MAX_DMA_DEVICE_NUM) return NULL;
   
   return &dma_devs[i];
}

int dma_device_release(struct dma_device_info* dev)
{
   dev->reserved=0;

   return 0;
}

int dma_device_register(struct dma_device_info* dev)
{
   int port_no=(int)(dev-dma_devs);
   int txbl, rxbl;
   int flag;

   switch(dev->tx_burst_len)
   {
      case 8:  txbl = 3; break;
      case 4:  txbl = 2; break;
      default: txbl = 1;
   }
   switch(dev->rx_burst_len)
   {
      case 8:  rxbl = 3; break;
      case 4:  rxbl = 2; break;
      default: rxbl = 1;
   }

   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_PS=port_no;
   *IFX_DMA_PCTRL = IFX_DMA_PCTRL_TXWGT_VAL(dev->tx_weight)
                  | IFX_DMA_PCTRL_TXENDI_VAL(dev->tx_endianness_mode)
                  | IFX_DMA_PCTRL_RXENDI_VAL(dev->rx_endianness_mode)
                  | IFX_DMA_PCTRL_PDEN_VAL(dev->drop_enable)
                  | IFX_DMA_PCTRL_TXBL_VAL(txbl)
                  | IFX_DMA_PCTRL_RXBL_VAL(rxbl);
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   local_irq_restore(flag);

   return 0;
}

int dma_device_unregister(struct dma_device_info* dev)
{
   int i;
   int port_no=(int)(dev-dma_devs);
   int flag;

   /* flush memcopy module; has no effect for other ports */
   local_irq_save(flag);
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_PS=port_no;
   *IFX_DMA_PCTRL |= IFX_DMA_PCTRL_GPC;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
   local_irq_restore(flag);

   for(i=0;i<dev->max_tx_chan_num;i++)
      reset_chan(dev->tx_chan[i]);

   for(i=0;i<dev->max_rx_chan_num;i++)
      reset_chan(dev->rx_chan[i]);

   return 0;
}


/**
 * Read Packet from DMA Rx channel.
 * The function gets the data from the current rx descriptor assigned
 * to the passed DMA device and passes it back to the caller.
 * The function is called in the context of DMA interrupt.
 * In detail the following actions are done:
 * - get current receive descriptor
 * - allocate memory via allocation callback function
 * - pass data from descriptor to allocated memory
 * - update channel weight
 * - release descriptor
 * - update current descriptor position
 *
 * \param *dma_dev    - pointer to DMA device structure
 * \param **dataptr   - pointer to received data
 * \param **opt
 * \return packet length - length of received data
 * \ingroup Internal
 */
int dma_device_read(struct dma_device_info* dma_dev, u8** dataptr, void** opt)
{
   int flag;
   u8* buf;
   int len;
   int byte_offset=0;
   void* p=NULL;

   struct dma_channel_info* pCh=dma_dev->rx_chan[dma_dev->current_rx_chan];

   struct rx_desc* rx_desc_p;

   /*get the rx data first*/
   rx_desc_p=(struct rx_desc*)pCh->desc_base+pCh->curr_desc;
   buf=(u8*)__va(rx_desc_p->Data_Pointer);
   *(u32*)dataptr=(u32)buf;
   len=rx_desc_p->status.field.data_length;
#ifndef	CONFIG_MIPS_UNCACHED
   dma_cache_inv((unsigned long) buf, len);
#endif //CONFIG_MIPS_UNCACHED
   if(opt)
   {
      *(int*)opt=(int)pCh->opt[pCh->curr_desc];
   }

   /*replace with a new allocated buffer*/
   buf=dma_dev->buffer_alloc(pCh->packet_size,&byte_offset,&p);
   if(buf){
      pCh->opt[pCh->curr_desc]=p;

      wmb();
      rx_desc_p->Data_Pointer=(u32)CPHYSADDR((u32)buf);
      rx_desc_p->status.word =(DMA_OWN<<31)  \
         |((byte_offset)<<23) \
         |pCh->packet_size;

      wmb();
   }
   else
   {
      *(u32*)dataptr = 0;
      if(opt) *(int*)opt = 0;
      //len = 0;
   }
#ifndef DMA_POLLING_MODE
   local_irq_save(flag);
   if (pCh->dur)
   {
      IFX_MBS_GRAB(IFX_MBSR_DMA);
      *IFX_DMA_CS=(int)(pCh-dma_chan);
      /* trigger descriptor fetching by clearing DUR interrupt */
      *IFX_DMA_CIS = IFX_DMA_CIS_DUR;
      IFX_MBS_RELEASE(IFX_MBSR_DMA);
      pCh->dur = 0;
   }
   local_irq_restore(flag);
#endif

   pCh->xfer_cnt++;
   /*increase the curr_desc pointer*/
   pCh->curr_desc++;
   if(pCh->curr_desc==pCh->desc_len) pCh->curr_desc=0;
   /*return the length of the received packet*/
   return len;
}



int dma_device_write(struct dma_device_info* dma_dev, u8* dataptr, int len,void* opt)
{
   int flag;
   u32 byte_offset;
   struct dma_channel_info* pCh;
   int chan_no;
   struct tx_desc* tx_desc_p;
   local_irq_save(flag);

   pCh=dma_dev->tx_chan[dma_dev->current_tx_chan];
   chan_no=(int)(pCh-dma_chan);

   if (pCh->control == IFX_DMA_CH_OFF)
   {
      local_irq_restore(flag);
      printk("%s: dma channel %d not enabled!\n", __FUNCTION__, chan_no);
      return 0;
   }

#ifndef USE_TX_CPT_INT
   tx_desc_p=(struct tx_desc*)pCh->desc_base+pCh->prev_desc;
   while((tx_desc_p->status.word & (DMA_DESC_OWN_DMA|DMA_DESC_CPT_SET)) == DMA_DESC_CPT_SET)
   {

      dma_dev->buffer_free((u8*)__va(tx_desc_p->Data_Pointer),pCh->opt[pCh->prev_desc]);
      memset(tx_desc_p,0,sizeof(struct tx_desc));
      if(dma_dev->intr_handler) dma_dev->intr_handler(dma_dev, TRANSMIT_CPT_INT);
      pCh->weight--;

      pCh->prev_desc=(pCh->prev_desc+1)%(pCh->desc_len);
      tx_desc_p=(struct tx_desc*)pCh->desc_base+pCh->prev_desc;
   }
#endif

   tx_desc_p=(struct tx_desc*)pCh->desc_base+pCh->curr_desc;
   /*Check whether this descriptor is available*/
   if (tx_desc_p->status.word & (DMA_DESC_OWN_DMA | DMA_DESC_CPT_SET))
   {
      /*if not , the tell the upper layer device*/
      dma_dev->intr_handler(dma_dev,TX_BUF_FULL_INT);
      local_irq_restore(flag);
      return 0;
   }
   pCh->opt[pCh->curr_desc]=opt;
   /*byte offset----to adjust the starting address of the data buffer, should be multiple of the burst length.*/
   byte_offset=((u32)CPHYSADDR((u32)dataptr))%((dma_dev->tx_burst_len)*4);
#ifndef	CONFIG_MIPS_UNCACHED
   if(KSEGX(dataptr) == KSEG0)
   {
      dma_cache_wback((unsigned long) dataptr,len);
      wmb();
   }
#endif //CONFIG_MIPS_UNCACHED

   tx_desc_p->Data_Pointer=(u32)CPHYSADDR((u32)dataptr)-byte_offset;
   wmb();
   tx_desc_p->status.word = (DMA_OWN<<31)
      |DMA_DESC_SOP_SET
      |DMA_DESC_EOP_SET
      |((byte_offset)<<23)
      |len;
   wmb();

   if (pCh->xfer_cnt == 0)
   {
      IFX_MBS_GRAB(IFX_MBSR_DMA);
      *IFX_DMA_CS=chan_no;
      *IFX_DMA_CCTRL |= IFX_DMA_CCTRL_ON_OFF;
      IFX_MBS_RELEASE(IFX_MBSR_DMA);
   }
#ifndef DMA_POLLING_MODE
   IFX_MBS_GRAB(IFX_MBSR_DMA);
   *IFX_DMA_CS=chan_no;
   /* trigger descriptor fetching by clearing DUR interrupt */
   *IFX_DMA_CIS = IFX_DMA_CIS_DUR;
   IFX_MBS_RELEASE(IFX_MBSR_DMA);
#endif
   pCh->xfer_cnt++;
   pCh->curr_desc++;
   if(pCh->curr_desc==pCh->desc_len)
      pCh->curr_desc=0;

#if 0
   /*Check whether this descriptor is available*/
   tx_desc_p=(struct tx_desc*)pCh->desc_base+pCh->curr_desc;
   if(tx_desc_p->status.field.OWN==DMA_OWN){
      /*if not , the tell the upper layer device*/
      dma_dev->intr_handler(dma_dev,TX_BUF_FULL_INT);
   }
#endif

   local_irq_restore(flag);
   return len;
}




/**
 * Display descriptor list via proc file
 *
 * \param chan_no   - logical channel number
 * \ingroup Internal
 */
int desc_list_proc_read(char *buf, char **start, off_t offset,
                         int count, int *eof, void *data)
{
    int len=0;
    int i;
    static int chan_no=0;
    u32 *p;

    if ((chan_no == 0) && (offset > count)) {
        *eof = 1;
        return 0;
    }

    if (chan_no != 0) {
        *start=buf;
    }
    else {
        buf=buf+offset;
        *start=buf;
    }

    p=(u32*)dma_chan[chan_no].desc_base;

    if (dma_chan[chan_no].dir == DIR_RX)
        len+=sprintf(buf+len,"channel %d %s Rx descriptor list:\n", chan_no, dma_chan[chan_no].dma_dev->device_name);
    else
        len+=sprintf(buf+len,"channel %d %s Tx descriptor list:\n", chan_no, dma_chan[chan_no].dma_dev->device_name);
    len+=sprintf(buf+len," no  address        data pointer command bits (Own, Complete, SoP, EoP, Offset) \n");
    len+=sprintf(buf+len,"---------------------------------------------------------------------------------\n");
    for(i=0; i<dma_chan[chan_no].desc_len; i++)
    {
        len+=sprintf(buf+len,"%3d  ",i);
        len+=sprintf(buf+len,"0x%08x     ", (u32)(p+(i*2)));
        len+=sprintf(buf+len,"%08x     ", *(p+(i*2+1)));
        len+=sprintf(buf+len,"%08x     ", *(p+(i*2)));

        if(*(p+(i*2)) & 0x80000000)
            len+=sprintf(buf+len,"D ");
        else
            len+=sprintf(buf+len,"C ");
        if(*(p+(i*2)) & 0x40000000)
            len+=sprintf(buf+len,"C ");
        else
            len+=sprintf(buf+len,"c ");
        if(*(p+(i*2)) & 0x20000000)
            len+=sprintf(buf+len,"S ");
        else
            len+=sprintf(buf+len,"s ");
        if(*(p+(i*2)) & 0x10000000)
            len+=sprintf(buf+len,"E ");
        else
            len+=sprintf(buf+len,"e ");

        /* byte offset is different for rx and tx descriptors*/
        if (dma_chan[chan_no].dir == DIR_RX)
        {
            len+=sprintf(buf+len,"%01x ", (*(p+(i*2)) & 0x01800000) >> 23);
        }
        else
        {
            len+=sprintf(buf+len,"%02x ", (*(p+(i*2)) & 0x0F800000) >> 23);
        }

        if( dma_chan[chan_no].curr_desc == i )
            len+=sprintf(buf+len,"<- CURR");

        if( dma_chan[chan_no].prev_desc == i )
            len+=sprintf(buf+len,"<- PREV");

        len+=sprintf(buf+len,"\n");

    }
    len+=sprintf(buf+len,"\n");
    chan_no++;
    if (chan_no > MAX_DMA_CHANNEL_NUM-1) {
        chan_no=0;
    }
    *eof=1;
    return len;
}

/**
 * Displays the weight of all DMA channels via proc file
 *
 *
 *
 * \param *buf
 * \param **start
 * \param offset
 * \param count
 * \param *eof
 * \param *data
 * \return len - amount of bytes written to file
 */
int channel_weight_proc_read(char *buf, char **start, off_t offset,
                         int count, int *eof, void *data)
{

    int i;
    int len=0;
    len+=sprintf(buf+len,"Qos dma channel weight list\n");
    len+=sprintf(buf+len,"channel_num default_weight current_weight    device    Tx/Rx\n");
    len+=sprintf(buf+len,"------------------------------------------------------------\n");
    for(i=0; i<MAX_DMA_CHANNEL_NUM; i++)
    {
        struct dma_channel_info* pCh=&dma_chan[i];

        if ( pCh->dir == DIR_RX) {
            len+=sprintf( buf+len,"     %2d      %08x        %08x      %10s   Rx\n",
                      i, pCh->default_weight, pCh->weight, pCh->dma_dev->device_name);
        }
        else {
            len+=sprintf( buf+len,"     %2d      %08x        %08x      %10s   Tx\n",
                      i, pCh->default_weight, pCh->weight, pCh->dma_dev->device_name);
        }
    }
}

/**
 * Provides DMA Register Content to proc file
 * This function reads the content of general DMA Registers, DMA Channel
 * Registers and DMA Port Registers and performs a structures output to the DMA proc file
 *
 *
 * \param *buf
 * \param **start
 * \param offset
 * \param count
 * \param *eof
 * \param *data
 * \return len - amount of bytes written to file
 */
int dma_register_proc_read(char *buf, char **start, off_t offset,
                         int count, int *eof, void *data)
{
    int len=0;
    int i;
    int limit=count;
    u32 flags=0;
    static int blockcount=0;
    static int channel_no=0;

    if ((blockcount == 0) && (offset > count)) {
        *eof=1;
        return 0;
    }

    switch (blockcount) {
        case 0:
            len+=sprintf(buf+len,"\nGeneral DMA Registers\n");
            len+=sprintf(buf+len,"-----------------------------------------\n");
            len+=sprintf(buf+len, "CLC=        %08x\n", *IFX_DMA_CLC );
            len+=sprintf(buf+len, "ID=         %08x\n", *IFX_DMA_ID );
            len+=sprintf(buf+len, "DMA_CPOLL=  %08x\n", *IFX_DMA_CPOLL );
            len+=sprintf(buf+len, "DMA_CS=     %08x\n", *IFX_DMA_CS );
            len+=sprintf(buf+len, "DMA_PS=     %08x\n", *IFX_DMA_PS );
            len+=sprintf(buf+len, "DMA_IRNEN=  %08x\n", *IFX_DMA_IRNEN );
            len+=sprintf(buf+len, "DMA_IRNCR=  %08x\n", *IFX_DMA_IRNCR );
            len+=sprintf(buf+len, "DMA_IRNICR= %08x\n", *IFX_DMA_IRNICR );
            len+=sprintf(buf+len,"\nDMA Channel Registers\n");
            blockcount = 1;
            return len;
            break;
        case 1:
            /* If we had an overflow start at beginnig of buffer otherwise use offset */
            if (channel_no != 0) {
                *start=buf;
            }
            else {
                buf=buf+offset;
                *start=buf;
            }

            save_and_cli( flags );
            for(i=channel_no;i<MAX_DMA_CHANNEL_NUM;i++)
            {
                struct dma_channel_info* pCh=&dma_chan[i];

                if (len+300 > limit)
                {
                    restore_flags( flags );
                    channel_no = i;
                    blockcount = 1;
                    return len;
                }
                len+=sprintf(buf+len,"-----------------------------------------\n");
                if ( pCh->dir == DIR_RX) {
                    len+=sprintf(buf+len,"Channel %d - Device %s Rx\n", i, pCh->dma_dev->device_name);
                }
                else {
                    len+=sprintf(buf+len,"Channel %d - Device %s Tx\n", i, pCh->dma_dev->device_name);
                }
                *IFX_DMA_CS=i;
                len+=sprintf(buf+len, "DMA_CCTRL=  %08x\n", *IFX_DMA_CCTRL);
                len+=sprintf(buf+len, "DMA_CDBA=   %08x\n", *IFX_DMA_CDBA);
                len+=sprintf(buf+len, "DMA_CIE=    %08x\n", *IFX_DMA_CIE);
                len+=sprintf(buf+len, "DMA_CIS=    %08x\n", *IFX_DMA_CIS);
                len+=sprintf(buf+len, "DMA_CDLEN=  %08x\n", *IFX_DMA_CDLEN);
             }
             restore_flags( flags );
             blockcount = 2;
             channel_no=0;
             return len;
             break;
         case 2:
            *start=buf;
            /*
             * display port dependent registers
             */
            len+=sprintf(buf+len,"\nDMA Port Registers\n");
            len+=sprintf(buf+len,"-----------------------------------------\n");
            save_and_cli( flags );
            for(i=0;i<MAX_DMA_DEVICE_NUM;i++)
            {
                *IFX_DMA_PS=i;
                len+=sprintf( buf+len,"Port %d DMA_PCTRL= %08x\n", i, *IFX_DMA_PCTRL );
            }
            restore_flags( flags );
            blockcount = 0;
            *eof=1;
            return len;
            break;
    }

    blockcount = 0;
    *eof=1;
    return 0;
}

/**
 * Open Method of DMA Device Driver
 * This function increments the device driver's use counter.
 *
 *
 * \param
 * \return
 */
static int dma_open(struct inode * inode, struct file * file)
{
   MOD_INC_USE_COUNT;
   return 0;
}

/**
 * Release Method of DMA Device driver.
 * This function decrements the device driver's use counter.
 *
 *
 * \param
 * \return
 */
static int dma_release(struct inode *inode, struct file *file)
{
   /*release the resources*/
   MOD_DEC_USE_COUNT;
   return 0;
}


/**
 * Ioctl Interface to DMA Module
 *
 * \param  None
 * \return 0  - initialization successful
 *         <0 - failed initialization
 */
static int dma_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
   int result=0;
   /*TODO: add some user controled functions here*/
   return result;
}



static struct file_operations dma_fops = {
owner:       THIS_MODULE,
open:        dma_open,
release:     dma_release,
ioctl:       dma_ioctl,
};


int map_dma_chan(struct dma_channel_info* map)
{
   int i;

   /* assign default values for channel settings */
   for(i=0;i<MAX_DMA_CHANNEL_NUM;i++)
   {
      dma_chan[i].byte_offset=0;
      dma_chan[i].open=&open_chan;
      dma_chan[i].close=&close_chan;
      dma_chan[i].reset=&reset_chan;
      dma_chan[i].enable_irq=&enable_ch_irq;
      dma_chan[i].disable_irq=&disable_ch_irq;
      dma_chan[i].tx_weight=1;
      dma_chan[i].control=0;
      dma_chan[i].default_weight=IFX_DMA_CH_DEFAULT_WEIGHT;
      dma_chan[i].weight=dma_chan[i].default_weight;
      dma_chan[i].curr_desc=0;
      dma_chan[i].prev_desc=0;
   }

   /* assign default values for port settings */
   for(i=0;i<MAX_DMA_DEVICE_NUM;i++)
   {
      dma_devs[i].num_tx_chan=1;/*set default tx channel number to be one*/
      dma_devs[i].num_rx_chan=1;/*set default rx channel number to be one*/
      dma_devs[i].buffer_alloc=&common_buffer_alloc;
      dma_devs[i].buffer_free=&common_buffer_free;
      dma_devs[i].intr_handler=NULL;
      dma_devs[i].tx_burst_len=4;
      dma_devs[i].rx_burst_len=4;
#ifdef CONFIG_CPU_LITTLE_ENDIAN
      dma_devs[i].tx_endianness_mode=0;
      dma_devs[i].rx_endianness_mode=0;
#else
      dma_devs[i].tx_endianness_mode=3;
      dma_devs[i].rx_endianness_mode=3;
#endif
   }

   return IFX_SUCCESS;
}

int dma_chip_init(void)
{
   int i;

#if 0
   *IFX_PMU_PWDCR &=~(1<<IFX_PMU_DMA_SHIFT);/*enable DMA from PMU*/
#endif
#ifdef CONFIG_SVIP
   *IFX_SYS1_CLKENR = IFX_SYS1_CLKENR_DMA;
   wmb();
#endif
   /* reset DMA */
   *IFX_DMA_CTRL = IFX_DMA_CTRL_RST;
   wmb();
   /* disable all the interrupts first */
   *IFX_DMA_IRNEN=0;

#ifdef DMA_POLLING_MODE
   /* enable polling for all channels */
   *IFX_DMA_CPOLL= IFX_DMA_CPOLL_EN | IFX_DMA_CPOLL_CNT_VAL(DMA_POLL_COUNTER);
#endif

   /****************************************************/
   for(i=0;i<MAX_DMA_CHANNEL_NUM;i++)
   {
      disable_ch_irq(&dma_chan[i]);
   }
   return IFX_SUCCESS;
}

int ifx_dma_init(void)
{
   int result=0;
   int i;
   int flag;
   static int ifx_dma_initialized = 0;

   if (ifx_dma_initialized == 1)
      return 0;
   ifx_dma_initialized = 1;

   result = register_chrdev(DMA_MAJOR,"dma-core",&dma_fops);
   if(result){
      IFX_DMA_EMSG("cannot register device dma-core!\n");
      return result;
   }

   dma_chip_init();
   map_dma_chan(dma_chan);
   g_desc_list=(u64*)KSEG1ADDR(__get_free_page(GFP_DMA));
   if (g_desc_list == NULL){
      IFX_DMA_EMSG("no memory for desriptor\n");
      return -ENOMEM;
   }
   memset(g_desc_list,0,PAGE_SIZE);
   for(i=0;i<MAX_DMA_CHANNEL_NUM;i++)
   {
      /* cross-link relative channels of a port to corresponding absolute channels */
      if (dma_chan[i].dir == DIR_RX)
      {
         ((struct dma_device_info*)(dma_chan[i].dma_dev))->
               rx_chan[dma_chan[i].rel_chan_no] = &dma_chan[i];
      }
      else
      {
         ((struct dma_device_info*)(dma_chan[i].dma_dev))->
               tx_chan[dma_chan[i].rel_chan_no] = &dma_chan[i];
      }
      dma_chan[i].abs_chan_no = i;

      dma_chan[i].desc_base=(u32)g_desc_list+i*IFX_DMA_DESCRIPTOR_OFFSET*8;
      dma_chan[i].curr_desc=0;
      dma_chan[i].desc_len=IFX_DMA_DESCRIPTOR_OFFSET;

      local_irq_save(flag);
      IFX_MBS_GRAB(IFX_MBSR_DMA);
      *IFX_DMA_CS=i;
      *IFX_DMA_CDBA=(u32)CPHYSADDR(dma_chan[i].desc_base);
      IFX_MBS_RELEASE(IFX_MBSR_DMA);
      local_irq_restore(flag);
   }

   g_ifx_dma_dir=proc_mkdir("driver/ifx_dma", NULL);

   create_proc_read_entry("dma_register",
      0,
      g_ifx_dma_dir,
      dma_register_proc_read,
      NULL);

   create_proc_read_entry("g_desc_list",
      0,
      g_ifx_dma_dir,
      desc_list_proc_read,
      NULL);

   create_proc_read_entry("channel_weight",
      0,
      g_ifx_dma_dir,
      channel_weight_proc_read,
      NULL);

   printk("IFX DMA engine initialized... \n");

   return 0;
}

/**
 * Cleanup DMA device
 * This function releases all resources used by the DMA device driver on
 * module removal.
 *
 *
 * \param  None
 * \return Nothing
 */
void dma_cleanup(void)
{
   int i;
   unregister_chrdev(DMA_MAJOR,"dma-core");
   free_page(KSEG0ADDR((unsigned long)g_desc_list));
   remove_proc_entry("channel_weight",g_ifx_dma_dir);
   remove_proc_entry("g_desc_list",g_ifx_dma_dir);
   remove_proc_entry("dma_register",g_ifx_dma_dir);
   remove_proc_entry("driver/ifx_dma", NULL);
   /*release the resources*/
   for(i=0;i<MAX_DMA_CHANNEL_NUM;i++)
      free_irq(dma_chan[i].irq,(void*)&dma_interrupt);
}

__initcall(ifx_dma_init);

EXPORT_SYMBOL(dma_device_reserve);
EXPORT_SYMBOL(dma_device_release);
EXPORT_SYMBOL(dma_device_register);
EXPORT_SYMBOL(dma_device_unregister);
EXPORT_SYMBOL(dma_device_read);
EXPORT_SYMBOL(dma_device_write);


MODULE_LICENSE ("GPL");
