/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/*
 * Initial version
 * Gary Jennejohn (C) 2003 <gj@denx.de>
 */
 
/* Group definitions for Doxygen */
/** \defgroup API API-Functions */
/** \defgroup Internal Internally used functions */

/*
 * Routines for generic manipulation of the interrupts found on the 
 * INCA-IP2 boards.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>

#include <asm/bootinfo.h>
#include <asm/irq.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/irq.h>
#include <asm/traps.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#include <linux/tqueue.h>

extern unsigned int incaip2_get_cpu_hz(void);
extern unsigned int incaip2_get_fpi_hz(void);
extern asmlinkage void mipsIRQ(void);
extern asmlinkage unsigned int do_IRQ(int irq, struct pt_regs *regs);
extern asmlinkage void incaip2_timer_interrupt(struct pt_regs *regs);


void incaip2_wdt_monitor_start(void);
void incaip2_wdt_monitor_stop(void);

#if 0
#define DEBUG_INT(x...) printk(x)
#else
#define DEBUG_INT(x...)
#endif

#ifdef CONFIG_KGDB
extern void breakpoint(void);
extern void set_debug_traps(void);
extern int remote_debug;
#endif

/**
 * Disable interrupt line.
 * This function disables the specified interrupt line in the ICU. 
 * The different interrupt modules of the Inca-IP2 are addressed by an offset. 
 * i.e. IM0 = 0,1,2...31, IM1 = 0+32,1+32,2+32...31+32 up to IM5 with an offset 
 * of 160. These offsets are already taken into account by the specific macros
 * defined in asm-mips/incaip2/irq.h 
 *
 * \param irq_nr - This is the interrupt number. (e.g. 129 for IM4-IRL1)
 */
void disable_incaip2_irq(unsigned int irq_nr)
{
	unsigned int flags=0;
	
	save_and_cli(flags);
	DEBUG_INT("disable_incaip_irq: %d ", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0)
	{
		/* access IM0 */
		*INCA_IP2_ICU_IM0_IER &= ~INCA_IP2_ICU_IM0_IR(irq_nr);
	}
	if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0)
	{
		/* access IM1 */
		*INCA_IP2_ICU_IM1_IER &= ~INCA_IP2_ICU_IM1_IR(irq_nr - 32);
	}
	if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0)
	{
		/* access IM2 */
		*INCA_IP2_ICU_IM2_IER &= ~INCA_IP2_ICU_IM2_IR(irq_nr - 64);
	}
	if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0)
	{
		/* access IM3 */
		*INCA_IP2_ICU_IM3_IER &= ~INCA_IP2_ICU_IM3_IR(irq_nr - 96);
	}
	if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0)
	{
		/* access IM4 */
		*INCA_IP2_ICU_IM4_IER &= ~INCA_IP2_ICU_IM4_IR(irq_nr - 128);
	}
	if (irq_nr <= INT_NUM_IM5_IRL31 && irq_nr >= INT_NUM_IM5_IRL0)
	{
		/* access IM5 */
		*INCA_IP2_ICU_IM5_IER &= ~INCA_IP2_ICU_IM5_IR(irq_nr - 160);
	}
	restore_flags(flags);
}

/**
 * Mask and acknowledge interrupt line.
 * This function will mask the specified interrupt line and then acknowledge it.
 * The different interrupt modules of the Inca-IP2 are addressed by an offset. 
 * i.e. IM0 = 0,1,2...31, IM1 = 0+32,1+32,2+32...31+32 up to IM5 with an offset 
 * of 160. These offsets are already taken into account by the specific macros
 * defined in asm-mips/incaip2/irq.h 
 *
 * \param irq_nr - This is the interrupt number. (e.g. 84 for IM2-IRL20)
 */
void mask_and_ack_incaip2_irq(unsigned int irq_nr)
{
	unsigned int flags=0;
	
	save_and_cli(flags);
	DEBUG_INT("mask_and_ack_incaip_irq: %d ", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0)
	{
		/* access IM0 */
		/* mask */
		*INCA_IP2_ICU_IM0_IER &= ~INCA_IP2_ICU_IM0_IR(irq_nr);
		/* ack */
		*INCA_IP2_ICU_IM0_ISR = INCA_IP2_ICU_IM0_IR(irq_nr);
	}
	if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0)
	{
		/* access IM1 */
		/* mask */
		*INCA_IP2_ICU_IM1_IER &= ~INCA_IP2_ICU_IM1_IR(irq_nr - 32);
		/* ack */
		*INCA_IP2_ICU_IM1_ISR = INCA_IP2_ICU_IM1_IR(irq_nr - 32);
	}
	if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0)
	{
		/* access IM2 */
		/* mask */
		*INCA_IP2_ICU_IM2_IER &= ~INCA_IP2_ICU_IM2_IR(irq_nr - 64);
		/* ack */
		*INCA_IP2_ICU_IM2_ISR = INCA_IP2_ICU_IM2_IR(irq_nr - 64);
	}
	if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0)
	{
		/* access IM3 */
		/* mask */
		*INCA_IP2_ICU_IM3_IER &= ~INCA_IP2_ICU_IM3_IR(irq_nr - 96);
		/* ack */
		*INCA_IP2_ICU_IM3_ISR = INCA_IP2_ICU_IM3_IR(irq_nr - 96);
	}
	if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0)
	{
		/* access IM4 */
		/* mask */
		*INCA_IP2_ICU_IM4_IER &= ~INCA_IP2_ICU_IM4_IR(irq_nr - 128);
		/* ack */
		*INCA_IP2_ICU_IM4_ISR = INCA_IP2_ICU_IM4_IR(irq_nr - 128);
	}
	if (irq_nr <= INT_NUM_IM5_IRL31 && irq_nr >= INT_NUM_IM5_IRL0)
	{
		/* access IM5 */
		/* mask */
		*INCA_IP2_ICU_IM5_IER &= ~INCA_IP2_ICU_IM5_IR(irq_nr - 160);
		/* ack */
		*INCA_IP2_ICU_IM5_ISR = INCA_IP2_ICU_IM5_IR(irq_nr - 160);
	}
	restore_flags(flags);
}

/**
 * Acknowledge interrupt line.
 * This function will acknowledge the spefified interrupt line.
 *
 * \param irq_nr - This is the interrupt number. (e.g. 84 for IM2-IRL20)
 */
void ack_incaip2_irq(unsigned int irq_nr)
{
	unsigned int flags=0;
	
	save_and_cli(flags);
	DEBUG_INT("ack_incaip_irq: %d ", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0)
	{
		/* access IM0 */
		/* ack */
		*INCA_IP2_ICU_IM0_ISR = INCA_IP2_ICU_IM0_IR(irq_nr);
	}
	if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0)
	{
		/* access IM1 */
		/* ack */
		*INCA_IP2_ICU_IM1_ISR = INCA_IP2_ICU_IM1_IR(irq_nr - 32);
	}
	if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0)
	{
		/* access IM2 */
		/* ack */
		*INCA_IP2_ICU_IM2_ISR = INCA_IP2_ICU_IM2_IR(irq_nr - 64);
	}
	if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0)
	{
		/* access IM3 */
		/* ack */
		*INCA_IP2_ICU_IM3_ISR = INCA_IP2_ICU_IM3_IR(irq_nr - 96);
	}
	if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0)
	{
		/* access IM4 */
		/* ack */
		*INCA_IP2_ICU_IM4_ISR = INCA_IP2_ICU_IM4_IR(irq_nr - 128);
	}
	if (irq_nr <= INT_NUM_IM5_IRL31 && irq_nr >= INT_NUM_IM5_IRL0)
	{
		/* access IM5 */
		/* ack */
		*INCA_IP2_ICU_IM5_ISR = INCA_IP2_ICU_IM5_IR(irq_nr - 160);
	}
	restore_flags(flags);
}

/**
 * Enable interrupt line.
 * This function enables the specified interrupt line in the ICU. 
 * The different interrupt modules of the Inca-IP2 are addressed by an offset. 
 * i.e. IM0 = 0,1,2...31, IM1 = 0+32,1+32,2+32...31+32 up to IM5 with an offset 
 * of 160. These offsets are already taken into account by the specific macros
 * defined in asm-mips/incaip2/irq.h 
 *
 * \param irq_nr - This is the interrupt number. (e.g. 54 for IM1-IRL22)
 */
void enable_incaip2_irq(unsigned int irq_nr)
{
	unsigned int flags=0;
	
	save_and_cli(flags);
	DEBUG_INT("enable_incaip_irq: %d ", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0)
	{
		/* access IM0 */
		*INCA_IP2_ICU_IM0_IER |= INCA_IP2_ICU_IM0_IR(irq_nr);
	}
	if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0)
	{
		/* access IM1 */
		*INCA_IP2_ICU_IM1_IER |= INCA_IP2_ICU_IM1_IR(irq_nr - 32);
	}
	if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0)
	{
		/* access IM2 */
		*INCA_IP2_ICU_IM2_IER |= INCA_IP2_ICU_IM2_IR(irq_nr - 64);
	}
	if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0)
	{
		/* access IM3 */
		*INCA_IP2_ICU_IM3_IER |= INCA_IP2_ICU_IM3_IR(irq_nr - 96);
	}
	if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0)
	{
		/* access IM4 */
		*INCA_IP2_ICU_IM4_IER |= INCA_IP2_ICU_IM4_IR(irq_nr - 128);
	}
	if (irq_nr <= INT_NUM_IM5_IRL31 && irq_nr >= INT_NUM_IM5_IRL0)
	{
		/* access IM5 */
		*INCA_IP2_ICU_IM5_IER |= INCA_IP2_ICU_IM5_IR(irq_nr - 160);
	}
	restore_flags(flags);
}

/**
 * Startup interrupt.
 * This function just enables the specified interrupt line. 
 *
 * \param irq - This is the interrupt number.
 */
static unsigned int startup_incaip2_irq(unsigned int irq)
{
	enable_incaip2_irq(irq);
	return 0; /* never anything pending */
}

/**
 * Shutdown interrupt.
 * in case of Inca-IP2 shutdown and disable are the same for interrupts.. 
 */
#define shutdown_incaip2_irq	disable_incaip2_irq

/**
 * End interrupt.
 * This function checks whether the interrupt is currently busy executing 
 * another handler and if this is not the case the interrupt is enabled again. 
 *
 * \param irq - This is the interrupt number.
 */
static void end_incaip2_irq(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED|IRQ_INPROGRESS)))
		enable_incaip2_irq(irq);
}

/**
 * Structure with ICU controller functions
 */
static struct hw_interrupt_type incaip2_irq_type = {
	"INCAIP2",
	startup_incaip2_irq,
	shutdown_incaip2_irq,
	enable_incaip2_irq,
	disable_incaip2_irq,
	ack_incaip2_irq,
	end_incaip2_irq,
	NULL
};

/**
 * Find least significant bit.
 * This function searches for the least significant bit in 32bit value.
 *
 * \param  x - The value to be checked.
 * \return Position of the least significant bit (0..31)
 */
static inline int ls1bit32(unsigned int x)
{
	int b = 31, s;

	s = 16; if (x << 16 == 0) s = 0; b -= s; x <<= s;
	s =  8; if (x <<  8 == 0) s = 0; b -= s; x <<= s;
	s =  4; if (x <<  4 == 0) s = 0; b -= s; x <<= s;
	s =  2; if (x <<  2 == 0) s = 0; b -= s; x <<= s;
	s =  1; if (x <<  1 == 0) s = 0; b -= s;

	return b;
}

/**
 * MIPS hardware interrupt #2 handler.
 * This function is called from the generic exception handler after it was 
 * determined which hardware interrupt occured. Hardware interrupt #2 is 
 * associated with IM0.
 *
 * \param regs - Original registers before interrupt context change.
 */
void incaip2_hw0_irqdispatch(struct pt_regs *regs)
{
	unsigned long int_status;
	int irq;

	int_status = *INCA_IP2_ICU_IM0_IOSR;

	/* if int_status == 0, then the interrupt has already been cleared */
	if (int_status == 0) {
		return;
	}

	irq = ls1bit32(int_status);

	DEBUG_INT("incaip2_hw0_irqdispatch: irq=%d\n", irq);
	do_IRQ(irq, regs);

	return;
}

/**
 * MIPS hardware interrupt #3 handler.
 * This function is called from the generic exception handler after it was 
 * determined which hardware interrupt occured. Hardware interrupt #3 is 
 * associated with IM1.
 *
 * \param regs - Original registers before interrupt context change.
 */
void incaip2_hw1_irqdispatch(struct pt_regs *regs)
{
	unsigned long int_status;
	int irq;

	int_status = *INCA_IP2_ICU_IM1_IOSR;

	/* if int_status == 0, then the interrupt has already been cleared */
	if (int_status == 0) {
		return;
	}

	irq = ls1bit32(int_status) + 32;

	DEBUG_INT("incaip2_hw1_irqdispatch: irq=%d\n", irq);
	do_IRQ(irq, regs);

	return;
}

/**
 * MIPS hardware interrupt #4 handler.
 * This function is called from the generic exception handler after it was 
 * determined which hardware interrupt occured. Hardware interrupt #4 is 
 * associated with IM2.
 *
 * \param regs - Original registers before interrupt context change.
 */
void incaip2_hw2_irqdispatch(struct pt_regs *regs)
{
	unsigned long int_status;
	int irq;

	int_status = *INCA_IP2_ICU_IM2_IOSR;

	/* if int_status == 0, then the interrupt has already been cleared */
	if (int_status == 0) {
		return;
	}
	
	irq = ls1bit32(int_status) + 64;

	DEBUG_INT("incaip2_hw2_irqdispatch: irq=%d\n", irq);
	do_IRQ(irq, regs);
	
	return;
}

/**
 * MIPS hardware interrupt #5 handler.
 * This function is called from the generic exception handler after it was 
 * determined which hardware interrupt occured. Hardware interrupt #5 is 
 * associated with IM3.
 *
 * \param regs - Original registers before interrupt context change.
 */
void incaip2_hw3_irqdispatch(struct pt_regs *regs)
{
	unsigned long int_status;
	int irq;

	int_status = *INCA_IP2_ICU_IM3_IOSR;

	/* if int_status == 0, then the interrupt has already been cleared */
	if (int_status == 0) {
		return;
	}
	
	irq = ls1bit32(int_status) + 96;

	DEBUG_INT("incaip2_hw3_irqdispatch: irq=%d\n", irq);
	do_IRQ(irq, regs);
	
	return;
}

/**
 * MIPS hardware interrupt #6 handler.
 * This function is called from the generic exception handler after it was 
 * determined which hardware interrupt occured. Hardware interrupt #6 is 
 * associated with IM4.
 *
 * \param regs - Original registers before interrupt context change.
 */
void incaip2_hw4_irqdispatch(struct pt_regs *regs)
{
	unsigned long int_status;
	int irq;

	int_status = *INCA_IP2_ICU_IM4_IOSR;

	/* if int_status == 0, then the interrupt has already been cleared */
	if (int_status == 0) {
		return;
	}
	
	irq = ls1bit32(int_status) + 128;

	DEBUG_INT("incaip2_hw4_irqdispatch: irq=%d\n", irq);
	do_IRQ(irq, regs);
	
	return;
}

/**
 * MIPS hardware interrupt #7 handler.
 * This function is called from the generic exception handler after it was 
 * determined which hardware interrupt occured. Hardware interrupt #7 is 
 * associated with IM5.
 *
 * \param regs - Original registers before interrupt context change.
 */
void incaip2_hw5_irqdispatch(struct pt_regs *regs)
{
	unsigned long int_status;
	int irq;

	int_status = *INCA_IP2_ICU_IM5_IOSR;

	/* if int_status == 0, then the interrupt has already been cleared */
	if (int_status == 0) {
		return;
	}
	
	if (int_status & INCA_IP2_ICU_IM5_IR(18))
	{
		incaip2_timer_interrupt(regs);
	}
	else
	{
		irq = ls1bit32(int_status) + 160;

		DEBUG_INT("incaip2_hw5_irqdispatch: irq=%d\n", irq);
		do_IRQ(irq, regs);
	}	
	return;
}

/**
 * MIPS bus error handler.
 * This function is called from the generic exception handler after it a 
 * bus error was detected. It will print some register values which might
 * help finding the reason for this exception.
 *
 * \param regs - Original registers before exception context change.
 * \param is_fixup - not used
 * \return Return either MIPS_BE_DISCARD if the error can be ignored or 
 *         MIPS_BE_FATAL if an oops needs to triggered.
 */
int incaip2_be_handler(struct pt_regs *regs, int is_fixup)
{
	u32 reg1,reg2,reg3,reg4,reg5,reg6,reg7,reg8;

	reg1 = *(INCA_IP2_BCU0_ECON);
	reg2 = *(INCA_IP2_BCU0_EADD);
	reg3 = *(INCA_IP2_BCU0_EDAT);
	reg4 = *(INCA_IP2_BCU1_ECON);
	reg5 = *(INCA_IP2_BCU1_EADD);
	reg6 = *(INCA_IP2_BCU1_EDAT);
	reg7 = *(INCA_IP2_MC_ERRCAUSE);
	reg8 = *(INCA_IP2_MC_ERRADDR);
	
	if (reg2 == 0x00000000) {
		if (reg4 == 0x00000000) {
			if (reg8 == 0x00000000) { 
				return MIPS_BE_DISCARD;
			}
		}
	}
	
	printk(KERN_ALERT "BCU0_ECON=0x%08x ", reg1); 
	printk("BCU0_EADD=0x%08x ", reg2); 
	printk("BCU0_EDAT=0x%08x\n", reg3);
	printk(KERN_ALERT "BCU1_ECON=0x%08x ", reg4); 
	printk("BCU1_EADD=0x%08x ", reg5); 
	printk("BCU1_EDAT=0x%08x\n", reg6);
	printk(KERN_ALERT "MC_ECAUSE=0x%08x ", reg7);
	printk("MC_EADDR=0x%08x\n", reg8);

	return MIPS_BE_FATAL;
}

/**
 * Enable external interrupt.
 * This function enables an external interrupt and sets the given mode.
 * valid values for mode are:
 *   - 0 = Interrupt generation disabled
 *   - 1 = Interrupt on rising edge
 *   - 2 = Interrupt on falling edge
 *   - 3 = Interrupt on rising and falling edge
 *   - 5 = Interrupt on high level detection
 *   - 6 = Interrupt on low level detection
 *
 * \param   exint - Number of external interrupt (0-2)
 * \param   mode  - Trigger mode
 * \ingroup API
 */
void incaip2_enable_external_int(u32 exint, u32 mode)
{
	u32 enables;

	/* Set mode in CR register */
	*(INCA_IP2_ICU_EIU_CR) &= ~(0x00000007 << (exint * 4));
	*(INCA_IP2_ICU_EIU_CR) |= ((mode & 0x00000007) << (exint * 4));

	/* Enable interrupt */
#if 0
	*(INCA_IP2_ICU_EIU_IRNEN) |= 0x00000001 << (exint & 0x00000003);
#else
	/* EIU doesn't behave as expected... */
	enables = (*(INCA_IP2_ICU_EIU_IRNEN) | (1 << exint)) & 0x7;
	*(INCA_IP2_ICU_EIU_IRNEN) = enables | (enables << 3);
	*(INCA_IP2_ICU_EIU_IRNICR) = enables;
#endif

}

/**
 * Disable external interrupt.
 * This function disables an external interrupt and sets mode to 0x00.
 *
 * \param   exint - Number of external interrupt (0-2)
 * \ingroup API
 */
void incaip2_disable_external_int(u32 exint)
{
	u32 enables;

	/* Set mode in CR register to 0x00 */
	*(INCA_IP2_ICU_EIU_CR) &= ~(0x00000007 << (exint * 4));
	/* Disable interrupt */
#if 0    
	*(INCA_IP2_ICU_EIU_IRNEN) &= 0xfffffffe << (exint & 0x00000003);
#else
	/* EIU doesn't behave as expected... */
	enables = (*(INCA_IP2_ICU_EIU_IRNEN) & ~(1 << exint)) & 0x7;
	*(INCA_IP2_ICU_EIU_IRNEN) = enables | (enables << 3);
	*(INCA_IP2_ICU_EIU_IRNICR) = enables;
#endif
}

/**
 * Get level of external interrupt line.
 * This function returns the level of an external interrupt line.
 *
 * \param   exint - Number of external interrupt (0-2)
 * \return  Level of external interrupt line. 
 * \ingroup API
 */
int incaip2_external_int_level(u32 exint)
{
	u32 mode;
	int exin_level;

	/* Check current interrupt mode */
	mode = (*(INCA_IP2_ICU_EIU_CR) >> (exint * 4)) & 0x7;

	/* Disable interrupt (if enabled) */
	if (mode)
		incaip2_disable_external_int(exint);

	switch (exint) 
	{
	case 0: 
		*INCA_IP2_GPIO_ALTSEL1 |= 0x00040000;
		__asm__("sync");
		exin_level = (*INCA_IP2_GPIO_PIN >> 25) & 1;
		*INCA_IP2_GPIO_ALTSEL1 &= ~0x00040000;
		break;
	case 1:
		*INCA_IP2_GPIO_ALTSEL1 = (*INCA_IP2_GPIO_ALTSEL1 & ~0x00003000) | 0x00001000;
		__asm__("sync");
		exin_level=(*INCA_IP2_GPIO_PIN >> 22) & 1;
		*INCA_IP2_GPIO_ALTSEL1 = (*INCA_IP2_GPIO_ALTSEL1 & ~0x00003000) | 0x00002000;
		break;
	case 2:
		*INCA_IP2_GPIO_ALTSEL1 = (*INCA_IP2_GPIO_ALTSEL1 & ~0x00000C00) | 0x00000400;
		__asm__("sync");
		exin_level=(*INCA_IP2_GPIO_PIN >> 21) & 1;
		*INCA_IP2_GPIO_ALTSEL1 = (*INCA_IP2_GPIO_ALTSEL1 & ~0x00000C00) | 0x00000800;
		break;
	default:
		exin_level = -1;
	}

	if (mode)
		incaip2_enable_external_int(exint, mode);

	return exin_level;
}


/**
 * Hardware interrupt initialization.
 * This function will initialize the handler structures for all ICU interrupt 
 * lines and the bus error exception.
 * All interrupt will be disabled in the ICU but all MIPS hardware interrupts
 * will be enabled. 
 */
void __init init_IRQ(void)
{
	int i;

	DEBUG_INT("init_IRQ ");

	board_be_handler = &incaip2_be_handler;
	
	init_generic_irq();

	/* mask all interrupt sources */
	*INCA_IP2_ICU_IM0_IER = 0;
	*INCA_IP2_ICU_IM1_IER = 0;
	*INCA_IP2_ICU_IM2_IER = 0;
	*INCA_IP2_ICU_IM3_IER = 0;
	*INCA_IP2_ICU_IM4_IER = 0;
	*INCA_IP2_ICU_IM5_IER = 0;

	/* Now safe to set the exception vector. */
	set_except_vector(0, mipsIRQ);

	for (i = 0; i <= INT_NUM_IM5_IRL31; i++) {
		irq_desc[i].status	= IRQ_DISABLED;
		irq_desc[i].action	= 0;
		irq_desc[i].depth	= 1;
		irq_desc[i].handler	= &incaip2_irq_type;
	}
	set_c0_status(IE_IRQ0 | IE_IRQ1 | IE_IRQ2 | IE_IRQ3 | IE_IRQ4 | IE_IRQ5);
#ifdef CONFIG_KGDB
	if (remote_debug) {
		set_debug_traps();
		breakpoint();
	}
#endif
}

#ifdef CONFIG_INCAIP2_WATCHDOG
/***********************************************************
 * Inca-IP2 Watchdog timer 
 ***********************************************************/

#ifdef CONFIG_INCAIP2_WATCHDOG_HEARTBEAT
#define INCA_IP2_TSF                  (KSEG1+0x1E001000)
#define INCA_IP2_TSF_TSF_CONF         ((volatile u32*)(INCA_IP2_TSF + 0x0000))
#define INCA_IP2_TSF_LEDMUX_CONF      ((volatile u32*)(INCA_IP2_TSF + 0x0018))
#define INCA_IP2_TSF_LED_REG		  ((volatile u32*)(INCA_IP2_TSF + 0x001C))
#define INCA_IP2_TSF_TSF_CONF_LEDEN   (1 << 1)
#endif

#define INCA_IP2_WDT_PW1 0x000000BE /**< First password for access */
#define INCA_IP2_WDT_PW2 0x000000DC /**< Second password for access */
#define INCA_IP2_WDT_CLKDIV0_VAL 1
#define INCA_IP2_WDT_CLKDIV1_VAL 64
#define INCA_IP2_WDT_CLKDIV2_VAL 4096
#define INCA_IP2_WDT_CLKDIV3_VAL 262144
#define INCA_IP2_WDT_CLKDIV0 0
#define INCA_IP2_WDT_CLKDIV1 1
#define INCA_IP2_WDT_CLKDIV2 2
#define INCA_IP2_WDT_CLKDIV3 3

#define WDT_MBX_SIZE_IN_WORDS 112

typedef struct
{
    volatile u32      MPS_BOOT_RVEC;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_NVEC;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_EVEC;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_CAUSE;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_EEPC;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_EPC;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_SR;   /**< CPU reset vector */
    volatile u32      MPS_BOOT_STATUS;   /**< CPU reset vector */
} wdt_mps_boot_cfg_reg;

typedef struct
{
    volatile u32      MBX[WDT_MBX_SIZE_IN_WORDS];
    wdt_mps_boot_cfg_reg  MBX_CPU0_BOOT_CFG;    /**< CPU0 Boot Configuration */
    wdt_mps_boot_cfg_reg  MBX_CPU1_BOOT_CFG;    /**< CPU1 Boot Configuration */
} wdt_mps_mbx;

u32 incaip2_wdt_frequency;

int incaip2_wdt_target_pid;

/**
 * Watchdog disable.
 * This function disables the Inca-IP2 watchdog. 
 */
void incaip2_wdt_disable(void)
{
	u32 wdt_cr=0;

	/* Write first part of password access */
	*INCA_IP2_WDT_BIU_WDT_CR |= INCA_IP2_WDT_BIU_WDT_CR_PW_SET(INCA_IP2_WDT_PW1) ;
	wdt_cr = *INCA_IP2_WDT_BIU_WDT_CR;
	/* Disable the watchdog in second password access (GEN=0) */
    wdt_cr &= ~(INCA_IP2_WDT_BIU_WDT_CR_GEN);
	*INCA_IP2_WDT_BIU_WDT_CR = INCA_IP2_WDT_BIU_WDT_CR_PW_SET(INCA_IP2_WDT_PW2) | wdt_cr; 

}

/**
 * Watchdog enable.
 * This function enables the Inca-IP2 watchdog. 
 */
void incaip2_wdt_enable(void)
{
	/* Write first part of password access */
	*INCA_IP2_WDT_BIU_WDT_CR |= INCA_IP2_WDT_BIU_WDT_CR_PW_SET(INCA_IP2_WDT_PW1) | INCA_IP2_WDT_BIU_WDT_CR_GEN;
	 /* Enable the watchdog in second password access (GEN=1) */
	*INCA_IP2_WDT_BIU_WDT_CR |= INCA_IP2_WDT_BIU_WDT_CR_PW_SET(INCA_IP2_WDT_PW2) | INCA_IP2_WDT_BIU_WDT_CR_GEN;
}

/**
 * Set watchdog reload value.
 * This function will set the reload value of the watchdog to a given delay in.
 * seconds.
 *
 * \param Delay - Delay in seconds
 * \return always 0
 */
int incaip2_wdt_set_reload_value(u32 Delay)
{
	u32 wdt_cr=0;
	u32 wdt_reload=0;
	u32 wdt_mask=0;

	/* Write first part of password access */
	*INCA_IP2_WDT_BIU_WDT_CR = INCA_IP2_WDT_BIU_WDT_CR_PW_SET(INCA_IP2_WDT_PW1) | INCA_IP2_WDT_BIU_WDT_CR_GEN;
	
	/*
	 * Calculate reload value and build register setting.
	 * Setting clk divider value to 3 gives us a divider of 262144 as our 
	 * fbi0 bus runns at 131MHz our counter decrements by 500 counts a second or 
	 * every 2ms.
	 */	
	wdt_reload = Delay * incaip2_wdt_frequency;
	
	wdt_cr = *INCA_IP2_WDT_BIU_WDT_CR;

	wdt_mask = ( ~INCA_IP2_WDT_BIU_WDT_CR_PW_SET(0xff) & ~INCA_IP2_WDT_BIU_WDT_CR_RELOAD_SET(0xffff) & ~INCA_IP2_WDT_BIU_WDT_CR_CLKDIV_SET(0x3) );

	wdt_cr &= wdt_mask;

	/*
	 * We implicitly enable the watchdog here so in reality
	 * there is no need for a seperate watchdog enable function
	 */
	wdt_cr |= ( INCA_IP2_WDT_BIU_WDT_CR_PW_SET(INCA_IP2_WDT_PW2) | 
	            INCA_IP2_WDT_BIU_WDT_CR_RELOAD_SET(wdt_reload) |
		    INCA_IP2_WDT_BIU_WDT_CR_CLKDIV_SET(INCA_IP2_WDT_CLKDIV3) | INCA_IP2_WDT_BIU_WDT_CR_GEN);

	/*
	 * Set the pre-warn timeout to be 0x1000 or 4096 counts (about 8 seconds)
	 * we should really chech that reload value is greater than the pre-warn
	 * and that it is greater than 32 seconds otherwise the pre-warn will go off
	 * every time before we reload the counter
	 */

	wdt_cr |= INCA_IP2_WDT_BIU_WDT_CR_PWL_SET(0x3);

	/* Set reload value in second password access */
	*INCA_IP2_WDT_BIU_WDT_CR = wdt_cr;
	return( 0 );
}

struct tq_struct incaip2_wdt_timer;
long incaip2_wdt_timer_count=0;
u32 incaip2_wdt_run = 0;
#ifdef CONFIG_INCAIP2_WATCHDOG_TIMEOUT
u32 incaip2_wdt_timeout = CONFIG_INCAIP2_WATCHDOG_TIMEOUT;
#else
u32 incaip2_wdt_timeout = 4;
#endif

/**
 * Watchdog timer task.
 * This function is called from the timer task every timer tick. It will check
 * whether the timer counter has reached a fourth of the timeout value. If this
 * is true the reload value will be set again with the original value.
 *
 * \param data - not used
 */
void incaip2_wdt_timer_task(void* data)
{

	if( incaip2_wdt_run == 0 )
	{
		/* We are being stopped so do nothing */
		return;
	}

	incaip2_wdt_timer_count++;
	if (incaip2_wdt_timer_count > ((incaip2_wdt_timeout*HZ) >> 2)) {
		incaip2_wdt_timer_count=0;
		incaip2_wdt_set_reload_value(incaip2_wdt_timeout);
#ifdef CONFIG_INCAIP2_WATCHDOG_HEARTBEAT
    	*INCA_IP2_TSF_LED_REG ^= 5;
#endif
			
	}
	queue_task(&incaip2_wdt_timer,&tq_timer);
}

#ifdef CONFIG_PROC_FS


int incaip2_wdt_proc_reset_reason(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    u32 reset_reason;
    char *p;
	int len;
	int code;

    /* This is a copy of the RCU Status register saved by the bootrom */
    wdt_mps_mbx *ifx_mps_reg = (wdt_mps_mbx*) INCA_IP2_MPS_SRAM;
    reset_reason = ifx_mps_reg->MBX_CPU0_BOOT_CFG.MPS_BOOT_SR;

    switch( reset_reason )
    {
    	case 0x9f300000:	p = "Power On";			code = 1;	break;
		case 0x99800000:	p = "Watchdog Fired";	code = 2;	break;
		case 0x98800000:	p = "Watchdog Fired";	code = 2;	break;
		case 0x98000000:	p = "Software";			code = 3;	break;
		default:			p = (reset_reason & 0x98000000)? "Unusual" : "Unknown - u-boot upgrade required"; code = 0; break;
    }

    len = sprintf(buf,"RR:%08x (%d)%s\n",reset_reason,code,p);
   	*eof = 1;
   	return len;
}

int incaip2_wdt_proc_read(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
   	int len=0;

    len+=sprintf(buf+len,"Inca-IP2 Watchdog\n");
        
   	len+=sprintf(buf+len,"WDT_CR=0x%08x\n", *INCA_IP2_WDT_BIU_WDT_CR);	
   	len+=sprintf(buf+len,"WDT_SR=0x%08x\n", *INCA_IP2_WDT_BIU_WDT_SR);
   	len+=sprintf(buf+len,"RCU_WD=0x%08x\n", *INCA_IP2_RCU_WDTRCR);

  	*eof = 1;
   	return len;
}

int incaip2_wdt_proc_write(struct file *file, const char *buffer,unsigned long count, void *data)
{
	char buff[50];
	int len;

	len = ( count < sizeof(buff) )? count : sizeof(buff);

	memcpy( buff, buffer,len);
	buff[len-1] = 0;

	if( strcmp("start monitor",buff) == 0 )
	{
		incaip2_wdt_monitor_start();
	}

	else if( strcmp("stop monitor",buff) == 0 )
	{
		incaip2_wdt_monitor_stop();
	}

	else if( strcmp("disable watchdog",buff) == 0 )
	{
		incaip2_wdt_disable();
	}

	else if( strcmp("kick watchdog",buff) == 0 )
	{
		incaip2_wdt_set_reload_value(incaip2_wdt_timeout);
	}

	else if( strncmp("set pid ", buff, 8) == 0 )
	{
		sscanf(buff, "set pid %d", &incaip2_wdt_target_pid);
		printk("Set hw wd target pid to %d\n",incaip2_wdt_target_pid);
	}

	else
	{
		printk("HDV (%s) unknown command\n",buff);
	}
	return count;


}
#endif /* CONFIG_PROC_FS */

/**
 * Watchdog start.
 * This function will setup the watchdog timer task and then enable th watchdog.
 */
void incaip2_wdt_start(void)
{
#ifdef CONFIG_INCAIP2_WATCHDOG_HEARTBEAT
    unsigned long flags;
#endif
    u32 version;
    wdt_mps_mbx *ifx_mps_reg = (wdt_mps_mbx*) INCA_IP2_MPS_SRAM;
 
   printk("Enabling watchdog with timeout of %d second(s)\n",incaip2_wdt_timeout);
    *INCA_IP2_PMS_PMS_GEN |= (1<<24);   
	if (incaip2_wdt_set_reload_value(incaip2_wdt_timeout) != 0) {
		printk("Error! Could not set reload value!\n");
	}
	else {
		incaip2_wdt_timer.routine = incaip2_wdt_timer_task;
		incaip2_wdt_timer.data = NULL;
		incaip2_wdt_frequency = (incaip2_get_fbs0clk() / INCA_IP2_WDT_CLKDIV3_VAL);
		queue_task(&incaip2_wdt_timer,&tq_timer);
		incaip2_wdt_enable();

#ifdef CONFIG_INCAIP2_WATCHDOG_HEARTBEAT
		/* enable the clock of the LED module in the Power Management Unit */
		save_flags(flags); cli();
		*INCA_IP2_PMS_PMS_GEN |= (1<<9);   
		*INCA_IP2_TSF_TSF_CONF |= INCA_IP2_TSF_TSF_CONF_LEDEN;
		restore_flags(flags);
		*INCA_IP2_TSF_LEDMUX_CONF = 0;
#endif
	}
    
    /* Set correct reset vector in bootrom */
    version=INCA_IP2_MPS_CHIPID_VERSION_GET(*INCA_IP2_MPS_CHIPID);
    if ( version == 1 ) { 
        ifx_mps_reg->MBX_CPU0_BOOT_CFG.MPS_BOOT_RVEC = (u32)0xBFC00550;
    }
    if ( version == 2 ) { 
        ifx_mps_reg->MBX_CPU0_BOOT_CFG.MPS_BOOT_RVEC = (u32)0xBFC0056C;
    }

    /* Enable reset of all modules if watchdog kicks in */ 
//    *INCA_IP2_RCU_WDTRCR =  INCA_IP2_RCU_WDT0RCR_CPU0 | INCA_IP2_RCU_WDT0RCR_CPU1 |
//                            INCA_IP2_RCU_WDT0RCR_CPUSU | INCA_IP2_RCU_WDT0RCR_HRST;
    *INCA_IP2_RCU_WDTRCR =  INCA_IP2_RCU_WDT0RCR_CPU0 | INCA_IP2_RCU_WDT0RCR_CPUSU | INCA_IP2_RCU_WDT0RCR_HRST;
    
#ifdef CONFIG_PROC_FS
    create_proc_read_entry("driver/ifx_wdt", 0, NULL, incaip2_wdt_proc_read , NULL);    
#endif    
}

void wdt_prewarn_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	struct task_struct* target_task_ptr;

	u32 wdt_pwl_value = INCA_IP2_WDT_BIU_WDT_CR_PWL_GET(*INCA_IP2_WDT_BIU_WDT_CR);

	wdt_pwl_value ^= 3;
	wdt_pwl_value = 0x1000 << wdt_pwl_value;

	printk(KERN_ALERT "\nWDT ISR pre-warn Fired!!! Phone will reset in %d seconds\n",wdt_pwl_value/incaip2_wdt_frequency);

	free_irq(INT_NUM_IM5_IRL0, 0);
	force_sig(SIGTRAP, current);

	target_task_ptr = find_task_by_pid(incaip2_wdt_target_pid);
	if(target_task_ptr != NULL)
	{
		send_sig(SIGTRAP, target_task_ptr, 0);
	}
}

void wdt_bad_access_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	printk(KERN_ALERT "\nWDT ISR bad access Fired\n");
	free_irq(INT_NUM_IM5_IRL1, 0);
}

int wdt_isr_init(void)
{
	int err;
	if( (err = request_irq(INT_NUM_IM5_IRL0, wdt_prewarn_isr, SA_INTERRUPT,"pre-warn-wdt", NULL)) != 0 )
	{
		printk(KERN_NOTICE "\nWDT ISR Init pre-warn failed\n");
	}

	if( (err = request_irq(INT_NUM_IM5_IRL1, wdt_bad_access_isr, SA_INTERRUPT,"bad-access-wdt", NULL)) != 0 )
	{
		printk(KERN_NOTICE "\nWDT ISR Init bad access failed\n");
	}

	return( err );
}

/**
 * Watchdog monitor start.
 * This function will start the watchdog timer task and enable the watchdog.
 */
void incaip2_wdt_monitor_start(void)
{
    printk(KERN_INFO "Enabling watchdog with timeout of %d second(s)\n",incaip2_wdt_timeout);

	incaip2_wdt_set_reload_value(incaip2_wdt_timeout);
	incaip2_wdt_run = 1;
	queue_task(&incaip2_wdt_timer,&tq_timer);
}

/**
 * Watchdog monitor stop.
 * This function will set a flag to stop the watchdog timer task.
 */
void incaip2_wdt_monitor_stop(void)
{
	printk(KERN_INFO "Letting watchdog timeout in %d second(s)\n",incaip2_wdt_timeout);

	/*
	 * kick the watchdog one more time so that we can be sure
	 * how long we have until the watchdog fires
	 */
	incaip2_wdt_set_reload_value(incaip2_wdt_timeout);
	incaip2_wdt_run = 0;
}

void incaip2_wdt_init(void)
{
	struct proc_dir_entry *entry;
	u32 reset_reason;
	char *p;
	int code;

	/* This is a copy of the RCU Status register saved by the bootrom */
	wdt_mps_mbx *ifx_mps_reg = (wdt_mps_mbx*) INCA_IP2_MPS_SRAM;
	reset_reason = ifx_mps_reg->MBX_CPU0_BOOT_CFG.MPS_BOOT_SR;

	incaip2_wdt_target_pid = 0;

	switch( reset_reason )
	{
		case 0x9f300000:	p = "Power On";			code = 1;	break;
		case 0x99800000:	p = "Watchdog Fired";	code = 2;	break;
		case 0x98800000:	p = "Watchdog Fired";	code = 2;	break;
		case 0x98000000:	p = "Software";			code = 3;	break;
		default:			p = (reset_reason & 0x98000000)? "Unusual" : "Unknown - u-boot upgrade required"; code = 0; break;
	}

	printk(KERN_INFO "RR:%08x (%d)%s\n",reset_reason,code,p);

	/* Patch up the reset vector */
	ifx_mps_reg->MBX_CPU0_BOOT_CFG.MPS_BOOT_RVEC = (u32)0xBFC0056C;

	*INCA_IP2_PMS_PMS_GEN |= (1<<24);


	incaip2_wdt_timer.routine = incaip2_wdt_timer_task;
	incaip2_wdt_timer.data = NULL;

	incaip2_wdt_frequency = (incaip2_get_fbs0clk() / INCA_IP2_WDT_CLKDIV3_VAL);

	wdt_isr_init();

	/* Enable reset of all modules if watchdog kicks in */
	*INCA_IP2_RCU_WDTRCR =  INCA_IP2_RCU_WDT0RCR_CPU0 | INCA_IP2_RCU_WDT0RCR_CPU1 | INCA_IP2_RCU_WDT0RCR_CPUSU | INCA_IP2_RCU_WDT0RCR_HRST;

#ifdef CONFIG_PROC_FS
	entry = create_proc_entry("driver/ifx_wdt", 0, 0);
	entry->read_proc  = incaip2_wdt_proc_read;
	entry->write_proc = incaip2_wdt_proc_write;
	create_proc_read_entry("reset-reason", 0, NULL, incaip2_wdt_proc_reset_reason , NULL);
#endif
}

EXPORT_SYMBOL(incaip2_wdt_enable);
EXPORT_SYMBOL(incaip2_wdt_disable);
EXPORT_SYMBOL(incaip2_wdt_monitor_start);
EXPORT_SYMBOL(incaip2_wdt_monitor_stop);

#endif /* CONFIG_INCAIP2_WATCHDOG */

EXPORT_SYMBOL(incaip2_disable_external_int);
EXPORT_SYMBOL(incaip2_enable_external_int);
EXPORT_SYMBOL(incaip2_external_int_level);
EXPORT_SYMBOL(enable_incaip2_irq);
EXPORT_SYMBOL(disable_incaip2_irq);
