/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
 
/*
 * port.c
 *
 * Inca-IP2 port driver
 *
 */

/* Group definitions for Doxygen */
/** \addtogroup API API-Functions */
/** \addtogroup Internal Internally used functions */
/** \addtogroup Obsolete Will be removed soon */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <asm/semaphore.h>
#include <asm/uaccess.h>

#if defined(CONFIG_INCAIP2)
#include <asm/incaip2/mux.h>
#include <asm/incaip2/port.h>

extern int incaip2_get_cpu_ver(void);

#endif

#include "port_defs.h"


static int port_major = 0;
static int ifx_port_pin_usage[MAX_PORTS][PINS_PER_PORT]; // Map for pin usage

static struct semaphore port_sem[MAX_PORTS];
static struct semaphore port_ioctl_sem;
#ifndef MODULE
static int inca_port_use_count = 0;
#endif

/***************************************************************************/
/**                                                                       **/
/** SSC chip select                                                       **/
/**                                                                       **/
/***************************************************************************/

/**
 * Set direction bit for chip select.
 * This function sets the direction bit for the SSC chip select depending on
 * the master/slave setting.
 *
 * \param   ssc_port SSC Port number (0 or 1)
 * \param   cs       Chip select (0 or 1)
 * \param   master   Master/Slave select (1=master, 0=slave)
 * \return  -EINVAL  Invalid ssc_port, cs or master value provided.
 * \return  0        OK
 * \ingroup API  
 */
int ifx_port_ssc_cs_set(int ssc_port, int cs, int master)
{
    u32 reg;
    int port=0;
    int pin=0;

    if ((ssc_port != 0) && (ssc_port != 1)) return -EINVAL;
    if ((cs != 0) && (cs != 1)) return -EINVAL;
    if ((master != 0) && (master != 1)) return -EINVAL;
    
    if (ssc_port == 0) {
        if (cs == 0) pin = 15;
        else pin = 14;
    }
    else {
        if (cs == 0) pin = 4;
        else pin = 3;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    reg &= ~(1 << pin);
    reg |= (master << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);

    return OK;
}

/**
 * Set direction bit for SSC pins.
 * This function sets the direction bits for the SSC clock and data pins 
 * depending on the master/slave setting.
 *
 * \param   ssc_port SSC Port number (0 or 1)
 * \param   master   Master/Slave select (1=master, 0=slave)
 * \return  -EINVAL  Invalid ssc_port or master value provided.
 * \return  0        OK
 * \ingroup API  
 */
int ifx_port_ssc_master_set(int ssc_port, int master)
{
    u32 reg;
    int port=0;
    int pin=0;

    if ((ssc_port != 0) && (ssc_port != 1)) return -EINVAL;
    if ((master != 0) && (master != 1)) return -EINVAL;
    
    if (ssc_port == 0) {
        /* Set clock */
        pin = 16;
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        reg |= (master << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);

        /* Set MTSR */
        pin = 17;
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        reg |= (master << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);

        /* Set MRST */
        pin = 18;
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        reg |= ((!master) << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    }
    else {
        /* Set clock */
        pin = 5;
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        reg |= (master << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);

        /* Set MTSR */
        pin = 6;
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        reg |= (master << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);

        /* Set MRST */
        pin = 7;
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        reg |= ((!master) << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    }
    
    return OK;
}

/***************************************************************************/
/**                                                                       **/
/** Port usage                                                            **/
/**                                                                       **/
/***************************************************************************/

/**
 * Reserve port pin for usage.
 * This function reserves a given pin for usage by the given module.
 *
 * \param port      Port number
 * \param pin       Pin to be reserved
 * \param module_id Module ID to identify the owner of a pin
 * \return -EBUSY   Pin allready used.
 * \return -EINVAL  Invalid port or pin provided.
 * \return 0        OK, pin reserved
 * \ingroup API  
 */
int ifx_port_reserve_pin(int port, int pin, int module_id)
{

    if (port < 0 || pin <0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;
        
    if (ifx_port_pin_usage[port][pin] == 0) {
        ifx_port_pin_usage[port][pin] = module_id;
    }
    else {
        return -EBUSY;
    }
    
    return OK;
}

/**
 * Free port pin.
 * This functions frees a port pin and thus clears the entry in the usage map.
 *
 * \param port      Port number
 * \param pin       Pin to be released
 * \param module_id Module ID to identify the owner of a pin
 * \return -EBUSY   Pin used by another module than the given ID.
 * \return -EINVAL  Invalid port or pin provided.
 * \return 0        OK, pin freed
 * \ingroup API  
 */
int ifx_port_free_pin(int port, int pin, int module_id)
{

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;
    ifx_port_pin_usage[port][pin] = 0;

    return OK;
}

/**
 * Enable Open Drain for given pin.
 * This function sets Open Drain mode for the given pin.
 *
 * \param port          Port number
 * \param pin           Pin to be used
 * \param module_id     Module ID to identify the owner of a pin
 * \return -EBUSY       Pin used by another module than the given ID.
 * \return -EINVAL      Invalid port or pin provided.
 * \return -ERESTARTSYS Another instance is using the driver.
 * \return 0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_set_open_drain(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_od_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & OD_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }
    
    PORT_READ_REG(ifx_port_bases[port] + PORT_OD_REG,reg);
    reg |= (1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_OD_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Disable Open Drain for given pin.
 * This function clears Open Drain mode for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_clear_open_drain(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_od_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & OD_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_OD_REG,reg);
    reg &= ~(1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_OD_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Enable PUDSEL for given pin.
 * This function sets the PUDSEL bit for the given pin.
 *
 * \param port          Port number
 * \param pin           Pin to be used
 * \param module_id     Module ID to identify the owner of a pin
 * \return -EBUSY       Pin used by another module than the given ID.
 * \return -EINVAL      Invalid port or pin provided.
 * \return -ERESTARTSYS Another instance is using the driver.
 * \return 0            OK, pin freed
 * \ingroup Obsolete  
 */
int ifx_port_set_pudsel(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_pudsel_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & PUDSEL_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_PUDSEL_REG,reg);
    reg |= (1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_PUDSEL_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Clear PUDSEL bit for given pin.
 * This function clears the PUDSEL bit for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup Obsolete  
 */
int ifx_port_clear_pudsel(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_pudsel_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & PUDSEL_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_PUDSEL_REG,reg);
    reg &= ~(1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_PUDSEL_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Set PUDEN bit for given pin.
 * This function sets the PUDEN bit for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_set_puden(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_puden_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & PUDEN_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_PUDEN_REG,reg);
    reg |= (1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_PUDEN_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Disable PUDEN bit for given pin.
 * This function clears the PUDEN bit for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_clear_puden(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_puden_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & PUDEN_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_PUDEN_REG,reg);
    reg &= ~(1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_PUDEN_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Enable Schmitt Trigger for given pin.
 * This function enables the Schmitt Trigger for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_set_stoff(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_stoff_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & STOFF_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_STOFF_REG,reg);
    reg |= (1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_STOFF_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Disable Schmitt Trigger for given pin.
 * This function disables the Schmitt Trigger for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_clear_stoff(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_stoff_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & STOFF_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_STOFF_REG,reg);
    reg &= ~(1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_STOFF_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Set direction to output for given pin.
 * This function sets the direction for the given pin to output.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_set_dir_out(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_dir_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & DIR_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    reg |= (1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Set direction to input for given pin.
 * This function sets the direction for the given pin to input.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_set_dir_in(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_dir_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has been registerd */
    if ((ifx_port_reg_avail[port] & DIR_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

#if defined(CONFIG_INCAIP2)

    if (incaip2_get_cpu_ver() < 2) {
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg &= ~(1 << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    }
    else {
        /* For A24 and A25 the meaning of DIR bits are inverted */
        if ( (pin == 8) || (pin == 9) ) {
            PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
            reg |= (1 << pin);
            PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        }
        else {
            PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
            reg &= ~(1 << pin);
            PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        }
    }

#else
    PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    reg &= ~(1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
#endif

    up (&port_sem[port]);

    return OK;
}

/**
 * Set output bit for given pin.
 * This function sets the output bit to 1 for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_set_output(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin <0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;
    
    /* Check for valid pin */
    if ((ifx_port_out_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & OUT_REG_MASK) == 0 ) return -EINVAL;

     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

#if defined(CONFIG_INCAIP2)

    if (incaip2_get_cpu_ver() < 2) {
        PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        reg |= (1 << pin);
        PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
    }
    else {
        /* For A24 and A25 the meaning of DIR bits are inverted */
        if ( (pin == 8) || (pin == 9) ) {
            PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
            reg &= ~(1 << pin);
            PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        }
        else {
            PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
            reg |= (1 << pin);
            PORT_WRITE_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
        }
    }

#else
    PORT_READ_REG(ifx_port_bases[port] + PORT_OUT_REG,reg);
    reg |= (1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_OUT_REG,reg);
#endif

    up (&port_sem[port]);

    return OK;
}

/**
 * Clear output bit for given pin.
 * This function clears the output bit for the given pin.
 *
 * \param   port         Port number
 * \param   pin          Pin to be used
 * \param   module_id    Module ID to identify the owner of a pin
 * \return  -EBUSY       Pin used by another module than the given ID.
 * \return  -EINVAL      Invalid port or pin provided.
 * \return  -ERESTARTSYS Another instance is using the driver.
 * \return  0            OK, pin freed
 * \ingroup API  
 */
int ifx_port_clear_output(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_out_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & OUT_REG_MASK) == 0 ) return -EINVAL;
    
     if (down_interruptible(&port_sem[port])) {
        return -ERESTARTSYS;
    }

    PORT_READ_REG(ifx_port_bases[port] + PORT_OUT_REG,reg);
    reg &= ~(1 << pin);
    PORT_WRITE_REG(ifx_port_bases[port] + PORT_OUT_REG,reg);

    up (&port_sem[port]);

    return OK;
}

/**
 * Get input bit for given pin.
 * This function gets the value of the given pin.
 *
 * \param   port      Port number
 * \param   pin       Pin to be used
 * \param   module_id Module ID to identify the owner of a pin
 * \return  -EBUSY    Pin used by another module than the given ID.
 * \return  -EINVAL   Invalid port or pin provided.
 * \return  0         OK, pin freed
 * \ingroup API  
 */
int ifx_port_get_input(int port, int pin, int module_id)
{
    u32 reg;

    if (port < 0 || pin < 0) return -EINVAL;
    if (port > MAX_PORTS || pin > PINS_PER_PORT) return -EINVAL;

    if (ifx_port_pin_usage[port][pin] != module_id) return -EBUSY;

    /* Check for valid pin */
    if ((ifx_port_in_pins[port] & (1 << pin)) == 0 ) return -EINVAL;
    
    /* Check if port has register */
    if ((ifx_port_reg_avail[port] & IN_REG_MASK) == 0 ) return -EINVAL;

    PORT_READ_REG(ifx_port_bases[port] + PORT_IN_REG,reg);
    reg &= (1 << pin);
    if (reg == 0x00) return 0;
    else return 1;
}

/**
 * Create proc file output.
 * This function creates the output for the port proc file, by reading
 * all appropriate registers and displaying the content as a table.
 *
 * \param   buf      Buffer to write the string to
 * \param   start    not used (Linux internal)
 * \param   offset   not used (Linux internal)
 * \param   count    not used (Linux internal)
 * \param   eof      Set to 1 when all data is stored in buffer
 * \param   data     not used (Linux internal)
 * \return  len      Lenght of data in buffer 
 * \ingroup Internal  
 */
int ifx_port_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    long    len = 0;
    int     t=0;
    int  port=0;
    u32 bit=0;
    u32 reg=0;

    len = sprintf(buf,"\nIFX Port Settings\n");

    len += sprintf(buf+len,"         3         2         1         0\n");
    len += sprintf(buf+len,"        10987654321098765432109876543210\n");
    len += sprintf(buf+len,"----------------------------------------\n");

    for (port = 0; port<MAX_PORTS; port++) {

        if ((ifx_port_reg_avail[port] & OUT_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-OUT: ",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_OUT_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");
        
        if ((ifx_port_reg_avail[port] & IN_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-IN:  ",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_IN_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");
        
        if ((ifx_port_reg_avail[port] & DIR_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-DIR: ",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_DIR_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");
        
/*
        if ((ifx_port_reg_avail[port] & STOFF_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-STO: ",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_STOFF_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");

        if ((ifx_port_reg_avail[port] & OD_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-OD:  ",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_OD_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");
*/
    
        if ((ifx_port_reg_avail[port] & PUDEN_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-PUDE:",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_PUDEN_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");

/*
        if ((ifx_port_reg_avail[port] & PUDSEL_REG_MASK) != 0 ) {
            len += sprintf(buf+len,"P%d-PUDS:",port);
            PORT_READ_REG(ifx_port_bases[port] + PORT_PUDSEL_REG,reg);
            bit = 0x80000000;
            for (t=0;t<32;t++) {
                if ((reg & bit) > 0) len += sprintf(buf+len,"X");
                else len += sprintf(buf+len," ");
                bit = bit >> 1;
            }
        }
        len += sprintf(buf+len,"\n");
*/

    }
    
    len += sprintf(buf+len,"\n");

    *eof = 1;   // No more data available
    return len;
}

/**
 * Open port device.
 * This function is called when the port device is opend and it will increase the
 * usage counter.
 *
 * \param   inode   Inode of device 
 * \param   filep   File structure of device
 * \return  OK      OK, device opened
 * \ingroup API  
 */
static int ifx_port_open(struct inode *inode, struct file *filep)
{
#ifdef MODULE
         /* increment module use counter */
         MOD_INC_USE_COUNT;
#else
        inca_port_use_count++;
#endif

    return OK;
}

/**
 * Release port device.
 * This function is called when the port device is closed and it will decrease the
 * usage counter.
 *
 * \param   inode   Inode of device 
 * \param   filelp  File structure of device
 * \return  0       OK, device opened
 * \ingroup API  
 */
static int ifx_port_release(struct inode *inode, struct file *filelp)
{
#ifdef MODULE
       /* increment module use counter */
       MOD_DEC_USE_COUNT;
#else
        inca_port_use_count--;
#endif
    
    return OK;
}

/**
 * Port driver IOCTL handler.
 * The following IOCTLs are supported for the port device. (All command use the arg 
 * parameter, which is an ifx_port_ioctl_parm struct defining port, pin and value to use.)
 * - #IFX_PORT_IOCPUDEN
 * - #IFX_PORT_IOCDIR
 * - #IFX_PORT_IOCOUTPUT
 * - #IFX_PORT_IOCINPUT
 * - #IFX_PORT_IOCOD
 * - #IFX_PORT_IOCSTOFF
 *
 * \param   inode        Inode of device 
 * \param   filp         File structure of device
 * \param   cmd          IOCTL command 
 * \param   arg          Argument for some IOCTL commands
 * \return  0            OK
 * \return  -EINVAL      An invalid command was specified
 * \return  -EFAULT      An error occured on accessing the argument
 * \return  -ERESTARTSYS The port driver is currently busy.
 * \ingroup API  
 */
static int ifx_port_ioctl(struct inode *inode, struct file *filp, 
                    unsigned int cmd, unsigned long arg) 
{
    int ret = 0;
    volatile struct ifx_port_ioctl_parm parm;
    
    if (_IOC_TYPE(cmd) != IFX_PORT_IOC_MAGIC) return -EINVAL;
    
    if (_IOC_DIR(cmd) & _IOC_WRITE) {
        if (!access_ok(VERIFY_READ, arg, sizeof(struct ifx_port_ioctl_parm))) 
            return -EFAULT;
        ret=copy_from_user((void*)&parm,(void*)arg,sizeof(struct ifx_port_ioctl_parm));
    }
    if (_IOC_DIR(cmd) & _IOC_READ) {
        if (!access_ok(VERIFY_WRITE, arg, sizeof(struct ifx_port_ioctl_parm))) 
            return -EFAULT;
    }

    if (down_interruptible(&port_ioctl_sem)) {
        return -ERESTARTSYS;
    }
    
    switch (cmd) {
        case IFX_PORT_IOCOD:
            if (parm.value == 0x00) {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_clear_open_drain);
            }
            else {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_open_drain);
            }
            break;
/*
        case IFX_PORT_IOCPUDSEL:
            if (parm.value == 0x00) {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_clear_pudsel);
            }
            else {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_pudsel);
            }
            break;
*/
        case IFX_PORT_IOCPUDEN:
            if (parm.value == 0x00) {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_clear_puden);
            }
            else {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_puden);
            }
            break;
        case IFX_PORT_IOCSTOFF:
            if (parm.value == 0x00) {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_clear_stoff);
            }
            else {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_stoff);
            }
            break;
        case IFX_PORT_IOCDIR:
            if (parm.value == 0x00) {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_dir_in);
            }
            else {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_dir_out);
            }
            break;
        case IFX_PORT_IOCOUTPUT:
            if (parm.value == 0x00) {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_clear_output);
            }
            else {
                PORT_IOC_CALL(ret,parm.port,parm.pin,ifx_port_set_output);
            }
            break;
        case IFX_PORT_IOCINPUT:
            ret=ifx_port_reserve_pin(parm.port,parm.pin,PORT_MODULE_ID); 
            if (ret == 0) 
                parm.value=ifx_port_get_input(parm.port,parm.pin,PORT_MODULE_ID); 
            ret=ifx_port_free_pin(parm.port,parm.pin,PORT_MODULE_ID); 
            copy_to_user((void*)arg,(void*)&parm,sizeof(struct ifx_port_ioctl_parm));
            break;
        default:
            ret = -EINVAL;
    }
    
    up( &port_ioctl_sem );

    return ret;
}

static struct file_operations port_fops = {
    open    : ifx_port_open,
    release : ifx_port_release,
    ioctl   : ifx_port_ioctl
};

/**
 * Initialize port structures.
 * This function initializes the internal data structures of the driver and 
 * will create the proc file entry and device.
 *
 * \return  0 OK, pin freed
 * \ingroup Internal  
 */
int __init ifx_port_init(void)
{
    int t = 0;
    int i = 0;
    int err = 0;
    u32 pins = 0;

    printk("IFX Port Initialization\n");

    sema_init(&port_ioctl_sem,1);
    for (i = 0; i < MAX_PORTS; i++ ) {
        sema_init(&port_sem[i],1);
    }
    
    for (t = 0; t<MAX_PORTS; t++) {
        for (i = 0;i<PINS_PER_PORT;i++) ifx_port_pin_usage[t][i] = -1;
    }

    /* Check ports for available pins */
    for (t = 0; t<MAX_PORTS; t++) {
        pins = IFX_GPIO_PINS(t);
        for (i = 0;i<PINS_PER_PORT;i++) {
            if ((pins & 0x00000001) > 0) 
                ifx_port_pin_usage[t][i] = 0;
            pins = pins >> 1;
        }
    }
    
    /* register port device */
    err = register_chrdev(0,"ifx-port", &port_fops);
    if (err > 0) port_major = err;
    else {
        printk("ifx-port: Error! Could not register port device. #%d\n",err);
        return err;
    }
    
    /* Create proc file */ 
    create_proc_read_entry("driver/ifx_port", 0, NULL, ifx_port_read_procmem, NULL);

    return OK;
}

/**
 * Remove port driver.
 * This function frees all allocated resources.
 *
 * \return  0 OK
 * \ingroup Internal  
 */
static void __exit ifx_port_exit(void)
{
    remove_proc_entry("driver/ifx_port", 0);
    unregister_chrdev(port_major, "ifx-port");

    printk("IFX Port driver removed\n");

    return;
}

module_init(ifx_port_init);
module_exit(ifx_port_exit);

EXPORT_SYMBOL(ifx_port_reserve_pin);
EXPORT_SYMBOL(ifx_port_free_pin);
EXPORT_SYMBOL(ifx_port_set_open_drain);
EXPORT_SYMBOL(ifx_port_clear_open_drain);
EXPORT_SYMBOL(ifx_port_set_pudsel);
EXPORT_SYMBOL(ifx_port_clear_pudsel);
EXPORT_SYMBOL(ifx_port_set_puden);
EXPORT_SYMBOL(ifx_port_clear_puden);
EXPORT_SYMBOL(ifx_port_set_stoff);
EXPORT_SYMBOL(ifx_port_clear_stoff);
EXPORT_SYMBOL(ifx_port_set_dir_out);
EXPORT_SYMBOL(ifx_port_set_dir_in);
EXPORT_SYMBOL(ifx_port_set_output);
EXPORT_SYMBOL(ifx_port_clear_output);
EXPORT_SYMBOL(ifx_port_get_input);

