/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */

#include <asm/types.h>
#include <asm/addrspace.h>

#include <asm/incaip2/incaip2.h>
#include "struct.h"
#include "const.h"
#include "cos.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"

/**
 * Assign CoS to ingress priority.
 * Set the CoS (class of service) for the Ingress Priority. 
 *
 * \param   chipId           Chip Id
 * \param   IngressPriority  IngressPriority(range is 0-7)
 * \param   CoS              CoS(class of service) (range is 0 - 3)
 * \return  PR_ERR_NONE      - Success                 
 * \return  PR_ERR_PRIORITY  - If the IngressPriority is not in the range of 0 to 7
 * \return  PR_ERR_CoS       - If the CoS is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR  - Base address is a null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_map_ingress_priority_CoS(u8 chipId,u32 IngressPriority, u32 CoS)
{
/* Define a pointer to Priority to COS conversion register */
	pri2_cos_reg_t* ptr_pri2cos_reg;
	
	u32 temp1=0;
	u32* ptr_reg1;
	s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

	if(IngressPriority > MAX_PRIORITY)
		return PR_ERR_PRIORITY;

	if(CoS > MAX_CoS)
		return PR_ERR_CoS;

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif

	ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + PRI2_COS_REG); 

	ptr_pri2cos_reg = (pri2_cos_reg_t*)&temp1;


	if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
    		return retflag;


	PR_Reg_Read(ptr_reg1,temp1);  
		
	
	switch(IngressPriority)
	{
	
	case 0:
		ptr_pri2cos_reg -> Pc0 = CoS;
		break; 	

	case 1:
		ptr_pri2cos_reg -> Pc1 = CoS;
		break;
	case 2:
		ptr_pri2cos_reg -> Pc2 = CoS;
		break;
	case 3:
		ptr_pri2cos_reg -> Pc3 = CoS;
		break;
	case 4:
		ptr_pri2cos_reg -> Pc4 = CoS;
		break;
	case 5:
		ptr_pri2cos_reg -> Pc5 = CoS;
		break;
	case 6:
		ptr_pri2cos_reg -> Pc6 = CoS;
		break;
	case 7:
		ptr_pri2cos_reg -> Pc7 = CoS;
		break;
	default:
		return PR_ERR_PRIORITY;


	}
	
	PR_Reg_Write(ptr_reg1,temp1);


	if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
    		return retflag;

	
	return PR_ERR_NONE ;
}

/**
 * Get ingress CoS assigned to priority.
 * Get the CoS (class of service) for the Ingress Priority. 
 *
 * \param   chipId           Chip Id
 * \param   IngressPriority  IngressPriority(range is 0-7)
 * \param   CoSPtr           Pointer to CoS(class of service) (range is 0 - 3)
 * \return  PR_ERR_NONE      - Success                 
 * \return  PR_ERR_PRIORITY  - If the IngressPriority is not in the range of 0 to 7
 * \return  PR_ERR_NULL_PTR  - Base address is a null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_ingress_priority_CoS( u8 chipId, u32 IngressPriority, u32* CoSPtr )
{

/* Define a pointer to Priority to COS conversion register */
	pri2_cos_reg_t* ptr_pri2cos_reg;
	
	u32 temp1=0;
	u32* ptr_reg1;	

#ifdef PARAMETER_CHECK

	if(IngressPriority > MAX_PRIORITY)
		return PR_ERR_PRIORITY;

	if( CoSPtr == NULL)
		return PR_ERR_NULL_PTR;
	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif

	ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + PRI2_COS_REG);	

	ptr_pri2cos_reg = (pri2_cos_reg_t*)&temp1;

	PR_Reg_Read(ptr_reg1,temp1); 
		
	
	switch(IngressPriority)
	{
	
	case 0:
		*CoSPtr =ptr_pri2cos_reg -> Pc0;
		break; 	

	case 1:
		*CoSPtr =ptr_pri2cos_reg -> Pc1;
		break;
	case 2:
		*CoSPtr =ptr_pri2cos_reg -> Pc2;
		break;
	case 3:
		*CoSPtr =ptr_pri2cos_reg -> Pc3;
		break;
	case 4:
		*CoSPtr =ptr_pri2cos_reg -> Pc4;
		break;
	case 5:
		*CoSPtr =ptr_pri2cos_reg -> Pc5;
		break;
	case 6:
		*CoSPtr =ptr_pri2cos_reg -> Pc6;
		break;
	case 7:
		*CoSPtr =ptr_pri2cos_reg -> Pc7;
		break;

	default:
		return PR_ERR_PRIORITY;
	}
	
	return PR_ERR_NONE ;
}

/**
 * Assign egress priority to CoS.
 * Set the Egress Priority for the CoS (class of service).
 *
 * \param   chipId           Chip Id
 * \param   CoS              CoS(class of service) (range is 0 - 3)
 * \param   EgressPriority   EgressPriority(range is 0-7)
 * \return  PR_ERR_NONE      - Success
 * \return  PR_ERR_CoS       - If the CoS is not in the range of 0 to 3
 * \return  PR_ERR_PRIORITY  - If the EgressPriority is not in the range of 0 to 7
 * \return  PR_ERR_NULL_PTR  - Base address is a null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_map_CoS_egress_priority(u8 chipId,u32 CoS, u32 EgressPriority)
{
	
/* Define a pointer to Class of service selection register */
	cos_sel_reg_t* ptr_cos_sel_reg;
	
	u32 temp1 =0;
	u32* ptr_reg1;
	s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

	if(CoS > MAX_CoS)
		return PR_ERR_CoS;
	
	if(EgressPriority > MAX_PRIORITY)
		return PR_ERR_PRIORITY;
	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif

	ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + COS_SEL_REG);

	ptr_cos_sel_reg = (cos_sel_reg_t*)&temp1;


	if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
    		return retflag;

	PR_Reg_Read(ptr_reg1,temp1);
		
	
	switch(CoS)
	{
	
	case 0:
		ptr_cos_sel_reg -> Cp0 = EgressPriority;
		break; 	

	case 1:
		ptr_cos_sel_reg -> Cp1 = EgressPriority;
		break;
	case 2:
		ptr_cos_sel_reg -> Cp2 = EgressPriority;
		break;
	case 3:
		ptr_cos_sel_reg -> Cp3 = EgressPriority;
		break;
	default:
		return PR_ERR_CoS;
	}
	
	PR_Reg_Write(ptr_reg1,temp1);



	if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
    		return retflag;

	
	return PR_ERR_NONE ;	
}

/**
 * Get egress priority assigned to CoS.
 * Get the Egress Priority for the CoS (class of service).
 *
 * \param   chipId            Chip Id
 * \param   CoS               COS queues(range is 0-3)
 * \param   EgressPriorityPtr Pointer to EgressPriority(range is 0-7)
 * \return  PR_ERR_NONE       - Success
 * \return  PR_ERR_CoS        - If the CoS is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR   - Base address is a null pointer
 * \return  PR_ERR_CHIP_ID    - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_CoS_egress_priority(u8 chipId,u32 CoS, u32* EgressPriorityPtr)
{

/* Define a pointer to Class of service selection register */
	cos_sel_reg_t* ptr_cos_sel_reg;
	
	u32 temp1 =0;
	u32* ptr_reg1;

#ifdef PARAMETER_CHECK	
	if(CoS > MAX_CoS)
		return PR_ERR_CoS;

	if(EgressPriorityPtr == NULL)
		return PR_ERR_NULL_PTR;

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif
	
	ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + COS_SEL_REG);
	
	ptr_cos_sel_reg = (cos_sel_reg_t*)&temp1;	

	PR_Reg_Read(ptr_reg1,temp1);
	

	switch(CoS)
	{
	
	case 0:
		*EgressPriorityPtr = ptr_cos_sel_reg -> Cp0;
		break; 	

	case 1:
		*EgressPriorityPtr = ptr_cos_sel_reg -> Cp1;
		break;
	case 2:
		*EgressPriorityPtr = ptr_cos_sel_reg -> Cp2;
		break;
	case 3:
		*EgressPriorityPtr = ptr_cos_sel_reg -> Cp3;
		break;
	default:
		return PR_ERR_CoS;
	}	

	return PR_ERR_NONE ;

}

/**
 * Set CoS reduction value for CoS.
 * Set the CoS Reduction value for the CoS(class of service).
 *
 * \param   chipId           Chip Id
 * \param   CoS              COS queues(range is 0-3)
 * \param   RedCoS           Reduced CoS(class of service) (range is 0 - 3)
 * \return  PR_ERR_NONE      - Success
 * \return  PR_ERR_REDU_CoS  - If the Reduced CoS not in the range of 0 to 3
 * \return  PR_ERR_CoS       - If the CoS is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR  - Base address is a null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_map_CoS_reduction(u8 chipId,u32 CoS, u32 RedCoS)
{	

/* Define a pointer to Class of service selection register */
	cos_sel_reg_t* ptr_cos_sel_reg;
	
	u32 temp1 =0;
	u32* ptr_reg1;
	s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
	
	if(CoS > MAX_CoS)
		return PR_ERR_CoS;

	if(RedCoS > MAX_RedCoS)
		return PR_ERR_REDU_CoS;

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif

	ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + COS_SEL_REG);
	ptr_cos_sel_reg = (cos_sel_reg_t*)&temp1;
	
	if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
    		return retflag;

	PR_Reg_Read(ptr_reg1,temp1);
	
	switch(CoS)
	{
		case 0:
		    ptr_cos_sel_reg -> Cr0 = RedCoS;
			break; 	

		case 1:
			ptr_cos_sel_reg -> Cr1 = RedCoS;
			break;
            
		case 2:
			ptr_cos_sel_reg -> Cr2 = RedCoS;
			break;
            
		case 3:
			ptr_cos_sel_reg -> Cr3 = RedCoS;
			break;
            
		default:
			return PR_ERR_CoS;
	}
	
	PR_Reg_Write(ptr_reg1,temp1);

	if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
    		return retflag;

	return PR_ERR_NONE ;
}

/**
 * Get CoS reduction value for CoS.
 * Get the CoS Reduction value for the CoS(class of service).
 *
 * \param   chipId           Chip Id
 * \param   CoS              COS queues(range is 0-3)
 * \param   RedCoSPtr        Pointer to Reduced CoS (range is 0 - 3)
 * \return  PR_ERR_NONE      - Success
 * \return  PR_ERR_CoS       - If the CoS is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR  - Base address is a null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_CoS_reduction(u8 chipId,u32 CoS, u32* RedCoSPtr)
{

/* Define a pointer to Class of service selection register */
	cos_sel_reg_t* ptr_cos_sel_reg;
	
	u32 temp1 =0;
	u32* ptr_reg1;

#ifdef PARAMETER_CHECK
	
	if(CoS > MAX_CoS)
		return PR_ERR_CoS;

	if(RedCoSPtr == NULL)
		return PR_ERR_NULL_PTR;
    
	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;
#endif

	ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + COS_SEL_REG);
	
	ptr_cos_sel_reg = (cos_sel_reg_t*)&temp1;

	PR_Reg_Read(ptr_reg1,temp1);
		
	
	switch(CoS)
	{
	
	case 0:
		*RedCoSPtr = ptr_cos_sel_reg -> Cr0;
		break; 	

	case 1:
		*RedCoSPtr = ptr_cos_sel_reg -> Cr1;
		break;
	case 2:
		*RedCoSPtr = ptr_cos_sel_reg -> Cr2;
		break;
	case 3:
		*RedCoSPtr = ptr_cos_sel_reg -> Cr3;
		break;
	default:
		return PR_ERR_CoS;
	}
	return PR_ERR_NONE ;
}
