/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */
/** \addtogroup Obsolete Will be removed soon */

#include <linux/kernel.h>
#include <asm/types.h>
#include <asm/addrspace.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/ifx_types.h>
#include "struct.h"
#include "const.h"
#include "vlan.h"
#include "vlan_prv.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"

extern s32 ifx_set_port_lock(u8 chipId,u32 portId, BOOL lock);
extern s32 ifx_set_SYM_VLAN(u8 chipId,u32 portId, BOOL IngressFilter);

/**
 * Set VlanAware Flag.
 * Enable or disable VLAN Awareness of the switch.
 *
 * \param   chipId                 Chip Id
 * \param   VlanAwareFlag          VLAN Awareness Flag
 *
 * \return  PR_ERR_NONE            - Success 
 * \return  PR_ERR_CHIP_ID         - Chip Id invalid
 * \return  PR_ERR_NULL_PTR        - base address is null pointer
 * \ingroup Internal
 */
s32 ifx_set_VLAN_aware( u8 chipId, BOOL VlanAwareFlag )
{
    u32 Temp=0;
    u32* ptr_stack_port;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    
    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;    
    
#endif

    ptr_stack_port = (u32*)(Chip_Base_Address[chipId] + STACK_PORT_REG);
    
    /* 
     * read current STACK_PORT_REG register content, modify VLAN awareness bit 
     * acc. to passed parameter VlanAwareFlag and write new configuration 
     * to STACK_PORT_REG
     */    
    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;  

    PR_Reg_Read(ptr_stack_port,Temp);
    Temp = Temp & STACK_PORT_VLAN_AWARE_MASK;
    Temp |= (VlanAwareFlag << VLAN_AWRFLG_POS); 
    PR_Reg_Write(ptr_stack_port,Temp);

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;
    
    return PR_ERR_NONE ;

}

/**
 * Get VlanAware Flag.
 * Get the VlanAwareness Status of the switch.
 *
 * \param   chipId           Chip Id
 * \param   VlanAwareFlagPtr Pointer to VLAN Awareness status flag
 * \return  PR_ERR_NONE      - Success 
 * \return  PR_ERR_NULL_PTR  - VlanAwareFlag or base address is null pointer 
 * \return  PR_ERR_CHIP_ID   - Chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_VLAN_aware( u8 chipId, BOOL* VlanAwareFlagPtr )
{   
    u32* ptr_stack_port;
    u32 Temp=0;

#ifdef PARAMETER_CHECK
    if(VlanAwareFlagPtr == NULL)
            return PR_ERR_NULL_PTR;
    
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_stack_port = (u32*)(Chip_Base_Address[chipId] + STACK_PORT_REG);

    /* 
     * read current STACK_PORT_REG register content, extract VLAN awareness bit 
     * and pass it to output parameter
     */        
    PR_Reg_Read(ptr_stack_port,Temp);
    Temp = Temp & (~STACK_PORT_VLAN_AWARE_MASK);
    *VlanAwareFlagPtr = (Temp !=0) ? TRUE:FALSE;

    return PR_ERR_NONE ;
    
}

#if 0
/**
 * Set port state for VLAN ID.
 * Set the port state for the individual port in the vlan table for the VID 
 * corresponding to vlanId if the port is member of it.
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portId               PortId(range is 0-2)
 * \param   portState            Port state
 *                                 - 0x1 Disabled   
 *                                 - 0x2 Blocking 
 *                                 - 0x3 Listening
 *                                 - 0x4 Learning
 *                                 - 0x5 Forward
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the Port Id is not in the range of 0 to 2
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0 to 4094
 * \return  PR_ERR_PORTSTATE     - If the Portstate is not in the range of 1 to 5
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_TIME_OUT      - Time out 
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Obsolete
 */
s32 ifx_set_PVST_port_state(u8 chipId,u32 vlanId, u32 portId, u32 portState)
{   
    
    u32 Temp =0;
    s32  i=0;
    u32 Vlan_Mask=1;
    u32 shift_count =0;
    u32 vid =0;
    s32 retflag =0; 
    BOOL IngressFilter = TRUE;
    
/* Define a pointer to a structure & assign the base address of 
   VLAN table command register */
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;

/* Define a pointer to VLAN Table Data register */
    u32* ptr_vlan_tbl_data;

    volatile u32 temp1 = 0;
    u32* ptr_reg1;


#ifdef PARAMETER_CHECK  
    if(portId > MAX_PORTID) 
        return PR_ERR_PORTID;

    if(portState > MAX_PORTSTATE)
        return PR_ERR_PORTSTATE;

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_vlan_tbl_data = (u32*)(Chip_Base_Address[chipId] + VLAN_TBL_DATA_REG);

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + VLAN_TBL_CMD_REG);

    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    
/* VLANID -> VID */ 
    retflag = ifx_get_VLAN_idx(chipId, vlanId, &vid);

    if(retflag == PR_ERR_VLANID)
        return PR_ERR_VLANID;

    if(retflag == PR_ERR_VLAN_NOTCONFG)
        return PR_ERR_VLAN_NOTCONFG;            /* vlanid is not configured  */

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;
    
    /* Write portstate to software table,PR_TABLE_PORTSTATE[1024][51] */
    if(PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,portId)].vlanmember == 0)
        return PR_ERR_NONE;
    
    PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,portId)].port_state = portState;

    if((PR_CHIP_MAJR_VER_ID[chipId] == MAJOR_ID_1) && (PR_CHIP_MINR_VER_ID[chipId] == MINOR_ID_3) )
    {

        /* Set the Symmetric Vlan bit to 1 for the specified port  */
        ifx_set_SYM_VLAN(chipId,portId,IngressFilter);
        switch(portState)
        {
            case IFX_PORT_DISABLED:
                    portState = 0x0;
                    break;

            case IFX_PORT_BLOCKING:
                    portState = 0x1;
                    break;
                    
            case IFX_PORT_LISTENING:
                    portState = 0x1;
                    break;

            case IFX_PORT_LEARNING:
                {
                    if(PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,portId)].vlanmember == 1)
                        portState = 0x2;
                    else
                        portState = 0x1;
                    break;
                }

            case IFX_PORT_FORWARDING:
                {               
                    if(PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,portId)].vlanmember == 1)
                        portState = 0x3;
                    else
                        portState = 0x1;
                    break;
                }
            default:
                        return PR_ERR_PORTSTATE;
            
        }
            
    }
    else
    {
        if(PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,portId)].vlanmember == 0)
            return PR_ERR_NONE;

/* Added this part of the code on 6/3/03 to over come the Port state dependency Hard ware problem  */
/* Disabled(1) - Set port state to 2(10) */
/* Blocking(2) & Listening(3) - Set port state to 2(10) */
/* Learning(4) - Set port state to 2(10) */
/* Forwarding(5) - Set port state to 3(11) */
            
        portState = (portState < IFX_PORT_FORWARDING) ? 2 : 3;

    }


/* Configuration of VLAN Table Command register for reading port state */
/* VLANID(9:0) - VID(0 -> 1023) */  
        ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    
/* Selection(29:27) - Select one of 32 bit sections of the VLAN Table entry */
/*  000 -> Bridge_State for ports 0 to 15(2 bits per port)
    001 -> Bridge_State for ports 16 to 31
    010 -> Bridge_State for ports 32 to 47
    011 -> Bridge_State for ports 48 to 50                                   */
 
   printk("%s: No select field in ptr_vlan_tbl_cmd structure!\n", __FUNCTION__);
/*
        ptr_vlan_tbl_cmd -> Select = (portId >> BRIDGE_STATE_POS) & BRIDGE_STATE_MASK;
*/
        
/* Read/Write(30) : 1-Read */
    
        ptr_vlan_tbl_cmd -> Read_write = READPR;

/* Request/Acknowledge(31): 1-Request */
            
        ptr_vlan_tbl_cmd -> Req_ack = REQ;

/* Added MACRO for indirect register read 7/7/2004 */

        PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);

/* Generate the mask depending upon the PortId */   
        Vlan_Mask = 3;
        shift_count = (portId & 0xf) << 1; /* (portId % 16)*2; */   
        Vlan_Mask <<= shift_count;
        Vlan_Mask = ~ Vlan_Mask;
    
        Temp &= Vlan_Mask;
    
    
        Temp |= (portState << shift_count); 


/* Configuration of VLAN Table Command register for writing */      
/* Read/Write(30) : 0-Write */
        ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    
/* Request/Acknowledge(31): 1-Request */
        ptr_vlan_tbl_cmd -> Req_ack = REQ;


/* Added MACRO for indirect register write 7/7/2004 */
        PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);



    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;


    return PR_ERR_NONE ;
}

/**
 * Get port state for VLAN ID.
 * Get the port state for the individual port from the vlan table for the VID 
 * corresponding to vlanId.
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portId               PortId(range is 0-2)
 * \param   portStatePtr         Port state
 *                                 - 0x1 Disabled   
 *                                 - 0x2 Blocking 
 *                                 - 0x3 Listening
 *                                 - 0x4 Learning
 *                                 - 0x5 Forward
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the Port Id is not in the range of 0 to 2
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0-4094
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_NULL_PTR      - portStatePtr is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Obsolete
 */
s32 ifx_get_PVST_port_state(u8 chipId,u32 vlanId, u32 portId, u32* portStatePtr)
{
    u32 vid =0;
    s32 retflag =0;


#ifdef PARAMETER_CHECK
    if(portId > MAX_PORTID) 
        return PR_ERR_PORTID;
    
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(portStatePtr == NULL ) 
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
#endif

/* VLANID -> VID */ 
    retflag = ifx_get_VLAN_idx(chipId, vlanId, &vid);

    if(retflag == PR_ERR_VLANID)
        return PR_ERR_VLANID;

    if(retflag == PR_ERR_VLAN_NOTCONFG)
        return PR_ERR_VLAN_NOTCONFG;            /* vlanid is not configured  */


    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

/* Reading the portstate from software table */
    *portStatePtr = PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,portId)].port_state;



    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    
        return PR_ERR_NONE;

}
#endif



/**
 * Get MIB counter for VLAN ID.
 * Get the MIB counter associated with the passed Vlan Id. 
 *
 * \param   chipId               Chip Id
 * \param   VlanIdx              Vlan Index (range is 0 - 63)
 * \param   counterId            MIB counterID
 *                                 - 0 Byte counter
 *                                 - 1 Packet counter
 *                                 - 2 Dropped packet counter
 *                                 - 3 Non unicast packet counter
 * \param   countPtr             Pointer to counter value of MIB selected
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_MIBIDX   - If the Vlan index is not in the range of 0 to 63 
 * \return  PR_ERR_MIB_CNTID     - If the MIB counterId is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR      - countPtr or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_VLAN_mib(u8 chipId, u32 VlanIdx, u32 counterId, u32* countPtr)
{
    volatile vlan_mibs_cmd_reg_t*  ptr_vlan_mibs_cmd; /* pointer to VLAN/Rule MIBs cmd register */
    u32* ptr_vlan_mibs_data;  /* pointer to VLAN/Rule MIBs data register */
    u32 Temp =0;
    s32 i=0;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(VlanIdx > MAX_VlanFlowIdx)
        return PR_ERR_VLAN_MIBIDX;

    if(counterId > MAX_MIB_CNTID)
        return PR_ERR_MIB_CNTID;

    if(countPtr == NULL)
        return PR_ERR_NULL_PTR;
    
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    ptr_vlan_mibs_data = (u32*) (Chip_Base_Address[chipId]+VLAN_MIBS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_MIBS_CMD_REG);
    ptr_vlan_mibs_cmd = (volatile vlan_mibs_cmd_reg_t*)&temp1;


	/* 
	 * Configuration of VLAN/Rule MIBs Command register for reading VLAN MIB counter 
	 * defined by VlanIdx and counterId
 	*/
    ptr_vlan_mibs_cmd -> Address = VlanIdx;
    ptr_vlan_mibs_cmd -> MIBId = counterId;
    ptr_vlan_mibs_cmd -> Read_write = READPR;
    ptr_vlan_mibs_cmd -> Req_ack = REQ;

	if((retflag = ifx_sem_take(chipId,SEM_VLAN_MIB)) != PR_ERR_NONE)
        return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_mibs_data,ptr_vlan_mibs_cmd,temp1,Temp,SEM_VLAN_MIB);

    if((retflag = ifx_sem_give(chipId,SEM_VLAN_MIB)) != PR_ERR_NONE)
            return retflag;

    *countPtr = Temp;
    return PR_ERR_NONE ;
}




#if 0
/**
 * Set VLAN Egress tag port list.
 * Set the VLAN Egress Tag port list in the VLAN Table entry associated with 
 * the vlanId. 
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portListPtr          Pointer to VLAN egress tag port list
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0-4094
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portListPtr or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_port_egress_VLAN_tag( u8 chipId, u32 vlanId, u32* portListPtr)
{
    u32 Temp =0;
    u32 vid =0;
    s32  i=0;
    s32 retflag=0;
    BOOL VlanAwareFlag;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;
    u32* ptr_vlan_tbl_data; 
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    /*
     * check if VLAN awareness is activated and 
     * get VLAN table index for passed VLAN ID
     */
    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    retflag = ifx_get_VLAN_idx(chipId,vlanId, &vid);

    if(retflag == PR_ERR_VLANID)
        return PR_ERR_VLANID;

    if(retflag == PR_ERR_VLAN_NOTCONFG)
        return PR_ERR_VLAN_NOTCONFG;                /* vlanid is not configured  */

    ptr_vlan_tbl_cmd -> Vlan_Id    = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;
    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);

    /* Set the new egress tagging information for ports 0-2 */    
    Temp &= ~EGRESS_TAGGING_MASK;
    Temp |= (((*portListPtr) & 0x3FFFFFF)<<6);

/* Configuration of VLAN Table Command register for writing */      
/* Read/Write(30) : 0-Write */
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    
/* Request/Acknowledge(31): 1-Request */
    ptr_vlan_tbl_cmd -> Req_ack = REQ;

/* Added MACRO for indirect register write 7/7/2004 */  

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);


    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;


    return PR_ERR_NONE ;    
   
}
#endif



/**
 * Get VLAN valid flag.
 * Get the VLAN valid flag from VLAN Table entry associated with the passed vlanId. 
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   valid_ptr            Pointer to VLAN valid flag
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0-4094
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portListPtr or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_VLAN_valid( u8 chipId, u32 vlanId, BOOL* valid_ptr)
{
    u32 Temp =0;
    s32 i=0;
    u32 vid =0;
    s32 retflag =0;
    BOOL VlanAwareFlag;

    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;
    u32* ptr_vlan_tbl_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;
    
    if(valid_ptr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    /* VLANID -> VLAN table index */ 
    retflag = ifx_get_VLAN_idx(chipId,vlanId, &vid);

    if(retflag == PR_ERR_VLANID)
        return PR_ERR_VLANID;

    if(retflag == PR_ERR_VLAN_NOTCONFG)
        return PR_ERR_VLAN_NOTCONFG;                /* vlanid is not configured  */
    
    /*
     * Read VLAN table entry
     */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);
   *valid_ptr = (BOOL)(Temp & VLAN_TABLE_VALID_MASK) >> VLAN_TABLE_VALID_SHIFT;
   
    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    return retflag;
   
}


/**
 * Get VLAN Egress tag port list.
 * Get the VLAN Egress Tag port list in the VLAN Table entry associated with 
 * the vlanId. 
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portListPtr          Pointer to VLAN egress tag port list
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0-4094
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portListPtr or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_port_egress_VLAN_tag( u8 chipId, u32 vlanId, u32* portListPtr )
{
    u32 Temp =0;
    s32 i=0;
    u32 vid =0;
    s32 retflag =0;
    BOOL VlanAwareFlag;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;
    u32* ptr_vlan_tbl_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;
    
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    /* VLANID -> VID */ 
    retflag = ifx_get_VLAN_idx(chipId,vlanId, &vid);

    if(retflag == PR_ERR_VLANID)
        return PR_ERR_VLANID;

    if(retflag == PR_ERR_VLAN_NOTCONFG)
        return PR_ERR_VLAN_NOTCONFG;                /* vlanid is not configured  */
    
    /*
     * Read port egress tagging information from VLAN table
     */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);
    *portListPtr = (Temp & VLAN_TABLE_EGRESS_MASK) >> VLAN_TABLE_EGRESS_SHIFT;

    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    return retflag;

}

#if 0
/**
 * Set VLAN port members.
 * Set the VLAN port member list in the VLAN Table entry associated with the vlanId. 
 *
 * \param   chipId                Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portListPtr           Bit 2:0  -> 0:Not a member,1:Member
 *                                Bit 63   -> 0:VLAN create,1:Delete
 * \return  PR_ERR_NONE           - Success 
 * \return  PR_ERR_VLAN_FULL      - Vlan table is full
 * \return  PR_ERR_VLANID         - Vlan Id is invalid
 * \return  PR_ERR_VLAN_NOTCONFG  - Vlan is not configured
 * \return  PR_ERR_VLAN_UNAWARE   - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR       - portListPtr or base address is null pointer
 * \return  PR_ERR_TIME_OUT       - Time out
 * \return  PR_ERR_CHIP_ID        - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_VLAN_port_members(u8 chipId,u32 vlanId, u32* portListPtr)
{   
    u32 Temp =0;
    BOOL VlanAwareFlag ;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    
    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

/* In portList Bit63  0 -> VLAN create
                      1 -> VLAN delete  */                  
    Temp = *(portListPtr+1) & VLAN_CRT_DEL_MSK;
    


    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;

    if(Temp == 0)
    {
        retflag = ifx_create_VLAN(chipId,vlanId,portListPtr);
    }
    else
    {
        if(PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0)
            retflag = PR_ERR_VLAN_NOTCONFG;             /* vlanid is not configured  */
        else
            retflag = ifx_delete_VLAN(chipId,vlanId);
    }
    

    ifx_sem_give(chipId,SEM_VLAN);


    return retflag;                 
}

/**
 * Get VLAN port members.
 * Get the VLAN port member list in the VLAN Table entry associated with the vlanId. 
 *
 * \param   chipId                Chip Id
 * \param   vlanId                VlanId(range is 0-4094)
 * \param   portListPtr           Pointer to portlist 
 * \return  PR_ERR_NONE           - Success 
 * \return  PR_ERR_VLANID         - Vlan table is full
 * \return  PR_ERR_VLAN_NOTCONFG  - Vlan is not configured
 * \return  PR_ERR_VLAN_UNAWARE   - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR       - portListPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID        - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_VLAN_port_members( u8 chipId, u32 vlanId, u32* portListPtr )
{   
    u32 vid =0;
    u32 Temp=0;
    s32 i =0;
    s32 retflag =0;
    BOOL VlanAwareFlag;
    u32 tmpLimit = TOTAL_PHY_PORTS;
    
#ifdef PARAMETER_CHECK
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;                   /* vlanId is invalid */

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif  
    ifx_get_VLAN_aware(chipId, &VlanAwareFlag);

    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    if((retflag = ifx_get_VLAN_idx(chipId, vlanId, &vid)) != PR_ERR_NONE)
        return retflag; 
    

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;
            
    /* Get the Vlan membership for the ports 0-31 */
    tmpLimit = ((tmpLimit < 31 )? tmpLimit : 31);
    for(i=tmpLimit; i>0; i--)
    {
        Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember;
        Temp <<= 1;
    }

    Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,0)].vlanmember;
    *portListPtr = Temp;

#if 0    
/* Get the Vlan membership for the ports 32-50 */
    Temp = 0;
    tmpLimit = ((TOTAL_PHY_PORTS < 50) ? TOTAL_PHY_PORTS : 50);
    for(i=tmpLimit; i>32; i--)
    {
        Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember;
        Temp <<= 1;
    }
    Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,32)].vlanmember;
    *(portListPtr+1) = Temp;
#endif

    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;
    
    
    return retflag;

}
#endif

/**
 * Get free VID list.
 * Get the free VlanIndex(VID) list and the count.
 *
 * \param   chipId                Chip Id
 * \param   vidList               Free VID list
 * \param   free_cnt              Free VID count 
 * \return  PR_ERR_NONE           - Success 
 * \return  PR_ERR_NOFREE_VID     - No free VID's
 * \return  PR_ERR_NULL_PTR       - vidList, free_cnt or base address is null pointer
 * \return  PR_ERR_CHIP_ID        - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_free_VID_list( u8 chipId, u32* vidList, u32* free_cnt )
{
    u32 index = readIndex[chipId];
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    if(vidList == NULL)
        return PR_ERR_NULL_PTR;
    if(free_cnt == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    *free_cnt = freeEntries[chipId];

    if(readIndex[chipId] == writeIndex[chipId])
        return PR_ERR_NOFREE_VID;
            
    while(index != writeIndex[chipId])
    {
        *vidList++ = PR_TABLE_VID[chipId][index];
        index += 1;
        if(index > MAX_VID)
            index = 0;
    }
    *vidList = PR_TABLE_VID[chipId ][ writeIndex[chipId]];

    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}


#if 0
/**
 * Set VID for VLAN ID.
 * Set the VLAN Index(VID) for the VlanId.
 *
 * \param   chipId            Chip Id
 * \param   VlanId            Vlan Id (range is 0-4094)
 * \param   VlanIndex         VlanIndex(range is 0-63)
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_VLANID     - If the Vlan Id is not in the range of 1-4094
 * \return  PR_ERR_TIME_OUT   - Time out
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Obsolete
 */
s32 ifx_set_VLAN_idx( u8 chipId, u32 VlanId, u32 VlanIndex )
{
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data ;
    volatile u32 temp1 = 0;
    s32 i=0;
    u32* ptr_reg1 ;
    s32 retflag = PR_ERR_NONE;
/*
    printk("%s( chipId=%d, VlanId=%d, VlanIndex=%d )\n",
                __FUNCTION__, chipId, VlanId, VlanIndex);
*/    
    
#ifdef PARAMETER_CHECK

    if( VlanId > MAX_VLANID )
        return PR_ERR_VLANID;
    if( chipId >= Total_Chips )
        return PR_ERR_CHIP_NUM ;

    if( Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID )
        return PR_ERR_NULL_PTR ;

#endif

    PR_VLAN_VID_TBL[chipId][VlanId] = VlanIndex;
/*
    printk("%s - VlanId(%d) -> table_pos(%d) )\n", 
           __FUNCTION__, VlanId, VlanIndex);
*/    
    return PR_ERR_NONE ;

}
#endif


/**
 * Get VID for VLAN ID.
 * Get the VLAN Index(VID) for the VlanId.
 *
 * \param   chipId               Chip Id
 * \param   VlanId               VlanId(range is 0-4094)
 * \param   VlanIndexPtr         Pointer to VlanIndex(range is 0-63)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0-4094
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_NULL_PTR      - portStatePtr is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Obsolete
 */
s32 ifx_get_VLAN_idx(u8 chipId , u32 VlanId, u32* VlanIndexPtr)
{
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK  
    
    if(VlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(VlanIndexPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif  

    /*
     * Get the VLAN table position of the passed VLAN ID from 
     * the PR_VLAN_VID_ASC_TBL and perform a validity check
     */
    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;
    
    *VlanIndexPtr = PR_VLAN_VID_TBL[chipId][VlanId];
    if(PR_VLAN_VID_ASC_TBL[chipId][VlanId] == 0)
        retflag = PR_ERR_VLAN_NOTCONFG;   /* vlanId is not configured  */ 
    else
    {
/*
        printk("%s - VlanId(%d) -> table_pos(%d)\n", 
               __FUNCTION__, VlanId, PR_VLAN_VID_TBL[chipId][VlanId]);
*/               
        retflag = PR_ERR_NONE;
    }

    ifx_sem_give(chipId,SEM_CPU);
    return retflag;
    
}

#if 0
/**
 * Set the VLAN port member list.
 * Set the VLAN port member list in the VLAN Table entry associated with the vlanId.
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portmemberList       Bit 2:0  -> 0:Not a member,1:Member
 *                               Bit 63   -> 0:VLAN create,1:Delete
 * \param   portegressList       Pointer to egress list
 * \param   mstp_stp             stp or mstp mode
 * \param   shar_ind             shared or independent VID  
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_FULL     - Vlan table is full
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_vlan_table( u8 chipId, u32 vlanId, 
                        u32* portmemberList, u32* portegressList,
                        u8 mstp_stp, BOOL shar_ind )
{   
    u32 vid =0;
    u32 VlanMIBIdx=0;
    s32 i=0,j=0;
    BOOL VlanAwareFlag = FALSE;
    u32 Temp = 0;
    s32 retflag = PR_ERR_NONE;  

    u32 portListPtr[2]={0,0};
    u32 portListtag[2]={0,0};

    u32 vid_tmp = 0;
    u32 vid_srch = 0;

    u8 mstp_stp_tmp = 0;
    BOOL shar_ind_tmp = FALSE;

    printk("%s( chipId=%d, vlanId=%d, portmemberList[0]=0x%08X, portmemberList[1]=0x%08X\n", 
           __FUNCTION__, chipId, vlanId, *portmemberList, *(portmemberList+1));
           
    printk("portegressList[0]=0x%08X, portegressList[1]=0x%08X, mstp_stp=%d, shar_ind=%d)\n", 
           *portegressList, *(portegressList+1), mstp_stp, shar_ind);
                      
#ifdef PARAMETER_CHECK  
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;
    
    if(portmemberList == NULL)
        return PR_ERR_NULL_PTR;

    if(portegressList == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif  

    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    /* In portList Bit63  0 -> VLAN create
                          1 -> VLAN delete  */                  
    Temp = *(portmemberList+1) & VLAN_CRT_DEL_MSK;

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;
    
    if(Temp == VLAN_CRT_DEL_MSK)
    {   
        /*
         * Delete VLAN entry
         */
        if(PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0)
        {
            ifx_sem_give(chipId,SEM_VLAN);
            printk("%s - PR_VLAN_VID_ASC_TBL[%d][%d] == 0\n", __FUNCTION__, chipId, vlanId);            
            return PR_ERR_VLAN_NOTCONFG;          /* vlanid is not configured  */
        }
        
        ifx_get_VLAN_idx(chipId,vlanId,&vid);

        PR_VLAN_MULTIPLE_VID[chipId][vid] -= 1;
        PR_VLAN_SHARE_IND[chipId][vlanId] = 0;
        PR_VLAN_STP_MSTP[chipId][vlanId] = 0;       

        if(PR_VLAN_MULTIPLE_VID[chipId][vid] == 0)
        {
            retflag = ifx_delete_VLAN( chipId, vlanId );
            ifx_sem_give(chipId,SEM_VLAN);
            return PR_ERR_NONE;
        }
            
        else
        {
            PR_VLAN_VID_ASC_TBL[chipId][vlanId] = 0;
#if 0            
            ifx_set_VLAN_idx(chipId,vlanId,DEFAULT_VID);
#endif            
            ifx_sem_give(chipId,SEM_VLAN);
            return PR_ERR_NONE; 
        }
    }
    else
    {
        /*
         * Create VLAN entry
         */        
        if(PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 1)
        {
            ifx_get_VLAN_idx(chipId,vlanId,&vid);
            PR_VLAN_MULTIPLE_VID[chipId][vid] -= 1;
        }
    }   

    PR_VLAN_SHARE_IND[chipId][vlanId] = shar_ind;
    PR_VLAN_STP_MSTP[chipId][vlanId] = mstp_stp;

    if(shar_ind == TRUE)
    {
        searchIndex(portListPtr,&vid_srch);

        printk("%s - searchIndex() returned %d \n",  __FUNCTION__, vid_srch);
        if(vid_srch > MAX_VID)
            vid_srch = 0;

        for(i=0;i <= MAX_VID;i++)
        {
            if( PR_VLAN_MULTIPLE_VID[chipId][vid_srch] >= 1 )
            {
                for(j=1;j<TOTAL_VLAN_IDS;j++)
                {
                    ifx_get_VLAN_idx(chipId,j,&vid_tmp);
                    if(vid_tmp == vid_srch)
                        break;
                    else
                        continue;
                }
                ifx_get_VLAN_STP_MSTP(chipId,j,&mstp_stp_tmp);
                ifx_get_VLAN_VID_share(chipId,j,&shar_ind_tmp);
                        
                if((mstp_stp_tmp == mstp_stp) && (shar_ind_tmp == shar_ind))
                {
                    ifx_get_VID_port_members(chipId,vid_srch,portListPtr);
                    ifx_get_port_egress_VID_tag(chipId,vid_srch,portListtag);
                    if((portListPtr[0]==portmemberList[0]) && 
                       (portListPtr[1] == (portmemberList[1] & 0x7FFFF)) &&
                       (portListtag[0] == portegressList[0]) && 
                       (portListtag[1] == portegressList[1]) )            
                        break;
                }   
            }
            vid_srch += 1;
            if(vid_srch > MAX_VID)
                vid_srch = 0;
        }
    }
        
    if(  (i == MAX_VID) || (shar_ind == FALSE) )
    {
        if((PR_VLAN_MULTIPLE_VID[chipId][vid] != 0) || 
           (PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0))
        {
//                ifx_get_VLAN_MIB_idx(chipId,&VlanMIBIdx);
            if((PR_TABLE_VID[chipId][readIndex[chipId]] < VlanMIBIdx) &&
                (freeEntries[chipId] > VlanMIBIdx))
            {
                ifx_VID_sort(chipId); 
            }
            
            if(ifx_get_vid(chipId,&vid) == PR_ERR_VLAN_FULL)
            {
                ifx_sem_give(chipId,SEM_VLAN);
                return PR_ERR_VLAN_FULL;
            }
        }   

        if((retflag = ifx_set_VID_port_members( chipId, vlanId,
                                                vid,portmemberList)) != PR_ERR_NONE)
        {
            printk("%s - ifx_set_VID_port_members() failed \n",  __FUNCTION__);
            ifx_sem_give(chipId,SEM_VLAN);
            return retflag;
        }

        if((retflag = ifx_set_port_egress_VID_tag(chipId,vid,portegressList)) != PR_ERR_NONE)
        {
            printk("%s - ifx_set_port_egress_VID_tag() failed \n",  __FUNCTION__);
            ifx_sem_give(chipId,SEM_VLAN);
            return retflag;
        }
        vid_srch = vid;
    }
    else
        if((PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 1) && 
           (PR_VLAN_MULTIPLE_VID[chipId][vid] == 0))
            ifx_delete_VLAN(chipId,vlanId);
        
    PR_VLAN_VID_ASC_TBL[chipId][vlanId] =1;
#if 0    
    printk("%s -> ifx_set_VLAN_idx( vlanId=%d, vid=%d ) \n",
            __FUNCTION__, vlanId, vid_srch);
    ifx_set_VLAN_idx(chipId,vlanId,vid_srch);
#endif    
    PR_VLAN_MULTIPLE_VID[chipId][vid_srch] += 1;

    
    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;
    
    return PR_ERR_NONE;

}
#endif

/**
 * Writes VLAN entry to HW VLAN table
 * Writes the passed VLAN entry at position vlanIdx into HW VLAN table .
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   vlanIdx              VLAN table index
 * \param   portmemberList       Bit 2:0  -> 0:Not a member,1:Member
 * \param   portegressList       Bit 2:0  -> 0:No egress tagging,1:egress tagging
 * \param   valid                Entry's valid information 
 *                               (0:VLAN invalid,1:VLAN valid)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_VLAN_INDEX    - VLAN table index is invalid
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id invalid
 * \return  PR_ERR_VLAN_VALID    - invalid valid-identifier (FALSE or TRUE)
 * \ingroup Internal
 */
s32 ifx_set_vlan_table_entry( u8 chipId, u32 vlanId, u32 vlanIdx,
                              u32* portmemberList, u32* portegressList, u32 valid  )
{  
    u32 Temp =0;
    s32 retflag=0;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;/* pointer to VLAN table command register */
    u32* ptr_vlan_tbl_data;            /* pointer to VLAN Table Data register */
    s32  i=0;   /* used by PR_Indirect_Reg_Write_Sem macro */
    
    
#ifdef PARAMETER_CHECK
    if(vlanId > MAX_VLANID+1)
        return PR_ERR_VLANID;          
    
    if(vlanIdx > MAX_VID) 
        return PR_ERR_VLAN_INDEX;

    if(portmemberList == NULL)
        return PR_ERR_NULL_PTR;
    
    if(portegressList == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
    
    if( (valid != TRUE) && (valid != FALSE) ) 
        return PR_ERR_VLAN_VALID;

#endif
    /*
     * set pointer to VLAN table access registers and
     * write new vlan entry to table 
     */
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1          = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd  = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    Temp = (vlanId & VLAN_TABLE_VLANID_MASK) |
                         ((*portmemberList & 0x00000007)<<VLAN_TABLE_MEMBER_SHIFT) |
                         ((*portegressList & 0x00000007)<<VLAN_TABLE_EGRESS_SHIFT) |
                         ((valid & 0x00000001)<<VLAN_TABLE_VALID_SHIFT);

/*                         
    printk("%s - ptr_vlan_tbl_data=0x%08X, ptr_reg1=0x%08X\n", 
               __FUNCTION__, ptr_vlan_tbl_data, ptr_reg1 );                             
    if( *ptr_vlan_tbl_data !=VLAN_ENTRY_INVALID)                         
        printk("%s - write value 0x%08X to HW VLAN table at pos %d\n", 
               __FUNCTION__, *ptr_vlan_tbl_data, vlanIdx );                         
*/                         
    ptr_vlan_tbl_cmd -> Vlan_Id    = vlanIdx;
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;  

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);

    
    if((retflag = ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;    

    
    ptr_vlan_tbl_cmd -> Vlan_Id    = vlanIdx;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;
    

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);
    
    if( (retflag=ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;        

/*    
    printk("%s - Read back from HW VLAN table for vid=%d returned 0x%08X\n", 
            __FUNCTION__, vlanIdx, Temp );
*/
    return retflag;

}




/**
 * Get VLAN entry from VLAN table
 * Reads the entry located at position table_index from VLAN table .
 *
 * \param   chipId               Chip Id
 * \param   vlanId               VlanId(range is 0-4094)
 * \param   portmemberList       Bit 2:0  -> 0:Not a member,1:Member
 * \param   portegressList       Bit 2:0  -> 0:No egress tagging,1:egress tagging
 * \param   valid_ptr            Pointer to entry's valid information 
 *                               (0:VLAN invalid,1:VLAN valid)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_FULL     - Vlan table is full
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList,  valid_ptr 
 *                                 or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_vlan_table_entry( u8 chipId, u32 vlanId,
                              u32* portmemberList, u32* portegressList,
                              BOOL* valid_ptr )
{  
    u32 Temp =0;
    s32 retflag=0;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;/* pointer to VLAN table command register */
    u32* ptr_vlan_tbl_data;            /* pointer to VLAN Table Data register */
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    BOOL VlanAwareFlag;
    u32 vid;
    s32 i;   /* needed by PR_Indirect_Reg_Read_Sem */

#ifdef PARAMETER_CHECK

    if(portmemberList == NULL)
        return PR_ERR_NULL_PTR;
    
    if(portegressList == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    /*
     * set pointer to VLAN table access registers and
     * write new vlan entry to table 
     */
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1          = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd  = (volatile vlan_tbl_cmd_reg_t*)&temp1;
    
    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    /* VLANID -> VLAN table index */ 
    retflag = ifx_get_VLAN_idx(chipId,vlanId, &vid);

    if(retflag == PR_ERR_VLANID)
        return PR_ERR_VLANID;

    if(retflag == PR_ERR_VLAN_NOTCONFG)
        return PR_ERR_VLAN_NOTCONFG;                /* vlanid is not configured  */
    
    /*
     * Read VLAN table entry and pass requested information to pointers 
     */
    ptr_vlan_tbl_cmd -> Vlan_Id    = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);
    *valid_ptr = (BOOL)(Temp & VLAN_TABLE_VALID_MASK) >> VLAN_TABLE_VALID_SHIFT;
    *portmemberList = (Temp & VLAN_TABLE_MEMBER_MASK) >> VLAN_TABLE_MEMBER_SHIFT;
    *portegressList = (Temp & VLAN_TABLE_EGRESS_MASK) >> VLAN_TABLE_EGRESS_SHIFT;
    
    if( (retflag=ifx_sem_give(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;        
/*    
    printk("%s - HW VLAN table access for vid=%d returned 0x%08X\n", 
            __FUNCTION__, vid, Temp );
*/            
    if( vlanId != (Temp & VLAN_TABLE_VLANID_MASK) )
    {
        printk("%s - Invalid VLAN ID %d returned from HW VLAN table for vid=%d\n", 
            __FUNCTION__, (Temp & VLAN_TABLE_VLANID_MASK), vid );
            
        printk("Is %d should be %d\n", (Temp & VLAN_TABLE_VLANID_MASK), vlanId );
        retflag=PR_ERR_VLANID;
    
    }
 
    return retflag;

}


#if 0
/**
 * Writes VLAN entry to HW VLAN table
 * Writes the passed VLAN entry at position vlanIdx into HW VLAN table .
 *
 * \param 
 * \param  
 * \param  
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_FULL     - Vlan table is full
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_delete_vlan_table_entry( u8 chipId, u32 vlanIdx )
{  
    u32 Temp =0;
    s32 retflag=0;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;/* pointer to VLAN table command register */
    u32* ptr_vlan_tbl_data;            /* pointer to VLAN Table Data register */
    s32  i=0;   /* used by PR_Indirect_Reg_Write_Sem macro */
    
    
#ifdef PARAMETER_CHECK
  
    if(vlanIdx > MAX_VID) 
        return PR_ERR_VLAN_INDEX;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
    
#endif
    /*
     * set pointer to VLAN table access registers and
     * write invalid vlan entry to table 
     */
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1          = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd  = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    *ptr_vlan_tbl_data = VLAN_ENTRY_INVALID;
   
    ptr_vlan_tbl_cmd -> Vlan_Id    = vlanIdx;
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;  
    
    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp,SEM_VLAN);

    retflag = ifx_sem_give(chipId,SEM_VLAN);

    return retflag;

}
#endif


/**
 * Controls VLAN tag insertion by PMAC
 * Enables or disables the VLAN tag insertion at switch CPU port by configuring 
 * the PMAC_HD_CTRL register.
 *
 * \param chipId               Chip Id
 * \param en_dis               Tagging enable(1)/disable(0) selector
 * \param vlanId               VlanId to be inserted to tag (0-4094)
 * \param prio                 priority to be inserted to tag (0...7)
 * \param cfi                  canonical format identifier value to be 
 *                             inserted to tag (0-little endian, 1-big endian)
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - Vlan Id is invalid 
 * \return  PR_ERR_PRIORITY      - Priority invalid
 * \return  PR_ERR_CFI           - CFI invalid
 * \return  PR_ERR_NULL_PTR      - Base address is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_pmac_vlan_tag_insertion( u8 chipId, BOOL en_dis, u32 vlanId, u8 prio, u8 cfi )
{  
    u32*          ptr_pmac_hd_ctl;   
    u32*          ptr_pmac_vlan;
    pmac_hd_ctl_t pmac_hd_ctl;   
    pmac_vlan_t   pmac_vlan;    

#ifdef PARAMETER_CHECK
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
    
    if( vlanId > MAX_VLANID )
        return PR_ERR_VLANID;  
    
    if( prio > MAX_PRIORITY )
        return PR_ERR_PRIORITY;
    
    if( (cfi != 0) && (cfi != 1) ) 
        return PR_ERR_CFI;
#endif
    /*
     * enable or disable VLAN tag insertion acc. to passed en_dis parameter.
     * In case of enabling write new VLAN parameters to be inserted to 
     * PMAC_VLAN register
     */
    ptr_pmac_hd_ctl = (u32*) ( Chip_Base_Address[chipId]+SWITCH_PMAC_HD_CTL );
    ptr_pmac_vlan   = (u32*) ( Chip_Base_Address[chipId]+SWITCH_PMAC_VLAN );

    PR_Reg_Read(ptr_pmac_hd_ctl,pmac_hd_ctl.val);
/*
    printk("%s - read 0x%08X from SWITCH_PMAC_HD_CTL register\n", 
            __FUNCTION__, pmac_hd_ctl.val);
*/            
    if( en_dis == 1 )
    {
        pmac_vlan.val         = 0x00000000;
        pmac_vlan.fld.pri     = prio;
        pmac_vlan.fld.cfi     = cfi;
        pmac_vlan.fld.vlan_id = vlanId;
        PR_Reg_Write( ptr_pmac_vlan, pmac_vlan.val);
/*
        printk("%s - wrote 0x%08X to SWITCH_PMAC_VLAN register\n", 
               __FUNCTION__, pmac_vlan.val);
*/               
        pmac_hd_ctl.fld.add_tag    = 1;
        pmac_hd_ctl.fld.add_eth_hd = 1;  
    }
    else
    {
        pmac_hd_ctl.fld.add_tag    = 0;  
/*        pmac_hd_ctl.fld.add_eth_hd = 0;  */  /* unsure if this needs to be disabled too */         

    }
    PR_Reg_Write( ptr_pmac_hd_ctl, pmac_hd_ctl.val );
/*
    printk("%s - wrote 0x%08X to SWITCH_PMAC_HD_CTL register\n", 
            __FUNCTION__, pmac_hd_ctl.val);    
*/            
    return PR_ERR_NONE;

}



/********************************* Internal APIs ****************************************/

#if 0
/**
 * Delete VLAN entry.
 * This function is used to delete VLAN entry in the hardware VLAN Table.
 *
 * \param   chipId               Chip Id
 * \param   vid                  VlanIndex(range is 0 - 63)
 * \param   vlanId               VlanId(range is 0-4094)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_NULL_PTR      - Base address is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_delete_VLAN_hardware( u8 chipId, u32 vid, u32 vlanId )
{       
    s32 index =0, i=0;
    u32 portList[2]={0,0};
    u32 Temp = 0;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;
    u32* ptr_vlan_tbl_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    /*
     * Configure VLAN table command and data register 
     */
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId] + VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;
    

    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);        

    /* Set the vlan egress tag for all the ports to zero */ 
    ifx_set_port_egress_VID_tag(chipId,vid,portList);
        
    return PR_ERR_NONE ;
    
}
#endif

#if 0
/**
 * Update VID table.
 * This function updates software Vlanindex(VID) Table with deleted vid
 *
 * \param   chipId               Chip Id
 * \param   vid                  VlanIndex(range is 0 - 63)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_add_vid(u8 chipId,u32 vid)
{

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
#endif
    
    writeIndex[chipId] = writeIndex[chipId] + 1; /* % 1024 */
    if(writeIndex[chipId] > MAX_VID)
        writeIndex[chipId] = 0;

    PR_TABLE_VID[chipId][writeIndex[chipId]] = vid;
    
    printk("%s - added new vid %d at pos. %d\n",
           __FUNCTION__, vid, writeIndex[chipId]);    
    freeEntries[chipId] += 1;           
    
    return PR_ERR_NONE ;
}

/**
 * Get VID from VID table.
 * This function is used to get the Vlanindex(VID) from the software VID Table.
 *
 * \param   chipId            Chip Id
 * \param   vidPtr            Pointer to vlanindex(VID)
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_VLAN_FULL  - Vlan table is full
 * \return  PR_ERR_NULL_PTR   - vidPtr is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_get_vid(u8 chipId,u32* vidPtr)
{

#ifdef PARAMETER_CHECK
    
    if(vidPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    
#endif

    if(freeEntries[chipId] == 0)
        return PR_ERR_VLAN_FULL;    /* Vlan table is full   */

    *vidPtr = PR_TABLE_VID[chipId][readIndex[chipId]];
    printk("%s - got position %d\n", __FUNCTION__, *vidPtr);
    
    readIndex[chipId] = readIndex[chipId] + 1;  /* % 1024  */
    if(readIndex[chipId] > MAX_VID)
        readIndex[chipId] = 0;

    printk("%s - new readIndex[%d] = %d\n",
           __FUNCTION__, chipId, readIndex[chipId]);    
    
    freeEntries[chipId] -= 1;
    
    return PR_ERR_NONE ;
}

#endif


#if 0
/**
 * Calculate search index.
 * Calculation of the Index into vlan table using the hashing algorithm
 *
 * \param   portListPtr       Port member list
 * \param   vidPtr            Pointer to Index
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_NULL_PTR   - portListPtr or vidPtr is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 searchIndex(u32* portListPtr,u32* vidPtr)
{
    u32 srch_index=0,crc =0;
    s32 i=0;
    u32 temp=0;

#ifdef PARAMETER_CHECK      
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(vidPtr == NULL)
        return PR_ERR_NULL_PTR;
#endif

/* x^9 + x^2 + 1*/

    temp = portListPtr[0];

    for(i=0 ;i<32; i++)
    {
        if(((crc>>9)  & 0x1) ^ (temp & 0x1)) 
            crc = ((crc ^ 2) << 1) | 1;
        else
            crc <<= 1;
            
        temp >>= 1;
    }
    
    temp = portListPtr[1];

    for(i=0 ;i<19; i++)
    {
        if(((crc>>9) & 0x1) ^ (temp & 0x1)) 
            crc = ((crc ^ 2) << 1) | 1;
        else
            crc <<= 1;
            
        temp >>= 1;
    }
    
    
/* swap the bit positions */

    srch_index = 0;
    for (i=0;i<10;i++)
    {
        srch_index <<=1;    
        temp = crc &0x1;
        srch_index |=temp;
        crc>>=1;                       
    } 

    *vidPtr = srch_index & 0x3FF;

/*  *vidPtr = crc & 0x3FF; */

    return PR_ERR_NONE ;
    
}
#endif


#if 0
/**
 * Set VLAN port member list.
 * Set the VLAN port member list in the VLAN Table entry associated with the 
 * vlanId. 
 *
 * \param   chipId            ChipId
 * \param   vlanId            VlanId(range is 0-4094)
 * \param   vid               VID(range is 1 - 63)
 * \param   portListPtr       Bit 2:0 -> 0:Not a member,1:Member
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_VLANID     - Vlan Id is invalid
 * \return  PR_ERR_TIME_OUT   - Time out
 * \return  PR_ERR_NULL_PTR   - portListPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_set_VID_port_members( u8 chipId, u32 vlanId, u32 vid, u32 *portListPtr )
{

    u32 Temp=0;
    u32 Status_Member[4]={0,0,0,0};
    s32 i=0;
    s32 start_limit=0,end_limit=0;
    u32 tmpLimit=0;
    u32 portId_limit =0;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd; /* pointer to VLAN table 
                                                       command register */
    u32* ptr_vlan_tbl_data;     /* pointer to VLAN Table Data register */
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK  

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif  


    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    /* Update the Hardware Vlan Table with new vlanmembership */
    Temp = *portListPtr;


/* For ports 0 to 15 */
    start_limit = (((TOTAL_FE_PORTS-1) < 15) ? (TOTAL_FE_PORTS-1) : 15);
    end_limit = 0;
    portId_limit = start_limit;
    if(ifx_port_state(chipId,start_limit,end_limit,&Status_Member[0],vid,Temp,portId_limit) != PR_ERR_NONE)
    {
        return PR_ERR_PORTSTATE; 
    }

#if 0
/* For ports 16 to 31 */
    /* changing for solving the GE port state problem */
    Temp = *portListPtr;
    start_limit = (((TOTAL_FE_PORTS-1) < 31) ? (TOTAL_FE_PORTS-1) : 31);
    end_limit = 16;
    portId_limit = start_limit; 
    if(ifx_port_state(chipId,start_limit,end_limit,&Status_Member[1],vid,Temp,portId_limit) != PR_ERR_NONE)
    {
        return PR_ERR_PORTSTATE; 
    }

/* For ports 32 to 47 */
    Temp = *(portListPtr+1);
    start_limit = (((TOTAL_FE_PORTS-1) < 15) ? (TOTAL_FE_PORTS-1) : 15);
    end_limit = 0;
    portId_limit = 47;
    if(ifx_port_state(chipId,start_limit,end_limit,&Status_Member[2],vid,Temp,portId_limit) != PR_ERR_NONE)
    {
        return PR_ERR_PORTSTATE; 
    }

/* For ports 48 to 50 */
    Temp = *(portListPtr+1);
    start_limit = 18;
    end_limit = 16;
    portId_limit = 50;
    if(ifx_port_state(chipId,start_limit,end_limit,&Status_Member[3],vid,Temp,portId_limit) != PR_ERR_NONE)
    {
        return PR_ERR_PORTSTATE; 
    }
#endif

/* Update software Vlan membership table for ports 0-31 */
    Temp = *portListPtr;
    tmpLimit = (((TOTAL_PORTS-1) < 31) ? (TOTAL_FE_PORTS-1) : 31);
    for(i=0; i<=tmpLimit; i++)
    {
        PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember = Temp & 0x1;
        Temp >>= 1;
    }

    
#if 0
/* Update software Vlan membership table for ports 32 - 50 */
    Temp = *(portListPtr+1);
    tmpLimit = (((TOTAL_PORTS-1) < 50) ? (TOTAL_FE_PORTS-1) : 50);
    for(i=32; i<=tmpLimit; i++)
    {
        PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember = Temp & 0x1;
        Temp >>= 1;
    }
#endif


    /* 
     * Read current VLAN configuration from VLAN table and 
     * update HW VLAN table with passed port membership configuration
     */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;     
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    Temp &= ~VLAN_TABLE_EGRESS_MASK; 
    Temp |= ((Status_Member[(TOTAL_FE_PORTS/16)] & 0x7) << VLAN_TABLE_EGRESS_SHIFT);   

    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;  
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    return PR_ERR_NONE ;

#if 0     

/* Configuration of VLAN Table Command register for writing */
/* VLANID(9:0) - VID(0 -> 1023) */  
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
        
/* Read/Write(30) : 1- Write */
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
/* LENA decrease to 24+2 */ 
    for( index=0 ; index<(TOTAL_FE_PORTS/16) ; index++)
    {
/* Selection(29:27) - Select one of 32 bit sections of the VLAN Table entry */
/*  000 -> Bridge_State for ports 0 to 15(2 bits per port)
    001 -> Bridge_State for ports 16 to 31
    010 -> Bridge_State for ports 32 to 47      */
/* 
        ptr_vlan_tbl_cmd -> Select = index;
*/      
   printk("%s: No select field in ptr_vlan_tbl_cmd structure!\n", __FUNCTION__);        
        
/* Request/Acknowledge(31): 1-Request */
            
        ptr_vlan_tbl_cmd -> Req_ack = REQ;

/* Added MACRO for indirect register write 7/7/2004 */
        PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Status_Member[index]);

    }   
    
/* Selection(29:27) - Select one of 32 bit sections of the VLAN Table entry */
/* 011 -> Bridge_State for ports 48 to 50                                   */
/* 
    ptr_vlan_tbl_cmd -> Select = 3;
*/
   printk("%s: No select field in ptr_vlan_tbl_cmd structure!\n", __FUNCTION__);        
/* Read/Write(30) : 1-Read */
    
    ptr_vlan_tbl_cmd -> Read_write = READPR;

/* Request/Acknowledge(31): 1-Request */
            
    ptr_vlan_tbl_cmd -> Req_ack = REQ;


/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);


    Temp &= 0xFFFFFFC0;
/* LENA decrease to 24+2 */
    Temp |= (Status_Member[(TOTAL_FE_PORTS/16)] & 0x3F);    

    
/* Configuration of VLAN Table Command register for writing */      
/* Read/Write(30) : 0-Write */
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    
/* Request/Acknowledge(31): 1-Request */
    ptr_vlan_tbl_cmd -> Req_ack = REQ;  
    
/* Added MACRO for indirect register write 7/7/2004 */
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    return PR_ERR_NONE ;
#endif
}
#endif

#if 0
/**
 * Get VLAN port member list.
 * Get the VLAN port member list in the VLAN Table entry associated with the vlanId.  
 *
 * \param   chipId            ChipId
 * \param   vid               VID(range is 0 - 63)
 * \param   portListPtr       Bit 2:0 -> 0:Not a member,1:Member
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_TIME_OUT   - Time out
 * \return  PR_ERR_NULL_PTR   - portListPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_get_VID_port_members( u8 chipId, u32 vid, u32* portListPtr )
{

    u32 Temp=0;
    s32 i =0;
    s32 retflag =0;
    u32 tmpLimit = TOTAL_PHY_PORTS;
    
#ifdef PARAMETER_CHECK
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif  
            
    /* Get the Vlan membership for the ports 0-31 */
    tmpLimit = ((tmpLimit < 31 )? tmpLimit : 31);
    for(i=tmpLimit; i>0; i--)
    {
        Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember;
        Temp <<= 1;
    }

    Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,0)].vlanmember;
    *portListPtr = Temp;

/* Get the Vlan membership for the ports 32-50 */
    Temp = 0;
    tmpLimit = ((TOTAL_PHY_PORTS < 50) ? TOTAL_PHY_PORTS : 50);
    for(i=tmpLimit; i>32; i--)
    {
        Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember;
        Temp <<= 1;
    }
    Temp |= PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,32)].vlanmember;
    *(portListPtr+1) = Temp;
    
    return retflag;

}
#endif


#if 0
/**
 * Set the VLAN Egress Tag port list.
 * Set the VLAN Egress Tag port list in the VLAN Table entry associated with 
 * the vlanId.  
 *
 * \param   chipId            ChipId
 * \param   vid               VID(range is 0 - 63)
 * \param   portListPtr       Pointer to VLAN egress tag port list
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_TIME_OUT   - Time out
 * \return  PR_ERR_NULL_PTR   - portListPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_set_port_egress_VID_tag(u8 chipId,u32 vid, u32* portListPtr)
{
    u32 Temp =0;
    s32  i=0;
    BOOL VlanAwareFlag;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd; /* pointer to VLAN table command register */

/* Define a pointer to VLAN Table Data register */
    u32* ptr_vlan_tbl_data; 

    volatile u32 temp1 = 0;
    u32* ptr_reg1;


#ifdef PARAMETER_CHECK
    
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    /* 
     * Read current egress tagging and membership information for passed vid
     * and overwrite egress tagging information with passed egress tagging 
     * information for ports 0,1 and 2
     */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    /* Set egress tagging information for ports 0-2  */ 
    Temp &= ~VLAN_TABLE_EGRESS_MASK; 
    Temp |= (((*portListPtr) & EGRESS_TAGGING_MASK) << VLAN_TABLE_EGRESS_SHIFT);
    
    /* write new configuration */
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;  
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    return PR_ERR_NONE ;
}
#endif

#if 0
/**
 * Get the VLAN Egress Tag port list.
 * Get the VLAN Egress Tag port list in the VLAN Table entry associated with 
 * the vlanId.
 *
 * \param   chipId            ChipId
 * \param   vid               VID(range is 0 - 63)
 * \param   portListPtr       Pointer to VLAN egress tag port list
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_TIME_OUT   - Time out
 * \return  PR_ERR_NULL_PTR   - portListPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_get_port_egress_VID_tag( u8 chipId, u32 vid, u32* portListPtr )
{
    u32 Temp =0;
    s32 i=0;
    BOOL VlanAwareFlag;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;
    u32* ptr_vlan_tbl_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK
        
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    ifx_get_VLAN_aware(chipId,&VlanAwareFlag);
    if(VlanAwareFlag == FALSE)
        return PR_ERR_VLAN_UNAWARE;

    ptr_vlan_tbl_cmd -> Vlan_Id    = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    /* Extract the egress tagging information for ports 0-2 */ 
    *portListPtr = (Temp & VLAN_TABLE_EGRESS_MASK) >> VLAN_TABLE_EGRESS_SHIFT;

    return PR_ERR_NONE;

}
#endif

#if 0
/**
 * Sort VID table.
 * This function is used to sort the software VID table.
 *
 * \param   chipId            ChipId
 * \ingroup Internal
 * \internal
 */
void ifx_VID_sort(u8 chipId)
{
    s32 i=0;
    s32 j=0;
    s32 k=0;
    s32 l=0;
    u32 Temp =0;
    u32 flag = 1;

    while((i<(freeEntries[chipId] - 1)) & flag)
    {
        flag = 0;
        for(j=readIndex[chipId],k=0;k < (freeEntries[chipId] - 1); j=l,k++)
        {
            l = (j+1)%NUMBER_VIDS;
 
            if(PR_TABLE_VID[chipId][ j] < PR_TABLE_VID[chipId][l])
            {
                Temp =  PR_TABLE_VID[chipId][j];
                PR_TABLE_VID[chipId ][j] = PR_TABLE_VID[chipId][l];
                PR_TABLE_VID[chipId][l] = Temp;
                flag = 1;
            }
        }
        i++;
    }
}
#endif

#if 0
/**
 * Create/Modify VLAN entry.
 * This function is used either to create new VLAN Table entry or modify the 
 * existing entry.
 *
 * \param   chipId            ChipId
 * \param   vlanId            VlanId(range is 0-4094)
 * \param   portListPtr       Bit 2:0 -> 0:Not a member,1:Member
 * \return  PR_ERR_NONE       - Success
 * \return  PR_ERR_VLAN_FULL  - Vlan table is full
 * \return  PR_ERR_TIME_OUT   - Time out
 * \return  PR_ERR_NULL_PTR   - portListPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_create_VLAN( u8 chipId, u32 vlanId, u32* portListPtr )
{   
    u32 vid =0;
    u32 Temp=0;
    u32 Status_Member[4]={0,0,0,0};
    u32 VlanMIBIdx=0;
    s32 i=0,j=0;
    s32 start_limit = 0;
    s32 end_limit = 0;
    u32 tmpLimit = 0;
    u32 portId_limit =0;
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;
    u32* ptr_vlan_tbl_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK  

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;
    if(portListPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif  

    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId]+VLAN_TBL_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId]+VLAN_TBL_CMD_REG);
    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

    /* If Vlanindex(vid) doesn't exist for vlanId create a new VLAN Table entry or modify
      the existing entry. */   
    if(ifx_get_VLAN_idx(chipId,vlanId,&vid) == PR_ERR_VLAN_NOTCONFG)
    {
//        ifx_get_VLAN_MIB_idx(chipId,&VlanMIBIdx);
        if((PR_TABLE_VID[chipId][readIndex[chipId]] < VlanMIBIdx) && (freeEntries[chipId] > VlanMIBIdx))
                ifx_VID_sort(chipId); 
                        
        if(ifx_get_vid(chipId,&vid) == PR_ERR_VLAN_FULL)
        {
            return PR_ERR_VLAN_FULL;
            
        }   
        else
        {
            PR_VLAN_VID_ASC_TBL[chipId][vlanId] =1;
#if 0            
            ifx_set_VLAN_idx(chipId,vlanId,vid);
#endif            
        }
    }

    /* Update the Hardware Vlan Table with new vlanmembership information */
    Temp = *portListPtr;

    /* Read port state for ports 0...2 */
    start_limit = (((TOTAL_FE_PORTS-1) < 15) ? (TOTAL_FE_PORTS-1) : 15);
    end_limit = 0;
    portId_limit = start_limit; 
    if(ifx_port_state( chipId, start_limit, end_limit,
                       &Status_Member[0], vid, Temp, portId_limit) != PR_ERR_NONE)
    {
        return PR_ERR_PORTSTATE; 
    }

    /* Update software Vlan membership table for ports 0-31 */
    Temp = *portListPtr;
    tmpLimit = (((TOTAL_PORTS-1) < 31) ? (TOTAL_FE_PORTS-1) : 31);
    for(i=0; i<=tmpLimit; i++)
    {
        PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember = Temp & 0x1;
        Temp >>= 1;
    }


#if 0 
    /* Configuration of VLAN Table Command register for writing */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;

/* LENA decrease to 24+2 */ 
    for( j=0 ; j<(TOTAL_FE_PORTS/16) ; j++)
    {
/* Selection(29:27) - Select one of 32 bit sections of the VLAN Table entry */
/*  000 -> Bridge_State for ports 0 to 15(2 bits per port)
    001 -> Bridge_State for ports 16 to 31
    010 -> Bridge_State for ports 32 to 47      */
/* 
        ptr_vlan_tbl_cmd -> Select = j;
*/
   printk("%s: No select field in ptr_vlan_tbl_cmd structure!\n", __FUNCTION__);
        
/* Request/Acknowledge(31): 1-Request */
            
        ptr_vlan_tbl_cmd -> Req_ack = REQ;

/* Added MACRO for indirect register write 7/7/2004 */
        PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Status_Member[j]);
    

    }   

    
/* Selection(29:27) - Select one of 32 bit sections of the VLAN Table entry */
/* 011 -> Bridge_State for ports 48 to 50                                   */
/* 
    ptr_vlan_tbl_cmd -> Select = 3;
*/
   printk("%s: No select field in ptr_vlan_tbl_cmd structure!\n", __FUNCTION__);
        
/* Read/Write(30) : 1-Read */
    
    ptr_vlan_tbl_cmd -> Read_write = READPR;

/* Request/Acknowledge(31): 1-Request */
            
    ptr_vlan_tbl_cmd -> Req_ack = REQ;

/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    Temp &= 0xFFFFFFC0;
/* LENA decrease to 24+2 */
    Temp |= (Status_Member[(TOTAL_FE_PORTS/16)] & 0x3F);    


/* Configuration of VLAN Table Command register for writing */      
/* Read/Write(30) : 0-Write */
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    
/* Request/Acknowledge(31): 1-Request */
    ptr_vlan_tbl_cmd -> Req_ack = REQ;  
    

/* Added MACRO for indirect register write 7/7/2004 */
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);
#endif
    
    /*
     * read VLAN table entry at pos. vid and update membership information
     */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
    ptr_vlan_tbl_cmd -> Read_write = READPR;
    ptr_vlan_tbl_cmd -> Req_ack = REQ;
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    Temp &= ~VLAN_TABLE_MEMBER_MASK;
    Temp |= ((Status_Member[0] & 0x7) << VLAN_TABLE_MEMBER_SHIFT);      

    ptr_vlan_tbl_cmd -> Vlan_Id    = vid;
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;
    ptr_vlan_tbl_cmd -> Req_ack    = REQ;
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Status_Member[0]);
   
    return PR_ERR_NONE ;
    
}
#endif

#if 0
/**
 * Delete VLAN entry.
 * This function is used to delete the VLAN entry.
 *
 * \param   chipId            ChipId
 * \param   vlanId            VlanId(range is 0-4094)
 * \return  PR_ERR_NONE       - Success
 * \return  PR_ERR_VLANID     - Invalid VlanId  
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_delete_VLAN(u8 chipId,u32 vlanId)
{   
    u32 vid =0;
    s32 i=0;

#ifdef PARAMETER_CHECK

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif
        
    ifx_get_VLAN_idx(chipId,vlanId,&vid);

/* Update Hardware VLANTABLE */
    ifx_delete_VLAN_hardware(chipId,vid,vlanId);

    
/* Release the vid to free vid list */  
    ifx_add_vid(chipId,vid);
    
/* Update the software vlanid to vid association Table, 
        1 -> vid exits for VlanId
        0 -> no vid for VlanId      */
    PR_VLAN_VID_ASC_TBL[chipId][vlanId] = 0;
    
/* Update Software Vlanmembership and Portstate table  */
    for(i=0; i<=(TOTAL_PORTS - 1); i++)
    {
        PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].vlanmember = 0;
        PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(vid,i)].port_state = IFX_PORT_FORWARDING;
    }
#if 0    
/* Update the Hardware VLAN -> VID table to the default value */
    ifx_set_VLAN_idx(chipId,vlanId,DEFAULT_VID);    
#endif    

    return PR_ERR_NONE ;
}
#endif

#if 0
/**
 * Set port state.
 * Set the port state for the individual port in the vlan table 
 *
 * \param   chipId         ChipId
 * \param   vid            VID(range is 0-4094)
 * \param   portId         PortId(range is 0-2)
 * \param   portState      Port state 
 *                           - 0x0 Port is not member of VLAN and is not tagged
 *                           - 0x1 Port is member of VLAN and is not tagged
 *                           - 0x2 Port is not member of VLAN and is tagged
 *                           - 0x3 Port is member of VLAN and is tagged
 * \return  PR_ERR_NONE    - Success
 * \return  PR_ERR_VLANID  - Invalid VlanId  
 * \return  PR_ERR_CHIP_ID - Chip Id is invalid
 * \return  PR_ERR_NULL_PTR - Base address is null pointer  
 * \ingroup Internal
 * \internal
 */
s32 ifx_set_Vlan_table_portstate(u32 chipId,u32 vid,u32 portId, u32 portState)
{
    u32 Temp =0;
    s32  i=0;
    u32 Vlan_Mask=1;
    u32 shift_count =0;
    


/* Define a pointer to a structure & assign the base address of
   VLAN table command register */
    volatile vlan_tbl_cmd_reg_t*  ptr_vlan_tbl_cmd;

/* Define a pointer to VLAN Table Data register */
    u32* ptr_vlan_tbl_data ;

    volatile u32 temp1 = 0;
    u32* ptr_reg1;


#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    if(vid > MAX_VLANID)
        return PR_ERR_VLANID;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    ptr_vlan_tbl_data = (u32*) (Chip_Base_Address[chipId] + VLAN_TBL_DATA_REG);

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + VLAN_TBL_CMD_REG);

    ptr_vlan_tbl_cmd = (volatile vlan_tbl_cmd_reg_t*)&temp1;

/* Configuration of VLAN Table Command register for reading port state */
    ptr_vlan_tbl_cmd -> Vlan_Id = vid;
   
/* Read/Write(30) : 1-Read */

    ptr_vlan_tbl_cmd -> Read_write = READPR;

/* Request/Acknowledge(31): 1-Request */

    ptr_vlan_tbl_cmd -> Req_ack = REQ;


/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);


/* Generate the mask depending upon the PortId */
    Vlan_Mask = 3;
    shift_count = (portId & 0x3); 
    portState &= Vlan_Mask; /* Clear all unused bits */
    Vlan_Mask <<= shift_count;
    Vlan_Mask = ~ Vlan_Mask;

    Temp &= Vlan_Mask;


    Temp |= (portState << shift_count);


/* Configuration of VLAN Table Command register for writing */
/* Read/Write(30) : 0-Write */
    ptr_vlan_tbl_cmd -> Read_write = WRITEPR;

/* Request/Acknowledge(31): 1-Request */
    ptr_vlan_tbl_cmd -> Req_ack = REQ;

    
/* Added MACRO for indirect register write 7/7/2004 */
    PR_Indirect_Reg_Write(ptr_reg1,ptr_vlan_tbl_data,ptr_vlan_tbl_cmd,temp1,Temp);

    return PR_ERR_NONE ;

}
#endif


/**
 * Check VLAN existance.
 * This checks whether the passed vlan is created or not
 *
 * \param   chipId         ChipId
 * \param   vlanId         VLAN ID
 * \param   create         Pointer to bool, TRUE if created, FALSE otherwise
 * \return  PR_ERR_NONE    - Success
 * \return  PR_ERR_VLANID  - Invalid VLAN ID
 * \return  PR_ERR_CHIP_ID - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_vlan_check( u8 chipId, u32 vlanId, BOOL* create)
{

#ifdef PARAMETER_CHECK  
    
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif

        *create = (PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 1) ? TRUE : FALSE;

        return PR_ERR_NONE;

}


#if 0
/**
 * Check VLAN existance.
 * Compute the Hardware portstate for the vlan members
 *
 * \param   chipId         ChipId
 * \param   start_limit    portId range
 * \param   end_limit      portId range
 * \param   Status_Member  Pointer to Hardware port state
 * \param   vid            VID
 * \param   Temp           Membership
 * \param   portId_limit   Max port ID
 * \return  PR_ERR_NONE    - Success
 * \ingroup Internal
 * \internal
 */
s32 ifx_port_state( u8 chipId, s32 start_limit, s32 end_limit,
                    u32* Status_Member, u32 vid, u32 Temp, u32 portId_limit) 
{
    s32 i=0;
    u32 portState = 0;
    u8  membership = 0;
    u8  old_membership=0;

    for(i=start_limit; i>=end_limit; i--)
    {

        membership = ((Temp >> i ) & 0x1);
        
        *Status_Member <<= 2;

        if((PR_CHIP_MAJR_VER_ID[chipId] == MAJOR_ID_1) && (PR_CHIP_MINR_VER_ID[chipId] == MINOR_ID_3) )
        {
            portState = PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(DEFAULT_VID,portId_limit)].port_state;
            switch(portState)
            {
                case IFX_PORT_DISABLED:
                        portState = 0x0;
                        break;

                case IFX_PORT_BLOCKING:
                        portState = 0x1;
                        break;
                    
                case IFX_PORT_LISTENING:
                        portState = 0x1;
                        break;

                case IFX_PORT_LEARNING:
                    {
                        if(membership == 1)
                            portState = 0x2;
                        else
                            portState = 0x1;
                        break;
                    }

                case IFX_PORT_FORWARDING:
                    {
                        if(membership == 1)
                            portState = 0x3;
                        else
                            portState = 0x1;
                        break;
                    }
                default:
                        return PR_ERR_PORTSTATE;
                    
            
            }
            *Status_Member |= portState;
            
        }
        else
        {

            portState = (PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(DEFAULT_VID,portId_limit)].port_state < IFX_PORT_FORWARDING) ? 2 : 3;

            if(membership == 1)
                *Status_Member |= (portState * membership);
            else
                *Status_Member |= 2;

        }
        portId_limit -= 1;

    }/*End of for loop  */
    return PR_ERR_NONE;
}
#endif

/**
 * Get free VID from VID table.
 * This function returns a free Vlanindex(VID) from the software VID Table.
 *
 * \param   chipId            Chip Id
 * \param   vidPtr            Pointer to vlanindex(VID)
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_VLAN_FULL  - Vlan table is full
 * \return  PR_ERR_NULL_PTR   - vidPtr is null pointer
 * \return  PR_ERR_CHIP_ID    - Chip Id is invalid
 * \ingroup Internal
 * \internal
 */
s32 ifx_get_free_vid( u8 chipId, u32* vidPtr )
{

#ifdef PARAMETER_CHECK
    if(vidPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
#endif

    if(fill_level[chipId] == SW_VLAN_TABLE_FULL+1)
        return PR_ERR_VLAN_FULL;

    *vidPtr = fill_level[chipId];
    fill_level[chipId]+=1;  
    
    return PR_ERR_NONE ;
}



/**
 * Deletes entry from SW VLAN table
 * Deletes the entry located at position vlanIdx from the SW VLAN table.
 *
 * \param   chipId         ChipId
 * \param   vlanIdx        VLAN table index(range is 0-63) 
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_INDEX    - VlanIndex is invalid
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_delete_sw_vlan_table_entry( u8 chipId, u32 vlanIdx )
{  
    
#ifdef PARAMETER_CHECK
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    
    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;    

    if(vlanIdx > MAX_VID) 
        return PR_ERR_VLAN_INDEX;
#endif
	/*
	 * indicated entry as invalid and reduce fill level
	 */
    PR_VLAN_SW_TBL[chipId][vlanIdx].val = VLAN_ENTRY_INVALID;
    fill_level[chipId]-=1;
    
    return PR_ERR_NONE;
}


/**
 * Writes VLAN entry to SW VLAN table
 * Writes the passed VLAN entry at position vlanIdx into SW VLAN table .
 *
 * \param   chipId         ChipId
 * \param   vlanId         VID(range is 0-4094) 
 * \param   vlanIdx        VLAN table index(range is 0-63) 
 * \param   portmemberList pointer to port member information 
 *                         (bit 0=LAN port, bit 1=PC port, bit 2=CPU port)
 * \param   portegressList pointer to port egress tagging information
 *                         (bit 0=LAN port, bit 1=PC port, bit 2=CPU port)
 * \param   valid          entry valid information   
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_FULL     - Vlan table is full
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_VLAN_INDEX    - VLAN table index invalid
 * \return  PR_ERR_VLAN_VALID    - Vlan valid flag invalid
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address
 *                                 is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_sw_vlan_table_entry( u8 chipId, u32 vlanId, u32 vlanIdx,
                                 u32* portmemberList, u32* portegressList, u32 valid  )
{  
/*    
    printk("%s( portmemberList[0]=0x%08X, portegressList[0]=0x%08X)\n",
           __FUNCTION__, *portmemberList, *portegressList);
*/    
#ifdef PARAMETER_CHECK
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    
    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;    

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;          
    
    if(vlanIdx > MAX_VID) 
        return PR_ERR_VLAN_INDEX;

    if(portmemberList == NULL)
        return PR_ERR_NULL_PTR;
    
    if(portegressList == NULL)
        return PR_ERR_NULL_PTR;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
    
    if( (valid != TRUE) && (valid != FALSE) ) 
        return PR_ERR_VLAN_VALID;

#endif

    /*
     * if an invalid entry is overwritten increase the fill level,
     * otherwise an existing entry is overwritten and fill level remains unchanged
     */
/*     
    if( PR_VLAN_SW_TBL[chipId][vlanIdx].val == VLAN_ENTRY_INVALID )
        fill_level[chipId]+=1;
*/
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.valid      = valid;
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.egr_cpu    = (*portegressList & 0x00000004)>>2;
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.egr_pc     = (*portegressList & 0x00000002)>>1;
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.egr_lan    = (*portegressList & 0x00000001);   
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.member_cpu = (*portmemberList & 0x00000004)>>2;
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.member_pc  = (*portmemberList & 0x00000002)>>1;
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.member_lan = (*portmemberList & 0x00000001);        
    PR_VLAN_SW_TBL[chipId][vlanIdx].field.Vlan_Id    = vlanId;
/*
    printk("%s - added value 0x%08X at pos. %d to SW VLAN table\n", 
           __FUNCTION__, PR_VLAN_SW_TBL[chipId][vlanIdx].val, vlanIdx);
*/    
    return PR_ERR_NONE;
}


/**
 * Get VLAN table position
 * Retrieves the VLAN table position of the passed VLAN ID if it exists in the table.
 *
 * \param   chipId         ChipId
 * \param   vlanId         VID(range is 0-4094) 
 * \param   vidPtr         Pointer to VLAN table index 
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_EMPTY    - VLAN table is empty
 * \return  PR_ERR_VLAN_ENTRY_NOT_FOUND - VLAN entry not found in table
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_NULL_PTR      - vidPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_pos_sw_vlan_table( u8 chipId, u32 vlanId, u32* vidPtr )
{  
    u32 search_index;
    
#ifdef PARAMETER_CHECK
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    
    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;    

    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;  
    
    if(vidPtr == NULL)
        return PR_ERR_NULL_PTR;    
    
#endif
    /*
     * If SW VLAN table is not empty search the table for the passed VLAN ID 
     */
    if( fill_level[chipId] != SW_VLAN_TABLE_EMPTY )
    {
        for( search_index=0; search_index<NUMBER_VIDS; search_index++)
        {
            if( PR_VLAN_SW_TBL[chipId][search_index].field.Vlan_Id == vlanId)
            {
                *vidPtr = search_index;
                return PR_ERR_NONE;
            }
        }
    }
	else
		return PR_ERR_VLAN_EMPTY;
		
    printk("%s - vlanid %d not found in PR_VLAN_SW_TBL\n", __FUNCTION__, vlanId);
    return PR_ERR_VLAN_ENTRY_NOT_FOUND;
}

#if 0
void print_sw_vlan_table(u8 chipId)
{
    s32 table_index;
    
    printk("SW VLAN table\n\n");        
    printk("                                Egress Tagging  Membership\n");    
    printk("index     VLAN ID      || valid   CPU  PC  LAN   CPU  PC  LAN    \n");
    printk("------------------------------------------------------\n");       
    for( table_index=0; table_index<NUMBER_VIDS; table_index++ )
    {
        printk("[%3d] |  0x%03x (%4d)  ||   %c   |%c|%c|%c    %c|%c|%c|\n", 
               table_index, 
               PR_VLAN_SW_TBL[chipId][table_index].field.Vlan_Id, 
               (int)PR_VLAN_SW_TBL[chipId][table_index].field.Vlan_Id, 
               (PR_VLAN_SW_TBL[chipId][table_index].field.valid) ?'1':'0',  /* valid */
               (PR_VLAN_SW_TBL[chipId][table_index].field.egr_cpu)?'+':' ',   /* egress CPU */           
               (PR_VLAN_SW_TBL[chipId][table_index].field.egr_pc)?'+':' ',   /* egress LAN */
               (PR_VLAN_SW_TBL[chipId][table_index].field.egr_lan)?'+':' ',   /* egress PC */
               (PR_VLAN_SW_TBL[chipId][table_index].field.member_cpu)?'+':' ',   /* Mmebership CPU */ 
               (PR_VLAN_SW_TBL[chipId][table_index].field.member_pc)?'+':' ',   /* Mmebership LAN */ 
               (PR_VLAN_SW_TBL[chipId][table_index].field.member_lan)?'+':' ');    /* Mmebership PC */         
        
    }

}
#endif

/**
 * Upodate PR_VLAN_SW_TBL table positions
 * Updates the table positions in the PR_VLAN_VID_TBL for the VLAN IDs used.
 * This function needs to be called everytime the PR_VLAN_SW_TBL has been sorted.
 * This updating is necessary because the table sort may have changed 
 * the positions of the vlan entries in the PR_VLAN_SW_TBL.   
 *
 * \param   chipId               ChipId
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_NOTCONFG - found an entry that has a valid VLAN-ID 
 *                                 but is indicated invalid in the 
 *                                 PR_VLAN_VID_ASC_TBL
 * \ingroup Internal
 */ 
s32 ifx_update_table_pos( u8 chipId )
{
    s32 sw_vlan_tbl_pos;
    s32 vlan_vid_tbl_pos=0;
    
    for( sw_vlan_tbl_pos=0; sw_vlan_tbl_pos<NUMBER_VIDS; sw_vlan_tbl_pos++ )
    {
        if(PR_VLAN_SW_TBL[chipId][sw_vlan_tbl_pos].val != 0x7FFFF)
        {
            /*
             * for valid PR_VLAN_SW_TBL entries get the assigned VLAN ID
             * which is the index for PR_VLAN_VID_ASC_TBL to check.
             * The PR_VLAN_VID_ASC_TBL table entry has to be valid and
             * its position in the PR_VLAN_VID_TBL table has to be updated.
             */
            vlan_vid_tbl_pos = PR_VLAN_SW_TBL[chipId][sw_vlan_tbl_pos].field.Vlan_Id;
            if(PR_VLAN_VID_ASC_TBL[chipId][vlan_vid_tbl_pos] == 1)
            {
                PR_VLAN_VID_TBL[chipId][vlan_vid_tbl_pos] = sw_vlan_tbl_pos;
/*
                printk("%s - Updated PR_VLAN_VID_TBL[chipId][%d]=%d\n", 
                       __FUNCTION__, vlan_vid_tbl_pos, sw_vlan_tbl_pos);
*/                       
            }
            else
            {   
                printk("%s - vlanid %d indicated inactive in PR_VLAN_VID_ASC_TBL\n",
                       __FUNCTION__, vlan_vid_tbl_pos);
                return PR_ERR_VLAN_NOTCONFG;
            }
            
        }    
    }
    return PR_ERR_NONE;
}


/**
 * Sort SW VLAN table
 * Sorts the entries of the SW VLAN table for ascending VLAN IDs.
 * Invalid entries indicated by VLAN-ID 0xFFF will be sorted to the tail of the table.
 * Bubble sort is used as sorting algorithm.
 *
 * \param   chipId               ChipId
 * \ingroup Internal
 */
void ifx_sort_sw_vlan_table( u8 chipId )
{
    s32 i=0;
    s32 j=0;
    vlan_tbl_entry_u temp;
/*
    printk("SW VLAN table before sort\n");
    print_sw_vlan_table( chipId );
*/    
    for( i = (NUMBER_VIDS-1); i >= 0; i--)
    {
        for( j = 1; j <= i; j++ )
        {
            if( PR_VLAN_SW_TBL[chipId][j-1].field.Vlan_Id > 
                PR_VLAN_SW_TBL[chipId][j].field.Vlan_Id )
            {
                temp = PR_VLAN_SW_TBL[chipId][j-1];
                PR_VLAN_SW_TBL[chipId][j-1] = PR_VLAN_SW_TBL[chipId][j];
                PR_VLAN_SW_TBL[chipId][j] = temp;
            }
        }
    }
    ifx_update_table_pos( chipId );
/*    
    printk("SW VLAN table after sort\n");
    print_sw_vlan_table( chipId );
*/


}

/**
 * Copy SW VLAN table content to hardware VLAN table
 * This function copies the valid entries of the PR_VLAN_SW_TBL table into the 
 * HW VLAN table. Afterwards the remaining entries of the HW table are filled 
 * with invalid data.
 * 
 *
 * \param   chipId               ChipId
 * \ingroup Internal
 */
s32 ifx_copy_sw_vlan_table_to_hw( u8 chipId )
{  
    int retval;
    u32 table_index;
    u32 port_members = 0x00000000;
    u32 port_egress  = 0x00000000;
/*
    printk("%s - fill_level[chipId]=%d\n", __FUNCTION__, fill_level[chipId]);
*/    
    for( table_index=0; table_index<fill_level[chipId]; table_index++ )
    {
        
        port_egress  = (u32) PR_VLAN_SW_TBL[chipId][table_index].field.egr_lan |
                       ((u32) PR_VLAN_SW_TBL[chipId][table_index].field.egr_pc)<<1 |
                       ((u32) PR_VLAN_SW_TBL[chipId][table_index].field.egr_cpu)<<2;
        port_members = (u32) PR_VLAN_SW_TBL[chipId][table_index].field.member_lan |
                       ((u32) PR_VLAN_SW_TBL[chipId][table_index].field.member_pc)<<1 |
                       ((u32) PR_VLAN_SW_TBL[chipId][table_index].field.member_cpu)<<2;
/*
        printk("%s -> ifx_set_vlan_table_entry( vlan_id=%d, table_index=%d, \n",
               __FUNCTION__, PR_VLAN_SW_TBL[chipId][table_index].field.Vlan_Id, table_index);
        printk("                       port_members=0x%08X, port_egress=0x%08X, valid=%d)\n",
                port_members, port_egress, PR_VLAN_SW_TBL[chipId][table_index].field.valid);
*/                
        if( (retval=ifx_set_vlan_table_entry( chipId, 
                                              (u32) PR_VLAN_SW_TBL[chipId][table_index].field.Vlan_Id,
                                              table_index,
                                              &port_members, 
                                              &port_egress, 
                                              (u32) PR_VLAN_SW_TBL[chipId][table_index].field.valid )) != PR_ERR_NONE )
        {
            printk("Writing VLAN entry to position %d failed !\n", table_index);            
            return retval;
        }
    
    
    }
    /*
     * fill all remaining entries with invalid data
     */
    port_members = 0x00000007;
    port_egress  = 0x00000007;
    for( table_index=fill_level[chipId]; table_index<TOTAL_VIDS; table_index++ )
    {
        if( (retval=ifx_set_vlan_table_entry( chipId, 
                                              0xFFF,
                                              table_index,
                                              &port_members, 
                                              &port_egress, 
                                              0x00000001 )) != PR_ERR_NONE )
        {
            printk("Writing invalid VLAN entry to position %d failed !\n", table_index);
            return retval;
        }
    }    
    return PR_ERR_NONE;
}



/**
 * Add VLAN entry to VLAN table
 * Adds the passed VLAN entry to the VLAN table. If the table already contains 
 * an entry that has the passed VLAN ID assigned this entry will be overwritten 
 * with the passed parameters.  
 *
 * \param   chipId         ChipId
 * \param   vlanId         VID(range is 0-4094) 
 * \param   portmemberList pointer to port member information 
 *                         (bit 0=LAN port, bit 1=PC port, bit 2=CPU port)
 * \param   portegressList pointer to port egress tagging information
 *                         (bit 0=LAN port, bit 1=PC port, bit 2=CPU port)
 * \param   valid          entry valid information   
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLAN_FULL     - Vlan table is full
 * \return  PR_ERR_VLANID        - Vlan Id is invalid
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_VLAN_UNAWARE  - Vlan Unaware mode is set
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address 
 *                                 is null pointer
 * \return  PR_ERR_TIME_OUT      - Time out
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_add_vlan_entry( u8 chipId, u32 vlanId, 
                        u32* portmemberList, u32* portegressList, u32 valid )
{
    s32 retval;
    u32 vlanIdx;
	
#ifdef PARAMETER_CHECK  
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
	
/*
    printk("%s( chipId=%d, vlanId=%d, \n", __FUNCTION__, chipId, vlanId);
    printk("portmemberList[0]=0x%08X, portegressList[0]=0x%08X, valid=%d )\n", 
           *portmemberList, *portegressList, valid);
*/    
    if( PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 1 ) 
    {
        printk("%s - VLAN-ID %d already in use -> overwrite entry\n", __FUNCTION__,vlanId);
        /*
         * VLAN ID is already known, get index of existing entry that has 
         * to be overwritten
         */ 
         if( (retval=ifx_get_pos_sw_vlan_table( chipId, vlanId, 
                                                &vlanIdx )) != PR_ERR_NONE )
             return retval;

        /*
         * update SW VLAN table with new entry and 
         * copy content to HW VLAN table
         */
        if( (retval=ifx_set_sw_vlan_table_entry( chipId, vlanId, vlanIdx,
                                                 portmemberList, portegressList, 
                                                 valid  )) != PR_ERR_NONE)
            return retval;   
        
        if( (retval=ifx_copy_sw_vlan_table_to_hw( chipId ))  != PR_ERR_NONE )
            return retval;              
        
        printk("VLAN-ID %d updated in VLAN table\n", vlanId);                    
    }    
    else
    {
/*
        printk("%s - VLAN-ID %d not in use -> add new VLAN table entry\n",
               __FUNCTION__,vlanId);
*/               
        /*
         * VLAN ID is unknown, get free table index for new entry that 
         * has to be added to table
         */
        if( (retval=ifx_get_free_vid( chipId, &vlanIdx )) != PR_ERR_NONE )
            return retval;       
/*        printk("%s - got free table pos. %d\n", __FUNCTION__,vlanIdx);  */      
        /*
         * update SW VLAN table with new entry, indicate usage, sort table and 
         * copy content to HW VLAN table
         */
        if( (retval=ifx_set_sw_vlan_table_entry( chipId, vlanId, vlanIdx,
                                                 portmemberList, portegressList, 
                                                 valid  )) != PR_ERR_NONE)
            return retval;        
        PR_VLAN_VID_ASC_TBL[chipId][vlanId] = 1;
        ifx_sort_sw_vlan_table( chipId );
        
        if( (retval=ifx_copy_sw_vlan_table_to_hw( chipId ))  != PR_ERR_NONE )
            return retval;  
        
        /*
         * get table index after table sort and update registration
         */
        if( (retval=ifx_get_pos_sw_vlan_table( chipId, vlanId, 
                                               &vlanIdx )) != PR_ERR_NONE )
            return retval;

        PR_VLAN_VID_TBL[chipId][vlanId] = vlanIdx;
        printk("VLAN-ID %d added to VLAN table\n", vlanId);        
    }        
    
    return retval;
    
    
}

/**
 * Delete VLAN entry from VLAN table
 * Removes the passed VLAN ID from the VLAN table. The function first gets 
 * the VLAN table position for the passed VLAN ID. Afterwards the entry 
 * is removed from the SW VLAN table, the table is re-sorted and 
 * copied to the HW VLAN table. Finally the removed entry is marked as invalid.
 *
 * \param   chipId         ChipId
 * \param   vlanId         VID(range is 0-4095) 
 *                              
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_VLANID        - Vlan Id is invalid or not existing in list
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured 
 * \return  PR_ERR_NULL_PTR      - portmemberList, portegressList or base address is null pointer
 * \return  PR_ERR_CHIP_ID       - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_delete_vlan_entry( u8 chipId, u32 vlanId )
{
    s32 retval=-1;
    u32 vlanIdx;

#ifdef PARAMETER_CHECK  
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    
    if( PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 1 ) 
    {
        /* 
         * VLAN ID to be deleted is in the PR_VLAN_VID_ASC_TBL 
         * -> get table position for passed VLAN ID
         * -> delete entry from SW VLAN table
         * -> re-sort SW VLAN table
         * -> copy content of SW VLAN table to HW VLAN table
         * -> remove deleted entry from registration lists           
         */
        if( (retval=ifx_get_pos_sw_vlan_table( chipId, vlanId, 
                                               &vlanIdx )) != PR_ERR_NONE )
            return retval; 
			       
        if( (retval=ifx_delete_sw_vlan_table_entry( chipId, vlanIdx ))  != PR_ERR_NONE )
            return retval; 
			
        ifx_sort_sw_vlan_table( chipId );   
		
        if( (retval=ifx_copy_sw_vlan_table_to_hw( chipId ))  != PR_ERR_NONE )
            return retval;  
        PR_VLAN_VID_ASC_TBL[chipId][vlanId] = 0;
        PR_VLAN_VID_TBL[chipId][vlanId]     = DEFAULT_VID;
    }
    else
    {
        /* the VLAN ID to be deleted does not exist in the PR_VLAN_VID_ASC_TBL */
        retval=PR_ERR_VLAN_NOTCONFG;
    }
    return retval; 

}

