/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
#include <linux/kernel.h>
#include <asm/types.h>
#include <asm/addrspace.h>
#include "struct.h"
#include "const.h"
#include "fe_port.h"
#include <asm/incaip2/incaip2.h>
#include "err.h"
#include "var.h"
#include <asm/incaip2/ifx_types.h>
#include "phy_prv.h"
#include "phy.h"
#include "ports.h"
#include "fe_port_prv.h"
#include "monitoring.h"
#include "api_init_prv.h"
#ifdef _VIPER
#include "vdslapi.h"
#endif

#undef INCAIP2_DEBUG_FEPORT

  
/**
 * Configure the operation mode of the individual FE port.
 *
 * \param   chipId          Chip Id
 * \param   portId          PortId(range is 0-1)
 * \param   feature		    Feature to be configured for the port
 *                          The allowed features are:
 *							    0.FE_PORT_AUTONEGOTIATION
 *							    2.FE_PORT_DUPLEX_MODE
 *							    3.FE_PORT_SPEED
 *							    4.FE_PORT_PAUSE_FLOW_CTRL
 *							    5.FE_PORT_PRIORITY
 *							    6.FE_PORT_MONITOR_INGRESS
 *							    7.FE_PORT_MONITOR_EGRESS
 *							    8.FE_PORT_AUTHENTICATION 
 * \param   mode		    Value of the feature 
 *
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the PortId is not in the range of 0 to 1
 * \return  PR_ERR_FE_PORT_SET	 - If feature is not in the range of 0 - 8 
 * \return  PR_ERR_TIME_OUT      - Time out 
 * \return  PR_ERR_NULL_PTR      - base address is null pointer
 * \return  PR_ERR_CHIP_ID       - If the chip Id invalid
 * \return  PR_ERR_PHY_RD			If Phy register read fails
 * \return  PR_ERR_PHY_WR			If Phy register write fails
 * \return  PR_ERR_PAUSE_HD		If the duplex mode is half then flow control can't be set
 * \return  PR_ERR_PAUSE_AN_DIS		If Auto-negotiation is disabled then flow control can't be set 
 * \ingroup Internal
 */  
s32 ifx_set_port_FE_mode(u8 chipId,u32 portId, u32 feature, u32 mode)
{
	u32	    dwTmp=0;
    u8      phy_addr=255;
	u32     shift=0;
	BOOL   	phy_auto;
	u32*	ptr_ge_mac;  
	u32*    ptr_port_auth;
	s32     retflag = PR_ERR_NONE;
#ifdef BLOCK_WHEN_LINK_DOWN
	u16		phyData1 =0; 
#endif

#ifdef INCAIP2_DEBUG_FEPORT
    printk("%s(chipId=%d, portId=%d, feature=%d, mode=0x%08X)\n", 
           __FUNCTION__, chipId, portId, feature, mode);		
#endif

#ifdef PARAMETER_CHECK

	if(portId >= TOTAL_FE_PORTS)
		return PR_ERR_PORTID;

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif  
	
    if( portId == INCAIP2_SWITCH_PORT_SELECT_LAN )
        ptr_ge_mac = (u32*) (Chip_Base_Address[chipId] + SWITCH_P0_GMAC_REG);
    else
        ptr_ge_mac = (u32*) (Chip_Base_Address[chipId] + SWITCH_P1_GMAC_REG);    

    ptr_port_auth = (u32*) (Chip_Base_Address[chipId] + PORT_AUTH); 

    if((retflag = ifx_get_portId_phyId( chipId,
                                        portId,
                                        &phy_addr )) != PR_ERR_NONE)
    {
        return retflag;
    }

#ifdef BLOCK_WHEN_LINK_DOWN
    /* If the link is down don't allow the following configurations */
    if((feature == FE_PORT_AUTONEGOTIATION) ||
       (feature == FE_PORT_DUPLEX_MODE) ||
       (feature == FE_PORT_SPEED) ||
       (feature == FE_PORT_PAUSE_FLOW_CTRL) )
    {
        if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
            return retflag;

        if (ifx_phy_read( chipId, phy_addr, PHY_STATUS, &phyData1 ) != PR_ERR_NONE)
        {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
        }
/*
        printk("%s -> ifx_phy_read(chipId=%d, phy_addr=%d, reg=%d) returned 0x%04X\n",
               __FUNCTION__, chipId, phy_addr, PHY_STATUS, phyData1);
*/
        if( ((phyData1 >> 2) & 0x01) == 0 )
        {
            printk("%s -> link is down\n", __FUNCTION__);
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_LINK_DOWN;
        }

        if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
            return retflag;
    }
#endif

	switch (feature) 
	{
		case FE_PORT_AUTONEGOTIATION:   
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_AUTONEGOTIATION\n", __FUNCTION__);
#endif
            mode &= 0x01;
            return(ifx_set_port_autoneg( chipId, portId, (BOOL)mode));

		case FE_PORT_DUPLEX_MODE:	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_DUPLEX_MODE\n", __FUNCTION__);
#endif
            /* advertising autonegotiation info */
			mode &= 0x1;
			
            if((retflag = ifx_get_port_autoneg(chipId,portId,&phy_auto)) != PR_ERR_NONE)
				return retflag;
            
			if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;
            
			if(phy_auto == FALSE)
			{
                /* 
                 * Autonegotiation is disabled 
                 * -> write duplex mode to PHY CTRL register and additionally
                 * re-disable ANEG to perform a PHY reset which forces the PHY
                 * to take over duplex configuration
                 */
#ifdef INCAIP2_DEBUG_FEPORT
                printk("%s - AUTONEGOTIATION OFF\n", __FUNCTION__);
#endif
				ifx_phy_set_duplex_mode( chipId, portId, (BOOL)mode );
                ifx_set_port_autoneg( chipId, portId, FALSE );
			}
			else
			{
                /* 
                 * Autonegotiation is enabled 
                 * -> update duplex/speed content of PHY Autonegotiation 
                 * Advertisement Register and internal registration 
                 */                
#ifdef INCAIP2_DEBUG_FEPORT
                printk("%s - AUTONEGOTIATION ON\n", __FUNCTION__);
#endif
				cache_duplex[chipId][portId] = mode;
				if(ifx_set_feport_duplex_speed_add ( chipId,  phy_addr, 
                                                     cache_speed[chipId][portId], 
                                                     cache_duplex[chipId][portId]) )
				{
					ifx_sem_give(chipId,SEM_MDIO);
					return PR_ERR_FE_PORT_SET;	
				}
			}
			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    			return retflag;
			break;

		case FE_PORT_SPEED:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_SPEED\n", __FUNCTION__);
#endif
			if (mode < 0 || mode > 1)
				return PR_ERR_FE_PORT_SET;
			mode &= 0x1;
			
			if((retflag = ifx_get_port_autoneg(chipId,portId,&phy_auto)) != PR_ERR_NONE)
				return retflag;

			if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;
			
			if(phy_auto == FALSE)
			{
                /* 
                 * Autonegotiation is disabled 
                 * -> write speed mode to PHY CTRL register
                 */                
				ifx_phy_speed_fe( chipId, portId, (u8)mode );
			}
			else
			{
                /* 
                 * Autonegotiation is enabled 
                 * -> update duplex/speed content of PHY Autonegotiation 
                 * Advertisement Register and internal registration 
                 */                     
				cache_speed[chipId][portId] = mode;
				if ( ifx_set_feport_duplex_speed_add( chipId, phy_addr, 
                                                      cache_speed[chipId][portId],
                                                      cache_duplex[chipId][portId]) )
				{
					ifx_sem_give(chipId,SEM_MDIO);
					return PR_ERR_FE_PORT_SET;
				}
			}
			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    			return retflag;
			break;


		case FE_PORT_PAUSE_FLOW_CTRL: 				
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_PAUSE_FLOW_CTRL\n", __FUNCTION__);
#endif
			if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;										
											
#ifdef _VIPER
			//enable FC,BP and BS in the VDSL port
			Mii_EnableDataFlowControl(portId,enLocal,mode);
            
#endif //_VIPER
			if((retflag = ifx_phy_pause_fe(chipId,portId,(BOOL)mode)) != PR_ERR_NONE)
			{
				ifx_sem_give(chipId,SEM_MDIO);
                printk("%s -> ifx_phy_pause_fe() failed, returned %d\n", 
                       __FUNCTION__, retflag);	
				return retflag;
			}
			
			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    			return retflag;
			break;

		case FE_PORT_PRIORITY: 			
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_PRIORITY\n", __FUNCTION__);
#endif
            return(ifx_set_port_CoS( chipId, portId, mode ) ) ;

		case FE_PORT_MONITOR_INGRESS: 	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_MONITOR_INGRESS\n", __FUNCTION__);
#endif
            return(ifx_set_ingress_monitor_flag( chipId, portId, mode) );
				
		case FE_PORT_MONITOR_EGRESS: 	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_MONITOR_EGRESS\n", __FUNCTION__);
#endif
            return(ifx_set_egress_monitor_flag( chipId, portId, mode) );
			
		case FE_PORT_AUTHENTICATION:	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_AUTHENTICATION\n", __FUNCTION__);
#endif
			if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;
            
			PR_Reg_Read( ptr_port_auth, dwTmp) ;
			shift = (portId & 0xF) << 1;
			dwTmp &= ~(0x03 << shift) ;
			dwTmp |= ((mode & 0x03) << shift);
			PR_Reg_Write(ptr_port_auth, dwTmp) ;
            
			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    			return retflag;
			break;

		default:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  Invalid Mode\n", __FUNCTION__);
#endif
			return PR_ERR_FE_PORT_SET;
	}

	return PR_ERR_NONE;
}


/**
 * Get the operation mode of the individual FE port.
 *
 * \param   chipId          Chip Id
 * \param   portId          PortId(range is 0-1)
 * \param   feature		    One of the below features that need to be read 
 *                          for the port. The allowed features are:
 *							    0.FE_PORT_AUTONEGOTIATION
 *                              1.FE_PORT_LINK_OK
 *							    2.FE_PORT_DUPLEX_MODE
 *							    3.FE_PORT_SPEED
 *							    4.FE_PORT_PAUSE_FLOW_CTRL
 *							    5.FE_PORT_PRIORITY
 *							    6.FE_PORT_MONITOR_INGRESS
 *							    7.FE_PORT_MONITOR_EGRESS
 *							    8.FE_PORT_AUTHENTICATION 
 * \param   modePtr		    Pointer to value of the feature
 *
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the PortId is not in the range of 0 to 1
 * \return  PR_ERR_FE_PORT_SET	 - If feature is not in the range of 0 - 8 
 * \return  PR_ERR_NULL_PTR      - base address is null pointer
 * \return  PR_ERR_CHIP_ID       - If the chip Id invalid
 * \return  PR_ERR_PHY_RD			If Phy register read fails
 * \return  PR_ERR_PHY_WR			If Phy register write fails
 * \ingroup Internal
 */  
s32 ifx_get_port_FE_mode( u8 chipId, u32 portId, u32 feature, u32 * modePtr )
{
	u32     dwTmp=0;
	u32     shift=0;
    BOOL    en_hw_mode=0;
    BOOL    phy_auto=0;
	u8      phy_addr=0;
	u32*	ptr_ge_mac;  
	u32*    ptr_port_auth;
    u32*    ptr_cpu_link_ok;
	s32     retflag = PR_ERR_NONE;
	u16		phyData4 =0;
	u16		phyData5 =0;
	
#ifdef INCAIP2_DEBUG_FEPORT
    printk("%s(chipId=%d, portId=%d, feature=%d)\n", 
           __FUNCTION__, chipId, portId, feature);
#endif
           
#ifdef PARAMETER_CHECK

	if(portId >= TOTAL_FE_PORTS)
		return PR_ERR_PORTID;
	if(modePtr == NULL)
		return PR_ERR_NULL_PTR;

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

	if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
		return PR_ERR_NULL_PTR;

#endif  
    
    ptr_cpu_link_ok = (u32*) (Chip_Base_Address[chipId] + CPU_LINK_OK_REG);
    
    if( portId == INCAIP2_SWITCH_PORT_SELECT_LAN )
        ptr_ge_mac = (u32*) (Chip_Base_Address[chipId] + SWITCH_P0_GMAC_REG);
    else
        ptr_ge_mac = (u32*) (Chip_Base_Address[chipId] + SWITCH_P1_GMAC_REG);
    
    ptr_port_auth = (u32*) (Chip_Base_Address[chipId] + PORT_AUTH); 


    /* get MDIO HW state machine state and ANEG state
       for determination of PHY link duplex and speed */
    if((retflag = ifx_get_MDIO_mode( chipId, portId,
                                     &en_hw_mode, &phy_auto)) != PR_ERR_NONE)
    {
        return retflag;
    }
    if((retflag = ifx_get_port_autoneg( chipId, portId, &phy_auto)) != PR_ERR_NONE)
    {
        printk("%s - failed to read ANEG state\n", __FUNCTION__ );
        return retflag;
    }

#ifdef INCAIP2_DEBUG_FEPORT
    if(en_hw_mode == TRUE)
        printk("%s - MDIO HW State Machine is enabled\n", __FUNCTION__ );
    else
        printk("%s - MDIO HW State Machine is disabled\n", __FUNCTION__ );
    if(phy_auto == TRUE)
        printk("%s - Auto Negotiation is enabled\n", __FUNCTION__ );
    else
        printk("%s - Auto Negotiation is disabled\n", __FUNCTION__ );
#endif

    if((retflag = ifx_get_portId_phyId( chipId, portId, &phy_addr )) != PR_ERR_NONE)
    {
        return retflag;
    }

    switch (feature)
    {
        case FE_PORT_AUTONEGOTIATION:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_AUTONEGOTIATION\n", __FUNCTION__);
#endif
            *((BOOL*)modePtr) = phy_auto;
            return PR_ERR_NONE;

		case FE_PORT_LINK_OK:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_LINK_OK\n", __FUNCTION__);
#endif
            if( en_hw_mode == TRUE && phy_auto == TRUE )
            { 
                /*
                 * HW autonegotiation enabled for this port
                 * -> read and return link status from CPU_LINK_OK register
                 */
				PR_Reg_Read(ptr_ge_mac, dwTmp);	
#ifdef INCAIP2_DEBUG_GEPORT
                printk("%s - read 0x%08X from SWITCH_GMAC_REG for port %d\n",
                       __FUNCTION__, dwTmp, portId);
#endif
				*modePtr = (dwTmp >> GE_PORT_LINK_POS) & 0x1;
				
#if 0				 
                PR_Reg_Read(ptr_cpu_link_ok, dwTmp);
#ifdef INCAIP2_DEBUG_FEPORT
                printk("%s - read 0x%08X from CPU_LINK_OK_REG\n", 
                       __FUNCTION__, dwTmp);
#endif
				shift = portId; /*(portId & PORTID_MASK) + PORT_LINK_POS;  */
				*modePtr = (dwTmp >> shift) & 0x01;
#endif				
			}
			else
			{
                /*
                 * HW autonegotiation disabled for this port 
                 * -> read link status from PHY status register and update 
                 * GMAC_REG link status bit
                 */                
				if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
					return retflag;
				 
                if (ifx_phy_read(chipId, phy_addr, PHY_STATUS, &phyData4) != 0)
					return PR_ERR_PHY_RD;											
				
				*modePtr = (phyData4 >> 2) & 0x01; 
            
				PR_Reg_Read(ptr_ge_mac, dwTmp);
#ifdef INCAIP2_DEBUG_GEPORT
                printk("%s - read 0x%08X from SWITCH_GMAC_REG for port %d\n",
                       __FUNCTION__, dwTmp, portId);
#endif
				shift = GE_PORT_LINK_POS; /* (portId & PORTID_MASK) + PORT_LINK_POS;  */
				dwTmp &= ~(1 << shift);
				dwTmp |= (*modePtr << shift);
				PR_Reg_Write(ptr_ge_mac, dwTmp);
            
#ifdef INCAIP2_DEBUG_GEPORT
                printk("%s - wrote 0x%08X to SWITCH_GMAC_REG for port %d\n",
                       __FUNCTION__, dwTmp, portId);
#endif

				if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
					return retflag;
			}
			break;

		case FE_PORT_DUPLEX_MODE:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_DUPLEX_MODE\n", __FUNCTION__);
#endif
			if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;
			
			if( (phy_auto == FALSE) && (en_hw_mode == FALSE) )
			{
                /*
                 * HW autonegotiation disabled and MDIO not used for this port
                 * -> read and return duplex status from PHY Control register.
                 * Update GMAC_REG duplex information accordingly.
                 */                
				if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &phyData4) != PR_ERR_NONE)
				{
					ifx_sem_give(chipId,SEM_MDIO);
					return PR_ERR_PHY_RD;
				}
				
				*modePtr = (phyData4 >> PHY_DUPLEX_MODE_POS) & 0x01;
				
				PR_Reg_Read(ptr_ge_mac, dwTmp);
				dwTmp &= ~(1 << GMAC_DUPLEX_POS);
				dwTmp |= (*modePtr << GMAC_DUPLEX_POS);
                PR_Reg_Write(ptr_ge_mac, dwTmp);
			}
			else
				if(en_hw_mode == TRUE)
				{
                    /*
                     * get duplex mode from GMAC_REG
                     */
					PR_Reg_Read(ptr_ge_mac, dwTmp) ;
					shift = GMAC_DUPLEX_POS;
					*modePtr = (dwTmp >> shift) & 0x01;
				}
				else
				{
                    /*
                     * Generate duplex mode from PHY_ANAR and PHY_ANLPA 
                     * registers content and update GMAC_REG duplex information
                     */                    
					if (ifx_phy_read( chipId, phy_addr, PHY_ANAR, &phyData4 ) != 0)
					{
						ifx_sem_give(chipId,SEM_MDIO);
						return PR_ERR_PHY_RD;
					}
				
					if (ifx_phy_read( chipId, phy_addr, PHY_ANLPA, &phyData5 ) != 0)
					{
						ifx_sem_give(chipId,SEM_MDIO);
						return PR_ERR_PHY_RD;
					}
                    /*
                     * perform bit AND combination of duplex/speed bits taken from 
                     * PHY_ANAR and PHY_ANLPA registers and extract duplex information
                     */
					dwTmp = ((phyData4 >> 5) & 0xF) & ((phyData5 >> 5) & 0xF);
                    /* check for 100 Base TX FD */
					if( ((dwTmp >> 3) & 0x1) == 0x1)
						*modePtr = IFX_FULL_DUPLEX;
					else
                        /* check for 100 Base TX HD */
						if( ((dwTmp >> 2) & 0x1) == 0x1)		
							*modePtr = IFX_HALF_DUPLEX;
						else
                            /* check for 10 Base T FD */
							if( ((dwTmp >> 1) & 0x1) == 0x1)		
								*modePtr = IFX_FULL_DUPLEX;
							else
								*modePtr = IFX_HALF_DUPLEX;
            
					PR_Reg_Read(ptr_ge_mac, dwTmp);
					shift = GMAC_DUPLEX_POS;
					dwTmp &= ~(1 << shift);
					dwTmp |= (*modePtr << shift);
					PR_Reg_Write(ptr_ge_mac, dwTmp);
					
				}
			
			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    			return retflag;
            
			break;

		case FE_PORT_SPEED:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_SPEED\n", __FUNCTION__);
#endif
            if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
                return retflag;
            
			if( (phy_auto == FALSE) && (en_hw_mode == FALSE) )
			{
				/* Read from Reg 0 */
                if (ifx_phy_read(chipId,phy_addr, PHY_CTRL, &phyData4) != PR_ERR_NONE)
				{
					ifx_sem_give(chipId,SEM_MDIO);
					return PR_ERR_PHY_RD;
				}
				*modePtr = (phyData4 >> PHY_FE_SPEED_POS) & 0x01;
            
                /* 
                 * update GMAC register with speed read from PHY 
                 */
				PR_Reg_Read(ptr_ge_mac, dwTmp);
/*				shift = (portId & PORTID_MASK) + 0;  */
				dwTmp &= ~(3 << shift);
				dwTmp |= ((*modePtr&0x1) << 1);
                PR_Reg_Write(ptr_ge_mac, dwTmp);
			}
			else
            {
				if( (en_hw_mode == TRUE) )  /* && ( phy_auto == TRUE ) )  */  
				{
                    PR_Reg_Read(ptr_ge_mac, dwTmp);
#ifdef INCAIP2_DEBUG_FEPORT
                    printk("%s - read 0x%08X from Port %d GMAC_REG\n",
                           __FUNCTION__, dwTmp, portId);
#endif
                    *modePtr = (dwTmp >> 1) & 0x03;
				}
                else
				{
                    /*
                     * Generate speed information from PHY_ANAR and PHY_ANLPA 
                     * registers content and update GMAC_REG speed information
                     */                     
					if (ifx_phy_read( chipId, phy_addr, PHY_ANAR, &phyData4 ) != 0)
					{
						ifx_sem_give(chipId,SEM_MDIO);
						return PR_ERR_PHY_RD;
					}
				
					if (ifx_phy_read(chipId, phy_addr, PHY_ANLPA, &phyData5) != 0)
					{
						ifx_sem_give(chipId,SEM_MDIO);
						return PR_ERR_PHY_RD;
					}
                    /*
                     * perform bit AND combination of duplex/speed bits taken from 
                     * PHY_ANAR and PHY_ANLPA registers and extract speed information
                     */            
					dwTmp = ((phyData4 >> 5) & 0xF) & ((phyData5 >> 5) & 0xF);
            
					if( ((dwTmp >> 3) & 0x1) == 0x1)
						*modePtr = IFX_SPEED_100_MBIT;
					else
						if( ((dwTmp >> 2) & 0x1) == 0x1)		
							*modePtr = IFX_SPEED_100_MBIT;
						else
							if( ((dwTmp >> 1) & 0x1) == 0x1)		
								*modePtr = IFX_SPEED_10_MBIT;
							else
								*modePtr = IFX_SPEED_10_MBIT;

                   /* 
                    * update GMAC register with speed read from PHY 
                    */
					PR_Reg_Read(ptr_ge_mac, dwTmp);
					shift = 1; /* (portId & PORTID_MASK) + 0;  */
					dwTmp &= ~(3 << shift);
					dwTmp |= (*modePtr << shift);
                    PR_Reg_Write(ptr_ge_mac, dwTmp);
				}  
			}

			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;
            break;

        case FE_PORT_PAUSE_FLOW_CTRL:
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s -  FE_PORT_PAUSE_FLOW_CTRL\n", __FUNCTION__);
#endif
            if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
                return retflag;

            if( (phy_auto == FALSE) && (en_hw_mode == FALSE) )
            {
                *modePtr = 0;
                PR_Reg_Read(ptr_ge_mac, dwTmp);
#ifdef INCAIP2_DEBUG_FEPORT
                printk("%s - read 0x%08X from Port %d GMAC_REG\n",
                           __FUNCTION__, dwTmp, portId);
#endif
				shift = (portId & PORTID_MASK) + PAUSE_CTRL_POS;
				dwTmp &= ~(0x01 << shift) ;
				dwTmp |= ((*modePtr & 0x01) << shift);
				PR_Reg_Write(ptr_ge_mac, dwTmp) ;
			}
			else
				if(en_hw_mode == TRUE)
				{ 
                    PR_Reg_Read(ptr_ge_mac, dwTmp);
#ifdef INCAIP2_DEBUG_FEPORT
                    printk("%s - read 0x%08X from Port %d GMAC_REG\n",
                           __FUNCTION__, dwTmp, portId);
#endif
					shift = PAUSE_CTRL_POS;  /*(portId & PORTID_MASK) + PAUSE_CTRL_POS;  */
                    if( ((dwTmp >> shift) & 0x03) == 0x03)
                       *modePtr = 1;
                    else
                        if( ((dwTmp >> shift) & 0x03) == 0x00)    
                            *modePtr = 0;
                        else
                        {
                            *modePtr = 0xFFFFFFFF;
                        }
				}
				else
				{
					if (ifx_phy_read(chipId, phy_addr, PHY_ANAR, &phyData4) != 0)
					{
						ifx_sem_give(chipId,SEM_MDIO);
						return PR_ERR_PHY_RD;
					}
				
					if (ifx_phy_read(chipId, phy_addr, PHY_ANLPA, &phyData5) != 0)
					{
						ifx_sem_give(chipId,SEM_MDIO);
						return PR_ERR_PHY_RD;
					}
            
					dwTmp = ((phyData4 >> 5) & 0xF) & ((phyData5 >> 5) & 0xF);
            
					if( ((dwTmp >> 3) & 0x1) == 0x1)
						*modePtr = 1;
					else
						if( ((dwTmp >> 2) & 0x1) == 0x1)		
							*modePtr = 0;
						else
							if( ((dwTmp >> 1) & 0x1) == 0x1)		
								*modePtr = 1;
							else
								*modePtr = 0;
            
					if(*modePtr == 1)
					{
						*modePtr = ( ((phyData4 >> 10) & 0x1) & ((phyData5 >> 10) & 0x1) );	
					}
					
					PR_Reg_Read(ptr_ge_mac, dwTmp) ;
					shift = (portId & PORTID_MASK) + PAUSE_CTRL_POS;
					dwTmp &= ~(0x01 << shift) ;
					dwTmp |= ((*modePtr & 0x01) << shift);
					PR_Reg_Write (ptr_ge_mac, dwTmp) ;
				}
			
			if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
    				return retflag;
            
			break;

		case FE_PORT_PRIORITY: 			
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_PRIORITY\n", __FUNCTION__);
#endif
            return( ifx_get_port_CoS(chipId,portId,modePtr) ) ;
		
		case FE_PORT_MONITOR_INGRESS: 	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_MONITOR_INGRESS\n", __FUNCTION__);
#endif
            return ( ifx_get_ingress_monitor_flag(chipId,portId,modePtr) );
			
		case FE_PORT_MONITOR_EGRESS: 	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_MONITOR_EGRESS\n", __FUNCTION__);
#endif
            return( ifx_get_egress_monitor_flag(chipId,portId,modePtr) );
			
		case FE_PORT_AUTHENTICATION:	
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - FE_PORT_AUTHENTICATION\n", __FUNCTION__);
#endif
            PR_Reg_Read(ptr_port_auth, dwTmp) ;
			shift = (portId & 0xF) << 1;
			*modePtr = (dwTmp >> shift) & 0x03;
			break;

		default:						
#ifdef INCAIP2_DEBUG_FEPORT
            printk("%s - Invalid Mode\n", __FUNCTION__);
#endif
            return PR_ERR_FE_PORT_SET;
	}

	return PR_ERR_NONE;
}





/*************************************** Internal APIs  **********************************/

/*****************************************************************************
  *
  *	FUNCTION:	ifx_set_feport_duplex_speed_add		 
  *
  *	DESCRIPTION: Set the phy speed and duplex mode	 
  *
  *	PARAMETERS:
  *		IN:
  *         chipId      Chip Id
  *			phy_addr	MDIO address of the PHY device
  *         speed 		Speed(100 or 10 mbps)
  *         duplex      Duplex mode(full or half) 		
  *			
  *		OUT:
  *			None
  *
  *	RETURNS:
  *		PR_ERR_NONE		     Success 
  *     PR_ERR_CHIP_ID       Invalid chip Id 
  *     PR_ERR_PHY_WR        If Phy write fails  
  *
  ***************************************************************************/

s32 ifx_set_feport_duplex_speed_add(u8 chipId, u8 phy_addr, u8 speed, u8 duplex)
{
	u16 wReg4=0;
#if 0
    u16 dwTmp=0;
#endif

#ifdef INCAIP2_DEBUG_FEPORT
    printk("%s(chipId=%d, phy_addr=%d, speed=%d, duplex=%d)\n", 
           __FUNCTION__, chipId, phy_addr, speed, duplex);	    
#endif
    
#ifdef PARAMETER_CHECK

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

#endif

    /* 
     * Read and update content of Autonegotiation Advertisement Register 
     * with combined speed and duplex information 
     */
    if (ifx_phy_read(chipId, phy_addr, PHY_ANAR, &wReg4) != 0)
		return PR_ERR_PHY_RD;
	
	if ( (speed == 1) && (duplex == 1) )
	{
		((PHY_REG4_T*)&wReg4)->full_100 = 1;
		((PHY_REG4_T*)&wReg4)->half_100 = 1;
		((PHY_REG4_T*)&wReg4)->full_10 = 1;
		((PHY_REG4_T*)&wReg4)->half_10 = 1;
	}
	else if ( (speed == 1) && (duplex == 0) )
	{
		((PHY_REG4_T*)&wReg4)->full_100 = 0;
		((PHY_REG4_T*)&wReg4)->half_100 = 1;
		((PHY_REG4_T*)&wReg4)->full_10 = 0;
		((PHY_REG4_T*)&wReg4)->half_10 = 1;
	}
	else if ( (speed == 0) && (duplex == 1) )
	{
		((PHY_REG4_T*)&wReg4)->full_100 = 0;
		((PHY_REG4_T*)&wReg4)->half_100 = 0;
		((PHY_REG4_T*)&wReg4)->full_10 = 1;
		((PHY_REG4_T*)&wReg4)->half_10 = 1;
	}
	else if ( (speed == 0) && (duplex == 0) )
	{
		((PHY_REG4_T*)&wReg4)->full_100 = 0;
		((PHY_REG4_T*)&wReg4)->half_100 = 0;
		((PHY_REG4_T*)&wReg4)->full_10 = 0;
		((PHY_REG4_T*)&wReg4)->half_10 = 1;
	}

    if (ifx_phy_write(chipId, phy_addr, PHY_ANAR, wReg4) != 0)
		return PR_ERR_PHY_WR;
#if 0
    /* Re-start Autonegotiation */
    if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != 0)
        return PR_ERR_PHY_RD;
#ifdef INCAIP2_DEBUG_FEPORT
    printk("%s - Auto Negotiation restarted ...\n", __FUNCTION__);	      
#endif
	dwTmp |= 0x2200; /* 8 bit Autonegotiation,9th bit Restart autonegotiation*/
	
    if (ifx_phy_write(chipId,  phy_addr, PHY_CTRL, dwTmp) != 0)
		return PR_ERR_PHY_WR;
#endif
	return PR_ERR_NONE;
}



/*****************************************************************************
  *
  *	FUNCTION:	ifx_phy_speed_fe		 
  *
  *	DESCRIPTION: Set the phy speed to 100 or 10 mbps	 
  *
  *	PARAMETERS:
  *		IN:
  *         chipId      Chip Id
  *			portId		port Id (0 - 49)
  *			phy_addr	MDIO address of the PHY device
  *         mode 		Speed(100 or 10 mbps)		
  *			
  *		OUT:
  *			None
  *
  *	RETURNS:
  *		PR_ERR_NONE		     Success 
  *     PR_ERR_CHIP_ID       Invalid chip Id 
  *     PR_ERR_PHY_WR        If Phy write fails
  *
  ***************************************************************************/

s32 ifx_phy_speed_fe( u8 chipId, u32 portId, u8 speed )
{
	u16   dwTmp=0;
    s32 retval;
    u8 phy_addr;

#ifdef PARAMETER_CHECK

	if(chipId >= Total_Chips)
		return PR_ERR_CHIP_ID ;

#endif
    if( (retval=ifx_get_portId_phyId(chipId, (u8) portId, &phy_addr)) != PR_ERR_NONE)
        return retval;
	
	if( ifx_phy_read( chipId, phy_addr, PHY_CTRL, &dwTmp ) != PR_ERR_NONE )
		return PR_ERR_PHY_RD;
	dwTmp &= ~(0x1 << PHY_FE_SPEED_POS);
	dwTmp |= ((speed & 0x1) << PHY_FE_SPEED_POS);  
	
    if (ifx_phy_write(chipId, phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
		return PR_ERR_PHY_WR;

	return PR_ERR_NONE ;	

}


