/************************************************************************
 *
 * Copyright (c) 2006
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
#include <linux/kernel.h>
#include <asm/types.h>
#include <asm/addrspace.h>
#include <asm/incaip2/ifx_types.h>
#include <asm/incaip2/incaip2.h>
#include "struct.h"
#include "const.h"
#include "ge_port.h"
#include "ge_port_prv.h"
#include "err.h"
#include "phy_prv.h"
#include "phy.h"
#include "var.h"
#include "api_init_prv.h"

#undef INCAIP2_DEBUG_GEPORT

extern s32 ifx_get_MDIO_mode(u8 chipId, u32 port_group, BOOL * mdioMode, BOOL * autoNegMode);
extern s32 ifx_get_switch_mode(u8 chipId,u32 feature, u32* modePtr);
extern s32 ifx_set_port_autoneg(u8 chipId,u32 portId, BOOL Enable);
extern s32 ifx_get_port_autoneg(u8 chipId,u32 portId, BOOL * EnablePtr);
extern s32 ifx_set_port_CoS(u8 chipId,u32 portId, u32 CoS);
extern s32 ifx_get_port_CoS(u8 chipId,u32 portId, u32* CoSPtr);
extern s32 ifx_set_ingress_monitor_flag(u8 chipId,u32 portId, BOOL ingressMonitorFlag);
extern s32 ifx_get_ingress_monitor_flag(u8 chipId,u32 portId, BOOL *ingressMonitorFlagPtr);
extern s32 ifx_set_egress_monitor_flag(u8 chipId,u32 portId, BOOL egressMonitorFlag);
extern s32 ifx_get_egress_monitor_flag(u8 chipId,u32 portId, BOOL *egressMonitorFlagPtr);



/**
 * Get Port Gigabit Ethernet Mode.
 * Gets the mode for a desired feature of an individual GE port.
 *
 *
 * \param   chipId          Chip Id
 * \param   portId          port Id (0 - 1)
 * \param   feature         Selector for the feature to be read for the port
 *                          The allowed features are:
 *                         0.GE_PORT_AUTONEGOTIATION
 *                         2.GE_PORT_DUPLEX_MODE
 *                         3.GE_PORT_SPEED
 *                         4.GE_RX_PORT_PAUSE_FLOW_CTRL
 *                         5.GE_TX_PORT_PAUSE_FLOW_CTRL
 *                         6.GE_PORT_PRIORITY
 *                         7.GE_PORT_MONITOR_INGRESS
 *                         8.GE_PORT_MONITOR_EGRESS
 *                         9.GE_PORT_AUTHENTICATION
 * \param   mode            Value of the feature
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_PORTID          - portId is not in the range of 0 - 1
 * \return  PR_ERR_GE_PORT_SET     - feature is not in the range of 0 - 9
 * \return  PR_ERR_NULL_PTR        - Base address for the chip ID is not set
 *                                   or modePtr is NULL.
 * \return  PR_ERR_PHY_WR          - Phy write fails
 * \return  PR_ERR_PHY_RD          - Phy read fails
 * \return  PR_ERR_LINK_DOWN       - Link is down on desired port
 * \return  PR_ERR_CHIP_ID         - Invalid Chip Id
 * \return  PR_ERR_PORTID          - Invalid Port Id
 * \ingroup Internal
 */
s32 ifx_set_port_GE_mode( u8 chipId, u32 portId, u32 feature, u32 mode )
{
   u32    dwTmp =0;
   u8     phy_addr=0;
   u32    shift=0;
   BOOL   phy_auto=0;
   u32*   ptr_gmac_reg ;
   u32*   ptr_port_auth ;
   s32    status = 0;
   u32    sw_feature = 0;
   u32    sw_mode = 0;
   u32    ifx_portId = portId;
   s32    retflag = PR_ERR_NONE;
#ifdef BLOCK_WHEN_LINK_DOWN
   u16    phyData1=0;
#endif

#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s(chipId=%d, portId=%d, feature=%d, mode=0x%08X)\n",
      __FUNCTION__, chipId, portId, feature, mode);
#endif
#ifdef PARAMETER_CHECK

   if(portId >= TOTAL_PHY_PORTS)
      return PR_ERR_PORTID;

   if(chipId >= Total_Chips)
      return PR_ERR_CHIP_ID ;

   if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
      return PR_ERR_NULL_PTR;

#endif

   switch(portId)
   {
   case INCAIP2_SWITCH_PORT_SELECT_LAN:
      ptr_gmac_reg = (u32*) SWITCH_P0_GMAC_REG;
      sw_feature = PORT0_GE_MODE;
      break;
   case INCAIP2_SWITCH_PORT_SELECT_PC:
      ptr_gmac_reg = (u32*) SWITCH_P1_GMAC_REG;
      sw_feature = PORT1_GE_MODE;
      break;

   default:
      return PR_ERR_PORTID;
   }
   ptr_port_auth = (u32*) PORT_AUTH;

   status = ifx_get_switch_mode(chipId,sw_feature,&sw_mode);

#ifdef PARAMETER_CHECK
   if(status != PR_ERR_NONE)
   {
      return status;
   }

   if(sw_mode == PORT_MII_MODE)
   {
      printk("%s - error: PORT_MII_MODE\n", __FUNCTION__ );
      return PR_ERR_GE_PORT_SET;
   }
#endif

   if((retflag = ifx_get_portId_phyId( chipId, ifx_portId, &phy_addr )) != PR_ERR_NONE)
   {
      return retflag;
   }

#ifdef BLOCK_WHEN_LINK_DOWN
   /* If the link is down don't allow the following configurations */
   if((feature == GE_PORT_AUTONEGOTIATION) ||
      (feature == GE_PORT_DUPLEX_MODE) ||
      (feature == GE_PORT_SPEED) ||
      (feature == GE_RX_PORT_PAUSE_FLOW_CTRL)  ||
      (feature == GE_TX_PORT_PAUSE_FLOW_CTRL))
   {
      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if (ifx_phy_read(chipId, phy_addr, PHY_STATUS, &phyData1) != 0)
      {
         ifx_sem_give(chipId,SEM_MDIO);
         return PR_ERR_PHY_RD;
      }

      if( ((phyData1 >> 2) & 0x01) == 0 )
      {
         printk("%s -> link is down\n", __FUNCTION__);
         ifx_sem_give(chipId,SEM_MDIO);
         return PR_ERR_LINK_DOWN;
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

   }
#endif

   switch (feature)
   {
   case GE_PORT_AUTONEGOTIATION:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_AUTONEGOTIATION\n", __FUNCTION__ );
#endif
      mode &= 0x01;
      return(ifx_set_port_autoneg(chipId,ifx_portId,(BOOL)mode));

   case GE_PORT_DUPLEX_MODE:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_DUPLEX_MODE\n", __FUNCTION__ );
#endif
      /* advertising autonegotiation info */
      mode &= 0x1;
      if((retflag = ifx_get_port_autoneg(chipId,ifx_portId,&phy_auto)) != PR_ERR_NONE)
         return retflag;

      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if(phy_auto == FALSE)
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> phy_auto == FALSE\n", __FUNCTION__);
#endif
         ifx_phy_set_duplex_mode( chipId, portId, (BOOL)mode );
      }
      else
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> phy_auto == TRUE\n", __FUNCTION__);
#endif
         cache_duplex[chipId][portId] = (0x01 & mode);
         if ( ifx_set_geport_duplex_speed_add( chipId,
            phy_addr,
            cache_speed[chipId][portId],
            cache_duplex[chipId][portId]) )
         {
            ifx_sem_give(chipId,SEM_MDIO);
            printk("%s - ifx_set_geport_duplex_speed_add() failed\n", __FUNCTION__);
            return PR_ERR_GE_PORT_SET;
         }
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;
      break;


   case GE_PORT_SPEED:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_SPEED\n", __FUNCTION__ );
#endif
      if((retflag = ifx_get_port_autoneg( chipId, ifx_portId,
         &phy_auto)) != PR_ERR_NONE)
      {
         printk("%s - ifx_get_port_autoneg() failed\n", __FUNCTION__);
         return retflag;
      }

      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
      {
         printk("%s - ifx_sem_take() failed\n", __FUNCTION__);
         return retflag;
      }

      if(phy_auto == FALSE)
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> phy_auto == FALSE\n", __FUNCTION__);
#endif
         ifx_phy_speed_ge(chipId,ifx_portId, mode);
      }
      else
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> phy_auto == TRUE\n", __FUNCTION__);
#endif
         cache_speed[chipId][portId] = (GE_PORT_SPEED_MASK & mode);
         if ( ifx_set_geport_duplex_speed_add( chipId,
            phy_addr,
            cache_speed[chipId][portId],
            cache_duplex[chipId][portId]) )
         {
            ifx_sem_give(chipId,SEM_MDIO);
            printk("%s - ifx_set_geport_duplex_speed_add() failed\n", __FUNCTION__);
            return PR_ERR_GE_PORT_SET;
         }
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;
      break;

   case GE_RX_PORT_PAUSE_FLOW_CTRL:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_RX_PORT_PAUSE_FLOW_CTRL\n", __FUNCTION__ );
#endif
      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if((retflag = ifx_phy_pause_rx(chipId,portId,(BOOL)mode)) != PR_ERR_NONE)
      {
         ifx_sem_give(chipId,SEM_MDIO);
         return retflag;
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;
      break;

   case GE_TX_PORT_PAUSE_FLOW_CTRL:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_TX_PORT_PAUSE_FLOW_CTRL\n", __FUNCTION__ );
#endif

      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if((retflag = ifx_phy_pause_tx(chipId,portId,(BOOL)mode)) != PR_ERR_NONE)
      {
         ifx_sem_give(chipId,SEM_MDIO);
         return retflag;
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;
      break;

   case GE_PORT_PRIORITY:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_PRIORITY\n", __FUNCTION__ );
#endif
      return( ifx_set_port_CoS(chipId,ifx_portId, mode )) ;

   case GE_PORT_MONITOR_INGRESS:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_MONITOR_INGRESS\n", __FUNCTION__ );
#endif
      return(ifx_set_ingress_monitor_flag( chipId,
         ifx_portId,
         (BOOL)mode) );


   case GE_PORT_MONITOR_EGRESS:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_MONITOR_EGRESS\n", __FUNCTION__ );
#endif
      return( ifx_set_egress_monitor_flag( chipId,
         ifx_portId,
         (BOOL)mode) );

   case GE_PORT_AUTHENTICATION:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_AUTHENTICATION\n", __FUNCTION__ );
#endif
      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      PR_Reg_Read( ptr_port_auth, dwTmp) ;
      shift = (portId & 0xF) << 1;
      dwTmp &= ~(GE_PORT_AUTHEN_MASK << shift) ;
      dwTmp |= ((mode & GE_PORT_AUTHEN_MASK) << shift);
      PR_Reg_Write(ptr_port_auth, dwTmp) ;

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;
      break;

   default:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> Invalid Mode passed\n", __FUNCTION__ );
#endif
      return PR_ERR_GE_PORT_SET;
   }

   return PR_ERR_NONE;
}


/**
 * Get Port Gigabit Ethernet Mode.
 * Gets the mode for a desired feature of an individual GE port.
 *
 *
 * \param   chipId          Chip Id
 * \param   portId          port Id (0 - 1)
 * \param   feature         Selector for the feature to be read for the port
 *                          The allowed features are:
 *                         0.GE_PORT_AUTONEGOTIATION
 *                         1.GE_PORT_LINK_OK
 *                         2.GE_PORT_DUPLEX_MODE
 *                         3.GE_PORT_SPEED
 *                         4.GE_RX_PORT_PAUSE_FLOW_CTRL
 *                         5.GE_TX_PORT_PAUSE_FLOW_CTRL
 *                         6.GE_PORT_PRIORITY
 *                         7.GE_PORT_MONITOR_INGRESS
 *                         8.GE_PORT_MONITOR_EGRESS
 *                         9.GE_PORT_AUTHENTICATION
 * \param   modePtr         Pointer to value of the feature
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_PORTID          - portId is not in the range of 0 - 1
 * \return  PR_ERR_GE_PORT_SET     - feature is not in the range of 0 - 9
 * \return  PR_ERR_NULL_PTR        - Base address for the chip ID is not set
 *                                   or modePtr is NULL.
 * \return  PR_ERR_PHY_WR          - Phy write fails
 * \return  PR_ERR_PHY_RD          - Phy read fails
 * \return  PR_ERR_CHIP_ID         - Invalid Chip Id
 * \return  PR_ERR_PORTID          - Invalid Port Id
 * \ingroup Internal
 */
s32 ifx_get_port_GE_mode( u8 chipId, u32 portId, u32 feature, u32 * modePtr )
{
   u32         dwTmp=0;
   u32         shift=0;
   BOOL        en_hw_mode=0;
   BOOL        phy_auto =0;
   gmac_reg_t* ptr_gmac_data; /* GMAC_REG register pointer */
   u32*        ptr_gmac_reg;  /* GMAC_REG data pointer */
   u32*        ptr_port_auth;
   s32         status = 0;
   u32         sw_feature = 0;
   u32         sw_mode = 0;
   u32         ifx_portId = portId;
   s32         retflag = PR_ERR_NONE;
   u16         phyData4=0, phyData5=0;
   u16         phyData9=0, phyData10=0;
   u8          phy_addr=255;
   u16           dwTmp_ge=0, dwTmp_fe=0;

#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s(chipId=%d, portId=%d, feature=%d)\n",
      __FUNCTION__, chipId, portId, feature );
#endif
#ifdef PARAMETER_CHECK

   if(portId >= IFX_TOTAL_PHY_PORTS)
      return PR_ERR_PORTID;

   if(modePtr == NULL)
      return PR_ERR_NULL_PTR;

   if(chipId >= Total_Chips)
      return PR_ERR_CHIP_ID ;

   if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
      return PR_ERR_NULL_PTR;

#endif

   /*
   * set pointer to GMAC_REG assigned to passed port
   */
   switch(portId)
   {
   case INCAIP2_SWITCH_PORT_SELECT_LAN:
      ptr_gmac_reg = (u32*) SWITCH_P0_GMAC_REG;
      sw_feature = PORT0_GE_MODE;
      break;
   case INCAIP2_SWITCH_PORT_SELECT_PC:
      ptr_gmac_reg = (u32*) SWITCH_P1_GMAC_REG;
      sw_feature = PORT1_GE_MODE;
      break;

   default:
      return PR_ERR_PORTID;
   }
   ptr_port_auth = (u32*) PORT_AUTH;
   ptr_gmac_data = (gmac_reg_t *)&dwTmp;


   if( (retflag = ifx_get_MDIO_mode( chipId, portId,
      &en_hw_mode, &phy_auto)) != PR_ERR_NONE )
   {
      printk("%s - failed to read ANEG/MDIO modes\n", __FUNCTION__ );
      return retflag;
   }
#if 1
   /* autonegotiation state in MDIO_SEL register is not correctly set by HW
   -> ANEG state is read from PHY CTRL register instead */
   if((retflag = ifx_get_port_autoneg( chipId,
      ifx_portId,
      &phy_auto)) != PR_ERR_NONE)
   {
      printk("%s - failed to read ANEG state\n", __FUNCTION__ );
      return retflag;
   }
#endif

#ifdef INCAIP2_DEBUG_GEPORT
   if(en_hw_mode == TRUE)
      printk("%s - MDIO HW State Machine is enabled\n", __FUNCTION__ );
   else
      printk("%s - MDIO HW State Machine is disabled\n", __FUNCTION__ );
   if(phy_auto == TRUE)
      printk("%s - Auto Negotiation is enabled\n", __FUNCTION__ );
   else
      printk("%s - Auto Negotiation is disabled\n", __FUNCTION__ );

#endif

   status = ifx_get_switch_mode(chipId, sw_feature, &sw_mode );

#ifdef PARAMETER_CHECK
   if(status != PR_ERR_NONE)
      return status ;
   if(sw_mode == PORT_MII_MODE)
   {
      printk("%s - error: PORT_MII_MODE enabled\n", __FUNCTION__ );
      return PR_ERR_GE_PORT_SET;
   }
#endif

   switch (feature)
   {
   case GE_PORT_AUTONEGOTIATION:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_AUTONEGOTIATION\n", __FUNCTION__ );
#endif
      *((BOOL*)modePtr) = phy_auto;
      return PR_ERR_NONE;

   case GE_PORT_LINK_OK:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_LINK_STATE\n", __FUNCTION__ );
#endif
      /*
      if((retflag = ifx_get_port_autoneg (chipId,
      ifx_portId,
      &phy_auto)) != PR_ERR_NONE)
      return retflag;
      */
      if( en_hw_mode == TRUE && phy_auto == TRUE)
      {
         PR_Reg_Read((u32*)ptr_gmac_reg, dwTmp);
         *modePtr = ptr_gmac_data->lg;
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - read 0x%08X from SWITCH_GMAC_REG for port %d\n",
            __FUNCTION__, dwTmp, ifx_portId);
#endif
      }
      else
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - no mdio state machine\n", __FUNCTION__);
         printk("   -> read link status from PHY STATUS register\n");
         printk("      and update GMAC_REG link status\n");
#endif

         /*
         * read link status from phy status register and
         * update link status in GMAC_REG
         */
         if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
            return retflag;

         ifx_get_portId_phyId( chipId, portId, &phy_addr );

         if (ifx_phy_read(chipId, phy_addr, PHY_STATUS, &phyData4) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         *modePtr = (phyData4 >> 2) & 0x01;
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - read link state %d from from PHY\n", __FUNCTION__, *modePtr);
#endif
         PR_Reg_Read((u32*)ptr_gmac_reg, dwTmp);
         ptr_gmac_data->lg = (*modePtr & 0x00000001);

         PR_Reg_Write(ptr_gmac_reg, dwTmp);

         if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
            return retflag;
      }
      break;

   case GE_PORT_DUPLEX_MODE:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_DUPLEX_MODE\n", __FUNCTION__ );
#endif
      /*
      if((retflag = ifx_get_port_autoneg (chipId,
      ifx_portId,
      &phy_auto)) != PR_ERR_NONE)
      return retflag;
      */
      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      /*
      * no automatic update -> read duplex status from PHY CTRL register
      * and update GMAC_REG duplex status
      */
      if( (phy_auto == FALSE) && (en_hw_mode == FALSE) )
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - no aneg, no mdio state machine\n", __FUNCTION__);
         printk("   -> read duplex status from PHY CTRL register\n");
         printk("      and update GMAC_REG duplex status\n");
#endif
         /* Read from Reg 0 */
         ifx_get_portId_phyId(chipId,portId, &phy_addr);
         if (ifx_phy_read( chipId, phy_addr,
            PHY_CTRL, &phyData4) != PR_ERR_NONE)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         *modePtr = (phyData4 >> PHY_DUPLEX_MODE_POS) & 0x01;

#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - read duplex state %d from from PHY\n", __FUNCTION__, *modePtr);
#endif

         PR_Reg_Read(ptr_gmac_reg, dwTmp);
         ptr_gmac_data->dg = *modePtr;
         /*
         dwTmp &= ~(1 << GE_PORT_DUPLEX_POS);
         dwTmp |= (*modePtr << GE_PORT_DUPLEX_POS);
         */
         PR_Reg_Write(ptr_gmac_reg, dwTmp);
      }
      else if(en_hw_mode == TRUE )
      {
         /* automatic update -> read duplex state from GMAC_REG */
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - en_hw_mode == TRUE -> read duplex state from GMAC_REG\n",
            __FUNCTION__ );
#endif
         PR_Reg_Read(ptr_gmac_reg, dwTmp);
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - read 0x%08X from SWITCH_GMAC_REG for port %d\n",
            __FUNCTION__, dwTmp, portId);
         printk("%s - Duplex state %d from SWITCH_GMAC_REG for port %d\n",
            __FUNCTION__, ptr_gmac_data->dg, portId);
#endif
         *modePtr = (u32)ptr_gmac_data->dg;
         /*               *modePtr = (dwTmp >> GE_PORT_DUPLEX_POS) & 0x01;  */
      }
      else /* en_hw_mode != TRUE */
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - en_hw_mode == FALSE -> determine duplex state from PHY states\n",
            __FUNCTION__ );
         printk("      and update GMAC_REG duplex status\n");
#endif
         ifx_get_portId_phyId( chipId, portId, &phy_addr );

         if (ifx_phy_read(chipId, phy_addr, 0x9, &phyData9) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, 0xa, &phyData10) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, PHY_ANAR, &phyData4) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, PHY_ANLPA, &phyData5) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         /*
         * Generate duplex mode from PHY_ANAR, PHY_ANLPA,
         * 1000Base-T Control and 1000Base-T Status registers content
         * and update GMAC_REG duplex information
         */
         dwTmp_ge = ((phyData9 >> 8) & 0x3) & ((phyData10 >> 10) & 0x3);
         dwTmp_fe = ((phyData4 >> 5) & 0x7F) & ((phyData5 >> 5) & 0x7F);

         if( ((dwTmp_ge >> 1) & 0x1) == 0x1)
            *modePtr = IFX_FULL_DUPLEX;
         else if( (dwTmp_ge & 0x1) == 0x1)
            *modePtr = IFX_HALF_DUPLEX;
         else if( ((dwTmp_fe >> 3) & 0x1) == 0x1)
            *modePtr = IFX_FULL_DUPLEX;
         else if( ((dwTmp_fe >> 2) & 0x1) == 0x1)
            *modePtr = IFX_HALF_DUPLEX;
         else if( ((dwTmp_fe >> 1) & 0x1) == 0x1)
            *modePtr = IFX_FULL_DUPLEX;
         else
            *modePtr = IFX_HALF_DUPLEX;
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - calculated duplex mode=%d \n", __FUNCTION__, *modePtr );
#endif
         PR_Reg_Read(ptr_gmac_reg, dwTmp);
         ptr_gmac_data->dg = *modePtr;
         /*
         dwTmp &= ~(1 << GE_PORT_DUPLEX_POS);
         dwTmp |= (*modePtr << GE_PORT_DUPLEX_POS);
         */
         PR_Reg_Write(ptr_gmac_reg, dwTmp);
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;
      break;

   case GE_PORT_SPEED:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_PORT_SPEED\n", __FUNCTION__ );
#endif
      /*
      if((retflag = ifx_get_port_autoneg(chipId,ifx_portId,&phy_auto)) != PR_ERR_NONE)
      return retflag;
      */

      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if((phy_auto == FALSE) && (en_hw_mode == FALSE))
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - (phy_auto == FALSE) && (en_hw_mode == FALSE)\n", __FUNCTION__ );
#endif
         /* Read from Reg 0 */
         ifx_get_portId_phyId(chipId,portId, &phy_addr);
         if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &phyData4) != PR_ERR_NONE)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         *modePtr = (phyData4 >> PHY_GE_SPEED1_POS) & 0x1;
         phyData4 =  ((phyData4 >> PHY_GE_SPEED2_POS) & 0x1) << 1 ;
         *modePtr |= phyData4 ;

         PR_Reg_Read(ptr_gmac_reg, dwTmp);
         dwTmp &= ~(GE_PORT_SPEED_MASK << 1);
         dwTmp |= (*modePtr << 1);
         PR_Reg_Write(ptr_gmac_reg, dwTmp);


      }
      else if(en_hw_mode == TRUE )
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - MDIO state machine enabled\n", __FUNCTION__ );
#endif
         PR_Reg_Read(ptr_gmac_reg, dwTmp);

#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - read 0x%08X from SWITCH_GMAC_REG for port %d\n",
            __FUNCTION__, dwTmp, portId);
         printk("%s - speed %d from SWITCH_GMAC_REG for port %d\n",
            __FUNCTION__, ptr_gmac_data->sg, portId);
#endif
         *modePtr = (u32) ptr_gmac_data->sg;
         /*                *modePtr = (u32) ((dwTmp >> 1) & GE_PORT_SPEED_MASK);  */
      }
      else
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s - MDIO state machine not enabled\n", __FUNCTION__ );
#endif
         ifx_get_portId_phyId(chipId,portId,/* &mdio_id, */&phy_addr);

         if (ifx_phy_read(chipId, phy_addr, 0x9, &phyData9) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, 0xa, &phyData10) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId,phy_addr, PHY_ANAR, &phyData4) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId,phy_addr, PHY_ANLPA, &phyData5) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> calculate speed information from PHY registers\n", __FUNCTION__ );
#endif
         dwTmp_ge = ((phyData9 >> 8) & 0x3) & ((phyData10 >> 10) & 0x3);
         dwTmp_fe = ((phyData4 >> 5) & 0x7F) & ((phyData5 >> 5) & 0x7F);

         if( ((dwTmp_ge >> 1) & 0x1) == 0x1)
            *modePtr = 2;
         else if( (dwTmp_ge & 0x1) == 0x1)
            *modePtr = 2;
         else if( ((dwTmp_fe >> 3) & 0x1) == 0x1)
            *modePtr = 1;
         else if( ((dwTmp_fe >> 2) & 0x1) == 0x1)
            *modePtr = 1;
         else if( ((dwTmp_fe >> 1) & 0x1) == 0x1)
            *modePtr = 0;
         else
            *modePtr = 0;

         printk("%s - calculated speed state %d from from PHY\n", __FUNCTION__, *modePtr);

         PR_Reg_Read(ptr_gmac_reg, dwTmp);
         dwTmp &= ~(GE_PORT_SPEED_MASK << 1);
         dwTmp |= (*modePtr << 1);
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> update GMAC REG with speed information\n", __FUNCTION__ );
#endif
         PR_Reg_Write(ptr_gmac_reg, dwTmp);

      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      break;

   case GE_RX_PORT_PAUSE_FLOW_CTRL:
#ifdef INCAIP2_DEBUG_GEPORT
      printk("%s -> GE_RX_PORT_PAUSE_FLOW_CTRL\n", __FUNCTION__ );
#endif
      /*
      if((retflag = ifx_get_port_autoneg(chipId,ifx_portId,&phy_auto)) != PR_ERR_NONE)
      return retflag;
      */
      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if( (phy_auto == FALSE) && (en_hw_mode == FALSE) )
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> (phy_auto == FALSE) && (en_hw_mode == FALSE)\n", __FUNCTION__ );
#endif
         *modePtr = 0;

         PR_Reg_Read(ptr_gmac_reg, dwTmp) ;
         dwTmp &= ~(0x01 << GE_RX_PORT_PAUSE_POS) ;
         dwTmp |= ((*modePtr & 0x01) << GE_RX_PORT_PAUSE_POS);
         PR_Reg_Write(ptr_gmac_reg, dwTmp) ;
      }
      else if(en_hw_mode == TRUE)
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> en_hw_mode == TRUE\n", __FUNCTION__ );
#endif
         PR_Reg_Read(ptr_gmac_reg, dwTmp) ;
         *modePtr = (dwTmp >> GE_RX_PORT_PAUSE_POS) & 0x01;
      }
      else
      {
#ifdef INCAIP2_DEBUG_GEPORT
         printk("%s -> en_hw_mode != TRUE\n", __FUNCTION__ );
#endif
         ifx_get_portId_phyId(chipId,portId, /* &mdio_id, */ &phy_addr);


         if (ifx_phy_read(chipId, phy_addr, 0x9, &phyData9) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, 0xa, &phyData10) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, PHY_ANAR, &phyData4) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, PHY_ANLPA, &phyData5) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }


         dwTmp_ge = ((phyData9 >> 8) & 0x3) & ((phyData10 >> 10) & 0x3);
         dwTmp_fe = ((phyData4 >> 5) & 0x7F) & ((phyData5 >> 5) & 0x7F);

         if( ((dwTmp_ge >> 1) & 0x1) == 0x1)
            *modePtr = 1;
         else if( (dwTmp_ge & 0x1) == 0x1)
            *modePtr = 0;
         else if( ((dwTmp_fe >> 3) & 0x1) == 0x1)
            *modePtr = 1;
         else if( ((dwTmp_fe >> 2) & 0x1) == 0x1)
            *modePtr = 0;
         else if( ((dwTmp_fe >> 1) & 0x1) == 0x1)
            *modePtr = 1;
         else
            *modePtr = 0;

         if(*modePtr == 1)
         {
            *modePtr =( ((phyData4 >> 10) & 0x1) & ((phyData5 >> 10) & 0x1) ) | ( ((phyData4 >> 10) & 0x1) & ((phyData4 >> 11) & 0x1) & ((phyData5 >> 11) & 0x1) );
         }

         PR_Reg_Read(ptr_gmac_reg, dwTmp) ;
         dwTmp &= ~(0x01 << GE_RX_PORT_PAUSE_POS) ;
         dwTmp |= ((*modePtr & 0x01) << GE_RX_PORT_PAUSE_POS);
         PR_Reg_Write(ptr_gmac_reg, dwTmp) ;

      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      break;

   case GE_TX_PORT_PAUSE_FLOW_CTRL:
      /*
      if((retflag = ifx_get_port_autoneg(chipId,ifx_portId,&phy_auto)) != PR_ERR_NONE)
      return retflag;
      */
      if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      if((phy_auto == FALSE) && (en_hw_mode == FALSE) )
      {
         *modePtr = 0;
         PR_Reg_Read(ptr_gmac_reg, dwTmp) ;
         dwTmp &= ~(0x01 << GE_TX_PORT_PAUSE_POS) ;
         dwTmp |= ((*modePtr & 0x01) << GE_TX_PORT_PAUSE_POS);
         PR_Reg_Write(ptr_gmac_reg, dwTmp) ;
      }
      else if(en_hw_mode == TRUE)
      {
         PR_Reg_Read(ptr_gmac_reg, dwTmp) ;
         *modePtr = (dwTmp >> GE_TX_PORT_PAUSE_POS) & 0x01;
      }
      else
      {
         ifx_get_portId_phyId(chipId,portId, /* &mdio_id, */ &phy_addr);

         if (ifx_phy_read(chipId,phy_addr, 0x9, &phyData9) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read( chipId,phy_addr,
            0xa, &phyData10) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr,
            PHY_ANAR, &phyData4) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         if (ifx_phy_read(chipId, phy_addr, PHY_ANLPA, &phyData5) != 0)
         {
            ifx_sem_give(chipId,SEM_MDIO);
            return PR_ERR_PHY_RD;
         }

         dwTmp_ge = ((phyData9 >> 8) & 0x3) & ((phyData10 >> 10) & 0x3);
         dwTmp_fe = ((phyData4 >> 5) & 0x7F) & ((phyData5 >> 5) & 0x7F);

         if( ((dwTmp_ge >> 1) & 0x1) == 0x1)
            *modePtr = 1;
         else if( (dwTmp_ge & 0x1) == 0x1)
            *modePtr = 0;
         else if( ((dwTmp_fe >> 3) & 0x1) == 0x1)
            *modePtr = 1;
         else if( ((dwTmp_fe >> 2) & 0x1) == 0x1)
            *modePtr = 0;
         else if( ((dwTmp_fe >> 1) & 0x1) == 0x1)
            *modePtr = 1;
         else
            *modePtr = 0;

         if(*modePtr == 1)
         {
            *modePtr =( ((phyData4 >> 10) & 0x1) & \
               ((phyData5 >> 10) & 0x1) ) | \
               ( ((phyData5 >> 10) & 0x1) & \
               ((phyData5 >> 11) & 0x1) & \
               ((phyData4 >> 11) & 0x1) );
         }

         PR_Reg_Read(ptr_gmac_reg, dwTmp) ;
         dwTmp &= ~(0x01 << GE_TX_PORT_PAUSE_POS) ;
         dwTmp |= ((*modePtr & 0x01) << GE_TX_PORT_PAUSE_POS);
         PR_Reg_Write (ptr_gmac_reg, dwTmp) ;
      }

      if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
         return retflag;

      break;

   case GE_PORT_PRIORITY:
      return(ifx_get_port_CoS(chipId,ifx_portId,modePtr )) ;

   case GE_PORT_MONITOR_INGRESS:
      return(ifx_get_ingress_monitor_flag(chipId,ifx_portId,(BOOL*)modePtr) );

   case GE_PORT_MONITOR_EGRESS:
      return(ifx_get_egress_monitor_flag(chipId,ifx_portId,(BOOL*)modePtr) );

   case GE_PORT_AUTHENTICATION:
      PR_Reg_Read(ptr_port_auth, dwTmp) ;
      shift = (portId & 0xF) << 1;
      *modePtr = (dwTmp >> shift) & GE_PORT_AUTHEN_MASK;
      break;

   default:
      return PR_ERR_GE_PORT_SET;
   }

   return PR_ERR_NONE;
}


/************************************************* Internal APIs  *************************************************/


/*****************************************************************************
 *
 *   FUNCTION:   ifx_set_geport_duplex_speed_add
 *
 *   DESCRIPTION: Set the phy speed and duplex mode
 *
 *   PARAMETERS:
 *      IN:
 *         chipId      Chip Id
 *         phy_addr   MDIO address of the PHY device
 *         speed       Speed(1000 or 100 or 10 mbps)
 *         duplex      Duplex mode(full or half)
 *
 *      OUT:
 *         None
 *
 *   RETURNS:
 *      PR_ERR_NONE           Success
 *     PR_ERR_CHIP_ID       Invalid chip Id
 *     PR_ERR_PHY_WR        If Phy write fails
 *     PR_ERR_PHY_RD       If Phy read fails
 *
 ***************************************************************************/
s32 ifx_set_geport_duplex_speed_add( u8 chipId, u8 phy_addr, u8 speed, u8 duplex)
{
   u16 wReg4=0, wReg9=0;
#if 0
   u16 dwTmp=0;
#endif

#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s -  phy_addr %u, speed = %u, duplex = %u\n",
      __FUNCTION__, phy_addr, speed, duplex);
#endif

#ifdef PARAMETER_CHECK
   if(chipId >= Total_Chips)
      return PR_ERR_CHIP_ID ;
#endif

   if (ifx_phy_read(chipId, phy_addr, PHY_ANAR, &wReg4) != PR_ERR_NONE)
      return PR_ERR_PHY_RD;
#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s -  read 0x%04X from PHY_ANAR Register\n", __FUNCTION__, wReg4);
#endif

   if (ifx_phy_read(chipId, phy_addr, 0x9, &wReg9) != PR_ERR_NONE)
      return PR_ERR_PHY_RD;
#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s -  read 0x%04X from GETH CTRL Register\n", __FUNCTION__, wReg9);
#endif

   if ( (speed == 2) && (duplex == 1) )
   {
      ((PHY_REG9_T*)&wReg9)->full_1000 = 1;
      ((PHY_REG9_T*)&wReg9)->half_1000 = 1;
      ((PHY_REG4_T*)&wReg4)->full_100 = 1;
      ((PHY_REG4_T*)&wReg4)->half_100 = 1;
      ((PHY_REG4_T*)&wReg4)->full_10 = 1;
      ((PHY_REG4_T*)&wReg4)->half_10 = 1;
   }
   else if ( (speed == 2) && (duplex == 0) )
   {
      ((PHY_REG9_T*)&wReg9)->full_1000 = 0;
      ((PHY_REG9_T*)&wReg9)->half_1000 = 1;
      ((PHY_REG4_T*)&wReg4)->full_100 = 0;
      ((PHY_REG4_T*)&wReg4)->half_100 = 1;
      ((PHY_REG4_T*)&wReg4)->full_10 = 0;
      ((PHY_REG4_T*)&wReg4)->half_10 = 1;
   }
   else if ( (speed == 1) && (duplex == 1) )
   {
      ((PHY_REG9_T*)&wReg9)->full_1000 = 0;
      ((PHY_REG9_T*)&wReg9)->half_1000 = 0;
      ((PHY_REG4_T*)&wReg4)->full_100 = 1;
      ((PHY_REG4_T*)&wReg4)->half_100 = 1;
      ((PHY_REG4_T*)&wReg4)->full_10 = 1;
      ((PHY_REG4_T*)&wReg4)->half_10 = 1;
   }
   else if ( (speed == 1) && (duplex == 0) )
   {
      ((PHY_REG9_T*)&wReg9)->full_1000 = 0;
      ((PHY_REG9_T*)&wReg9)->half_1000 = 0;
      ((PHY_REG4_T*)&wReg4)->full_100 = 0;
      ((PHY_REG4_T*)&wReg4)->half_100 = 1;
      ((PHY_REG4_T*)&wReg4)->full_10 = 0;
      ((PHY_REG4_T*)&wReg4)->half_10 = 1;
   }
   else if ( (speed == 0) && (duplex == 1) )
   {
      ((PHY_REG9_T*)&wReg9)->full_1000 = 0;
      ((PHY_REG9_T*)&wReg9)->half_1000 = 0;
      ((PHY_REG4_T*)&wReg4)->full_100 = 0;
      ((PHY_REG4_T*)&wReg4)->half_100 = 0;
      ((PHY_REG4_T*)&wReg4)->full_10 = 1;
      ((PHY_REG4_T*)&wReg4)->half_10 = 1;
   }
   else if ( (speed == 0) && (duplex == 0) )
   {
      ((PHY_REG9_T*)&wReg9)->full_1000 = 0;
      ((PHY_REG9_T*)&wReg9)->half_1000 = 0;
      ((PHY_REG4_T*)&wReg4)->full_100 = 0;
      ((PHY_REG4_T*)&wReg4)->half_100 = 0;
      ((PHY_REG4_T*)&wReg4)->full_10 = 0;
      ((PHY_REG4_T*)&wReg4)->half_10 = 1;
   }

   if (ifx_phy_write(chipId,  phy_addr, PHY_ANAR, wReg4) != PR_ERR_NONE)
      return PR_ERR_PHY_WR;
#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s -  wrote 0x%04X to PHY_ANAR Register\n", __FUNCTION__, wReg4);
#endif

   if (ifx_phy_write(chipId,  phy_addr, 0x9, wReg9) != PR_ERR_NONE)
      return PR_ERR_PHY_WR;
#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s -  wrote 0x%04X to GETH CTRL Register\n", __FUNCTION__, wReg9);
#endif

#if 0
   /* Re-start Autonegotiation */
#ifdef INCAIP2_DEBUG_GEPORT
   printk("%s -  Re-start Autonegotiation\n", __FUNCTION__);
#endif
   if (ifx_phy_read(chipId,  phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
      return PR_ERR_PHY_RD;

   dwTmp |= 0x2200; /* 8 bit Autonegotiation,9th bit Restart autonegotiation*/

   if (ifx_phy_write(chipId,  phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
      return PR_ERR_PHY_WR;
#endif
#ifdef INCAIP2_DEBUG_GEPORT
   if (ifx_phy_read(chipId, phy_addr, PHY_ANAR, &wReg4) != PR_ERR_NONE)
      return PR_ERR_PHY_RD;
   printk("%s -  read 0x%04X from PHY_ANAR Register\n", __FUNCTION__, wReg4);
#endif

   return PR_ERR_NONE;
}



/*****************************************************************************
 *
 *   FUNCTION:   ifx_phy_speed_ge
 *
 *   DESCRIPTION: Set the phy speed to 1000 or 100 or 10 mbps
 *
 *   PARAMETERS:
 *      IN:
 *         chipId      Chip Id
 *         portId      port Id (0 - 49)
 *         mdio_id      MDIO Id
 *         phy_addr   MDIO address of the PHY device
 *         mode       Speed(1000 or 100 or 10 mbps)
 *      OUT:
 *         None
 *
 *   RETURNS:
 *      PR_ERR_NONE      Success
 *     PR_ERR_CHIP_ID       Invalid chip Id
 *     PR_ERR_PHY_WR        If Phy write fails
 *     PR_ERR_PHY_RD       If Phy read fails
 *
 ***************************************************************************/

s32 ifx_phy_speed_ge( u8 chipId, u32 portId, u32 speed )
{
   u16   dwTmp=0;
   u8 phy_addr;
   s32 retflag;

#ifdef PARAMETER_CHECK

   if(chipId >= Total_Chips)
      return PR_ERR_CHIP_ID ;

#endif
   if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
      return retflag;

   if((retflag = ifx_get_portId_phyId( chipId, portId, &phy_addr)) != PR_ERR_NONE)
   {
      return retflag;
   }

   if (ifx_phy_read( chipId, phy_addr, PHY_CTRL, &dwTmp ) != PR_ERR_NONE)
   {
      ifx_sem_give(chipId,SEM_MDIO);
      return PR_ERR_PHY_RD;
   }

   dwTmp &= ~((0x1 << PHY_GE_SPEED1_POS) | (0x1 << PHY_GE_SPEED2_POS));
   dwTmp |= ((speed & 0x1) << PHY_GE_SPEED1_POS);
   dwTmp |= (((speed >> 1) & 0x1) << PHY_GE_SPEED2_POS) ;

   dwTmp |= (0x1 << PHY_GE_RESET_POS);  /* PHY Reset*/

   if (ifx_phy_write(chipId,  phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
   {
      ifx_sem_give(chipId,SEM_MDIO);
      return PR_ERR_PHY_WR;
   }

   if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
      return retflag;

   return PR_ERR_NONE ;

}
