/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup API API-Functions */
/** \addtogroup Internal Internally used functions */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/smp_lock.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/uaccess.h>
#include <linux/init.h>

#include <asm/incaip2/ifx_types.h>
#include "switch_api.h"
#include "ioctl.h"
#include "ioctl_struct.h"

#include <fe_port.h>
#include <ge_port.h>
#include <mib_counters.h>
#include <monitoring.h>
#include <phy.h>
#include <ports.h>
#include <trunk.h>
#include <phy_prv.h>
#include <ports_prv.h>


/**
 * Port IOCTL handler.
 * IOCTL handler for port configuration modification. 
 * The following IOCTLs are supported:
 * - #IFX_GET_PORT_MIB_COUNTERS
 * - #IFX_SET_INGRESS_MONITOR_FLAG
 * - #IFX_GET_INGRESS_MONITOR_FLAG
 * - #IFX_SET_EGRESS_MONITOR_FLAG
 * - #IFX_GET_EGRESS_MONITOR_FLAG
 * - #IFX_PHY_READ
 * - #IFX_PHY_WRITE
 * - #IFX_SET_PORT_LOCK
 * - #IFX_GET_PORT_LOCK
 * - #IFX_SET_PORT_VLANID
 * - #IFX_GET_PORT_VLANID
 * - #IFX_SET_PORT_INGRESS_VLAN_TAG
 * - #IFX_GET_PORT_INGRESS_VLAN_TAG
 * - #IFX_SET_PORT_INGRESS_VLAN_FILTER
 * - #IFX_GET_PORT_INGRESS_VLAN_FILTER
 * - #IFX_SET_PORT_COS
 * - #IFX_GET_PORT_COS
 * - #IFX_SET_PORT_JUMBO_ENABLE
 * - #IFX_GET_PORT_JUMBO_ENABLE

 *
 * \param   file   - File structure of device
 * \param   cmd    - IOCTL command 
 * \param   arg    - Argument for some IOCTL commands
 * \return  0      - OK
 * \return  ENOTTY - Invalid command
 * \ingroup API  
 */
int port_ioctl(struct file *file,unsigned int cmd, unsigned long arg)
{
    int err=0;
    /*structures for arg*/
    portMode_api    pm_api;
    mibCounters_api mc_api;
    Flag_api        fg_api;
    phyAction_api   pa_api;
    portVlan_api pv_api;
    portCoS_api pc_api;
    mdiomode_api mm_api;
    port_eth_conf_api p_eth_conf_api;
    

    switch(cmd)
    {

        case IFX_SET_PORT_FE_MODE:
            DPRINTK("%s: IFX_SET_PORT_FE_MODE\n",__FUNCTION__);
            copy_from_user(&pm_api,(portMode_api *) arg,sizeof (portMode_api));
            err = ifx_set_port_FE_mode( pm_api.chipId,
                                        pm_api.portId, 
                                        pm_api.feature, 
                                        pm_api.mode);
            break;
        case IFX_GET_PORT_FE_MODE:
            DPRINTK("%s: IFX_GET_PORT_FE_MODE\n",__FUNCTION__);
            copy_from_user(&pm_api,(portMode_api *) arg,sizeof (portMode_api));
            err = ifx_get_port_FE_mode(pm_api.chipId,
                                       pm_api.portId, 
                                       pm_api.feature,
                                       &pm_api.mode);
            if(err == 0)
                copy_to_user((portMode_api *) arg, &pm_api, sizeof(portMode_api));
            break;

        case IFX_SET_PORT_GE_MODE:
            DPRINTK("%s: IFX_SET_PORT_GE_MODE\n",__FUNCTION__);
            copy_from_user(&pm_api,(portMode_api *) arg,sizeof (portMode_api));
            err = ifx_set_port_GE_mode(pm_api.chipId,pm_api.portId, pm_api.feature, pm_api.mode);
            break;
        case IFX_GET_PORT_GE_MODE:
            DPRINTK("%s: IFX_GET_PORT_GE_MODE\n",__FUNCTION__);
            copy_from_user(&pm_api,(portMode_api *) arg,sizeof (portMode_api));
            err = ifx_get_port_GE_mode(pm_api.chipId,pm_api.portId, pm_api.feature,&pm_api.mode);
            if(err == 0)
                      copy_to_user((portMode_api *) arg, &pm_api, sizeof(portMode_api));
            break;

        case IFX_GET_PORT_MIB_COUNTERS:
        case IFX_GET_MAC_MIB_COUNTERS:
            copy_from_user(&mc_api,(mibCounters_api *) arg,sizeof (mibCounters_api));
            if( cmd == IFX_GET_PORT_MIB_COUNTERS )
            {
                DPRINTK("%s: IFX_GET_PORT_MIB_COUNTERS\n",__FUNCTION__);
                err = ifx_get_port_MIB_counters(mc_api.chipId,mc_api.portId, mc_api.counterId,&mc_api.count);
            }
            else
            {
                DPRINTK("%s: IFX_GET_MAC_MIB_COUNTERS\n",__FUNCTION__);                
                err = ifx_get_mac_MIB_counters(mc_api.chipId,mc_api.portId, mc_api.counterId,&mc_api.count);
            }
            if(err == 0)
                copy_to_user((mibCounters_api *) arg, &mc_api, sizeof(mibCounters_api));
            break;
            
        case IFX_SET_INGRESS_MONITOR_FLAG:
            DPRINTK("%s: IFX_SET_INGRESS_MONITOR_FLAG\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_set_ingress_monitor_flag(fg_api.chipId,fg_api.portId, fg_api.flag);
            break;
        case IFX_GET_INGRESS_MONITOR_FLAG:
            DPRINTK("%s: IFX_GET_INGRESS_MONITOR_FLAG\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err =  ifx_get_ingress_monitor_flag(fg_api.chipId,fg_api.portId,  &fg_api.flag);
            if(err == 0)
                      copy_to_user((Flag_api *) arg, &fg_api, sizeof(Flag_api));
            break;
        case IFX_SET_EGRESS_MONITOR_FLAG:
            DPRINTK("%s: IFX_SET_EGRESS_MONITOR_FLAG\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            /*DPRINTK("flag = %d\n", fg_api.flag);*/
            err =  ifx_set_egress_monitor_flag(fg_api.chipId,fg_api.portId, fg_api.flag);
            break;
        case IFX_GET_EGRESS_MONITOR_FLAG:
            DPRINTK("%s: IFX_GET_EGRESS_MONITOR_FLAG\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err =  ifx_get_egress_monitor_flag(fg_api.chipId,fg_api.portId, &fg_api.flag);
            if(err == 0)
                      copy_to_user((Flag_api *) arg, &fg_api, sizeof(Flag_api));
            break;
        case IFX_PHY_READ:
            DPRINTK("%s: IFX_PHY_READ\n",__FUNCTION__);
            copy_from_user(&pa_api,(phyAction_api *) arg,sizeof (phyAction_api));
            err = ifx_phy_read( pa_api.chipId,
                                /* pa_api.mdio_id,  */ 
                                pa_api.phy_addr, 
                                pa_api.RegOffset, 
                                &pa_api.Data );
            if(err == 0)
                      copy_to_user((phyAction_api *) arg, &pa_api, sizeof(phyAction_api));
            break;
        case IFX_PHY_WRITE:
            DPRINTK("%s: IFX_PHY_WRITE\n",__FUNCTION__);
            copy_from_user(&pa_api,(phyAction_api *) arg,sizeof (phyAction_api));
            err =  ifx_phy_write( pa_api.chipId,
                                  /* pa_api.mdio_id,  */ 
                                   pa_api.phy_addr, 
                                   pa_api.RegOffset, 
                                   pa_api.Data);
            break;
            
        case IFX_SET_PORT_ETH_CONF:
            DPRINTK("%s: IFX_SET_PORT_ETH_CONF\n",__FUNCTION__);
            copy_from_user( &p_eth_conf_api,(port_eth_conf_api *) arg,
                            sizeof (port_eth_conf_api));
            err =  ifx_set_port_eth_conf( p_eth_conf_api.chipId, 
                                          p_eth_conf_api.portId, 
                                          p_eth_conf_api.speed, 
                                          p_eth_conf_api.duplex, 
                                          p_eth_conf_api.autoneg,
                                          p_eth_conf_api.int_ext_phy);  
                                          
            if( err )
                printk("%s -> ifx_set_port_eth_conf returned %d\n", __FUNCTION__, err);
            break;            

            
        case IFX_GET_PORT_ETH_CONF:
            DPRINTK("%s: IFX_GET_PORT_ETH_CONF\n",__FUNCTION__);
            copy_from_user( &p_eth_conf_api,(port_eth_conf_api *) arg,
                            sizeof (port_eth_conf_api));
            err =  ifx_get_port_eth_conf( p_eth_conf_api.chipId, 
                                          p_eth_conf_api.portId, 
                                          (u32*)&p_eth_conf_api.speed, 
                                          (u32*)&p_eth_conf_api.duplex, 
                                          (BOOL*)&p_eth_conf_api.autoneg,
                                          (u32*)&p_eth_conf_api.int_ext_phy,
                                          (u32*)&p_eth_conf_api.link_status);   
            if(err == 0)
                copy_to_user((port_eth_conf_api *) arg, &p_eth_conf_api,
                              sizeof(port_eth_conf_api));                                          
            break;                        
            
            
        case IFX_SET_PORT_LOCK:
            DPRINTK("%s: IFX_SET_PORT_LOCK\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_set_port_lock(fg_api.chipId,fg_api.portId, fg_api.flag);
            break;
        case IFX_GET_PORT_LOCK:
            DPRINTK("%s: IFX_GET_PORT_LOCK\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_get_port_lock(fg_api.chipId,fg_api.portId, &fg_api.flag);
            if(err == 0)
                      copy_to_user((Flag_api *) arg, &fg_api, sizeof(Flag_api));
            break;
        case IFX_SET_PORT_VLANID:
            DPRINTK("%s: IFX_SET_PORT_VLANID\n",__FUNCTION__);
            copy_from_user(&pv_api,(portVlan_api *) arg,sizeof (portVlan_api));
            err = ifx_set_port_VLANId(pv_api.chipId,pv_api.portId, pv_api.vlanid);
            break;
        case IFX_GET_PORT_VLANID:
            DPRINTK("%s: IFX_GET_PORT_VLANID\n",__FUNCTION__);
            copy_from_user(&pv_api,(portVlan_api *) arg,sizeof (portVlan_api));
            err = ifx_get_port_VLANId(pv_api.chipId,pv_api.portId, &pv_api.vlanid);
            if(err == 0)
                      copy_to_user((portVlan_api *) arg, &pv_api, sizeof(portVlan_api));
            break;
        case IFX_SET_PORT_INGRESS_VLAN_TAG:
            DPRINTK("%s: IFX_SET_PORT_INGRESS_VLAN_TAG\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_set_port_ingress_VLAN_tag(fg_api.chipId,fg_api.portId, fg_api.flag);
            break;
        case IFX_GET_PORT_INGRESS_VLAN_TAG:
            DPRINTK("%s: IFX_GET_PORT_INGRESS_VLAN_TAG\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_get_port_ingress_VLAN_tag(fg_api.chipId,fg_api.portId, &fg_api.flag);
            if(err == 0)
                      copy_to_user((Flag_api *) arg, &fg_api, sizeof(Flag_api));
            break;
        case IFX_SET_PORT_INGRESS_VLAN_FILTER:
            DPRINTK("%s: IFX_SET_PORT_INGRESS_VLAN_FILTER\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_set_port_ingress_VLAN_filter(fg_api.chipId,fg_api.portId, fg_api.flag);
            break;
        case IFX_GET_PORT_INGRESS_VLAN_FILTER:
            DPRINTK("%s: IFX_GET_PORT_INGRESS_VLAN_FILTER\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_get_port_ingress_VLAN_filter(fg_api.chipId,fg_api.portId, &fg_api.flag);
            if(err == 0)
                      copy_to_user((Flag_api *) arg, &fg_api, sizeof(Flag_api));
            break;
        case IFX_SET_PORT_COS:
            DPRINTK("%s: IFX_SET_PORT_COS\n",__FUNCTION__);
            copy_from_user(&pc_api,(portCoS_api *) arg,sizeof (portCoS_api));
            err = ifx_set_port_CoS(pc_api.chipId,pc_api.portId, pc_api.CoS);
            break;
        case IFX_GET_PORT_COS:
            DPRINTK("%s: IFX_GET_PORT_COS\n",__FUNCTION__);
            copy_from_user(&pc_api,(portCoS_api *) arg,sizeof (portCoS_api));
            err = ifx_get_port_CoS(pc_api.chipId,pc_api.portId, &pc_api.CoS);
            if(err == 0)
                      copy_to_user((portCoS_api *) arg, &pc_api, sizeof(portCoS_api));
            break;
        case IFX_SET_PORT_JUMBO_ENABLE:
            DPRINTK("%s: IFX_SET_PORT_JUMBO_ENABLE\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_set_port_jumbo_enable(fg_api.chipId,fg_api.portId, fg_api.flag);
            break;
        case IFX_GET_PORT_JUMBO_ENABLE:
            DPRINTK("%s: IFX_GET_PORT_JUMBO_ENABLE\n",__FUNCTION__);
            copy_from_user(&fg_api,(Flag_api *) arg,sizeof (Flag_api));
            err = ifx_get_port_jumbo_enable(fg_api.chipId,fg_api.portId, &fg_api.flag);
            if(err == 0)
                      copy_to_user((Flag_api *) arg, &fg_api, sizeof(Flag_api));
            break;
       case IFX_GET_PORT_CONFIGURATION:
            DPRINTK("%s: IFX_GET_PORT_CONFIGURATION\n",__FUNCTION__);
            copy_from_user(&pc_api,(portCoS_api *) arg,sizeof (portCoS_api));
            err = ifx_get_port_default_configuration(pc_api.chipId,pc_api.portId, &pc_api.CoS);
            if(err == 0)
                      copy_to_user((portCoS_api *) arg, &pc_api, sizeof(portCoS_api));
            break;            
            
/*
        case IFX_SET_PORT_SVLAN:
            DPRINTK("%s: IFX_SET_PORT_SVLAN\n",__FUNCTION__);
            copy_from_user(&pv_api,(portVlan_api *) arg,sizeof (portVlan_api));
            err = ifx_set_port_sVLAN(pv_api.chipId,pv_api.portId, pv_api.vlanid);
            break;
        case IFX_GET_PORT_SVLAN:
            DPRINTK("%s: IFX_GET_PORT_SVLAN\n",__FUNCTION__);
            copy_from_user(&pv_api,(portVlan_api *) arg,sizeof (portVlan_api));
            err = ifx_get_port_sVLAN(pv_api.chipId,pv_api.portId, &pv_api.vlanid);
            if(err == 0)
                      copy_to_user((portVlan_api *) arg, &pv_api, sizeof(portVlan_api));
            break;
        case IFX_PORT_GROUP_TBI_INIT:
            DPRINTK("%s: IFX_PORT_GROUP_TBI_INIT\n",__FUNCTION__);
            copy_from_user(&tbi_api,(tbiInit_api *) arg,sizeof (tbiInit_api));
            err = ifx_port_group_tbi_init(tbi_api.chipId,tbi_api.portGroupId);
            break;
        case IFX_SET_PORT_TRUNK_MEMBER:
            DPRINTK("%s: IFX_SET_PORT_TRUNK_MEMBER\n",__FUNCTION__);
            copy_from_user(&tm_api,(trunckMember_api *) arg,sizeof (trunckMember_api));
            err =  ifx_set_port_trunk_member(tm_api.chipId,tm_api.portId, tm_api.groupSel);
            break;
        case IFX_GET_PORT_TRUNK_MEMBER:
            DPRINTK("%s: IFX_GET_PORT_TRUNK_MEMBER\n",__FUNCTION__);
            copy_from_user(&tm_api,(trunckMember_api *) arg,sizeof (trunckMember_api));
            err =  ifx_get_port_trunk_member(tm_api.chipId,tm_api.portId, &tm_api.groupSel);
            if(err == 0)
                      copy_to_user((trunckMember_api *) arg, &tm_api, sizeof(trunckMember_api));
            break;
*/
        case IFX_GET_MDIO_MODE:
            DPRINTK("%s: IFX_GET_MDIO_MODE\n",__FUNCTION__);
            copy_from_user(&mm_api, (mdiomode_api *) arg, sizeof (mdiomode_api));
            err = ifx_get_MDIO_mode(mm_api.chipId, mm_api.port_group, &mm_api.mdioMode, &mm_api.autoNegMode);
            if(err == 0)
                copy_to_user((mdiomode_api *) arg, &mm_api, sizeof(mdiomode_api));
            break;
        case IFX_SET_MDIO_MODE:
            DPRINTK("%s: IFX_SET_MDIO_MODE\n",__FUNCTION__);
            copy_from_user(&mm_api, (mdiomode_api *) arg, sizeof (mdiomode_api));
            err = ifx_set_MDIO_mode(mm_api.chipId, mm_api.port_group, mm_api.mdioMode, mm_api.autoNegMode);
            break;

        default:
            return -ENOTTY;
        break;

    }

    return err;
}
