/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */

#include <asm/types.h>
#include "struct.h"
#include "const.h"
#include <asm/incaip2/ifx_types.h>
#include "mib_counters.h"
#include "err.h"
#include "var.h"
#include <asm/addrspace.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/switch.h>

#define TX_DEFERED_CNT          26
#define TX_LATE_COLISION_CNT    25



/**
 * Get MAC MIB counters
 * Access the MIB counters for each individual port for a specified MIB counter Id.
 *
 * \param   chipId            Chip Id
 * \param   portId            port Id (0 - 2)
 * \param   counterId         Specifies the MIB counter that need to be accessed
 * \param   countPtr          Pointer to value of MIB counter for the specified 
 *                            counter Id of specified port.
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_PORTID     - If portId is not in the range of 0 - 2
 * \return  PR_ERR_COUNTER_ID - If MIB counter Id is not in the range of 0 - 26
 * \return  PR_ERR_NULL_PTR   - countPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Invalid Chip Id
 * \ingroup Internal
 */
s32 ifx_get_mac_MIB_counters(u8 chipId, u32 portId, u32 counterId, u32 *countPtr)
{
    u32 *ptr_mac_mib;
    u32 Temp=0;

#ifdef PARAMETER_CHECK
    
    if(portId >= TOTAL_PORTS-1) 
        return PR_ERR_PORTID;

    if(counterId > TOTAL_MAC_MIB_COUNT) 
        return PR_ERR_COUNTER_ID;   

    if(countPtr== NULL)  
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips) 
       return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID) 
        return PR_ERR_NULL_PTR;
#endif

    switch(portId)
    {
        case INCAIP2_SWITCH_PORT_SELECT_LAN: 
            ptr_mac_mib = (u32*)(SWITCH_P0_GMAC_MIB_REG_0 + counterId * 4); 
            break;  /* PC port base address */
        case INCAIP2_SWITCH_PORT_SELECT_PC: 
            ptr_mac_mib = (u32*)(SWITCH_P1_GMAC_MIB_REG_0 + counterId * 4); 
            break;  /* LAN port base address */
        default: 
            return PR_ERR_PORTID;
    }

    PR_Reg_Read(ptr_mac_mib, Temp);
/*    
    printk("%s - portId=%d, counterId=%d, counterVal=0x%08X\n ",__FUNCTION__,portId, counterId, Temp);
*/    
    *countPtr = Temp;
    return PR_ERR_NONE ;
}


s32 INCAIP2_SwitchMacGetMIBCounters( u32 portId, INCAIP2_SwitchMac_MIB_Table_t *MIB_TablePtr)
{
   u32 MIBbase;
   int index = 0;
   int value;
   int ret_value = -1;

   if ((portId < TOTAL_PORTS-1) && (MIB_TablePtr != NULL ))
   {
     /* get semaphore */
/*     semTake(sw21553_SwitchAccessSem, WAIT_FOREVER) ;  */

        switch(portId)
        {
            case INCAIP2_SWITCH_PORT_SELECT_LAN: 
                MIBbase = SWITCH_P0_GMAC_MIB_REG_0; 
                break;  /* PC port base address */
            case INCAIP2_SWITCH_PORT_SELECT_PC: 
                MIBbase = SWITCH_P1_GMAC_MIB_REG_0; 
                break;  /* LAN port base address */
            default: 
                return(ret_value);
        }

        for ( index = 0; index < sizeof(INCAIP2_SwitchMac_MIB_Table_t); index +=4 )
        {
            /* read MIB counter values and copy them to MIB counter structure */
            PR_Reg_Read( (u32*)(MIBbase + index), value );
            *((u32*)((u32)MIB_TablePtr+index)) = (u32)value;
        }

     /* free semaphore before return */
/*     semGive(sw21553_SwitchAccessSem) ;  */
        ret_value = 0;
   }
   return(ret_value);
}

/**
 * Get Port MIB counters
 * Access the MIB counters for each individual port for a specified MIB counter Id.
 *
 * \param   chipId            Chip Id
 * \param   portId            port Id (0 - 2)
 * \param   counterId         Specifies the MIB counter that need to be accessed
 * \param   countPtr          Pointer to value of MIB counter for the specified 
 *                            counter Id of specified port.
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_PORTID     - If portId is not in the range of 0 - 2
 * \return  PR_ERR_COUNTER_ID - If MIB counter Id is not in the range of 0 - 26
 * \return  PR_ERR_NULL_PTR   - countPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - Invalid Chip Id
 * \ingroup Internal
 */
s32 ifx_get_port_MIB_counters(u8 chipId, u32 portId, u32 counterId, u32 *countPtr)
{
    u32 *ptr_port_mib;
    u32 Temp=0;

#ifdef PARAMETER_CHECK
    
    if(portId >= TOTAL_PORTS-1) 
        return PR_ERR_PORTID;
    

    if(counterId > TOTAL_PORT_MIB_COUNT) 
        return PR_ERR_COUNTER_ID;   
    

    if(countPtr== NULL)  
        return PR_ERR_NULL_PTR;
    

    if(chipId >= Total_Chips) 
       return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID) 
        return PR_ERR_NULL_PTR;

#endif

    switch(portId)
    {
        case INCAIP2_SWITCH_PORT_SELECT_LAN: 
            ptr_port_mib = (u32*)(SWITCH_P0_PORT_MIB_REG_0 + counterId * 4); 
            break;  /* PC port base address */
        case INCAIP2_SWITCH_PORT_SELECT_PC: 
            ptr_port_mib = (u32*)(SWITCH_P1_PORT_MIB_REG_0 + counterId * 4); 
            break;  /* LAN port base address */
        default: 
            return PR_ERR_PORTID;
    }

    PR_Reg_Read(ptr_port_mib, Temp);
    *countPtr = Temp;
/*    
    printk("%s - portId=%d, counterId=%d, counterVal=0x%08X\n ",__FUNCTION__,portId, counterId, Temp);
*/    
    return PR_ERR_NONE ;
}

s32 INCAIP2_SwitchPortGetMIBCounters( u32 portId, INCAIP2_SwitchPort_MIB_Table_t *MIB_TablePtr)
{
   u32 MIBbase;
   int index = 0;
   int value;
   int ret_value = -1;

   if ((portId < TOTAL_PORTS-1) && (MIB_TablePtr != NULL ))
   {
     /* get semaphore */
/*     semTake(sw21553_SwitchAccessSem, WAIT_FOREVER) ;  */

        switch(portId)
        {
            case INCAIP2_SWITCH_PORT_SELECT_LAN: 
                MIBbase = SWITCH_P0_PORT_MIB_REG_0; 
                break;  /* PC port base address */
            case INCAIP2_SWITCH_PORT_SELECT_PC: 
                MIBbase = SWITCH_P1_PORT_MIB_REG_0; 
                break;  /* LAN port base address */
            default: 
                return(ret_value);
        }

        for ( index = 0; index < sizeof(INCAIP2_SwitchPort_MIB_Table_t); index +=4 )
        {
            /* read MIB counter values and copy them to MIB counter structure */
            PR_Reg_Read( (u32*)(MIBbase + index), value );
            *((u32*)((u32)MIB_TablePtr+index)) = (u32)value;
        }

     /* free semaphore before return */
/*     semGive(sw21553_SwitchAccessSem) ;  */
        ret_value = 0;
   }
   return(ret_value);
}


