/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */

#include <asm/types.h>
#include <asm/addrspace.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/ifx_types.h>
#include "struct.h"
#include "const.h"
#include "monitoring.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"

/**
 * Set egress monitor flag
 * Set the Egress monitor flag for the specified port.
 *
 * \param   chipId              Chip Id
 * \param   portId              port Id (range is 0 - 2)
 * \param   egressMonitorFlag   Value of TRUE will cause all packets from the 
 *                              portId to be monitored on the monitoring port
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_PORTID       - If portId is not in the range of 0 - 2
 * \return  PR_ERR_NULL_PTR     - base address is null pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32  ifx_set_egress_monitor_flag(u8 chipId,u32 portId, BOOL egressMonitorFlag)
{

    u32 *ptr_egress_monitor;
    u32 Temp=0;
    u32 pos=0;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_egress_monitor = (u32*)(Chip_Base_Address[chipId] + EG_MON_REG) + (portId >> 5);
    pos = (portId < 32) ? portId : (portId - 32) ;

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    PR_Reg_Read(ptr_egress_monitor,Temp);

    Temp &=  ~(1 << pos);
    Temp |= egressMonitorFlag << pos ;

    PR_Reg_Write(ptr_egress_monitor, Temp);

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE ;    
}

/**
 * Get egress monitor flag
 * Get the Egress monitor flag for the specified port.
 *
 * \param   chipId                Chip Id
 * \param   portId                port Id (range is 0 - 2)
 * \param   egressMonitorFlagPtr  Pointer to Egress monitor flag.Value of TRUE 
 *                                will cause all packets from the portId to 
 *                                be monitored on the monitoring port
 * \return  PR_ERR_NONE           - Success 
 * \return  PR_ERR_PORTID         - If portId is not in the range of 0 - 2
 * \return  PR_ERR_NULL_PTR       - egressMonitorFlagPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID        - If the chip Id invalid
 * \ingroup Internal
 */
s32  ifx_get_egress_monitor_flag(u8 chipId,u32 portId, BOOL *egressMonitorFlagPtr)
{

    u32 *ptr_egress_monitor;
    u32 Temp=0;
    u32 pos=0;
#ifdef PARAMETER_CHECK

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(egressMonitorFlagPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_egress_monitor = (u32*)(Chip_Base_Address[chipId] + EG_MON_REG) + (portId >> 5);
    pos = (portId < 32) ? portId : (portId - 32) ;

    PR_Reg_Read(ptr_egress_monitor,Temp);

    *egressMonitorFlagPtr = (Temp >> pos) & 0x01 ;

    return PR_ERR_NONE ;
}

/**
 * Set ingress monitor flag
 * Set the Ingress monitor flag for the specified port.
 *
 * \param   chipId              Chip Id
 * \param   portId              port Id (range is 0 - 2)
 * \param   ingressMonitorFlag  Value of TRUE will cause all packets from the 
 *                              portId to be monitored on the monitoring port
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_PORTID       - If portId is not in the range of 0 - 2
 * \return  PR_ERR_TIME_OUT     - Time out
 * \return  PR_ERR_NULL_PTR     - base address is null pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32  ifx_set_ingress_monitor_flag(u8 chipId,u32 portId, BOOL ingressMonitorFlag)
{

    u32 Temp=0;
    s32 i = 0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1=0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    /* set pointers to ACS_CTRL and DATA register */
    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /* 
     * Port Configuration table access to set ingress monitor information 
     * for passed portId. Read current configuration first before overwriting 
     * ingress monitoring information.
     */
    ptr_cpu_acs_ctrl -> Table_sel  = PORT_CONFIG_TBL;
    ptr_cpu_acs_ctrl -> Index      = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack    = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    Temp &= PORTX_MIRROR_MASK;
    Temp |=  ingressMonitorFlag << PORTX_MIRROR_POS;

    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE ;
}

/**
 * Get ingress monitor flag
 * Get the Ingress monitor flag for the specified port.
 *
 * \param   chipId                Chip Id
 * \param   portId                port Id (range is 0 - 2)
 * \param   ingressMonitorFlagPtr Pointer to Ingress monitor flag.Value of TRUE 
 *                                will cause all packets from the portId to 
 *                                be monitored on the monitoring port
 * \return  PR_ERR_NONE           - Success 
 * \return  PR_ERR_PORTID         - If portId is not in the range of 0 - 2
 * \return  PR_ERR_TIME_OUT       - Time out
 * \return  PR_ERR_NULL_PTR       - ingressMonitorFlagPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID        - If the chip Id invalid
 * \ingroup Internal
 */
s32  ifx_get_ingress_monitor_flag(u8 chipId,u32 portId, BOOL *ingressMonitorFlagPtr)
{

    u32 Temp =0;
    s32 i = 0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1=0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId >= TOTAL_PORTS) 
        return PR_ERR_PORTID;

    if(ingressMonitorFlagPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
 
    /* set pointers to ACS_CTRL and DATA register */
    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;


    /* 
     * Port Configuration table access to get ingress monitor information 
     * for passed portId
     */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFIG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    *ingressMonitorFlagPtr = (Temp >> PORTX_MIRROR_POS) & 1;

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE ;
}
