 /************************************************************************
 *
 * Copyright (c) 2006
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */
/** \addtogroup Obsolete Will be removed soon */
#include <linux/kernel.h>
#include <asm/types.h>
#include <asm/incaip2/ifx_types.h>
#include <asm/system.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/switch.h>
#include "const.h"
#include "phy.h"
#include "phy_prv.h"
#include "struct.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"

#include "fe_port.h"
#include "ge_port.h"
#include "fe_port_prv.h"
#include "ge_port_prv.h"


#undef INCAIP2_DEBUG_PHY

extern s32 ifx_get_switch_mode(u8 chipId,u32 feature, u32* modePtr);


/**
 * Set the Auto-Negotiation Enable/Disable Status for the specified port.
 * This function writes the passed autonegotiation parameter to the
 * PHY_CTRL register.
 *
 * \param   chipId          Chip Id
 * \param   portId          PortId(range is 0-1)
 * \param   Enable          Enable/Disable autonegotiation for the specified port
 *
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - portId is not in the range of 0 - 1
 * \return  PR_ERR_NULL_PTR     - base address is null pointer
 * \return  PR_ERR_CHIP_ID      - Invalid Chip Id
 * \return  PR_ERR_PHY_RD       - Phy read fails
 * \return  PR_ERR_PHY_WR       - Phy write fails
 * \ingroup Internal
 */
s32 ifx_set_port_autoneg(u8 chipId,u32 portId, BOOL Enable)
{
    u16 dwTmp=0;
    u8  phy_addr=0;
    s32 retflag =0;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, portId=%d, Enable=%d )\n",
          __FUNCTION__, chipId, portId, Enable);
#endif
#ifdef PARAMETER_CHECK

    if(portId >= (IFX_TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
#endif

    if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    ifx_get_portId_phyId( chipId, portId, &phy_addr);

    if( ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE )
    {
        ifx_sem_give(chipId,SEM_MDIO);
        return PR_ERR_PHY_RD;
    }

    dwTmp &= ~(0x1 << PHY_AUTONEGO_POS);
    dwTmp |= ((Enable & 0x1) << PHY_AUTONEGO_POS) ;

    if(Enable == TRUE)
        dwTmp |= (0x1 << PHY_RESTRT_AUTONEGO_POS);  /* Restart Auto negotiation */
#if 0
#ifndef _VIPER
    else
    {
        /* in case of disabled autonegotiation a PHY reset is necessary
           to force the PHY to take over new settings
           (e.g. 10|HD|NOANEG -> 10|FD|NOANEG) */
#ifdef INCAIP2_DEBUG_PHY
        printk("%s - Reset PHY\n",  __FUNCTION__);
#endif
        dwTmp |= (0x1 << PHY_FE_RESET_POS); /* Reset PHY */
    }
#endif
#endif

#ifdef INCAIP2_DEBUG_PHY
    printk("%s - write 0x%04X to PHY_CTRL register\n",
          __FUNCTION__, dwTmp);
#endif
    if (ifx_phy_write(chipId, phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
    {
        ifx_sem_give(chipId,SEM_MDIO);
        return PR_ERR_PHY_WR;
    }

    if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;
#ifdef INCAIP2_DEBUG_PHY
    if(Enable == TRUE)
      printk("Port %d PHY ANEG enabled and restarted\n", portId);
    else
      printk("Port %d PHY ANEG disabled \n", portId);
#endif
    return PR_ERR_NONE;
}


/**
 * Get the Auto-Negotiation Enable/Disable Status for the specified port.
 * This function reads the autonegotiation status from the PHY_CTRL register.
 *
 * \param   chipId          Chip Id
 * \param   portId          PortId(range is 0-1)
 * \param   EnablePtr       Pointer to Enable/Disable Status of autonegotiation
 *                          for the specified port
 *
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - portId is not in the range of 0 - 1
 * \return  PR_ERR_NULL_PTR     - base address is null pointer
 * \return  PR_ERR_CHIP_ID      - Invalid Chip Id
 * \return  PR_ERR_PHY_RD       - Phy read fails
 * \ingroup Internal
 */
s32 ifx_get_port_autoneg( u8 chipId, u32 portId, BOOL * EnablePtr )
{
    u16    dwTmp=0;
    u8     phy_addr=0;
    s32   retflag =0;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, portId=%d )\n", __FUNCTION__, chipId, portId);
#endif
#ifdef PARAMETER_CHECK

    if(EnablePtr == NULL)
        return PR_ERR_NULL_PTR;

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
#endif

    if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    ifx_get_portId_phyId(chipId,portId, &phy_addr) ;

    if(ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
    {
        ifx_sem_give(chipId,SEM_MDIO);
        return PR_ERR_PHY_RD;
    }
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read 0x%08X from PHY_CTRL register of phy_addr %d\n",
           __FUNCTION__, dwTmp, phy_addr);
#endif
    *EnablePtr = (BOOL) ((dwTmp >> PHY_AUTONEGO_POS) & 0x0001);

    if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE ;
}

/*****************************************************************************
  *
  * FUNCTION:       ifx_start_port_autoneg
  *
  * DESCRIPTION:    Restart the Auto-Negotiation for the specified port.
  *
  * PARAMETERS:
  *     IN:
  *         chipId      Chip Id
  *         portId      port Id (0 - 49)
  *     OUT:
  *         None
  *
  * RETURNS:
  *     PR_ERR_NONE     Success
  *     PR_ERR_PORTID   If portId is not in the range of 0 - 49
  *     PR_ERR_CHIP_ID  Invalid Chip Id
  *     PR_ERR_PHY_RD   If Phy register read fails
  *     PR_ERR_PHY_WR   If Phy register write fails
  *
  ***************************************************************************/

s32 ifx_start_port_autoneg( u8 chipId, u32 portId )
{
    u16   dwTmp=0;
    u8    phy_addr=0;
    s32   retflag =0;

#ifdef PARAMETER_CHECK

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif

    if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    ifx_get_portId_phyId(chipId,portId, /* &mdio_id, */ &phy_addr) ;

    if (ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_RD;

    dwTmp |= (0x01 << PHY_AUTONEGO_POS) ;  /* Auto negotiation enable */
    dwTmp |= (0x01 << PHY_RESTRT_AUTONEGO_POS) ;  /* Restart Auto-Nego bit */

    if (ifx_phy_write(chipId, /* mdio_id,*/ phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
    {
        ifx_sem_give(chipId,SEM_MDIO);
        return PR_ERR_PHY_WR;
    }

    if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE ;
}

/**
 * Get MDIO and autonegotiation mode.
 * Get MDIO mode and Hardware Autonegotiation mode.
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id(range is 0-1)
 * \param   mdioMode        0 - MDIO interface is not used for Port configuration
 *                          1 - MDIO interface is used for Port configuration
 * \param   autoNegMode     0 - Hardware Auto negotiation is not used
 *                          1 - Hardware Auto negotiation is used
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Obsolete
 */
s32 ifx_get_MDIO_mode( u8 chipId, u32 portId,
                        BOOL * mdioMode, BOOL * autoNegMode)
{
    u32  dwTmp=0;
    u32* ptr_mdio_mod_sel;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, portId=%d)\n",
           __FUNCTION__, chipId, portId);
#endif

#ifdef PARAMETER_CHECK

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_mdio_mod_sel = (u32*) (Chip_Base_Address[chipId] + MDIO_MOD_SEL );

    PR_Reg_Read(ptr_mdio_mod_sel, dwTmp) ;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read value 0x%08X from MDIO_MOD_SEL register at address 0x%08X\n",
           __FUNCTION__, dwTmp, MDIO_MOD_SEL);
#endif
    *mdioMode = (dwTmp & MDIO_MODE_MASK) >> MDIO_MODE_POS ;
    *autoNegMode =  (dwTmp >> portId) & AUTONEG_MODE_MASK;
#ifdef INCAIP2_DEBUG_PHY
   printk("%s - MDIO Mode=%d\n", __FUNCTION__, *mdioMode);
   printk("%s - ANEG Mode=%d\n", __FUNCTION__, *autoNegMode);
#endif
    return PR_ERR_NONE;
}

/**
 * Set MDIO phy usage.
 * Set the usage of either internal or external PHY in the MDIO_MOD_SEL register.
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id(range is 0-1)
 * \param   int_ext_phy     Internal/External PHY select (0-internal 1-external)
 *
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Obsolete
 */
s32 ifx_set_MDIO_phy( u8 chipId, u32 portId, BOOL int_ext_phy )
{
    u32  dwTmp = 0;
    u32  *ptr_mdio_mod_sel;
    s32   retflag =0;

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_mdio_mod_sel = (u32*) (Chip_Base_Address[chipId] + MDIO_MOD_SEL );

    if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    PR_Reg_Read(ptr_mdio_mod_sel, dwTmp) ;

    dwTmp &= ~((1 << MDIO_INT_EXT_POS ) );
    dwTmp |= (int_ext_phy << MDIO_INT_EXT_POS );

    PR_Reg_Write(ptr_mdio_mod_sel, dwTmp) ;

    if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}

/**
 * Get MDIO phy usage.
 * Get the PHY usage information (internal or external PHY) from the
 * MDIO_MOD_SEL register.
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id(range is 0-1)
 * \param   int_ext_phy_ptr Pointer to phy usage information (0-internal 1-external)
 *
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Obsolete
 */
s32 ifx_get_MDIO_phy( u8 chipId, u32 portId, BOOL* int_ext_phy_ptr )
{
    u32  dwTmp = 0;
    u32  *ptr_mdio_mod_sel;
    s32   retflag =0;

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if( int_ext_phy_ptr == NULL )
        return PR_ERR_NULL_PTR;

#endif

    ptr_mdio_mod_sel = (u32*) (Chip_Base_Address[chipId] + MDIO_MOD_SEL );

    if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    PR_Reg_Read(ptr_mdio_mod_sel, dwTmp) ;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - MDIO_MOD_SEL=0x%08X\n", __FUNCTION__, dwTmp );
#endif
    *int_ext_phy_ptr = (BOOL)((dwTmp & MDIO_INT_EXT_MASK) >> MDIO_INT_EXT_POS);

    if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}



/**
 * Set MDIO and autonegotiation mode.
 * Set MDIO mode and Hardware Autonegotiation mode (per port).
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id(range is 0-1)
 * \param   mdioMode        0 - MDIO interface is not used for Port configuration
 *                          1 - MDIO interface is used for Port configuration
 * \param   autoNegMode     0 - Hardware Auto negotiation is not used
 *                          1 - Hardware Auto negotiation is used
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Obsolete
 */
s32 ifx_set_MDIO_mode(u8 chipId, u32 portId, BOOL mdioMode, BOOL autoNegMode)
{
    u32  dwTmp = 0;
    u32  *ptr_mdio_mod_sel;
    s32   retflag =0;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s(chipId=%d, portId=%d, mdioMode=%d, autoNegMode=%d)\n",
           __FUNCTION__, chipId, portId, mdioMode, autoNegMode );
#endif

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_mdio_mod_sel = (u32*) (Chip_Base_Address[chipId] + MDIO_MOD_SEL );

    if((retflag = ifx_sem_take(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    PR_Reg_Read(ptr_mdio_mod_sel, dwTmp) ;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read 0x%08X from MDIO_MOD_SEL register\n",
                       __FUNCTION__, dwTmp);
#endif

    dwTmp &= ~((1 << MDIO_MODE_POS ) | (1 << portId));
    dwTmp |= (mdioMode << MDIO_MODE_POS ) | (autoNegMode << portId);

    PR_Reg_Write(ptr_mdio_mod_sel, dwTmp) ;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s - wrote 0x%08X to MDIO_MOD_SEL register\n",
                       __FUNCTION__, dwTmp);
#endif

    if((retflag = ifx_sem_give(chipId,SEM_MDIO)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}



/********************************************* Internal APIs  **************************************************/

/**
 * Read PHY register.
 * Read from Phy register.
 *
 * \param   chipId          Chip Id
 * \param   phy_addr        MDIO address of the PHY device
 * \param   RegOffset       PHY register address
 * \param   pwData          Pointer to Data read from phy register
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_PHY_RD   - If phy read fails
 * \return  PR_ERR_NULL_PTR - pwData or base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_phy_read(u8 chipId, u8 phy_addr, u8 RegOffset, u16 * pwData)
{


    volatile u32      dwTmp=0;
    u32* ptr_mdio_acc;
    s32 i=0;
    unsigned int flags;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, phy_addr=%d, RegOffset=%d)\n",
           __FUNCTION__, chipId, phy_addr, RegOffset);
#endif
#ifdef PARAMETER_CHECK
    if(pwData == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    save_and_cli(flags);
    ptr_mdio_acc = (u32*)(Chip_Base_Address[chipId] + SWITCH_MDIO_ACC);
/*  ptr_mdio_acc = (u32*)(Chip_Base_Address[chipId] + SWITCH_MDIO_ACC + (mdio_id << 2)); */

    dwTmp = (0x00 << MDIO_ACC_READY_POS) |                  /* Clear Ready bit */
            (0x01 << MDIO_ACC_READ_POS) |                   /* Read */
            ((phy_addr & 0x1F) << MDIO_ACC_PHYID_POS) |     /* PHY address */
            ((RegOffset & 0x1F) << SWITCH_MDIO_ACCID_POS);  /* register index */

    PR_Reg_Write(ptr_mdio_acc,dwTmp);

    for ( i = 0 ; i < PHY_RETRY_TIME ; i++ )
    {
        PR_Reg_Read(ptr_mdio_acc,dwTmp);
        if ( ((dwTmp >> MDIO_ACC_READY_POS) & 0x01) == 0x1 )
            break ;
    }

    if (((dwTmp >> MDIO_ACC_READY_POS) & 0x01) == 0)
    {
        printk("%s - access ready timeout !!!\n", __FUNCTION__);
        restore_flags(flags);
        return PR_ERR_PHY_RD;
    }
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read 0x%04X from MDIO_ACC_REG\n", __FUNCTION__,dwTmp);
#endif
    *pwData = dwTmp & MDIO_ACC_DATA_MASK;
    restore_flags(flags);
    return PR_ERR_NONE;
}

/**
 * Write to PHY register.
 * Write 16 bit value wData to Phy register defined by passed
 * phy_addr and RegOffset.
 *
 * \param   chipId          Chip Id
 * \param   phy_addr        MDIO address of the PHY device
 * \param   RegOffset       PHY register address
 * \param   wData           Data to be written into phy register
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_PHY_WR   - If Phy write fails
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_phy_write( u8 chipId, u8 phy_addr, u8 RegOffset, u16 wData )
{
    volatile u32        dwTmp=0;
    u32* ptr_mdio_acc;
    s32 i=0;
    unsigned int flags;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, phy_addr=%d, RegOffset=%d, wData=0x%04X)\n",
           __FUNCTION__, chipId, phy_addr, RegOffset, wData);
#endif
#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    save_and_cli(flags);
    ptr_mdio_acc = (u32*)(Chip_Base_Address[chipId] + SWITCH_MDIO_ACC); /* + (mdio_id << 2));  */

    dwTmp = (0x00 << MDIO_ACC_READY_POS) |                      /* clear Ready bit */
            (0x01 << MDIO_ACC_WRITE_POS) |                      /* Write */
            ((phy_addr & 0x1F) << MDIO_ACC_PHYID_POS) |         /* PHY address */
            ((RegOffset & 0x1F) << SWITCH_MDIO_ACCID_POS) |     /* register index */
            (wData & MDIO_ACC_DATA_MASK) ;                      /* write data */

    PR_Reg_Write(ptr_mdio_acc,dwTmp);

    for ( i = 0 ; i < PHY_RETRY_TIME ; i++ )
    {
        PR_Reg_Read(ptr_mdio_acc,dwTmp);
        if ( ((dwTmp >> MDIO_ACC_READY_POS) & 0x01) == 0x1 )
            break ;
    }

    if (((dwTmp >> MDIO_ACC_READY_POS) & 0x01) == 0x0)
    {
        printk("%s - access ready timeout !!!\n", __FUNCTION__);
        restore_flags(flags);
        return PR_ERR_PHY_WR;
    }


    restore_flags(flags);
    return PR_ERR_NONE;
}

/**
 * Get PHY Id for passed portId.
 * Returns the PHY ID that is assigned to the passed port ID.
 *
 * \param   chipId                 Chip Id
 * \param   portId                 port Id (0 - 1)
 * \param   phy_adr                Pointer to PHY ID
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_CHIP_ID         - Invalid Chip Id
 * \return  PR_ERR_PORTID          - Invalid Port Id
 * \ingroup Internal
 */
s32 ifx_get_portId_phyId(u8 chipId, u8 portId, u8* phy_adr)
{
    u32     dwTmp=0;
    u32*    ptr_mdio_ID_reg;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s(chipId=%d, portId=%d)\n", __FUNCTION__, chipId, portId);
#endif
#ifdef PARAMETER_CHECK
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;
#endif

    /*
     * read MDIO ID Register content and extract PHY ID for passed portId
     */
    ptr_mdio_ID_reg = (u32*) SWITCH_MDIO_ID;
    PR_Reg_Read( ptr_mdio_ID_reg, dwTmp );
#ifdef INCAIP2_DEBUG_PHY
    printk("%s- read 0x%08X from SWITCH_MDIO_ID register at address 0x%08X\n",
           __FUNCTION__, dwTmp, SWITCH_MDIO_ID);
#endif
    *phy_adr = (u8)((dwTmp >> (portId*5)) & 0x1F);

#ifdef INCAIP2_DEBUG_PHY
    printk("%s- return phy_addr %d for port %d\n",
           __FUNCTION__, *phy_adr, portId);
#endif

    return PR_ERR_NONE;
}


/**
 * Set PHY Duplex Mode.
 * Set the Duplex Mode for the PHY assigned to the passed port Id to full duplex
 * or half duplex.
 *
 * \param   chipId                 Chip Id
 * \param   portId                 port Id (0 - 1)
 * \param   mode                   Full(1)/Half(0) duplex mode
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_PHY_WR          - Phy write fails
 * \return  PR_ERR_PHY_RD          - Phy read fails
 * \return  PR_ERR_CHIP_ID         - Invalid Chip Id
 * \return  PR_ERR_PORTID          - Invalid Port Id
 * \ingroup Internal
 */
s32 ifx_phy_set_duplex_mode( u8 chipId, u32 portId, BOOL mode )
{
    u16   dwTmp=0;
    u8    phy_addr;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s(chipId=%d, portId=%d, mode=%d)\n", __FUNCTION__, chipId, portId, mode);
#endif
#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;
#endif

    /*
     * Get phy ID for passed port id and get current PHY CTRL register content.
     * Write passed duplex mode parameter to PHY CTRL register.
     */
    ifx_get_portId_phyId( chipId, portId, &phy_addr) ;

    if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_RD;

    dwTmp &= ~(0x1 << PHY_DUPLEX_MODE_POS);
    dwTmp |= ((mode & 0x1) << PHY_DUPLEX_MODE_POS);
#if 0
#ifndef _VIPER
    printk("%s- reset PHY at port %d\n", __FUNCTION__, portId);
    dwTmp |= (0x1 << PHY_FE_RESET_POS);  /* PHY Reset*/
#endif
#endif
    if (ifx_phy_write(chipId, phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

    return PR_ERR_NONE ;
}


/**
 * Get PHY Duplex Mode.
 * Get the Duplex Mode (half duplex or full duplex) for the PHY assigned
 * to the passed port Id.
 *
 * \param   chipId                 Chip Id
 * \param   portId                 port Id (0 - 1)
 * \param   mode                   Pointer to PHY duplex status
 *                                 (Full(1)/Half(0) duplex mode)
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_CHIP_ID         - Chip Id invalid
 * \return  PR_ERR_PHY_RD          - Phy read fails
 * \ingroup Internal
 */
s32 ifx_phy_get_duplex_mode(u8 chipId, u32 portId, BOOL* mode)
{
    u16   dwTmp=0;
    u8    phy_addr;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s(chipId=%d, portId=%d)\n", __FUNCTION__, chipId, portId);
#endif
#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
#endif
    /*
     * get phy ID for passed port id and read current PHY CTRL register content.
     * Extract duplex mode and pass to output parameter mode.
     */
    ifx_get_portId_phyId( chipId, portId, &phy_addr ) ;

    if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_RD;

    *mode = (BOOL) ((dwTmp & PHY_DUPLEX_MODE_MASK) >> PHY_DUPLEX_MODE_POS);

    return PR_ERR_NONE ;
}




/*****************************************************************************
  *
  * FUNCTION:   ifx_phy_pause_fe
  *
  * DESCRIPTION: Set the pause control in the PHY.
  *
  * PARAMETERS:
  *     IN:
  *         chipId      Chip Id
  *         portId      Port Id (range is 0 - 49)
  *         Enable      Pause control flag
  *     OUT:
  *         None
  *
  * RETURNS:
  *     PR_ERR_NONE     Success
  *     PR_ERR_PORTID   If portId is not in the range of 0 - 49
  *     PR_ERR_PHY_WR   If Phy write fails
  *     PR_ERR_PHY_RD   If Phy read fails
  *
  ***************************************************************************/

s32 ifx_phy_pause_fe(u8 chipId,u32 portId, BOOL Enable)
{
    u16  dwTmp=0;
    u8   phy_addr=0;
    u16  dwTmp4=0;
    BOOL mode = 0;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s( portId=%d, Enable=%d ) \n", __FUNCTION__, portId, Enable);
#endif

#ifdef PARAMETER_CHECK

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif

    ifx_get_portId_phyId(chipId,portId, &phy_addr) ;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read PHY CTRL Register\n", __FUNCTION__);
#endif
    /* Read reg 0 */
    if(ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    if( ((dwTmp >> PHY_AUTONEGO_POS) & 0x01) == 0x0)
        return PR_ERR_PAUSE_AN_DIS; /* If Auto-negotiation is disabled */

    /*
     * Read PHY status register and check whether we are autonegotiation capable
     */
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read PHY STATUS Register\n", __FUNCTION__);
#endif
    if(ifx_phy_read(chipId, phy_addr, PHY_STATUS, &dwTmp) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    if( ((dwTmp >> PHY_AUTONEGO_CP_POS) & 0x01) == 0x0)
        return PR_ERR_PAUSE_AN_NC;

    /* Read reg 4 */
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - read PHY ANAR Register\n", __FUNCTION__);
#endif
    if(ifx_phy_read(chipId, phy_addr, PHY_ANAR, &dwTmp4) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    /* in half duplex mode pause control has to be disabled */
    dwTmp = ((dwTmp4 >> 5) & 0xF);

    if( ((dwTmp >> 3) & 0x1) == 0x1)
        mode = 1;
    else
        if( ((dwTmp >> 2) & 0x1) == 0x1)
            mode = 0;
        else
            if( ((dwTmp >> 1) & 0x1) == 0x1)
                mode = 1;
            else
                mode = 0;

    if(mode == 0)
        Enable = FALSE ;

    /*
     * write pause parameter to PHY ANAR and PHY CTRL register and
     * restart autonegotiation
     */
    dwTmp4 &= ~(0x1 << PHY_PAUSE_POS);
    dwTmp4 |= ((Enable & 0x1) << PHY_PAUSE_POS) ;
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - write to PHY ANAR Register\n", __FUNCTION__);
#endif
    if (ifx_phy_write(chipId, phy_addr, PHY_ANAR, dwTmp4) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s -  wrote 0x%04X to PHY_ANAR Register\n", __FUNCTION__, dwTmp4);
#endif

    if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_RD;

    dwTmp |= (0x01 << PHY_RESTRT_AUTONEGO_POS) ;  /* Restart Auto-Nego bit */
#ifdef INCAIP2_DEBUG_PHY
    printk("%s - Restart AUTONEGOTIATION\n", __FUNCTION__);
#endif
   if (ifx_phy_write(chipId, phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

    if(mode == 0)
        return PR_ERR_PAUSE_HD;

    return PR_ERR_NONE;
}



/*****************************************************************************
  *
  * FUNCTION:   ifx_phy_pause_rx
  *
  * DESCRIPTION: Set the Rx pause control in the PHY.
  *
  * PARAMETERS:
  *     IN:
  *         chipId      Chip Id
  *         portId      Port Id (range is 0,8,16,24,32,40,48,49)
  *         Enable      Pause control flag
  *     OUT:
  *         None
  *
  * RETURNS:
  *     PR_ERR_NONE     Success
  *     PR_ERR_PORTID   If portId is not in the range
  *     PR_ERR_PHY_WR   If Phy write fails
  *     PR_ERR_PHY_RD   If Phy read fails
  *
  ***************************************************************************/

s32 ifx_phy_pause_rx(u8 chipId,u32 portId, BOOL Enable)
{
    u16  dwTmp=0;
    u8   phy_addr=0;
    u16  dwTmp4=0;
    u16  dwTmp9=0;
    u16  dwTmp_ge=0,dwTmp_fe=0;

    BOOL mode = 0;


#ifdef PARAMETER_CHECK

    if(portId > GE_PORT_1)
        return PR_ERR_PORTID;


    if(portId < GE_PORT_1)
    {
        if((portId & 7) != 0 )
            return PR_ERR_PORTID;
    }

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif


    ifx_get_portId_phyId( chipId, portId, &phy_addr );

   /* Read reg 0 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    if( ((dwTmp >> PHY_AUTONEGO_POS) & 0x01) == 0x0)
        return PR_ERR_PAUSE_AN_DIS; /* If Auto-negotiation is disabled */

   /* Read reg 1 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_STATUS, &dwTmp) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    if( ((dwTmp >> PHY_AUTONEGO_CP_POS) & 0x01) == 0x0)
        return PR_ERR_PAUSE_AN_NC;  /* If Auto negotiation is not completed */

   /* Read reg 9 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, 0x9, &dwTmp9) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

   /* Read reg 4 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_ANAR, &dwTmp4) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    dwTmp_ge = ((dwTmp9 >> 8) & 0x3);
    dwTmp_fe = ((dwTmp4 >> 5) & 0x7F);

    if( ((dwTmp_ge >> 1) & 0x1) == 0x1)
        mode = 1;
    else
        if( (dwTmp_ge & 0x1) == 0x1)
            mode = 0;
        else
            if( ((dwTmp_fe >> 3) & 0x1) == 0x1)
                mode = 1;
            else
                if( ((dwTmp_fe >> 2) & 0x1) == 0x1)
                        mode = 0;
                else
                    if( ((dwTmp_fe >> 1) & 0x1) == 0x1)
                        mode = 1;
                    else
                        mode = 0;

    if(mode == 0)
        Enable = FALSE;


    if(Enable == TRUE)
    {
        dwTmp4 |= (0x1 << PHY_PAUSE_POS);
    }
    else
    {
        dwTmp4 &= ~(0x1 << PHY_PAUSE_POS);
    }


    if (ifx_phy_write(chipId, phy_addr, PHY_ANAR, dwTmp4) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s -  wrote 0x%04X to PHY_ANAR Register\n", __FUNCTION__, dwTmp4);
#endif

    if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_RD;

    dwTmp |= (0x01 << PHY_RESTRT_AUTONEGO_POS) ;  /* Restart Auto-Nego bit */

    if (ifx_phy_write(chipId, phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

    if(mode == 0)
        return PR_ERR_PAUSE_HD;

    return PR_ERR_NONE;
}



/*****************************************************************************
  *
  * FUNCTION:   ifx_phy_pause_tx
  *
  * DESCRIPTION: Set the Tx pause control in the PHY.
  *
  * PARAMETERS:
  *     IN:
  *         chipId      Chip Id
  *         portId      Port Id (range is 0,8,16,24,32,40,48,49)
  *         Enable      Pause control flag
  *     OUT:
  *         None
  *
  * RETURNS:
  *     PR_ERR_NONE     Success
  *     PR_ERR_PORTID   If portId is not in the range
  *     PR_ERR_PHY_WR   If Phy write fails
  *     PR_ERR_PHY_RD   If Phy read fails
  *
  ***************************************************************************/

s32 ifx_phy_pause_tx(u8 chipId,u32 portId, BOOL Enable)
{
    u16  dwTmp=0;
    u8   phy_addr=0;
    u16  dwTmp4=0;
    u16  dwTmp9=0;
    u16  dwTmp_ge=0,dwTmp_fe=0;
    BOOL mode = 0;


#ifdef PARAMETER_CHECK

    if(portId > GE_PORT_1)
        return PR_ERR_PORTID;


    if(portId < GE_PORT_1)
    {
        if((portId & 7) != 0 )
            return PR_ERR_PORTID;
    }

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif


    ifx_get_portId_phyId( chipId, portId, &phy_addr );

   /* Read reg 0 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    if( ((dwTmp >> PHY_AUTONEGO_POS) & 0x01) == 0x0)
        return PR_ERR_PAUSE_AN_DIS; /* If Auto-negotiation is disabled */

   /* Read reg 1 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_STATUS, &dwTmp) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    if( ((dwTmp >> PHY_AUTONEGO_CP_POS) & 0x01) == 0x0)
        return PR_ERR_PAUSE_AN_NC;  /* If Auto negotiation is not completed */

   /* Read reg 9 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, 0x9, &dwTmp9) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

   /* Read reg 4 */
    if(ifx_phy_read(chipId,/* mdio_id,*/ phy_addr, PHY_ANAR, &dwTmp4) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;

    dwTmp_ge = ((dwTmp9 >> 8) & 0x3);
    dwTmp_fe = ((dwTmp4 >> 5) & 0x7F);

    if( ((dwTmp_ge >> 1) & 0x1) == 0x1)
        mode = 1;
    else
        if( (dwTmp_ge & 0x1) == 0x1)
            mode = 0;
        else
            if( ((dwTmp_fe >> 3) & 0x1) == 0x1)
                mode = 1;
            else
                if( ((dwTmp_fe >> 2) & 0x1) == 0x1)
                        mode = 0;
                else
                    if( ((dwTmp_fe >> 1) & 0x1) == 0x1)
                        mode = 1;
                    else
                        mode = 0;

    if(mode == 0)
        Enable = FALSE;


    if(Enable == TRUE)
    {
        dwTmp4 |= (0x1 << PHY_ASYM_PAUSE_POS);
    }
    else
    {
        dwTmp4 &= ~(0x1 << PHY_ASYM_PAUSE_POS);
    }


    if (ifx_phy_write(chipId, phy_addr, PHY_ANAR, dwTmp4) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s -  wrote 0x%04X to PHY_ANAR Register\n", __FUNCTION__, dwTmp4);
#endif

    if (ifx_phy_read(chipId, phy_addr, PHY_CTRL, &dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_RD;

    dwTmp |= (0x01 << PHY_RESTRT_AUTONEGO_POS) ;  /* Restart Auto-Nego bit */

    if (ifx_phy_write(chipId, phy_addr, PHY_CTRL, dwTmp) != PR_ERR_NONE)
        return PR_ERR_PHY_WR;

    if(mode == 0)
        return PR_ERR_PAUSE_HD;

    return PR_ERR_NONE;
}


/**
 * Set Port Ethernet Configuration.
 * Set Ethernet Configuration parameters for the passed portId.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port select(0(LAN)-1(PC))
 * \param   speed               0-10Mbit/s, 1-100Mbit/s, 2-1000Mbit/s
 * \param   duplex              half(0) or full duplex mode(1)
 * \param   autoneg             disable(0) or enable(1) autonegotiation
 * \param   int_ext_phy         use internal(0) or external(1) PHY
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_CHIP_VLAN_AWARE - VlanAwareFlag is invalid
 * \return  PR_ERR_CHIP_ID         - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_port_eth_conf( u8 chipId, u32 portId,
                           u32 speed, u32 duplex, BOOL autoneg,
                           u32 int_ext_phy)
{
    s32 retval;
    u16 temp_reg;
    u8 phy_addr;
    u32 tmp_reg_val;
    BOOL int_ext_current;
    BOOL mdioMode;
    BOOL autoNegMode;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, portId=%d, speed=%d, duplex=%d, autoneg=%d, int_ext_phy=%d)\n",
           __FUNCTION__, chipId, portId, speed, duplex, autoneg, int_ext_phy);
#endif
#ifdef PARAMETER_CHECK

    if(portId >= (IFX_TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID;

    if((duplex != IFX_HALF_DUPLEX) && (duplex != IFX_FULL_DUPLEX))
        return PR_ERR_ETH_DUPLEX_MODE;

    if(speed > IFX_SPEED_1000_MBIT)
        return PR_ERR_ETH_SPEED;

    if((autoneg != TRUE) && (autoneg != FALSE))
        return PR_ERR_ETH_ANEG;

    if(int_ext_phy > IFX_USE_EXTERNAL_PHY)
        return PR_ERR_ETH_INT_EXT_PHY;

#endif
    /*
     * internal PHY can be operated at 100 or 100MBit/s only
     */
    if( (int_ext_phy == IFX_USE_INTERNAL_PHY) &&
        (speed > IFX_SPEED_100_MBIT) )
    {
        printk("%s - Invalid speed+PHY combination passed \n", __FUNCTION__);
        return PR_ERR_ETH_SPEED;
    }

    /*
     * check if PHY change is desired by user by comparison of desired and
     * active PHY. In this case the active PHY has to be powered down before
     * the desired PHY can be activated.
     */
    if( (retval=ifx_get_MDIO_phy( chipId, portId, &int_ext_current )) != PR_ERR_NONE )
        return retval;

    if( int_ext_current != int_ext_phy )
    {
        printk("%s - Change to ", __FUNCTION__);
        if( int_ext_phy )
            printk("external PHY\n");
        else
            printk("internal PHY\n");

        if (ifx_phy_read( chipId, portId, PHY_CTRL, &temp_reg ) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;
        temp_reg |= (1<<PHY_POWER_DOWN_POS);
#ifdef INCAIP2_DEBUG_PHY
        printk("Power down active PHY\n");
#endif
        if (ifx_phy_write( chipId, portId, PHY_CTRL, temp_reg ) != PR_ERR_NONE)
            return PR_ERR_PHY_WR;
        /*
        * Configure MDIO_MOD_SEL to select desired PHY (internal/external) and
        * power up the desired PHY
        */
#ifdef INCAIP2_DEBUG_PHY
        printk("Switch to requested PHY\n");
#endif
        if( (retval=ifx_set_MDIO_phy( chipId, portId,  (BOOL) int_ext_phy )) != PR_ERR_NONE )
            return retval;

#ifdef INCAIP2_DEBUG_PHY
        printk("Power up requested PHY\n");
#endif
        if (ifx_phy_read( chipId, portId, PHY_CTRL, &temp_reg ) != PR_ERR_NONE)
            return PR_ERR_PHY_RD;
        temp_reg &= ~(1<<PHY_POWER_DOWN_POS);
        if (ifx_phy_write( chipId, portId, PHY_CTRL, temp_reg ) != PR_ERR_NONE)
            return PR_ERR_PHY_WR;

    }

    if( int_ext_phy == IFX_USE_INTERNAL_PHY)
    {
#ifdef INCAIP2_DEBUG_PHY
        printk("Configure internal PHY with passed settings\n");
#endif

        /* Configure passed port's MAC to operate in MII mode with passed speed */
        if( portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
        {
            tmp_reg_val=INCAIP2_SW_REG32( SWITCH_P0_GMAC_REG );
            INCAIP2_SW_REG32( SWITCH_P0_GMAC_REG ) = (tmp_reg_val & (~0x0000000F)) |
                                                     ((speed & 0x00000003) << 1) |
                                                     duplex << 3 | GMAC_MII;
        }
        else
        {
            tmp_reg_val= INCAIP2_SW_REG32( SWITCH_P1_GMAC_REG );
            INCAIP2_SW_REG32( SWITCH_P1_GMAC_REG ) = (tmp_reg_val & (~0x0000000F)) |
                                                     ((speed & 0x00000003) << 1) |
                                                     duplex << 3 | GMAC_MII;
        }

        if( (retval=ifx_set_port_FE_mode( chipId, portId,
                                          FE_PORT_SPEED, speed )) != PR_ERR_NONE )
        {
            printk("%s -> ifx_set_port_FE_mode() failed for FE_PORT_SPEED\n", __FUNCTION__);
            return retval;
        }

        if( (retval=ifx_set_port_FE_mode( chipId, portId,
                                          FE_PORT_DUPLEX_MODE, duplex )) != PR_ERR_NONE )
        {
            printk("%s -> ifx_set_port_FE_mode() failed for FE_PORT_DUPLEX\n", __FUNCTION__);
            return retval;
        }

        if( (retval=ifx_phy_speed_fe( chipId, portId, (u8) speed )) != PR_ERR_NONE )
            return retval;
    }
    else
    {
#ifdef INCAIP2_DEBUG_PHY
        printk("Configure external PHY with passed settings\n");
#endif

        /* Configure passed port's MAC to operate in RGMII mode with passed speed  */
        if( portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
       {
            tmp_reg_val=INCAIP2_SW_REG32( SWITCH_P0_GMAC_REG );
            INCAIP2_SW_REG32( SWITCH_P0_GMAC_REG ) = (tmp_reg_val & (~0x0000000F)) |
                                                     ((speed & 0x00000003) << 1) |
                                                     duplex << 3 | GMAC_RGMII;
        }
        else
        {
            tmp_reg_val= INCAIP2_SW_REG32( SWITCH_P1_GMAC_REG );
            INCAIP2_SW_REG32( SWITCH_P1_GMAC_REG ) = (tmp_reg_val & (~0x0000000F)) |
                                                     ((speed & 0x00000003) << 1) |
                                                     duplex << 3 | GMAC_RGMII;
        }

        /* Add delay to Rx and Tx Clock w.r.t. data */
#ifdef INCAIP2_DEBUG_PHY
        printk("Add delay to Rx and Tx Clock\n");
#endif
        ifx_get_portId_phyId( chipId, portId, &phy_addr);
        ifx_phy_read( chipId, phy_addr, 20, &temp_reg);
        ifx_phy_write( chipId, phy_addr, 20, temp_reg | 0x82);

        if( (retval=ifx_set_port_GE_mode( chipId, portId,
                                          GE_PORT_SPEED, speed )) != PR_ERR_NONE )
        {
            printk("%s -> ifx_set_port_GE_mode() failed for speed, returned %d \n",
                    __FUNCTION__, retval);
            return retval;
        }

        if( (retval=ifx_set_port_GE_mode( chipId, portId,
                                          GE_PORT_DUPLEX_MODE,
                                          duplex )) != PR_ERR_NONE )
        {
            printk("%s -> ifx_set_port_GE_mode() failed for duplex mode, returned %d \n",
                    __FUNCTION__, retval);
            return retval;
        }

        if( (retval=ifx_phy_speed_ge( chipId, portId, speed )) != PR_ERR_NONE )
        {
            printk("%s -> ifx_phy_speed_ge() failed for speed, returned %d \n",
                    __FUNCTION__, retval);
            return retval;
        }

    }

    /*
     * configure autonegotiation and duplex mode according to passed parameters
     */
    if( (retval= ifx_phy_set_duplex_mode( chipId, portId, (BOOL) duplex)) != PR_ERR_NONE )
    {
        printk("%s -> ifx_phy_set_duplex_mode() failed \n", __FUNCTION__);
        return retval;
    }

    ifx_get_MDIO_mode( chipId, portId, &mdioMode, &autoNegMode);


    if( (retval= ifx_set_port_autoneg( chipId, portId, autoneg)) != PR_ERR_NONE )
    {
        printk("%s -> ifx_set_port_autoneg() failed \n", __FUNCTION__);
        return retval;
    }

    ifx_set_MDIO_mode( chipId, portId, mdioMode, autoneg);

    return retval;

}


/**
 * Get Port Ethernet Configuration.
 * Get Ethernet Configuration parameters for the passed portId.
 *
 * \param   chipId                 Chip Id
 * \param   portId                 Port select(0(LAN)-1(PC))
 * \param   speed_ptr              0-10Mbit/s, 1-100Mbit/s, 2-1000Mbit/s
 * \param   duplex_ptr             half(0) or full duplex mode(1)
 * \param   autoneg_ptr            disable(0) or enable(1) autonegotiation
 * \param   int_ext_phy_ptr        use internal(0) or external(1) PHY
 * \param   link_status_ptr        Pointer to link status
 * \return  PR_ERR_NONE            - Success
 * \return  PR_ERR_CHIP_VLAN_AWARE - EnablePtr is NULL
 * \return  PR_ERR_NULL_PTR        - VlanAwareFlag is invalid
 * \return  PR_ERR_CHIP_ID         - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_eth_conf( u8 chipId, u32 portId,
                           u32* speed_ptr, u32* duplex_ptr, BOOL* autoneg_ptr,
                           u32* int_ext_phy_ptr, u32* link_status_ptr)
{
    s32 retval;
    BOOL int_ext_phy;
    BOOL autoneg;
    u32 duplex;
    u32 speed;
   u32 link_status;

#ifdef INCAIP2_DEBUG_PHY
    printk("%s( chipId=%d, portId=%d )\n",
           __FUNCTION__, chipId, portId );
#endif

#ifdef PARAMETER_CHECK

    if(portId >= (IFX_TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID;

    if( (speed_ptr == NULL) ||
        (duplex_ptr == NULL) ||
        (autoneg_ptr == NULL) ||
        (int_ext_phy_ptr == NULL) )
        return PR_ERR_NULL_PTR;

#endif

    /*
     * get phy usage (internal/external)
     */
    if( (retval=ifx_get_MDIO_phy( chipId, portId, &int_ext_phy )) != PR_ERR_NONE )
        return retval;
    *int_ext_phy_ptr = (u32)int_ext_phy;

    /*
     * get autonegotiation, duplex and speed information for passed port and
    * retrieved PHY usage
     */
    if( int_ext_phy == IFX_USE_INTERNAL_PHY )
    {
        if( (retval=ifx_get_port_FE_mode( chipId, portId,
                                          FE_PORT_SPEED, &speed )) != PR_ERR_NONE )
            return retval;

        if( (retval=ifx_get_port_FE_mode( chipId, portId,
                                          FE_PORT_DUPLEX_MODE, &duplex )) != PR_ERR_NONE )
            return retval;

        if( (retval=ifx_get_port_FE_mode( chipId, portId,
                                          FE_PORT_AUTONEGOTIATION, &autoneg )) != PR_ERR_NONE )
            return retval;

        if( (retval=ifx_get_port_FE_mode( chipId, portId,
                                          FE_PORT_LINK_OK, &link_status )) != PR_ERR_NONE )
            return retval;

    }
    else
    {
        if( (retval=ifx_get_port_GE_mode( chipId, portId,
                                          GE_PORT_SPEED, &speed )) != PR_ERR_NONE )
            return retval;

        if( (retval=ifx_get_port_GE_mode( chipId, portId,
                                          GE_PORT_DUPLEX_MODE, &duplex )) != PR_ERR_NONE )
            return retval;

        if( (retval=ifx_get_port_GE_mode( chipId, portId,
                                          GE_PORT_AUTONEGOTIATION, &autoneg )) != PR_ERR_NONE )
            return retval;

        if( (retval=ifx_get_port_GE_mode( chipId, portId,
                                          GE_PORT_LINK_OK, &link_status )) != PR_ERR_NONE )
            return retval;

    }
    *speed_ptr       = speed;
    *autoneg_ptr     = (BOOL) autoneg;
    *duplex_ptr      = duplex;
    *link_status_ptr = link_status;

    return retval;
}


