/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */
/** \addtogroup Obsolete Will be removed soon */

/*pr_port.c --  Port Configuration API */
#include <asm/types.h>
#include <asm/addrspace.h>
#include <asm/incaip2/ifx_types.h>
#include <asm/incaip2/incaip2.h>

#include <asm/incaip2/switch.h>
#include "struct.h"
#include "const.h"
#include "ports.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"
#include "ports_prv.h"
#ifdef _VIPER
#include "vdslapi.h"
#endif //_VIPER


/**
 * Set port locking bit
 * Set the Port Locking bit for each switch port. If the port is locked, no new 
 * address can be learned from this port.
 *
 * \param   chipId              Chip Id
 * \param   portId              PortId(range is 0-2)
 * \param   lock                1 - Association of port Id with MAC address is locked
  *                             0 - Not locked
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_PORTID       - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_LOCK_INVALID - If the Lock bit is invalid
 * \return  PR_ERR_TIME_OUT     - Time out 
 * \return  PR_ERR_NULL_PTR     - Base address is null pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_port_lock(u8 chipId,u32 portId, BOOL lock)
{
    u32 Temp=0;
    s32 i = 0;
    
/* Define a pointer to a structure & assign the base address of 
   CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;

/* Define a pointer to CPU access data register */
    u32* ptr_cpu_acs_data;


    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;


#ifdef PARAMETER_CHECK
    
    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;
        
    if((lock != 1) && (lock != 0))
        return PR_ERR_LOCK_INVALID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif


    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);

    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;


/* Configuration of CPU Access control register for reading */

/* Table_sel(20:18) :110 -Port Configuration Table,PORT_CONFG_TBL = 0x6 */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    
/* index(17:6):PortID */
    ptr_cpu_acs_ctrl -> Index = portId;
    
/* Read/Write(1):1-Read */
    ptr_cpu_acs_ctrl -> Read_write = READPR;

/* Request/Acknowledge(0): 1-Request */
    ptr_cpu_acs_ctrl -> Req_ack = REQ;


    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);


/* PortX_locked:2,PORTX_LOCKED_MASK = 0xFFFF FFFB*/

    Temp &= PORTX_LOCKED_MASK;
    Temp |= (lock << 2);

    
/* Configuration of CPU Access control register for writing */  
/* Read/Write(1):0-Write */
    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    
/* Request/Acknowledge(0): 1-Request */
    ptr_cpu_acs_ctrl -> Req_ack = REQ;
    
/* Added MACRO for indirect register write 7/7/2004 */
    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);


    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE ;
}

/**
 * Get port locking bit
 * Get the Port Locking bit for each switch port. If the port is locked, no new 
 * address can be learned from this port.
 *
 * \param   chipId              Chip Id
 * \param   portId              PortId(range is 0-2)
 * \param   lockPtr             Pointer to Port locking bit
 *                              1 - Association of port Id with MAC address is locked
 *                              0 - Not locked
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_PORTID       - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_TIME_OUT     - Time out 
 * \return  PR_ERR_NULL_PTR     - lockPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_lock(u8 chipId,u32 portId, BOOL* lockPtr)
{

    u32 Temp=0;
    s32 i = 0;

/* Define a pointer to a structure & assign the base address of 
   CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;

/* Define a pointer to CPU access data register */
    u32* ptr_cpu_acs_data;

    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK  
    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;
    if(lockPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif


    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);

/*  PR_Reg_Read(ptr_reg1,temp1); */
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;


/* Configuration of CPU Access control register for reading */

/* Table_sel(20:18) :110 -Port Configuration Table,PORT_CONFG_TBL = 0x6 */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    
/* index(17:6):PortID */
    ptr_cpu_acs_ctrl -> Index = portId;
    
/* Read/Write(1):1-Read */
    ptr_cpu_acs_ctrl -> Read_write = READPR;

/* Request/Acknowledge(0): 1-Request */
    ptr_cpu_acs_ctrl -> Req_ack = REQ;


    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);



    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

/* PortX_locked:2 bit position */

    *lockPtr = ((Temp >> 2) & 0x1) ? TRUE : FALSE;
    
    return PR_ERR_NONE ;

}

/**
 * Set port VLAN ID
 * Set the Port VLAN ID (PVID) for switch port.
 *
 * \param   chipId               Chip Id
 * \param   portId               PortId(range is 0-2)
 * \param   vlanId               VlanId(range is 0-4094)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_VLANID        - If the Vlan Id is not in the range of 0 to 4094
 * \return  PR_ERR_VLAN_NOTCONFG - Vlan is not configured
 * \return  PR_ERR_TIME_OUT      - Time out 
 * \return  PR_ERR_NULL_PTR      - Base address is null pointer
 * \return  PR_ERR_CHIP_ID       - If the chip Id invalid
 * \ingroup Internal
 */
s32  ifx_set_port_VLANId( u8 chipId, u32 portId, u32 vlanId )
{
    u32 Temp =0;
    s32  i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    
    if(vlanId > MAX_VLANID)
        return PR_ERR_VLANID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    
    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /* 
     * Update the VLAN ID bits of the port configuration entry 
     * for the passed port id 
     */
    ptr_cpu_acs_ctrl -> Table_sel  = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index      = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack    = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
        return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    Temp &= PORTX_VLANID_MASK;
    Temp = Temp | (vlanId << PORTX_VLANID_POS);

    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;
    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);
#if 0
    if(PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0) 
    {
        printk("%s - PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0\n", __FUNCTION__);
        ifx_sem_give(chipId,SEM_CPU);
        return PR_ERR_VLAN_NOTCONFG;                /* vlanid is not configured  */
    }
    else
    {
        ifx_sem_give(chipId,SEM_CPU);
        return PR_ERR_NONE ;
    }
#endif    
    return( ifx_sem_give(chipId,SEM_CPU));

}

/**
 * Get port VLAN ID
 * Get the Port VLAN ID (PVID) for the switch port.
 *
 * \param   chipId               Chip Id
 * \param   portId               PortId(range is 0-2)
 * \param   vlanIdPtr            Pointer to VlanId(range is 0-4094)
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_TIME_OUT      - Time out 
 * \return  PR_ERR_NULL_PTR      - vlanIdPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID       - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_VLANId(u8 chipId,u32 portId, u32* vlanIdPtr)
{

    u32 Temp =0;
    s32  i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID) 
        return PR_ERR_PORTID;

    if(vlanIdPtr == NULL) 
            return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    *vlanIdPtr = (Temp >> PORTX_VLANID_POS) & 0xFFF;

    return PR_ERR_NONE ;
}

/**
 * Set VLAN Ingress Tag
 * Set the VLAN Ingress Tag (Tagged packets only) for the switch port. If it is
 * enabled then untagged packets are dropped on that port. 
 *
 * \param   chipId                  Chip Id
 * \param   portId                  PortId(range is 0-2)
 * \param   IngressTag              FALSE - Accept tagged and untagged frames
 *                                  TRUE  - Accept tagged frames only
 * \return  PR_ERR_NONE             - Success 
 * \return  PR_ERR_PORTID           - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_VLAN_INGRESS_TAG - IngressTag is invalid
 * \return  PR_ERR_TIME_OUT         - Time out 
 * \return  PR_ERR_NULL_PTR         - Base address is null pointer
 * \return  PR_ERR_CHIP_ID          - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_port_ingress_VLAN_tag(u8 chipId,u32 portId, BOOL IngressTag)
{
    u32 Temp =0;
    s32  i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
        
    if((IngressTag != TRUE) && (IngressTag != FALSE))
        return PR_ERR_VLAN_INGRESS_TAG;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    Temp &= PORTX_INGRESS_CTRL;
    Temp |= IngressTag ? (1 << PORTX_INGRESS_POS) : 0;

    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    return( ifx_sem_give(chipId,SEM_CPU) );

}

/**
 * Get VLAN Ingress Tag
 * Get the VLAN Ingress Tag (Tagged packets only) for the switch port. If it is 
 * enabled then untagged packets are dropped on that port.
 *
 * \param   chipId                  Chip Id
 * \param   portId                  PortId(range is 0-2)
 * \param   IngressTagPtr           Pointer to Ingress Tag
 *                                    - FALSE - Accept tagged and untagged frames
 *                                    - TRUE  - Accept tagged frames only
 * \return  PR_ERR_NONE             - Success 
 * \return  PR_ERR_PORTID           - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_TIME_OUT         - Time out 
 * \return  PR_ERR_NULL_PTR         - IngressTagPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID          - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_ingress_VLAN_tag(u8 chipId,u32 portId, BOOL* IngressTagPtr)
{

    u32 Temp=0;
    s32  i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK  
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    if(IngressTagPtr ==NULL )
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;
    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;
    
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    Temp &= (~PORTX_INGRESS_CTRL);
    *IngressTagPtr = Temp != 0 ? TRUE : FALSE;
    
    return PR_ERR_NONE ;

}

/**
 * Set VLAN Ingress Filter
 * Sets the VLAN Ingress Filter (Symmetric vlan) flag for the switch port in the
 * Software table PR_INGRESS_FILTER. It will have effect only if the port 
 * state is Forwarding in STP mode.
 *
 * \param   chipId                     Chip Id
 * \param   portId                     PortId(range is 0-2)
 * \param   IngressFilter              TRUE  - Packet is dropped
 *                                     FALSE - Packet is handled 
 * \return  PR_ERR_NONE                - Success 
 * \return  PR_ERR_PORTID              - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_VLAN_INGRESS_FILTER - IngressFilter is invalid 
 * \return  PR_ERR_NULL_PTR            - Base address is null pointer
 * \return  PR_ERR_CHIP_ID             - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_port_ingress_VLAN_filter(u8 chipId,u32 portId, BOOL IngressFilter)
{

        s32 retflag = PR_ERR_NONE;
        
#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    
    if((IngressFilter != TRUE) && (IngressFilter != FALSE))
        return PR_ERR_VLAN_INGRESS_FILTER;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif


    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;

    PR_INGRESS_FILTER[chipId][portId] = IngressFilter;  
    if(PR_TABLE_PORTSTATUS[chipId][PR_Table_Portstatus(DEFAULT_VID,portId)].port_state == IFX_PORT_FORWARDING)
        ifx_set_SYM_VLAN(chipId,portId,IngressFilter);
        
    return ( ifx_sem_give(chipId,SEM_VLAN) );
}

/**
 * Get VLAN Ingress Filter
 * Get the VLAN Ingress Filter Flag for the switch port.
 *
 * \param   chipId            Chip Id
 * \param   portId            PortId(range is 0-2)
 * \param   IngressFilterPtr  Pointer to VLAN Ingress filter flag
 * \return  PR_ERR_NONE       - Success 
 * \return  PR_ERR_PORTID     - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR   - IngressFilterPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID    - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_ingress_VLAN_filter(u8 chipId,u32 portId, BOOL* IngressFilterPtr)
{
    
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    
    if(IngressFilterPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    if((retflag = ifx_sem_take(chipId,SEM_VLAN)) != PR_ERR_NONE)
        return retflag;

    *IngressFilterPtr = (PR_INGRESS_FILTER[chipId][portId] !=0) ? TRUE : FALSE;

    return( ifx_sem_give(chipId,SEM_VLAN) );
}

/**
 * Set default CoS
 * Set the default CoS(class of service) for the switch port.
 *
 * \param   chipId           Chip Id
 * \param   portId           PortId(range is 0-2)
 * \param   CoS              Default CoS(class of service) for port (range is 0 - 3)
 * \return  PR_ERR_NONE      - Success 
 * \return  PR_ERR_PORTID    - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_CoS       - If the CoS is not in the range 0 to 3
 * \return  PR_ERR_TIME_OUT  - Time out
 * \return  PR_ERR_NULL_PTR  - Base address is null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_port_CoS(u8 chipId,u32 portId, u32 CoS)
{
    u32 Temp =0 ;
    s32  i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;

    if(CoS > MAX_CoS)
        return PR_ERR_CoS;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /*
     * Read current port configuration via port configuration table access
     */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    /*
     * update default CoS for passed portId and 
     * write it to port configuration table
     */
    Temp &= PORTX_COS_MASK;
    Temp = Temp | (CoS << PORTX_COS_POS);

    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    return( ifx_sem_give(chipId,SEM_CPU) );
}

/**
 * Get default CoS
 * Get the default CoS(class of service) for the switch port.
 *
 * \param   chipId           Chip Id
 * \param   portId           PortId(range is 0-2)
 * \param   CoSPtr           Pointer to CoS(range is 0-3)
 * \return  PR_ERR_NONE      - Success 
 * \return  PR_ERR_PORTID    - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_TIME_OUT  - Time out
 * \return  PR_ERR_NULL_PTR  - CoSPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID   - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_CoS(u8 chipId,u32 portId, u32* CoSPtr)
{
    u32 Temp =0, i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID) 
        return PR_ERR_PORTID;
    if(CoSPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    *CoSPtr = (Temp >> PORTX_COS_POS) & 0x3;
    
    return PR_ERR_NONE ;

}



/**
 * Set Jumbo Enable flag
 * Set the Jumbo Enable flag for the switch port. If it is enabled then that 
 * port accepts Jumbo frames.
 *
 * \param   chipId           Chip Id
 * \param   portId           PortId(range is 0-2)
 * \param   JumboEnable      TRUE  - Jumbo frames are allowed on this port.
 *                           FALSE - Jumbo frames are not allowed on this port.
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_PORTID       - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_JUMBO_ENABLE - If the Jumbo Enable flag is invalid
 * \return  PR_ERR_NULL_PTR     - Base address is null pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_port_jumbo_enable(u8 chipId,u32 portId, BOOL JumboEnable)
{

    u32 Temp =0;
    u32 Mask;    /* mask to filter selected port */
    u32* ptr_jumbo_enable;
    s32 retflag = PR_ERR_NONE;
    
#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    
    if((JumboEnable != TRUE) && (JumboEnable != FALSE))
        return PR_ERR_JUMBO_ENABLE;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    /* 
     * read current jumbo enable configuration of ports 0...2 and 
     * update the configuration for the passed portid 
     */
    ptr_jumbo_enable = (u32*) (Chip_Base_Address[chipId] + JUMBO_ENABLE_REG);

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    PR_Reg_Read(ptr_jumbo_enable,Temp);

    Mask = (u32) (1<<portId);
    Temp &= ~Mask;
    Temp |= (JumboEnable << portId);   
    PR_Reg_Write(ptr_jumbo_enable,Temp);

    return( ifx_sem_give(chipId,SEM_GENRL) );
 
}

/**
 * Get Jumbo Enable flag
 * Get the Jumbo Enable flag for the switch port. 
 *
 * \param   chipId          Chip Id
 * \param   portId          PortId(range is 0-2)
 * \param   JumboEnablePtr  Pointer to Jumbo enable flag
 *                            - TRUE  - Jumbo frames are allowed on this port.
 *                            - FALSE - Jumbo frames are not allowed on this port.
 * \return  PR_ERR_NONE     - Success 
 * \return  PR_ERR_PORTID   - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR - JumboEnablePtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_jumbo_enable(u8 chipId,u32 portId, BOOL* JumboEnablePtr)
{
    u32 Temp =0;
    u32 Mask = 1;
    u32* ptr_jumbo_enable;
    
#ifdef PARAMETER_CHECK
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;

    if(JumboEnablePtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    /* 
     * read current jumbo enable configuration of ports 0...2 and 
     * return the jumbo enable information of the passed portid 
     */
    ptr_jumbo_enable = (u32*) (Chip_Base_Address[chipId] + JUMBO_ENABLE_REG);
    PR_Reg_Read(ptr_jumbo_enable,Temp);

    Mask <<= portId;
    Temp &= Mask;
    *JumboEnablePtr = (Temp !=0) ? TRUE : FALSE;

    return PR_ERR_NONE ;
    
}
#if 0
/******************************************************************************
  *
  * FUNCTION:    ifx_set_port_sVLAN
  *
  * DESCRIPTION: This function is used to set the Stacked VLAN configuration
  *              for the switch port. 
  *              
  * PARAMETERS:
  *     IN:
  *         chipId          Chip Id
  *         portId          PortId(range is 0-50)
  *         Svlan           Stacked VLAN configuration(range is 0-3)
  *                         00 - No change
  *                         01 - Replace VLANID of tag
  *                         10 - Replace Priority of tag
  *                         11 - Add tag
  *     OUT:
  *         None
  *                          
  * RETURNS:
  *     PR_ERR_NONE                 Success                         
  *     PR_ERR_PORTID               If the PortId is not in the range of 0 to 50
  *     PR_ERR_SVLAN                If the Stacked VLAN configuration is invalid
  *     PR_ERR_TIME_OUT             Time out
  *     PR_ERR_CHIP_ID              Invalid chip Id
  *     PR_ERR_NULL_PTR             Null pointer
  *     
  ******************************************************************************/

s32 ifx_set_port_sVLAN(u8 chipId,u32 portId, u32 Svlan)
{
    u32 Temp =0, i=0;

/* Define a pointer to a structure & assign the base address of 
    CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;

/* Define a pointer to CPU access data register */
    u32* ptr_cpu_acs_data;

    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;



#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;

    if(Svlan > 3)
        return PR_ERR_SVLAN;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif


    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);

    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    
/* Configuration of CPU Access control register for reading */

/* Table_sel(20:18) :110 -Port Configuration Table,PORT_CONFG_TBL = 0x6 */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    
/* index(17:6):PortID */
    ptr_cpu_acs_ctrl -> Index = portId;
    
/* Read/Write(1):1-Read */
    ptr_cpu_acs_ctrl -> Read_write = READPR;

/* Request/Acknowledge(0): 1-Request */
    ptr_cpu_acs_ctrl -> Req_ack = REQ;


    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);


/* Reset the old Svlan cfg,PORTX_SVLAN_MASK = 0xFFFF FFFC */
    Temp &= PORTX_SVLAN_MASK;
    
/* Set the new Svlan cfg */
    Temp = Temp | Svlan;

    
/* Configuration of CPU Access control register for writing */  
/* Read/Write(1):0-Write */
    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    
/* Request/Acknowledge(0): 1-Request */
    ptr_cpu_acs_ctrl -> Req_ack = REQ;
    
/* Added MACRO for indirect register write 7/7/2004 */
    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);


    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE ;
}


/******************************************************************************
  *
  * FUNCTION:    ifx_get_port_sVLAN
  *
  * DESCRIPTION: Get Stacked VLAN configuration for the switch port. 
  *              
  * PARAMETERS:
  *     IN:
  *         chipId          Chip Id
  *         portId          PortId(range is 0-50)
  *     OUT:
  *         SvlanPtr        Pointer to Stacked VLAN configuration(range is 0-3)
  *                          
  * RETURNS:
  *     PR_ERR_NONE         Success                         
  *     PR_ERR_PORTID       If the PortId is not in the range of 0 to 50
  *     PR_ERR_NULL_PTR     Null pointer
  *     PR_ERR_TIME_OUT     Time out
  *     PR_ERR_CHIP_ID      Invalid chip Id
  *     
  ******************************************************************************/

s32 ifx_get_port_sVLAN(u8 chipId,u32 portId, u32* SvlanPtr)
{
    u32 Temp =0;
    s32 i=0;

/* Define a pointer to a structure & assign the base address of 
    CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;

/* Define a pointer to CPU access data register */
    u32* ptr_cpu_acs_data;

    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;


#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID) 
        return PR_ERR_PORTID;
    if(SvlanPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif


    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;


/* Configuration of CPU Access control register for reading Qpriority*/
/* Table_sel(20:18) :110 -Port Configuration Table,PORT_CONFG_TBL = 0x6 */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    
/* index(17:6):PortID */
    ptr_cpu_acs_ctrl -> Index = portId;
    
/* Read/Write(1):1-Read */
    ptr_cpu_acs_ctrl -> Read_write = READPR;

/* Request/Acknowledge(0): 1-Request */
    ptr_cpu_acs_ctrl -> Req_ack = REQ;


    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);



    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

/* PortX_Svlan_cfg(1:0) */
    *SvlanPtr = Temp & 0x3;
    
    return PR_ERR_NONE ;

}
#endif



/************************************ Internal APIs  ************************/

/**
 * Set VLAN symmetric flag.
 * Set the Symmetric vlan flag for the switch port. 
 *
 * \param   chipId                     Chip Id
 * \param   portId                     PortId (range is 0-2)
 * \param   symflag                    Symmetric flag value (TRUE or FALSE)
 * \return  PR_ERR_NONE                - Success 
 * \return  PR_ERR_NULL_PTR            - Base address is null pointer 
 * \return  PR_ERR_CHIP_ID             - If the chip Id invalid
 * \return  PR_ERR_PORTID              - Invalid port ID
 * \return  PR_ERR_VLAN_INGRESS_FILTER - Invalid symflag value
 * \ingroup Internal
 */
s32 ifx_set_SYM_VLAN(u8 chipId,u32 portId, BOOL symflag)
{
/* Checked - OK */
    u32 Temp,Vlan_Mask = 1;
    u32 shift_count =0;
    u32* ptr_symm_vlan;
    
#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    
    if((symflag != TRUE) && (symflag != FALSE))
        return PR_ERR_VLAN_INGRESS_FILTER;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;


#endif

    /*
     * 
     * 
     */
    ptr_symm_vlan = (u32*) SYMM_VLAN_REG;   
    
    PR_Reg_Read(ptr_symm_vlan,Temp);
    
    shift_count = portId & 0x3;
    Vlan_Mask <<= shift_count;
    Vlan_Mask = ~ Vlan_Mask;
    
    Temp &= Vlan_Mask;
    
    Temp |= (symflag << shift_count);
    
    PR_Reg_Write(ptr_symm_vlan,Temp);
    return PR_ERR_NONE ;    
    
}



/******************************************************************************
  *
  * FUNCTION:    ifx_get_SYM_VLAN
  *
  * DESCRIPTION: Get the Symmetric vlan Flag for the switch port. 
  *              
  * PARAMETERS:
  *     IN:
  *         chipId              Chip Id
  *         portId              PortId(range is 0-2)
  *     OUT:
  *         IngressFilterPtr    Pointer to VLAN Ingress filter flag  
  *
  * RETURNS:
  *     PR_ERR_NONE             Success                 
  *     PR_ERR_PORTID           If the Port Id is not in the range of 0 to 50 
  *     PR_ERR_NULL_PTR         Null pointer
  *     PR_ERR_CHIP_ID          Invalid chip Id
  *
  ******************************************************************************/
/**
 * Get VLAN symmetric flag.
 * Get the Symmetric vlan flag for the switch port. 
 *
 * \param   chipId                     Chip Id
 * \param   portId                     PortId (range is 0-2)
 * \param   symflagPtr                 Pointer to Symmetric flag value 
 *                                     (TRUE or FALSE)
 * \return  PR_ERR_NONE                - Success 
 * \return  PR_ERR_NULL_PTR            - Base address is null pointer 
 * \return  PR_ERR_CHIP_ID             - Invalid Chip Id
 * \return  PR_ERR_PORTID              - Invalid port ID
 * \return  PR_ERR_VLAN_INGRESS_FILTER - Invalid symflag value
 * \ingroup Internal
 */
s32 ifx_get_SYM_VLAN( u8 chipId, u32 portId, BOOL* symflagPtr)
{
    
    u32 Temp =0;
    u32 Vlan_Mask = 1;
    u32 shift_count =0;
    u32* ptr_symm_vlan;
    
#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID)
        return PR_ERR_PORTID;
    
    if(symflagPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_symm_vlan = (u32*) (Chip_Base_Address[chipId] + SYMM_VLAN_REG);
    
    PR_Reg_Read(ptr_symm_vlan,Temp);

    /*
     * filter symm_vlan bit for passed port id
     */
    shift_count = portId;
    Vlan_Mask <<= shift_count;
    Temp &= Vlan_Mask;

    *symflagPtr = (Temp !=0) ? TRUE : FALSE;

    return PR_ERR_NONE ;
    
}

#if 0  
/**
 * Init TBI
 * Init the tbi for each group of port 
 *
 * \param   chipId          Chip Id
 * \param   portGroupId     group Id(range is 0-7)
 * \return  PR_ERR_NONE     - Success 
 * \return  PR_ERR_PORTID   - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Obsolete
 * \internal
 */
s32 ifx_port_group_tbi_init(u8 chipId,u8 portGroupId)
{
    volatile u32 temp = 0;
    u32 * ptr_Gmac_Reg = NULL;
    u32 * ptr_PmaLpm_Reg = NULL;
    u32 * ptr_GmacCtrl_Reg = NULL;

#ifdef PARAMETER_CHECK
    
    if(portGroupId > MAX_PORTID) 
        return PR_ERR_PORTID;
    
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_Gmac_Reg = (u32*)(Chip_Base_Address[chipId] + portGroupId * TOTAL_PORT_REGS + GMAC_REG );

    ptr_PmaLpm_Reg = (u32*)(Chip_Base_Address[chipId] + portGroupId * TOTAL_PORT_REGS + PMA_LPM_REG );

    ptr_GmacCtrl_Reg = (u32*)(Chip_Base_Address[chipId] + portGroupId * TOTAL_PORT_REGS + GMAC_CTRL_REG );

    temp = *ptr_GmacCtrl_Reg;
                
    if((temp & PMA_MODE) == PMA_MODE)
    {
        temp = *ptr_PmaLpm_Reg;
        if ((temp & PMA_SYNC_LINK_STATUS) == PMA_SYNC_LINK_STATUS)
        {
            temp = *ptr_Gmac_Reg;
            temp = temp | GMAC_LINK_OK_MASK;
            *ptr_Gmac_Reg = temp;   
        }
        else
        {   
            temp = *ptr_Gmac_Reg;
            temp = temp & ~(GMAC_LINK_OK_MASK);
            *ptr_Gmac_Reg = temp;
        }
    }
    
    return PR_ERR_NONE;

}
#endif   

/**
 * Get port default configuration
 * Get the default configuration of the selected switch port by accessing the 
 * port configuration table.
 *
 * \param   chipId               Chip Id
 * \param   portId               PortId(range is 0-2)
 * \param   portConfPtr          Pointer to Port Configuration word of selected port
 * \return  PR_ERR_NONE          - Success 
 * \return  PR_ERR_PORTID        - If the PortId is not in the range of 0 to 2
 * \return  PR_ERR_TIME_OUT      - Time out 
 * \return  PR_ERR_NULL_PTR      - vlanIdPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID       - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_port_default_configuration( u8 chipId, u32 portId, u32* portConfPtr)
{
    u32 Temp =0;
    s32  i=0;
    /* pointer to CPU access control register */
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;  /* pointer to CPU access data register */
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK
    
    if(portId > MAX_PORTID) 
        return PR_ERR_PORTID;

    if(portConfPtr == NULL) 
            return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /* 
     * Configuration CPU Access control register for reading from 
     * port configuration table 
     */
    ptr_cpu_acs_ctrl -> Table_sel = PORT_CONFG_TBL;
    ptr_cpu_acs_ctrl -> Index = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    *portConfPtr = Temp;

    return PR_ERR_NONE ;
}



