/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */

#include <linux/kernel.h>
#include <linux/string.h>
#include <asm/types.h>
#include <asm/addrspace.h>
#include <asm/incaip2/ifx_types.h>
#include <asm/incaip2/incaip2.h>
#include "struct.h"
#include "const.h"
#include "flow_engine.h"
#include "flow_engine_prv.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"

#undef DIRECT_RULE_CONFIGURATION

#ifdef DIRECT_RULE_CONFIGURATION
u32 cam_bank[512][24];
#endif

void generate_rule_mem( u16 start_data_val, u16 end_data_val, u32 rule_no, u32 key_no );



/* Implemented for software read from Ikey selection register , (..this register write only) */
/* static u32   Ikey_Sel_Reg[8] ;*/

/**
 * Set port keys.
 * Set the 8 IKeys for one of 3 ports. IKEY value and the corresponding
 * desription is given below.
 *
 * - 0  Corresponding 16 bit Layer 2 field.
 * - 1  Port Rule match result
 * - 8  IP DA [ 0 - 15 ]
 * - 9  IP DA [ 16 - 32 ]
 * - 10 IP SA [ 0 - 15 ]
 * - 11 IP SA [ 16 - 32 ]
 * - 12 TCP / UDP Destination Port
 * - 13 TCP / UDP Source Port
 * - 14 IP Protocol ID / TCP Flag
 * - 15 Diff Serv / SPID
 *
 * \param   chipId          Chip Id
 * \param   portId          Port (0 - 2)
 * \param   IKey            8 IKeys, where each IKey is of 4 bits
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_PORTID   - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_port_key( u8 chipId, u32 portId, u32 IKey )
{
    ikey_sel_reg_t*   ptr_ikey_sel;
    s32 retflag = PR_ERR_NONE;
/*
    printk("%s( chipId=%d, portId=%d, IKey=0x%08X )\n",
           __FUNCTION__, chipId, portId, IKey);
*/
#ifdef PARAMETER_CHECK
    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_ikey_sel = (ikey_sel_reg_t*) SWITCH_P0_PORT_IKEY_SEL;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_ikey_sel = (ikey_sel_reg_t*) SWITCH_P1_PORT_IKEY_SEL;
        }
        else
        {
            ptr_ikey_sel = (ikey_sel_reg_t*) SWITCH_P2_PORT_IKEY_SEL;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    PR_Reg_Write(ptr_ikey_sel,IKey);
/*
    printk("%s - wrote value 0x%08X to ptr_ikey_sel at 0x%08X\n",
           __FUNCTION__, IKey, (u32)ptr_ikey_sel);
*/
    Ikey_Sel_Reg[chipId][portId] = IKey;

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}

/**
 * Get port keys.
 * Get the 8 IKeys for one of 3 port. IKEY value and the corresponding
 * desription is given below.
 *
 * - 0  Corresponding 16 bit Layer 2 field.
 * - 1  Port Rule match result
 * - 8  IP DA [ 0 - 15 ]
 * - 9  IP DA [ 16 - 32 ]
 * - 10 IP SA [ 0 - 15 ]
 * - 11 IP SA [ 16 - 32 ]
 * - 12 TCP / UDP Destination Port
 * - 13 TCP / UDP Source Port
 * - 14 IP Protocol ID / TCP Flag
 * - 15 Diff Serv / SPID
 *
 * \param   chipId           Chip Id
 * \param   portId           Port  ID (0 - 2)
 * \param   IKeyPtr          Pointer to 8 IKeys, where each IKey is of 4 bits
 * \return  PR_ERR_NONE      - Success
 * \return  PR_ERR_PORTID    - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR  - IKeyPtr or base address is null pointer
 * \return  PR_ERR_CHIP_ID   - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_port_key(u8 chipId,u32 portId, u32 *IKeyPtr)
{
    ikey_sel_reg_t*   ptr_ikey_sel;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;
    if(IKeyPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_ikey_sel = (ikey_sel_reg_t*) SWITCH_P0_PORT_IKEY_SEL;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_ikey_sel = (ikey_sel_reg_t*) SWITCH_P1_PORT_IKEY_SEL;
        }
        else
        {
            ptr_ikey_sel = (ikey_sel_reg_t*) SWITCH_P2_PORT_IKEY_SEL;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    /* Register implemented in software keeps the same copy as hardware register */
    *IKeyPtr = Ikey_Sel_Reg[chipId][portId];

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}

/**
 * Set rule for flow ID.
 * Set the Rule for a flow Id. This function generates 512
 * Rule bits for the given flow Id using Key Vector and Mask Vector by
 * calling function ifx_generate_512_rule_bits.
 * It stores the generated 512 bits in the Rule Memory
 * by calling function ifx_set_rule_mem.
 *
 * \param   chipId          Chip Id
 * \param   flowId          flow Id (0 - 23)
 * \param   IKeyData        Key Vector (0 - 127)
 * \param   IKeyMask        Mask Vector (0 - 127)
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_FLOWID   - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_NULL_PTR - IKeyData or IKeyMask is null pointer
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_flow_pattern( u8 chipId, u32 flowId, u32* IKeyData, u32* IKeyMask )
{

#ifndef DIRECT_RULE_CONFIGURATION
    u8 RuleVector_512[RULE_VECTOR_SIZE] ;
    u32 rule32_val=0;
    u32 rule_pos=0;  /* selects the bit in the rule to be accessed */
    u32 section=0;
    u32 bank=0;
    u32 address=0;
    u32 k=0;
    s32 retflag = PR_ERR_NONE;

/*
   printk("%s( chipId=%d, flowId=%d, IKeyData=0x%08X 0x%08X 0x%08X 0x%08X\n",
           __FUNCTION__, chipId, flowId,
           *IKeyData, *(IKeyData+1), *(IKeyData+2), *(IKeyData+3) );
   printk("      IKeyMask=0x%08X 0x%08X 0x%08X 0x%08X )\n",
                *IKeyMask, *(IKeyMask+1), *(IKeyMask+2), *(IKeyMask+3) );
*/
#ifdef PARAMETER_CHECK

    if(flowId >= TOTAL_PORT_RULE_ID)
        return PR_ERR_FLOWID;

    if(IKeyData == NULL)
        return PR_ERR_NULL_PTR;

    if(IKeyMask == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif

    /* Initialize rule vector with zeros*/
    memset( &RuleVector_512, 0, RULE_VECTOR_SIZE*sizeof(u8));

    /*  generate 512 rule bits from Rule Vector and Mask Vector
        and store it in RuleVector_512  */
    ifx_generate_512_rule_bits(IKeyData, IKeyMask, RuleVector_512);

    /*
     * calculate the section and the bit position in the section
     * to which the rule belongs to
     */
    section = (flowId / RULES_PER_SECTION ) % RULE_MEM_NUM;
    rule_pos = flowId % RULES_PER_SECTION ;

/*    printk("section=%d, rule_pos=%d\n", section, rule_pos);  */

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    /*
     * Access 4 banks of 128 entries each to write new 512 bit column
     * to rule memory
     */
    for(k=0; k<RULE_MEM_NUM; k++)
    {
        bank = k;
        for(address=0; address<ADDRESS_PER_RULEMEM; address++)
        {
            if( (retflag=ifx_get_rule_mem(chipId, bank, address,section,&rule32_val))
                                != PR_ERR_NONE )
                return retflag;
/*
           printk("k=%d, bank=%3d, section=%3d, address=%5d, rule32_val=0x%08X\n",
                   k, bank, section, address, rule32_val);
           printk("RuleVector_512[%d]=%d\n", address+k*ADDRESS_PER_RULEMEM, RuleVector_512[address+k*ADDRESS_PER_RULEMEM]);
*/
           rule32_val &= ~(1 << rule_pos);
           rule32_val |= (RuleVector_512[address  + k * ADDRESS_PER_RULEMEM]) << rule_pos ;
/*           printk("-> rule32_val=0x%08X\n", rule32_val);  */
            if( (retflag=ifx_set_rule_mem(chipId, bank, address,section,rule32_val))
                                != PR_ERR_NONE )
                return retflag;
       } /* for all 128 addresses */
    } /* for all 512 addresses, 4 banks */

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;

#else  /* DIRECT_RULE_CONFIGURATION */
    u32 rule32_val;
    u16 start_data_val=(*IKeyData&0xFFFF);
    u16 end_data_val=(*IKeyData&0xFFFF0000)>>16;
    u32 rule_no=flowId;
    u32 key_no=(*IKeyMask&0x7);
    s32 address;


    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;
    volatile u32 Temp=0;
    u32 i=0, j=0;
    u32* ptr_reg;

    /* compare for MAC source address 00:00:00:00:00:04 at LAN port */
    generate_rule_mem( 0x0000, 0xFFFF, 0, 0 );
    generate_rule_mem( 0x0000, 0xFFFF, 0, 1 );
    generate_rule_mem( 0x0000, 0xFFFF, 0, 2 );
    generate_rule_mem( 0x0004, 0x0004, 0, 3 );
    generate_rule_mem( 0x0000, 0x0000, 0, 4 );
    generate_rule_mem( 0x0000, 0x0000, 0, 5 );
    generate_rule_mem( 0x0000, 0xFFFF, 0, 6 );
    generate_rule_mem( 0x0000, 0xFFFF, 0, 7 );


   /* compare for MAC source address 00:00:00:00:00:03 at PC port*/
    generate_rule_mem( 0x0000, 0xFFFF, 1, 0 );
    generate_rule_mem( 0x0000, 0xFFFF, 1, 1 );
    generate_rule_mem( 0x0000, 0xFFFF, 1, 2 );
    generate_rule_mem( 0x0003, 0x0003, 1, 3 );
    generate_rule_mem( 0x0000, 0x0000, 1, 4 );
    generate_rule_mem( 0x0000, 0x0000, 1, 5 );
    generate_rule_mem( 0x0000, 0xFFFF, 1, 6 );
    generate_rule_mem( 0x0000, 0xFFFF, 1, 7 );

    print_cam_bank();


    for(address=0; address<512; address++)
    {
        rule32_val=0;
        for(j=0; j<24; j++)
            rule32_val |= ((cam_bank[address][j])<<j);

        ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
        ptr_reg = (u32 *) (Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
        ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)(&Temp);

        /*  Place the control in cpu access control register */

        ptr_cpu_acs_ctrl->Req_ack = REQ;  /* Request */
        ptr_cpu_acs_ctrl->Read_write = WRITEPR;  /* write */
        ptr_cpu_acs_ctrl->Table_specific = 0;  /* TB counter */
        ptr_cpu_acs_ctrl->Index = address ;  /* one of 512 Flowt Ids */
        ptr_cpu_acs_ctrl->Table_sel = RULE_TB_SEL  ;/* Rule Engine memory */

        PR_Indirect_Reg_Write_Sem(ptr_reg,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,Temp,rule32_val,SEM_CPU);
    }
    print_rule_mem();
    return PR_ERR_NONE;
#endif

}

/**
 * Get rule for flow ID.
 * Get the Key Vector and Mask Vector for a given flow Id.
 * This function reads the 512 rule bits from rule memory by calling
 * function ifx_get_rule_mem, then generates key vectors and mask
 * vectors by calling function ifx_generate_rule_mask_vectors.
 * Depending on mask vector there can be more than one key vectors.
 * Hence this function returns the key Vector for which mask is zero.
 * For example: Just consider the following 4 bits.
 * Mask bits = 1100
 * Rule bits = 0010
 * Possible Rule Vectors :
 *
 * mask bits   rule bits
 * 0000        0010
 * 0100        0110
 * 1000        1010
 * 1100        1110
 *
 * Rule bits considered = 0010 (2)
 *
 * \param   chipId          Chip Id
 * \param   flowId          flow Id (0 - 23)
 * \param   IKeyDataPtr     Pointer to Key Vector (0 - 127)
 * \param   IKeyMaskPtr     Pointer to Mask Vector (0 - 127)
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_FLOWID   - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_NULL_PTR - IKeyDataPtr or IKeyMaskPtr is null pointer
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_flow_pattern( u8 chipId, u32 flowId, u32* IKeyDataPtr, u32* IKeyMaskPtr)
{
    u8 RuleVector_512[RULE_VECTOR_SIZE] ;
    u32 rule32_val=0;
    u32 rule_pos=0;
    u32 section=0;
    u32 bank=0;
    u32 address=0;
    u32 k=0;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(flowId > MAX_FLOW_ID)
        return PR_ERR_FLOWID;
    if(IKeyDataPtr== NULL)
        return PR_ERR_NULL_PTR;
    if(IKeyMaskPtr== NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif
   /* Initialize rule vector with zeros*/
    memset( &RuleVector_512, 0, RULE_VECTOR_SIZE*sizeof(u8));


    /* calculate the section and the bit position within the section
       to which the rule belongs */
    section = (flowId / RULES_PER_SECTION ) % RULE_MEM_NUM;
    rule_pos = flowId % RULES_PER_SECTION ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    for(k=0; k<RULE_MEM_NUM; k++)
    {
        bank = k;  /* (flowId / ADDRESS_PER_RULEMEM) + k * RULE_MEM_NUM;  */

        for(address=0; address < ADDRESS_PER_RULEMEM; address++)
        {
            /*
             * Read from Rule Memory coresponding to a flow Id and
             * store the rule in RuleVector_512
             */
            ifx_get_rule_mem(chipId, bank,  address,section,&rule32_val) ;
            RuleVector_512[address + k * ADDRESS_PER_RULEMEM] =  (rule32_val >> rule_pos) & 1;
        } /* for all 128 addresses */
    } /* for all 512 addresses, 4 banks */

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
        return retflag;

    /* Generate IKeyData(0:127) and IKeyMask(0:127) from RuleVector_512 */
    ifx_generate_rule_mask_vectors(IKeyDataPtr, IKeyMaskPtr, RuleVector_512) ;

    return PR_ERR_NONE;
}



/**
 * Delete action parameters for flow ID.
 * Delete one or more actions specified by feature for the flow specified by flow Id.
 *
 * \param   chipId          Chip Id
 * \param   flowId          flow Id (0 - 23)
 * \param   feature         One of the below features
 *                          - REDIRECTION: disable redirection action for this rule.
 *                          - MIRROR: disable mirroring for this rule.
 *                          - COS_ASSIGNMENT: disable CoS assignment for this rule.
 *                          - VLAN_ASSIGNMENT: disable vlan assignment action
 *                                             for this rule
 *                          - TKN_BKT_ID: disable token bucket action for this rule
 *                          - DELETE_ACTION_ENTRY: Delete action from table
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_FLOWID      - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_FLOW_ACTION - If the feature is not in the range of 1- 6
 * \return  PR_ERR_TIME_OUT    - Time out
 * \return  PR_ERR_NULL_PTR    - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_delete_flow_action_param( u8 chipId, u32 flowId, u32 feature )
{
    volatile cpu_acs_ctrl_reg_t*    ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*             ptr_cpu_acs_data;
    u32                             Temp=0;
    u32                             i=0;
    volatile u32                    temp1 = 0;
    u32*                            ptr_reg1;
    s32                             retflag = PR_ERR_NONE;
    u32                             disable_bit=0x00000000;

#ifdef PARAMETER_CHECK

    if(flowId > MAX_FLOW_ID)
        return PR_ERR_FLOWID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *)&temp1;

    switch(feature)
    {
        case REDIRECTION:
            disable_bit |= AT_REDIRECT_EN_DIS_MASK;
            break;

        case MIRROR:
            disable_bit |= AT_MIRROR_EN_DIS_MASK;
            break;

        case COS_ASSIGNMENT:
             disable_bit |= AT_COS_EN_DIS_MASK;
            break;

        case VLAN_ASSIGNMENT:
            disable_bit |= AT_VLAN_EN_DIS_MASK;
            break;

        case TKN_BKT_ID:
            disable_bit |= AT_TOKEN_EN_DIS_MASK;
            break;

        case DELETE_ACTION_ENTRY:
            /* remove entry from action table */
            return(ifx_set_flow_action_all(0,flowId,0));

        default:
            return PR_ERR_FLOW_ACTION ;
    }
    /*  Place the command in cpu access control register */
    ptr_cpu_acs_ctrl->Req_ack    = REQ;  /* Request */
    ptr_cpu_acs_ctrl->Read_write = READPR;  /* read */
    ptr_cpu_acs_ctrl->Index      = flowId;  /* one of 512 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel  = ACTION_TB_SEL ; /* Flow Action Table */

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
        return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    /* set disable mask to disable desired feature */
    Temp &= ~disable_bit;

    /*  Place the command in cpu access control register */
    ptr_cpu_acs_ctrl->Req_ack    = REQ;  /* Request */
    ptr_cpu_acs_ctrl->Read_write = WRITEPR;  /* write */
    ptr_cpu_acs_ctrl->Index      = flowId;  /* one of 512 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel  = ACTION_TB_SEL ; /* Flow Action Table */

/*
    printk("%s - write value 0x%08X to cpu_acs_data register\n",
           __FUNCTION__,Temp);
*/

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);
/*
    printk("%s - wrote value 0x%08X to cpu_acs_ctrl register\n",
           __FUNCTION__,*ptr_reg1);
*/
    retflag = ifx_sem_give(chipId,SEM_CPU);

    return retflag;
}






/**
 * Set action parameters for flow ID.
 * Set a particular action specified by feature for the flow specified by flow Id.
 *
 * \param   chipId          Chip Id
 * \param   flowId          flow Id (0 - 23)
 * \param   feature         One of the below features
 *                          - REDIRECT_MCAST:
 *                              - bit 0 = LAN port
 *                              - bit 1 = PC Port
 *                              - bit 2 = CPU Port
 *                          - MIRROR:   1 = Mirror Enable
 *                                      0 = Mirror Disable
 *                          - COS_ASSIGNMENT: 0-3 CoS Value to be assigned to this rule.
 *                          - VLAN_ASSIGNMENT: 0-4094 Vlan value to be associated
 *                                             for this rule
 *                          - TKN_BKT_ID: 0-23 Token bucket id for this rule.
 * \param   flowAction      Action Value for the selected feature
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_FLOWID      - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_FLOW_ACTION - If the feature is not in the range of 1- 6
 * \return  PR_ERR_TIME_OUT    - Time out
 * \return  PR_ERR_NULL_PTR    - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_flow_action_param(u8 chipId,u32 flowId, u32 feature, u32 flowAction)
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;
    u32 Temp=0;
    u32 mask=0;
    u32 pos=0, i=0;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;
    u32 enable_bit;

#ifdef PARAMETER_CHECK

    if(flowId > MAX_FLOW_ID )
        return PR_ERR_FLOWID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *)&temp1;

    switch(feature)
    {
        case REDIRECTION:
            pos = AT_REDIRECT_POS ;
            mask = AT_REDIRECT_MASK_W ;
            enable_bit = AT_REDIRECT_EN_DIS_MASK;
            break;

        case MIRROR:
            pos = AT_MIRROR_POS;
            mask = AT_MIRROR_MASK_W;
            enable_bit = AT_MIRROR_EN_DIS_MASK;
            break;

        case COS_ASSIGNMENT:
            pos = AT_COS_POS;
            mask = AT_COS_MASK_W;
            enable_bit = AT_COS_EN_DIS_MASK;
            break;

        case VLAN_ASSIGNMENT:
            pos = AT_VLAN_POS;
            mask =AT_VLAN_MASK_W;
            enable_bit = AT_VLAN_EN_DIS_MASK;
            break;

        case TKN_BKT_ID:
            pos = AT_TKN_BKT_ID_POS;
            mask = AT_TKN_BKT_ID_MASK_W;
            enable_bit = AT_TOKEN_EN_DIS_MASK;
            break;

        default:
            return PR_ERR_FLOW_ACTION ;
    }

    /*
     * configure cpu access control and data registers for action table read access
     */
    ptr_cpu_acs_ctrl->Req_ack    = REQ;     /* Request */
    ptr_cpu_acs_ctrl->Read_write = READPR;  /* read */
    ptr_cpu_acs_ctrl->Index      = flowId;  /* one of 24 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel  = ACTION_TB_SEL ; /* Flow Action Table */

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    /*
     * set new action and configure cpu access control and data registers
     * for action table write access
     */
    Temp &= mask;
    Temp |= (flowAction << pos) | enable_bit;

    ptr_cpu_acs_ctrl->Req_ack    = REQ;      /* Request */
    ptr_cpu_acs_ctrl->Read_write = WRITEPR;  /* write */
    ptr_cpu_acs_ctrl->Index      = flowId;   /* one of 24 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel  = ACTION_TB_SEL ; /* Flow Action Table */
/*
    printk("%s - write value 0x%08X to cpu_acs_data register\n",
           __FUNCTION__,Temp);
*/
    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);
/*
    printk("%s - wrote value 0x%08X to cpu_acs_ctrl register\n",
           __FUNCTION__,*ptr_reg1);
*/
    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}

/**
 * Get action for flow ID.
 * Get a particular action for the flow specified by flow Id.
 *
 * \param   chipId          Chip Id
 * \param   flowId          flow Id (0 - 23).
 * \param   feature         One of the below features
 *                          - REDIRECT_MCAST:
 *                              - bit 0 = CPU port
 *                              - bit 1 = PC Port
 *                              - bit 2 = LAN Port
 *                          - MIRROR:   1 = Mirror Enable
 *                                      0 = Mirror Disable
 *                          - COS_ASSIGNMENT: 0-3 CoS Value to be assigned to this rule.
 *                          - VLAN_ASSIGNMENT: 0-4094 Vlan value to be associated
 *                                             for this rule
 *                          - TKN_BKT_ID: 0-62 Token bucket id for this rule.
 *                                        if this 63, token bucket Id is flow Id
 * \param   flowActionPtr   Pointer to Action Value and activation status
 *                          for the selected feature. Bit 31 holds the activation
 *                          state of the selected feature.
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_FLOWID      - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_FLOW_ACTION - If the feature is not in the range of 1- 6
 * \return  PR_ERR_TIME_OUT    - Time out
 * \return  PR_ERR_NULL_PTR    - flowActionPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_flow_action_param( u8 chipId, u32 flowId, u32 feature,
                               u32* flowActionPtr)
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*            ptr_cpu_acs_data ;
    u32                            Temp=0;
    u32                            mask=0, i=0;
    u32                            pos=0;
    volatile u32                   temp1 =0;
    u32*                           ptr_reg1;
    s32                            retflag = PR_ERR_NONE;
    u32                            enable_bit;
    u32                            enable_bit_pos;

#ifdef PARAMETER_CHECK

    if(flowId > MAX_FLOW_ID)
        return PR_ERR_FLOWID;

    if(flowActionPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *)&temp1;

    switch(feature)
    {
        case REDIRECTION:
            pos = AT_REDIRECT_POS ;
            mask = AT_REDIRECT_MASK_R ;
            enable_bit = AT_REDIRECT_EN_DIS_MASK;
            enable_bit_pos = AT_REDIRECT_EN_DIS_POS;
            break;

        case MIRROR:
            pos = AT_MIRROR_POS;
            mask = AT_MIRROR_MASK_R;
            enable_bit = AT_MIRROR_EN_DIS_MASK;
            enable_bit_pos = AT_MIRROR_EN_DIS_POS;
            break;

        case COS_ASSIGNMENT:
            pos = AT_COS_POS;
            mask = AT_COS_MASK_R;
            enable_bit = AT_COS_EN_DIS_MASK;
            enable_bit_pos = AT_COS_EN_DIS_POS;
            break;

        case VLAN_ASSIGNMENT:
            pos = AT_VLAN_POS;
            mask =AT_VLAN_MASK_R;
            enable_bit = AT_VLAN_EN_DIS_MASK;
            enable_bit_pos = AT_VLAN_EN_DIS_POS;
            break;

        case TKN_BKT_ID:
            pos = AT_TKN_BKT_ID_POS;
            mask = AT_TKN_BKT_ID_MASK_R;
            enable_bit = AT_TOKEN_EN_DIS_MASK;
            enable_bit_pos = AT_TOKEN_EN_DIS_POS;
            break;

        default:
            return PR_ERR_FLOW_ACTION ;
    }

    /*
     * configure cpu access control and data registers for action table read access
     */
    ptr_cpu_acs_ctrl->Req_ack    = REQ;        /* Request */
    ptr_cpu_acs_ctrl->Read_write = READPR;     /* Read */
    ptr_cpu_acs_ctrl->Index      = flowId;     /* one of 24 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel  = ACTION_TB_SEL ; /* Flow Action Table */

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
        return retflag;
/*
    printk("%s - read 0x%08X from CPU_ACS_DATA_REG\n", __FUNCTION__, Temp);
*/
    /* extract mode bits and enable bit for passed feature */
    *flowActionPtr = (Temp >> pos) & mask;
    *flowActionPtr |=(((Temp & enable_bit) << (31-enable_bit_pos)) );

    return PR_ERR_NONE;
}

#if 0
/**
 * Enable flow action.
 * Enable or disable the 6 actions for a particular flow set. Each Flow Set
 * contains 24 Rules.
 *
 * \param   chipId              Chip Id
 * \param   flowsetId           flow set Id (0 -3)
 * \param   flowsetSelection    6 bit Flow action value for the specified flow set
 *                                - bit 5 : MIB Counting
 *                                - bit 4 : Bandwidth Policing
 *                                - bit 3 : CoS Assignement
 *                                - bit 2 : VLAN Assignment
 *                                - bit 1 : Redirection
 *                                - bit 0 : Mirror
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_RULESET_ID  - If the flowSetId is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR    - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Obsolete
 */
s32 ifx_set_flow_action( u8 chipId, u32 flowsetId, u32 flowsetSelection )
{
    u32 Temp =0;
    s32  i=0;
    volatile cpu_acs_ctrl_reg_t*  ptr_cpu_acs_ctrl;
    u32* ptr_cpu_acs_data;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

    printk("%s( chipId=%d, flowsetId=%d, flowsetSelection=0x%08X )\n",
           __FUNCTION__, chipId, flowsetId, flowsetSelection);

#ifdef PARAMETER_CHECK

    if(flowsetId > RULE_SET3)
        return PR_ERR_RULESET_ID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_cpu_acs_data = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)&temp1;

    /*
     *
     *
     */
    ptr_cpu_acs_ctrl -> Table_sel  = RULE_TBL_VAL;
    ptr_cpu_acs_ctrl -> Index      = portId;
    ptr_cpu_acs_ctrl -> Read_write = READPR;
    ptr_cpu_acs_ctrl -> Req_ack    = REQ;

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
        return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    Temp &= PORTX_VLANID_MASK;
    Temp = Temp | (vlanId << PORTX_VLANID_POS);

    ptr_cpu_acs_ctrl -> Read_write = WRITEPR;
    ptr_cpu_acs_ctrl -> Req_ack = REQ;
    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);
#if 0
    if(PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0)
    {
        printk("%s - PR_VLAN_VID_ASC_TBL[chipId][vlanId] == 0\n", __FUNCTION__);
        ifx_sem_give(chipId,SEM_CPU);
        return PR_ERR_VLAN_NOTCONFG;                /* vlanid is not configured  */
    }
    else
    {
        ifx_sem_give(chipId,SEM_CPU);
        return PR_ERR_NONE ;
    }
#endif
    return( ifx_sem_give(chipId,SEM_CPU));



#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + RULE_SEL_REG);

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    PR_Reg_Read(ptr_reg1,temp1);
    ptr_rule_sel = (rule_sel_reg_t*)&temp1;

    switch( flowsetId )
    {
        case RULE_SET0:
            ptr_rule_sel->Rule_set0 = flowsetSelection;
            break;
        case RULE_SET1:
            ptr_rule_sel->Rule_set1 = flowsetSelection;
            break;
        case RULE_SET2:
            ptr_rule_sel->Rule_set2 = flowsetSelection;
            break;
        case RULE_SET3:
            ptr_rule_sel->Rule_set3 = flowsetSelection;
            break;

        default:
            return PR_ERR_RULESET_ID;
    }

    PR_Reg_Write(ptr_reg1,temp1);


    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}
#endif

#if 0
/**
 * Get flow action.
 * Get the 6 actions for a particular flow set.
 *
 * \param   chipId              Chip Id
 * \param   flowsetId           flow set Id (0 -3)
 * \param   flowsetSelectionPtr Pointer to 6 bit Flow action value for the
 *                              specified flow set.
 *                              6 bit Flow action value for the specified flow set
 *                                - bit 5 : MIB Counting
 *                                - bit 4 : Bandwidth Policing
 *                                - bit 3 : CoS Assignement
 *                                - bit 2 : VLAN Assignment
 *                                - bit 1 : Redirection
 *                                - bit 0 : Mirror
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_RULESET_ID  - If the flowSetId is not in the range of 0 to 3
 * \return  PR_ERR_NULL_PTR    - flowsetSelectionPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Obsolete
 */
s32 ifx_get_flow_action(u8 chipId,u32 flowsetId, u32* flowsetSelectionPtr)
{


    rule_sel_reg_t*  ptr_rule_sel;

    u32 temp1=0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK

    if(flowsetSelectionPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + RULE_SEL_REG);

    PR_Reg_Read(ptr_reg1,temp1);
    ptr_rule_sel = (rule_sel_reg_t*)&temp1;


        switch( flowsetId )
        {
            case RULE_SET0:
                *flowsetSelectionPtr = ptr_rule_sel->Rule_set0 ;
                break;
            case RULE_SET1:
                *flowsetSelectionPtr = ptr_rule_sel->Rule_set1 ;
                break;
            case RULE_SET2:
                *flowsetSelectionPtr = ptr_rule_sel->Rule_set2 ;
                break;
            case RULE_SET3:
                *flowsetSelectionPtr = ptr_rule_sel->Rule_set3 ;
                break;
            default:
                return PR_ERR_RULESET_ID ;
        }

        return PR_ERR_NONE;
    printk("%s: RULE_SEL_REGISTER not defined for INCAIP2 switch !!!\n", __FUNCTION__);
    return -1;

}
#endif


/**
 * Set rule for port ID.
 * Set the rule associated with a port ID. The Parser extracts 128 bit from
 * the packet pointed by 8 offsets. Each offset extracts 16 bit info from parser
 * These extracted bits are masked as specified by port mask, unmasked bits are
 * compared with port rule. The result of the comparision results in 8 bits.
 * (one bit for each 16 bit info) these are sign extended to 16 bits. The
 * resultant 128 bits can be used as an index to Rule memory.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   portRulePtr         Pointer to 128 bit Port Rule
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR     - portRulePtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_port_rule(u8 chipId,u32 portId, u32 *portRulePtr)
{

    port_rule_regs_t*    ptr_port_rule ;
    int i=0;
    u32 Temp=0;
    s32 retflag = PR_ERR_NONE;


#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(portRulePtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_port_rule = (port_rule_regs_t*)  SWITCH_P0_PORT_RULE0_REG;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_port_rule = (port_rule_regs_t*)  SWITCH_P1_PORT_RULE0_REG;
        }
        else
        {
            ptr_port_rule = (port_rule_regs_t*)  SWITCH_P2_PORT_RULE0_REG;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;


    for(i=0; i<PORT_RULE_REGS_NUM; i++)
    {
        Temp = *(portRulePtr + i);
        PR_Reg_Write((ptr_port_rule + i),Temp);
    }

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}

/**
 * Get rule for port ID.
 * Get the rule associated with a port ID. Parser extracts 128 bit from
 * the packet pointed by 8 offsets. Each offset extracts 16 bit info from parser
 * These extracted bits are masked as specified by port mask, unmasked bits are
 * compared with port rule. The result of the comparision results in 8 bits.
 * (one bit for each 16 bit info) these are sign extended to 16 bits. The
 * resultant 128 bits can be used as an index to Rule memory.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   portRulePtr         Pointer to 128 bit Port Rule
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR     - portRulePtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_port_rule(u8 chipId,u32 portId, u32 *portRulePtr)
{

    port_rule_regs_t*    ptr_port_rule ;
    int i=0;
    u32 Temp=0;
    s32 retflag = PR_ERR_NONE;


#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(portRulePtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_port_rule = (port_rule_regs_t*)  SWITCH_P0_PORT_RULE0_REG;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_port_rule = (port_rule_regs_t*)  SWITCH_P1_PORT_RULE0_REG;
        }
        else
        {
            ptr_port_rule = (port_rule_regs_t*)  SWITCH_P2_PORT_RULE0_REG;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    for(i=0; i<PORT_RULE_REGS_NUM; i++)
    {
        PR_Reg_Read((ptr_port_rule + i),Temp);
        *(portRulePtr + i) = Temp;
    }

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;

}

/**
 * Set rule mask for port ID.
 * Set the mask associated with a port ID. Parser extracts 128 bit from
 * the packet pointed by 8 offsets. Each offset extracts 16 bit info from parser
 * These extracted bits are masked as specified by port mask, unmasked bits are
 * compared with port rule. The result of the comparision results in 8 bits.
 * (one bit for each 16 bit info) these are sign extended to 16 bits. The
 * resultant 128 bits can be used as an index to Rule memory.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   portMaskPtr         Pointer to 128 bit Port mask
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR     - portMaskPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_port_mask( u8 chipId, u32 portId, u32 *portMaskPtr)
{

    port_mask_regs_t*    ptr_port_mask ;
    int i=0;
    u32 Temp=0;
    s32 retflag = PR_ERR_NONE;
/*
   printk("%s( chipId=%d, portId=%d, portMaskPtr=0x%08X 0x%08X 0x%08X 0x%08X)\n",
           __FUNCTION__, chipId, portId, *portMaskPtr,
           *(portMaskPtr+1), *(portMaskPtr+2), *(portMaskPtr+3) );
*/
#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(portMaskPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_port_mask = (port_mask_regs_t*)  SWITCH_P0_PORT_MASK0_REG;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_port_mask = (port_mask_regs_t*)  SWITCH_P1_PORT_MASK0_REG;
        }
        else
        {
            ptr_port_mask = (port_mask_regs_t*)  SWITCH_P2_PORT_MASK0_REG;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    for(i=0; i<PORT_MASK_REGS_NUM; i++)
    {
        Temp = *(portMaskPtr + i);
        PR_Reg_Write((ptr_port_mask + i),Temp);
/*
        printk("%s - wrote 0x%08X to address 0x%08X\n",
               __FUNCTION__, Temp, (u32)ptr_port_mask);
*/
    }

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}

/**
 * Get rule mask for ID.
 * Get the mask associated with a port ID. Parser extracts 128 bit from
 * the packet pointed by 8 offsets. Each offset extracts 16 bit info from parser
 * These extracted bits are masked as specified by port mask, unmasked bits are
 * compared with port rule. The result of the comparision results in 8 bits.
 * (one bit for each 16 bit info) these are sign extended to 16 bits. The
 * resultant 128 bits can be used as an index to Rule memory.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   portMaskPtr         Pointer to 128 bit Port mask
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR     - portMaskPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_port_mask( u8 chipId ,u32 portId, u32 *portMaskPtr)
{

    port_mask_regs_t*    ptr_port_mask ;
    int i=0;
    u32 Temp=0;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(portMaskPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_port_mask = (port_mask_regs_t*)  SWITCH_P0_PORT_MASK0_REG;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_port_mask = (port_mask_regs_t*)  SWITCH_P1_PORT_MASK0_REG;
        }
        else
        {
            ptr_port_mask = (port_mask_regs_t*)  SWITCH_P2_PORT_MASK0_REG;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    for(i=0; i<PORT_MASK_REGS_NUM; i++)
    {
        PR_Reg_Read((ptr_port_mask + i),Temp);
        *(portMaskPtr + i) = Temp;
    }

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
}

/**
 * Set offset for port ID.
 * Set the offset associated with a port ID. Parser extracts 128 bit from
 * the packet pointed by 8 offsets. Each offset extracts 16 bit info from parser
 * These extracted bits are masked as specified by port mask, unmasked bits are
 * compared with port rule. The result of the comparision results in 8 bits.
 * (one bit for each 16 bit info) these are sign extended to 16 bits. The
 * resultant 128 bits can be used as an index to Rule memory.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   portOffsetPtr       Pointer to 64 bit Port Offset
 *                              It contains Offsets from Offset 7 to Offset 0
 *                              Each offset has the following fields
 *                                - Bit 7 :
 *                                    - 1 = Layer 3 Extraction
 *                                    - 0 = Layer 2 Extraction
 *                                - Bit 6 - Bit 0 :
 *                                    - 7 bit offset into L3 / l2 fields. (in
 *                                      bytes up to max 128 bytes)
 *                              Note: Layer 3 start is zero by default.
 *                                    It can be changed to more than zero.
 *                                    If it is changed to more than zero, then
 *                                    offset should be subtracted from
 *                                    Layer 3 start value .
 *                                    (see ifx_set_Rx_config )
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR     - portOffsetPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_port_offset( u8 chipId, u32 portId, u32* portOffsetPtr )
{

    port_offset_regs_t*    ptr_port_offset ;
    int i=0;
    u32 Temp=0;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(portOffsetPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_port_offset = (port_offset_regs_t*)  SWITCH_P0_OFFSET0_REG;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_port_offset = (port_offset_regs_t*)  SWITCH_P1_OFFSET0_REG;
        }
        else
        {
            ptr_port_offset = (port_offset_regs_t*)  SWITCH_P2_OFFSET0_REG;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    for(i=0; i<PORT_OFFSET_REGS_NUM; i++)
    {
        Temp = *(portOffsetPtr + i);
        PR_Reg_Write((ptr_port_offset + i),Temp);
    }

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;


    return PR_ERR_NONE;
}

/**
 * Get offset for port ID.
 * Get the offset associated with a port group ID.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   portOffsetPtr       Pointer to 64 bit Port Offset
 *                              It contains Offsets from Offset 7 to Offset 0
 *                              portOffsetPtr[0]=offs3 offs2 offs1 offs0
 *                              portOffsetPtr[1]=offs7 offs6 offs5 offs4
 *                              Each offset has the following fields
 *                                - Bit 7 :
 *                                    - 1 = Layer 3 Extraction
 *                                    - 0 = Layer 2 Extraction
 *                                - Bit 6 - Bit 0 :
 *                                    - 7 bit offset into L3 / l2 fields. (in
 *                                      bytes up to max 128 bytes)
 *                              Note: Layer 3 start is by default is zero.
 *                                    It can be changed to more than zero.
 *                                    If it is changed to more than zero, then
 *                                    offset should be subtracted from
 *                                    Layer 3 start value .
 *                                    (see ifx_set_Rx_config )
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR     - portOffsetPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_port_offset( u8 chipId, u32 portId, u32 *portOffsetPtr )
{

    port_offset_regs_t*    ptr_port_offset ;
    int i=0;
    u32 Temp=0;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId > (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(portOffsetPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif
    if(portId == INCAIP2_SWITCH_PORT_SELECT_LAN)
    {
        ptr_port_offset = (port_offset_regs_t*)  SWITCH_P0_OFFSET0_REG;
    }
    else
        if(portId == INCAIP2_SWITCH_PORT_SELECT_PC)
        {
            ptr_port_offset = (port_offset_regs_t*)  SWITCH_P1_OFFSET0_REG;
        }
        else
        {
            ptr_port_offset = (port_offset_regs_t*)  SWITCH_P2_OFFSET0_REG;
        }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;
    /* configure OFFSET0 and OFFSET1 registers with passed offset values */
    for(i=0; i<PORT_OFFSET_REGS_NUM; i++)
    {
        PR_Reg_Read((ptr_port_offset + i),Temp);
        *(portOffsetPtr + i) = Temp;
    }

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}

/**
 * Set RX configuration.
 * Set the RX configuration.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   L3Program           If set, the L3_start and L3_start_vld bits of
 *                              this register are used by the RX parser else
 *                              the RX parser parses the packet to determine
 *                              L3 start.
 * \param   L3StartVld          Layer3 start valid
 * \param   L3Start             Actual layer2 start
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - Invalid Port Id
 * \return  PR_ERR_NULL_PTR     - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_Rx_config( u8 chipId, u32 portId, BOOL L3Program,
                       BOOL L3StartVld, u32 L3Start)
{
    rx_config_reg_t*    ptr_rx_config;
    u32                 temp1=0;
    u32*                ptr_reg1=NULL;
    s32                 retflag = PR_ERR_NONE;
/*
    printk("%s( portId=%d, L3Program=%d, L3StartVld=%d, L3Start=0x%08X ) \n",
        __FUNCTION__, portId, L3Program, L3StartVld, L3Start);
*/
#ifdef PARAMETER_CHECK

    if(portId >= (TOTAL_PORTS - 1))
        return PR_ERR_PORTID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if( (L3Program != TRUE) &&  (L3Program != FALSE) )
        return PR_ERR_L3_PROG;

    if( (L3StartVld != TRUE) &&  (L3StartVld != FALSE) )
        return PR_ERR_L3_START;

    if( L3Start > L3_MAX_VAL)
        return PR_ERR_L3_START;

#endif

    switch(portId)
    {
        case INCAIP2_SWITCH_PORT_SELECT_LAN:
            ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SWITCH_P0_RX_CONF);
            break;
        case INCAIP2_SWITCH_PORT_SELECT_PC:
            ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SWITCH_P1_RX_CONF);
            break;
        case INCAIP2_SWITCH_PORT_SELECT_CPU:
            ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SWITCH_P2_RX_CONF);
            break;
    }

    if((retflag = ifx_sem_take(chipId,SEM_GENRL)) != PR_ERR_NONE)
            return retflag;

    PR_Reg_Read(ptr_reg1,temp1);
    ptr_rx_config = (rx_config_reg_t*)&temp1;

    ptr_rx_config->L3Program  = L3Program ;
    ptr_rx_config->L3StartVld = L3StartVld ;
    ptr_rx_config->L3Start    = L3Start;
/*
    printk("%s - write 0x%08X to SWITCH_RX_CONF at address 0x%08X \n",
        __FUNCTION__, temp1, (u32)ptr_reg1);
*/
    PR_Reg_Write(ptr_reg1,temp1);

    if((retflag = ifx_sem_give(chipId,SEM_GENRL)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;

}

/**
 * Get RX configuration.
 * Get the RX configuration.
 *
 * \param   chipId              Chip Id
 * \param   portId              Port Id (0 - 2)
 * \param   L3ProgramPtr        Pointer to Layer3 mode
 * \param   L3StartVldPtr       Pointer to Layer3 start valid
 * \param   L3StartPtr          Pointer to actual layer2 start
 * \return  PR_ERR_NONE         - Success
 * \return  PR_ERR_PORTID       - If the portId is not in the range of 0-2
 * \return  PR_ERR_NULL_PTR     - L3ProgramPtr, L3StartVldPtr, L3StartPtr or
 *                                base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_Rx_config( u8 chipId, u32 portId, BOOL* L3ProgramPtr,
                       BOOL* L3StartVldPtr, u32* L3StartPtr)
{
    rx_config_reg_t*    ptr_rx_config;
    u32                 temp1=0;
    u32*                ptr_reg1 = NULL;

/*    printk("%s( portId=%d ) \n", __FUNCTION__, portId);  */

#ifdef PARAMETER_CHECK

    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;

    if(L3ProgramPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(L3StartVldPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(L3StartPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    switch(portId)
    {
        case INCAIP2_SWITCH_PORT_SELECT_LAN:
            ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SWITCH_P0_RX_CONF);
            break;
        case INCAIP2_SWITCH_PORT_SELECT_PC:
            ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SWITCH_P1_RX_CONF);
            break;
        case INCAIP2_SWITCH_PORT_SELECT_CPU:
            ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SWITCH_P2_RX_CONF);
            break;
    }

    PR_Reg_Read(ptr_reg1,temp1);
    ptr_rx_config = (rx_config_reg_t*)&temp1;
/*
    printk("%s - read 0x%08X from SWITCH_RX_CONF at address 0x%08X \n",
        __FUNCTION__, temp1, ptr_reg1);
*/
    *L3ProgramPtr = ptr_rx_config->L3Program ;
    *L3StartVldPtr = ptr_rx_config->L3StartVld ;
    *L3StartPtr = ptr_rx_config->L3Start ;

    return PR_ERR_NONE;
}

/********************************************* Internal API Definition ***********************************************/

#ifdef DIRECT_RULE_CONFIGURATION
void print_cam_bank( void )
{
    s32 i=0;
    s32 j=0;

    for(i=0; i< 512; i++)
    {
        printk("cam_bank[%d] = ", i);
        for(j=0; j<24; j++)
            printk("%1d ",cam_bank[i][j]);
        printk("\n");
    }


}

void print_rule_mem( void )
{
    s32 i, j, address;
    u32 rule32_val;

    for(address=0; address<512; address++)
    {
        ifx_get_rule_mem(0,  0,  address, 0, (u32*) &rule32_val);
        printk("0x%08X\n",rule32_val);

    }
}

void generate_rule_mem( u16 start_data_val, u16 end_data_val, u32 rule_no, u32 key_no )
{
    u32 start_nibble, end_nibble;
    s32 i, j;
    static u32 first_call=1;

    if(first_call)
    {
        for( i=0; i<512; i++ )
        {
            for( j=0; j< 24; j++ )
                cam_bank[i][j] = 0;
        }
        first_call=0;
    }

    start_nibble = (start_data_val&0xF) + key_no * 64;
    end_nibble   = (end_data_val&0xF) + key_no * 64;

    for( i=start_nibble; i<=end_nibble; i++)
    {
        cam_bank[i][rule_no] = 1;
    }

    start_nibble = ((start_data_val>>4)&0xF) + 16 + key_no * 64;
    end_nibble   = ((end_data_val>>4)&0xF)  +16 + key_no * 64;

    for( i=start_nibble; i<=end_nibble; i++ )
    {
        cam_bank[i][rule_no] = 1;
    }

    start_nibble = ((start_data_val>>8)&0xF) + 32 + key_no * 64;
    end_nibble   = ((end_data_val>>8)&0xF)  + 32 + key_no * 64;

    for( i=start_nibble; i<=end_nibble; i++ )
    {
        cam_bank[i][rule_no] = 1;
    }

    start_nibble = ((start_data_val>>12)&0xF) + 48 + key_no * 64;
    end_nibble   = ((end_data_val>>12)&0xF)  + 48 + key_no * 64;

    for( i=start_nibble; i<=end_nibble; i++ )
    {
        cam_bank[i][rule_no] = 1;
    }
}
#endif

/**
 * Generate 512 bits from the 128 bit IkeyMask and the 128 bit IkeyData.
 * The resulting vector will be used for the generation of the 512
 * table entries for the rule memory. *
 *
 * The algorithm is as follows.
 * 1.For each bank (4 such abnks to which rule corresponds to )
 *     2. For each successive 16 addresses (8 such addresses in a bank )
 *         extract 4 rule bits & 4 mask bits
 *         write one in the position that corresponds to (rule | mask)
 *     2. end of for loop 2
 * 1.end of for loop 1
 *
 * Example:
 * Rule bits = 0010
 * Mask bits = 1100
 * Write 1 in only the locations 2,6,10,14 among the corresponding 16 locations.
 * Write zero in all other 12 locations.
 *
 * \param   IKeyData        Pointer to 8 IKeys, where each IKey is of 4 bits
 *                          IKeyData[0] = Ikey7...Ikey0
 *                          IKeyData[1] = Ikey15...Ikey8
 *                          IKeyData[2] = Ikey23...Ikey16
 *                          IKeyData[3] = Ikey31...Ikey24
 * \param   IKeyMask        Pointer to 8 IKeyMasks, where each mask is of 4 bits
 *                          IKeyMask[0] = mask7...mask0
 *                          IKeyMask[1] = mask15...mask8
 *                          IKeyMask[2] = mask23...mask16
 *                          IKeyMask[3] = mask31...mask24
 * \param   RuleVector_512  Pointer to rule memory
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_PORTID   - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32  ifx_generate_512_rule_bits(u32* IKeyData, u32* IKeyMask, u8* RuleVector_512)
{
    u32 mask_4=0; /*  4 bit mask (for 16 addresses) */
    u32 rule_4=0; /*  4 bit rule (for 16 addresses) */
    int i=0;
    int j=0;
    int k=0;
/*
    printk("%s - IKeyData= 0x%08X 0x%08X 0x%08X 0x%08X,\n",
           __FUNCTION__, *IKeyData, *(IKeyData+1), *(IKeyData+2),*(IKeyData+3) );
    printk("     IKeyMask=0x%08X 0x%08X 0x%08X 0x%08X\n",
           *IKeyMask, *(IKeyMask+1), *(IKeyMask+2),*(IKeyMask+3));
*/
#ifdef PARAMETER_CHECK

    if(IKeyData== NULL)
        return PR_ERR_NULL_PTR;
    if(IKeyMask== NULL)
        return PR_ERR_NULL_PTR;
    if(RuleVector_512 == NULL)
        return PR_ERR_NULL_PTR;
#endif
    for(k=0; k<RULE_MEM_NUM; k++)   /* for each bank */
    {
        for(i=0; i<2*BITS_PER_IKEY; i++)
        {
            rule_4 = (IKeyData[k] >> (i*4)) & 0xF ; /* for each Ikey (Ikey0 - Ikey7) */
            mask_4 = (IKeyMask[k] >> (i*4)) & 0xF ; /* for each Mask corresponding to Ikey (Ikey0 - Ikey7) */
            for(j=0; j<16; j++)         /* for each Ikey(j) */
            {
                if((j | mask_4) == (rule_4 | mask_4))
                    RuleVector_512[j + ADDRESS_PER_IKEY*i + ADDRESS_PER_RULEMEM*k] = 1 ;
                else
                    RuleVector_512[j + ADDRESS_PER_IKEY*i + ADDRESS_PER_RULEMEM*k] = 0 ;
            }/* for all 16 addresses */
        } /* for all 128 addresses 1 bank */
    } /* for all 512 addresses, 4 banks */
/*
    for(j=0;j<512;j++)
        printk("RuleVector_512[%d]=0x%02X\n",j, RuleVector_512[j]);
*/
    return PR_ERR_NONE;
}


/**
 * Set rule memory.
 * This function writes 24 rule bits (incl. in rule32_val) to
 * the rule memory table at position flowid.
 *
 * \param   chipId          Chip Id
 * \param   bank            bank select: 0...3
 * \param   address         table index select: 0...127
 * \param   section         rule section select
 *                          0: rules 31...0
 *                          1: rules 63...32
 *                          2: rules 95...64
 *                          3: rules 127...96
 * \param   rule32_val      24 bit rule value
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_FLOWID   - Flow Id is invalid
 * \return  PR_ERR_RULE_MEM_ID - Rule Section is invalid
 * \return  PR_ERR_NULL_PTR - rule32_valPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_rule_mem(u8 chipId, u32 bank,  u32 address, u32 section, u32 rule32_val)
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;
    volatile u32 Temp=0;
    u32 i=0;
    u32* ptr_reg ;
/*    u32 bank=0;  */
/*
    printk("%s( address=%d, section=%d, rule32_val=0x%08X )\n",
           __FUNCTION__, address, section, rule32_val );
*/
#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if(section > MAX_RULE_MEM_ID)
        return PR_ERR_RULE_MEM_ID ;

    if(address > MAX_ADDRESS_PER_RULEMEM )
        return PR_ERR_RULE_MEM_ADDRESS;

#endif

    ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg = (u32 *) (Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)(&Temp);

    /*
     * configure cpu access and control register
     * for rule engine memory table write access
     */
    ptr_cpu_acs_ctrl->Req_ack        = REQ;      /* Request */
    ptr_cpu_acs_ctrl->Read_write     = WRITEPR;  /* write */
    ptr_cpu_acs_ctrl->Table_specific = section;  /* section select */
    ptr_cpu_acs_ctrl->Index          = (bank  << 7) | address ;  /* one of 24 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel      = RULE_TB_SEL  ; /* Rule Engine memory */

    PR_Indirect_Reg_Write_Sem(ptr_reg,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,Temp,rule32_val,SEM_CPU);
    return PR_ERR_NONE ;
}

/**
 * Get rule memory.
 * This function retrieves 24 rule bits from the rule memory table
 * at position flowid.
 *
 * \param   chipId          Chip Id
 * \param   bank            bank select: 0...3
 * \param   address         table index select: 0...127
 * \param   section         rule section select
 *                          0: rules 31...0
 *                          1: rules 63...32
 *                          2: rules 95...64
 *                          3: rules 127...96
 * \param   rule32_valPtr      - pointer to 24 bit rule value
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_FLOWID      - Flow Id is invalid
 * \return  PR_ERR_RULE_MEM_ID - Rule Section is invalid
 * \return  PR_ERR_NULL_PTR    - rule32_valPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_get_rule_mem(u8 chipId, u32 bank, u32 address, u32 section, u32* rule32_valPtr)
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;
    volatile u32 Temp=0;
    u32 temp1 = 0;
    u32 i=0;
    u32* ptr_reg ;

#ifdef PARAMETER_CHECK
    if(rule32_valPtr== NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if(section > MAX_RULE_MEM_ID)
        return PR_ERR_RULE_MEM_ID ;

    if(address > MAX_ADDRESS_PER_RULEMEM )
        return PR_ERR_RULE_MEM_ADDRESS;
#endif

    ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg = (u32*) (Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t*)(&Temp);

    /*
     * configure cpu access and control register
     * for rule engine memory table read access
     */
    ptr_cpu_acs_ctrl->Req_ack        = REQ;  /* Request */
    ptr_cpu_acs_ctrl->Read_write     = READPR;  /* write */
    ptr_cpu_acs_ctrl->Table_specific = section;  /* TB counter */
    ptr_cpu_acs_ctrl->Index          = (bank  << 7) | address ;  /* one of 24 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel      = RULE_TB_SEL  ;/* Rule Engine memory */

    PR_Indirect_Reg_Read_Sem(ptr_reg,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,Temp,temp1,SEM_CPU);

    *rule32_valPtr = temp1;

    return PR_ERR_NONE ;

}

/**
 * Generate IKeyData and IkeyMask from rule vector of a flow id.
 * IKeyData bits having relative mask bits of IKeyMask as 1 are masked to zero.
 * Depending on mask vector there can be more than one rule vectors.
 * Hence this function returns the Rule Vector for which mask is zero.
 * For example:
 * Just consider the following 4 bits.
 * Mask bits = 1100
 * Rule bits = 0010
 * Possible Rule Vectors :
 *
 * mask bits   rule bits
 *  0000        0010
 *  0100        0110
 *  1000        1010
 *  1100        1110
 *
 *  Rule bits considered = 0010 (2)
 *
 *  Algorithm:
 *  1.For each bank (4 such abnks to which rule corresponds to )
 *      2. For each successive 16 addresses (8 such addresses in a bank )
 *
 *          Find  4 bit address corresponding to first location and
 *          last location in which value stored is 1.
 *          4 mask bits = first location address  XOR  last location address
 *          4 rule bits = first location address  AND  last location address
 *
 *      2. end of for loop 2
 *  1.end of for loop 1
 *
 *  Example:
 *  first location address = 0010
 *  last location address  = 1100
 *  Write 1 in only the locations 2,6,10,14 among the corresponding 16 locations.
 *  Write zero in all other 12 locations.
 *
 *  Mask bits considered = First Mask bits XOR Last mask bits
 *                       = 0010 XOR 1100
 *                       = 1100
 *
 * \param   IKeyData        Pointer to key Vector (0 - 127)
 * \param   IKeyMask        Pointer to kay mask Vector (0 - 127)
 * \param   RuleVector_512  Pointer to 512 bit Rule Vector (input parameter)
 * \return  PR_ERR_NONE        - Success
 * \return  PR_ERR_FLOWID      - Flow Id is invalid
 * \return  PR_ERR_RULE_MEM_ID - Rule Section is invalid
 * \return  PR_ERR_NULL_PTR    - rule32_valPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - Chip Id is invalid
 * \ingroup Internal
 */
s32  ifx_generate_rule_mask_vectors(u32* IKeyData, u32* IKeyMask, u8* RuleVector_512)
{

    u32 mask_4=0; /*  4 bit mask (for 16 addresses) */
    u32 rule_4=0; /*  4 bit rule (for 16 addresses) */
    u32 mask_first=0; /* first position in which RuleVector_512[j]=1 for each 16 positions */
    u32 mask_last=0;   /* last  position in which RuleVector_512[j]=1 for each 16 positions */
    u32 i=0;
    u32 j=0;
    u32 k=0;

#ifdef PARAMETER_CHECK

    if(IKeyData== NULL)
        return PR_ERR_NULL_PTR;
    if(IKeyMask== NULL)
        return PR_ERR_NULL_PTR;
    if(RuleVector_512 == NULL)
        return PR_ERR_NULL_PTR;
#endif

    /* Generate IKeyMask(0:127) from RuleVector_512[j] for j= 0 to j=511 */

    for(k=0; k<RULE_MEM_NUM; k++)
    {
        IKeyMask[k] = 0;
        IKeyData[k] = 0;
        for(i=0; i<2*BITS_PER_IKEY; i++)
        {
            j = 0;
            mask_first = 0;
            mask_last = 0;

            while((RuleVector_512[j + ADDRESS_PER_IKEY*i + ADDRESS_PER_RULEMEM*k] != 1) &&
                  (j<ADDRESS_PER_IKEY))
            {
                j++;
            }

            if (j == 16)
                break ; /* rule corresponding to this IKEY position is zero */

            mask_first =j;
            mask_last = j;
            j++;

            while(j<ADDRESS_PER_IKEY)
            {
                if(RuleVector_512[j + ADDRESS_PER_IKEY*i + ADDRESS_PER_RULEMEM*k] ==1)
                    mask_last = j;
                j++;
            }

            mask_4 = mask_first ^ mask_last; /* 4 bit mask for each Ikey(i) */
            rule_4 = mask_first & mask_last; /* 4 bit rule for each Ikey(i) */
            IKeyMask[k] |= mask_4 << 4*i ; /* 32bit mask for one bank(k) */
            IKeyData[k] |= rule_4 << 4*i ; /* 32bit mask for one bank(k) */
        }
    }/* 128 bit mask and rule for one rule */

    return PR_ERR_NONE;

}

/**
 * Delete rule for flow ID.
 * This function is used to delete the rule specified by flowId.
 *
 * \param   chipId          Chip Id
 * \param   flowId          Rule Id that need to be deleted.
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_FLOWID   - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_delete_flow_pattern( u8 chipId, u32 flowId )
{
    u32 rule32_val=0;
    u32 rule_pos=0;
    u32 section=0;
    u32 bank=0;
    u32 address=0;
    u32 k=0;
    s32 retflag;


#ifdef PARAMETER_CHECK

    if(flowId >MAX_FLOW_ID )
        return PR_ERR_FLOWID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

#endif
/*  Find the section and the bit position in section to which the rule belongs to */

    section = (flowId / RULES_PER_SECTION ) % RULE_MEM_NUM;
    rule_pos = flowId % RULES_PER_SECTION ;

    for(k=0; k<RULE_MEM_NUM; k++)
    {
        bank = (flowId / ADDRESS_PER_RULEMEM ) + k * RULE_MEM_NUM;

        for(address=0; address<ADDRESS_PER_RULEMEM; address++)
        {
            if( (retflag=ifx_get_rule_mem(chipId, bank, address,section,&rule32_val))
                                != PR_ERR_NONE )
                return retflag;
            /* write zero in the corresponding rule id position*/
            rule32_val &= ~(1 << rule_pos) ;

            if( (retflag=ifx_set_rule_mem(chipId,bank,address,section,rule32_val))
                                != PR_ERR_NONE )
                return retflag;
        } /* for all 128 addresses */
    } /* for all 512 addresses, 4 banks */

    return PR_ERR_NONE;
}



/*****************************************************************************
 *
 * FUNCTION:  ifx_set_flow_action_all
 *
 * DESCRIPTION: This function is used to set the Flow action table.
 *
 * PARAMETERS:
 *      IN:
 *              ChipId      ChipId.
 *              flowId
 *              FlowAction
 *      OUT:
 *             None
 *
 *
 * RETURNS:
 *       PR_ERR_NONE            Success
 *       PR_ERR_NULL_PTR        Null Pointer
 *       PR_ERR_CHIP_ID         Invalid Chip Id
 *
 ***************************************************************************/
/**
 * Write flow action to flow action table.
 * This function writes the passed value flowAction into the flow action table
 * The table position is specified by flowId.
 *
 * \param   chipId          Chip Id
 * \param   flowId          Rule Id that need to be deleted.
 * \param   flowAction      Value for flow action table
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_FLOWID   - If the flowId is not in the range of 0 to 23
 * \return  PR_ERR_CHIP_ID  - Chip Id is invalid
 * \ingroup Internal
 */
s32 ifx_set_flow_action_all( u8 chipId, u32 flowId, u32 flowAction )
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);

    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    u32  i=0;

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if(flowId >= TOTAL_PORT_RULE_ID)
        return PR_ERR_FLOWID;

#endif

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *) (&temp1);

    PR_Reg_Write(ptr_cpu_acs_data,flowAction);

    /*
     * configure cpu access control and data register for action table write access
     */
    ptr_cpu_acs_ctrl->Req_ack = REQ;  /* Request */
    ptr_cpu_acs_ctrl->Read_write = WRITEPR;  /* write */
    ptr_cpu_acs_ctrl->Index = flowId;  /* one of 24 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel = ACTION_TB_SEL ; /* Flow Action Table */

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,flowAction,SEM_CPU);

    return PR_ERR_NONE;
}


#if 0
/*****************************************************************************
 *
 * FUNCTION:  ifx_get_flow_action_all
 *
 * DESCRIPTION: This function is used to gete the flow action by flowId.
 *
 * PARAMETERS:
 *      IN:
 *              ChipId      ChipId.
 *              flowId      Rule Id that need to be deleted.
 *      OUT:
 *             FlowActionPtr
 *
 *
 * RETURNS:
 *       PR_ERR_NONE            Success
 *       PR_ERR_NULL_PTR        If the flowId is not in the range of 0 to 511
 *       PR_ERR_CHIP_ID         Invalid Chip Id
 *
 ***************************************************************************/


s32 ifx_get_flow_action_all(u8 chipId,u32 flowId,u32 *flowActionPtr )
{


    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;

    volatile u32 temp1 = 0;
    u32 Temp = 0;
    u32* ptr_reg1;
    u32 i=0;

#ifdef PARAMETER_CHECK

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

    if(flowActionPtr == NULL)
        return PR_ERR_NULL_PTR;

#endif

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_data =(cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *) (&temp1);

/*  Place the command in cpu access control register */
    ptr_cpu_acs_ctrl->Req_ack = REQ;  /* Request */
    ptr_cpu_acs_ctrl->Read_write = READPR;  /* write */
    ptr_cpu_acs_ctrl->Index = flowId;  /* one of 512 Flow Ids */
    ptr_cpu_acs_ctrl->Table_sel = ACTION_TB_SEL ; /* Flow Action Table */


/* Added MACRO for indirect register read 7/7/2004 */
    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);


    *flowActionPtr = Temp;

    return PR_ERR_NONE;
}
#endif

