/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup Internal Internally used functions */

#include <linux/kernel.h>
#include <asm/types.h>
#include <asm/addrspace.h>
#include <asm/incaip2/ifx_types.h>
#include <asm/incaip2/incaip2.h>
#include "struct.h"
#include "const.h"
#include "qos.h"
#include "err.h"
#include "var.h"
#include "api_init_prv.h"

#undef INCAIP2_DEBUG_QOS

extern s32 ifx_get_PORT_rate_shape( u8 chipId, BOOL* en_dis_ptr, BOOL* PREPtr );

/**
 * Set values for token bucket ID.
 * Set the Token Bucket Count, Replenish Rate Index, Burst Size Index associated
 * with a Token Bucket ID.
 *
 * \param   chipId              Chip Id
 * \param   tBucketId           Token Bucket Id (0 - 23)
 * \param   tBucketCnt          32 bit Bucket Count
 * \param   tReplenishRateIdx   Index to one of 8 Replinish Rate Registers (0 - 7)
 *                              to be assigned to bucket ID
 * \param   tBurstSizeIdx       Index into one of 8 Burst Registers (0 - 7) 
 *                              to be assigned to bucket ID
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_BKT_ID_NUM   - If the Bucket Id is not in the range of 0 to 63    
 * \return  PR_ERR_RATE_NUM     - If index into Replenish Rate registers is not in the range of 0 to 7
 * \return  PR_ERR_BURST_NUM    - If index into Burst registers is not in the range of 0 to 7
 * \return  PR_ERR_TIME_OUT     - Time out 
 * \return  PR_ERR_NULL_PTR     - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_QoS_bucket(u8 chipId,u32 tBucketId, u32 tBucketCnt,
                      u32 tReplenishRateIdx, u32 tBurstSizeIdx)
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;
    u32 Temp=0;
    u32 i=0;
    volatile u32 temp1=0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;
	
#ifdef INCAIP2_DEBUG_QOS
    printk("%s( chipId=%d, tBucketId=%d, tBucketCnt=0x%08X, tReplenishRateIdx=0x%08X, tBurstSizeIdx=0x%08X )\n", 
    __FUNCTION__, chipId, tBucketId, tBucketCnt, tReplenishRateIdx, tBurstSizeIdx ); 
#endif  
    
#ifdef PARAMETER_CHECK
    
    if(tBucketId >= TOTAL_BUCKET_ID)
    return PR_ERR_BKT_ID_NUM;

    if(tReplenishRateIdx >= TOTAL_RATE_REGS)
    return PR_ERR_RATE_NUM;

    if(tBurstSizeIdx >= TOTAL_BURST_REGS)
    return PR_ERR_BURST_NUM;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif  

    ptr_cpu_acs_data = (cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *)&temp1;

    /*
     * configure ACS_CTRL_REG for writing bucket counter value for passed bucket ID 
     */
    ptr_cpu_acs_ctrl->Req_ack        = REQ;  /* Request */
    ptr_cpu_acs_ctrl->Read_write     = WRITEPR;  /* write */
    ptr_cpu_acs_ctrl->Table_specific = 0;          /* TB counter */
    ptr_cpu_acs_ctrl->Index          = tBucketId;  /* one of 512 Bucket Ids */
    ptr_cpu_acs_ctrl->Table_sel      = TOKEN_BKT_TB_SEL ;  /* Token Bucket */

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,tBucketCnt,SEM_CPU);

    /* 
     * write replenish rate register index and burst size register index 
     * to cpu_acs_data_reg for passed bucket ID.  
     */
    Temp = (tBurstSizeIdx << BURST_POS) | tReplenishRateIdx;
    
    ptr_cpu_acs_ctrl->Req_ack        = REQ;  
    ptr_cpu_acs_ctrl->Read_write     = WRITEPR; 
    ptr_cpu_acs_ctrl->Table_specific = 1;  /* Control Bits */
    ptr_cpu_acs_ctrl->Index          = tBucketId;  /* one of 512 Bucket Ids */
    ptr_cpu_acs_ctrl->Table_sel      = TOKEN_BKT_TB_SEL ; /* Token Bucket */

    PR_Indirect_Reg_Write_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    return PR_ERR_NONE;
            
}

/**
 * Get values for token bucket ID.
 * Get the Token Bucket Count, Replenish Rate Index, Burst Size Index associated 
 * with a Token Bucket ID.
 *
 * \param   chipId               Chip Id
 * \param   tBucketId            Token Bucket Id (0 - 63)
 * \param   tBucketCntPtr        Pointer to 32 bit Bucket Count
 * \param   tReplenishRateIdxPtr Pointer to Index into one of 8 Replinish Rate 
 *                               Registers(0-7)
 * \param   tBurstSizeIdxPtr     Pointer to Index into one of 8 Burst Registers(0 - 7) 
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_BKT_ID_NUM   - If the Bucket Id is not in the range of 0 to 63    
 * \return  PR_ERR_TIME_OUT     - Time out 
 * \return  PR_ERR_NULL_PTR     - tBucketCntPtr, tReplenishRateIdxPtr, tBurstSizeIdxPtr
 *                                or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_get_QoS_bucket(u8 chipId,u32 tBucketId,u32 * tBucketCntPtr,
                  u32 * tReplenishRateIdxPtr,u32 * tBurstSizeIdxPtr)
{
    volatile cpu_acs_ctrl_reg_t*   ptr_cpu_acs_ctrl;
    cpu_acs_data_reg_t*   ptr_cpu_acs_data;
    u32 Temp=0, i=0;
    volatile u32 temp1 = 0;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef INCAIP2_DEBUG_QOS
    printk("%s( chipId=%d, tBucketId=%d)\n", __FUNCTION__, chipId, tBucketId ); 
#endif

#ifdef PARAMETER_CHECK
    if(tBucketId >= TOTAL_BUCKET_ID)
        return PR_ERR_BKT_ID_NUM;
    if(tBucketCntPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(tReplenishRateIdxPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(tBurstSizeIdxPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    
    ptr_cpu_acs_data = (cpu_acs_data_reg_t *) (Chip_Base_Address[chipId] + CPU_ACS_DATA_REG);
    ptr_cpu_acs_ctrl = (volatile cpu_acs_ctrl_reg_t *)&temp1;
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + CPU_ACS_CTRL_REG);

    /*
     * Read token bucket counter value via cpu access control register 
     * (lower 32 bits access) for passed token bucket ID
     */
    ptr_cpu_acs_ctrl->Req_ack        = REQ; 
    ptr_cpu_acs_ctrl->Read_write     = READPR;  
    ptr_cpu_acs_ctrl->Table_specific = 0;  /* select TB counter access (lower 32 bits) */
    ptr_cpu_acs_ctrl->Index          = tBucketId;          /* one of 24 Bucket Ids */
    ptr_cpu_acs_ctrl->Table_sel      = TOKEN_BKT_TB_SEL  ;/* Token Bucket */

    if((retflag = ifx_sem_take(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);
    *tBucketCntPtr = Temp ;

    /*
     * Read replenish rate register index and burst size register index 
     * via cpu access control register (upper 32 bits access) 
     * for passed token bucket ID
     */
    ptr_cpu_acs_ctrl->Req_ack        = REQ;            
    ptr_cpu_acs_ctrl->Read_write     = READPR;
    ptr_cpu_acs_ctrl->Table_specific = 1;     /* repl. rate index and burst size index
                                                 (upper 32 bits access)*/
    ptr_cpu_acs_ctrl->Index          = tBucketId;     /* one of 24 Bucket Ids */
    ptr_cpu_acs_ctrl->Table_sel      = TOKEN_BKT_TB_SEL ; /* Token Bucket Table */

    PR_Indirect_Reg_Read_Sem(ptr_reg1,ptr_cpu_acs_data,ptr_cpu_acs_ctrl,temp1,Temp,SEM_CPU);

    if((retflag = ifx_sem_give(chipId,SEM_CPU)) != PR_ERR_NONE)
            return retflag;

    *tReplenishRateIdxPtr = Temp & REPL_IDX_MASK;    
    *tBurstSizeIdxPtr = (Temp >> BURST_POS) & BURST_MASK;

    return PR_ERR_NONE;
            
}

/**
 * Set flow control for token bucket ID.
 * Set the Token Bucket Flow Control Actions for loosely-conforming packets 
 * (actionLC) and non-conforming Packets (actionNC).
 *
 * PortRuleBaseIndex,actionLC,actionNC are set for the whole switch. Since Flow
 * Control is asserted for a port,the Flow Control is used only for PRBI to 
 * PRBI+ 50. Other flows out of this range will not be able to perform Flow 
 * Control,but they can still reduce CoS.So,set either both Flow Control and 
 * CoS - or - Flow Control, and drop packets during programming.
 * For example,if flow control and CoS are set,PRBI to PRBI+50 will have both.
 * Other ranges will have only CoS.
 *
 * \param   chipId           Chip Id
 * \param   portRuleBaseIdx  If the action is Assert Flow Control, then from 
 *                           portRuleBaseIdx to portRuleBaseIdx + 50 are used 
 *                           for flow control for corresponding port. 
 *                           portRuleBaseIdx range is 0-511 If the portRuleBaseIdx 
 *                           X, and if  flow control need to be asserted on 
 *                           port Y, then the token bucket considered is X + Y.
 * \param   actionLC         4 bit Flow Control Action for loosley confirming packets
 * \param   actionNC         4 bit Flow Control Action for non confirming packets
 *                           actionLC & actionNC are 4 bit value indicating one 
 *                           or more actions.
 *                             - bit 3 : forward
 *                             - bit 2 : reduce CoS
 *                             - bit 1 : assert flow control
 *                             - bit 0 : drop
 *                           When bit3 or bit 2 is set both bit 1 and bit 0 can 
 *                           not be set. Hence actions can be any one of the 
 *                           following values.
 *                             - forward
 *                             - forward & reduce CoS
 *                             - assert flow control
 *                             - assert flow control and drop
 *                             - drop
 * \return  PR_ERR_NONE            - Success 
 * \return  PR_ERR_PORTRULE_ID_NUM - If the portRuleBaseIdx is not in the range 
 *                                   of 0 to 23
 * \return  PR_ERR_LC_BITS         - The error condition occurs in folowing cases:
 *                                     - 1. If both Forwarding bit and Drop bit 
 *                                          of actionLC are set
 *                                     - 2. If both Reduce CoS and Drop bits of 
 *                                          actionLC are set.
 * \return  PR_ERR_NC_BITS         - The error condition occurs in folowing cases:
 *                                     - 1. If both Forwarding bit and Drop bit 
 *                                          of actionNC are set
 *                                     - 2. If both Reduce CoS and Drop bits of 
 *                                          actionNC are set.
 * \return  PR_ERR_NULL_PTR        - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID         - If the chip Id invalid
 * \ingroup Internal
 */
s32 ifx_set_QoS_bucket_flow_control(u8 chipId,u32 portRuleBaseIdx, u32 actionLC,
                                 u32 actionNC)
{
    tb_ctrl_reg_t*   ptr_tb_ctrl;
    u32 temp1 = 0;
    u32* ptr_reg1;

#ifdef PARAMETER_CHECK

    if(portRuleBaseIdx >= TOTAL_PORT_RULE_ID)
        return PR_ERR_PORTRULE_ID_NUM;

    if(((actionLC & TB_CTRL_INVALID1)==9) || ((actionLC & TB_CTRL_INVALID2)==5))
        return PR_ERR_LC_BITS;

    if(((actionNC & TB_CTRL_INVALID1)==9) || ((actionNC & TB_CTRL_INVALID2)==5))
        return PR_ERR_NC_BITS;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;


#endif

    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + TB_CTRL_REG);

    ptr_tb_ctrl = (tb_ctrl_reg_t *)&temp1;
        
    ptr_tb_ctrl->rule_base_idx = portRuleBaseIdx ;
    ptr_tb_ctrl->actionLC      = actionLC;
    ptr_tb_ctrl->actionNC      = actionNC;
 
    PR_Reg_Write(ptr_reg1,temp1);
    
    return PR_ERR_NONE;
            
}

/**
 * Get flow control for token bucket ID.
 * Get the Token Bucket Flow Control Actions for loosely-conforming packets 
 * (actionLC) and nonconforming Packets (actionNC). 
 *
 * \param   chipId              Chip Id
 * \param   portRuleBaseIdxPtr  If the action is Assert Flow Control, then from 
 *                              portRuleBaseIdx to portRuleBaseIdx + 50 are used
 *                              for flow control for corresponding port. 
 *                              portRuleBaseIdx range is 0-511. If the 
 *                              portRuleBaseIdx X, and if  flow control need to 
 *                              be asserted on 
 * \param   actionLCPtr         4 bit Flow Control Action for loosley confirming 
 *                              packets
 * \param   actionNCPtr         4 bit Flow Control Action for non confirming 
 *                              packets actionLC & actionNC are 4 bit value 
 *                              indicating one or more actions.
 *                                - bit 3 : forward
 *                                - bit 2 : reduce CoS
 *                                - bit 1 : assert flow control
 *                                - bit 0 : drop
 *                              When bit3 or bit 2 is set both bit 1 and bit 0 
 *                              can not be set. Hence actions can be any one of 
 *                              the following values.
 *                                - forward
 *                                - forward & reduce CoS
 *                                - assert flow control
 *                                - assert flow control and drop
 *                                - drop
 * \return  PR_ERR_NONE         - Success 
 * \return  PR_ERR_NULL_PTR     - portRuleBaseIdxPtr, actionLCPtr, actionNCPtr 
 *                                or base address is null Pointer
 * \return  PR_ERR_CHIP_ID      - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_get_QoS_bucket_flow_control(u8 chipId,u32* portRuleBaseIdxPtr, 
               u32* actionLCPtr, u32* actionNCPtr)
{
    tb_ctrl_reg_t*   ptr_tb_ctrl;
    u32 temp1=0;
    u32* ptr_reg1;
    

#ifdef PARAMETER_CHECK
    
    if(portRuleBaseIdxPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(actionLCPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(actionNCPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + TB_CTRL_REG);

    PR_Reg_Read(ptr_reg1,temp1);
    ptr_tb_ctrl = (tb_ctrl_reg_t *)&temp1;
    
        
    *portRuleBaseIdxPtr = ptr_tb_ctrl->rule_base_idx ;
    *actionLCPtr = ptr_tb_ctrl->actionLC ;
    *actionNCPtr = ptr_tb_ctrl->actionNC ;
    
    return PR_ERR_NONE;
            
}

/**
 * Set parameters for WFQ.
 * Set the Weight Factor & Deficit Counter for the Weighted Fair Queue.  
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id (0 - 2) 
 * \param   queueId         Queue id (0 -3)
 * \param   weightFactor    16 bit Weight Factor associated with each queue for 
 *                          WFQ scheduling
 * \param   deficitCount    17 bit DeficitCount for WFQ scheduling
 * \return  PR_ERR_NONE     - Success 
 * \return  PR_ERR_PORTID   - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_QUEUEID  - If the queueId is not in the range of 0 to 3
 * \return  PR_ERR_TIME_OUT - Time out
 * \return  PR_ERR_NULL_PTR - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_set_QoS_WF_queue( u8 chipId,u32 portId, u32 queueId, 
                          u32 weightFactor, u32 deficitCount)
{
    volatile sd_cmd_reg_t*    ptr_sd_cmd ;
    volatile sd_data_reg2_t*  ptr_sd_data2 ;
    volatile sd_data_reg1_t*  ptr_sd_data1 ;

    volatile u32 temp = 0;
    volatile u32 temp2 = 0,temp1 = 0;
    u32 i=0;
    u32* ptr_reg;
    u32* ptr_reg2;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;

#ifdef PARAMETER_CHECK

    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;

    if(queueId >= TOTAL_EGRESS_QUEUES)
        return PR_ERR_QUEUEID;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(weightFactor > MAX_WEIGHT_FACTOR)
        return PR_ERR_MAX_WEIGHT ;

    if(deficitCount > MAX_DEF_COUNT)
        return PR_ERR_DEF_COUNT ;    

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_reg = (u32*)(Chip_Base_Address[chipId] + SD_CMD_REG);
    ptr_reg2 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS2);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS1);

    ptr_sd_cmd = (volatile sd_cmd_reg_t*)&temp;
    ptr_sd_data2 = (sd_data_reg2_t*)&temp2; 
    ptr_sd_data1 = (sd_data_reg1_t*)&temp1;
            
	/*  Place the weightFactor, deficitCount in Scheduler Data Structure  
        Data Registers */
    ptr_sd_data2->Weight = (weightFactor >> SD_WFQ_SHF) ;
    ptr_sd_data1->Weight = weightFactor & SD_WFQ_MSK;
    ptr_sd_data2->Deficit_cnt = deficitCount;

    if((retflag = ifx_sem_take(chipId,SEM_SCHED)) != PR_ERR_NONE)
            return retflag;

#ifdef INCAIP2_DEBUG_QOS 
    printk("%s - write 0x%08X to SD_DATA_REG1\n", __FUNCTION__, temp1);    
    printk("%s - write 0x%08X to SD_DATA_REG2\n", __FUNCTION__, temp2);       
#endif      		

    PR_Reg_Write(ptr_reg2,temp2);
    PR_Reg_Write(ptr_reg1,temp1);

	/*  Place the command in Scheduler Data Structure Command Register */
    ptr_sd_cmd->Req_ack = REQ;
    ptr_sd_cmd->Read_write = WRITEPR; 
    ptr_sd_cmd->field_sel =  SD_DEFICIT_CNT_SEL | SD_WFQ_SEL ;
    ptr_sd_cmd->Queue_id = queueId ;
    ptr_sd_cmd->Port_id = portId ;

#ifdef INCAIP2_DEBUG_QOS
    printk("%s - write 0x%08X to SD_CMD_REG\n", __FUNCTION__, temp);       
#endif  
	/*  Wait for acknowledgement */
    PR_Reg_Write(ptr_reg,temp);

    for (i=0; i<PR_TIME_OUT;i++)
    {
        PR_Reg_Read(ptr_reg,temp);
        if(ptr_sd_cmd -> Req_ack == ACK) 
            break;

    }

    if((ptr_sd_cmd -> Req_ack) != ACK) 
        return PR_ERR_TIME_OUT;

    if((retflag = ifx_sem_give(chipId,SEM_SCHED)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}

/**
 * Get parameters for WFQ.
 * Get the Weight Factor & Deficit Counter for the Weighted Fair Queue.  
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id (0 - 2) 
 * \param   queueId         Queue id (0 -3)
 * \param   weightFactorPtr Pointer to 16 bit Weight Factor associated with each 
 *                          queue for WFQ scheduling
 * \param   deficitCountPtr Pointer to 21 bit DeficitCount for WFQ scheduling
 * \return  PR_ERR_NONE     - Success 
 * \return  PR_ERR_PORTID   - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_QUEUEID  - If the queueId is not in the range of 0 to 3
 * \return  PR_ERR_TIME_OUT - Time out
 * \return  PR_ERR_NULL_PTR - weightFactorPtr, deficitCountPtr or base address 
 *                            is null Pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_get_QoS_WF_queue(u8 chipId,u32 portId, u32 queueId, 
        u32* weightFactorPtr, u32* deficitCountPtr)
{
    volatile sd_cmd_reg_t*    ptr_sd_cmd ;
    volatile sd_data_reg2_t*  ptr_sd_data2 ;
    volatile sd_data_reg1_t*  ptr_sd_data1 ;

    volatile u32 temp = 0;
    volatile u32 temp2 = 0,temp1 = 0;
    u32 i=0;
    u32* ptr_reg;
    u32* ptr_reg2;
    u32* ptr_reg1;
    s32 retflag = PR_ERR_NONE;


#ifdef PARAMETER_CHECK

    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;
    if(queueId >= TOTAL_EGRESS_QUEUES)
        return PR_ERR_QUEUEID;
    if(weightFactorPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(deficitCountPtr == NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif


    ptr_reg = (u32*)(Chip_Base_Address[chipId] + SD_CMD_REG);
    ptr_reg2 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS2);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS1);

    ptr_sd_cmd = (volatile sd_cmd_reg_t*)&temp;
    ptr_sd_data2 = (sd_data_reg2_t*)&temp2; 
    ptr_sd_data1 = (sd_data_reg1_t*)&temp1;

	/*  Place the command in Scheduler Data Structure Command Register */
    ptr_sd_cmd->Req_ack = REQ;
    ptr_sd_cmd->Read_write = READPR; 
    ptr_sd_cmd->field_sel = SD_DEFICIT_CNT_SEL | SD_WFQ_SEL  ;
    ptr_sd_cmd->Queue_id = queueId ;
    ptr_sd_cmd->Port_id = portId ;

    if((retflag = ifx_sem_take(chipId,SEM_SCHED)) != PR_ERR_NONE)
            return retflag;

#ifdef INCAIP2_DEBUG_QOS
    printk("%s - write 0x%08X to SD_CMD_REG\n", __FUNCTION__, temp);       
#endif  	
	
	/*  Write read command and wait for acknowledgement */
    PR_Reg_Write(ptr_reg,temp);

    for (i=0; i<PR_TIME_OUT;i++)
    {
        PR_Reg_Read(ptr_reg,temp);
        if(ptr_sd_cmd -> Req_ack == ACK)
            break;
    }

    if((ptr_sd_cmd -> Req_ack) != ACK)
        return PR_ERR_TIME_OUT;

	/*  Place the weightFactor, deficitCount from Scheduler Data Structure
        Data Registers */
    PR_Reg_Read(ptr_reg2,temp2);
    PR_Reg_Read(ptr_reg1,temp1);

#ifdef INCAIP2_DEBUG_QOS
    printk("%s - read 0x%08X from SD_DATA_REG1\n", __FUNCTION__, temp1);    
    printk("%s - read 0x%08X from SD_DATA_REG2\n", __FUNCTION__, temp2);       
#endif    	
	
    if((retflag = ifx_sem_give(chipId,SEM_SCHED)) != PR_ERR_NONE)
        return retflag;

    *weightFactorPtr = (ptr_sd_data2->Weight << SD_WFQ_SHF ) |
                       (ptr_sd_data1->Weight & SD_WFQ_MSK) ;
    *deficitCountPtr =   ptr_sd_data2->Deficit_cnt   ;
    
    return PR_ERR_NONE;
}

/**
 * Set parameters for rate shaping.
 * Set the buckeCnt, replenishRate & tokenValue and Replenish timer for Rate 
 * Shaping. If Rate shaping mode is port based, set the same values for all 4 
 * queues.If it is queue based,set the values for the specified queue. 
 *
 * \param   chipId          Chip Id
 * \param   portId          Port Id (0 - 2) 
 * \param   queueId         Queue id (0 -3)
 * \param   bucketCnt       21 bit Bucket count for Rateshaping
 * \param   replenishRate   20 bit Replinish Rate for Rateshaping
 * \param   tokenValue      20 bit Token value for Rateshaping
 * \param   tReplenishTimer Replinish rate timer for Rateshaping.
 *                          Its value is 0 for 1/128 ms, value 1 for 1/64 ms, 
 *                          value 2 for 1/8 ms & value 3 for 1ms
 * \return  PR_ERR_NONE     - Success 
 * \return  PR_ERR_PORTID   - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_QUEUEID  - If the queueId is not in the range of 0 to 3
 * \return  PR_ERR_TIME_OUT - Time out
 * \return  PR_ERR_NULL_PTR - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_set_QoS_shaping_queue( u8 chipId,u32 portId, u32 queueId, u32 bucketCnt,
                               u32 replenishRate, u32 tokenValue, 
                               u32 tReplenishTimer )
{
    volatile sd_cmd_reg_t*   ptr_sd_cmd ;
    volatile sd_data_reg1_t*  ptr_sd_data1 ;
    volatile sd_data_reg0_t*  ptr_sd_data0 ;
    volatile sd_data_reg2_t*  ptr_sd_data2 ;
    volatile u32 temp = 0;
    volatile u32 temp0 = 0,temp1 = 0, temp2 = 0;
    u32 i=0;
    u32* ptr_reg;
    u32* ptr_reg1;
    u32* ptr_reg2;
    u32* ptr_reg0;

    BOOL PRE = FALSE ;
    BOOL en_dis;    
    s32 status = 0 ;
    s32 retflag = PR_ERR_NONE;
#ifdef INCAIP2_DEBUG_QOS
    printk("%s( chipId=%d, portId=%d, queueId=%d, bucketCnt=0x%08X, \n",
           __FUNCTION__, chipId, portId, queueId, bucketCnt);
    printk("          replenishRate=0x%08X, tokenValue=0x%08X, tReplenishTimer=%d\n", 
            replenishRate, tokenValue, tReplenishTimer);
#endif 
#ifdef PARAMETER_CHECK

    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;

    if(queueId >= TOTAL_EGRESS_QUEUES)
        return PR_ERR_QUEUEID;

    if(bucketCnt >= MAX_BUCKET_COUNT)
        return PR_ERR_BKT_CNT_NUM;  
    
    if(replenishRate >= MAX_REPLENISH_RATE)
        return PR_ERR_RATE_NUM;    
    
    if(tokenValue >= MAX_TOKEN_VALUE)
        return PR_ERR_TOKEN_VALUE;  
    
    if(tReplenishTimer > MAX_REPL_TIMER_VALUE)
        return PR_ERR_REPL_TIMER_VALUE;  

    
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif

    ptr_reg =  (u32*)(Chip_Base_Address[chipId] + SD_CMD_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS1);
    ptr_reg2 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS2);
    ptr_reg0 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS0);

    ptr_sd_cmd = (volatile sd_cmd_reg_t*)&temp;
    ptr_sd_data0 = (sd_data_reg0_t*)&temp0;
    ptr_sd_data1 = (sd_data_reg1_t*)&temp1; 
    ptr_sd_data2 = (sd_data_reg2_t*)&temp2;
            
    /* chech whether port based rate shaping or queue based rate shaping */

    status = ifx_get_PORT_rate_shape( chipId, &en_dis, &PRE );
    if(status != PR_ERR_NONE)
        return status ;

	/*  Place the weightFactor, deficitCount in Scheduler Data Structure 
        Data Registers */
    ptr_sd_data1->Bucket_cnt = bucketCnt;
    ptr_sd_data1->Replinish_rate = replenishRate >> SD_REP_RATE_SHF ;
    ptr_sd_data0->Replinish_rate = replenishRate &  SD_REP_RATE_MSK;
    ptr_sd_data0->Token_val = tokenValue;
    ptr_sd_data2->Replinish_timer = tReplenishTimer ;

    if((retflag = ifx_sem_take(chipId,SEM_SCHED)) != PR_ERR_NONE)
            return retflag;

    PR_Reg_Write(ptr_reg0,temp0);
    PR_Reg_Write(ptr_reg1,temp1);
    PR_Reg_Write(ptr_reg2,temp2);
#ifdef INCAIP2_DEBUG_QOS
    printk("%s - SD_DATA_REG0=0x%08X\n", __FUNCTION__, temp0);    
    printk("%s - SD_DATA_REG1=0x%08X\n", __FUNCTION__, temp1);    
    printk("%s - SD_DATA_REG2=0x%08X\n", __FUNCTION__, temp2);       
#endif    
    
/*  Place the command in Scheduler Data Structure Command Register */

    ptr_sd_cmd->Read_write = WRITEPR; 
    ptr_sd_cmd->field_sel = SD_BKT_CNT_SEL | SD_TIMER_SEL |
                            SD_REP_RATE_SEL | SD_TKN_VAL_SEL;
    ptr_sd_cmd->Port_id = portId ;

    if(PRE == TRUE) /*port based rate shaping */
    {
        for(queueId = 0 ; queueId < TOTAL_EGRESS_QUEUES ; queueId ++ )
        {
            ptr_sd_cmd->Queue_id = queueId ;
            ptr_sd_cmd->Req_ack = REQ;
#ifdef INCAIP2_DEBUG_QOS
            printk("%s - write 0x%08X to SD_CMD_REG\n", __FUNCTION__, temp);
#endif			
            /*  Wait for acknowledgement */
            PR_Reg_Write(ptr_reg,temp);

            for (i=0; i<PR_TIME_OUT;i++)
            {
                PR_Reg_Read(ptr_reg,temp);
                if(ptr_sd_cmd -> Req_ack == ACK) 
                    break;
            }
#ifdef INCAIP2_DEBUG_QOS			
            printk("%s - SD_CMD_REG=0x%08X\n", __FUNCTION__, temp);
#endif			
            if((ptr_sd_cmd -> Req_ack) != ACK) 
                return PR_ERR_TIME_OUT;

        } /* for each queue, queue 0 - 3, */

    } /* end of port based rate shaping */
    else /* queue based rate shaping */
    {

        ptr_sd_cmd->Queue_id = queueId ;
        ptr_sd_cmd->Req_ack = REQ;
    
        /*  Wait for acknowledgement */
#ifdef INCAIP2_DEBUG_QOS		
        printk("%s - write 0x%08X to SD_CMD_REG\n", __FUNCTION__, temp);
#endif		
        PR_Reg_Write(ptr_reg,temp);
    
        for (i=0; i<PR_TIME_OUT;i++)
        {
            PR_Reg_Read(ptr_reg,temp);
            if(ptr_sd_cmd -> Req_ack == ACK) 
                break;
        }
#ifdef INCAIP2_DEBUG_QOS		
        printk("%s - SD_CMD_REG=0x%08X\n", __FUNCTION__, temp);
#endif		
        if((ptr_sd_cmd -> Req_ack) != ACK) 
            return PR_ERR_TIME_OUT;

    }/* end of queue based rate shaping */


    if((retflag = ifx_sem_give(chipId,SEM_SCHED)) != PR_ERR_NONE)
        return retflag;

    return PR_ERR_NONE;
}

/**
 * Get parameters for rate shaping.
 * Get the buckeCnt, replenishRate, tokenValue and Replenish timer for Rate Shaping.
 *
 * \param   chipId             Chip Id
 * \param   portId             Port Id (0 - 2) 
 * \param   queueId            Queue id (0 -3)
 * \param   bucketCntPtr       Pointer to 21 bit Bucket count for Rateshaping
 * \param   replenishRatePtr   Pointer to 20 bit Replinish Rate for Rateshaping
 * \param   tokenValuePtr      Pointer to 20 bit Token value for Rateshaping
 * \param   tReplenishTimerPtr Pointer to Replinish rate timer for Rateshaping.
 *                             Its value is 0 for 1/128 ms,value 1 for 1/64 ms, 
 *                             value 2 for 1/8 ms & value 3 for 1ms    
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_PORTID      - If the portId is not in the range of 0 to 2
 * \return  PR_ERR_QUEUEID     - If the queueId is not in the range of 0 to 3
 * \return  PR_ERR_TIME_OUT    - Time out
 * \return  PR_ERR_NULL_PTR    - bucketCntPtr, replenishRatePtr, tokenValuePtr, 
 *                               tReplenishTimerPtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_get_QoS_shaping_queue(u8 chipId,u32 portId, u32 queueId, u32* bucketCntPtr, 
         u32* replenishRatePtr, u32* tokenValuePtr,u32* tReplenishTimerPtr)
{
    volatile sd_cmd_reg_t*   ptr_sd_cmd ;
    volatile sd_data_reg1_t*  ptr_sd_data1 ;
    volatile sd_data_reg0_t*  ptr_sd_data0 ;
    volatile sd_data_reg2_t*  ptr_sd_data2 ;
    volatile u32 temp = 0;
    volatile u32 temp0 = 0,temp1 = 0, temp2 = 0;
    u32 i=0;
    u32* ptr_reg;
    u32* ptr_reg1;
    u32* ptr_reg2;
    u32* ptr_reg0;
    s32 retflag = PR_ERR_NONE;


#ifdef PARAMETER_CHECK

    if(portId >= TOTAL_PORTS)
        return PR_ERR_PORTID;
    if(queueId >= TOTAL_EGRESS_QUEUES)
        return PR_ERR_QUEUEID;
    if(bucketCntPtr == NULL)
        return PR_ERR_NULL_PTR;
    if(replenishRatePtr== NULL)
        return PR_ERR_NULL_PTR;
    if(tokenValuePtr == NULL)
        return PR_ERR_NULL_PTR;
    if(tReplenishTimerPtr== NULL)
        return PR_ERR_NULL_PTR;

    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;

#endif


    ptr_reg = (u32*)(Chip_Base_Address[chipId] + SD_CMD_REG);
    ptr_reg1 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS1);
    ptr_reg2 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS2);
    ptr_reg0 = (u32*)(Chip_Base_Address[chipId] + SD_DATA_REGS0);

    ptr_sd_cmd = (volatile sd_cmd_reg_t*)&temp;

    ptr_sd_data0 = (sd_data_reg0_t*)&temp0;
    ptr_sd_data1 = (sd_data_reg1_t*)&temp1; 
    ptr_sd_data2 = (sd_data_reg2_t*)&temp2;

/*  Place the command in Scheduler Data Structure Command Register */
    ptr_sd_cmd->Req_ack = REQ;
    ptr_sd_cmd->Read_write = READPR; 
    ptr_sd_cmd->field_sel = (SD_BKT_CNT_SEL | SD_TIMER_SEL | 
                             SD_REP_RATE_SEL | SD_TKN_VAL_SEL) ;
    ptr_sd_cmd->Queue_id = queueId ;
    ptr_sd_cmd->Port_id = portId ;

    if((retflag = ifx_sem_take(chipId,SEM_SCHED)) != PR_ERR_NONE)
            return retflag;
#ifdef INCAIP2_DEBUG_QOS
    printk("%s - write 0x%08X to SD_CMD_REG\n", __FUNCTION__, temp);
#endif	
    PR_Reg_Write(ptr_reg,temp);

    for (i=0; i<PR_TIME_OUT;i++)
    {
        PR_Reg_Read(ptr_reg,temp);
        if(ptr_sd_cmd -> Req_ack == ACK) 
            break;

    }
#ifdef INCAIP2_DEBUG_QOS	
    printk("%s - SD_CMD_REG=0x%08X\n", __FUNCTION__, temp);
#endif	
    if((ptr_sd_cmd -> Req_ack) != ACK) 
        return PR_ERR_TIME_OUT;

/*  Place the weightFactor, deficitCount in Scheduler Data Structure 
    Data Registers */
    PR_Reg_Read(ptr_reg0,temp0);
    PR_Reg_Read(ptr_reg1,temp1);
    PR_Reg_Read(ptr_reg2,temp2);
#ifdef INCAIP2_DEBUG_QOS
    printk("%s - SD_DATA_REG0=0x%08X\n", __FUNCTION__, temp0);    
    printk("%s - SD_DATA_REG1=0x%08X\n", __FUNCTION__, temp1);    
    printk("%s - SD_DATA_REG2=0x%08X\n", __FUNCTION__, temp2);        
#endif

    if((retflag = ifx_sem_give(chipId,SEM_SCHED)) != PR_ERR_NONE)
        return retflag;

    *bucketCntPtr = ptr_sd_data1->Bucket_cnt ;
    *replenishRatePtr = ptr_sd_data1->Replinish_rate << SD_REP_RATE_SHF
                        | ptr_sd_data0->Replinish_rate  ;
    *tokenValuePtr = ptr_sd_data0->Token_val;
    *tReplenishTimerPtr = ptr_sd_data2->Replinish_timer;
    
    return PR_ERR_NONE;
}

/**
 * Set token bucket replenish rate.
 * Set the Token Bucket Replenish Rate.
 *
 * \param   chipId             Chip Id
 * \param   tReplenishRateIdx  Index into one of 8 Replenish Rate registers  
 * \param   tReplenishRate     Replenish Rate which is to be programmed in 
 *                             selected register
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_RATE_NUM    - If index into Replenish Rate registers is not  
 *                               in the range of 0 to 7
 * \return  PR_ERR_NULL_PTR    - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_set_QoS_replenish_rate( u8 chipId, u32 tReplenishRateIdx, 
                                u32 tReplenishRate )
{
    rate_reg_t*   ptr_rate ;

#ifdef PARAMETER_CHECK

    if(tReplenishRateIdx >= TOTAL_RATE_REGS)
        return PR_ERR_RATE_NUM ;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_rate = (rate_reg_t*) (Chip_Base_Address[chipId] + RATE_REG ) + tReplenishRateIdx;

    PR_Reg_Write(ptr_rate,tReplenishRate);
   
    return PR_ERR_NONE;
}

/**
 * Get token bucket replenish rate.
 * Get the Token Bucket Replenish Rate.
 *
 * \param   chipId             Chip Id
 * \param   tReplenishRateIdx  Index into one of 8 Replenish Rate registers  
 * \param   tReplenishRatePtr  Pointer to Replenish Rate
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_RATE_NUM    - If index into Replenish Rate registers is not  
 *                               in the range of 0 to 7
 * \return  PR_ERR_NULL_PTR    - tReplenishRatePtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_get_QoS_replenish_rate(u8 chipId,u32 tReplenishRateIdx, 
                            u32 * tReplenishRatePtr)
{
    rate_reg_t*   ptr_rate ;
    u32 Temp; 

#ifdef PARAMETER_CHECK

    if(tReplenishRateIdx >= TOTAL_RATE_REGS)
        return PR_ERR_RATE_NUM ;
    if(tReplenishRatePtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    
    ptr_rate =  (rate_reg_t*) (Chip_Base_Address[chipId] + RATE_REG )+ tReplenishRateIdx;
    
    PR_Reg_Read(ptr_rate,Temp);
    *tReplenishRatePtr = Temp;

    return PR_ERR_NONE;
}

/**
 * Set token bucket burst size.
 * Set the Token Bucket Burst size.
 *
 * \param   chipId             Chip Id
 * \param   tBurstSizeIdx      Index into one of 8 Burst registers
 * \param   tBurstSize         Burst Size
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_BURST_NUM   - If index into Burst registers is not in the 
 *                               range of 0 to 7
 * \return  PR_ERR_NULL_PTR    - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_set_QoS_burst_size(u8 chipId,u32 tBurstSizeIdx, u32 tBurstSize)
{
    burst_reg_t*   ptr_burst ;

#ifdef PARAMETER_CHECK

    if(tBurstSizeIdx >= TOTAL_BURST_REGS)
        return PR_ERR_BURST_NUM ;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_burst  =   (burst_reg_t*) (Chip_Base_Address[chipId] + BURST_REG ) + tBurstSizeIdx;

    PR_Reg_Write(ptr_burst,tBurstSize);

    
    return PR_ERR_NONE;
}

/**
 * Get token bucket burst size.
 * Get the Token Bucket Burst size.
 *
 * \param   chipId             Chip Id
 * \param   tBurstSizeIdx      Index into one of 8 Burst registers
 * \param   tBurstSizePtr      Pointer to Burst Size
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_BURST_NUM   - If index into Burst registers is not in the 
 *                               range of 0 to 7
 * \return  PR_ERR_NULL_PTR    - tBurstSizePtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_get_QoS_burst_size(u8 chipId,u32 tBurstSizeIdx, u32 *tBurstSizePtr)
{
    burst_reg_t*   ptr_burst ;
    u32 Temp=0;

#ifdef PARAMETER_CHECK

    if(tBurstSizeIdx >= TOTAL_BURST_REGS)
        return PR_ERR_BURST_NUM ;
    if(tBurstSizePtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif
    
    ptr_burst  =   (burst_reg_t*) (Chip_Base_Address[chipId] + BURST_REG )+ tBurstSizeIdx;
    
    PR_Reg_Read(ptr_burst,Temp);
    
    *tBurstSizePtr = Temp;
        
    return PR_ERR_NONE;
}

/**
 * Set token bucket extended burst size.
 * Set the Token Bucket Extended Burst size.
 *
 * \param   chipId             Chip Id
 * \param   tEBurstSizeIdx     Index into one of 8 Extended Burst registers
 * \param   tEBurstSize        Extended Burst Size which is to be programmed
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_EBURST_NUM  - If index into Extended Burst registers is not 
 *                               in the range of 0 to 7
 * \return  PR_ERR_NULL_PTR    - Base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_set_QoS_extend_burst_size(u8 chipId,u32 tEBurstSizeIdx, u32 tEBurstSize)
{
    eburst_reg_t*   ptr_eburst ;

#ifdef PARAMETER_CHECK

    if(tEBurstSizeIdx >= TOTAL_EBURST_REGS)
        return PR_ERR_EBURST_NUM ;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_eburst  =   (eburst_reg_t*) (Chip_Base_Address[chipId] + EBURST_REG ) + tEBurstSizeIdx;

    PR_Reg_Write(ptr_eburst,tEBurstSize);

    
    return PR_ERR_NONE;
}

/**
 * Get token bucket extended burst size.
 * Get the Token Bucket Extended Burst size.  
 *
 * \param   chipId             Chip Id
 * \param   tEBurstSizeIdx      Index into one of 8 Extended Burst registers
 * \param   tEBurstSizePtr      Pointer to Extended Burst Size
 * \return  PR_ERR_NONE        - Success 
 * \return  PR_ERR_EBURST_NUM  - If index into Extended Burst registers is not 
 *                               in the range of 0 to 7
 * \return  PR_ERR_NULL_PTR    - tEBurstSizePtr or base address is null Pointer
 * \return  PR_ERR_CHIP_ID     - If the chip Id invalid
 * \ingroup Internal
 */ 
s32 ifx_get_QoS_extend_burst_size(u8 chipId,u32 tEBurstSizeIdx, u32 *tEBurstSizePtr)
{
    eburst_reg_t*   ptr_eburst ;
    u32 Temp=0;

#ifdef PARAMETER_CHECK

    if(tEBurstSizeIdx >= TOTAL_EBURST_REGS)
        return PR_ERR_EBURST_NUM ;
    if(tEBurstSizePtr == NULL)
        return PR_ERR_NULL_PTR;
    if(chipId >= Total_Chips)
        return PR_ERR_CHIP_ID ;

    if(Chip_Base_Address[chipId] == CHIP_ADDRESS_INVALID)
        return PR_ERR_NULL_PTR;
#endif

    ptr_eburst  =   (eburst_reg_t*) (Chip_Base_Address[chipId] + EBURST_REG ) + tEBurstSizeIdx;

    PR_Reg_Read(ptr_eburst,Temp);
    
    *tEBurstSizePtr = Temp;

    
    return PR_ERR_NONE;
}


