/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/smp_lock.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/uaccess.h>
#include <linux/init.h>

#include <asm/incaip2/ifx_types.h>
#include <asm/incaip2/switch.h>
#include "switch_api.h"
#include "ioctl.h"
#include "api_init_prv.h"

static int dev_initialized;

extern s32 ifx_phy_read( u8 chipId, u8 phy_addr, u8 RegOffset, u16 * pwData );
extern s32 ifx_phy_write( u8 chipId, u8 phy_addr, u8 RegOffset, u16 wData );

s32 ifx_set_STP_port_state(u8 chipId,u32 portId, u32 portState);
s32 ifx_get_STP_port_state(u8 chipId,u32 portId, u32* portStatePtr);
s32 ifx_map_ingress_priority_CoS(u8 chipId,u32 IngressPriority, u32 CoS);
s32 ifx_get_ingress_priority_CoS(u8 chipId,u32 IngressPriority, u32* CoSPtr);

#if 0
static struct switch_api_command {
    const char *name;
    const void *function;
    const char *param;          /* i=int, l=long, s=string */
} commands[] = {
    { "ifx_map_ingress_priority_CoS", ifx_map_ingress_priority_CoS, "lll" }
};

static const char *s_api_get_string(const char **buffer, int *len)
{
    const char *s1 = *buffer;
    const char *s2;
    int quoted = 0;

    while (*s1 == ' ') { s1++; }

    if ((*s1 == '\0') || (*s1 == '\n')) /* end of string */
    {
        *len = 0;
        return NULL;
    }

    if (*s1 == '"')
    {
        quoted = 1;
        s1++;
    }

    s2 = s1;
    while ((*s2) && (*s2 != '\n'))
    {
        if (((*s2 == ' ') && (!quoted)) ||
            ((*s2 == '"') && ( quoted)) ||
            ((*s2 == ',') && (!quoted)))
        {
            break;
        }
        s2++;
    }

    *len = (s2 - s1);

    if (quoted && (*s2 == '"')) s2++;
    if (*s2 == ',') s2++;       /* remove comma between arguments */
    *buffer = s2;               /* save start of next token */

    return s1;
}

static long s_api_atoi(const char *buffer, int len)
{
    long val = 0;
    int hex = 0;

    if ((len > 2) && (buffer[0]=='0') && (buffer[1]=='x'))
    {
        hex = 1;
        len -= 2;
        buffer += 2;
    }

    while (*buffer && len)
    {
        if ((*buffer >= '0') && (*buffer <= '9'))
        {
            val = val * ((hex) ? 16 : 10) + (*buffer - '0');
        }
        else if ((hex && (*buffer >= 'A') && (*buffer <= 'F')) ||
                 (hex && (*buffer >= 'a') && (*buffer <= 'f')))
        {
            val = val * 16 + ((*buffer & ~0x20) - 'A' + 10);
        }
        else
        {
            printk("atoi: invalid character: '%c'\n", *buffer);
            return 0;
        }
        len--;
        buffer++;
    }

    return val;
}
#endif


/**
 * Read MDIO register.
 * Read from Phy register.
 *
 * \param   chipId          Chip Id
 * \param   phy_addr        MDIO address of the PHY device
 * \param   RegOffset       PHY register address
 * \param   pwData          Pointer to Data read from phy register
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_PHY_RD   - If phy read fails
 * \return  PR_ERR_NULL_PTR - pwData or base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup API
 */
s32 ifx_mdio_read(u8 chipId, u8 phy_addr, u8 RegOffset, u16 * pwData)
{
    return ifx_phy_read(chipId, phy_addr, RegOffset, pwData );
}

/**
 * Write to MDIO register.
 * Write 16 bit value wData to Phy register defined by passed phy_addr and
 * RegOffset.
 *
 * \param   chipId          Chip Id
 * \param   phy_addr        MDIO address of the PHY device
 * \param   RegOffset       PHY register address
 * \param   wData           Data to be written into phy register
 * \return  PR_ERR_NONE     - Success
 * \return  PR_ERR_PHY_WR   - If Phy write fails
 * \return  PR_ERR_NULL_PTR - Base address is null pointer
 * \return  PR_ERR_CHIP_ID  - If the chip Id invalid
 * \ingroup API
 */
s32 ifx_mdio_write( u8 chipId, u8 phy_addr, u8 RegOffset, u16 wData )
{
    return ifx_phy_write(chipId, phy_addr, RegOffset, wData );
}


/**
 * Open switch API device.
 * This function increases the usage count of the switch API device.
 *
 * \param   inode - Inode of device
 * \param   file  - File structure of device
 * \return  0     - OK, device opened
 * \ingroup API
 */
static int s_api_open (struct inode *inode, struct file *file)
{
    unsigned int minor = MINOR(inode->i_rdev);
        /*increamet the number of device users so it will not be removed*/
    MOD_INC_USE_COUNT;

    DPRINTK("opened: s_api-%d\n", minor);

    return 0;
}

/**
 * Close switch API device.
 * This function decreases the usage count of the switch API device.
 *
 * \param   inode - Inode of device
 * \param   file  - File structure of device
 * \return  0     - OK, device closed
 * \ingroup API
 */
static int s_api_release (struct inode *inode, struct file *file)
{
    unsigned int minor = MINOR(inode->i_rdev);

    DPRINTK("closed: s_api-%d\n", minor);

        /*decrement the number of device users*/
    MOD_DEC_USE_COUNT;
    /* here we should realse the API table memory*/
    return 0;
}



/**
 * Read switch API device.
 * The switch API device does not support direct reading, thus this function
 * does not return any data.
 *
 * \param   file   - File structure of device node
 * \param   buf    - Destination buffer to copy data to
 * \param   count  - Number of bytes to be read
 * \param   offset - Current position in file
 * \return  0      - Always zero
 * \ingroup API
 */
static ssize_t s_api_read (struct file *file, char *buf, size_t count, loff_t *offset)
{
    DPRINTK("enter %s\n", __FUNCTION__);
        return 0;
}

/**
 * Write to switch API device.
 * The switch API device does not support writing, thus this function
 * does use any provided data.
 *
 * \param   file   - File structure of device node
 * \param   buf    - Destination buffer to copy data to
 * \param   count  - Number of bytes to be read
 * \param   offset - Current position in file
 * \return  0      - Always zero
 * \ingroup API
 */
static ssize_t s_api_write (struct file *file, const char *buf, size_t count, loff_t *offset)
{
    DPRINTK("enter %s\n", __FUNCTION__);
    return 0;
}

/**
 * Switch API IOCTL handler.
 * The IOCTL handler checks the given command and will call the appropriate
 * subhandler. If an error occurs in the subhandler, its error code will also be
 * returned from this function.
 * The following subhandlers are available:
 *   - Address resolution configuration (#addrresl_ioctl)
 *   - MAC configuration (#mac_ioctl)
 *   - General port configuration (#port_ioctl)
 *   - QOS configuration (#qos_ioctl)
 *   - General switch configuration (#switch_ioctl)
 *
 * \param   inode  - Inode of device
 * \param   file   - File structure of device
 * \param   cmd    - IOCTL command
 * \param   arg    - Argument for some IOCTL commands
 * \return  0      - OK
 * \return  ENOTTY - Wrong IOC magic number or IOC number too large or
 *                   invalid command
 * \return  EFAULT - Error accessing data from user space
 * \ingroup API
 */
int s_api_ioctl(struct inode *inode, struct file *file,
                            unsigned int cmd, unsigned long arg)
{
    unsigned int minor = MINOR(inode->i_rdev);
    int err=0;
    if (_IOC_TYPE(cmd) != IFX_IOC_MAGIC)
    {
        DPRINTK("%s: Wrong magic number=%d\n",__FUNCTION__,_IOC_TYPE(cmd));
        return -ENOTTY;
    }
    if (_IOC_NR(cmd) > IFX_IOC_MAXNR)
    {
        DPRINTK("%s: NR exceed max number=%d\n",__FUNCTION__,_IOC_NR(cmd));
        return -ENOTTY;
    }

    if (_IOC_DIR(cmd) & _IOC_READ)
        err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
    else if (_IOC_DIR(cmd) & _IOC_WRITE)
        err =  !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
    if (err)
    {
        DPRINTK("%s: read/write error\n", __FUNCTION__);
        return -EFAULT;
    }

    /*Implement IOCTL*/
    if(cmd==IFX_IOC_HARDRESET)
    {
        DPRINTK("%s: IFX_IOC_HARDRESET\n",__FUNCTION__);
        while (MOD_IN_USE)
                MOD_DEC_USE_COUNT;
        MOD_INC_USE_COUNT;
    }
    else if((_IOC_NR(cmd) >= ADDRRESL_BEGIN) && (_IOC_NR(cmd)<=ADDRRESL_END))
        err = addrresl_ioctl(file,cmd,arg);
/*
        else if((_IOC_NR(cmd) >= CPUPORT_BEGIN) && (_IOC_NR(cmd)<=CPUPORT_END))
            err = cpuport_ioctl(file,cmd,arg);
        else if((_IOC_NR(cmd) >= LED_BEGIN) && (_IOC_NR(cmd)<=LED_END))
            err = led_ioctl(file,cmd,arg);
*/
    else if((_IOC_NR(cmd) >= MAC_BEGIN) && (_IOC_NR(cmd)<=MAC_END))
        err = mac_ioctl(file,cmd,arg);
    else if((_IOC_NR(cmd) >= PORT_BEGIN) && (_IOC_NR(cmd)<=PORT_END))
        err = port_ioctl(file,cmd,arg);
    else if((_IOC_NR(cmd) >= QOS_BEGIN) && (_IOC_NR(cmd)<=QOS_END))
        err = qos_ioctl(file,cmd,arg);
    else if((_IOC_NR(cmd) >= SWITCH_BEGIN) && (_IOC_NR(cmd)<=SWITCH_END))
        err = switch_ioctl(file,cmd,arg);
#ifdef CONFIG_TEST_API
   else if(cmd==IFX_UPDATE_TEST_INDEX)
    {
        copy_from_user (&param_index,(u8 *) arg,sizeof (u8));
        DPRINTK("%s: IFX_UPDATE_TEST_INDEX, param_index = %d\n",__FUNCTION__,param_index);
    }
#endif //CONFIG_TEST_API
    else
        return -ENOTTY;
/*
    if(err)
        printk("%s returns %d\n", __FUNCTION__, err);
*/
    return -err;
}

static struct file_operations s_api_fops = {
    owner:          THIS_MODULE,
    open:           s_api_open,
    release:        s_api_release,
    read:           s_api_read,
    write:          s_api_write,
    ioctl:      s_api_ioctl,
};

#ifdef MODULE
/**
 * Remove switch API module.
 * This function is called upon removal of the switch API driver. It unregisters
 * the device and will free all allocated memory.
 *
 * \return  0        - OK, no problem
 * \return  EFAULT   - There was a problem while freeing memory
 * \ingroup Internal
 */
static int s_api_cleanup(void)
{

    u8 No_of_Chips =1;
    int res;

    dev_initialized --;

    if ((res = devfs_unregister_chrdev(S_API_MAJOR,"incaip2_switch_API"))) {
        printk("unable to release major %d for s_api\n", S_API_MAJOR);
    }

    /*remove memory allocated for the use of the switch api*/
    if(ifx_release_API_mem(No_of_Chips))
        return -EFAULT;

    return 0;
}
#endif

/**
 * Initialize switch API module.
 * This function registers the switch API device.
 *
 * \return  0        - OK, no problem
 * \return  ENOMEM   - An error occured during registration
 * \ingroup Internal
 */
int  s_api_init(void)
{

    int res;

    printk("INCAIP2 Switch API initializing\n");
    dev_initialized = 0;

    /* We set the base address to 0x00000000, since all register macros already
       include the correct base address */

    if( ifx_set_chip_number(1) )
        printk("ifx_set_chip_number() failed\n");
    if( ifx_set_base_address(0,0x00000000) )
        printk("ifx_set_base_address() failed\n");

#ifdef CONFIG_DEVFS_FS
    res = devfs_register_chrdev(S_API_MAJOR,"incaip2_switch_API",&s_api_fops);
    if(res < 0)
    {
        DPRINTK("%s: can't get major\n", __FUNCTION__);
        goto ERR_REG;
    }

    res = (int) devfs_register( NULL, "incaip2_switch_API",
                                DEVFS_FL_NONE, S_API_MAJOR, 0,
                                S_IFCHR | S_IRUGO | S_IWUGO,
                                &s_api_fops, NULL);
    if (res == 0)
    {
        DPRINTK("Unable to register switch API device\n");
        goto ERR_REG;
    }
#else
    if ((res = register_chrdev(S_API_MAJOR, "incaip2_switch_API", &s_api_fops)) < 0)
    {
        DPRINTK("Unable to register major %d for INCAIP2 switch API\n", S_API_MAJOR);
        goto ERR_REG;
    }
#endif

    DPRINTK("%s: registered incaip2_switch_API major_id=%d\n",__FUNCTION__, S_API_MAJOR);
    dev_initialized ++;

    return 0;

ERR_REG:
    /*an error ocurred. unregister the driver*/
    devfs_unregister_chrdev(S_API_MAJOR,"incaip2_switch_API");
    dev_initialized --;
    DPRINTK("%s:unregister device after error \n",__FUNCTION__);
    return -ENOMEM;
}


/**
 * Module init function.
 * This function calls the switch API init function if it is compiled as a
 * kernel module.
 *
 * \return  0        - OK, no problem
 * \return  ENOMEM   - An error occured during registration
 * \ingroup Internal
 */
static int __init init_module(void)
{

        return s_api_init();
}

#ifdef MODULE
/**
 * Module cleanup function.
 * This function calls the switch API cleanup function, when the module is
 * removed.
 *
 * \ingroup Internal
 */
static void __exit cleanup_module(void)

{
    /*remove memory allocated for the use of the switch api*/
        s_api_cleanup();

}

MODULE_LICENSE("GPL");

/* Macros for mapping the load and remove functions */
module_exit(cleanup_module)
#endif
module_init(init_module)


EXPORT_SYMBOL(s_api_ioctl);
EXPORT_SYMBOL(ifx_mdio_read);
EXPORT_SYMBOL(ifx_mdio_write);

