/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/
/* Group definitions for Doxygen */
/** \addtogroup API API-Functions */
/** \addtogroup Internal Internally used functions */

#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <asm/addrspace.h>
#include <asm/uaccess.h>
#include <asm/incaip2/mux.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/irq.h>
#include "tsf_defs.h"

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
static devfs_handle_t devfs_handle_ledmatrix;
static devfs_handle_t devfs_handle_keypad;
#endif

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Inca-IP2 TSF Driver");

/* number of supported PWM devices */
#define MAX_PWM     2

/*
 * Parameter Definitions for Module
 */
 
/* PWM Parameters */
static int pwm1 = 1;       // Availability of PWM1 (0 or 1) 
MODULE_PARM(pwm1, "i");
MODULE_PARM_DESC(pwm1, "Enable PWM1 (0 or 1)");

static int pwm2 = 1;       // Availability of PWM2 (0 or 1) 
MODULE_PARM(pwm2, "i");
MODULE_PARM_DESC(pwm2, "Enable PWM2 (0 or 1)");

static int pwm1_value = 0;  // Initial value for PWM1
MODULE_PARM(pwm1_value, "i");
MODULE_PARM_DESC(pwm1_value, "Initial value for PWM1 (0..255)");

static int pwm2_value = 255;  // Initial value for PWM2
MODULE_PARM(pwm2_value, "i");
MODULE_PARM_DESC(pwm2_value, "Initial value for PWM2 (0..255)");

static int pwm_major = 0;   // Major ID of PWM device (default: 0=dynamic)
MODULE_PARM(pwm_major, "i");
MODULE_PARM_DESC(pwm_major, "Major ID of PWM device (default: 0=dynamic)");

/* LED Parameters */
static int led_pins = -1;   // Number of LED pins (5..10) 
MODULE_PARM(led_pins, "i");
MODULE_PARM_DESC(led_pins, "Number of LED pins (5..10)");

static int ncol = -1;       // Number of columns in LED array (2 or 4)
MODULE_PARM(ncol, "i");
MODULE_PARM_DESC(ncol, "Number of columns in LED array (2 or 4)");

static int inv  = -1;       // Invert Polarity of Column Driver Signal
MODULE_PARM(inv, "i");
MODULE_PARM_DESC(inv, "Invert Polarity of Column Driver Signal (0 for external NPN transistor, 1 for direct LED drive)");

static int led_eth_lan = 0; // LEDs controlled by Ethernet LAN Port
MODULE_PARM(led_eth_lan, "i");
MODULE_PARM_DESC(led_eth_lan, "LEDs controlled by Ethernet LAN Port (bits[4:0] = [SPD, ACT, DPX, STA, TL])");

static int led_eth_pc = 0;  // LEDs controlled by Ethernet PC Port
MODULE_PARM(led_eth_pc, "i");
MODULE_PARM_DESC(led_eth_pc, "LEDs controlled by Ethernet PC Port (bits[4:0] = [SPD, ACT, DPX, STA, TL])");

static int led_major = 0;   // Major ID of ledmatrix device (default: 0=dynamic)
MODULE_PARM(led_major, "i");
MODULE_PARM_DESC(led_major, "Major ID of ledmatrix device (default: 0=dynamic)");

/* Keypad Parameters */
static int key_pins = -1;   // Number of KEY pins (2..15) 
MODULE_PARM(key_pins, "i");
MODULE_PARM_DESC(key_pins, "Number of KEY pins (2..15)");

static int key_major = 0;   // Major ID of keypad device (default: 0=dynamic)
MODULE_PARM(key_major, "i");
MODULE_PARM_DESC(key_major, "Major ID of keypad device (default: 0=dynamic)");

static char *dev_name[MAX_PWM]  = { "pwm1", "pwm2" };
static int  *dev_avail[MAX_PWM] = { &pwm1, &pwm2 };
static int   dev_id[MAX_PWM]    = {0, 1};

static u32 led_mask = 0x00FFFFFF;

static int opened_by_user = 0;  /* TODO: replace with spin_lock */
static int opened_by_kernel = 0;

keypad_callback_t keypad_callback = NULL;

static u16 scan_lines_buf0[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static u16 scan_lines_buf1[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static u16 *last_scan_lines = scan_lines_buf0;
static u16 *curr_scan_lines = scan_lines_buf1;

static key_event_t key_fifo[KEY_FIFO_DEPTH];
static key_event_t *rd_ptr = key_fifo;
static key_event_t *wr_ptr = key_fifo;

static struct proc_dir_entry *ifx_pwm_proc[MAX_PWM];
static struct proc_dir_entry *ifx_led_proc;
static struct proc_dir_entry *ifx_keypad_proc;

/**
 * Convert String to Decimal.
 * Converts an ascii string of decimal digits to a positive integer.
 *
 * \param   s       pointer to the ascii string
 * \return  >=0     converted integer value
 * \return  -EINVAL unexpected character in ascii string
 * \ingroup Internal  
 */
static int ascii2int(char* s)
{
    int temp_val = 0;

    while (*s >= '0' && *s <= '9') {
        temp_val = temp_val * 10 + (*s - '0');
        s++;
    }
    
    if (*s != 0xA) { // string should end with line feed (0xA)
        printk(KERN_ERR "unexpected character in ascii string!\n");
        return -EINVAL;
    }
    return temp_val;
}

/**
 * Convert Hex to unsigned integer.
 * Converts an ascii hex string (max. 32bit) to an unsigned integer.
 *
 * \param   s       pointer to the ascii string
 * \return  >=0     converted integer value
 * \return  0       unexpected character in ascii string
 * \ingroup Internal  
 */
static u32 hexstr2uint(char* s)
{
    int digit = 0;
    u32 temp_val = 0;

    while (*s != 0xA) { /* last character should be <LF> */
        if (digit >= 8) {
            printk(KERN_ERR "Enter a Hex value (max 32 bit)!\n");
            return 0; /* results in keeping LEDs as-is! */
        }
        
        if (*s >= '0' && *s <= '9') {
            temp_val = (temp_val << 4) + (*s - '0');
        }
        else if (*s >= 'a' && *s <= 'f') {
            temp_val = (temp_val << 4) + (*s - 'a') + 10;
        }
        else if (*s >= 'A' && *s <= 'F') {
            temp_val = (temp_val << 4) + (*s - 'A') + 10;
        }
        else {
            printk(KERN_ERR "Enter a Hex value!\n");
            return 0; /* results in keeping LEDs as-is! */
        } 
        digit++;
        s++;
    }

    return temp_val;
}

/***** PWM ********************************************************************/

/**
 * Set PWM value.
 * This function sets the Pulse Width value of a PWM device.
 *
 * \param   id  ID of the addressed PWM device
 * \param   value   PWM value in the range 0..255
 * \return  0       OK
 * \return  -EINVAL Device not available or invalid value
 * \ingroup API  
 */
int ifx_pwm_set(int id, u8 value) {
    u32 temp;
    
    if (id < MAX_PWM) {
        if (*dev_avail[id]) {
            temp = (*INCA_IP2_TSF_PWM12) & ~(0xFFFF << 16*id);
            *INCA_IP2_TSF_PWM12 = temp | (value << 16*id);
        }
        else {
            printk(KERN_ERR "Device %s not available!\n", 
                                dev_name[id]);
            return -EINVAL;
        }
    }
    else {
        printk(KERN_ERR "PWM ID out of range!\n");
        return -EINVAL;
    }
    return 0;
}

/**
 * Get PWM value.
 * This function gets the Pulse Width value of a PWM device.
 *
 * \param   id  ID of the addressed PWM device
 * \return  >=0 current PWM value
 * \return  -EINVAL Device not available or invalid ID
 * \ingroup API
 */
int ifx_pwm_get(int id) {
    if (id < MAX_PWM) {
        if (*dev_avail[id]) {
            return (int) ((*INCA_IP2_TSF_PWM12 >> 16*id) & 0xFFFF);
        }
        else {
            printk(KERN_ERR "Device %s not available!\n", 
                                dev_name[id]);
            return -EINVAL;
        }
    }
    else {
        printk(KERN_ERR "PWM ID out of range!\n");
        return -EINVAL;
        
    }
}

/**
 * Read PWM from device.
 * This function is called when a user space program read from the PWM device. 
 * It will copy the current PWM value into the provided buffer.
 *
 * \param   filp    File structure of device node
 * \param   buf     Destination buffer to copy data to
 * \param   count   Number of bytes to be read
 * \param   f_pos   Current position in file
 * \return  count   Number of bytes read
 * \return  -EINVAL     Invalid minor ID provided
 * \return  -EFAULT Error on copying data to user space
 * \ingroup API  
 */
static ssize_t ifx_pwm_read(struct file *filp, char *buf,
                     size_t count, loff_t *f_pos)
{
    struct inode *inode = filp->f_dentry->d_inode;
    int id = MINOR(inode->i_rdev);
    char pwm_val;
    
    if (id < MAX_PWM) {
        pwm_val = ifx_pwm_get(id);
    }
    else {
        printk(KERN_ERR "PWM ID out of range!\n");
        return -EINVAL;
    }
    
    count = 1;
    if (copy_to_user(buf, &pwm_val, count)) {
        return -EFAULT;
    }
    return count;
}

/**
 * Write PWM to device.
 * This function is called when a user space program writes to the PWM device. 
 * It will read the new PWM value from the buffer and then set it. If more than 
 * one byte shall be written and error will occur.
 *
 * \param   filp    File structure of device node
 * \param   buf     Source buffer to read value from 
 * \param   count   Number of bytes to be written (must be 1)
 * \param   f_pos   Current position in file
 * \return  count   Number of bytes written
 * \return  -EINVAL Invalid minor ID provided or to manny bytes written.
 * \return  -EFAULT Error on copying data from user space
 * \ingroup API  
 */
static ssize_t ifx_pwm_write(struct file *filp, const char *buf,
                     size_t count, loff_t *f_pos)
{
    struct inode *inode = filp->f_dentry->d_inode;
    int id = MINOR(inode->i_rdev);
    char pwm_val;
    
    if (count != 1) {
        return -EINVAL;
    }
    
    if (copy_from_user(&pwm_val, buf, count)) {
        return -EFAULT;
    }
    
    if (id < MAX_PWM) {
        ifx_pwm_set(id, pwm_val);
    }
    else {
        printk(KERN_ERR "PWM ID out of range!\n");
        return -EINVAL;
    }
    return count;
}

struct file_operations pwm_fops = {
    read  : ifx_pwm_read,
    write : ifx_pwm_write,
};

/***** LED ********************************************************************/
#define INCAIP2_MUX_LED_INV    0    
#define INCAIP2_MUX_LED_NCOL   4    
#define inca_led_check(indexLED) (indexLED & ~led_mask)

/**
 * Clear LED bits.
 * Switch off the LED(s) of the Matrix. All set bits in the 'indexLED' are
 * switched off. Other bits are ignored
 *
 * \param   indexLED    bit vector as in register LED_REG 
 * \return  0           If clearing the LED bits was successfull
 * \return  indexLED    If the chip or the implementation doesn't support to 
 *                      clear one or more LEDs, the function is returning the 
 *                      indexLED bits that lead to the error.
 * \ingroup API
 */
u32 ifx_led_off(u32 indexLED)
{
    int res;
    
    /* check if all selected LED pins are supported by the hardware */
    res = inca_led_check(indexLED);
    if (res == 0) {
        /* clear the bit in the LED register. */
        /* All other LED are not changed */
        *INCA_IP2_TSF_LED_REG &= ~indexLED;
    }
    return res;
}

/**
 * Set LED bits.
 * Switch on the LED(s) of the Matrix. All set bits in the 'indexLED' are
 * switched on. Other bits are ignored
 *
 * \param   indexLED    bit vector as in register LED_REG 
 * \return  0           If setting the LED bits was successfull
 * \return  indexLED    If the chip or the implementation doesn't support to 
 *                      set one or more LEDs, the function is returning the 
 *                      indexLED bits that lead to the error.
 * \ingroup API
 */
u32 ifx_led_on(u32 indexLED)
{
    int res;

    /* check if all selected LED pins are supported by the hardware */
    res = inca_led_check(indexLED);
    if (res == 0) {
        /* set the bit in the LED register. */
        /* All other LED are not changed */
        *INCA_IP2_TSF_LED_REG |= indexLED;
    }
    return res;
}

/**
 * Configure Ethernet LEDs.
 * Set the bits for Ethernet LEDs for Speed, Activity, Duplex Mode, Status 
 * and Transmission.
 *
 * \param   eth_port    Selected Ethernet port (LED_ETHPORT_PC or LED_ETHPORT_LAN)
 * \param   eth_leds    u8 defining the bits to be set
 * \return  0           Setting the LED bits was successfull
 * \return  eth_leds    An invalid port was specified
 * \ingroup Internal  
 */
u8 ifx_led_eth_ctrl(int eth_port, u8 eth_leds)
{
    switch (eth_port) {
        case LED_ETHPORT_PC:
            *INCA_IP2_TSF_LEDMUX_CONF &= 0x03E00003;
            *INCA_IP2_TSF_LEDMUX_CONF |= ((eth_leds & (LED_ETH_SPD |
                                  LED_ETH_ACT |
                                  LED_ETH_DPX |
                                  LED_ETH_STA |
                                  LED_ETH_TL) 
                                  ) << 16);
            led_mask &= ~((u32) eth_leds &
                    (LED(0,4)|LED(1,4)|LED(2,4)|LED(3,4)) );
            led_mask |= (~((u32) eth_leds) &
                    (LED(0,4)|LED(1,4)|LED(2,4)|LED(3,4)) );
            break;
        case LED_ETHPORT_LAN:
            *INCA_IP2_TSF_LEDMUX_CONF &= 0x001F0003;
            *INCA_IP2_TSF_LEDMUX_CONF |= ((eth_leds & (LED_ETH_SPD |
                                  LED_ETH_ACT |
                                  LED_ETH_DPX |
                                  LED_ETH_STA |
                                  LED_ETH_TL) 
                                  ) << 21);
            led_mask &= ~((u32) (eth_leds << 4) & 
                    (LED(0,5)|LED(1,5)|LED(2,5)|LED(3,5)) );
            led_mask |= (~((u32) (eth_leds << 4)) & 
                    (LED(0,5)|LED(1,5)|LED(2,5)|LED(3,5)) );
            break;
        default:
            printk(KERN_ERR "Invalid Ethernet Port Number!\n");
            return eth_leds;
    }   
    return 0;
}

/**
 * LED matrix IOCTL handler.
 * The following IOCTLs are supported for the LED device.
 * \li #LED_ON
 * \li #LED_OFF
 * \li #LED_ETH_PC
 * \li #LED_ETH_LAN
 *
 * \param   inode   Inode of device 
 * \param   filp    File structure of device
 * \param   cmd     IOCTL command 
 * \param   arg     Argument for some IOCTL commands
 * \return  0               Setting the LED bits was successfull
 * \return  -ENOIOCTLCMD    An invalid command was specified
 * \ingroup API  
 */
static int ifx_led_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg) {
    int err;
    
    switch (cmd) {
        case LED_ON:
            err = ifx_led_on((u32) arg);
            break;
        case LED_OFF:
            err = ifx_led_off((u32) arg);
            break;
        case LED_ETH_PC:
            err = ifx_led_eth_ctrl(LED_ETHPORT_PC, (u8) arg);
            break;
        case LED_ETH_LAN:
            err = ifx_led_eth_ctrl(LED_ETHPORT_LAN, (u8) arg);
            break;
        default:
            err = -ENOIOCTLCMD;
    }
    return err;
}

struct file_operations led_fops = {
    ioctl : ifx_led_ioctl
};

/***** KEY ********************************************************************/

/**
 * Read scan lines.
 * Read keypad scan results from registers and prepare for further processing.
 *
 * \param scan_lines Pointer to the scan lines array
 * \ingroup Internal  
 */
static void get_scanlines(u16 *scan_lines) 
{
    scan_lines[ 0] = (u16) (~(*INCA_IP2_TSF_SREG01)        ) & 0x3FFF;
    scan_lines[ 1] = (u16) (~(*INCA_IP2_TSF_SREG01)    >>15) & 0x3FFE;
    scan_lines[ 2] = (u16) (~(*INCA_IP2_TSF_SREG23)    << 2) & 0x3FFC;
    scan_lines[ 3] = (u16) (~(*INCA_IP2_TSF_SREG23)    >>13) & 0x3FF8;
    scan_lines[ 4] = (u16) (~(*INCA_IP2_TSF_SREG456)   << 4) & 0x3FF0;
    scan_lines[ 5] = (u16) (~(*INCA_IP2_TSF_SREG456)   >> 7) & 0x3FE0;
    scan_lines[ 6] = (u16) (~(*INCA_IP2_TSF_SREG456)   >>18) & 0x3FC0;
    scan_lines[ 7] = (u16) (~(*INCA_IP2_TSF_SREG7TO13) << 7) & 0x3F80;
    scan_lines[ 8] = (u16) (~(*INCA_IP2_TSF_SREG7TO13)     ) & 0x3F00;
    scan_lines[ 9] = (u16) (~(*INCA_IP2_TSF_SREG7TO13) >> 7) & 0x3E00;
    scan_lines[10] = (u16) (~(*INCA_IP2_TSF_SREG7TO13) >>11) & 0x3C00;
    scan_lines[11] = (u16) (~(*INCA_IP2_TSF_SREG7TO13) >>14) & 0x3800;
    scan_lines[12] = (u16) (~(*INCA_IP2_TSF_SREG7TO13) >>16) & 0x3000;
    scan_lines[13] = (u16) (~(*INCA_IP2_TSF_SREG7TO13) >>17) & 0x2000;
};                                   

/**
 * Validate scan lines.
 * Check keypad scan results wrt. row/column rules (refer to User's Manual 
 * for details). On error, reset all resulting scan lines.
 *
 * \param scan_lines Pointer to the scan lines array
 * \return 0 OK
 * \return -1 A row column/conflict occured
 * \ingroup Internal 
 */
static int validate_scanlines(u16 *scan_lines) 
{
    int i, first_set, err = 0;
    u16 dead_clm = 0;
    
    for (i = 0; i < key_pins -1; i++) {
        if (scan_lines[i]) {
            first_set = ffs(scan_lines[i]) -1;
//          printk(KERN_INFO "ffs(0x%08X)=%d\n", scan_lines[i], first_set); 
            /* only one bit in the row? 
               and no conflict with other column yet? */
            if ((1 << first_set) == scan_lines[i] && 
                    !(scan_lines[i] & dead_clm)) {

                dead_clm |= scan_lines[i];
            }
            else {
                printk(KERN_INFO "scan_lines[i]=0x%08X first_set=%d dead_clm=0x%08X\n", scan_lines[i], first_set, dead_clm); 
                printk(KERN_ERR "row/column conflict!\n"); 
                err = -1; 
                break;
            }
        }
    }
    
    return err; 
};

#ifdef KEYPAD_DEBUG
static u32 keymap = 0;
#endif

/**
 * Add new keycodes to FIFO.
 * Compare the last scan with the current scan and derive key events. Add 
 * these key events to FIFO.
 *
 * \param last_scan Pointer to the old scan lines array
 * \param curr_scan Pointer to the new scan lines array
 * \return keycount Number of added key events.
 * \ingroup Internal  
 */
static int add_keycodes(u16 *last_scan, u16 *curr_scan) 
{
    int i, first_set, keybase = 0, key_state, keycount = 0;
    u16 delta;
    char keycode;
    unsigned long time_stamp;
    
    for (i = 0; i < key_pins -1; i++) {
        delta = last_scan[i] ^ curr_scan[i];
        while (delta) {
            first_set = ffs(delta) -1;
            keycode = keybase + first_set;
            key_state = (  (curr_scan[i] >> first_set) & 1
                    ? KEY_PRESSED
                    : KEY_RELEASED);
            time_stamp = jiffies;
            
            if (!FIFO_FULL(rd_ptr, wr_ptr)) {
                /* add keycodes to FIFO */
                wr_ptr->key_code = keycode;
                wr_ptr->key_state = key_state;
                wr_ptr->time_stamp = time_stamp;
                PTR_INCR(wr_ptr);
                keycount++;
            } 
            else {
                printk(KERN_ERR "Key buffer full. "
                    "Ignoring key event!\n");
                break;
            }
            
#ifdef KEYPAD_DEBUG
            if (key_state) {
                keymap |= (1 << keycode);
            }
            else {
                keymap &= ~(1 << keycode);
            }
#endif          
            delta &= ~(1 << first_set);
        }
        keybase += (key_pins-i-2);
    }
#ifdef KEYPAD_DEBUG
    printk("\r");
    for (i=0; i<32;i++) {
        if ((keymap >> i) & 1) printk("%2d", i);
        else printk("  ");
    }
    printk("           ");
#endif  
    return keycount;
}


/**
 * Keypad interrupt handler.
 * The interrupt handler will read the new scan lines and check for new
 * keycodes. If any new keycodes were fond the kernel callback function
 * is called or the user space select call is triggerd.
 * 
 * \param irq Interrupt number
 * \param dev_id Pointer to interrupt specific data (set upon register)
 * \param regs Current MIPS register context
 * \ingroup Internal  
 */
void ifx_keypad_isr(int irq, void *dev_id, struct pt_regs *regs)
{
    int count;
    keypad_dev_t *dev = (keypad_dev_t *) dev_id;
    
    get_scanlines(curr_scan_lines);
    
    if (validate_scanlines(curr_scan_lines) == 0) { /* i.e. valid! */
    
        count = add_keycodes(last_scan_lines, curr_scan_lines);
        if (count == -1) return; /* leave scan codes unchanged! */
        
        /* issue callbacks to other kernel module, if available */
        if (opened_by_kernel) {
            while (rd_ptr != wr_ptr) {
                (*keypad_callback) (rd_ptr->key_code,
                            rd_ptr->key_state,
                            rd_ptr->time_stamp);
                PTR_INCR(rd_ptr);
            }
        }

        /* wake up user process, if available */
        if (opened_by_user) {
            wake_up_interruptible(&dev->waitq);
        }
        /* concept: buffer switching */
        if (last_scan_lines == scan_lines_buf1) {
            last_scan_lines = scan_lines_buf0;
            curr_scan_lines = scan_lines_buf1;
        }
        else {
            last_scan_lines = scan_lines_buf1;
            curr_scan_lines = scan_lines_buf0;
        }
    }
    else return; /* leave scan codes unchanged! */
    
}

/**
 * Open keypad device.
 * This function is called when the keypad device is opend. If the device is
 * not opened from kernel already private data for the interrupt handler is 
 * initialized and then the interrupt is registerd.
 *
 * \param   inode   Inode of device 
 * \param   filp    File structure of device
 * \return  0       OK, device opened
 * \return  -EINVAL There was a problem with request_irq
 * \return  -ENOMEM No more memory available
 * \return  -EBUSY  Device is already opened from kernel
 * \ingroup API  
 */
static int ifx_keypad_open(struct inode *inode, struct file *filp)
{
    keypad_dev_t *dev;
    int err =0;
    
    opened_by_user = 1;
    if (!opened_by_kernel) {
        filp->private_data = kmalloc(sizeof(keypad_dev_t), GFP_KERNEL);
        if (filp->private_data == NULL) {
            opened_by_user = 0;
            return -ENOMEM;
        }
        dev = (keypad_dev_t *) filp->private_data;
        init_waitqueue_head(&dev->waitq);
        
        KEYPAD_INTERRUPTS_CLEAR();
        /* Interrupt */
        err = request_irq(INCAIP2_KEY_IRQ_LINE, ifx_keypad_isr, 0,
                            "keypad", dev);
        if (err) {
            printk(KERN_ERR "Could not request_irq().\n");
            kfree(filp->private_data);
            opened_by_user = 0;
            return err;
        }
        printk(KERN_INFO "Keypad interrupt enabled.\n");
        return 0;
    }
    else {
        opened_by_user = 0;
        printk(KERN_ERR "Keypad driver already registered by "
                        "other kernel module.\n");
        return -EBUSY;
    }
}

/**
 * Release keypad device.
 * This function is called when the keypad device is closed. The interrupt 
 * will be released and the memory freed.
 *
 * \param   inode   Inode of device 
 * \param   filp    File structure of device
 * \return  0       OK, device opened
 * \ingroup API  
 */
static int ifx_keypad_release(struct inode *inode, struct file *filp)
{
    keypad_dev_t *dev;
    
    dev = (keypad_dev_t *) filp->private_data;
    free_irq(INCAIP2_KEY_IRQ_LINE, dev);

    kfree(filp->private_data);
    opened_by_user = 0;

    return 0;
}

/**
 * Keypad IOCTL handler.
 * The following IOCTL is supported for the keypad device.
 * \li #KEYPAD_GET
 *
 * \param   inode   Inode of device 
 * \param   filp    File structure of device
 * \param   cmd     IOCTL command 
 * \param   arg     Argument for some IOCTL commands
 * \return  0           OK
 * \return  -ENOIOCTLCMD    An invalid command was specified
 * \return  -EAGAIN     No key event stored
 * \return  -EFAULT     Error on copying data to user space
 * \ingroup API  
 */
static int ifx_keypad_ioctl(struct inode *inode, struct file *filp, 
                    unsigned int cmd, unsigned long arg) 
{
    int res = 0;
    
    switch (cmd) {
    case KEYPAD_GET:
        if (rd_ptr != wr_ptr) {
            if (copy_to_user((void *) arg, (void *) rd_ptr,
                    sizeof(rd_ptr[0])))
                res = -EFAULT;
#ifdef KEYPAD_DEBUG                 
            printk(KERN_INFO "keycode=%02d, state=%d, "
                "timestamp=%d, res=%08X\n",
                rd_ptr->key_code, rd_ptr->key_state,
                (u32) rd_ptr->time_stamp, res);
#endif          
            PTR_INCR(rd_ptr);
        }
        else {
            res = -EAGAIN; /* no key event stored */
        }
        break;
    default:
        res = -ENOIOCTLCMD;
    }
    return res;
}

/**
 * Poll handler.
 * For select calls from user space this function is called. It will sleep 
 * on the device wait queue until a key event is available.
 *
 * \param   filp File structure of device
 * \param   wait Internal table of poll wait queues
 * \return  mask If new data is available the POLLPRI bit is set, 
 *               triggering an exception indication
 * \ingroup API  
 */
static unsigned int ifx_keypad_poll(struct file *filp, poll_table *wait)
{
    unsigned int mask = 0;
    keypad_dev_t *dev = filp->private_data;
    
    poll_wait(filp, &dev->waitq, wait);
    if (rd_ptr != wr_ptr) mask |= POLLPRI; /* FIFO is not empty */
    
    return mask;
    
}

static struct file_operations keypad_fops = {
    open    : ifx_keypad_open,
    release : ifx_keypad_release,
    ioctl   : ifx_keypad_ioctl,
    poll    : ifx_keypad_poll
};

/**
 * Register kaypad callback.
 * This function is called by another kernel module to register a callback 
 * function for key events.
 * 
 * \param fp Pointer to callback function
 * \return -EBUSY Keypad device is already opended from user space.
 * \return 0 OK
 * \ingroup API  
 */
int ifx_key_register_callback(keypad_callback_t fp) {
    int err =0;
    
    opened_by_kernel = 1;
    if (!opened_by_user) {
        keypad_callback = fp;
        
        KEYPAD_INTERRUPTS_CLEAR();
        /* Interrupt */
        err = request_irq(INCAIP2_KEY_IRQ_LINE, ifx_keypad_isr, 0,
                            "keypad", NULL);
        if (err) {
            printk(KERN_ERR "Could not request_irq().\n");
            return -EBUSY;
        }
        printk(KERN_INFO "Keypad interrupt enabled.\n");
        return 0;
    }
    else {
        opened_by_kernel = 0;
        printk(KERN_ERR "Keypad driver already opened by user "
                            "application.\n");
        return -EBUSY;
    }
}

/**
 * Unregsiter callback function.
 * This function is called by another kernel module to unregister its 
 * callback function. 
 *
 * \return 0 OK
 * \ingroup API  
 */
int ifx_key_unregister_callback(void) {
    free_irq(INCAIP2_KEY_IRQ_LINE, NULL);
    
    keypad_callback = NULL;
    opened_by_kernel = 0;
    return 0;
}

/***** proc *******************************************************************/

/**
 * Create PWM proc file output.
 * This function creates the output for the PWM proc file
 *
 * \param   buf         Buffer to write the string to
 * \param   start   not used (Linux internal)
 * \param   offset  not used (Linux internal)
 * \param   count   not used (Linux internal)
 * \param   eof         Set to 1 when all data is stored in buffer
 * \param   data    not used (Linux internal)
 * \return len  Lenght of data in buffer 
 * \ingroup Internal  
 */
static int ifx_pwm_read_procmem(char *buf, char **start, off_t offset, 
                    int count, int *eof, void *data)
{
    long    len = 0;
    int     id  = *((int *) data);
    
    len += sprintf(buf+len,"PWM_device=/dev/%s\n", dev_name[id]);
    len += sprintf(buf+len,"PWM_value=%d\n", ifx_pwm_get(id));

    *eof = 1;   // No more data available

    return len;
}

/**
 * Process PWM proc file input.
 * This function evaluates the input of the PWM proc file and sets the PWM 
 * value if it is with 0..255
 *
 * \param   file    file structure for proc file
 * \param   buffer  buffer holding the data
 * \param   count   number of characters in buffer
 * \param   data    ID of PWM device (set during registration)
 * \return count    Number of processed characters 
 * \ingroup Internal  
 */
static int ifx_pwm_write_procmem(struct file *file, const char *buffer,
                    unsigned long count, void *data)
{
    char*   ibuf = (char*) buffer;
    int     val = 0;
    int     id  = *((int *) data);

    val = ascii2int(ibuf);
    
    if (val >= 0 && val < 256) {
        ifx_pwm_set(id, val);
    }
    else {
        printk(KERN_ERR "enter a value in the range 0..255!\n"); 
        return -EINVAL;
    }
    
    return count;
    
}

/**
 * Create LED proc file output.
 * This function creates the output for the LED proc file
 *
 * \param   buf         Buffer to write the string to
 * \param   start   not used (Linux internal)
 * \param   offset  not used (Linux internal)
 * \param   count   not used (Linux internal)
 * \param   eof         Set to 1 when all data is stored in buffer
 * \param   data    not used (Linux internal)
 * \return len  Lenght of data in buffer 
 * \ingroup Internal  
 */
static int ifx_led_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    long    len = 0;

    len  = sprintf(buf+len,"Inca-IP2 LED_REG register: 0x%08X\n", 
                    *INCA_IP2_TSF_LED_REG);
    len += sprintf(buf+len,"  Mask of available LEDs: 0x%08X\n", 
                    led_mask);
    len += sprintf(buf+len,"    LEDMUX_CONF register: 0x%08X\n", 
                    *INCA_IP2_TSF_LEDMUX_CONF);

    *eof = 1;   // No more data available
    return len;
}

/**
 * Process LED proc file input.
 * This function evaluates the input of the LED proc file and changes LED 
 * settings accordingly.
 *
 * \param   file    file structure for proc file
 * \param   buffer  buffer holding the data
 * \param   count   number of characters in buffer
 * \param   data    not used
 * \return count    Number of processed characters 
 * \ingroup Internal  
 */
static int ifx_led_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
    char*   ibuf = (char*) buffer;
    u32     val = 0;
    int err = 0;

    if (strncmp(ibuf, "led_set 0x", 10) == 0) {
        printk(KERN_INFO "...decoding led_set command...\n");
        ibuf += 10;
        val = hexstr2uint(ibuf);
        if (val == 0) 
            err = -EINVAL;
        else
            err = ifx_led_on(val);
    }
    else if (strncmp(ibuf, "led_clr 0x", 10) == 0) {
        printk(KERN_INFO "...decoding led_clr command...\n");
        ibuf += 10;
        val = hexstr2uint(ibuf);
        if (val == 0) 
            err = -EINVAL;
        else
            err = ifx_led_off(val);
    }
    else if (strncmp(ibuf, "led_eth 0x", 10) == 0) {
        printk(KERN_INFO "...decoding led_eth command...\n");
        ibuf += 10;
        val = hexstr2uint(ibuf);
        ifx_led_eth_ctrl(LED_ETHPORT_PC, val & 0x1F);
        ifx_led_eth_ctrl(LED_ETHPORT_LAN, (val >> 16) & 0x1F);
        err = 0;
    }
    else {
        printk(KERN_ERR "command syntax unknown!\n");
        err = -EINVAL;
    }
    
    if (err) {
        printk(KERN_ERR "Usage: led_set|led_clr|led_eth 0x<hex_value>\n"
            "Currently available LEDs: 0x%08X (Mask)\n", led_mask); 
    }
    
    printk(KERN_INFO "Inca-IP2 LED_REG register: 0x%08X\n", 
                            *INCA_IP2_TSF_LED_REG);
    return count;
    
}

/**
 * Create keypad proc file output.
 * This function creates the output for the keypad proc file
 *
 * \param   buf         Buffer to write the string to
 * \param   start   not used (Linux internal)
 * \param   offset  not used (Linux internal)
 * \param   count   not used (Linux internal)
 * \param   eof     Set to 1 when all data is stored in buffer
 * \param   data    not used (Linux internal)
 * \return len  Lenght of data in buffer 
 * \ingroup Internal  
 */
static int ifx_keypad_read_procmem(char *buf, char **start, off_t offset, 
                    int count, int *eof, void *data)
{
    long    len = 0;
    int i;
    char *prefix;

    len  = sprintf(buf+len,"key_pins=%d\n",key_pins);
    len  += sprintf(buf+len,"key_major=%d\n\n",key_major);

    for (i = 0; i < KEY_FIFO_DEPTH; i++) {
        if (&key_fifo[i] == rd_ptr && &key_fifo[i] == wr_ptr) 
            prefix = "rw->";
        else if (&key_fifo[i] == rd_ptr) prefix = "rd->";
        else if (&key_fifo[i] == wr_ptr) prefix = "wr->";
        else prefix = "    ";
        len += sprintf(buf+len,"%skeycode[%02d]=%02d, state=%d, "
                    "timestamp=%u\n", prefix, i,
                    key_fifo[i].key_code,
                    key_fifo[i].key_state,
                    (u32) key_fifo[i].time_stamp);
    }
    *eof = 1;   // No more data available
    return len;
}

/**
 * Create TSF proc file output.
 * This function creates the output for the TSF proc file
 *
 * \param   buf         Buffer to write the string to
 * \param   start   not used (Linux internal)
 * \param   offset  not used (Linux internal)
 * \param   count   not used (Linux internal)
 * \param   eof         Set to 1 when all data is stored in buffer
 * \param   data    not used (Linux internal)
 * \return len  Lenght of data in buffer 
 * \ingroup Internal  
 */
int ifx_tsf_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    long    len = 0;

    len = sprintf(buf,"\nINCA-IP2 TSF Settings\n");

    len=len+sprintf(buf+len,"ON\n");
    *eof = 1;   // No more data available

    return len;
}

/***** init *******************************************************************/

/**
 * Initialize TSF unit.
 * This function configures the TSF modules for keypad, ledmatrix and pwm
 * and creates several proc file entries.
 *
 * \ingroup API  
 */

#ifdef CONFIG_FB
extern int panelReadType( void );
#endif

/* this required to to support different levels of WP4 hardware */

int ifx_tsf_init(void)
{
    int err = 0, res;
    unsigned long flags;
    char tmp_string[27];
    
    printk("Inca-IP2 TSF Initialization\n");

    /* Enable TSF module in PMS */
    *INCA_IP2_PMS_PMS_GEN |= 0x00000200;
    /* PWM */

    /* Check module parameters *dev_avail[x] = pwmx */
    if ((pwm1 == 1) && (ifx_mux_pwm1_pins() == 0)) {
        pwm1 = 0;
        printk(KERN_ERR "PWM1 device requested by user but not available!\n");
    }
    if ((pwm2 == 1) && (ifx_mux_pwm2_pins() == 0)) {
        pwm2 = 0;
        printk(KERN_ERR "PWM2 device requested by user but not available!\n");
    }

    printk(KERN_NOTICE "Device availability: %s=%d and %s=%d\n",
         dev_name[0], *dev_avail[0], dev_name[1], *dev_avail[1]);

    if (pwm1 == 1 || pwm2 == 1) {
        
        if ((unsigned) pwm1_value > 255 || (unsigned) pwm2_value > 255) {
            printk(KERN_ERR "PWM value(s) out of range 0..255!\n");
            return -EINVAL;
        }
    
        /* allocate Linux resources */
        if (request_region((u32) INCA_IP2_TSF_PWM12, 0x04, 
                    "INCA-IP2 Pulse Width Modulator") == NULL) {
            printk(KERN_ERR "request_region() failed!\n");
            err = -EBUSY;
            goto FAIL_REQUEST_PWM_REGION;
        }
        
        /* enable the clock for Terminal Specific Functions in the 
                            Power Management Unit */
                            
#ifdef FIXME
// FIXME: use real PMS calls
        save_flags(flags); cli();
        *INCA_IP2_PMU_PM_GEN |= INCA_IP2_PMU_PM_GEN_EN9;   
        restore_flags(flags);
#endif

#ifdef CONFIG_FB
        /* Initialize the PWMs */
#warning PANEL READ TYPE REMOVED WILL NOT WORK
#ifdef CONFIG_FB_EPSON
        if( panelReadType() == 0 )
        {
        	pwm2_value = 0;
        }
#endif
#endif
        ifx_pwm_set(1, pwm1_value);
        ifx_pwm_set(0, pwm2_value);
        
        /* enable the PWM on the INCA-IP2 */
        save_flags(flags); cli();
        *INCA_IP2_TSF_TSF_CONF |= INCA_IP2_TSF_TSF_CONF_PWMEN;
        restore_flags(flags);
        
        /* register PWM devices */
        res = register_chrdev(pwm_major, "pwm", &pwm_fops);
        if (res < 0) {
            printk(KERN_ERR "Could not register PWM device!\n");
            goto FAIL_REGISTER_PWM_DEVICE;
        }
        if (pwm_major == 0) pwm_major = res;
    
        printk(KERN_NOTICE "PWM device registered. major: %d.\n", pwm_major);
        
        SET_MODULE_OWNER(&pwm_fops);
            
        /* create PROC entries */
        if (*dev_avail[0]) {
            sprintf(tmp_string, "driver/%s", dev_name[0]);
            ifx_pwm_proc[0] = create_proc_entry(tmp_string, 0, 0);
            if (!ifx_pwm_proc[0]) {
                printk(KERN_ERR "could not create "
                    "/proc/driver/%s!\n", dev_name[0]);
                err = -ENOENT;
                goto FAIL_CREATE_PWM1_PROC;
            }
            ifx_pwm_proc[0]->read_proc  = ifx_pwm_read_procmem;
            ifx_pwm_proc[0]->write_proc = ifx_pwm_write_procmem;
            ifx_pwm_proc[0]->data = &dev_id[0];
            printk(KERN_NOTICE "/proc/driver/%s created!\n",dev_name[0]);
        }
        if (*dev_avail[1]) {
            sprintf(tmp_string, "driver/%s", dev_name[1]);
            ifx_pwm_proc[1] = create_proc_entry(tmp_string, 0, 0);
            if (!ifx_pwm_proc[1]) {
                printk(KERN_ERR "could not create "
                    "/proc/driver/%s!\n", dev_name[1]);
                err = -ENOENT;
                goto FAIL_CREATE_PWM2_PROC;
            }
            ifx_pwm_proc[1]->read_proc  = ifx_pwm_read_procmem;
            ifx_pwm_proc[1]->write_proc = ifx_pwm_write_procmem;
            ifx_pwm_proc[1]->data = &dev_id[1];
            printk(KERN_NOTICE "/proc/driver/%s created!\n",dev_name[1]);
        }
        
        /* success */
        printk(KERN_NOTICE "INCA-IP2 Pulse Width Modulator initialized\n");
    }
    
    
    /* LED */
    led_pins = ifx_mux_led_pins();
        
    if (led_pins >= 5) {
        led_mask = ~((u32) 0xFFFFFFFF << (4*(led_pins-4)));
    }
    else {  /* too few LED pins requested by user! */
        printk(KERN_ERR "too few LED pins requested by user (min: 5)!\n");
        err=-EINVAL;
        goto FAIL_REQUEST_LED_REGION;
    }
    
    /* allocate Linux resources */
    if (request_region((u32) INCA_IP2_TSF_LEDMUX_CONF, 0x08, 
                    "INCA-IP2 LED Matrix") == NULL) {
        printk(KERN_ERR "request_region() failed!\n");
        res = -EBUSY;
        goto FAIL_REQUEST_LED_REGION;
    }
    
#ifdef FIXME
// FIXME: use real PMS calls
    /* enable the clock of the LED module in the Power Management Unit */
    save_flags(flags); cli();
    *INCA_IP2_PMU_PM_GEN |= INCA_IP2_PMU_PM_GEN_EN9;   
    restore_flags(flags);
#endif

    if (inv == -1) inv = INCAIP2_MUX_LED_INV;
    if (inv) {
        *INCA_IP2_TSF_LEDMUX_CONF |= INCA_IP2_TSF_LEDMUX_CONF_INV;
    } 
    else {
        *INCA_IP2_TSF_LEDMUX_CONF &= ~INCA_IP2_TSF_LEDMUX_CONF_INV;
    }

    /* clear all LEDs */
    *INCA_IP2_TSF_LED_REG = 0x00000000;

    /* define which LEDs are controlled autonomously by Ethernet HW */ 
    ifx_led_eth_ctrl(LED_ETHPORT_LAN, led_eth_lan);
    ifx_led_eth_ctrl(LED_ETHPORT_PC,  led_eth_pc);
    
    if (ncol == -1) ncol = INCAIP2_MUX_LED_NCOL;
    if (ncol == 2) {     
        *INCA_IP2_TSF_LEDMUX_CONF |= INCA_IP2_TSF_LEDMUX_CONF_NCOL;

        /* LEDs L24, L34, L25, L35, ..., L29 & L39 not available! */
        led_mask &= 0x00333333;
    } 
    else {
        *INCA_IP2_TSF_LEDMUX_CONF &= ~INCA_IP2_TSF_LEDMUX_CONF_NCOL;
    }

    // MRH01193 - (WP1/2 only) Configure Multiplexor for 2 columns and not 4 
    // allowing longer pulse with associated increase in brightness
    int myVersion =  (*INCA_IP2_GPIO_PIN >> 12) & 0x0000000F;
    int myType = (myVersion & 0x00000003);
    if (myVersion==4)
    {
        printk(KERN_ERR "tsf: OS15 detected, LED Matrix configured for 4 columns, INCA_IP2_TSF_LEDMUX_CONF is %x\n", *INCA_IP2_TSF_LEDMUX_CONF);
    }
    else
    {
        if (myType < 2)
        {
            printk(KERN_ERR "tsf: WP1/WP2 LED Matrix configured for 2 columns\n");
            *INCA_IP2_TSF_LEDMUX_CONF |= INCA_IP2_TSF_LEDMUX_CONF_NCOL;
        }
        else
        {
            printk(KERN_ERR "tsf: LED Matrix configured for 4 columns, INCA_IP2_TSF_LEDMUX_CONF is %x\n", *INCA_IP2_TSF_LEDMUX_CONF);
        }
    }

    /* enable the LED multiplexer on the board */
    save_flags(flags); cli();

    *INCA_IP2_TSF_TSF_CONF |= INCA_IP2_TSF_TSF_CONF_LEDEN;
    restore_flags(flags);
   

#ifdef CONFIG_DEVFS_FS
    res=devfs_register_chrdev(led_major , "ledmatrix", &led_fops);
    if (led_major==0) led_major=res;
    devfs_handle_ledmatrix=(int) devfs_register(NULL, "ledmatrix", DEVFS_FL_NONE, led_major, 0, S_IFCHR | S_IRUGO | S_IWUGO, &led_fops, NULL);
    if (devfs_handle_ledmatrix == 0) {
        err=0;
        printk(KERN_ERR "Could not register ledmatrix device!\n");
        goto FAIL_REGISTER_LED_DEVICE;
    }
#else
    /* register ledmatrix device */
    res = register_chrdev(led_major, "ledmatrix", &led_fops);
    if (res < 0)  {
        printk(KERN_ERR "Could not register ledmatrix device!\n");
        err=res;
        goto FAIL_REGISTER_LED_DEVICE;
    }
    if (led_major == 0) led_major = res;
#endif

    
    SET_MODULE_OWNER(&led_fops);
    
    printk(KERN_NOTICE "Ledmatrix device registered. major: %d.\n", led_major);
    
    
    /* create PROC entry */
    ifx_led_proc = create_proc_entry( "driver/ledmatrix", 0, 0);
    if (!ifx_led_proc) {
        printk(KERN_ERR "could not create /proc/driver/ledmatrix!\n");
        err = -ENOENT;
        goto FAIL_CREATE_LED_PROC;
    }
    ifx_led_proc->read_proc  = ifx_led_read_procmem;
    ifx_led_proc->write_proc = ifx_led_write_procmem;

    printk(KERN_NOTICE "/proc/driver/ledmatrix created!\n");
    
    /* success */
    printk(KERN_NOTICE "Inca-IP2 LED Matrix initialized with pins "
                    "LED[%d:0].\n", led_pins-1);

    /* KEY */
    if (key_pins == -1) key_pins = ifx_mux_key_pins();
    
    if (key_pins < 2) { /* too few KEY pins requested by user! */
        printk(KERN_ERR "too few KEY pins requested by user "
                            "(min: 2)!\n");
        err = -EINVAL;
        goto FAIL_REQUEST_KEY_REGION;
    }
    
    /* allocate Linux resources */
    if (request_region((u32) INCA_IP2_TSF_KEY_CONF, 0x14, 
                    "INCA-IP2 Keypad") == NULL) {
        printk(KERN_ERR "request_region() failed!\n");
        err = -EBUSY;
        goto FAIL_REQUEST_KEY_REGION;
    }
    
#ifdef FIXME
// FIXME: use real PMS calls
    /* enable clock of the TSF (LED, Key, PWM) in Power Management Unit */
    save_flags(flags); cli();
    *INCA_IP2_PMU_PM_GEN |= INCA_IP2_PMU_PM_GEN_EN9;   
    restore_flags(flags);
#endif   


    /* enable the keypad */
    save_flags(flags); cli();
    *INCA_IP2_TSF_TSF_CONF |= INCA_IP2_TSF_TSF_CONF_KEYEN;
//  *INCA_IP_Ports_P2_PUDSEL |= 0x6000; /* pull-up characteristic */
//  *INCA_IP_Ports_P2_PUDEN  |= 0x6000; /* enable pull(-up) transistor */
    restore_flags(flags);
    *INCA_IP2_TSF_KEY_CONF = key_pins -1; /* scan KEY0-7 */
    
#ifdef CONFIG_DEVFS_FS
    res=devfs_register_chrdev(key_major , "keypad", &keypad_fops);
    if (key_major==0) key_major=res;
    devfs_handle_ledmatrix=(int) devfs_register(NULL, "keypad", DEVFS_FL_NONE, key_major, 0, S_IFCHR | S_IRUGO | S_IWUGO, &keypad_fops, NULL);
    if (devfs_handle_ledmatrix == 0) {
        err=0;
        printk(KERN_ERR "Could not register keypad device!\n");
        goto FAIL_REGISTER_LED_DEVICE;
    }
#else
    /* register keypad device */
    res = register_chrdev(key_major, "keypad", &keypad_fops);
    if (res < 0) {
        printk(KERN_ERR "Could not register keypad device!\n");
        err=res;
        goto FAIL_REGISTER_KEY_DEVICE;
    }
    if (key_major == 0) key_major = res;
#endif
    
    SET_MODULE_OWNER(&keypad_fops);
    
    printk(KERN_NOTICE "Keypad device registered. major: %d.\n", key_major);
    
    /* create PROC entry */
    ifx_keypad_proc = create_proc_entry( "driver/keypad", 0, 0);
    if (!ifx_keypad_proc) {
        printk(KERN_ERR "could not create /proc/driver/keypad!\n");
        err = -ENOENT;
        goto FAIL_CREATE_KEY_PROC;
    }
    ifx_keypad_proc->read_proc  = ifx_keypad_read_procmem;

    printk(KERN_NOTICE "/proc/driver/keypad created!\n");

    
    printk(KERN_NOTICE "Inca-IP2 keypad initialized with pins KEY[%d:0].\n", 
                                key_pins-1);

    /* TSF */
    create_proc_read_entry("driver/ifx_tsf", 0, NULL, ifx_tsf_read_procmem, NULL);

    return 0; /* success */


    /* TSF Error handling */
    /* if initialization fails, release all resources aquired so far! */

FAIL_CREATE_KEY_PROC:
    unregister_chrdev(key_major, "keypad");
FAIL_REGISTER_KEY_DEVICE:
    release_region((u32) INCA_IP2_TSF_KEY_CONF, 0x14);
FAIL_REQUEST_KEY_REGION:

remove_proc_entry("driver/ledmatrix", 0);
FAIL_CREATE_LED_PROC:
    unregister_chrdev(led_major, "ledmatrix");
FAIL_REGISTER_LED_DEVICE:
    release_region((u32) INCA_IP2_TSF_LEDMUX_CONF, 0x08);
FAIL_REQUEST_LED_REGION:

    if (*dev_avail[1]) {
        sprintf(tmp_string, "driver/%s", dev_name[1]);
        remove_proc_entry(tmp_string, 0);
    }
FAIL_CREATE_PWM2_PROC:
    if (*dev_avail[0]) {
        sprintf(tmp_string, "driver/%s", dev_name[0]);
        remove_proc_entry(tmp_string, 0);
    }
FAIL_CREATE_PWM1_PROC:
    unregister_chrdev(pwm_major, "pwm");
FAIL_REGISTER_PWM_DEVICE:
            
#ifdef FIXME    
// FIXME: use real PMS calls
    save_flags(flags); cli();
    *INCA_IP2_TSF_TSF_CONF    &= ~INCA_IP2_TSF_TSF_CONF_PWMEN;
    restore_flags(flags);
#endif

    ifx_pwm_set(0, 0);
    ifx_pwm_set(1, 0);
    
    release_region((u32) INCA_IP2_TSF_PWM12, 0x04);
    
FAIL_REQUEST_PWM_REGION:
    return err;

}

/**
 * Remove TSF module.
 * Upon removal of the TSF module this function will free all allocated resources
 * and unregister devices.
 *
 * \ingroup API  
 */
static void __exit ifx_tsf_exit(void)
{
    unsigned long flags;
    char tmp_string[27];
    
    remove_proc_entry("driver/ifx_tsf", 0);
    
#ifdef CONFIG_DEVFS_FS
    devfs_unregister(devfs_handle_keypad);
#else
    unregister_chrdev(key_major, "keypad");
#endif    
    release_region((u32) INCA_IP2_TSF_KEY_CONF, 0x14);
    remove_proc_entry("driver/ledmatrix", 0);
#ifdef CONFIG_DEVFS_FS
    devfs_unregister(devfs_handle_ledmatrix);
#else
    unregister_chrdev(led_major, "ledmatrix");
#endif    
    release_region((u32) INCA_IP2_TSF_LEDMUX_CONF, 0x08);
    if (*dev_avail[1]) {
        sprintf(tmp_string, "driver/%s", dev_name[1]);
        remove_proc_entry(tmp_string, 0);
    }
    if (*dev_avail[0]) {
        sprintf(tmp_string, "driver/%s", dev_name[0]);
        remove_proc_entry(tmp_string, 0);
    }
    unregister_chrdev(pwm_major, "pwm");
            
#ifdef FIXME    
// FIXME: use real PMS calls
    save_flags(flags); cli();
    *INCA_IP2_TSF_TSF_CONF    &= ~INCA_IP2_TSF_TSF_CONF_PWMEN;
    restore_flags(flags);
#endif

    ifx_pwm_set(0, 0);
    ifx_pwm_set(1, 0);
    release_region((u32) INCA_IP2_TSF_PWM12, 0x04);
}

module_init(ifx_tsf_init);
module_exit(ifx_tsf_exit);

EXPORT_SYMBOL(ifx_key_register_callback);
EXPORT_SYMBOL(ifx_key_unregister_callback);
EXPORT_SYMBOL(ifx_keypad_open);
EXPORT_SYMBOL(ifx_keypad_release);
EXPORT_SYMBOL(ifx_keypad_ioctl);
EXPORT_SYMBOL(ifx_pwm_set);
EXPORT_SYMBOL(ifx_pwm_get);
EXPORT_SYMBOL(ifx_led_off);
EXPORT_SYMBOL(ifx_led_on);
EXPORT_SYMBOL(ifx_led_eth_ctrl);

