/************************************************************************
 *
 * Copyright (c) 2004
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/*
 *  linux/drivers/char/serial_incaipasc.c
 *
 *  Driver for INCAIPASC serial ports
 *
 *  Initial Version
 *  Copyright (C) 2002 Gary Jennejohn (gj@denx.de)
 *
 *  Based on drivers/char/serial.c, by Linus Torvalds, Theodore Ts'o.
 *  Based on drivers/serial/serial_s3c2400.c
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/circ_buf.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/irq.h>
#include <asm/incaip/serial.h>

#if defined(CONFIG_SERIAL_INCAIPASC_CONSOLE) && defined(CONFIG_MAGIC_SYSRQ)
#define SUPPORT_SYSRQ
#endif


#include <linux/serial_core.h>

#define UART_NR		1

#define SERIAL_INCAIPASC_MAJOR	TTY_MAJOR
#define CALLOUT_INCAIPASC_MAJOR	TTYAUX_MAJOR
#define SERIAL_INCAIPASC_MINOR	64
#define SERIAL_INCAIPASC_NR	UART_NR

static struct tty_driver normal, callout;
static struct tty_struct *incaipasc_table[UART_NR];
static struct termios *incaipasc_termios[UART_NR], *incaipasc_termios_locked[UART_NR];
static struct uart_port incaipasc_ports[UART_NR];
static struct uart_info *incaipasc_info;
#ifdef CONFIG_SERIAL_INCAIPASC_CONSOLE /*SUPPORT_SYSRQ*/
static struct console incaipasc_console;
#endif
static unsigned int uartclk = 0;

int rx_bytes_count=0;
int tx_bytes_count=0;
int rx_parity_error_count=0;
int rx_frame_error_count=0;
int rx_overrun_error_count=0;

#define SET_BIT(reg, mask)    *(reg) |= (mask)
#define CLEAR_BIT(reg, mask)  *(reg) &= (~(mask)) 
#define CLEAR_BITS(reg, mask) CLEAR_BIT(reg, mask)
#define SET_BITS(reg, mask)   SET_BIT(reg, mask)
#define SET_BITFIELD(reg, mask, off, val) \
				{*(reg) &= (~(mask)); *(reg) |= ((val) << (off));}

/* forward declaration to shut the compiler up */
static void incaipasc_rx_int(int, void *, struct pt_regs *);
static void incaipasc_tx_chars(struct uart_info *info);
static void incaip_serial_out(char ch);

/* fake flag to indicate CREAD was not set -> throw away all bytes */
#define UART_DUMMY_UER_RX 1

#define ENABLE_IRQ(n) enable_irq((n))
#define DISABLE_IRQ(n) disable_irq((n))

/* TX interrupt */
#define INCAIPASC_TIR INT_NUM_IM2_IRL16
/* RX interrupt */
#define INCAIPASC_RIR INT_NUM_IM2_IRL17
/* ERROR interrupt */
#define INCAIPASC_EIR INT_NUM_IM2_IRL18

/* macro to set the bit corresponding to an interrupt number */
#define BIT_NO(irq) (1 << ((irq) - 64))

#if (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 1200)
#define INCAIPASC_DEFAULT_BAUDRATE  B1200
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 2400)
#define INCAIPASC_DEFAULT_BAUDRATE  B2400
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 4800)
#define INCAIPASC_DEFAULT_BAUDRATE  B4800
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 9600)
#define INCAIPASC_DEFAULT_BAUDRATE  B9600
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 19200)
#define INCAIPASC_DEFAULT_BAUDRATE  B19200
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 38400)
#define INCAIPASC_DEFAULT_BAUDRATE  B38400
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 57600)
#define INCAIPASC_DEFAULT_BAUDRATE  B57600
#elif (CONFIG_INCAIPASC_DEFAULT_BAUDRATE == 115200)
#define INCAIPASC_DEFAULT_BAUDRATE  B115200
#else
#error CONFIG_INCAIPASC_DEFAULT_BAUDRATE not set to standard baudrate!
#endif


static u_int incaipasc_tx_empty(struct uart_port *port);

extern unsigned int incaip_get_fpi_hz(void);

static void incaipasc_stop_tx(struct uart_port *port, u_int from_tty)
{
	DISABLE_IRQ(INCAIPASC_TIR);
}

static void incaipasc_start_tx(struct uart_port *port, u_int nonempty, u_int from_tty)
{
	ENABLE_IRQ(INCAIPASC_TIR);
	if (incaipasc_info)
		incaipasc_tx_chars(incaipasc_info);
}

static void incaipasc_stop_rx(struct uart_port *port)
{
	/* clear the RX enable bit */
	*INCA_IP_ASC_ASC_WHBCON = ASCWHBCON_CLRREN;
}

static void incaipasc_enable_ms(struct uart_port *port)
{
	/* no modem signals */
	return;
}

static void
#ifdef SUPPORT_SYSRQ
incaipasc_rx_chars(struct uart_info *info, struct pt_regs *regs)
#else
incaipasc_rx_chars(struct uart_info *info)
#endif
{
	struct tty_struct *tty = info->tty;
	unsigned int ch = 0, rsr = 0, fifocnt;
	struct uart_port *port = info->port;

	fifocnt = (*INCA_IP_ASC_FSTAT) & ASCFSTAT_RXFFLMASK;
	while (fifocnt--)
	{
		ch = *INCA_IP_ASC_ASC_RBUF;
		rsr = (*INCA_IP_ASC_ASC_CON & ASCCON_ANY) | UART_DUMMY_UER_RX;

		if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
			tty->flip.tqueue.routine((void *)tty);
			if (tty->flip.count >= TTY_FLIPBUF_SIZE) {
				printk(KERN_WARNING "TTY_DONT_FLIP set\n");
				return;
			}
		}

		*tty->flip.char_buf_ptr = ch;
		*tty->flip.flag_buf_ptr = TTY_NORMAL;
		port->icount.rx++;
		rx_bytes_count++;
		
		/*
		 * Note that the error handling code is
		 * out of the main execution path
		 */
		if (rsr & ASCCON_ANY) {
			if (rsr & ASCCON_PE) {
				port->icount.parity++;
				rx_parity_error_count++;
				SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_CLRPE);
			} else if (rsr & ASCCON_FE) {
				port->icount.frame++;
				rx_frame_error_count++;
				SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_CLRFE);
			}
			if (rsr & ASCCON_OE) {
				port->icount.overrun++;
				rx_overrun_error_count++;
				SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_CLROE);
			}

			rsr &= port->read_status_mask;

			if (rsr & ASCCON_PE)
				*tty->flip.flag_buf_ptr = TTY_PARITY;
			else if (rsr & ASCCON_FE)
				*tty->flip.flag_buf_ptr = TTY_FRAME;
		}

#ifdef SUPPORT_SYSRQ
		if (uart_handle_sysrq_char(info, ch, regs))
			continue;
#endif

		if ((rsr & port->ignore_status_mask) == 0) {
			tty->flip.flag_buf_ptr++;
			tty->flip.char_buf_ptr++;
			tty->flip.count++;
		}
		if ((rsr & ASCCON_OE) &&
		    tty->flip.count < TTY_FLIPBUF_SIZE) {
			/*
			 * Overrun is special, since it's reported
			 * immediately, and doesn't affect the current
			 * character
			 */
			*tty->flip.char_buf_ptr++ = 0;
			*tty->flip.flag_buf_ptr++ = TTY_OVERRUN;
			tty->flip.count++;
		}
	}
	if (ch != 0)
		tty_flip_buffer_push(tty);
	return;
}

static void incaipasc_tx_chars(struct uart_info *info)
{
	struct uart_port *port = info->port;

	if (info->xmit.head == info->xmit.tail
	    || info->tty->stopped
	    || info->tty->hw_stopped) {
		incaipasc_stop_tx(port, 0);
		return;
	}

	if (port->x_char) {
		incaip_serial_out(port->x_char);
		port->icount.tx++;
		port->x_char = 0;
		tx_bytes_count++;
		return;
	}
	incaip_serial_out(info->xmit.buf[info->xmit.tail]);
	info->xmit.tail = (info->xmit.tail + 1) & (UART_XMIT_SIZE - 1);
	port->icount.tx++;
	tx_bytes_count++;

	if (CIRC_CNT(info->xmit.head, info->xmit.tail, UART_XMIT_SIZE) <
			WAKEUP_CHARS)
		uart_event(info, EVT_WRITE_WAKEUP);
	if (info->xmit.head == info->xmit.tail)
	{
		incaipasc_stop_tx(info->port, 0);
	}
}

static void incaipasc_tx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_info *info = dev_id;
	
	/* this interrupt tells us that the TXB is empty - fill it */
	incaipasc_tx_chars(info);
	/* clear any pending interrupts */
	*INCA_IP_ICU_IM2_ISR = BIT_NO(INCAIPASC_TIR);
}

static void incaipasc_er_int(int irq, void *dev_id, struct pt_regs *regs)
{
	
	/* XXX */
	/* clear any pending interrupts */
	*INCA_IP_ICU_IM2_ISR = BIT_NO(INCAIPASC_EIR);
	SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_CLRPE);
	SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_CLRFE);
	SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_CLROE);
	return;
}

static void incaipasc_rx_int(int irq, void *dev_id, struct pt_regs *regs)
{
	struct uart_info *info = dev_id;

#ifdef SUPPORT_SYSRQ
	incaipasc_rx_chars(info, regs);
#else
	incaipasc_rx_chars(info);
#endif
	/* clear any pending interrupts */
	*INCA_IP_ICU_IM2_ISR = BIT_NO(INCAIPASC_RIR);
}

static u_int incaipasc_tx_empty(struct uart_port *port)
{
	int status;

	/*
	 * FSTAT tells exactly how many bytes are in the FIFO.
	 * The question is whether we really need to wait for all
	 * 16 bytes to be transmitted before reporting that the
	 * transmitter is empty.
	 */
	status = *INCA_IP_ASC_FSTAT & ASCFSTAT_TXFFLMASK;
	return status ? 0 : TIOCSER_TEMT;
}

static u_int incaipasc_get_mctrl(struct uart_port *port)
{
	/* no modem control signals - the readme says to pretend all are set */
	return TIOCM_CTS|TIOCM_CAR|TIOCM_DSR;
}

static void incaipasc_set_mctrl(struct uart_port *port, u_int mctrl)
{
	/* no modem control - just return */
	return;
}

static void incaipasc_break_ctl(struct uart_port *port, int break_state)
{
	/* no way to send a break */
	return;
}

static int incaipasc_startup(struct uart_port *port, struct uart_info *info)
{
	int retval;
	u_int con = 0;
	unsigned long flags;

	/* this assumes: CON.BRS = CON.FDE = 0 */
	if (uartclk == 0)
		uartclk = incaip_get_fpi_hz()/2;

	incaipasc_ports[0].uartclk = uartclk;
	incaipasc_info = info;

	/* block the IRQs */
	local_irq_save(flags);

	/* this setup was probably already done in u-boot */
	/* enable the ASC in the PMU */
	*INCA_IP_PMU_PM_GEN |= INCA_IP_PMU_PM_GEN_EN13;
	
	/* set up the CLC */
	CLEAR_BIT(INCA_IP_ASC_ASC_CLC, INCA_IP_ASC_ASC_CLC_DISS);
	SET_BITFIELD(INCA_IP_ASC_ASC_CLC, ASCCLC_RMCMASK, ASCCLC_RMCOFFSET, 1);
	/* asynchronous mode */
	con = ASCCON_M_8ASYNC;
	/* set error signals  - framing and overrun */
	con |= ASCCON_FEN;
	con |= ASCCON_OEN;
	con |= ASCCON_PEN;
	/* choose the line - there's only one */
	*INCA_IP_ASC_ASC_PISEL = 0;
#if 1
	*INCA_IP_ASC_TXFCON = (((INCAASC_TXFIFO_FL<<ASCTXFCON_TXFITLOFF)&ASCTXFCON_TXFITLMASK) | ASCTXFCON_TXFEN |ASCTXFCON_TXFFLU);
	*INCA_IP_ASC_RXFCON = (((INCAASC_RXFIFO_FL<<ASCRXFCON_RXFITLOFF)&ASCRXFCON_RXFITLMASK) | ASCRXFCON_RXFEN |ASCRXFCON_RXFFLU);
	wmb();
#else
	/* TXFIFO's fill level */
	SET_BITFIELD(INCA_IP_ASC_TXFCON, ASCTXFCON_TXFITLMASK,
		ASCTXFCON_TXFITLOFF, INCAASC_TXFIFO_FL);
	/* enable TXFIFO */
	SET_BIT(INCA_IP_ASC_TXFCON, ASCTXFCON_TXFEN);
	/* RXFIFO's fill level */
	SET_BITFIELD(INCA_IP_ASC_RXFCON, ASCRXFCON_RXFITLMASK,
		ASCRXFCON_RXFITLOFF, INCAASC_RXFIFO_FL);
	/* enable RXFIFO */
	SET_BIT(INCA_IP_ASC_RXFCON, ASCRXFCON_RXFEN);
	/* now really set CON */
#endif
	*INCA_IP_ASC_ASC_CON = con;

	/*
	 * Allocate the IRQs
	 */
	retval = request_irq(INCAIPASC_RIR, incaipasc_rx_int, 0, "asc_rx", info);
	if (retval)
		return retval;
	retval = request_irq(INCAIPASC_TIR, incaipasc_tx_int, 0, "asc_tx", info);
	if (retval)
	{
		free_irq(INCAIPASC_RIR, info);
		return retval;
	}
	retval = request_irq(INCAIPASC_EIR, incaipasc_er_int, 0, "asc_er", info);
	if (retval)
	{
		free_irq(INCAIPASC_RIR, info);
		free_irq(INCAIPASC_TIR, info);
		return retval;
	}
	/* unblock the IRQs */
	local_irq_restore(flags);

	return 0;
}

static void incaipasc_shutdown(struct uart_port *port, struct uart_info *info)
{
	/*
	 * Free the interrupts
	 */
	free_irq(INCAIPASC_RIR, info);
	free_irq(INCAIPASC_TIR, info);
	free_irq(INCAIPASC_EIR, info);
	/*
	 * disable the baudrate generator to disable the ASC
	 */
	*INCA_IP_ASC_ASC_CON = 0;

	/* flush and then disable the fifos */
	SET_BIT(INCA_IP_ASC_RXFCON, ASCRXFCON_RXFFLU);
	CLEAR_BIT(INCA_IP_ASC_RXFCON, ASCRXFCON_RXFEN);
	SET_BIT(INCA_IP_ASC_TXFCON, ASCTXFCON_TXFFLU);
	CLEAR_BIT(INCA_IP_ASC_TXFCON, ASCTXFCON_TXFEN);
}

static void incaipasc_change_speed(struct uart_port *port, u_int cflag, u_int iflag, u_int quot)
{
	u_int con = 0;
	unsigned long flags;
	u32 temp_reg=0;
	int t=0;

	/* block the IRQs */
	save_and_cli(flags);

	/* Wait for FIFO to be empty */
	while (((*INCA_IP_ASC_FSTAT)&0x0000FF00)>0) {
		asm("nop;nop;nop;nop;nop;");
	}
	temp_reg=*INCA_IP_ASC_TXFCON;
	/* Wait for characters to be sent out */
	for (t=0;t<60000;t++) {
		asm("nop;nop;nop;nop;nop;");
	}

	/* byte size and parity */
	switch (cflag & CSIZE) {
	/* 7 bits are always with parity */
	case CS7: con = ASCCON_M_7ASYNCPAR; break;
	/* the ASC only suports 7 and 8 bits */
	case CS5:
	case CS6:
	default:
		if (cflag & PARENB)
			con = ASCCON_M_8ASYNCPAR;
		else
			con = ASCCON_M_8ASYNC;
		break;
	}
	if (cflag & CSTOPB)
		con |= ASCCON_STP;
	if (cflag & PARENB) {
		if (!(cflag & PARODD))
			con &= ~ASCCON_ODD;
		else
			con |= ASCCON_ODD;
	}

	port->read_status_mask = ASCCON_OE;
	if (iflag & INPCK)
		port->read_status_mask |= ASCCON_FE | ASCCON_PE;
	/* the ASC can't really detect or generate a BREAK */
#if 0
	if (iflag & (BRKINT | PARMRK))
		port->read_status_mask |= UERSTAT_BREAK;
#endif

	/*
	 * Characters to ignore
	 */
	port->ignore_status_mask = 0;
	if (iflag & IGNPAR)
		port->ignore_status_mask |= ASCCON_FE | ASCCON_PE;
#if 0
	/* always ignore breaks - the ASC can't handle them XXXX */
	port->ignore_status_mask |= UERSTAT_BREAK;
#endif
	if (iflag & IGNBRK) {
		/*port->ignore_status_mask |= UERSTAT_BREAK;*/
		/*
		 * If we're ignoring parity and break indicators,
		 * ignore overruns too (for real raw support).
		 */
		if (iflag & IGNPAR)
			port->ignore_status_mask |= ASCCON_OE;
	}

	/*
	 * Ignore all characters if CREAD is not set.
	 */
	if ((cflag & CREAD) == 0)
		port->ignore_status_mask |= UART_DUMMY_UER_RX;

	/* set error signals  - framing, parity  and overrun */
	con |= ASCCON_FEN;
	con |= ASCCON_OEN;
	con |= ASCCON_PEN;

	/* disable the baudrate generator */
	CLEAR_BIT(INCA_IP_ASC_ASC_CON, ASCCON_R);
	temp_reg=*INCA_IP_ASC_ASC_CON;
	
	/* set up CON */
	*INCA_IP_ASC_ASC_CON = con;
	/* Set baud rate - take a divider of 2 into account */
	quot = quot/2 - 1;
	/* the next 3 probably already happened when we set CON above */
	/* make sure the fractional divider is off */
	CLEAR_BIT(INCA_IP_ASC_ASC_CON, ASCCON_FDE);
	/* set up to use divisor of 2 */
	CLEAR_BIT(INCA_IP_ASC_ASC_CON, ASCCON_BRS);
	/* now we can write the new baudrate into the register */
	*INCA_IP_ASC_ASC_BTR = quot;
	temp_reg=*INCA_IP_ASC_ASC_BTR;

	/* turn the baudrate generator back on */
	SET_BIT(INCA_IP_ASC_ASC_CON, ASCCON_R);
	temp_reg=*INCA_IP_ASC_ASC_CON;
	/* Enable receiver */
	SET_BIT(INCA_IP_ASC_ASC_WHBCON, ASCWHBCON_SETREN);
	temp_reg=*INCA_IP_ASC_ASC_CON;
	/* Wait for changes to take effect */
	for (t=0;t<3000;t++) {
		asm("nop;nop;nop;nop;nop;");
	}
	/* unblock the IRQs */
	restore_flags(flags);
}

static const char *incaipasc_type(struct uart_port *port)
{
	return port->type == PORT_INCAIPASC ? "INCAIPASC" : NULL;
}

/*
 * Release the memory region(s) being used by 'port'
 */
static void incaipasc_release_port(struct uart_port *port)
{
	return;
}

/*
 * Request the memory region(s) being used by 'port'
 */
static int incaipasc_request_port(struct uart_port *port)
{
	return 0;
}

/*
 * Configure/autoconfigure the port.
 */
static void incaipasc_config_port(struct uart_port *port, int flags)
{
	if (flags & UART_CONFIG_TYPE) {
		port->type = PORT_INCAIPASC;
		incaipasc_request_port(port);
	}
}

/*
 * verify the new serial_struct (for TIOCSSERIAL).
 */
static int incaipasc_verify_port(struct uart_port *port, struct serial_struct *ser)
{
	int ret = 0;
	if (ser->type != PORT_UNKNOWN && ser->type != PORT_INCAIPASC)
		ret = -EINVAL;
	if (ser->irq < 0 || ser->irq >= NR_IRQS)
		ret = -EINVAL;
	if (ser->baud_base < 9600)
		ret = -EINVAL;
	return ret;
}

static struct uart_ops incaipasc_pops = {
	tx_empty:	incaipasc_tx_empty,
	set_mctrl:	incaipasc_set_mctrl,
	get_mctrl:	incaipasc_get_mctrl,
	stop_tx:	incaipasc_stop_tx,
	start_tx:	incaipasc_start_tx,
	stop_rx:	incaipasc_stop_rx,
	enable_ms:	incaipasc_enable_ms,
	break_ctl:	incaipasc_break_ctl,
	startup:	incaipasc_startup,
	shutdown:	incaipasc_shutdown,
	change_speed:	incaipasc_change_speed,
	type:		incaipasc_type,
	release_port:	incaipasc_release_port,
	request_port:	incaipasc_request_port,
	config_port:	incaipasc_config_port,
	verify_port:	incaipasc_verify_port,
};

static struct uart_port incaipasc_ports[UART_NR] = {
	{
		membase:	(void *)INCA_IP_ASC,
		mapbase:	INCA_IP_ASC,
		iotype:		SERIAL_IO_MEM,
		irq:		INT_NUM_IM2_IRL17, /* RIR */
		uartclk:	0, /* filled in dynamically */
		fifosize:	16,
		unused:		{ INT_NUM_IM2_IRL16, INT_NUM_IM2_IRL18, INT_NUM_IM2_IRL19 }, /* xmit/error/xmit-buffer-empty IRQ */
		type:		PORT_INCAIPASC,
		ops:		&incaipasc_pops,
		flags:		ASYNC_BOOT_AUTOCONF,
	},
};

#ifdef CONFIG_SERIAL_INCAIPASC_CONSOLE
#ifdef used_and_not_const_char_pointer
static int incaipasc_console_read(struct uart_port *port, char *s, u_int count)
{
	unsigned int status;
	int fifocnt;
	int c;
	unsigned long flags;
#if DEBUG
	printk("incaipasc_console_read() called\n");
#endif

	c = 0;
	/* block the IRQ */
	local_irq_save(flags);

	fifocnt = *INCA_IP_ASC_FSTAT & ASCFSTAT_RXFFLMASK;
	while (c < count && fifocnt) {
		*s++ = INCA_IP_ASC_ASC_RBUF;
		c++;
		fifocnt--;
	}
	/* clear any pending interrupts */
	*INCA_IP_ICU_IM2_ISR = BIT_NO(INCAIPASC_RIR);
	/* unblock the IRQ */
	local_irq_restore(flags);
	/* return the count */
	return c;
}
#endif

static void incaip_serial_out(char ch)
{
	int fifocnt;
	unsigned long flags;
	
	save_and_cli(flags);
	
	do
	{
		fifocnt = (*INCA_IP_ASC_FSTAT & ASCFSTAT_TXFFLMASK)
				>> ASCFSTAT_TXFFLOFF;
	} while (fifocnt == INCAASC_TXFIFO_FULL);
	*INCA_IP_ASC_ASC_TBUF = ch;
	restore_flags(flags);
}

static void incaipasc_console_write(struct console *co, const char *s, u_int count)
{
	int i;
	char scr;
	unsigned long flags;
#ifdef EARLY_PRINTK_HACK
	extern int do_early_printk;

	if (do_early_printk == 1)
		do_early_printk = 0;
#endif
	/* block the IRQ */
	local_irq_save(flags);
	/*
	 *	Now, do each character
	 */
	scr = '\0';
	for (i = 0; i < count;i++)
	{
		if (s[i] == '\n')
			incaip_serial_out('\r');
		incaip_serial_out(s[i]);

	} /* for */

	/* clear any pending interrupts */
	*INCA_IP_ICU_IM2_ISR = BIT_NO(INCAIPASC_TIR);
	/* restore the IRQ */
	local_irq_restore(flags);
}

static kdev_t incaipasc_console_device(struct console *co)
{
	return mk_kdev(SERIAL_INCAIPASC_MAJOR,
		SERIAL_INCAIPASC_MINOR + co->index);
}

static void __init
incaipasc_console_get_options(struct uart_port *port, int *baud, int *parity, int *bits)
{
	u_int lcr_h;

	lcr_h = *INCA_IP_ASC_ASC_CON;
	/* do this only if the ASC is turned on */
	if (lcr_h & ASCCON_R) {
		u_int quot, div, fdiv, frac;

		*parity = 'n';
		if ((lcr_h & ASCCON_MODEMASK) == ASCCON_M_7ASYNCPAR ||
		            (lcr_h & ASCCON_MODEMASK) == ASCCON_M_8ASYNCPAR) {
			if (lcr_h & ASCCON_ODD)
				*parity = 'o';
			else
				*parity = 'e';
		}

		if ((lcr_h & ASCCON_MODEMASK) == ASCCON_M_7ASYNCPAR)
			*bits = 7;
		else
			*bits = 8;

		quot = *INCA_IP_ASC_ASC_BTR + 1;
		/* this gets hairy if the fractional divider is used */
		if (lcr_h & ASCCON_FDE)
		{
			div = 1;
			fdiv = *INCA_IP_ASC_ASC_FDV;
			if (fdiv == 0)
				fdiv = 512;
			frac = 512;
		}
		else
		{
			div = lcr_h & ASCCON_BRS ? 3 : 2;
			fdiv = frac = 1;
		}
		/*
		 * This doesn't work exactly because we use integer
		 * math to calculate baud which results in rounding
		 * errors when we try to go from quot -> baud !!
		 * Try to make this work for both the fractional divider
		 * and the simple divider. Also try to avoid rounding
		 * errors using integer math.
		 */
		*baud = frac * (port->uartclk / (div * 512 * 16 * quot));
		if (*baud > 1100 && *baud < 2400)
			*baud = 1200;
		if (*baud > 2300 && *baud < 4800)
			*baud = 2400;
		if (*baud > 4700 && *baud < 9600)
			*baud = 4800;
		if (*baud > 9500 && *baud < 19200)
			*baud = 9600;
		if (*baud > 19000 && *baud < 38400)
			*baud = 19200;
		if (*baud > 38400 && *baud < 57600)
			*baud = 38400;
		if (*baud > 57600 && *baud < 115200)
			*baud = 57600;
		if (*baud > 115200 && *baud < 230400)
			*baud = 115200;
	}
}

static int __init incaipasc_console_setup(struct console *co, char *options)
{
	struct uart_port *port;
	int baud = CONFIG_INCAIPASC_DEFAULT_BAUDRATE;
	int bits = 8;
	int parity = 'n';
	int flow = 'n';

	/* this assumes: CON.BRS = CON.FDE = 0 */
	if (uartclk == 0)
		uartclk = incaip_get_fpi_hz();

	incaipasc_ports[0].uartclk = uartclk;
	incaipasc_ports[0].type = PORT_INCAIPASC;

	/*
	 * Check whether an invalid uart number has been specified, and
	 * if so, search for the first available port that does have
	 * console support.
	 */
	port = uart_get_console(incaipasc_ports, UART_NR, co);

	if (options)
		uart_parse_options(options, &baud, &parity, &bits, &flow);
	else
		incaipasc_console_get_options(port, &baud, &parity, &bits);

	return uart_set_options(port, co, baud, parity, bits, flow);
}

static struct console incaipasc_console = {
	name:		"ttyS",
	write:		incaipasc_console_write,
#ifdef used_and_not_const_char_pointer
	read:		incaipasc_console_read,
#endif
	device:		incaipasc_console_device,
	setup:		incaipasc_console_setup,
	flags:		CON_PRINTBUFFER,
	index:		0,
};

void __init incaipasc_console_init(void)
{
	register_console(&incaipasc_console);
}

int incaipasc_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	long	len = 0;

	len = sprintf(buf,"\nInca-IP ASC Debug\n");

	len=len+sprintf(buf+len,"rx_bytes_count=%d\n",rx_bytes_count);
	len=len+sprintf(buf+len,"\ntx_bytes_count=%d\n",tx_bytes_count);
	len=len+sprintf(buf+len,"\nrx_overrun_error_count=%d\n",rx_overrun_error_count);
	len=len+sprintf(buf+len,"rx_parity_error_count=%d\n",rx_parity_error_count);
	len=len+sprintf(buf+len,"rx_frame_error_count=%d\n",rx_frame_error_count);

	len=len+sprintf(buf+len,"\n");

	*eof = 1;	// No more data available
	return len;
}

#define INCAIPASC_CONSOLE	&incaipasc_console
#else
#define INCAIPASC_CONSOLE	NULL
#endif

static struct uart_driver incaipasc_reg = {
	owner:			NULL,
	normal_major:		SERIAL_INCAIPASC_MAJOR,
#ifdef CONFIG_DEVFS_FS
	normal_name:		"ttyS%d",
	callout_name:		"cua%d",
#else
	normal_name:		"ttyS",
	callout_name:		"cua",
#endif
	normal_driver:		&normal,
	callout_major:		CALLOUT_INCAIPASC_MAJOR,
	callout_driver:		&callout,
	table:			incaipasc_table,
	termios:		incaipasc_termios,
	termios_locked:		incaipasc_termios_locked,
	minor:			SERIAL_INCAIPASC_MINOR,
	nr:			UART_NR,
	port:			incaipasc_ports,
	cons:			INCAIPASC_CONSOLE,
};

static int __init incaipasc_init(void)
{
	int ret;

	ret = uart_register_driver(&incaipasc_reg);
        /* Overwrite init_termios speed value - console will do some prints
           with UART default baudrate (38400) otherwise... */
        normal.init_termios.c_cflag = INCAIPASC_DEFAULT_BAUDRATE | CS8 | CREAD | HUPCL | CLOCAL;
#if 0
	if (ret == 0)
		uart_add_one_port(&incaipasc_reg, &incaipasc_ports[0]);
#endif

	/* Create proc file */ 
	create_proc_read_entry("driver/inca_asc", 0, NULL, incaipasc_read_procmem, NULL);

	return ret;
}

static void __exit incaipasc_exit(void)
{
#if 0
	uart_remove_one_port(&incaipasc_reg, &incaipasc_ports[0]);
#endif
	uart_unregister_driver(&incaipasc_reg);
}

module_init(incaipasc_init);
module_exit(incaipasc_exit);

EXPORT_NO_SYMBOLS;

MODULE_AUTHOR("Gary Jennejohn");
MODULE_DESCRIPTION("MIPS INCAIPASC serial port driver");
MODULE_LICENSE("GPL");
