/* 
 * Cryptographic API.
 *
 * Support for Infineon DEU hardware crypto engine.
 *
 * Copyright (c) 2005  Johannes Doering <info@com-style.de>, INFINEON
 *
 * Key expansion routine taken from crypto/aes.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * ---------------------------------------------------------------------------
 * Copyright (c) 2002, Dr Brian Gladman <brg@gladman.me.uk>, Worcester, UK.
 * All rights reserved.
 *
 * LICENSE TERMS
 *
 * The free distribution and use of this software in both source and binary
 * form is allowed (with or without changes) provided that:
 *
 *   1. distributions of this source code include the above copyright
 *      notice, this list of conditions and the following disclaimer;
 *
 *   2. distributions in binary form include the above copyright
 *      notice, this list of conditions and the following disclaimer
 *      in the documentation and/or other associated materials;
 *
 *   3. the copyright holder's name is not used to endorse products
 *      built using this software without specific written permission.
 *
 * ALTERNATIVELY, provided that this notice is retained in full, this product
 * may be distributed under the terms of the GNU General Public License (GPL),
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 * DISCLAIMER
 *
 * This software is provided 'as is' with no explicit or implied warranties
 * in respect of its properties, including, but not limited to, correctness
 * and/or fitness for purpose.
 * ---------------------------------------------------------------------------
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/crypto.h>
#include <linux/interrupt.h>
#include <asm/byteorder.h>
#include <asm/semaphore.h>
#define INCA_IP_AES_AES                          (KSEG1+0x18000880)

#define AES_MIN_KEY_SIZE	16	/* in uint8_t units */
#define AES_MAX_KEY_SIZE	32	/* ditto */
#define AES_BLOCK_SIZE		16	/* ditto */
#define AES_EXTENDED_KEY_SIZE	64	/* in uint32_t units */
#define AES_EXTENDED_KEY_SIZE_B	(AES_EXTENDED_KEY_SIZE * sizeof(uint32_t))


#define DEU_AESIR INT_NUM_IM0_IRL28

//#define CRYPTO_DEBUG


#ifdef CONFIG_CRYPTO_DEV_INCAIP1_AES
#include <asm/incaip/inca-ip.h>
#include <asm/incaip/inca-ip-deu-structs.h>
#define AES_START INCA_IP_AES_AES
#endif
#ifdef CONFIG_CRYPTO_DEV_INCAIP2_AES
#include <asm/incaip2/incaip2-deu.h>
#include <asm/incaip2/incaip2-deu-structs.h>
#define AES_START AES_CON
#endif
#ifdef CONFIG_CRYPTO_DEV_DANUBE_AES
#include <asm/danube/danube.h>
#include <asm/danube/danube_deu.h>
#include <asm/danube/danube_deu_structs.h>
#define AES_START DANUBE_AES_CON
#endif
#ifdef CONFIG_CRYPTO_DEV_IFX_DMA
#ifdef CONFIG_DANUBE
#include <asm/danube/irq.h>
#endif
#ifdef CONFIG_INCAIP2
#include <asm/incaip2/irq.h>
#endif

#include "ifxdeu-dma.h"
extern struct dma_device_info _dma_device;
extern char* deu_dma_read_buffer;
extern int deu_dma_nbytes;
extern wait_queue_head_t deu_dma_rx_wait;
extern volatile int deu_dma_receive_flag;
extern volatile int deu_dma_receive_sleeping_flag;
extern struct semaphore deu_dma_sem;

void swapu32(u32 *addri,u32 *addrj)
{
  u32 h;

  h = *addri;
  *addri = *addrj;
  *addrj = h;
}

#endif

struct aes_ctx {
	uint32_t e_data[AES_EXTENDED_KEY_SIZE+4];
	uint32_t d_data[AES_EXTENDED_KEY_SIZE+4];
	uint32_t *E;
	uint32_t *D;
	int key_length;
};

extern  int disable_multiblock;

static int aes_set_key(void *ctx_arg, const uint8_t *in_key, unsigned int key_len, uint32_t *flags)
{
	volatile struct aes_t *aes = (struct aes_t*) AES_START;
	volatile register u32 tmp;

// start crypto engine with write to ILR 
	aes->controlr.SM=1;
#ifndef CONFIG_CRYPTO_DEV_IFX_DMA
	aes->controlr.ARS=1;
#endif

	/* 128, 192 or 256 bit key length */
	aes->controlr.K =key_len/8 - 2;

	//printk("keylen: %d\n",key_len);

	if (key_len != 16 && key_len != 24 && key_len != 32) {
		*flags |= CRYPTO_TFM_RES_BAD_KEY_LEN;
		return -EINVAL;
	}

	if (key_len == 128 / 8) {

		aes->K3R = *((u32 *) in_key +0);
		aes->K2R = *((u32 *) in_key +1);
		aes->K1R = *((u32 *) in_key +2);
		aes->K0R = *((u32 *) in_key +3);


	} else if (key_len == 192 / 8) {

		aes->K5R = *((u32 *) in_key +0);
		aes->K4R = *((u32 *) in_key +1);
		aes->K3R = *((u32 *) in_key +2);
		aes->K2R = *((u32 *) in_key +3);
		aes->K1R = *((u32 *) in_key +4);
		aes->K0R = *((u32 *) in_key +5);

	} else if (key_len == 256 / 8) {

		aes->K7R = *((u32 *) in_key +0);
		aes->K6R = *((u32 *) in_key +1);
		aes->K5R = *((u32 *) in_key +2);
		aes->K4R = *((u32 *) in_key +3);
		aes->K3R = *((u32 *) in_key +4);
		aes->K2R = *((u32 *) in_key +5);
		aes->K1R = *((u32 *) in_key +6);
		aes->K0R = *((u32 *) in_key +7);

	} else
		return -EINVAL;

	/* let HW pre-process DEcryption key in any case (even if
	   ENcryption is used). Key Valid (KV) bit is then only
	   checked in decryption routine! */
	aes->controlr.PNK =1;

	tmp = aes->controlr.BUS;
	/*-----------------------------------------------------------------*/
	do {
		tmp = aes->controlr.BUS;
	}
	while (aes->controlr.BUS);
	if(!aes->controlr.KV)
		return -EINVAL;

#ifdef CONFIG_CRYPTO_DEV_IFX_DMA

#if 0
    struct dma_device_info *dma_device=&_dma_device;
	volatile struct deu_dma_t *dma = (struct deu_dma_t*) DMA_CON;

	if(!dma_device) return -1;        

	ifx_deu[0].dma_device=dma_device;

	dma->controlr.ALGO = 1;  //AES
	dma->controlr.BS = 0;
	dma->controlr.EN=1;
#endif

#endif
	
    return 0;
}

static void aes_ifxdeu(void *ctx_arg, uint8_t *out_arg, const uint8_t *in_arg,
			uint8_t *iv_arg, size_t nbytes, int encdec,
			int mode)
{

#ifdef CONFIG_CRYPTO_DEV_IFX_DMA
    struct dma_device_info *dma_device=&_dma_device;
	volatile struct deu_dma_t *dma = (struct deu_dma_t*) DMA_CON;
	int x=0;
	u32* out_dmaarr=out_arg;
	u32* in_dmaarr=in_arg;
	int rlen=0;
    int retry_count = 100;
    int waitcount = 0;
    u32* out_dma=deu_dma_read_buffer;
	u8* incopy=NULL;
#else
	volatile register u32 tmp;
#endif

	volatile struct aes_t *aes = (struct aes_t*) AES_START;
	int i=0;
	int nblocks=0;
	int size=16;

#ifdef CRYPTO_DEBUG
    printk("hardware crypto in aes_ifxdeu\n");
	printk("hardware is running\n");
	printk("encdec %x\n",encdec);
	printk("mode %x\n",mode);
	printk("nbytes %u\n",nbytes);
#endif

	aes->controlr.E_D=!encdec;  //encryption
	aes->controlr.O=mode-1;  //0 ECB 1 CBC 2 OFB 3 CFB 4 CTR	hexdump(prin,sizeof(*des));
	//aes->controlr.F = 128; //default  maybe needs to be updated to a proper value calculated from nbytes

	if(mode>1)
		{
		aes->IV3R = (*(u32*)iv_arg);  	
		aes->IV2R = (*((u32*)iv_arg+1)) ;
		aes->IV1R = (*((u32*)iv_arg+2)) ;
		aes->IV0R = (*((u32*)iv_arg+3)) ;
		};

#ifndef CONFIG_CRYPTO_DEV_IFX_DMA
	switch (nbytes) {
	case 16:
		aes->ID3R = *((u32 *) in_arg +0);
		aes->ID2R = *((u32 *) in_arg +1);
		aes->ID1R = *((u32 *) in_arg +2);
		aes->ID0R = *((u32 *) in_arg +3); /* start crypto */
		break;
	case 8:
		aes->ID1R = *((u32 *) in_arg +0);
		aes->ID0R = *((u32 *) in_arg +1); /* start crypto */
		break;
	case 4:
		aes->ID0R = *((u32 *) in_arg); /* start crypto */
		break;
	case 2:
		aes->ID0R = (u32) *((u16 *) in_arg); /* start crypto */
		break;
	case 1:
		aes->ID0R = (u32) *in_arg; /* start crypto */
		break;
	default:
		printk(KERN_ERR "size = %d\n", size);
		;
	}

	/* wait for busy bit to clear */
	/*--- Workaround ----------------------------------------------------
	  do a dummy read to the busy flag because it is not raised early
	  enough in CFB (OFB?) AES modes */
	tmp = aes->controlr.BUS;
	/*-----------------------------------------------------------------*/
	do {
		tmp = aes->controlr.BUS;
	}
	while (aes->controlr.BUS);

	switch (size) {
	case 16:
		*((u32 *) out_arg +0) = aes->OD3R;
		*((u32 *) out_arg +1) = aes->OD2R;
		*((u32 *) out_arg +2) = aes->OD1R;
		*((u32 *) out_arg +3) = aes->OD0R;
		break;
	case 8:
		*((u32 *) out_arg +0) = aes->OD1R;
		*((u32 *) out_arg +1) = aes->OD0R;
		break;
	case 4:
		*((u32 *) out_arg +0) = aes->OD0R;
		break;
	case 2:
		*((u16 *) out_arg) = (u16) aes->OD0R;
		break;
	case 1:
		*out_arg = (u8) aes->OD0R;
		break;
	default:
		;
		}

#else //dma

   
    incopy=KSEG1ADDR(kmalloc(nbytes,GFP_KERNEL));
	if(incopy==NULL) return;
	else memcpy(incopy,in_arg,nbytes);

    
	 _ifx_deu_device* pDev=ifx_deu;

    deu_dma_nbytes=nbytes;
    aes->controlr.DAU=0;  
    aes->controlr.SM=1;  
    aes->controlr.ARS=1; 
    dma->controlr.BS=2;  
	dma->controlr.ALGO = 1; 
	dma->controlr.EN=1;
    deu_dma_init_device(1,0,0);
        
    rlen=dma_device_write(dma_device,(u8 *)incopy,nbytes,NULL);
    while ((rlen != nbytes ) && (retry_count != 0) ) {
        rlen=dma_device_write(dma_device,(u8 *)incopy,nbytes,NULL);
        retry_count--;
    }

    if ( (retry_count == 0) && (rlen != nbytes) ) {
        printk("%s DMA write failed for several times. Giving up.\n",__FUNCTION__);
        /* Clear allocated memory before kfree */
        memset(incopy,0,nbytes);
        kfree(KSEG0ADDR(incopy));
        dma->controlr.EN=0;
        aes->controlr.SM=0;  //needed for other modules
        deu_dma_release();
        return;
    }
 
    down_interruptible(&deu_dma_sem);

    for(i=0;i<(nbytes/4);i++)
	{
        x=i^0x3;
		out_dmaarr[i] = out_dma[x];
	}

	if(incopy) {
        /* Clear allocated memory before kfree */
        memset(incopy,0,nbytes);
        kfree(KSEG0ADDR(incopy));
    }
    else printk("TRACE %s %s %d incopy is NULL cannot free!\n",__FILE__,__FUNCTION__,__LINE__);

/*
    if(out_dma)	{
        free_page((u32)out_dma);
	}
*/

	aes->controlr.SM=0;  //needed for other modules
    dma->controlr.EN=0;
    deu_dma_release();

#endif

}


static void
aes_ifxdeu_ecb(void *ctx, uint8_t *dst, const uint8_t *src,
		uint8_t *iv, size_t nbytes, int encdec, int inplace)
{
	aes_ifxdeu(ctx, dst, src, NULL, nbytes, encdec,
		    CRYPTO_TFM_MODE_ECB);
}

static void
aes_ifxdeu_cbc(void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv,
		size_t nbytes, int encdec, int inplace)
{
	aes_ifxdeu(ctx, dst, src, iv, nbytes, encdec,
		    CRYPTO_TFM_MODE_CBC);
}

static void
aes_ifxdeu_cfb(void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv,
		size_t nbytes, int encdec, int inplace)
{
	aes_ifxdeu(ctx, dst, src, iv, nbytes, encdec,
		    CRYPTO_TFM_MODE_CFB);
}

static void
aes_ifxdeu_ofb(void *ctx, uint8_t *dst, const uint8_t *src, uint8_t *iv,
		size_t nbytes, int encdec, int inplace)
{
	aes_ifxdeu(ctx, dst, src, iv, nbytes, encdec,
		    CRYPTO_TFM_MODE_OFB);
}

static void
aes_encrypt(void *ctx_arg, uint8_t *out, const uint8_t *in)
{
	//printk("aes_encrypt\n");
	aes_ifxdeu(ctx_arg, out, in, NULL, AES_BLOCK_SIZE,
		    CRYPTO_DIR_ENCRYPT, CRYPTO_TFM_MODE_ECB);
	
}

static void
aes_decrypt(void *ctx_arg, uint8_t *out, const uint8_t *in)
{
	//printk("aes_decrypt\n");
	aes_ifxdeu (ctx_arg, out, in, NULL, AES_BLOCK_SIZE,
		     CRYPTO_DIR_DECRYPT, CRYPTO_TFM_MODE_ECB);
}

static struct crypto_alg aes_alg = {
	.cra_name		=	"aes",
	.cra_preference		=	CRYPTO_PREF_HARDWARE,
	.cra_flags		=	CRYPTO_ALG_TYPE_CIPHER,
	.cra_blocksize		=	AES_BLOCK_SIZE,
	.cra_ctxsize		=	sizeof(struct aes_ctx),
#ifdef MODULE
    .cra_module		=	THIS_MODULE,
#endif
	.cra_list		=	LIST_HEAD_INIT(aes_alg.cra_list),
	.cra_u			=	{
		.cipher = {
			.cia_min_keysize	=	AES_MIN_KEY_SIZE,
			.cia_max_keysize	=	AES_MAX_KEY_SIZE,
			.cia_setkey	   	= 	aes_set_key,
			.cia_encrypt	 	=	aes_encrypt,
			.cia_decrypt	  	=	aes_decrypt
		}
	}
};


int __init ifxdeu_init_aes(void)
{
	if (!disable_multiblock) {
		aes_alg.cra_u.cipher.cia_max_nbytes = (size_t)-1;
		aes_alg.cra_u.cipher.cia_req_align  = 16;
		aes_alg.cra_u.cipher.cia_ecb        = aes_ifxdeu_ecb;
		aes_alg.cra_u.cipher.cia_cbc        = aes_ifxdeu_cbc;
		aes_alg.cra_u.cipher.cia_cfb        = aes_ifxdeu_cfb;
		aes_alg.cra_u.cipher.cia_ofb        = aes_ifxdeu_ofb;
	}

	printk(KERN_NOTICE
		"Using Infineon DEU for AES algorithm%s.\n", 
		disable_multiblock ? "" : " (multiblock)");

	//gen_tabs();
	return crypto_register_alg(&aes_alg);
}

void __exit ifxdeu_fini_aes(void)
{
	crypto_unregister_alg(&aes_alg);
}

