/*
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 * danube_nand.c
 *
 *  Description:
 *    device driver of NAND controller of Danube chip
 *  Author:
 *    Samuels Xu Liang
 *  Created:
 *    6 Sep 2005
 *  History & Modification Tag:
 *  ___________________________________________________________________________
 *  |  Tag   |                  Comments                   | Modifier & Time  |
 *  |--------+---------------------------------------------+------------------|
 *  |  S0.0  | First version of this driver and the tag is | Samuels Xu Liang |
 *  |        | implied.                                    |    6 Jul 2005    |
 *  ---------------------------------------------------------------------------
 *  ___________________________________________________________________________
 *  |  Tag   |                  Comments                   | Modifier & Time  |
 *  |--------+---------------------------------------------+------------------|
 *  |  S0.1  | Second version of this driver for INCAIP2   | Winder Sung      |
 *  |        |                                             |    3 Nov 2005    |
 *  ---------------------------------------------------------------------------
 */


/*
 * ####################################
 *              Head File
 * ####################################
 */

/*
 *  Common Head File
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <linux/errno.h>
#include <asm/io.h> /* Winder add this. Without this, we cannot compile successfully. But why?? */
#include <asm/delay.h>

/*
 *  Chip Specific Head File
 */
 #if defined (CONFIG_INCAIP2)         /* FIXME */
#include <asm/incaip2/incaip2.h>
#define IFX_EBU_ADDR_REGION0_WP12 0x10800000 /* Default CS0 used by EBU for external boot after reset. */
#define IFX_EBU_ADDR_REGION0_WP34 0x10000000 /* Default CS0 used by EBU for external boot after reset. */
#define IFX_EBU_ADDR_REGION1 0x19000000 /* CS1 */
#define IFX_EBU_ADDR_REGION2 0x1B000000 /* PC card */
#define IFX_EBU_ADDR_REGION3 0x1C000000
#define IFX_EBU_REGISTER_BASE 0x14102000
#else
#include <danube.h>
#define IFX_EBU_ADDR_REGION0 0x10000000 /* Default CS0 used by EBU for external boot after reset. */
#define IFX_EBU_ADDR_REGION1 0x1A000000 /* CS1 */
#define IFX_EBU_ADDR_REGION2 0x19000000
#define IFX_EBU_ADDR_REGION3 0x1C000000
#define IFX_EBU_REGISTER_BASE 0x10105300
#define IFX_NAND_ADDR_REGION_BASE IFX_EBU_ADDR_REGION1
#endif

unsigned int incaip2_get_cpu_ver(void);

/*
 * ####################################
 *     Parameters to Configure NAND
 * ####################################
 */
static volatile u8 *ebu_base_address;

#ifdef MODULE
    static int  ebu_address_region  = 1;    /*  number of address region (0-4)      */
    static int  ebu_config          = 0;    /*  1 to config EBU, else to do nothing */

    MODULE_PARM(ebu_address_region, "i");
    MODULE_PARM_DESC(ebu_address_region, "Number of address region (0-4).");
    MODULE_PARM(ebu_config, "i");
    MODULE_PARM_DESC(ebu_config, "Set 1 to config EBU, else to do nothing.");
#else
    static int ebu_address_region    = 0;

    /* ebu_config done on fly dependant on WP type */
    static int ebu_config            = 0;	/* don't reconfigure ebu */
	static unsigned int	nand_region_base;
    static int wpType;
#endif

#if 1
#define NUM_PARTITIONS                  6

// winder add. 20 Nov 2005, Duisburg.
// This is partitioning for Workpoint 4
static struct mtd_partition partition_info[NUM_PARTITIONS] = {
	{
		.name   = "environment variables",
		.offset = 0xA0000,   /* from 640k */
		.size   = 0xC000    /* size 48k */
	},
	{
		.name   = "critical data section",
		.offset = 0xAC000,   /* from 688k */
		.size   = 0x8000    /* size 32k */
	},
	{
		.name   = "FT-Image",
		.offset = 0xB4000,   /* from 720k */
		.size   = 0x3CC000   /* size 3.88MB */
	},
	{
		.name   = "Application Kernel",
		.offset = 0x480000,   /* from 4.5MB */
		.size   = 0x200000    /* size 2MB */
	},

	/* The size of the following partion will be modified if Workpoint 1 or 2 */
	{
		.name   = "Application",
		.offset = 0x680000,   /* from 6.5MB */
		.size   = 0x3000000    /* size 48MB */
	},

	/* The base/size of the following partion will be modified if Workpoint 1 or 2 */
	/* See ifx_nand_init() below */
	{
		.name   = "Application Data",
		.offset = 0x3680000,   /* from 54.5MB */
		.size   = 0x0980000    /* size 9.5MB */
	}

};
#endif


/*
 * ####################################
 *              Definition
 * ####################################
 */

#define IFX_NAND_DEBUG_ON                 0

/*
 *  Bits Operation
 */
#define GET_BITS(x, msb, lsb)           (((x) & ((1 << ((msb) + 1)) - 1)) >> (lsb))
#define SET_BITS(x, msb, lsb, value)    (((x) & ~(((1 << ((msb) + 1)) - 1) ^ ((1 << (lsb)) - 1))) | (((value) & ((1 << (1 + (msb) - (lsb))) - 1)) << (lsb)))

/*
 *  EBU Registers Mapping
 */
#define IFX_EBU_REG                      (KSEG1 + IFX_EBU_REGISTER_BASE)

#define IFX_EBU_ADDR_SEL             ((volatile u32*)(IFX_EBU_REG + 0x20 + (ebu_address_region * 4)))
#define IFX_EBU_CONFIG                  ((volatile u32*)(IFX_EBU_REG + 0x60 + (ebu_address_region * 4)))

/*
 *  NAND Registers Mapping
 */
#define IFX_NAND_REG                     IFX_EBU_REG

#define IFX_NAND_CONTROL                 ((volatile u32*)(IFX_NAND_REG + 0x00B0))
#define IFX_NAND_WAIT                ((volatile u32*)(IFX_NAND_REG + 0x00B4))
#define IFX_NAND_ECC0                ((volatile u32*)(IFX_NAND_REG + 0x00B8))
#define IFX_NAND_ECC_AC              ((volatile u32*)(IFX_NAND_REG + 0x00BC))

/*
 *  NAND IO Port Mapping
 */
#define NAND_CMD_DECODER_ALE(x)         ((x) ? (1 << 2) : 0)
#define NAND_CMD_DECODER_CLE(x)         ((x) ? (1 << 3) : 0)
#define NAND_CMD_DECODER_CS(x)          ((x) ? (1 << 4) : 0)
#define NAND_CMD_DECODER_SE(x)          ((x) ? (1 << 5) : 0)
#define NAND_CMD_DECODER_WP(x)          ((x) ? (1 << 6) : 0)
#define NAND_CMD_DECODER_PRE(x)         ((x) ? (1 << 7) : 0)
#define NAND_CLE_BIT	NAND_CMD_DECODER_CLE(1);	
#define NAND_ALE_BIT	NAND_CMD_DECODER_ALE(1);

#define IFX_NAND_CMD                 ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(1) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_WRITE               ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_ADDR                ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(1) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_READ                ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(1) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))

#define IFX_NAND_WRITE_OVER          ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(0) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_ADDR_OVER           ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(0) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))
#define IFX_NAND_READ_OVER           ((volatile u8*)((u32)ebu_base_address + NAND_CMD_DECODER_ALE(0) + NAND_CMD_DECODER_CLE(0) + NAND_CMD_DECODER_CS(0) + NAND_CMD_DECODER_SE(0) + NAND_CMD_DECODER_WP(0) + NAND_CMD_DECODER_PRE(0)))

/*
 *  EBU Address Select Register
 */
#define EBU_ADDR_SEL_BASE_SET(value)    (((value) & ((1 << (1 + 31 - 12)) - 1)) << 12)
#define EBU_ADDR_SEL_MASK_SET(value)    SET_BITS(0, 7, 4, (value))
#define EBU_ADDR_SEL_MRME_SET(value)    ((value) ? (1 << 1) : 0)
#define EBU_ADDR_SEL_REGEN_SET(value)   ((value) ? (1 << 0) : 0)

/*
 *  EBU Configuration Register
 */
#define EBU_CON_WRDIS_SET(value)        ((value) ? (1 << 31) : 0)
#define EBU_CON_ADSWP_SET(value)        ((value) ? (1 << 30) : 0)
#define EBU_CON_AGEN_SET(value)         SET_BITS(0, 26, 24, (value))
#define EBU_CON_SETUP_SET(value)        ((value) ? (1 << 22) : 0)
#define EBU_CON_WAIT_SET(value)         SET_BITS(0, 21, 20, (value))
#define EBU_CON_WINV_SET(value)         ((value) ? (1 << 19) : 0)
#define EBU_CON_PW_SET(value)           SET_BITS(0, 17, 16, (value))
#define EBU_CON_ALEC_SET(value)         SET_BITS(0, 15, 14, (value))
#define EBU_CON_BCGEN_SET(value)        SET_BITS(0, 13, 12, (value))
#define EBU_CON_WAITWRC_SET(value)      SET_BITS(0, 10, 8, (value))
#define EBU_CON_WAITRDC_SET(value)      SET_BITS(0, 7, 6, (value))
#define EBU_CON_HOLDC_SET(value)        SET_BITS(0, 5, 4, (value))
#define EBU_CON_RECOVC_SET(value)       SET_BITS(0, 3, 2, (value))
#define EBU_CON_CMULT_SET(value)        SET_BITS(0, 1, 0, (value))

/*
 *  NAND Flash Controller Control Register
 */
#define NAND_CON_LAT_EN                 GET_BITS(*IFX_NAND_CONTROL, 23, 18)

#define NAND_CON_ECC_ON_SET(value)      ((value) ? (1 << 31) : 0)
#define NAND_CON_PRE_LAT_EN_SET(value)  ((value) ? (1 << 23) : 0)
#define NAND_CON_WP_LAT_EN_SET(value)   ((value) ? (1 << 22) : 0)
#define NAND_CON_SE_LAT_EN_SET(value)   ((value) ? (1 << 21) : 0)
#define NAND_CON_CS_LAT_EN_SET(value)   ((value) ? (1 << 20) : 0)
#define NAND_CON_CLE_LAT_EN_SET(value)  ((value) ? (1 << 19) : 0)
#define NAND_CON_ALE_LAT_EN_SET(value)  ((value) ? (1 << 18) : 0)
#define NAND_CON_OUT_CS_S_SET(value)    SET_BITS(0, 11, 10, (value))
#define NAND_CON_IN_CS_S_SET(value)     SET_BITS(0, 9, 8, (value))
#define NAND_CON_PRE_P_SET(value)       ((value) ? (1 << 7) : 0)
#define NAND_CON_WP_P_SET(value)        ((value) ? (1 << 6) : 0)
#define NAND_CON_SE_P_SET(value)        ((value) ? (1 << 5) : 0)
#define NAND_CON_CS_P_SET(value)        ((value) ? (1 << 4) : 0)
#define NAND_CON_CLE_P_SET(value)       ((value) ? (1 << 3) : 0)
#define NAND_CON_ALE_P_SET(value)       ((value) ? (1 << 2) : 0)
#define NAND_CON_CSMUX_E_SET(value)     ((value) ? (1 << 1) : 0)
#define NAND_CON_NANDM_SET(value)       ((value) ? (1 << 0) : 0)


/*
 *  NAND Flash Device RD/BY State Register
 */
#define NAND_WAIT_WR_C          (*IFX_NAND_WAIT & (1 << 3))
#define NAND_WAIT_RYBY          ((*IFX_NAND_WAIT & 0x7) == 0x7)
#define NAND_WAIT_CLEAR_ALL()   do{*IFX_NAND_WAIT = 0x0;}while(0)

#define NAND_WAIT_READY()       do{while(!NAND_WAIT_RYBY);NAND_WAIT_CLEAR_ALL();}while(0)


/*
 *  NAND Flash ECC Register
 */
#define NAND_ECC0_B2                    GET_BITS(*IFX_NAND_ECC0, 23, 16)
#define NAND_ECC0_B1                    GET_BITS(*IFX_NAND_ECC0, 15, 8)
#define NAND_ECC0_B0                    GET_BITS(*IFX_NAND_ECC0, 7, 0)

#define NAND_ECC0_CLEAR()               do { *IFX_NAND_ECC0 = 0; } while (0)


/*
 * ####################################
 * Preparation of Debug trick
 * ####################################
 */

/*
 *  If try module on Amazon chip, prepare some tricks to prevent invalid memory write.
 */
#if defined(IFX_NAND_DEBUG_ON) && IFX_NAND_DEBUG_ON
#endif  //  defined(IFX_NAND_DEBUG_ON) && IFX_NAND_DEBUG_ON


/*
 * ####################################
 *             Declaration
 * ####################################
 */

/*
 *  MTD Device Operations to Replace Default Routines
 */
static void ifx_hwcontrol(struct mtd_info *mtd, int cmd);
static int ifx_dev_ready(struct mtd_info *);
static void ifx_write_buf(struct mtd_info *mtd, const uint8_t *buf, int len);
static void ifx_write_byte(struct mtd_info *mtd, u_char byte);
static int ifx_calculate_ecc(struct mtd_info *, const u_char *, u_char *);
static void ifx_enable_hwecc(struct mtd_info *, int);

/*
 *  Init Functions
 */
static inline void check_parameter(void);
static inline void config_ebu(void);
static inline void config_nand(void);


/*
 * ####################################
 *            Local Variable
 * ####################################
 */

static struct mtd_info mtd = {0};
static struct nand_chip chip = {0};
static u_char data_buf[512 + 16];

static int f_mtd_partitions = 0;

static int f_ecc_write = 0;


/*
 * ####################################
 *           Global Variable
 * ####################################
 */


/*
 * ####################################
 *            Local Function
 * ####################################
 */
/* On the openStage Hardware we have lock out the FPGA interrupts
 * while we access the NAND device as the latching of the WP bit on
 * address line 6 stops us from writting to the FPGA as the address line
 * is held by the nand controller */
 
#include <asm/incaip2/irq.h>
 
#define FPGA_INT_NUM INT_NUM_IM5_IRL6
#define FPGA_INT_NR 6
u32 oldValue;

/* Hardware control function */
static void ifx_hwcontrol_latch (struct mtd_info *mtd, int cmd)
{
	struct nand_chip *this = (struct nand_chip *) mtd->priv;
	u32 port = (u32) this->IO_ADDR_W;

	switch (cmd) {
		case NAND_CTL_SETCLE: port |=  NAND_CLE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_CLRCLE: port &= ~NAND_CLE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_SETALE: port |=  NAND_ALE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_CLRALE: port &= ~NAND_ALE_BIT; this->IO_ADDR_W = (void *)port; break;

		case NAND_CTL_SETNCE:
			/* lock FPGA ints here */
			oldValue = *INCA_IP2_ICU_IM5_IER;
			*INCA_IP2_ICU_IM5_IER &= ~(INCA_IP2_ICU_IM5_IR(FPGA_INT_NR));
			*IFX_NAND_CONTROL |= NAND_CON_NANDM_SET(1);
			break;
		case NAND_CTL_CLRNCE:
			*IFX_NAND_CONTROL &= ~NAND_CON_NANDM_SET(1);
			/* unlock ints here */
			*INCA_IP2_ICU_IM5_IER |= (oldValue & INCA_IP2_ICU_IM5_IR(FPGA_INT_NR));
			break;
    }
}

static void ifx_hwcontrol(struct mtd_info *mtd, int cmd)
{
	/* Infineon use STMicro NAND flash which is interruptible at all times (i.e. also during RDBY# is low).
	 * This is also true of the Samsung NAND flash data sheet (k9f1208x0b), our flash should behave the
	 * same as we are using the FBGA package K9F1208X0B-J.
	 * For other packages it is saying: "On K9F1208X0B-Y,P or K9F1208X0B-V,F
	 * CE# must be held low during tR", saying that the NAND chip select must be held active during RDBY# low phases.
	 */

	struct nand_chip *this = (struct nand_chip *) mtd->priv;

	u32 port = (u32) this->IO_ADDR_W;
	switch (cmd) {
		case NAND_CTL_SETCLE: port |=  NAND_CLE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_CLRCLE: port &= ~NAND_CLE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_SETALE: port |=  NAND_ALE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_CLRALE: port &= ~NAND_ALE_BIT; this->IO_ADDR_W = (void *)port; break;
		case NAND_CTL_SETNCE: break;
		case NAND_CTL_CLRNCE: break;
    }
}

/*
 *  Description:
 *    Check if the MTD device is ready to accesss.
 *  Input:
 *    mtd --- struct mtd_info *, MTD device to check.
 *  Output:
 *    none
 */
static int ifx_dev_ready(struct mtd_info *mtd)
{
	int ret_val = NAND_WAIT_RYBY;

	if(ret_val)
	{
		NAND_WAIT_CLEAR_ALL();
	}

	return ret_val;
}


static void ifx_write_buf(struct mtd_info *mtd, const uint8_t *buf, int len)
{
	int i;
	struct nand_chip *this = mtd->priv;

	for (i=0; i<len; i++)
	{
		writeb(buf[i], this->IO_ADDR_W);

		while(!NAND_WAIT_WR_C);
	}
}

static void ifx_write_byte(struct mtd_info *mtd, u_char byte)
{
	struct nand_chip *this = mtd->priv;

	writeb(byte, this->IO_ADDR_W);

	while(!NAND_WAIT_WR_C);

	return;
}

/*
 *  Description:
 *    Caculate ECC of a page. For read, ECC is calculated by hardware, however,
 *    for write, a standard function implemented in source file "nand_ecc.c" is
 *    called to produce the ECC.
 *  Input:
 *    mtd      --- struct mtd_info *, MTD device to do ECC calculation.
 *    dat      --- const u_char *, page data.
 *    ecc_code --- u_char *, buffer to put ECC.
 *  Output:
 *    none
 */
static int ifx_calculate_ecc(struct mtd_info *mtd, const u_char *dat, u_char *ecc_code)
{
    if ( f_ecc_write )
        nand_calculate_ecc(mtd, dat, ecc_code);
    else
    {
        ecc_code[0] = NAND_ECC0_B0;
        ecc_code[1] = NAND_ECC0_B1;
        ecc_code[2] = NAND_ECC0_B2;
        *IFX_NAND_CONTROL &= ~NAND_CON_ECC_ON_SET(1);
    }
    return 0;
}

void ifx_nand_waitcmd(void)
{
    while(!NAND_WAIT_WR_C);
}

/*
 *  Description:
 *    Send signal to prepare for ECC calculation. For read, need to enable
 *    ECC hardware.
 *  Input:
 *    mtd  --- struct mtd_info *, MTD device to enable hardware ECC calculation.
 *    mode --- u_char *, signal ECC calculation for read or write.
 *  Output:
 *    none
 */
static void ifx_enable_hwecc(struct mtd_info *mtd, int mode)
{
    switch ( mode )
    {
    case NAND_ECC_READ:
        f_ecc_write = 0;
        NAND_ECC0_CLEAR();
        *IFX_NAND_CONTROL |= NAND_CON_ECC_ON_SET(1);
        break;
    case NAND_ECC_WRITE:
        f_ecc_write = 1;
        break;
    }
}

/*
 *  Description:
 *    Check and adjust parameters passed in by command "insmod".
 *  Input:
 *    none
 *  Output:
 *    none
 */
#ifdef MODULE
    static inline void check_parameter(void)
    {
        if ( ebu_address_region < 0 || ebu_address_region > 4 )
            ebu_address_region = 1;
        else if ( ebu_address_region == 4 )
            ebu_address_region = 3;

        if ( ebu_config != 1 )
            ebu_config = 0;

        switch ( ebu_address_region )
        {
        case 0:
            ebu_base_address = (volatile u8*)(KSEG1 + IFX_EBU_ADDR_REGION0); break;
        case 1:
            ebu_base_address = (volatile u8*)(KSEG1 + IFX_EBU_ADDR_REGION1); break;
        case 2:
            ebu_base_address = (volatile u8*)(KSEG1 + IFX_EBU_ADDR_REGION2); break;
        case 3:
            ebu_base_address = (volatile u8*)(KSEG1 + IFX_EBU_ADDR_REGION3); break;
        }
    }
#else
    static inline void check_parameter(void) {}
#endif

/*
 *  Description:
 *    Config EBU region.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static inline void config_ebu(void)
{
    *IFX_EBU_ADDR_SEL = EBU_ADDR_SEL_BASE_SET(nand_region_base >> 12)
                         | EBU_ADDR_SEL_MASK_SET(15)
                         | EBU_ADDR_SEL_MRME_SET(0)
                         | EBU_ADDR_SEL_REGEN_SET(1);

    *IFX_EBU_CONFIG = EBU_CON_WRDIS_SET(0) | EBU_CON_ADSWP_SET(1) | EBU_CON_AGEN_SET(0x00)
                    | EBU_CON_SETUP_SET(1) | EBU_CON_WAIT_SET(0x00) | EBU_CON_WINV_SET(0)
                    | EBU_CON_PW_SET(0x00) | EBU_CON_ALEC_SET(3) | EBU_CON_BCGEN_SET(0x01)
                    | EBU_CON_WAITWRC_SET(2) | EBU_CON_WAITRDC_SET(2) | EBU_CON_HOLDC_SET(1)
                    | EBU_CON_RECOVC_SET(1) | EBU_CON_CMULT_SET(0x01);

}

/*
 *  Description:
 *    Config NAND controller.
 *  Input:
 *    none
 *  Output:
 *    none
 */
static inline void config_nand(void)
{
    /*
     *  Samsung NAND Chip connections to INCA-IP2
     *  CLE - EBU_AD3
     *  ALE - EBU_AD2
     *  CE  - EBU_CS0
     *  WE  - EBU_WR
     *  RE  - EBU_RD
     *  WP  - FPGA (AD6 on old boards) 
     *  R/B - NAND_READY (driven by nand chip)
     * 
     *  NAND_ALE - EBU_AD2
     *  NAND_CLE - EBU_AD3
     *  NAND_CS  - EBU_CS0
     *  NAND_SE  - EBU_AD5
     *  NAND_WP  - EBU_AD6
     *  NAND_PRE - EBU_AD7
     * 
     *  ECC disabled
     *  CLE, ALE and CS are pulse, all other signal are latches based
     *  OUT_CS is enable and mapped EBU_CS0
     *  PRE, WP, SE and CS/CE are active low
     *  CLE and ALE are active high
     *  CS Mux is enabled
     *  NAND mode is disabled
     */

    *IFX_NAND_CONTROL = NAND_CON_ECC_ON_SET(0)
                     | NAND_CON_PRE_LAT_EN_SET(0) | NAND_CON_WP_LAT_EN_SET(0)
                     | NAND_CON_SE_LAT_EN_SET(0) | NAND_CON_CS_LAT_EN_SET(0)
                     | NAND_CON_CLE_LAT_EN_SET(0) | NAND_CON_ALE_LAT_EN_SET(0)
                     | NAND_CON_OUT_CS_S_SET(ebu_address_region)
                     | NAND_CON_IN_CS_S_SET(ebu_address_region)
                     | NAND_CON_PRE_P_SET(1) | NAND_CON_WP_P_SET(1)
                     | NAND_CON_SE_P_SET(1) | NAND_CON_CS_P_SET(1)
                     | NAND_CON_CLE_P_SET(0) | NAND_CON_ALE_P_SET(0)
                     | NAND_CON_CSMUX_E_SET(0)
                     | NAND_CON_NANDM_SET(0);
                     
}


/*
 * ####################################
 *           Global Function
 * ####################################
 */


/*
 * ####################################
 *           Init/Cleanup API
 * ####################################
 */

/*
 *  Description:
 *    Register MTD partitions or device.
 *  Input:
 *    none
 *  Output:
 *    0    --- successful
 *    else --- failure, usually it is negative value of error code
 */
int __init ifx_nand_init(void)
{

    wpType = get_WP_type();

    if( wpType < 3 ){
		nand_region_base = IFX_EBU_ADDR_REGION0_WP12;
		ebu_base_address = (volatile u8*)(KSEG1 + nand_region_base);
		ebu_config = 0;
	}else{
		nand_region_base = IFX_EBU_ADDR_REGION0_WP34;
		ebu_base_address = (volatile u8*)(KSEG1 + nand_region_base);
		ebu_config = 1;
	}



    /*
     *  Remap to uncached address
     */

    /*
     *  Configure hardware
     */
    check_parameter();

    if ( ebu_config )
        config_ebu();

    config_nand();
    
    printk("OpenStage%d\n", get_WP_type()*20);

    if( get_WP_type() < 3 )
    {
    	chip.hwcontrol      = ifx_hwcontrol;
    	/* Disable NAND write protect which is controlled
    	 * by bit 4 of inca-ip2 gpio pin registers
    	 * on the openStage20 and OpenStage40 phones
    	 */
    	*INCA_IP2_GPIO_POUT |= (1<<4);
    	
    }
    else if (incaip2_get_cpu_ver() < 3)
    {
    	/* is this an old WP or WP4 */
    	printk("Nand WP connected to EBU_AD6\n");
    	chip.hwcontrol      = ifx_hwcontrol_latch;
    }
    else
    {
    	printk("Nand WP connected to FPGA\n");
    	chip.hwcontrol      = ifx_hwcontrol;
    }

    /*
     *  Setup structures
     */
    chip.IO_ADDR_R      = (void *)IFX_NAND_READ;
    chip.IO_ADDR_W      = (void *)IFX_NAND_WRITE;
    chip.dev_ready      = ifx_dev_ready;
    chip.write_buf      = ifx_write_buf;
    chip.write_byte     = ifx_write_byte;
    chip.calculate_ecc  = ifx_calculate_ecc;
    chip.correct_data   = nand_correct_data;
    chip.enable_hwecc   = ifx_enable_hwecc;
//    chip.eccmode        = NAND_ECC_HW3_256;
    chip.eccmode        = NAND_ECC_SOFT;
//    chip.eccmode        = NAND_ECC_NONE;
    chip.data_buf       = data_buf;
//    chip.options        |= NAND_SKIP_BBTSCAN;

    mtd.priv = &chip;

//	printk( "Nand Control    = 0x%08x\n",*IFX_NAND_CONTROL);
//	printk( "Nand Read Addr  = 0x%08x\n",(u32)chip.IO_ADDR_R);
//	printk( "Nand Write Addr = 0x%08x\n",(u32)chip.IO_ADDR_W);
    /*
     *  Scan to find existance of the device
     */
    if ( nand_scan(&mtd, 1) )
    {
//        iounmap(ebu_base_address);
        return -ENXIO;
    }

    /*
     *  Register MTD partitions or device
     */
#ifdef CONFIG_MTD_CMDLINE_PARTS
	{
		int mtd_parts_nb = 0;
		struct mtd_partition *mtd_parts = NULL;

		mtd_parts_nb = parse_cmdline_partitions(&mtd, &mtd_parts, "ifx nand controller");
		if ( mtd_parts_nb > 0 )
		{
			add_mtd_partitions(&mtd, mtd_parts, mtd_parts_nb);
			f_mtd_partitions = 1;
		}
		else
			add_mtd_device(&mtd);
	}
#else
  #if defined(NUM_PARTITIONS) && NUM_PARTITIONS > 0

    if( get_WP_type() < 3 ){
	/* We only have a 32MB device available for Workpoints 1 and 2 */
//	partition_info[4].size = 0x1800000;	/* 24MB App partition */
//	partition_info[5].offset = 0x1e80000;	/* 30.5MB */
//	partition_info[5].size = 0x180000;	/* 1.5MB */
	partition_info[4].size = 0x1500000;	/* 20MB App partition */
	partition_info[5].offset = 0x1b80000;	/* 27.5MB */
	partition_info[5].size = 0x480000;	/* 4.5MB */
    }
    add_mtd_partitions(&mtd, partition_info, NUM_PARTITIONS);
    f_mtd_partitions = 1;
  #else
    add_mtd_device(&mtd);
  #endif
#endif

    return 0;
}

/*
 *  Description:
 *    Deregister MTD partitions or device.
 *  Input:
 *    none
 *  Output:
 *    none
 */
void __exit ifx_nand_exit(void)
{
    if ( f_mtd_partitions )
        del_mtd_partitions(&mtd);
    else
        del_mtd_device(&mtd);

//    iounmap(ebu_base_address);
}

module_init(ifx_nand_init);
module_exit(ifx_nand_exit);
