/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* Backlight dimming 
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/port.h>


#include <linux/tqueue.h> /* task queues */
#include <linux/time.h>
#include "Quantum_drivers.h"
#include "fpga_defs.h"
#include "fpga.h"

#define WP_TYPE_MASK		3

int get_WP_version(void);
int wpv;
/**
* Time out for poll timer is 20mS
*/
#define TIMER_INTERVAL_IN_MILLISECS 20
#define BACKLIGHT_LEVELS 255
#define MAX_SLIDER_BRIGHTNESS 80
#define SLIDER_BRIGHTNESS_OFFSET 30
typedef struct  {
	struct timer_list timer;
	struct tq_struct *task;
	int RunAquisition;
	spinlock_t lock;
	backlight_data_t fade_setting;
	unsigned char fade_stepsize;

} Backlight_device_t;



static Backlight_device_t Backlight_data;




/** Major ID of Backlight device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID Backlight (default: 0=dynamic)");



/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;


ssize_t Backlight_write(struct file *fp, const char *buf, size_t size, loff_t *off);
int Backlight_open(struct inode *inode, struct file *fp);
void BacklightUpdate(Backlight_device_t *dev);




/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations Backlight_fops = {
	open:		Backlight_open,  /* open */
	write:		Backlight_write, /* write */


};

/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */
int Backlight_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "Backlight_open: major = %d minor = %d\n", omajor, ominor);
	
	fp->private_data = &Backlight_data;



	return 0;
}


static void Backlight_startTimer(Backlight_device_t *dev)
{

	dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ)/1000;
	add_timer(&dev->timer);
}


/**
* Poll timer time out function which calls the BacklightUpdate function to set the new 
* backlight level then reschedules its self.
*/
static void Backlight_timeout(unsigned long data)
{
	Backlight_device_t *dev = (Backlight_device_t*)data;
	
	BacklightUpdate(&Backlight_data);
	

	dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ)/1000;
	add_timer(&dev->timer);

}

/**
* This function calculates the from-to range that the backlight is required change the level, an appropriate
* incremental step is dervied from the range, the time to fade, and the current poll timeout. The display backlight 
* and LED's then fade in or out as required over the current range with the derived step to the stop value.
*/
void BacklightUpdate(Backlight_device_t *dev)
{
	unsigned char stepsize;
	unsigned char user_stepsize;
	static int setlevel=255;
	static int stoplevel=255;
	int range;
	int user_range;
	static int slider_setlevel=MAX_SLIDER_BRIGHTNESS;
	static int user_setlevel=80;
	static int slider_stoplevel=80;
	
	if(dev->fade_setting.fadeTime <= 0) dev->fade_setting.fadeTime = 1;
	
	if(dev->fade_setting.lampLevel <= 0) dev->fade_setting.lampLevel = 0;
	

	range = (stoplevel - dev->fade_setting.fadeLevel);
	user_range = (slider_stoplevel - dev->fade_setting.fadeLevel - SLIDER_BRIGHTNESS_OFFSET);
	
	if (range < 0) range = -range;
	if (user_range < 0) user_range = -user_range;
	
	stepsize= (range * TIMER_INTERVAL_IN_MILLISECS)/dev->fade_setting.fadeTime ;
	user_stepsize= (user_range * TIMER_INTERVAL_IN_MILLISECS)/dev->fade_setting.fadeTime ;

	if(stepsize <= 0) stepsize = 1;
	if(user_stepsize <= 0) user_stepsize = 1;
	
	
	if((wpv & WP_TYPE_MASK) > 1)	/* WP3 or WP4? */
	{
	
		if(dev->fade_setting.lampLevel != user_setlevel)
		{
			user_setlevel = dev->fade_setting.lampLevel;
			slider_setlevel = user_setlevel;
			*FPGA_MMI_PWM = slider_setlevel;
		}
	
	
		if((dev->fade_setting.fadeDirection == 1) && (setlevel > dev->fade_setting.fadeLevel))
		{
			setlevel= setlevel - stepsize;
			
			if(setlevel <= dev->fade_setting.fadeLevel)
			{
				if(setlevel <= 0)setlevel = 0;
				setlevel = dev->fade_setting.fadeLevel;
				stoplevel = setlevel;
				if(stoplevel <= 5) ifx_pwm_set(1,0xFF);
			}

			ifx_pwm_set(0,setlevel);

		}
		else if((dev->fade_setting.fadeDirection == 0) && (setlevel < dev->fade_setting.fadeLevel))
		{
			if(stoplevel <= 5) ifx_pwm_set(1,0);
	
			setlevel= setlevel + stepsize;
				
			if(setlevel >= dev->fade_setting.fadeLevel) 
			{
				if(setlevel >= BACKLIGHT_LEVELS)setlevel =BACKLIGHT_LEVELS;
				
				setlevel= dev->fade_setting.fadeLevel;
				stoplevel = setlevel;
			}
			ifx_pwm_set(0,setlevel);

		}
	
		int level = (dev->fade_setting.fadeLevel - SLIDER_BRIGHTNESS_OFFSET);
		if (level < 0) level = 0;
		if((dev->fade_setting.fadeDirection == 1) && (slider_setlevel > level ))
		{
			
			slider_setlevel = slider_setlevel - user_stepsize;
						
			if(slider_setlevel <= (dev->fade_setting.fadeLevel - SLIDER_BRIGHTNESS_OFFSET))
			{
				if(slider_setlevel <= 0)slider_setlevel = 0;
				slider_setlevel = (dev->fade_setting.fadeLevel - SLIDER_BRIGHTNESS_OFFSET);
				slider_stoplevel = slider_setlevel;
			}
	
			if(slider_setlevel < 0)slider_setlevel=0;
			if(slider_setlevel >= user_setlevel)
			{
				*FPGA_MMI_PWM = user_setlevel;
				
			}
			else
			{
				
				*FPGA_MMI_PWM = slider_setlevel;
			}
	
			
		}
		else if((dev->fade_setting.fadeDirection == 0) && (slider_setlevel < user_setlevel))
		{
			
			slider_setlevel = slider_setlevel + user_stepsize;
					
			if(slider_setlevel >= user_setlevel) 
			{
				if(slider_setlevel >= BACKLIGHT_LEVELS)slider_setlevel =BACKLIGHT_LEVELS;
				
				slider_setlevel= user_setlevel;
				slider_stoplevel = slider_setlevel;
	
			}
			
			if(slider_setlevel >= user_setlevel)
			{
				*FPGA_MMI_PWM = user_setlevel;
				
			}
			else
			{
				
				*FPGA_MMI_PWM = slider_setlevel;
			}
		
	
		}
	}
	else /*WP1 and WP2 */
	{
		if(dev->fade_setting.lampLevel != user_setlevel)
		{
			user_setlevel = dev->fade_setting.lampLevel;
			slider_setlevel = user_setlevel;
			ifx_pwm_set(1,slider_setlevel);
		}
		
		if((dev->fade_setting.fadeDirection == 1) && (setlevel > dev->fade_setting.fadeLevel))
		{
			setlevel= setlevel - stepsize;
			
			if(setlevel <= dev->fade_setting.fadeLevel)
			{
				if(setlevel <= 0)setlevel = 0;
				setlevel = dev->fade_setting.fadeLevel;
				stoplevel = setlevel;
			}
	
			ifx_pwm_set(0,setlevel);
			
		}
		else if((dev->fade_setting.fadeDirection == 0) && (setlevel < dev->fade_setting.fadeLevel))
		{
	
			setlevel= setlevel + stepsize;
				
			if(setlevel >= dev->fade_setting.fadeLevel) 
			{
				if(setlevel >= BACKLIGHT_LEVELS)setlevel =BACKLIGHT_LEVELS;
				
				setlevel= dev->fade_setting.fadeLevel;
				stoplevel = setlevel;
			}
			ifx_pwm_set(0,setlevel);	

		}

	}
}



/**
 * Aliased write function called from user space
 *\param *fp pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns Number of data items sent or error state
 */

ssize_t Backlight_write(struct file *fp, const char *buf, size_t size, loff_t *off)
{

	Backlight_device_t* dev = (Backlight_device_t*)fp->private_data;


	if (size != sizeof(backlight_data_t))
	{
		printk(KERN_WARNING "Backlight_write: illegal size %d\n", size);
		return 0;
	}
	
	if (copy_from_user(&dev->fade_setting, buf, sizeof(backlight_data_t)))
	{
		return( -EFAULT );
	}
	return size;

}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *Backlight_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/Backlight'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int Backlight_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "Backlight Device version 1.0\n";
    printk(KERN_INFO "/proc/drivers/Backlight %d\n",count);

    len = sizeof(info);

    if( count < sizeof(info))
    	len = count;
    memcpy(buf,info,len);

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int Backlight_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init Backlight_init(void)
{
	int res = 0;
	
	wpv = get_WP_version();

		/* create and register device */

	if( (input_devfs_handle = devfs_mk_dir(NULL, "input", NULL)) == 0 )
		return -EBUSY;

	devfs_register(input_devfs_handle, "Backlight",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&Backlight_fops,0);


	if (major == 0) major = res;

	SET_MODULE_OWNER(&Backlight_fops);

	printk(KERN_NOTICE "\nBacklight device registered.\n");

	/* create PROC entry */
	if( (Backlight_proc = create_proc_entry( "driver/Backlight", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/Backlight!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	Backlight_proc->read_proc  = Backlight_read_procmem;
	Backlight_proc->write_proc = Backlight_write_procmem;

	printk(KERN_NOTICE "/proc/driver/Backlight created!\n");
	
	spin_lock_init(&Backlight_data.lock);

	printk(KERN_INFO "initializing timer for the first time %d ticks\n",(TIMER_INTERVAL_IN_MILLISECS * HZ)/1000);

	init_timer(&Backlight_data.timer);
	
	Backlight_data.timer.data = (int)&Backlight_data;
	Backlight_data.timer.function = Backlight_timeout;
	
	Backlight_data.fade_setting.fadeTime = 1;
	Backlight_data.fade_setting.fadeDirection = 0;
	Backlight_data.fade_setting.fadeLevel = 255;
	Backlight_data.fade_setting.lampLevel = 80;

	ifx_pwm_set(0, Backlight_data.fade_setting.fadeLevel);

	Backlight_startTimer(&Backlight_data);


	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit Backlight_exit(void)
{


	remove_proc_entry("driver/Backlight", 0);

	devfs_unregister(input_devfs_handle);

	/* Delete the timer */
	del_timer_sync( &Backlight_data.timer );

	printk(KERN_NOTICE "Backlight_exit\n");
}


#ifdef MODULE
module_init(Backlight_init);
module_exit(Backlight_exit);
#endif


/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
