/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* 
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/port.h>



#include <linux/tqueue.h> /* task queues */
#include <linux/time.h>
#include "Quantum_drivers.h"
#include "fpga_defs.h"
#include "fpga.h"


/** Major ID of Contrast device (default: 0=dynamic)
*/
static int major = 0;

/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;


int Contrast_write(struct file *fp, const char *buf, size_t size, loff_t *off);
int Contrast_open(struct inode *inode, struct file *fp);
int Contrast_close(struct inode *inode, struct file *fp);
/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations Contrast_fops = {
	open:		Contrast_open,  /* open */
	write:		Contrast_write, /* write */
	release:	Contrast_close, /* release */

};

/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */
int Contrast_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;
		
	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "Contrast_open: major = %d minor = %d\n", omajor, ominor);
		
	return 0;
}
/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int Contrast_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "Contrast_close:\n");
	err = 0;
	return err;
}

static void panelSPIDelay( u8 data )
{
	udelay( data );
}

static void panelSPIPanelWrite( u8 output )
{

	*FPGA_PAN_CNTRL = output;
	*FPGA_PAN_CNTRL = output;
}


static void panelSPIBitWrite( u8 data )
{
	int i;
	u8 output;
	u8 base = FPGA_PANEL_DEFAULT_WRITE_BITS;
	for( i=0; i < 8; i++ )
	{
		/*MSB first*/
		output = base | (((data<<i) & 0x80) >> 4);
		panelSPIPanelWrite( output );
		output |= FPGA_PANEL_SPI_CLK_BIT;
		panelSPIPanelWrite( output );
	}
}


static void panelSPIBitWriteReg( u8 reg )
{
	/* When wrtting register set bit 2, the write bit */
	panelSPIBitWrite( (reg << 2) | 0x2 );
}


static void panelSPISetDir( void )
{
	u8 output = FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_DIR_BIT | FPGA_PANEL_SPI_CLK_BIT;
	panelSPIPanelWrite( output );

}

static void panelWriteReg( u8 reg, u8 value)
{

	panelSPISetDir();
	panelSPIBitWriteReg( reg );
	panelSPIBitWrite( value );
	panelSPISetDir();
	panelSPIDelay(2);
}

/**
 * Aliased write function called from user space
 *\param *fp pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns Number of data items sent or error state
 */

int Contrast_write(struct file *fp, const char *buf, size_t size, loff_t *off)
{

	static unsigned char contrast_setting = 3;
	
		if (size != sizeof(contrast_setting))
	{
		printk(KERN_WARNING "Contrast_write: illegal size %d\n", size);
		return 0;
	}
	
	if (copy_from_user(&contrast_setting, buf, sizeof(contrast_setting)))
	{
		return( -EFAULT );
	}
	
//	panelWriteReg(0x08, contrast_setting);
//	panelWriteReg(0x28, contrast_setting);

	return size;

}




/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *Contrast_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/Contrast'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int Contrast_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "Contrast Device version 1.0\n";
    printk(KERN_INFO "/proc/drivers/Contrast %d\n",count);

    len = sizeof(info);

    if( count < sizeof(info))
    	len = count;
    memcpy(buf,info,len);

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int Contrast_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	static char procfs_buffer[1024];
	static unsigned long procfs_buffer_size = 0;


	/* get buffer size */
	procfs_buffer_size = count;
	
//	if (procfs_buffer_size > PROCFS_MAX_SIZE ) {
//		procfs_buffer_size = PROCFS_MAX_SIZE;
//	}
	
	/* write data to the buffer */
	if ( copy_from_user(procfs_buffer, buffer, procfs_buffer_size) ) {
		return -EFAULT;
	}
	printk("Value written to /proc %s\n",procfs_buffer);
	
		
	return procfs_buffer_size;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init Contrast_init(void)
{
	int res = 0;

	/* create and register device */

	if( (input_devfs_handle = devfs_mk_dir(NULL, "input", NULL)) == 0 )
		return -EBUSY;

	devfs_register(input_devfs_handle, "Contrast",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&Contrast_fops,0);


	if (major == 0) major = res;

	SET_MODULE_OWNER(&Contrast_fops);

	printk(KERN_NOTICE "\nContrast device registered.\n");

	/* create PROC entry */
	if( (Contrast_proc = create_proc_entry( "driver/Contrast", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/Contrast!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	Contrast_proc->read_proc  = Contrast_read_procmem;
	Contrast_proc->write_proc = Contrast_write_procmem;

	printk(KERN_NOTICE "/proc/driver/Contrast created!\n");
	

	
	return(0);
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit Contrast_exit(void)
{

	remove_proc_entry("driver/Contrast", 0);

	devfs_unregister(input_devfs_handle);


	printk(KERN_NOTICE "Contrast_exit\n");
}


#ifdef MODULE
module_init(Contrast_init);
module_exit(Contrast_exit);
#endif


/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
