/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/** \file Headset.c
*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/mps.h>
#include <asm/incaip2/irq.h>
#include <asm/incaip2/incaip2.h>
#include <linux/delay.h>
#include <linux/tqueue.h> /* task queues */
#include <linux/time.h>
#include "fpga.h"
#include "fpga_defs.h"
#include "keyinput.h"
#include "mps_event.h"

#define CORDLESS_RXD_BIT	(1<<22)		// GP22 / EXIN1 is CORDLESS_RXD (in)	to be used on T40

#define CORDLESS_TXD_BIT	(1<<27)		// GP27 is CORDLESS_TXD (out)
#define CORDLESS_TXD_BIT_OS40Core	(1)		    // GP0  is CORDLESS_TXD (out) for OS40Core

/**
* Time out for poll timer is 450mS
*/
#define TIMER_INTERVAL_IN_MILLISECS 450


/**
* Length of wp2 transmit command queue buffer
*/

#define TX_QUEUE_LEN		(1<<6)
#define TX_QUEUE_MASK	(TX_QUEUE_LEN-1)

#define PTR_INCR(x) x = ((x == TX_QUEUE_MASK) ? 0 : x +1)
#define FIFO_FULL(rp,wp) ((((rp) + TX_QUEUE_LEN - (wp)) & TX_QUEUE_MASK) == 2)

// using EXIN1 (IM3_IRL29)
#define HEADSET_RXD_INT_NUM INT_NUM_IM3_IRL29 	/* CORDLESS_RXD interrupt line */

#define 	WP2_RXD_ONHOOK 		0XC4
#define		WP2_RXD_OFFHOOK 	0XA2

/** Major ID of Headset device (default: 0=dynamic)
*/
static int major = 0;

/**
* Workpoint type (i.e. WP1, WP2, WP3 or WP4)
*/
static int wpType = 0;
static int wp2_is_T40;
/**
* Internal variables for transmitter operation
*/
static struct semaphore transmit_lock;
static struct semaphore txTimer_wait;
static int pulse_count = 0;			// (non-zero pulse_count means transmit timer is running)
static u8 signal = 0;

/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID Headset (default: 0=dynamic)");

/**
* Structure for Headset device
*/
typedef struct  {
	struct timer_list timer;
	struct tq_struct *task;
	int timer_running;
	int dummy;
	spinlock_t lock;
	int rxd_pulse_count;
	int rxd_command;
	int rxd_edge;
	int rxd_pause;
	int rxd_commandByte[32];
	unsigned int txd_command;
	int wqPtr;
	int rqPtr;
	unsigned int txCommandQueue[TX_QUEUE_LEN];
} Headset_device_t;
//MCS
#define TRUE 1
#define FALSE 0
/**
* An instance of the headset driver data structure
*/
static Headset_device_t Headset_data;


//typedef int (*keypad_callback_t_ul) (unsigned long key, int updown, unsigned long timestamp);
static keypad_callback_t_ul Headset_callback = NULL;


/* Timer for ed processing - e.g. we want to wait for transmitter to become available */
struct timer_list timer;

/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;

// Allowed commands for headset
typedef enum {
	HEADSET_NULL_CMD = 0,
	HEADSET_RINGER,
	HEADSET_CALL_ON,
	HEADSET_CALL_OFF
} eFpgaDhsgCmd;

#define HEADSET_CMD_RETRY_LIMIT	10

int get_WP_type(void);
int get_WP_version(void);
static void headset_RXD_isr(int irq, void *dev_id, struct pt_regs *regs);
static void setup_RXD_interrupt(void);


/**
* Declartion of a task queue structure
*/
struct tq_struct TX_command_task; /* global: initialized to zero */

/**
* Transmit command timer start which schedules the Transmit command
*/
static void TX_command_startTimer(Headset_device_t *dev)
{

	schedule_task(&TX_command_task);
	
	dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ)/1000;
	add_timer(&dev->timer);
	dev->timer_running = 1;
}


/**
* Transmit command timer time out function which schedules the Transmit command 
* to the headset and reschedules its self.
*/
static void TX_command_timeout(unsigned long data)
{
	Headset_device_t *dev = (Headset_device_t*)data;
	
	dev->timer_running = 0;
	
	/* Check if there is anything on the comand queue */
		
	if( dev->wqPtr != dev->rqPtr )
	{
		dev->txd_command = dev->txCommandQueue[dev->rqPtr];
		PTR_INCR(dev->rqPtr);
		TX_command_startTimer(&Headset_data);
	}

}

/**
* Callback function called by MPS for headset events (i.e. plugged in, unplugged).
*/
static int HeadsetEventCallback( int state )
{
	// ### Need to check for WP1/2 operation ###

	// Check the received events

//	printk(KERN_NOTICE "/proc/driver/Headset received event from MPS\n");

	if( state == MPS_HEADSET_DISCONNECTED)
	{
		// Headset disconnected event
		// Pass it to the keyboard input handler
		printk(KERN_NOTICE "/proc/driver/Headset disconnected event\n");
		Headset_callback( KERNEL_HEADSET_KEY, KEY_RELEASED, jiffies );
	}
	else
	{
		// Headset connected event
		// Pass it to the keyboard input handler
		printk(KERN_NOTICE "/proc/driver/Headset connected event\n");
		Headset_callback( KERNEL_HEADSET_KEY, KEY_PRESSED, jiffies );
	}

	return( 0 );
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *Headset_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/Headset'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/
static int Headset_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "Headset Device version 2.0\n";
	printk(KERN_INFO "/proc/driver/Headset %d\n",count);

	len = sizeof(info);

	if( count < sizeof(info))
		len = count;
	memcpy(buf,info,len);

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int Headset_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declaration of the file operations structure which effectively aliases application
* functions such as read, write, etc. to equivalent kernel funtion
*/
int Headset_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int Headset_open(struct inode *, struct file *);
int Headset_close(struct inode *, struct file *);

struct file_operations headset_fops = {
	ioctl:		Headset_ioctl,	/* ioctl */
	open:		Headset_open,	/* open */
	release:	Headset_close,	/* release */
};

// Note that there are no read or write functions for this device.
// The command are sent usint the ioctl function and the received
// events are read via the keyInput device,

/***************************/
/* WP2 IOCTL code */

static void start_GPTC_TX_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	pulse_count = 1;
	//  Reload and run the timer
	*INCA_IP2_GPTC_RUN_3A  		=   0x00000002;		// Ensure the timer is stopped

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000010;		// Kill any pending timer interrupt (just in case)
	*INCA_IP2_GPTC_IRNEN   	   	|=  0x00000010;		// Enable timer interrupt

	spin_unlock_irqrestore(i_lock, i_flags);

	*INCA_IP2_GPTC_RUN_3A  		=   0x00000005;		// Now reload and start the timer
}

static void stop_GPTC_TX_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	// Stop the timer running
	// (Note: this function is called from interrupt code)

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNEN   		&= ~0x00000010;		// Disable timer interrupt
	*INCA_IP2_GPTC_RUN_3A  		=   0x00000002;		// Ensure the timer is stopped
	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000010;		// Kill any pending timer interrupt (just in case)

	spin_unlock_irqrestore(i_lock, i_flags);

	pulse_count = 0;
}

static void restart_GPTC_TX_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	*INCA_IP2_GPTC_RUN_3A  		=   0x00000002;		// Ensure the timer is stopped

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000010;		// Kill any pending timer interrupt (just in case)
	*INCA_IP2_GPTC_IRNEN   	   	|=  0x00000010;		// Enable timer interrupt

	spin_unlock_irqrestore(i_lock, i_flags);

	*INCA_IP2_GPTC_RUN_3A  		=   0x00000005;		// Now reload and start the timer
}

static void start_GPTC_RX_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	pulse_count = 1;
	//  Reload and run the timer
	*INCA_IP2_GPTC_RUN_3B  		=   0x00000002;		// Ensure the timer is stopped

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000020;		// Kill any pending timer interrupt (just in case)
	*INCA_IP2_GPTC_IRNEN   	   	|=  0x00000020;		// Enable timer interrupt

	spin_unlock_irqrestore(i_lock, i_flags);

	*INCA_IP2_GPTC_RUN_3B  		=   0x00000005;		// Now reload and start the timer
}

static void stop_GPTC_RX_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	// Stop the timer running
	// (Note: this function is called from interrupt code)
	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNEN   		&= ~0x00000020;		// Disable timer interrupt
	*INCA_IP2_GPTC_RUN_3B  		=   0x00000002;		// Ensure the timer is stopped
	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000020;		// Kill any pending timer interrupt (just in case)

	spin_unlock_irqrestore(i_lock, i_flags);
}

static void wp2_cordless_txd_on(void)
{
	if(get_WP_version() == 5)
		*INCA_IP2_GPIO_POUT &= ~CORDLESS_TXD_BIT_OS40Core;
	else
	*INCA_IP2_GPIO_POUT &= ~CORDLESS_TXD_BIT;
}

static void wp2_cordless_txd_off(void)
{
	if(get_WP_version() == 5)
		*INCA_IP2_GPIO_POUT |= CORDLESS_TXD_BIT_OS40Core;
	else
	*INCA_IP2_GPIO_POUT |= CORDLESS_TXD_BIT;
}

static void wp2_HeadsetTxPulse_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	int bitnum;


	// Clear the interrupt first
 	*INCA_IP2_GPTC_IRNCR	|=  0x00000010;

	// Now process any output actions (but don't do
	// anything unless the pulse count has been started -
	// an extra safeguard).

	if (pulse_count > 0)
	{
		// Should get a pulse (timer interrupt) every 625uS.
		// Do nothing until pulse 2 - let the timer stabilise.
		// Send bits on even pulses from 2 to 16 inclusive
		// ... with inter-bit lows on odd pulses from 3 to 17 inclusive.
		// Leave line low for pulses 18 to 50 inclusive (2 bytes).

		if ((pulse_count >= 2) && (pulse_count <= 17))
		{
			if ((pulse_count % 2) == 0)
			{	// Even pulse - send signal bit
				// (top bit first)
				bitnum = 8 - (pulse_count/2);	// (bitnum is a value from 0 to 7)
				if (((signal >> bitnum) & 1) == 1)
				{						// bit is 1 - switch on
					wp2_cordless_txd_on();
					
					// H36092 - In case interrupt was late arriving now ensure data bit is not too short
					restart_GPTC_TX_timer();
				}
				else
				{						// bit is 0 - switch off
					wp2_cordless_txd_off();
				}
			}
			else
			{	// Odd pulse - send inter-bit low
				wp2_cordless_txd_off();
			}
		}
		else
		{
			// We are either in the lead-in or the post-signal pause,
			// so ensure the line is low and do nothing else.
			wp2_cordless_txd_off();
		}

		pulse_count++;

		// Now check for end
		if (pulse_count > 50)		// 10ms command plus two null bytes
		{							// Finished
			signal = 0;				// The signal has been sent
			stop_GPTC_TX_timer();	// No more pulses
			up(&txTimer_wait);		// Release the transmitter
		}
	}
	// else do nothing
}
static void wp2_HeadsetRxPulse_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	// 8 bit commands; command length 10ms; every 1.25ms we get 1 bit plus 1 space (.625ms);
	// by pulsing every .3125ms we get every potential edge and the middle of each bit / space
	// after catching the first edge the first (and every 4th after) pulse will hit the middle 
	// of each bit
    //     v       v       v       v       v       v       v       v------------  Each potential edge
	//-----+   +---+   +---+   +---+   +---+   +---+   +---+   +---+   +--------  Quiescent State (0)
	//     |   |   |0/1|   |0/1|   |0/1|   |0/1|   |0/1|   |0/1|   |0/1|    
	//     | 1 |   | ? |   | ? |   | ? |   | ? |   | ? |   | ? |   | ? |    
	//     +---+   +---+   +---+   +---+   +---+   +---+   +---+   +---+          Low (1)
    //       ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^
    //pulse: 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	//NOTE the above diagram relates to the signal on the line between the headset and
	//the phone the signal at the INCAIP2 is inverted i.e. Low = 0
	int i;
	unsigned long flags;
	spin_lock_irqsave(&Headset_data.lock,flags);
	if(Headset_data.rxd_pulse_count < 31)
	{	  
	  //at every 4th pulse we check for edge capture
	  if ((Headset_data.rxd_pulse_count % 4) == 0){
		//did we receive an edge?
		if(Headset_data.rxd_edge != 0){
			Headset_data.rxd_command = Headset_data.rxd_command | 1;
		}
		//get ready for the next edge
		Headset_data.rxd_command <<= 1;
	  }
      //clear the edge every 4th plus 2
	  if((Headset_data.rxd_pulse_count % 4) == 2){
		Headset_data.rxd_edge = 0;
	  }
	  Headset_data.rxd_pulse_count++;
	}
	else
	{
	  //we've read all bits - stop the timer, clear the count and wait for the next command
	  Headset_data.rxd_pulse_count = 0;
      //we've shifted 8 bits, move back one
	  Headset_data.rxd_command >>= 1;
	  Headset_data.rxd_pause = TRUE;
	  stop_GPTC_RX_timer();
	  for(i = 0; i < 32;i++)
	  {
		Headset_data.rxd_commandByte[i] = 0;  
	  }

	  // Assume any command is a toggle of the hookswitch
	  Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_PRESSED, jiffies ); 
	  Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_RELEASED, jiffies ); 

	  switch(Headset_data.rxd_command)
	  {
		case WP2_RXD_ONHOOK:    
		{
			// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_PRESSED, jiffies ); 
			// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_RELEASED, jiffies ); 
			break;
		}
		case WP2_RXD_OFFHOOK:
		{	
			// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_PRESSED, jiffies ); 
			// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_RELEASED, jiffies ); 
			break;
		}
		default:
		{
	  		printk(KERN_NOTICE "Headset:wp2_HeadsetRxPulse_isr - invalid command = %0x\n", Headset_data.rxd_command);
		}
	  }
	  Headset_data.rxd_command = 0;
	}
  	spin_unlock_irqrestore(&Headset_data.lock,flags);
  	return;
}

static int send_cmd_to_wp2(unsigned int cmd)
{
	int err = 0;

	// Determine the actual signal to send to the headset
	switch(cmd)
	{
		case HEADSET_RINGER:	signal = 0xc4;	break;
		case HEADSET_CALL_ON:	signal = 0xa2;	break;
		case HEADSET_CALL_OFF:	signal = 0x91;	break;
		case HEADSET_NULL_CMD:	// No command to send
		{
			return(0);			// So don't want to do anything else
			break;
		}
		default:
		{
			printk(KERN_ERR "Headset_ioctl (WP2): invalid cmd (%02x)\n", cmd);
			return( -EINVAL );
		}
	}

//	printk("Signal to headset is 0x%02x\n",signal);

	// Got a signal, now send it if we can
	// Let the timer interrupt handler do all the work (see above)
	start_GPTC_TX_timer();	// Start timer ints going
	down(&txTimer_wait);	// And wait here until the signal has been strobed out
							// (by the timer interrupt handler)

//	printk("Signal sent\n");
	return(err);			// Finished!
}
static void queue_tx_cmd(unsigned int cmd)
{
	Headset_data.txd_command = cmd;

	if(	Headset_data.timer_running == 0)
	{/* nothing being sent start transmission immediatly */
		TX_command_startTimer(&Headset_data);
	}
	else
	{/* a command is in the process of being sent queue command*/
		if( !FIFO_FULL(Headset_data.rqPtr,Headset_data.wqPtr) )
		{
			Headset_data.txCommandQueue[Headset_data.wqPtr]= cmd;
			PTR_INCR(Headset_data.wqPtr);
		}
		else
		{
			printk(KERN_NOTICE "Headset TX command queue: FIFO full\n");
		}
	}
}

/***************************/
/* WP3/4 IOCTL code */

/**
* Interrupt handler for WP3/4 (FPGA) headset events (i.e. onhook, offhook).
*/
static int wp4_Headset_isr( eFpgaInterruptNum vector, void *data )
{
	u8 status;

	status = *FPGA_DHSG_STATCON;

	// Handle receive interrupts first

	if ((status & FPGA_DHSG_RX_ERROR) == 0)
	{
		// If there were no receive errors

		if ((status & FPGA_DHSG_RX_READY) == FPGA_DHSG_RX_READY)
		{
			// And we have received a signal

			// Assume any command is a toggle of the hookswitch
			Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_PRESSED, jiffies ); 
			Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_RELEASED, jiffies ); 

			if ((status & FPGA_DHSG_RX_SIG_MASK) == FPGA_DHSG_RX_SIG_NOSIG)
			{
				// No received signal/erroneous received signal
				// (do nothing here)
				printk(KERN_ERR "DHSG No received signal/erroneous received signal, statcon = 0x%02x\n",status);
			}
			else if ((status & FPGA_DHSG_RX_SIG_MASK) == FPGA_DHSG_RX_SIG_OFFHOOK)
			{
				// Headset off-hook event
				// Pass it to the keyboard input handler
				// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_PRESSED, jiffies );
				// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_RELEASED, jiffies );
				//printk(KERN_NOTICE "Headset::off-hook, press/release sent, statcon = 0x%02x\n",status);
				//printk(KERN_NOTICE "Headset::off-hook, statcon = 0x%02x\n",status);
			}
			else if ((status & FPGA_DHSG_RX_SIG_MASK) == FPGA_DHSG_RX_SIG_ONHOOK)
			{
				// Headset on-hook event
				// Pass it to the keyboard input handler
				// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_PRESSED, jiffies );
				// Headset_callback( KERNEL_HEADSET_HOOKSW, KEY_RELEASED, jiffies );
				//printk(KERN_NOTICE "Headset::on-hook, press/release sent, statcon = 0x%02x\n",status);
				//printk(KERN_NOTICE "Headset::on-hook, statcon = 0x%02x\n",status);
			}
		}
		// else not a receive interrupt
	}
	else
	{
		// Receive error detected
		printk(KERN_ERR "DHSG Receive error, statcon = 0x%02x\n",status);
	}

	// Clear the receive interrupt events
	*FPGA_DHSG_STATCON = (FPGA_DHSG_RX_READY | FPGA_DHSG_RX_ERROR);

	// Transmit operation does not generate interrupts
	// (so nothing to do here)

	// Now clear the DHSG interrupt event
	*FPGA_INT_EVENT = (1<<vector);

	return 0;
}

static int send_cmd_to_wp4(unsigned int cmd)
{
	int err = 0;
	u8 status;

	// Send the command via the FPGA
	status = *FPGA_DHSG_STATCON;
	if ((status & FPGA_DHSG_TX_RUNNING) == 0)
	{
		// No transmission currently in progress, so OK to start this one

		switch(cmd)
		{
			// send appropriate command to DHSG interface
			case HEADSET_RINGER:	*FPGA_DHSG_STATCON = FPGA_DHSG_TX_SIG_RINGER;	break;
			case HEADSET_CALL_ON:	*FPGA_DHSG_STATCON = FPGA_DHSG_TX_SIG_CALL_ON;	break;
			case HEADSET_CALL_OFF:	*FPGA_DHSG_STATCON = FPGA_DHSG_TX_SIG_CALL_OFF;	break;
			case HEADSET_NULL_CMD:	break;	// No command to send
			default:
				printk(KERN_ERR "Headset_ioctl (WP3/4): invalid cmd (%02x)\n", cmd);
				return( -EINVAL );
		}
	}
	else	// Transmitter busy - can't send this command yet.
	{

		printk(KERN_ERR "Headset_ioctl: transmitter busy, statcon = 0x%02x\n", status);
		return( -EBUSY );
	}

	return(err);
}

/***************************/
/* Common IOCTL code */

/**
 * Aliased ioctl function called from user space
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *filp pointer to a kernel file data structure
 *\param cmd ioctl command
 *\param arg associated command arguments
 * Currently not used
 *\returns 0 (success) or -EINVAL/-EFAULT (error)
 */

int Headset_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err = 0;

	// First, get exclusive use of the transmitter
	// (Can't allow two threads to transmit at the same time!)
	down(&transmit_lock);

//	printk(KERN_INFO "Headset_ioctl: cmd=%02x\n", cmd);

	switch(wpType)
	{
		case 1:
		{
			printk(KERN_ERR "ioctl: Headset not supported on WP1\n");
			err = -EINVAL;		// ### CHECK RETURN CODE REQUIRED ###
			break;
		}
		case 2:
		{
			err = 0;

			if( wp2_is_T40)
			{
				queue_tx_cmd(cmd);
		 	}
			break;
		}
		case 3:		// WP3 and WP4 cases are the same
		case 4:
		{
				queue_tx_cmd(cmd);
			break;
		}
		default:	// Invalid phone type
		{
			printk(KERN_ERR "Failed to send command to headset\n");
			err = -EINVAL;
		}
	}

	// Now release the transmitter
	up(&transmit_lock);

	return err;
}


/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */

int Headset_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;
	Headset_device_t* dev;

	if (wpType == 1)
	{
		printk(KERN_ERR "open: Headset not supported on WP1\n");
		return(-1);		// Not allowed on workpoint 1
	}

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "Headset_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &Headset_data;
	dev = (Headset_device_t*)fp->private_data;

	return 0;
}

/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int Headset_close(struct inode *inode, struct file *fp)
{
	int err;

	if (wpType == 1)
	{
		printk(KERN_ERR "close: Headset not supported on WP1\n");
		return(-1);		// Not allowed on workpoint 1
	}

	printk(KERN_NOTICE "Headset_close:\n");
	err = 0;
	return err;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>> Init/Shutdown Section <<<<<<<<<<<<<<<<<<<<<<<<<<<*/

// This function carries out the required actions when the headset
// device is installed, either on startup, or via insmod.

static void setup_GPTC_timers(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	// Set up headset transmit timer to interrupt every 625uS when enabled,

	if( request_irq(INT_NUM_IM3_IRL26, wp2_HeadsetTxPulse_isr, SA_INTERRUPT, "headsetTxTimer", NULL) != 0 )
	{ 
		printk(KERN_NOTICE "\nHeadset:setup_GPTC_timers - TX ISR Init Failed\n");
	}

	if(request_irq(INT_NUM_IM3_IRL27,wp2_HeadsetRxPulse_isr, SA_INTERRUPT,"headsetRxTimer", NULL) != 0 )
	{ 
		printk(KERN_NOTICE "\nHeadset:setup_GPTC_timers - RX ISR Init Failed\n");
	}
	
	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	// The GEN and CLC registers are common to all timers, so be careful
	// about setting them up. Make sure the values used are compatible with
	// other modules that use the timers.
	*INCA_IP2_PMS_PMS_GEN		|=	0x00000400;		// Power supply for GPTC module
	*INCA_IP2_GPTC_CLC 			=   0x00010114;		// 1x bus speed (8kHz?)

	// The remaining modules are specific to the timer,
	// or contain individual bits that are specific to the timer.
	*INCA_IP2_GPTC_CON_3A  		=   0x000001c0;		// sync, any edge, 16-bit, auto-reload, count down
	*INCA_IP2_GPTC_RELOAD_3A	=   0x0000a000;		// Period is 0.625mS
	*INCA_IP2_GPTC_COUNT_3A		=   0x0000a000;
	*INCA_IP2_GPTC_RUN_3A  		=   0x00000002;		// Ensure the timer is stopped
	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000010;		// Kill any pending timer interrupt (just in case)
	*INCA_IP2_GPTC_IRNEN   	   	|=  0x00000010;		// Enable timer interrupt

	// for the RX timer
	*INCA_IP2_GPTC_CON_3B  		=   0x000001c0;		// sync, any edge, 16-bit, auto-reload, count down
	*INCA_IP2_GPTC_RELOAD_3B	=   0x00005000;		// Period is 0.3125
	*INCA_IP2_GPTC_COUNT_3B		=   0x00005000;
	*INCA_IP2_GPTC_RUN_3B  		=   0x00000002;		// Ensure the timer is stopped
	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000020;		// Kill any pending timer interrupt (just in case)
	*INCA_IP2_GPTC_IRNEN   	   	|=  0x00000020;		// Enable timer interrupt

	spin_unlock_irqrestore(i_lock, i_flags);
}

static void destroy_GPTC_timers(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	// for the TXD timer

	free_irq(INT_NUM_IM3_IRL26,0);

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNEN   		&= ~0x00000010;		// Disable timer interrupt
	*INCA_IP2_GPTC_RUN_3A  		=   0x00000002;		// Ensure the timer is stopped
	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000010;		// Kill any pending timer interrupt (just in case)

	spin_unlock_irqrestore(i_lock, i_flags);

	*INCA_IP2_GPTC_CON_3A  		=   0x00000000;
	*INCA_IP2_GPTC_RELOAD_3A	=   0x00000000;

	// and for the RXD timer

	free_irq(INT_NUM_IM3_IRL27,0);

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNEN   		&= ~0x00000020;		// Disable timer interrupt
	*INCA_IP2_GPTC_RUN_3B  		=   0x00000002;		// Ensure the timer is stopped
	*INCA_IP2_GPTC_IRNCR   	   	|=  0x00000020;		// Kill any pending timer interrupt (just in case)

	spin_unlock_irqrestore(i_lock, i_flags);

	*INCA_IP2_GPTC_CON_3B  		=   0x00000000;
	*INCA_IP2_GPTC_RELOAD_3B	=   0x00000000;


}

static void setup_RXD_interrupt(void)
{
	  int err;

	  /* associate the ISR */
	  if( (err = request_irq(HEADSET_RXD_INT_NUM,headset_RXD_isr, SA_INTERRUPT,"headsetRXD-rising-edge", NULL)) == 0 )
	  {
				
	  /* enable EXIN1 for rising edge */
		incaip2_enable_external_int(1,1);				
	  }
	  else
	  {
		printk(KERN_NOTICE "ERROR:%d Headset:setup_RXD_interrupt - failed to create Headset interrupt\n",err);
	  }

	
}
static void destroy_RXD_interrupt(void){
  	  /* disable EXIN1 interrupt */
	  incaip2_disable_external_int(1);
	  free_irq(HEADSET_RXD_INT_NUM,0);
}

static void headset_RXD_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	// we've detected a rising edge
	// by reference to pulse count we detect the command 

	// Acknowledge the interrupt on EXIN1 (this screws up something)
	*INCA_IP2_ICU_IM3_ISR = INCA_IP2_ICU_IM3_IR(29);
	unsigned long flags;
	int level;
	
	//At least 10us of filtering
	
	udelay(10);
	level = incaip2_external_int_level(1);
	
	// Check whether the interrupt was a real signal
	// or just noise.
	
	if(level == 1)
	{
	spin_lock_irqsave(&Headset_data.lock,flags);
	if(Headset_data.rxd_pause == TRUE)
	{
		// this is a new command
		Headset_data.rxd_pulse_count = 0;
		Headset_data.rxd_command = 0;
		Headset_data.rxd_edge = 0;
		Headset_data.rxd_pause = FALSE;
		start_GPTC_RX_timer();
	}
		
	Headset_data.rxd_edge++;
  	spin_unlock_irqrestore(&Headset_data.lock,flags);
	}

	return;
}

void headset_TX_ServiceTask(void *ptr)
{
	
	Headset_device_t* dev = (Headset_device_t*) ptr;
	
	if( wpType == 2 && wp2_is_T40 )
	{
		send_cmd_to_wp2(dev->txd_command);
	}
	else if((wpType == 3) || (wpType == 4))
	{
		send_cmd_to_wp4(dev->txd_command);
	}
	else
	{
		printk(KERN_ERR "headset_TX_ServiceTask: Unkonwn Headset type\n");
	}
		

}

static void headset_startup(void)
{
	Headset_data.rxd_pulse_count = 0;
	keyRegInterface_t keyRegistration;

	init_MUTEX(&transmit_lock);			// Controls access to transmitter
	// ### May need a mutex for for receive timer ###

	/**********************/
	// Don't do anything for workpoint 1.
	// Headsets are not supported on this variant

	if (wpType == 1)
	{
		printk(KERN_ERR "startup: Headset not supported on WP1\n");
		return;		// Nothing to do
	}

	/**********************/
	// Register for key input callback
	// (We register with keyInput.c, which supplies a function to call back.
	// We then call this function to pass on any headset events.)

	keyRegistration.keyInputEvent = 0;
	keyRegistration.name = "Headset";

	if( keyInputRegister( &keyRegistration ) != 0 )	// Get the function to call back (in keyInput)
	{
		printk(KERN_NOTICE "incaip2_headset_register_callback failed:\n");
	}
	else
	{
		Headset_callback = keyRegistration.keyInputEvent;
	}


	if( mps_register_headset_callback( HeadsetEventCallback ) != 0)
	{
		printk(KERN_NOTICE "mps_register_headset_callback failed\n");
	}

	/***************************/
	// Set up required IO lines (WP2 only)


	/***************************/
	// Set up the headset timer(s) (WP2 only)
	int i;
	init_MUTEX_LOCKED(&txTimer_wait);	// Used with GPTC timer to send signals on WP2
	pulse_count = 0;

	if( wpType == 2 && wp2_is_T40 )
	{
        printk(KERN_ERR "headset_startup: allocating resource for WP2 with T40 hardware\n");
			/* cordless rx interrupt on T40 is on EXINT1
		 * the altselt and ports should already be set up by mux
		 * Set correct initial level for CORDLESS_TXD pin
		 */
		wp2_cordless_txd_off();

		// CORDLESS_RXD
		Headset_data.rxd_command = 0;
		Headset_data.rxd_edge	 = 0;
  		Headset_data.rxd_pause	 = TRUE;
		setup_GPTC_timers();
		setup_RXD_interrupt();
		for(i = 0; i < 32; i++)
  		{
			Headset_data.rxd_commandByte[i] = 0;  
		}
		//the spinlock is only to handle contention betwen the RXD timer and edge interrupts
		spin_lock_init(&Headset_data.lock);
	}

	/***************************/
	// Set up the FPGA Interrupt (WP3&4 only)

	if ((wpType == 3) || (wpType == 4))
	{
		// Clear any pending DHSG interrupt events first (just in case)
		*FPGA_INT_EVENT = (1 << FPGA_DHSG_INT);

		// Register the interrupt handler and enable DHSG interrupts
		request_fpga_irq( FPGA_DHSG_INT, wp4_Headset_isr, &Headset_data );
		enable_fpga_irq( FPGA_DHSG_INT );

//		printk(KERN_INFO "Headset interrupts enabled\n");
	}

	/**********************/
}


// This function carries out the required actions when the headset
// device is uninstalled, either on shutdown, or via rmmod. Currently,
// this means unregistering the required callback functions.

static void headset_shutdown(void)
{
	keyRegInterface_t keyRegistration;
	keyRegistration.name = "Headset";

	/**********************/
	// Don't do anything for workpoint 1.
	// Headsets are not supported on this variant

	if (wpType == 1)
	{
		printk(KERN_ERR "shutdown: Headset not supported on WP1\n");
		return;		// Nothing to do
	}

	/***************************/
	// Shut down the headset timer(s) (WP2 only)

	if( wpType == 2 && wp2_is_T40 )
	{
        printk(KERN_ERR "headset_shutdown: releasing resource for WP2 with T40 hardware\n");
		up(&txTimer_wait);		// Release the transmitter
		up(&transmit_lock);		// Release the transmitter
		destroy_GPTC_timers();
		destroy_RXD_interrupt();
	}

	/***************************/
	// Unsetup the FPGA interrupt (WP3&4 only)

	if ((wpType == 3) || (wpType == 4))
	{
		// Disable DHSG interrupts and unregister the interrupt handler
		disable_fpga_irq( FPGA_DHSG_INT );
		free_fpga_irq( FPGA_DHSG_INT );

//		printk(KERN_INFO "Headset interrupts disabled\n");
	}

	/**********************/
	// Unregister for headset event callback
	if( mps_unregister_headset_callback() != 0)
	{
		printk(KERN_NOTICE "mps_unregister_headset_callback failed\n");
	}
	else
	{
//		printk(KERN_NOTICE "incaip2_mps_unregister_callback OK\n");
	}

	// Unregister for key input callback
	if( keyInputUnRegister( &keyRegistration ) != 0 )
	{
		printk(KERN_NOTICE "incaip2_headset_unregister_callback failed\n");
	}
	else
	{
//		printk(KERN_NOTICE "incaip2_headset_unregister_callback OK\n");
	}
	
	Headset_callback = NULL;
}


/**
* Kernel initialisation function run when the module is installed via insmod
*/
static int __init Headset_init(void)
{
	int res = 0;
	/* create and register device */

	if( (input_devfs_handle = devfs_register(0, "input/Headset",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&headset_fops,0)) == 0 )
		return -EBUSY;

	if (major == 0) major = res;


	wpType = get_WP_type();
    wp2_is_T40 = (incaip2_get_cpu_ver() < 3) ? 0 : 1;

	SET_MODULE_OWNER(&headset_fops);

	printk(KERN_NOTICE "\nHeadset device registered.\n");

	/* create PROC entry */
	if( (Headset_proc = create_proc_entry( "driver/Headset", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/Headset!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	Headset_proc->read_proc  = Headset_read_procmem;
	Headset_proc->write_proc = Headset_write_procmem;

	printk(KERN_NOTICE "/proc/driver/Headset created!\n");

	// Carry out required startup actions
	headset_startup();

	Headset_data.wqPtr = 0;
	Headset_data.rqPtr = 0;
	
	TX_command_task.routine = headset_TX_ServiceTask; /*scheduled function*/	
	TX_command_task.data = &Headset_data; /*argument to function*/
	
	init_timer(&Headset_data.timer);
	Headset_data.timer.data = (int)&Headset_data;
	Headset_data.timer.function = TX_command_timeout;
	Headset_data.timer_running = 0;
	
	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/
static void __exit Headset_exit(void)
{
	// Carry out required shutdown actions
	headset_shutdown();

	del_timer_sync( &Headset_data.timer );

	remove_proc_entry("driver/Headset", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_NOTICE "Headset module exit\n");
}


module_init(Headset_init);
module_exit(Headset_exit);

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");


