/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/** \file LANstatus.c
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#else
#error CONFIG_DEVFS_FS must be set
#endif

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>
#include <asm/incaip2/irq.h>
#include <asm/incaip2/incaip2.h>
#include "fpga.h"
#include "fpga_defs.h"

#include "keyinput.h"

//u8 ifx_led_eth_ctrl( int, u8 );	/* to be removed! */

s32 ifx_mdio_read(u8, u8, u8, u16*);
s32 ifx_mdio_write(u8, u8, u8, u16);
int get_WP_version(void);

static keypad_callback_t_ul LANstatus_callback = NULL;

#define GIGA_PHY_MASK		8
#define WP_TYPE_MASK		3

#define CHIP_ID				0
#define PHY0_ADDR			0
#define PHY1_ADDR			1

#define EPHY_CS_SR1			17	/* external PHY copper specific status register 1 */
#define EPHY_IER			18	/* external PHY interrupt enable register */
#define EPHY_CS_SR2			19	/* external PHY copper specific status register 2 */
#define EPHY_PAR			22	/* external PHY page address register */
#define IPHY_IER			20	/* internal PHY interrupt enable register */
#define IPHY_SSR			23	/* internal PHY specific status register */
#define IPHY_ISR			25	/* internal PHY interrupt status register */

#define PHY_STATUS_REG		1	/* Standard PHY Status Register - Common to both internal and external PHYS */
#define PHY_STATUS_LINK_MASK  0x04	/* bit 2 Link Status 0-down 1-up; must be read twice */

#define EPHY_LINK_STATUS_MASK	0x0400
#define IPHY_LINK_STATUS_MASK	0x0020

#define EPHY_RTL_STATUS_MASK	0x0400
#define IPHY_RTL_STATUS_MASK	0x0010

#define PHY0_INT_NUM		INT_NUM_IM2_IRL24
#define PHY1_INT_NUM		INT_NUM_IM2_IRL25

#define LAN_STAT_INT_NUM   	INT_NUM_IM5_IRL6 	/* External Interrupt for LAN Status */

#define PC_EXINT            (1) 
#define PC_STAT_INT_NUM     INT_NUM_IM3_IRL29   /* External Interrupt for PC Status */
#define IM_OFFSET           (96)
#define INCA_IP2_ICU_IM_IR  INCA_IP2_ICU_IM3_IR
#define INCA_IP2_ICU_IM_ISR INCA_IP2_ICU_IM3_ISR

#define SIGIFUP             (SIGUSR1)
#define SIGIFDOWN           (SIGUSR2)
#define IOCTL_CONNECT       (1) 
#define IOCTL_DISCONNECT    (2) 

#if !defined TRUE
#define TRUE                (1 == 1)
#define FALSE               (1 == 0)
#endif

static void PCstatus_extFpgaISR( eFpgaInterruptNum vector, void *data );
static void PCstatus_intPhyISR( int irq, void *dev_id, struct pt_regs *regs );
static void PCstatus_extPhyISR( int irq, void *dev_id, struct pt_regs *regs );
static int  PCstatus_installIntPhyISR( void );
static void PCstatus_destroyIntPhyISR( void );
static int  PCstatus_installExtPhyISR( void );
static void PCstatus_destroyExtPhyISR( void );
static void PCstatus_installFpgaPhyISR( void );
static void PCstatus_destroyFpgaPhyISR( void );
static void PCstatus_handleInterrupt( int ifUp );
static int  PCstatus_open( struct inode * pInode, struct file * pFile );
static int  PCstatus_close( struct inode * pInode, struct file * pFile );
static int  PCstatus_ioctl( struct inode * pInode, struct file * pFile, unsigned int cmd, unsigned long arg );
static int  PCstatus_init( void );
static void PCstatus_exit( void );

static int                ifIsUp;
static struct task_struct *userProcess;
static  devfs_handle_t    pcstatus_devfs_handle;
/**
* Callback function called for external PHY Link status (FPGA) interrupts.
*/
static void LANstatusFPGAIntHandler( eFpgaInterruptNum vector, void *data )
{
	u16 regdata;

	/* Acknowledge the interrupt so that it doesn't keep occurring (i.e. read page 0, reg 19) */
	if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, EPHY_CS_SR2, &regdata) != 0)
		printk(KERN_NOTICE "LANstatusFPGAIntHandler: EPHY_CS_SR2 ifx_mdio_read failed!\n");

	*FPGA_INT_EVENT = (1<<vector);

	/* Confirm which interrupt triggered */
	if(regdata & EPHY_LINK_STATUS_MASK)
	{
		if( LANstatus_callback != NULL )
		{
			/* link up/down indication (page 0 reg 17 bit 10) */
			if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, EPHY_CS_SR1, &regdata) != 0)
				printk(KERN_NOTICE "LANstatusFPGAIntHandler: EPHY_CS_SR1 ifx_mdio_read failed!\n");

			printk(KERN_NOTICE "LANstatusFPGAIntHandler: Link is %s!\n", ((regdata & EPHY_RTL_STATUS_MASK) ? "UP" : "DOWN"));

			LANstatus_callback( KERNEL_LANSTATUS_KEY, ((regdata & EPHY_RTL_STATUS_MASK) ? KEY_PRESSED : KEY_RELEASED), jiffies );
		}
		else
			printk(KERN_NOTICE "LANstatusFPGAIntHandler: No keyInput function to call!\n");
	}
}

static void PCstatus_extFpgaISR( eFpgaInterruptNum vector, void *data )
{
    u16 regdata;

    /* Acknowledge the interrupt so that it doesn't keep occurring (i.e. read page 0, reg 19) */
    if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, EPHY_CS_SR2, &regdata ) != 0 )
    {
        printk( KERN_NOTICE "externalLANstatusIntHandler: EPHY_CS_SR2 ifx_mdio_read failed!\n" );
    }

    /* Acknowledge the interrup on FPGA */
    *FPGA_INT_EVENT = ( 1 << vector );

    /* Confirm which interrupt triggered */
    if ( regdata & EPHY_LINK_STATUS_MASK )
    {
        /* link up/down indication (page 0 reg 17 bit 10) */
        if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, EPHY_CS_SR1, &regdata) != 0 )
        {
            printk( KERN_NOTICE "externalLANstatusIntHandler: EPHY_CS_SR1 ifx_mdio_read failed!\n" );
        }

        printk( KERN_NOTICE "externalLANstatusIntHandler: Link is %s!\n", ((regdata & EPHY_RTL_STATUS_MASK) ? "UP" : "DOWN" ) );

        PCstatus_handleInterrupt( regdata & EPHY_RTL_STATUS_MASK );
    }
}


/**
* Callback function called for internal PHY Link status interrupts.
*/
static void LANstatusIntHandler(int irq, void *dev_id, struct pt_regs *regs)
{
	u16 regdata;

	/* Acknowledge the interrupt so that it does not keep occurring (i.e. read reg 25) */
	if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, IPHY_ISR, &regdata) != 0)
		printk(KERN_NOTICE "LANstatusIntHandler: IPHY_ISR ifx_mdio_read failed!\n");

	/* ack the INCA int */
	*INCA_IP2_ICU_IM2_ISR = INCA_IP2_ICU_IM2_IR(PHY0_INT_NUM - 64);

	/* Confirm which interrupt triggered */
	if(regdata & IPHY_LINK_STATUS_MASK)
	{
		if( LANstatus_callback != NULL )
		{
			/* Why did Richie use the PSSR rather than PSR???? */
			/* H62801 - Use PHY Status Register Link Status rather that PSSR */
			/* link up/down indication (reg 01 bit 2) */
			if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, PHY_STATUS_REG, &regdata) != 0)
				printk(KERN_NOTICE "LANstatusIntHandler: PHY Status Reg ifx_mdio_read failed!\n");

			regdata = 0;

			if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, PHY_STATUS_REG, &regdata) != 0)
				printk(KERN_NOTICE "LANstatusIntHandler: PHY Status Reg ifx_mdio_read failed!\n");

			printk(KERN_NOTICE "LANstatusIntHandler: Link is %s!\n", ((regdata & PHY_STATUS_LINK_MASK) ? "UP" : "DOWN"));

			LANstatus_callback( KERNEL_LANSTATUS_KEY, ((regdata & PHY_STATUS_LINK_MASK) ? KEY_PRESSED : KEY_RELEASED), jiffies );
		}
		else
			printk(KERN_NOTICE "LANstatusIntHandler: No keyInput function to call!\n");
	}
}

static void PCstatus_intPhyISR( int irq, void *dev_id, struct pt_regs *regs )
{
    u16 regdata;

    /* Acknowledge the interrupt so that it does not keep occurring (i.e. read reg 25) */
    if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, IPHY_ISR, &regdata ) != 0 )
    {
        printk( KERN_INFO "PCstatus_intPhyISR: IPHY_ISR ifx_mdio_read failed!\n" );
    }

    /* ack the INCA int */
    *INCA_IP2_ICU_IM2_ISR = INCA_IP2_ICU_IM2_IR( PHY1_INT_NUM - 64 );

    /* Confirm which interrupt triggered */
    if ( regdata & IPHY_LINK_STATUS_MASK )
    {
        /* link up/down indication (reg 23 bit 4) */
        if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, IPHY_SSR, &regdata ) != 0 )
        {
            printk( KERN_INFO "PCstatus_intPhyISR: IPHY_SSR ifx_mdio_read failed!\n" );
        }

        printk( KERN_INFO "PCstatus_intPhyISR: Link is %s!\n", ( ( regdata & IPHY_RTL_STATUS_MASK ) ? "UP" : "DOWN" ) );

        PCstatus_handleInterrupt( regdata & IPHY_RTL_STATUS_MASK );
    }
}


static void externalLANstatusIntHandler(int irq, void *dev_id, struct pt_regs *regs)
{

	u16 regdata;

	// acknowledge the interrupt
	*INCA_IP2_ICU_IM5_ISR = INCA_IP2_ICU_IM5_IR(INT_NUM_IM5_IRL6 - 160);


	/* Acknowledge the interrupt so that it doesn't keep occurring (i.e. read page 0, reg 19) */
	if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, EPHY_CS_SR2, &regdata) != 0)
		printk(KERN_NOTICE "externalLANstatusIntHandler: EPHY_CS_SR2 ifx_mdio_read failed!\n");


	/* Confirm which interrupt triggered */
	if(regdata & EPHY_LINK_STATUS_MASK)
	{
		if( LANstatus_callback != NULL )
		{
			/* link up/down indication (page 0 reg 17 bit 10) */
			if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, EPHY_CS_SR1, &regdata) != 0)
				printk(KERN_NOTICE "externalLANstatusIntHandler: EPHY_CS_SR1 ifx_mdio_read failed!\n");

			printk(KERN_NOTICE "externalLANstatusIntHandler: Link is %s!\n", ((regdata & EPHY_RTL_STATUS_MASK) ? "UP" : "DOWN"));

			LANstatus_callback( KERNEL_LANSTATUS_KEY, ((regdata & EPHY_RTL_STATUS_MASK) ? KEY_PRESSED : KEY_RELEASED), jiffies );
		}
		else
			printk(KERN_NOTICE "externalLANstatusIntHandler: No keyInput function to call!\n");
	}
	return;
}

static void PCstatus_extPhyISR( int irq, void *dev_id, struct pt_regs *regs )
{
    u16 regdata;

    // acknowledge the interrupt
    *INCA_IP2_ICU_IM_ISR = INCA_IP2_ICU_IM_IR( PC_STAT_INT_NUM - IM_OFFSET );


    /* Acknowledge the interrupt so that it doesn't keep occurring (i.e. read page 0, reg 19) */
    if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, EPHY_CS_SR2, &regdata ) != 0 )
    {
        printk( KERN_NOTICE "externalLANstatusIntHandler: EPHY_CS_SR2 ifx_mdio_read failed!\n" );
    }

    /* Confirm which interrupt triggered */
    if ( regdata & EPHY_LINK_STATUS_MASK )
    {
        /* link up/down indication (page 0 reg 17 bit 10) */
        if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, EPHY_CS_SR1, &regdata) != 0 )
        {
            printk( KERN_NOTICE "externalLANstatusIntHandler: EPHY_CS_SR1 ifx_mdio_read failed!\n" );
        }

        printk( KERN_NOTICE "externalLANstatusIntHandler: Link is %s!\n", ((regdata & EPHY_RTL_STATUS_MASK) ? "UP" : "DOWN" ) );

        PCstatus_handleInterrupt( regdata & EPHY_RTL_STATUS_MASK );
    }
}


static void setupExtInterruptForExternalPHY(void)
{
		   

	  int err;
	  /* associate the ISR */
	  if( (err = request_irq(LAN_STAT_INT_NUM,externalLANstatusIntHandler, SA_INTERRUPT,"ExtPHY", NULL)) == 0 )
	  {
		/* enable EXIN0 for falling edge */
		incaip2_enable_external_int(0,2);				
		printk(KERN_NOTICE "setupExtInterruptForExternalPHY - created external PHY interrupt\n");
	  }
	  else
	  {
		printk(KERN_NOTICE "setupExtInterruptForExternalPHY - failed to create external PHY interrupt\n");
	  }

}
static void destroyExtInterruptForExternalPHY()
{

  	  /* disable EXIN0 */
	  incaip2_disable_external_int(0);
	  free_irq(LAN_STAT_INT_NUM,0);
	

}

static int PCstatus_installIntPhyISR()
{
    u16 regdata;

    printk( KERN_INFO "PCstatus_init: Internal PHY\n" );

    if ( request_irq( PHY1_INT_NUM, PCstatus_intPhyISR, SA_INTERRUPT, "PHY1", NULL ) < 0 )
    {
        printk( KERN_INFO "PCstatus_installIntPhyISR: Internal PHY1 PC interrupt registration failed\n" );
        return -1;
    }

    /* Enable the 'Link status changed' interrupt (i.e. set bit 5, reg 20) */
    if ( ifx_mdio_write( CHIP_ID, PHY1_ADDR, IPHY_IER, IPHY_LINK_STATUS_MASK ) != 0 )
    {
        printk( KERN_INFO "PCstatus_installIntPhyISR: IPHY_IER ifx_mdio_write failed!\n" );
        return -1;
    }

    /* Make sure the interrupt status is cleared (i.e. read reg 25) */
    if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, IPHY_ISR, &regdata ) != 0 )
    {
        printk( KERN_NOTICE "PCstatus_installIntPhyISR: IPHY_ISR ifx_mdio_read failed!\n" );
        return -1;
    }

    /* ack the INCA int */
    *INCA_IP2_ICU_IM2_ISR = INCA_IP2_ICU_IM2_IR( PHY1_INT_NUM - 64 );

    printk( KERN_NOTICE "PCstatus_installIntPhyISR - created internal PHY interrupt\n" );
    return 0;
}

static void PCstatus_destroyIntPhyISR()
{
    u16 regdata;

    printk( KERN_INFO "PCstatus_destroyIntPhyISR: Internal PHY\n" );

    /* disable interrupt */
    if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, IPHY_IER, &regdata ) != 0 )
    {
        printk(KERN_NOTICE "PCstatus_exit: IPHY_IER ifx_mdio_read failed!\n");
    }

    regdata &= ~IPHY_LINK_STATUS_MASK;

    if ( ifx_mdio_write( CHIP_ID, PHY1_ADDR, IPHY_IER, regdata ) != 0 )
    {
        printk( KERN_NOTICE "PCstatus_exit: IPHY_IER ifx_mdio_write failed!\n" );
    }

    free_irq( PHY1_INT_NUM, NULL );
    printk( KERN_NOTICE "PCstatus_exit: Internal PHY1 PC interrupt unregistration OK\n" );
}

static int PCstatus_installExtPhyISR()
{
    printk( KERN_NOTICE "PCstatus_installExtPhyISR: External PHY on WP1 or WP2\n" );

    /* associate the ISR */
    if ( request_irq( PC_STAT_INT_NUM, PCstatus_extPhyISR, SA_INTERRUPT, "ExtPHY1", NULL ) < 0 )
    {
        printk( KERN_NOTICE "setupExtInterruptForExternalPHY - failed to create external PHY interrupt\n" );
        return -1;
    }

    /* enable EXINT for falling edge */
    incaip2_enable_external_int( PC_EXINT, 2 );             

    printk( KERN_NOTICE "setupExtInterruptForExternalPHY - created external PHY interrupt\n" );

    return 0;
}

static void PCstatus_destroyExtPhyISR()
{
    printk( KERN_NOTICE "PCstatus_destroyExtPhyISR: WP1 or WP2\n" );

    /* disable EXINT */
    incaip2_disable_external_int( PC_EXINT );

    free_irq( PC_STAT_INT_NUM, NULL );

    printk( KERN_NOTICE "PCstatus_exit: External PC interrupt unregistration OK\n" );
}

static void PCstatus_installFpgaPhyISR()
{
    printk( KERN_NOTICE "PCstatus_installFpgaPhyISR: WP3 or WP4\n" );

    /* Register with FPGA for LAN status interrupt */
    if ( request_fpga_irq( FPGA_SPARE_9_INT, (fpgaIntHander)PCstatus_extFpgaISR, NULL ) )
    {
        printk(KERN_NOTICE "LANstatus_init: FPGA LAN interrupt registration OK\n");
    }

    enable_fpga_irq( FPGA_SPARE_9_INT );
}

static void PCstatus_destroyFpgaPhyISR()
{
    printk( KERN_NOTICE "PCstatus_destroyFpgaPhyISR: WP3 or WP4\n" );
    /* Unregister with FPGA for LAN status interrupt */

    disable_fpga_irq( FPGA_SPARE_9_INT );
    free_fpga_irq( FPGA_SPARE_9_INT );

    printk( KERN_NOTICE "PCstatus_destroyFpgaPhyISR: FPGA PC interrupt unregistration OK\n" );
}

static void PCstatus_handleInterrupt( int ifUp )
{
    if ( ifUp )
    {
        ifIsUp = TRUE;

        /* if we already know the user process we can send a signal to it */
        /* if not, we tell the IF information in ioctl() */
        if ( userProcess != 0 )
        {
            printk( KERN_INFO "PCstatus: sending signal SIGIFUP to pid %d\n", userProcess->pid );
            /*( *my_sys_kill )( userProcess->pid, SIGIFUP );*/
            /*sys_kill( userProcess->pid, SIGIFUP );*/
            kill_proc( userProcess->pid, SIGIFUP, 1 );
        }
        else printk( KERN_INFO "PCstatus: signal SIGIFUP postponed\n" );
    }
    else
    {
        ifIsUp = FALSE;

        if ( userProcess != 0 )
        {
            printk( KERN_INFO "PCstatus: sending signal SIGIFDOWN to pid %d\n", userProcess->pid );
            /*( *my_sys_kill )( userProcess->pid, SIGIFDOWN );*/
            /*sys_kill( userProcess->pid, SIGIFDOWN );*/
            kill_proc( userProcess->pid, SIGIFDOWN, 1 );
        }
        else printk( KERN_INFO "PCstatus: signal SIGIFDOWN postponed\n" );
    }
}

static int PCstatus_open( struct inode * pInode, struct file * pFile )
{
    return 0;
}

static int PCstatus_close( struct inode * pInode, struct file * pFile )
{
    return 0;
}

static int PCstatus_ioctl( struct inode * pInode, struct file * pFile, unsigned int cmd, unsigned long arg )
{
    int retVal = 0;

    switch( cmd )
    {
    case IOCTL_CONNECT:
        printk( KERN_INFO "PCstatus: ** IOCTL_CONNECT\n" );
        /* get the task struct pointer to the calling userspace program, because we want to send signals to it */
        userProcess = current;
        if ( ifIsUp )
        {
            /* tell the caller that the interface is already up */
            retVal = 1;
        }
        break;

    case IOCTL_DISCONNECT:
        printk( KERN_INFO "PCstatus: ** IOCTL_DISCONNECT\n" );
        /* delete the pointer to the calling userspace program, because it's terminating */
        userProcess = 0;
        break;

    default:
        printk( KERN_INFO "PCstatus: unknown IOCTL 0x%X\n", cmd );
        retVal = -EINVAL;
    }
    return retVal;
}


static struct file_operations fops =
{
    .owner   = THIS_MODULE,
    .open    = PCstatus_open, 
    .release = PCstatus_close,
    .ioctl   = PCstatus_ioctl, 
};

/* Initialize monitoring of the PC port */
static int PCstatus_init()
{
    u16 regdata;
    int wpv;

    printk( KERN_NOTICE "PCstatus: mod_init called\n" );

    ifIsUp         = 0;
    userProcess    = 0;

    /* get workpoint version */
    wpv = get_WP_version();

    printk( KERN_INFO "PCstatus_init: get_WP_version returned %04x\n", wpv );

    if ( wpv & GIGA_PHY_MASK )  /* external PHY? */
    {
        printk( KERN_INFO "PCstatus_init: External PHY detected\n" );
        printk( KERN_INFO "Using an external PHY is not supported in this version, Pc port monitoring not established!\n" );
#if 0
        /* Set page address register to 0 (reg 22) */
        if ( ifx_mdio_write( CHIP_ID, PHY1_ADDR, EPHY_PAR, 0 ) != 0 )
        {
            printk( KERN_NOTICE "PCstatus_init: EPHY_PAR ifx_mdio_write failed!\n" );
        }

        /* Enable the 'Link status changed' interrupt (i.e. set bit 10, page 0, reg 18) */
        if ( ifx_mdio_write( CHIP_ID, PHY1_ADDR, EPHY_IER, EPHY_LINK_STATUS_MASK ) != 0 )
        {
            printk( KERN_NOTICE "PCstatus_init: EPHY_IER ifx_mdio_write failed!\n" );
        }

        /* Make sure the interrupt status is cleared (i.e. read page 0, reg 19) */
        if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, EPHY_CS_SR2, &regdata ) != 0 )
        {
            printk( KERN_NOTICE "PCstatus_init: EPHY_CS_SR2 ifx_mdio_read failed!\n" );
        }

        if ( ( wpv & WP_TYPE_MASK ) > 1 )   /* WP3 or WP4? */
        {
            PCstatus_installFpgaPhyISR();
        }
        else    /* WP1 or WP2 */
        {
            if ( PCstatus_installExtPhyISR() < 0 )
            {
                printk( KERN_INFO "PCstatus_init: installExtPhyISR failed\n" );
                return -1;
            }
        }
#endif
    }
    else    /* internal PHY */
    {
        if ( PCstatus_installIntPhyISR() < 0 )
        {
            printk( KERN_INFO "PCstatus_init: installInternalPhyISR failed\n" );
            return -1;
        }
    }

    printk( KERN_INFO "PCstatus_init: register_chrdev\n" );
    /*                                      /dev   name        get MAJ auto         maj min chardev   fops  spare */
    pcstatus_devfs_handle = devfs_register( NULL, "pc_status", DEVFS_FL_AUTO_DEVNUM, 0,  0, S_IFCHR, &fops, 0 );

    if ( pcstatus_devfs_handle == NULL )
    {
        printk( KERN_INFO "PCstatus_init: register_chrdev failed\n" );
        return -1;
    }

    printk( KERN_INFO "PCstatus_init: success pcstatus_devfs_handle = %p\n", pcstatus_devfs_handle );
    return 0;
}

static void PCstatus_exit()
{
    u16 regdata;
    int wpv;

    printk( "PCstatus_exit: mod_exit called\n" );

    devfs_unregister( pcstatus_devfs_handle );

    wpv = get_WP_version();

    printk( KERN_NOTICE "PCstatus_exit: get_WP_version returned %04x\n", wpv );

    if ( wpv & GIGA_PHY_MASK )  /* external PHY? */
    {
        printk( KERN_INFO "PCstatus_exit: External PHY\n" );
        printk( KERN_INFO "Using an external PHY is not supported in this version\n" );
#if 0

        /* disable interrupt */
        if ( ifx_mdio_read( CHIP_ID, PHY1_ADDR, EPHY_IER, &regdata ) != 0 )
        {
            printk(KERN_NOTICE "PCstatus_exit: EPHY_IER ifx_mdio_read failed!\n");
        }

        regdata &= ~EPHY_LINK_STATUS_MASK;

        if ( ifx_mdio_write( CHIP_ID, PHY1_ADDR, EPHY_IER, regdata ) != 0 )
        {
            printk( KERN_NOTICE "PCstatus_exit: EPHY_IER ifx_mdio_write failed!\n" );
        }

        if ( ( wpv & WP_TYPE_MASK ) > 1 )   /* WP3 or WP4? */
        {
            PCstatus_destroyFpgaPhyISR();
        }
        else    /* WP1 or WP2 */
        {
            PCstatus_destroyExtPhyISR();
        }
#endif
    }
    else
    {
        /* Internal PHY */
        PCstatus_destroyIntPhyISR();
    }
}



/**
* Kernel initialisation function run when the module is installed via insmod
*/
static int __init LANstatus_init(void)
{
	keyRegInterface_t keyRegistration;
	u16 regdata;
	int wpv;

	/* Register for keyInput callback */
	keyRegistration.keyInputEvent = 0;
	keyRegistration.name = "LAN";

	if( keyInputRegister( &keyRegistration ) != 0 )	/* Get the function to call back (in keyInput) */
	{
		printk(KERN_NOTICE "LANstatus_init: LAN status keyInput registration failed:\n");
	}
	else
	{
		printk(KERN_NOTICE "LANstatus_init: LAN status keyInput registration OK\n");
		LANstatus_callback = keyRegistration.keyInputEvent;
	}

	wpv = get_WP_version();
	printk(KERN_NOTICE "LANstatus_init: get_WP_version returned %04x\n", wpv);
	if(wpv & GIGA_PHY_MASK)	/* external PHY? */
	{
		printk(KERN_NOTICE "LANstatus_init: External PHY\n");
		/* Set page address register to 0 (reg 22) */
		if(ifx_mdio_write(CHIP_ID, PHY0_ADDR, EPHY_PAR, 0) != 0)
			printk(KERN_NOTICE "LANstatus_init: EPHY_PAR ifx_mdio_write failed!\n");

		/* Enable the 'Link status changed' interrupt (i.e. set bit 10, page 0, reg 18) */
		if(ifx_mdio_write(CHIP_ID, PHY0_ADDR, EPHY_IER, EPHY_LINK_STATUS_MASK) != 0)
			printk(KERN_NOTICE "LANstatus_init: EPHY_IER ifx_mdio_write failed!\n");

		/* Make sure the interrupt status is cleared (i.e. read page 0, reg 19) */
		if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, EPHY_CS_SR2, &regdata) != 0)
			printk(KERN_NOTICE "LANstatus_init: EPHY_CS_SR2 ifx_mdio_read failed!\n");

		if((wpv & WP_TYPE_MASK) > 1)	/* WP3 or WP4? */
		{
			printk(KERN_NOTICE "LANstatus_init: WP3 or WP4\n");
			/* Register with FPGA for LAN status interrupt */
			if(request_fpga_irq( FPGA_LAN_INT, (fpgaIntHander)LANstatusFPGAIntHandler, NULL ))
				printk(KERN_NOTICE "LANstatus_init: FPGA LAN interrupt registration OK\n");

			enable_fpga_irq( FPGA_LAN_INT );
		}
		else	/* WP1 or WP2 */
		{
			printk(KERN_NOTICE "LANstatus_init: WP1 or WP2\n");
			setupExtInterruptForExternalPHY();
		}
	}
	else	/* internal PHY */
	{
		printk(KERN_NOTICE "LANstatus_init: Internal PHY\n");

		/* Enable the 'Link status changed' interrupt (i.e. set bit 5, reg 20) */
		if(ifx_mdio_write(CHIP_ID, PHY0_ADDR, IPHY_IER, IPHY_LINK_STATUS_MASK) != 0)
			printk(KERN_NOTICE "LANstatus_init: IPHY_IER ifx_mdio_write failed!\n");

		/* Make sure the interrupt status is cleared (i.e. read reg 25) */
		if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, IPHY_ISR, &regdata) != 0)
			printk(KERN_NOTICE "LANstatus_init: IPHY_ISR ifx_mdio_read failed!\n");

		/* ack the INCA int */
		*INCA_IP2_ICU_IM2_ISR = INCA_IP2_ICU_IM2_IR(PHY0_INT_NUM - 64);

		if(request_irq(PHY0_INT_NUM, LANstatusIntHandler, SA_INTERRUPT, "PHY0", NULL) == 0)
			printk(KERN_NOTICE "LANstatus_init: Internal PHY0 LAN interrupt registration OK\n");

	}

    /* Functions for monitoring the satus of the LAN port installed, now do the same for the PC port */
    return( PCstatus_init() );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/
static void __exit LANstatus_exit(void)
{
	keyRegInterface_t keyRegistration;
	u16 regdata;
	int wpv;

	/* Unregister keyInput callback */
	keyRegistration.name = "LAN";
	if( keyInputUnRegister( &keyRegistration ) != 0 )
	{
		printk(KERN_NOTICE "LANstatus_exit: LAN status keyInput unregistration failed\n");
	}
	else
	{
		printk(KERN_NOTICE "LANstatus_exit: LAN status keyInput unregistration OK\n");
	}
	LANstatus_callback = NULL;

	wpv = get_WP_version();
	printk(KERN_NOTICE "LANstatus_exit: get_WP_version returned %04x\n", wpv);
	if(wpv & GIGA_PHY_MASK)	/* external PHY? */
	{
		printk(KERN_NOTICE "LANstatus_exit: External PHY\n");
		/* disable interrupt */
		if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, EPHY_IER, &regdata) != 0)
			printk(KERN_NOTICE "LANstatus_exit: EPHY_IER ifx_mdio_read failed!\n");
		regdata &= ~EPHY_LINK_STATUS_MASK;
		if(ifx_mdio_write(CHIP_ID, PHY0_ADDR, EPHY_IER, regdata) != 0)
			printk(KERN_NOTICE "LANstatus_exit: EPHY_IER ifx_mdio_write failed!\n");

		if((wpv & WP_TYPE_MASK) > 1)	/* WP3 or WP4? */
		{
			printk(KERN_NOTICE "LANstatus_exit: WP3 or WP4\n");
			/* Unregister with FPGA for LAN status interrupt */
			disable_fpga_irq( FPGA_LAN_INT );
			free_fpga_irq( FPGA_LAN_INT );
			printk(KERN_NOTICE "LANstatus_exit: FPGA LAN interrupt unregistration OK\n");
		}
		else	/* WP1 or WP2 */
		{
			printk(KERN_NOTICE "LANstatus_exit: WP1 or WP2\n");
			destroyExtInterruptForExternalPHY();
			printk(KERN_NOTICE "LANstatus_exit: External LAN interrupt unregistration OK\n");
		}
	}
	else
	{
			printk(KERN_NOTICE "LANstatus_exit: WP3 or WP4\n");
			/* disable interrupt */
			if(ifx_mdio_read(CHIP_ID, PHY0_ADDR, IPHY_IER, &regdata) != 0)
				printk(KERN_NOTICE "LANstatus_exit: IPHY_IER ifx_mdio_read failed!\n");
			regdata &= ~IPHY_LINK_STATUS_MASK;
			if(ifx_mdio_write(CHIP_ID, PHY0_ADDR, IPHY_IER, regdata) != 0)
				printk(KERN_NOTICE "LANstatus_exit: IPHY_IER ifx_mdio_write failed!\n");

			free_irq(PHY0_INT_NUM, NULL);
			printk(KERN_NOTICE "LANstatus_exit: Internal PHY0 LAN interrupt unregistration OK\n");
	}

    /* Quit also monitoring of the PC port */
    PCstatus_exit();
}


module_init(LANstatus_init);
module_exit(LANstatus_exit);

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
