/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* The MMI and DMI led driver interfaces to the 595 shift-registers
* 
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>
#include "fpga.h"
#include "Quantum_drivers.h"
#include "latch.h"

#define WP_TYPE_MASK		3

int get_WP_version(void);
int wpv;


/** Major ID of QLed device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID QLed (default: 0=dynamic)");




typedef struct  {
	spinlock_t lock;
	led_data_t QLed_setting;
	led_data_t QLed_readData;
} QLed_device_t;

static QLed_device_t QLed_data;

int mmi_variant;


/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;

void QLedServiceTask(void *ptr);

ssize_t QLed_write(struct file *fp, const char *buf, size_t size, loff_t *off);
ssize_t QLed_read(struct file *file, char *buf, size_t size, loff_t * off);
int QLed_open(struct inode *inode, struct file *fp);
int QLed_close(struct inode *inode, struct file *fp);


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations QLed_fops = {
	read:		QLed_read,	/*read */
	write:		QLed_write, /* write */
	open:		QLed_open,  /* open */
	release:	QLed_close, /* release */

};



led_data_t QLedUpdate(led_data_t *ledinfo)
{
	led_data_t indata;
	
	typedef struct {
		unsigned int ledmap[2];
	} outled_t;
	
	typedef struct {
		unsigned char dmledmap[2];	
	} outdm_t;
	
	static outled_t outled;
	static outdm_t outdm;
	outled_t readled;
	outdm_t readdm;

	outled.ledmap[0] = ledinfo->ledmap[0];
	outled.ledmap[1] = ledinfo->ledmap[1];
	outdm.dmledmap[0] = ledinfo->dmledmap[0];
	outdm.dmledmap[1] = ledinfo->dmledmap[1];
	
	if((wpv & WP_TYPE_MASK) > 1)	/* WP3 or WP4? */
	{

		if( spi_ss_set(SS_MM_LED) == 0 )
		{
			printk("QLedUpdate Data not ready MMI_LED\n");
	//		return( -EBUSY );
		}
		
		udelay(15);
	
		
		quantum_write_spi(&outled,sizeof(outled_t));
	
	
		/* at 32Khz it takes 250us to transmit 8 bits */
	
		/* read response back from scc driver */
		
		
		quantum_read_spi(&readled,sizeof(outled_t));
	
		udelay(50);
	
		/* raise /SS */
		spi_ss_clear(SS_MM_LED);
	
		udelay(15);
		
			if( spi_ss_set(SS_DM_LED) == 0 )
		{
			printk("QLedUpdate Data not ready DM_LED\n");
	//		return( -EBUSY );
		}
		udelay(15);
	
		quantum_write_spi(&outdm,sizeof(outdm_t));
	
		/* at 32Khz it takes 250us to transmit 8 bits */
	
	
	
		/* read response back from scc driver */
		quantum_read_spi(&readdm,sizeof(outdm_t));
	
		udelay(50);
		/* raise /SS */
		spi_ss_clear(SS_DM_LED);
	
		udelay(15);
	}
	else	/* WP1 or WP2 */
	{
		
		spi_ss_set(SS_DM_LED);
		udelay(15);
	
		quantum_write_spi(&outdm,sizeof(outdm_t));
		
		/* read response back from scc driver */
		quantum_read_spi(&readdm,sizeof(outdm_t));
	
		udelay(50);
		/* raise /SS */
		spi_ss_clear(SS_DM_LED);
		udelay(15);
	}

	
	indata.ledmap[0] = readled.ledmap[0];
	indata.ledmap[1] = readled.ledmap[1];
	indata.dmledmap[0] = readdm.dmledmap[0];
	indata.dmledmap[1] = readdm.dmledmap[1];
		
	return indata;
}

int QLedOutput(void)
{
	QLedServiceTask(&QLed_data);

	return( 0 );
}

/**
 * Aliased read function called from user space
 * Reads availale command acknowledges else during normal operation
 * it gets polled data off the incoming data queue buffer manipulated it and returns the result to usre space.
 *\param *file pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns either number of data items transferred or error state
 */

ssize_t QLed_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	ssize_t ret;
	
	if (size != sizeof(int))
	{
		printk(KERN_WARNING "QLed_read: illegal size %d\n", size);
		return 0;
	}
		
	ret=sizeof(int);

	if(copy_to_user(buf, &mmi_variant, ret))
	{
		printk(KERN_ERR "QLed_read: copy_to_user failed\n");
		ret = -ERESTARTSYS;
	}

	return( ret );

}


ssize_t QLed_write(struct file *fp, const char *buf, size_t size, loff_t *off)
{

	QLed_device_t* dev = (QLed_device_t*)fp->private_data;

//	printk(KERN_ERR "QLed_write: len=%d\n", size);

	if (size != sizeof(led_data_t))
	{
		printk(KERN_WARNING "QLed_write: illegal size %d\n", size);
		return 0;
	}

	if (copy_from_user(&dev->QLed_setting, buf, sizeof(led_data_t)))
	{
		return( -EFAULT );
	}

	return size;

}


/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */

int QLed_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;
	QLed_device_t* dev;


	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "QLed_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &QLed_data;
	dev = (QLed_device_t*)fp->private_data;
		

	return 0;
}

/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int QLed_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "QLed_close:\n");
	err = 0;
	return err;
}



void QLedServiceTask(void *ptr)
{
	led_data_t updateData;


	QLed_device_t* dev = (QLed_device_t*) ptr;

	updateData.ledmap[0] = dev->QLed_setting.ledmap[0];
	updateData.ledmap[1] = dev->QLed_setting.ledmap[1];
	
	updateData.dmledmap[0] = dev->QLed_setting.dmledmap[0]; 
	updateData.dmledmap[1] = dev->QLed_setting.dmledmap[1];
	
	dev->QLed_readData = QLedUpdate(&updateData);
	

}




/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *QLed_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/QLed'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int QLed_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "QLed Device version 1.0\n";
    printk(KERN_INFO "/proc/drivers/QLed %d\n",count);

    len = sizeof(info);

    if( count < sizeof(info))
    	len = count;
    memcpy(buf,info,len);

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int QLed_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init QLed_init(void)
{
	int res = 0;
	
	wpv = get_WP_version();
	
	/* create and register device */

	if( (input_devfs_handle = devfs_register(0, "input/QLed",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&QLed_fops,0)) == 0 )
		return -EBUSY;


	if (major == 0) major = res;

	SET_MODULE_OWNER(&QLed_fops);

	printk(KERN_NOTICE "\nQLed device registered.\n");

	/* create PROC entry */
	if( (QLed_proc = create_proc_entry( "driver/QLed", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/QLed!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	QLed_proc->read_proc  = QLed_read_procmem;
	QLed_proc->write_proc = QLed_write_procmem;

	printk(KERN_NOTICE "/proc/driver/QLed created!\n");
	
	QLed_data.QLed_setting.ledmap[0] = 0xaa;
	QLedServiceTask(&QLed_data);
	QLed_data.QLed_setting.ledmap[0] = 0;
	QLedServiceTask(&QLed_data);
	
	if(QLed_data.QLed_readData.ledmap[0] == 0x1540000){
		mmi_variant=1;
	}
	else mmi_variant=0;
	
	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit QLed_exit(void)
{


	remove_proc_entry("driver/QLed", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_NOTICE "QLed_exit\n");
}


#ifdef MODULE
module_init(QLed_init);
module_exit(QLed_exit);
EXPORT_SYMBOL(QLedOutput);
#endif


/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
