/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* The touch Slide driver interfaces to the QT5C03(QSlide Touch Slider) device
* using the SPI hardware of the INCA-IP2. This device reports a 7 bit number
* that represents the absolute position on the Slide that is touched. This
* into acceleration data by reporting the change in absolute position between
* one absolute position is  converted update and the next. Updated position
* data is reported every 40ms.
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/incaip2.h>
#include "Quantum_drivers.h"
#include "latch.h"
#include "fpga.h"
#include "keyinput.h"


MODULE_DESCRIPTION("Quantum Slider Device");


/* FPK sensitivity Parameters - WP4*/
int fpkset_WP4 = 8; 
MODULE_PARM(fpkset_WP4, "i");
MODULE_PARM_DESC(fpkset_WP4, "Set FPK sensitivity (1..63)");

/* FPK sensitivity Parameters - WP2 & WP3*/
int fpkset_WP = 9; 
MODULE_PARM(fpkset_WP, "i");
MODULE_PARM_DESC(fpkset_WP, "Set FPK sensitivity (1..63)");

/* SLIDER sensitivity Parameters */
int sliderset = 5; 
MODULE_PARM(sliderset, "i");
MODULE_PARM_DESC(sliderset, "Set SLIDER sensitivity (1..63)");

/* WHEEL sensitivity Parameters */
int wheelset = 10; 
MODULE_PARM(wheelset, "i");
MODULE_PARM_DESC(wheelset, "Set WHEEL sensitivity (1..63)");

int wheelperiod = 16; 
MODULE_PARM(wheelperiod, "i");
MODULE_PARM_DESC(wheelperiod, "Set WHEEL period (1..63)");

/**
* Time out for poll timer is 40mS
*/
#define TIMER_INTERVAL_IN_MILLISECS 40

/**
* Number of polls before we send any pending volume key indication
*/
#define SLIDER_POLL_FREQ	3
#define SLIDER_HISTORY_SIZE (1<<8)


/**
* Structure for kernel timer device used for the QSlide poll 'tick' and mouse emulation
*/

typedef struct {
	u8 start;
	u8 end;
	u8 centre;
	u8 tolerence;
} slider_map_t;


typedef struct {
	int	sensitivity;	/* acceleration component */
	int	setting;		/* initial distance required for event */
	int	de;				/* distance since last event sent */
	int	polls;			/* number of polls since last event */
	int	zeroTotal;		/* number of consecutive zeros */
	int	badsamples;		/* number of zeros identified as missed codes*/
	int	flip;			/* number of incorrect reversals identified as missed codes*/

	int touched;
	int	deltaIndex;			/* index to history */
	int	deltaTotal;			/* running total of history */
	char deltas[8];			/* history */
	u8 lastWrite;			/* last reported position */
	u8 lastPoll;			/* last reported position */
} QWheel_state_t;

typedef void (*actionFnPtr_t) ( void *pollFnData, u8 poll );

typedef struct {
	actionFnPtr_t pollFn; 
	void *pollFnData; 
	volatile u16 History[SLIDER_HISTORY_SIZE];
	volatile int HistoryCount;
	volatile int FreezeCount;
	char *name;
	int errorPollCount;
	int totalPolls;
	int totalCalibrate;
	int totalDriftCompensate;
	int totalErrorStatus;
	int totalErrorDetect;
	int totalPollErrorBit;
	int totalSetSensitivity;
	int totalDataNotReady;
	int totalDataReady;
	spi_ss_t ss;
	u8 type;
	u8 state;
	u8 send;
	u8 sent;
	u8 recv;
	u8 sensitivity;
	u8 driftComp;
	u8 pollcount;
	u8 firstDetection;
    u8 firstCalibration;
} slider_state_t;

typedef struct  {
	struct timer_list timer;
	struct tq_struct *task;
	int RunAquisition;
	spinlock_t lock;
} QSlide_device_t;



static QSlide_device_t QSlide_data;
keypad_callback_t_ul QSlide_callback = 0;
QWheel_state_t QWheelstate;

QWheel_callback_t QWheel_callback = 0;
QWheel_callback_t QWheelTestIfc_callback = 0;


slider_state_t FPKmachine;
slider_state_t VOLmachine;
slider_state_t WHLmachine;

u8 FPKmachineStartTable[6] = { SS_DM_KEY,    0x00,0x00,0x80,0x05,0x0A };
u8 VOLmachineStartTable[6] = { SS_MM_SLIDER, 0x00,0x00,0x80,0x05,0x0A };
u8 WHLmachineStartTable[6] = { SS_MM_WHEEL,  0x40,0x00,0x80,0x05,0x0A };

/* FPK range values taken from HSI document: Note: WP2/WP3 values inverted for consistency */
const int FPK_calibration_values[3][9][2] =
{
	{	{0,21},		/*WP2 FPK 0 range*/
		{23,42},	/*WP2 FPK 1 range*/
		{44,64},	/*WP2 FPK 2 range*/
		{66,85},	/*WP2 FPK 3 range*/
		{87,106},	/*WP2 FPK 4 range*/
		{108,126},	/*WP2 FPK 5 range*/
		{0,0},		/* N/A */
		{0,0},		/* N/A */
		{0,0}},		/* N/A */
	{	{0,15},	/*WP3 FPK 0 range*/
		{17,30},	/*WP3 FPK 1 range*/
		{32,46},	/*WP3 FPK 2 range*/
		{48,62},	/*WP3 FPK 3 range*/
		{64,77},	/*WP3 FPK 4 range*/
		{79,93},	/*WP3 FPK 5 range*/
		{95,111},	/*WP3 FPK 6 range*/
		{113,126},	/*WP3 FPK 7 range*/
		{0,0}},		/*N/A*/
		
	{	{0,11},		/*WP4 FPK 0 range*/
		{13,23},	/*WP4 FPK 1 range*/
		{25,39},	/*WP4 FPK 2 range*/
		{41,55},	/*WP4 FPK 3 range*/
		{57,71},	/*WP4 FPK 4 range*/
		{73,88},	/*WP4 FPK 5 range*/
		{90,103},	/*WP4 FPK 6 range*/
		{105,117},	/*WP4 FPK 7 range*/
		{119,127}} 	/*WP4 FPK 8 range*/
};

static u8 lastVolKeySent;
static u8 newVolKey;
static int sliderPollCount;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/

DECLARE_WAIT_QUEUE_HEAD(qSlide_wait_queue);

/**
* Declartion of a task queue structure
*/
struct tq_struct QSlide_task; /* global: initialized to zero */



unsigned char numberOfFPKs;

static slider_map_t FPK_map[9];

u32 FPKeyStateTable;
u32 volumeStateTable;


void QSlideServiceTask(void *ptr);
void QSlideServiceTask_WP2(void *ptr);
int get_WP_version(void);
int get_WP_type(void);

int QSlideEventNotify( u8 event, int state )
{
	if( QSlide_callback != 0 )
	{
	//	printk("QSlider, QSlideEventNotify(), key = 0x%02x, state = %d\n", event, state);
		return( QSlide_callback( (unsigned long)event, state, jiffies ) );
	}
	return( 0 );
}

int QWheelEventNotify( u8 event )
{
	if( QWheel_callback != 0 )
	{
	//	printk("QSlider, QWheelEventNotify(), key = 0x%02x\n", event);
		return( QWheel_callback( event, jiffies ) );
	}
	return( 0 );
}

int QWheelTestIfcEventNotify( u8 event )
{
	if( QWheelTestIfc_callback != 0 )
	{
	//	printk("QSlider, QWheelTestIfcEventNotify(), key = 0x%02x\n", event);
		return( QWheelTestIfc_callback( event, jiffies ) );
	}
	return( 0 );
}

int QWheelSetSensitivity( int sensitivity )
{
	printk("QSlider, QWheelSetSensitivity(%d)\n", sensitivity );
	QWheelstate.setting = sensitivity;
	return( 0 );
}

int QWheelSetMark( int sensitivity )
{
	printk("QSlider, QWheelSetMark(%d)\n", sensitivity );
	QWheelstate.sensitivity = sensitivity;
	return( 0 );
}


void initQWheelState( QWheel_state_t *state )
{
	state->de = 0;
	state->setting = 0;
	state->polls = 0;
	state->sensitivity = wheelperiod;
	state->lastWrite = 0;
	state->touched = 0;	
}



static void init_FPK_map( void )
{
	int i;
	int WP_index;

//    printk("Hardware version = 0x%x, number of FPKs = 0x%x\n", get_WP_version(), numberOfFPKs);
	if(get_WP_type() == 2)
	{
		/* WP2 */
		WP_index = 0;
	}
	else if(get_WP_type() == 3)
	{
		/* WP3 */
		WP_index = 1;
	}		
	else
	{
		/* WP4 */
		WP_index = 2;
	}


	for( i=0 ; i < numberOfFPKs;i++ )
	{
		FPK_map[i].start = FPK_calibration_values[WP_index][i][0];	
		FPK_map[i].end = FPK_calibration_values[WP_index][i][1];
//		printk("FPK_%d %3d <-> %3d\n",i,FPK_map[i].start,FPK_map[i].end);
	}
}

int FPKValueStats[265];

static void clearFPKvaluesStats( void )
{
	int i;
	for( i=0 ; i<128; i++ )
	{
		FPKValueStats[i] = 0;
	}
}

static void clearFPKeyStats( int start, int end )
{
	int i;
	for( i=start ; i<=end; i++ )
	{
		FPKValueStats[i] = 0;
	}
}

static void showFPKvaluesStats( void )
{
	int i;
	int key = 0;
	int score = 0;
	int meanscore = 0;
	int entries = 0;
	int totalHits = 0;
	int hits = 0;
	int miss = 0;
	int centre = 0;
	int tempmean = 0;
	int max = 0;
	int maxat = 0;
	for( i=0 ; i<128; i++ )
	{
		if( FPK_map[key].start == i )
		{
			max = 0;
			maxat = i;
			entries = 0;
			score = 0;
			hits = 0;
			meanscore = 0;
			centre = (FPK_map[key].end - FPK_map[key].start)/2 + FPK_map[key].start;
			printk("FPK_%d start=<%d> centre=<%d> end=<%d> miss=%d\n",key,i,centre,FPK_map[key].end, miss);
		}

		if( (i >= FPK_map[key].start) && ( i <= FPK_map[key].end) )
		{
			if( max < FPKValueStats[i])
			{
				max = FPKValueStats[i];
				maxat = i;
			}
			entries++;
			totalHits += FPKValueStats[i];
			hits += FPKValueStats[i];
			score += ( i - centre)*FPKValueStats[i];
			tempmean = (centre > i)? (centre - i):(i - centre);
			meanscore += FPKValueStats[i]*tempmean;
		}
		else if( FPKValueStats[i] != 0 )
		{
			miss++;
			printk("FPK_%d miss @ %d %d<->%d  hits=%d\n",key,i,FPK_map[key].start,FPK_map[key].end,FPKValueStats[i]);
			clearFPKeyStats(i,i);
		}
		
		if( FPK_map[key].end == i )
		{
			miss = 0;
			if( hits != 0 )
			{
				meanscore = meanscore/hits;
				score = score/hits;
			}
//			printk("       entries=%d hits=%d \n",entries,hits );
//			printk("       score=%d Q=%d max=%d <%d>\n\n",score,meanscore,max,maxat);
			if( score < 0 && hits > 64)
			{
//				printk("FPK_%d Moved <- start=<%d> centre=<%d> end=<%d>\n",key,FPK_map[key].start,centre,FPK_map[key].end);
				clearFPKeyStats(FPK_map[key].start,FPK_map[key].end);
				if( FPK_map[key].start > 0 )
					FPK_map[key].start--;
				FPK_map[key].end--;
			}
			if( score > 0 && hits > 64)
			{
//				printk("FPK_%d Moved -> start=<%d> centre=<%d> end=<%d>\n",key,FPK_map[key].start,centre,FPK_map[key].end);
				clearFPKeyStats(FPK_map[key].start,FPK_map[key].end);
				FPK_map[key].start++;
				if( FPK_map[key].end < 127 )
					FPK_map[key].end++;
			}
			if( score == 0 && hits > 64)
			{
				clearFPKeyStats(FPK_map[key].start,FPK_map[key].end);
			}
			key++;
		}
	}
}


static u8 getFPKeyNumer( u8 val )
{
int i;
static int count;
int score;


	if(get_WP_type() == 3 || get_WP_type() == 2)
	{
		/* WP3 so invert and mask out top bit */
		val = (~val) & 0x7f;
	}
	else
	{
		/* WP4, just mask out top bit */
		val = val & 0x7f;
	}

	FPKValueStats[val]++;
	
	count++;
	
	if( (count & 0x3f ) == 0 )
	{
	//	showFPKvaluesStats();
	//	clearFPKvalues`	Stats();
	}
	
	for( i=0; i < sizeof(FPK_map)/sizeof(slider_map_t); i++ )
	{
		if( val >= FPK_map[i].start && val <= FPK_map[i].end )
		{
			score = (val - FPK_map[i].start)- (FPK_map[i].end - FPK_map[i].start)/2;
//			printk("FPK_%d value %d score=%d\n",i,val, score);
			return( i );
		}
		if( val < FPK_map[i].end )
			break;
	}
//	printk("FPK_%d value %d %02x missed\n", i, val,inVal );
	return( -1 );
}


static void FPKReleaseAction(void)
{
	u8 key;
	u32 keyBit;

	// If there are any FPK keys to release
	if(FPKeyStateTable != 0 )
	{
		// Then release all of them
		for( key=0; key < KERNEL_FPK_KEY_RANGE; key++ )
		{
			keyBit = (1<<key);
			if( (FPKeyStateTable & keyBit) != 0 )
			{
//				printk("QSlider, FPKReleaseAction(), key = 0x%02x, state = %d\n", (KERNEL_FPK_KEY+key), KEY_RELEASED);
				QSlideEventNotify( KERNEL_FPK_KEY+key, KEY_RELEASED );
				FPKeyStateTable &= ~keyBit;
			}
		}
	}
}

static void FPKTouchAction( u8 value )
{
	u8 key;
	u32 keyBit;

	// First, convert slider scale value to FPK key number
	key = getFPKeyNumer(value);

	// Now process the keypress
	if( key < KERNEL_FPK_KEY_RANGE )
	{
		keyBit = 1<<key;
		/* check if we have already notified for this key */
		if( (keyBit & FPKeyStateTable) == 0 )
		{
			/* we have a new key */
			if( FPKeyStateTable != 0 )
			{
				/* we also have an old key */
				FPKReleaseAction();
			}
			FPKeyStateTable |= keyBit;
//			printk("QSlider, FPKTouchAction(), key = 0x%02x, state = %d\n", (KERNEL_FPK_KEY+key), KEY_PRESSED);
			QSlideEventNotify( KERNEL_FPK_KEY+key, KEY_PRESSED );
		}
	}
}

static u8 getVolumeKeyNumer( u8 val )
{
	val = (~val) & 0x7f;
//	printk("VOL value %d \n",val);
	return( val >> 3 );
}

static void SLDClearKeys(void)
{
	u8 key;
	u32 keyBit;

	if(volumeStateTable != 0 )
	{
		for( key=0; key < KERNEL_VOLUME_KEY_RANGE; key++ )
		{
			keyBit = (1<<key);
			if( (volumeStateTable & keyBit) != 0 )
			{
//				printk("QSlider, SLDClearKeys(), key = 0x%02x, state = %d\n", (KERNEL_VOLUME_KEY+key), KEY_RELEASED);
				QSlideEventNotify( KERNEL_VOLUME_KEY+key, KEY_RELEASED );
				volumeStateTable &= ~keyBit;
			}
		}
	}
}

static void SLDReleaseAction(void)
{
	u8 key;
	u32 keyBit;

	if (volumeStateTable != 0)
	{
		SLDClearKeys();
		if (volumeStateTable == 0)
		{
			/* The user has removed their finger */
			/* First, see if the slider position is up to date */
			if ((newVolKey != lastVolKeySent) && (newVolKey != KERNEL_NULL_KEY))
			{
				/* No it isn't, so send the last key press */
				key = newVolKey - KERNEL_VOLUME_KEY;
				keyBit = 1<<key;
				volumeStateTable |= keyBit;
//				printk("QSlider, SLDReleaseAction() volume, key = 0x%02x, state = %d\n", newVolKey, KEY_PRESSED);
				QSlideEventNotify( newVolKey, KEY_PRESSED );
				lastVolKeySent = newVolKey;

				// Then clear it
				SLDClearKeys();
			}

			/* Now release the slider */
//			printk("QSlider, SLDReleaseAction(), key = 0x%02x, state = %d\n", KERNEL_SLIDER_KEY, KEY_RELEASED);
			QSlideEventNotify( KERNEL_SLIDER_KEY, KEY_RELEASED );

			lastVolKeySent = KERNEL_NULL_KEY;
			newVolKey = KERNEL_NULL_KEY;
		}
	}
}

static void SLDTouchAction( u8 value )
{
	u8 key;
	u32 keyBit;

	if( (key = getVolumeKeyNumer(value)) < KERNEL_VOLUME_KEY_RANGE )
	{
		keyBit = 1<<key;
		newVolKey = KERNEL_VOLUME_KEY + key;
		/* Check if we have already notified for this key */
		if (newVolKey != lastVolKeySent)
		{
			/* We have a new key - so process it */

			if( lastVolKeySent == KERNEL_NULL_KEY )
			{
				/* There isn't an old key. therefore the slider has just been touched.
				 * Need to send a slider keypress as well as a volume keypress
				 */
//				printk("QSlider, SLDTouchAction() slider, key = 0x%02x, state = %d\n", KERNEL_SLIDER_KEY, KEY_PRESSED);
				QSlideEventNotify( KERNEL_SLIDER_KEY, KEY_PRESSED );

				sliderPollCount = 0;	// Force update this time
			}

			/* Now update the slider every third poll (if there is
			 * an update to do, that is). Note that this is not done
			 * for every poll in order to avoid sending too many
			 * messages too quickly to the application.
			 */
			if( sliderPollCount == 0 )
			{
				/* Slider position has moved since last update */
				/* See if there is an old key */
				if( lastVolKeySent != KERNEL_NULL_KEY )
				{
					/* There is - get rid of this first */
					SLDClearKeys();
				}

				/* Send new key press */
				volumeStateTable |= keyBit;
//				printk("QSlider, SLDTouchAction() volume, key = 0x%02x, state = %d\n", newVolKey, KEY_PRESSED);
				QSlideEventNotify( newVolKey, KEY_PRESSED );
				lastVolKeySent = newVolKey;
			}
			else
			{	// Debug only
//				printk("QSlider, SLDTouchAction() update skipped, key = 0x%02x\n", newVolKey);
			}
		}
	}
}

/**
* Poll timer time out function which schedules the QSlide hardware data poll
* and reschedules its self.
*/
static void QSlide_timeout(unsigned long data)
{
	QSlide_device_t *dev = (QSlide_device_t*)data;
	static int lost_flag = 0;
	unsigned long flags;

	/* Protect the following critical region */

	spin_lock_irqsave(&dev->lock,flags);

/**
* Diagnostic: print a warning message if the QSlide_task failed to run
* when it was queued last time.
*/
	if (dev->RunAquisition == 0)
	{
		lost_flag = 0;
		dev->RunAquisition = 1;
	}
	else
	{
		if (!lost_flag )
		{
//			printk(KERN_WARNING "Run Aquisition will be lost! (QSlider)\n");
			lost_flag = 1;
		}
	}

	/* Slider poll count is used to deteremine when to send updated slider position */
	sliderPollCount++;
	sliderPollCount %= SLIDER_POLL_FREQ;

	/* Critical region ends here */
	spin_unlock_irqrestore(&dev->lock,flags);

	/* Wake up the reading process */

	schedule_task(&QSlide_task);

//	wake_up_interruptible( &tl_wait_queue );
	/*
	 * Start a timer that expires in
	 * TIMER_INTERVAL_IN_MILLISECS ms and that
	 * calls tl_timeout on expiration. Hint: tl->timer
	 */

	dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ)/1000;
	add_timer(&dev->timer);

}

/**
* Poll timer time out function which schedules the QSlide hardware data poll
* and reschedules its self.
*/
static void QSlide_startTimer(QSlide_device_t *dev)
{

	dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ)/1000;
	add_timer(&dev->timer);
	spi_ss_sync(SS_MM_SLIDER);
	spi_ss_sync(SS_DM_KEY);
	spi_ss_sync(SS_MM_WHEEL);
	
}


/**
 * Function called by sheduled task queue by the timer timeout task.
 * It controls the physical hardware via inca_ip API function calls to poll the
 * hardware for data.
 *\param ptr pointer to kernel device timer structure
 *\returns pollData a byte of data read from the QSlide device.
 */

static poll_data_t QSlideMsg(spi_ss_t slaveSelect, poll_data_t msg)
{
	static poll_data_t pollData;
	poll_data_t nullPoll;

	nullPoll.poll = 0;
	pollData.poll = 0x7f;

	if( spi_ss_set(slaveSelect) == 0 )
	{
		return( pollData );
	}

	/* give time for chip to get its act together */
	udelay(15); /* Q411 this should be <= 13us */
	quantum_write_spi(&msg,sizeof(poll_data_t));

	/* read resposcc driver */
	quantum_read_spi(&pollData,sizeof(poll_data_t));

	/* at 32Khz it takes 250us to transmit 8 bits */
	udelay(50);

	/* raise /SS */
	spi_ss_clear(slaveSelect);

	udelay(15);  /* device holds onto SPI bus <= 12 us */
//	printk("polldata = %x\n",pollData.poll);

	return ( pollData );
}

static void QSlidePollActionFPK(void* pollFnData, u8 poll)
{
	// MR H00674 remove any possibility of 0XFF signal produced incorrectly when
	// chip decides to drop DRDY incorrectly
	if( ( poll != 0xFF ) && ((poll & 0x80 ) == 0x80) )
	{
		FPKTouchAction( poll );
	}
	else if( poll == 0 )
	{
		FPKReleaseAction();
	}
	else
	{
		printk(KERN_NOTICE "FPKSlider: Error %02X %d\n",poll, poll);
	}
}


static void QSlidePollActionSLD(void* pollFnData, u8 poll)
{
	if( (poll & 0x80 ) == 0x80 )
	{
		SLDTouchAction( poll );
	}
	else if( poll == 0 )
	{
		SLDReleaseAction();
	}
	else
	{
		printk(KERN_NOTICE "VolumeSlider: Error %02X %d\n",poll, poll);
	}
}


static int QwheelSampleUpdate( QWheel_state_t *state, char move )
{
	int i;
	int abs;
	char prev;
	char x;


	char *deltas = state->deltas;
	int	deltaIndex = state->deltaIndex;
	int	zeroTotal = state->zeroTotal;
	int	deltaTotal = state->deltaTotal;
	
	if( move == 0 )
	{
		zeroTotal++;
	}
	else
	{
		/* deltas[ deltaIndex -1] points at the newest entry */

		if( (prev = deltas[ (deltaIndex - zeroTotal -1) & 7 ]) != 0 )
		{
			x = move/(zeroTotal+1);
			if( ((prev ^ move) & 0x80) != 0 )
			{
				int dx;
				int dy;
				char z = move ^ 0x80;
				char y = z /(zeroTotal+1);
				
				dx = prev - x; /* dx is difference between two values */
				dy = prev - y;
				
				if( dx < 0 ) dx = -dx;
				if( dy < 0 ) dy = -dy;


				if( dy > dx )
				{
					deltaTotal = 0;
					for( i=0 ; i < 8; i++ )
						deltas[i] = 0;
				}
				else
				{
					x = y;
					move = z;
					state->flip++;
				}
			}
			abs = (deltaTotal > 0) ? deltaTotal : -deltaTotal;
			abs = abs >> 3;
			if( zeroTotal != 0 && abs != 0 && x != 0)
			{
				state->badsamples += zeroTotal;
				i = deltaIndex & 7;
//				printk(KERN_NOTICE "missed code? p=%d x=%d av=%d z=%d  ",prev, x, abs ,zeroTotal);
//				printk("<%d %d %d %d", deltas[i],deltas[(i+1)&7],deltas[(i+2)&7],deltas[(i+3)&7]);
//				printk(" %d %d %d %d> %d\n",deltas[(i+4)&7],deltas[(i+5)&7],deltas[(i+6)&7],deltas[(i+7)&7], move);
//
//				if( deltaIndex )
//					printk("polls=%d b=%d a=%d\n",deltaIndex, state->badsamples,(1000*state->badsamples)/deltaIndex);

				for( i=0;i<zeroTotal;i++ )
				{
					deltas[(deltaIndex - zeroTotal+i)& 7 ] = x;
					deltaTotal += x;
					move -= x;
				}				
			}
		}
		zeroTotal = 0;
	}

	deltaTotal -= deltas[ deltaIndex & 7 ]; /* subtract old value */
	deltaTotal += move; 					/* add new value in */
	deltas[ deltaIndex & 7 ] = move; 		/* and save it in the bucket */

	state->deltaIndex++;							/* now move on */
	state->deltaTotal = deltaTotal;
	state->zeroTotal = zeroTotal;
	return( deltaTotal );
}


static int QWheelSpeedABS( int deltaTotal )
{
	int abstotal = ( deltaTotal > 0 )? deltaTotal : -deltaTotal;
	int speed = (abstotal * 3125)/128;

	return( speed );
}


static int QWheelStep( int speed, int sensativity )
{
	static int table [5][16] = 
	{
		/* .  .  .  1  .  .  .  2  .  .  .  3  .  .  .  4 revs/second */
		{ 32,32,24,24,22,20,18,16,12, 8, 6, 5, 4, 3, 2, 1 },
		{ 24,24,16,16,16,16,16, 8, 8, 8, 6, 5, 4, 3, 2, 1 },
		{ 16,15,14,13,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1 },
		{ 12,12,12,12,11,10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1 },
		{  8, 8, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 1, 1 }
	};
	
	int step = speed / 250;
	
	if( step > 15 ) step = 15;

	if( sensativity > 4 ) sensativity = 4;


	return( table[sensativity][step] );
}

static void QWheelTouched( QWheel_state_t *state )
{
	int i;
	state->de = 0;
	state->polls = 0;
	state->zeroTotal = 0;
	state->badsamples = 0;
	state->flip = 0;
	state->deltaIndex = 0;
	state->deltaTotal = 0;
	state->zeroTotal = 0;
	state->lastPoll = 0;
	for( i=0 ; i < 8; i++ )
		state->deltas[i] = 0;
	QSlideEventNotify( KERNEL_WHEEL_KEY, KEY_PRESSED );

}

static void QWheelMotionStepped( QWheel_state_t *state, char move )
{
	int s,av, step;
	char event;
	char x;
	s = 0;

	av = QwheelSampleUpdate( state, move );
	s = QWheelSpeedABS( av );
	step = QWheelStep( s , state->setting );

	x = move;

	/* smooth the move once we reach 0.5 rev/sec. This will filter out reverse moves and spikes */
	if( s > 500 && move != 0 )
	{
		x = (av + 3) >>3;
//		if( x > 0 && move > x )
//			printk("spike:x=%d m=%d s=%d\n", x, move, s );				
//		if( x < 0 && move < x )
//			printk("spike:x=%d m=%d s=%d\n", x, move, s );				
//		if( move == 0 )
//			printk("miss:x=%d m=%d s=%d\n", x, move, s );				
	}

	state->de += x;
	state->polls++;
	
	event = state->de / step;

	if( event != 0 )
	{
//		printk("QWheel:e=%d st=%d p=%d x=%d m=%d de=%d sp=%d\n", event,step,state->polls, x, move, state->de, s );				
		state->de -= event * step;
		state->polls = 0;
		QWheelEventNotify( event );
		QWheelTestIfcEventNotify( event );
	}
}



static void QSlidePollActionWheel( void* pollFnData, u8 poll )
{
	char last,current;
	char move,imove;

	QWheel_state_t *state = (QWheel_state_t *) pollFnData;

	if( IS_TOUCHED(poll) )
	{
		if(state->touched == 0)
		{
			state->touched = 1;
			QWheelTouched( state );
		}
		else
		{
			last = QWHEEL_DATA(state->lastWrite) << 1;
			current =  QWHEEL_DATA(poll) << 1;
			imove = (last - current);
			move = imove >> 1;

			QWheelMotionStepped( state, move );

		}
		state->lastWrite = poll;
	}
	else if(state->touched != 0)
	{
		if( poll == 0x40 )
		{
			state->touched = 0;
			QSlideEventNotify( KERNEL_WHEEL_KEY, KEY_RELEASED );
		}
		else
		{
			printk("QWheel:bad poll in touch %02x\n", poll );				
		}
	}
}


void stateInit(slider_state_t *machine, u8 *initTable)
{
	machine->ss = *initTable++;
	machine->type = *initTable++;
	machine->send = *initTable++;
	machine->state = *initTable++;
	machine->sent = 0;
	machine->recv = 0;
	machine->sensitivity = *initTable++;
	machine->driftComp = *initTable++;
	machine->pollcount = machine->driftComp;
	machine->totalPolls = 0;
	machine->totalCalibrate = 0;
	machine->totalDriftCompensate = 0;
	machine->totalErrorStatus = 0;
	machine->totalErrorDetect = 0;
	machine->totalSetSensitivity = 0;
	machine->totalDataNotReady = 0;
	machine->totalDataReady = 0;
	machine->firstDetection = 0;
    machine->firstCalibration = 0;
//	machine->FreezeCount = SLIDER_HISTORY_SIZE; /* use to capture initialisation */
	machine->FreezeCount = -1;
	machine->HistoryCount = 0;
//	printk("stateInit:ss=%d send=%02x state=%02x\n",machine->ss,machine->send, machine->state);
}

void statePoll(slider_state_t *machine)
{
	machine->totalPolls++;
    
	if( machine->recv == (0x02 | machine->type))
	{
		machine->totalPollErrorBit++;
	}
	
    if (machine->firstCalibration == 0) //haven't tried to calibrate yet, so do it
    {
        //printk("Calibration required for ss %d\n", machine->ss);
        machine->firstCalibration = 1;
        machine->state = machine->send = 0x01;
    }
    else if( ((0x80 & machine->recv) == 0x80)
        && machine->firstDetection == 0) //first touch, so check for calib. status
    {
        machine->state = machine->send = 0x04;
        machine->firstDetection = 1;
        //printk("First touch on ss %d\n", machine->ss);
    }
	else if( machine->pollcount == 0 )
	{
		machine->state = machine->send = 0x03;
	}
	else
	{
		machine->pollcount--;
		machine->send = 0x00;
	}
	
	if(machine->recv & 0x80)
	{
//		printk("statePoll:ss=%d recv=%d\n",machine->ss,machine->recv & 0x7f);
	}
	else if( machine->recv != machine->type)
	{
		//printk("statePoll:%s ss=%d recv=%02x bad poll\n",machine->name,machine->ss,machine->recv );
	}

	if( machine->pollFn != 0 )
	{
		machine->pollFn(machine->pollFnData,machine->recv);
	}
}

void stateCalibrate(slider_state_t *machine)
{
	printk("Calibrating: %s ss=%d\n",machine->name,machine->ss);

	machine->totalCalibrate++;
    machine->firstDetection = 0;
    machine->firstCalibration = 1;
	machine->state = machine->send = 0x00;
}

void stateCompensate(slider_state_t *machine)
{
	machine->totalDriftCompensate++;
//	if( (machine->totalDriftCompensate % 10 ) == 0 )
//		printk("Drift Compensation: %s ss=%d Number %d\n",machine->name,machine->ss,machine->totalDriftCompensate/10);
	machine->pollcount = machine->driftComp;
	machine->state = 0x00;
	statePoll(machine);
}

void stateErrorStatus(slider_state_t *machine)
{
	if( machine->sent == 0x04 )
	{
		machine->totalErrorStatus++;
		machine->send = 0x00;
	}
	else if( machine->sent == 0x00 )
	{
		if( (machine->recv & 0x02) == 0x02)
		{
			printk("Calibration Error: %s ss=%d\n",machine->name,machine->ss);
			machine->state = machine->send = 0x01;
		}
		else
		{
            printk("Calibration OK\n");
			machine->send = 0x80 | machine->sensitivity;
			machine->state = 0x80;
		}
	}
}

void stateSetThreshold(slider_state_t *machine)
{
	machine->totalSetSensitivity++;
	machine->state = 0x00;
	machine->send = 0x00;
	statePoll(machine);
}

void statePollFirstTouch(slider_state_t *machine)
{
	if( machine->sent == 0x04 )
	{
		machine->send = 0x00;
	}
	else
	{
		machine->send = 0x04;
		if( machine->recv & 0x02 )
		{
			machine->state = machine->send = 0x01;
			printk("Re-Calibrating: %s ss=%d recv=%02x\n",machine->name,machine->ss,machine->recv);
		}
		else if( machine->recv & 0x80 )
		{
			machine->state = machine->send = 0x00;
			printk("Polling: %s ss=%d\n",machine->name,machine->ss);
		}
	}
}

void QuantumStateMachine( slider_state_t *machine )
{

	if( machine->recv == 0x7f )
	{
		machine->totalDataNotReady++;
		return;
	}

	machine->sent = machine->send;
	machine->totalDataReady++;

	switch( machine->state )
	{
		case 0x00: statePoll(machine); 			break;
		case 0x01: stateCalibrate(machine); 	break;
		case 0x03: stateCompensate(machine); 	break;
		case 0x04: stateErrorStatus(machine); 	break;
		case 0x05: statePollFirstTouch(machine); 	break;
		case 0x80: stateSetThreshold(machine); 	break;
		default: printk(KERN_NOTICE "QuantumStateMachine: Error %d\n",machine->state); return;
	}
}


void QSliderHistory( slider_state_t *machine )
{
/*	if( machine->recv == 0x42 || machine->recv == 0x02 )
		machine->FreezeCount = SLIDER_HISTORY_SIZE;
*/
	if( machine->FreezeCount > 0 )
	{
		machine->FreezeCount--;		
	}
	
	if( machine->FreezeCount != 0 )
	{
		machine->History[machine->HistoryCount++ & (SLIDER_HISTORY_SIZE-1)] = (machine->send << 8) | machine->recv;
	}
}


void QuantumRead( slider_state_t *machine )
{
	poll_data_t send;
	poll_data_t recv;

	send.poll = machine->send;
	recv = QSlideMsg(machine->ss, send );
	machine->recv = recv.poll;
	QSliderHistory(machine);
}

/**
 * Function placed on the sheduled task queue by the timer timeout task.
 * It polls the hardware for data and puts it into the incoming data queue buffer.
 */
void QSlideServiceTask(void *ptr)
{
	unsigned long flags;
	
	QSlide_device_t* dev = (QSlide_device_t*) ptr;

	spin_lock_irqsave(&dev->lock,flags);
	dev->RunAquisition = 0;
	/* Critical Region ends here */
	spin_unlock_irqrestore(&dev->lock,flags);

	QuantumRead( &FPKmachine );
	QuantumRead( &WHLmachine );
	QuantumRead( &VOLmachine );
	QLedOutput();
	QuantumStateMachine( &FPKmachine );
	QuantumStateMachine( &WHLmachine );
	QuantumStateMachine( &VOLmachine );
}

void QSlideServiceTask_WP2(void *ptr)
{
	unsigned long flags;
	
	QSlide_device_t* dev = (QSlide_device_t*) ptr;

	spin_lock_irqsave(&dev->lock,flags);
	dev->RunAquisition = 0;
	/* Critical Region ends here */
	spin_unlock_irqrestore(&dev->lock,flags);

	QuantumRead( &FPKmachine );
	QLedOutput();
	QuantumStateMachine( &FPKmachine );
}




/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *Qvol_proc;
static struct proc_dir_entry *Qfpk_proc;
static struct proc_dir_entry *Qwhl_proc;
static struct proc_dir_entry *QWc_proc;


static int QSlide_proc_one(char *buf, int len, slider_state_t *machine )
{
	unsigned long flags;
	int i;
	int historySize;
	int historyIndex;
	spinlock_t lock;

	spin_lock_init(&lock);

	spin_lock_irqsave(&lock,flags);

	historyIndex = 0;
	historySize = machine->HistoryCount;

	len += sprintf( buf+len,"%s history count %d\n",machine->name, historySize );

	if( historySize > SLIDER_HISTORY_SIZE )
	{
		historySize = SLIDER_HISTORY_SIZE;
		historyIndex = machine->HistoryCount;
		/* list starting at oldest */
	}

	for( i=0; i < historySize; i++ )
	{
		if( (i & 0xf) == 0xf )
			len += sprintf(buf+len,"%04X\n",machine->History[ historyIndex & (SLIDER_HISTORY_SIZE-1) ] );
		else
			len += sprintf(buf+len,"%04X ",machine->History[ historyIndex & (SLIDER_HISTORY_SIZE-1) ] );

		historyIndex++;
	}

	spin_unlock_irqrestore(&lock,flags);
	
	len += sprintf(buf+len,"\nCalibrated(0x1)=%d, ErrorQuery(0x4)=%d Error=%d ErrorPoll=%d\n\n" , machine->totalCalibrate, machine->totalErrorStatus, machine->totalErrorDetect, machine->totalPollErrorBit);

	return len;
}

static int QSlide_cooked_proc_one(char *buf, int len, slider_state_t *machine, QWheel_state_t *state  )
{
	unsigned long flags;
	int i;
	int historySize;
	int historyIndex;
	char last = 0;
	char col = 0 ; /* blk=0 red=1 grn=2 yel=3 blue=4 mage=5 cian=6 white=7 */
	int x = 0;

	spinlock_t lock;

	spin_lock_init(&lock);

	spin_lock_irqsave(&lock,flags);

	historyIndex = 0;
	historySize = machine->HistoryCount;

	len += sprintf( buf+len,"%s history count %d\n",machine->name, historySize );

	if( historySize > SLIDER_HISTORY_SIZE )
	{
		historySize = SLIDER_HISTORY_SIZE;
		historyIndex = machine->HistoryCount;
		/* list starting at oldest */
	}

	for( i=0; i < historySize; i++ )
	{
		char next =  machine->History[ historyIndex & (SLIDER_HISTORY_SIZE-1) ];

		if( ((next & 0x80) != 0) &&  ((last & 0x80) != 0) )
		{
			char current;
			char imove;
			char move;

			current = (next & 0x7f) << 1;
			last = (last & 0x7f) << 1;
			imove = (last - current);
			move = imove >> 1;

			x = x + (int)move;
			if( (x > 127) || (x < -127) )
			{
				col = (col == 0) ? 4 : 0;
				x = 0;
				len += sprintf(buf+len,"\033[3%dm",col);
			}
			
			len += sprintf(buf+len,"% 3d ",move );
		}
		else if((next & 0x80) != 0)
		{
				len += sprintf(buf+len,"  T " );
				x = 0;
		}
		else
				len += sprintf(buf+len,"  . " );

		if( (i & 0xf) == 0xf )
			len += sprintf(buf+len,"\n" );
		historyIndex++;
		last = next;
	}

	spin_unlock_irqrestore(&lock,flags);
	
	len += sprintf(buf+len,"\033[0m\nCalibrated(0x1)=%d, ErrorQuery(0x4)=%d Error=%d ErrorPoll=%d\n" , machine->totalCalibrate, machine->totalErrorStatus, machine->totalErrorDetect, machine->totalPollErrorBit);
	if( state->deltaIndex )
			len += sprintf(buf+len,"Samples=%d bad=%d av=%d flip=%d\n\n",state->deltaIndex, state->badsamples,(1000*state->badsamples)/state->deltaIndex,state->flip);


	return len;
}

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/QSlide'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int Qvol_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	len = QSlide_proc_one( buf, len, &VOLmachine );
	return len;
}

static int Qfpk_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	len = QSlide_proc_one( buf, len, &FPKmachine );
	return len;
}

static int Qwhl_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	len = QSlide_proc_one( buf, len, &WHLmachine );
	return len;
}

static int Qwhl_cooked_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	len = QSlide_cooked_proc_one( buf, len, &WHLmachine, &QWheelstate );
	return len;
}


static int asciitoint( char *s )
{
	int val = 0;
	if( s )
	{
		while( (*s >= '0') && (*s <= '9') )
		{
			val *= 10;
			val += *s - '0';
			s++;
		}
	}
	return( val );
}
/**
* Function to handle a proc entry write not used
*/

static int QSlide_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	char buff[50];
	int len;
	
	len = ( count < sizeof(buff) )? count : sizeof(buff);
			
	memcpy( buff, buffer,len);
	buff[len-1] = 0;
	printk(KERN_INFO "QSlide Write procmem (%s)\n",buff);

	if( strcmp("calibrate wheel",buff) == 0 )
	{
		WHLmachine.firstCalibration = 0;
	}
	if( strcmp("calibrate vol",buff) == 0 )
	{
		VOLmachine.firstCalibration = 0;
	}
	if( strcmp("calibrate fpk",buff) == 0 )
	{
		FPKmachine.firstCalibration = 0;
	}

	if( strcmp("status wheel",buff) == 0 )
	{
		WHLmachine.state = WHLmachine.send = 0x04;
	}
	if( strcmp("status vol",buff) == 0 )
	{
		VOLmachine.state = VOLmachine.send = 0x04;
	}
	if( strcmp("status fpk",buff) == 0 )
	{
		FPKmachine.state = FPKmachine.send = 0x04;
	}

	if( strncmp("set ",buff,4) == 0 )
	{
		QWheelSetSensitivity( asciitoint( &buff[4] ) );
	}

	if( strncmp("mark ",buff,5) == 0 )
	{
		QWheelSetMark( asciitoint( &buff[5] ) );
	}

	return count;
}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init QSlide_init(void)
{
	keyRegInterface_t registration;
	QWheelRegInterface_t wheelRegistration;
	QWheelTestIfRegInterface_t wheelTestIfcRegistration;

	/* create PROC entry */
	if( (Qvol_proc = create_proc_entry( "driver/Qvol", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/Qvol!\n");
		return( -ENOENT );
	}

	if( (Qfpk_proc = create_proc_entry( "driver/Qfpk", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/Qfpk!\n");
		remove_proc_entry("driver/Qvol", 0);
		return( -ENOENT );
	}

	if( (Qwhl_proc = create_proc_entry( "driver/Qwhl", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/Qwhl!\n");
		remove_proc_entry("driver/Qvol", 0);
		remove_proc_entry("driver/Qfpk", 0);
		return( -ENOENT );
	}

	if( (QWc_proc = create_proc_entry( "driver/QWc", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/QWc!\n");
		remove_proc_entry("driver/Qvol", 0);
		remove_proc_entry("driver/Qfpk", 0);
		remove_proc_entry("driver/Qwhl", 0);
		return( -ENOENT );
	}

	Qvol_proc->read_proc  = Qvol_read_procmem;
	Qvol_proc->write_proc = QSlide_write_procmem;
	Qfpk_proc->read_proc  = Qfpk_read_procmem;
	Qfpk_proc->write_proc = QSlide_write_procmem;
	Qwhl_proc->read_proc  = Qwhl_read_procmem;
	Qwhl_proc->write_proc = QSlide_write_procmem;
	QWc_proc->read_proc  = Qwhl_cooked_read_procmem;
	QWc_proc->write_proc = QSlide_write_procmem;

	
	spin_lock_init(&QSlide_data.lock);

	printk(KERN_INFO "QSlider Poll Rate %dms\n",TIMER_INTERVAL_IN_MILLISECS);

	if(get_WP_type() == 2)
	{
		/*This is a WP2 display*/
		
		 numberOfFPKs = 6;	
	}
	else if(get_WP_type() == 3)
	{
		/*This is a WP3 display*/
		
		 numberOfFPKs = 8;	
	}
	else numberOfFPKs = 9;

	init_timer(&QSlide_data.timer);
	init_FPK_map();

	QSlide_data.RunAquisition = 0;

	
	QSlide_data.timer.data = (int)&QSlide_data;
	QSlide_data.timer.function = QSlide_timeout;

	if( (get_WP_type() == 3) || (get_WP_type() == 4) )
	{
	initQWheelState( &QWheelstate );
		volumeStateTable = 0;
	QSlide_task.routine = QSlideServiceTask; /*scheduled function*/
	}
	else
	{
		QSlide_task.routine = QSlideServiceTask_WP2; /*scheduled function*/
	}
	
	QSlide_task.data = &QSlide_data; /*argument to function*/
	FPKeyStateTable = 0;
	
	QSlide_callback = 0;
	registration.keyInputEvent = 0;
	registration.name = "QSlider";
	if( keyInputRegister( &registration ) == 0 )
	{
		QSlide_callback = registration.keyInputEvent;
	}

	if( (get_WP_type() == 3) || (get_WP_type() == 4) )
	{
		wheelRegistration.QWheelInputEvent = 0;
		wheelRegistration.name = "QSlider";
	
		if( QWheelInputRegister( &wheelRegistration ) == 0 )
		{
			QWheel_callback = wheelRegistration.QWheelInputEvent;
		}
	
		wheelTestIfcRegistration.QWheelInputEvent = 0;
		wheelTestIfcRegistration.QWheelSetSensitivity = QWheelSetSensitivity;
		wheelTestIfcRegistration.name = "QSlider";
	
		if( QWheelTestIfcRegister( &wheelTestIfcRegistration ) == 0 )
		{
			QWheelTestIfc_callback = wheelTestIfcRegistration.QWheelInputEvent;
		}
		VOLmachine.name = "slider: ";
		WHLmachine.name = "wheel:  ";
		VOLmachine.pollFn = QSlidePollActionSLD;
		WHLmachine.pollFn = QSlidePollActionWheel;
		VOLmachine.pollFnData = 0;
		WHLmachine.pollFnData = &QWheelstate;
		stateInit( &VOLmachine, VOLmachineStartTable );
		stateInit( &WHLmachine, WHLmachineStartTable );
	
		VOLmachine.sensitivity = sliderset;
		WHLmachine.sensitivity = wheelset;
	
		VOLmachine.send = 0x80 | sliderset;
		WHLmachine.send = 0x80 | wheelset;
	
		printk("Wheel  sensitivity (%d)\n",wheelset);
	
		lastVolKeySent = KERNEL_NULL_KEY;
		newVolKey = KERNEL_NULL_KEY;

	}

	FPKmachine.name = "fpk:    ";
	FPKmachine.pollFn = QSlidePollActionFPK;	
	FPKmachine.pollFnData = 0;
		
	stateInit( &FPKmachine, FPKmachineStartTable );

	if(get_WP_type() == 4)
	{
		FPKmachine.sensitivity = fpkset_WP4;
		FPKmachine.send = 0x80 | fpkset_WP4;
		printk("FPK    sensitivity (%d)\n",fpkset_WP4);
	}
	else	// it must be WP2 or WP3
	{
		FPKmachine.sensitivity = fpkset_WP;
		FPKmachine.send = 0x80 | fpkset_WP;
		printk("FPK    sensitivity (%d)\n",fpkset_WP);
	}

	printk("Slider sensitivity (%d)\n",sliderset);


	sliderPollCount = 0;

	QSlide_startTimer(&QSlide_data);
	
	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit QSlide_exit(void)
{
	keyRegInterface_t registration;
	QWheelRegInterface_t wheelRegistration;
	QWheelTestIfRegInterface_t wheelTestIfcRegistration;

	/* Delete the timer */
	del_timer_sync( &QSlide_data.timer );
	printk(KERN_INFO "QSlider timer stopped\n");

	remove_proc_entry("driver/Qvol", 0);
	remove_proc_entry("driver/Qfpk", 0);
	remove_proc_entry("driver/Qwhl", 0);
	remove_proc_entry("driver/QWc", 0);

	if( (get_WP_type() == 3) || (get_WP_type() == 4) )
	{		
		wheelRegistration.name = "QSlider";
		if( QWheelInputUnRegister( &wheelRegistration ) == 0 )
		{
		}
		
		wheelTestIfcRegistration.name = "QSlider";
		if( QWheelTestIfcUnRegister( &wheelTestIfcRegistration ) == 0 )
		{
		}
	}
	
	registration.name = "QSlider";
	if( keyInputUnRegister( &registration ) == 0 )
	{
	}
	
	QSlide_callback = 0;

	printk(KERN_NOTICE "QSlide_exit\n");
}



module_init(QSlide_init);
module_exit(QSlide_exit);

//EXPORT_SYMBOL(fpkset);

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
