/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/** \file QTouch.c
* The QTouch driver interfaces to the QT1100A (Key QTouch sensor) device using
* the SPI hardware of the INCA-IP2. Once the QTouch device is reset and the SetupBlock
* (36 bytes: that controls the internal operation of the QTouch device) is sent to
* the device, the device reports the key status (touched/released) by means of interrupts
*  using QTouch CRDY signal.
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */
#include <asm/incaip2/tsf.h>


#include "fpga.h"
#include "fpga_defs.h"
#include "Quantum_drivers.h"
#include "keyinput.h"


/**
* Development debug print switch
*/
//#define QT_DEBUG

#define QTOUCH_ACTON_ERRORS	(1) /* Ignore Error bit in KEY */



/**
* Structure for QTouch key & state machine data
*/
typedef struct {
	spinlock_t lock;
	int IsrIgnorecount;

	/* used to progress our state machine */
	u32 keystate;
	int deviceState;
	int index;
	int count;
	const u8  *list;
	u8  reply[40];
	u32 errorActionCount;
	u32 notUnderStoodCount;
	u32 notReadyCount;
	u32 KeyErrorCount;
	u32 MulipleKeyCount;
	u32 LateIsrCount;
	u32 TotalIsrCount;
	u32 TotalKeyCount;
	u32 KeyCounts[10];
} QTouchDevice_t;

/**
* An instance of the Qtouch key & driver state machine data structure
*/
static QTouchDevice_t QTouch_Device;
keypad_callback_t_ul QTouch_callback = 0;
/**
* Pointer to an instance of the Qtouch key & driver state machine data structure
*/
QTouchDevice_t* qtDevPtr;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/
DECLARE_WAIT_QUEUE_HEAD(qTouch_wait_queue);

/**
* Declartion of a task queue structure
*/
struct tq_struct QTouch_task; /* global: initialized to zero */




static const unsigned char QTouch_default_SetupBlock[35] = {
	0x93,0x93,0x93,0x93,0x93,0x93,0x93,0x93,0x93,0x93,
	0x75,0x75,0x75,0x75,0x75,0x75,0x75,0x75,0x75,0x75,
	0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,
	0x51,0x57,0x50,0x03,0x69
	};



static void QTouchKernelJob(void *ptr);


#define NUM_QTOUCH_STATES 20
#define NUM_QTOUCH_EVENTS 4
#define NA -1
#define RESULT int

/**
* action table function pointer definition
*/
typedef	RESULT	(*QT_ACTION)(QTouchDevice_t* devPtr,...);


static RESULT QTouchStartStatusCmd( QTouchDevice_t *cmd );
static RESULT QTouchGettingStatus( QTouchDevice_t *cmd );
static RESULT QTouchStartRetLastCmd( QTouchDevice_t *cmd );
static int QTouchGettingLastCmd( QTouchDevice_t *cmd );
static int QTouchPrepareSetup( QTouchDevice_t *cmd );
static int QTouchStartSetupCmd( QTouchDevice_t *cmd );
static int QTouchStartDownloadCmd( QTouchDevice_t *cmd );
static int QTouchDownloading( QTouchDevice_t *cmd );
static int QTouchFinish( QTouchDevice_t *cmd );
static int QTouchStartCalibrationCmd( QTouchDevice_t *cmd );
static int QTouchCalibrating( QTouchDevice_t *cmd );
static int QTouchStartKeyAquisitionCmd( QTouchDevice_t *cmd );
static int QTouchKeyAquisition( QTouchDevice_t *cmd );
/**
* QTouch driver state table, each row indicating the current state, and each column
* indicating the next state entry depending on the event that has occured.
*/
static const int qTouchFsm[NUM_QTOUCH_STATES][NUM_QTOUCH_EVENTS] = {
/* EVENT
 *     0, 1, 2, 3 */
	{ NA, 1,NA,NA },	/* STATE 0 */
	{ NA, 2,NA,NA },	/* STATE 1 QTouchStartStatusCmd */
	{ NA, 2, 3,NA },	/* STATE 2 QTouchGettingStatus */
	{ NA, 4,NA,NA },	/* STATE 3 QTouchStartRetLastCmd */
	{ NA, 4, 5,NA },	/* STATE 4 QTouchGettingLastCmd */
	{ NA, 6,NA,NA },	/* STATE 5 QTouchStartSetupCmd */
	{ NA, 6, 7,NA },	/* STATE 6 QTouchPrepareSetup */
	{ NA, 8,NA,NA },	/* STATE 7 QTouchStartDownloadCmd */
	{  3, 8, 9,NA },	/* STATE 8 QTouchDownloading */
	{ NA,10,NA,NA },	/* STATE 9 QTouchStartCalibrationCmd */
	{ NA,10,11,NA },	/* STATE 10 QTouchCalibrating */
	{ NA,12,NA,NA },	/* STATE 11 QTouchStartStatusCmd */
	{ NA,12,18,NA },	/* STATE 12 QTouchGettingStatus*/
	{ NA,14,NA,NA },	/* STATE 13 QTouchStartKeyAquisitionCmd */
	{ NA,14,1,NA },		/* STATE 14 QTouchKeyAquisition */
	{ NA,NA,NA,NA }, 	/* STATE 15 QTouchFinish */
	{ NA,17,NA,NA }, 	/* STATE 16 QTouchStartStatusCmd */
	{ NA,17,13,NA }, 	/* STATE 17 QTouchGettingStatus */
	{ NA,19,NA,NA }, 	/* STATE 18 QTouchStartRetLastCmd */
	{ NA,19,16,NA }, 	/* STATE 19 QTouchGettingLastCmd */
};

/****************************************************************/
/*	Action Table for QTouch Control Object						*/
/****************************************************************/


static  QT_ACTION qTouchActions[NUM_QTOUCH_STATES] =
{
	(QT_ACTION)0,							/* state 0 */
	(QT_ACTION)QTouchStartStatusCmd,		/* state 1 */
	(QT_ACTION)QTouchGettingStatus,			/* state 2 */
	(QT_ACTION)QTouchStartRetLastCmd,		/* state 3 */
	(QT_ACTION)QTouchGettingLastCmd,		/* state 4 */
	(QT_ACTION)QTouchStartSetupCmd,			/* state 5 */
	(QT_ACTION)QTouchPrepareSetup,			/* state 6 */
	(QT_ACTION)QTouchStartDownloadCmd,		/* state 7 */
	(QT_ACTION)QTouchDownloading,			/* state 8 */
	(QT_ACTION)QTouchStartCalibrationCmd,	/* state 9 */
	(QT_ACTION)QTouchCalibrating,			/* state 10 */
	(QT_ACTION)QTouchStartStatusCmd,		/* state 11 */
	(QT_ACTION)QTouchGettingStatus,			/* state 12 */
	(QT_ACTION)QTouchStartKeyAquisitionCmd,	/* state 13 */
	(QT_ACTION)QTouchKeyAquisition,			/* state 14 */
	(QT_ACTION)QTouchFinish,				/* state 15 */
	(QT_ACTION)QTouchStartStatusCmd,		/* state 16 */
	(QT_ACTION)QTouchGettingStatus,			/* state 17 */
	(QT_ACTION)QTouchStartRetLastCmd,		/* state 18 */
	(QT_ACTION)QTouchGettingLastCmd,		/* state 19 */
};

/**
* QT_NULL_ACTION routine return an valid index 0 ie to the state function pointer array qTouchActions
*
*/
static RESULT QT_NULL_ACTION( QTouchDevice_t* devPtr,... )
{
	return( 0 );
}

/**
* QT_ERROR_ACTION routine return an invalid index (-1) to the state function pointer array qTouchActions
*/
static RESULT QT_ERROR_ACTION( QTouchDevice_t* devPtr,... )
{
	/*
	printk("fsm error: s=%d\n",devPtr->deviceState );
	devPtr->deviceState = 0;
	*/
	devPtr->errorActionCount++;
	return( -1 );
}

/**
* qTouchNextAction routines acquires the next state depending on the
* event from the QTouch driver state table "qTouchFsm" row = current
* sate, column = event.
*\param devPtr pointer to Qtouch key & driver state machine data structure
*\param event actin event
*\return index to state function pointer array (qTouchActions)
*/
static QT_ACTION qTouchNextAction( QTouchDevice_t* devPtr, int event )
{
	static int error;
	int state = qTouchFsm[devPtr->deviceState][event];

//	if( error == 0 )
//		printk("fsm ev=%d s=%d->%d\n",event,devPtr->deviceState,state );

	if (state > 0)
	{
		devPtr->deviceState = state;
		return( qTouchActions[state] );
	}
	else if (state < 0)
	{
		error = 1;
		return( QT_ERROR_ACTION );
	}
	return( QT_NULL_ACTION );
}


/****************************************************************/
/*	Example Event handler for events							*/
/****************************************************************/
/**
* qTouchErrorEvent routine is called when error occurs on sending
* command or setup message to the device. The next state to run is
* taken from the QTouch driver state table "qTouchFsm" row =
* curren sate; column = 0.
*\param devPtr pointer to Qtouch key & driver state machine data structure
*\return index to state function pointer array (qTouchActions)
*/
static RESULT	qTouchErrorEvent( QTouchDevice_t* devPtr )
{
	return( (qTouchNextAction( devPtr, 0 ))( devPtr ) );
}
/**
* qTouchDrdyEvent routine is called at every interrupt. The next state
* to run is taken from the QTouch driver state table "qTouchFsm"
* row = current sate; column = 1.
*\param devPtr pointer to Qtouch key & driver state machine data structure
*\return next state
*/
static RESULT	qTouchDrdyEvent( QTouchDevice_t* devPtr )
{
	return( (qTouchNextAction( devPtr, 1 ))( devPtr ) );
}
/**
* qTouchDoneEvent routine is called when sending the curren command or
* setup message is sent to device successfully. The next state to run is
* taken from the QTouch driver state table "qTouchFsm" row =
* curren sate; column = 2.
*\param devPtr pointer to Qtouch key & driver state machine data structure
*\return next state
*/
static RESULT	qTouchDoneEvent( QTouchDevice_t* devPtr )
{
	return( (qTouchNextAction( devPtr, 2 ))( devPtr ) );
}

int QTouchEventNotify( u8 event, int state )
{
	if( QTouch_callback != 0 )
	{
		return( QTouch_callback( (unsigned long)event, state, jiffies ) );
	}
	else
	{
//		printk(KERN_WARNING "QSlideEventNotify no callback Event will be lost!\n");
	}
	return( 0 );
}


/**
* Called from a task queue to read_write from_to device
*/
static void QTouchKernelJob(void *ptr)
{
QTouchDevice_t *devPtr = (QTouchDevice_t *)ptr;
static int oldPortState;
int portState;


	portState = spi_drdy(SS_MM_KEY);

	devPtr->TotalIsrCount++;
	if( portState == 0 )
	{
#ifdef QT_DEBUG
		printk("QTouchKernelJob ISR Late %02x\n", (unsigned char)portState);
#endif
		devPtr->LateIsrCount++;
	}

	if( portState != oldPortState )
	{
#ifdef QT_DEBUG
//		printk("QTouchKernelJob Device Not Ready %02x\n", (unsigned char)portState);
#endif
		oldPortState = portState;
	}

	if( portState != 0 )
	{
#ifdef QT_DEBUG_DRDY
//		printk("QTouch DRDY event %d\n", (unsigned char)portState);
#endif
		qTouchDrdyEvent(devPtr);
	}
}

/**
* QTouch_isr routine is called when QTouch device is ready to talk
* (i.e. upon interrupt from QTouch device). This routine clears
* the interrupt and puts the QTouch_task in the task queue.
*\param irq the interrupt number being requested
*\param dev_id pointer used for shared interrupt lines
*\param *regs holds a snapshot of the processor's context before
*the processor entered interrupt code
*/
int QTouch_isr( eFpgaInterruptNum vector, void *data )
{
QTouchDevice_t * devPtr = (QTouchDevice_t *)data;
/*
	inca_port_clear_output(CS_PORT,CSQ_KEY_PIN,PORT_MODULE_ID);
	*/

//	printk(KERN_NOTICE "QTouch_isr vector %d  ISR count %d\n",vector,devPtr->IsrIgnorecount);

	if( devPtr->IsrIgnorecount == 0 )
		schedule_task(&QTouch_task);
	else
		devPtr->IsrIgnorecount--;

	return( 0 );
}

int QTouchKey_isr( eFpgaInterruptNum vector, void *data )
{
	printk(KERN_NOTICE "QTouch Key vector %d\n",vector);
	disable_fpga_irq( FPGA_MM_KEY_INT );

	return( 0 );
}


/**
* QTouch_isr_init routine installs the CDRY (QTouch device Data Ready)
* as an external interrupt (EXIN2).
*\returns 0 to indicate the interrupt is installed successfully, or a
* negative code to indicate an error status.
*/
int QTouch_isr_init(QTouchDevice_t *dev)
{
	int err=0;
	request_fpga_irq( FPGA_MM_KEY_DRDY_INT, QTouch_isr, dev );
	request_fpga_irq( FPGA_MM_KEY_INT, QTouchKey_isr, dev );

	enable_fpga_irq( FPGA_MM_KEY_DRDY_INT );
	enable_fpga_irq( FPGA_MM_KEY_INT );
	
	return( err );
}


/**
* 8-bit crc calculation. Initial value is 0.
* polynomial = X8 + X5 + X4 + 1
* data is an 8 bit number; crc is an 8-bit number
*\param crc calculated CRC so far
*\param data further data
*\return temp calculated crc
*/
static int QTouchCalcOctetsCRC(int crc, int data)
{
	int index; // shift counter
	int fb;
	index = 8; // initialise the shift counter
	do
	{
		fb = (crc ^ data) & 0x01;
		data >>= 1;
		crc >>= 1;
		if(fb)
		{
			crc ^= 0x8c;
		}
	} while(--index);
	return crc;
}
/**
* QTouchCRC routine calculates the CRC
*\param buf pointer to data buffer
*\param len length of data buffer
*\return final Calculated CRC
*/

static int QTouchCRC(char *buf, int len)
{
	unsigned char crc;
	int i;
	for( i=0, crc=0; i < len; i++ )
	{
		crc = QTouchCalcOctetsCRC( crc, buf[i] );
	}
	return( crc );
}


/*****************************************************************
 * QTouchSend returns octets sent
 *****************************************************************/
/**
* QTouchSend routine reads and writes data to/from the QTouch device
*\param txByte transmit byte
*\param rxByte receive byte
*\return 1 if Qtuch device is ready, else 0.
*/
static int QTouchSend(u8 txByte, u8 *rxByte)
{
	int result = 1;

	if( spi_ss_set(SS_MM_KEY) == 0 )
	{
		/* QTouch device not ready. */
#ifdef QT_DEBUG
		printk("QTouch device not ready for %x\n",txByte );
#endif
		return( 0 );
	}
	/* write poll to ssc driver*/
	quantum_write_spi(&txByte,sizeof(txByte));

	/* at 32Khz it takes 250us to transmit 8 bits */


	*rxByte = 0xff;

	/* read response back from scc driver */
	quantum_read_spi(rxByte,sizeof(txByte));
	/* raise /SS */
	udelay(250);

	spi_ss_clear(SS_MM_KEY);
/*
	printk("sent=%x reply=%x\n",txByte, *rxByte );
*/
	return result;
}

/**
* QTouchSendCommand routine sends command that is in the list buffer
* to the device and places the response from the device in the reply
* buffer.
*\param cmd pointer to QTouch key & driver state machine data structure
*\returns 0 to indicate send is complete, 1 to indicate it is still sending
*/
static int QTouchSendCommand( QTouchDevice_t *cmd )
{
	if( cmd->index < cmd->count )
	{
		if( QTouchSend(cmd->list[cmd->index],&cmd->reply[cmd->index]) != 0 )
			cmd->index++;
		else
			cmd->notReadyCount++;
		return( 1 );
	}
	else
	{
		cmd->index = 0;
		return( 0 );
	}
}

/**
* QTouchCheckCRC routine checks the CRC for the message from the device.
*\param cmd pointer to QTouch key & driver state machine data structure
*\return 0 if valid CRC, 1 if invalid CRC.
*/
static int QTouchCheckCRC( QTouchDevice_t *cmd )
{
	int result = 0;
	/* We have finished transmission now work out the results
	 * currently cmd->reply[0] contains scrap byte, this should be 0x55
	 * we over write this byte with our command so that we have a convenient
	 * buffer for calculating the crc
	 */
	cmd->reply[0] = cmd->list[0];
	if( (result = QTouchCRC( cmd->reply, cmd->count )) == 0 )
	{
#ifdef QT_DEBUG
		printk("0x%02X\tCRC OK\n",cmd->reply[0] );
#endif
	}
	else
	{
#ifdef QT_DEBUG
		printk("0x%02X\tCRC  0x%02X 0x%02X count=%d\n",cmd->reply[0], cmd->reply[1], cmd->reply[2], cmd->count );
#endif
	}
	return( result );
}

/**
* QTouchGettingStatus routine sends a 0xC2 (device status) to
* device. The Device responds with the general device status.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return 0
*/
static int QTouchGettingStatus( QTouchDevice_t *cmd )
{
	u8 status;
	int i;
	static const char *errorTable[] = {
		"Cal error\n",
		"RAM CRC error\n",
		"EEPROM CRC error\n",
		"Sync error\n",
		"Extreme signal error\n",
		"Reset occured\n",
		"EEPROM error\n",
		"key(s) detecting\n",
	};

	if( QTouchSendCommand( cmd ) != 0 )
	{
		/* still sending */
		return( 0 );
	}

	/* command sent so now check CRC */
	if( QTouchCheckCRC( cmd ) != 0 )
	{
		/* CRC bad so now go round again */
		return( 0 );
	}

	/* Ok CRC is good so lets look at status */

	status = cmd->reply[1];
#ifdef QT_DEBUG
	printk("0xC2\tDevice Status 0x%02X\n",status);
#endif
	for(i=0; i<8 ;i++ )
	{
		if( (status & (1<<i)) != 0 )
		{
			printk("\t%s",errorTable[i]);
		}
	}
	qTouchDoneEvent(cmd);
	return( 0 );
}

/**
* QTouchStartStatusCmd routine creates a 0xC2 (Device Status) command.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchStartStatusCmd( QTouchDevice_t *cmd )
{
	static const u8 statusSeq[] = { 0xC2, 0, 0 };

	cmd->list = statusSeq;
	cmd->index = 0;
	cmd->count = sizeof(statusSeq);

	return( 0 );
}
/**
* QTouchGettingLastCmd routine processes the last command sent to the device.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchGettingLastCmd( QTouchDevice_t *cmd )
{
	u8 status;

	if( QTouchSendCommand( cmd ) != 0 )
	{
		/* still sending */
		return( 0 );
	}

	/* command sent so find what last went */

	status = ~cmd->reply[1];
#ifdef QT_DEBUG
	printk("0xC7\tLast Command was 0x%02X\n",status);
#endif
	qTouchDoneEvent(cmd);
	return( 0 );
}
/**
* QTouchStartRetLastCmd routine creates a 0xC7 (Return Last Command) command
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchStartRetLastCmd( QTouchDevice_t *cmd )
{
	static const u8 retLastSeq[] = { 0xC7, 0 };

	cmd->list = retLastSeq;
	cmd->index = 0;
	cmd->count = sizeof(retLastSeq);
	return( 0 );
}

/**
* QTouchPrepareSetup routine checks the status of the QTouch device
* and only enters the QTouchStartDownloadCmd state if the device is idle.
*\param cmd Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchPrepareSetup( QTouchDevice_t *cmd )
{
	u8 status;

	if( QTouchSendCommand( cmd ) != 0 )
	{
		/* still sending */
		return( 0 );
	}

	/* command sent so find what last went */

	status = cmd->reply[2];

	if( status == 0x53 )
	{
		/* ok device ready for download */
		printk("0x01\tSetup reply=0x%02X Ready for download\n",status);
		qTouchDoneEvent(cmd);
	}
	else
	{
		printk("0x01\tSetup reply=0x%02X Failed\n",status);
		qTouchErrorEvent(cmd);
	}

	return( 0 );
}

/**
* QTouchStartSetupCmd routine Creates the Enter Setups Load Mode command
* to load the setup block into the device.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchStartSetupCmd( QTouchDevice_t *cmd )
{
	static const u8 setupSeq[] = { 0x01,0x01, 0 };

	cmd->list = setupSeq;
	cmd->index = 0;
	cmd->count = sizeof(setupSeq);

	return( 0 );
}

/**
* QTouchDownloading routine sends the SetupBlock (36 bytes) message to
* the device and then reads the device response. If the response is
* 0xFE i.e. success then the next state is set to QTouchStartCalibrationCmd
* otherwise the next state is set to QTouchStartRetLastCmd to re-send
* the SetupBlock.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchDownloading( QTouchDevice_t *cmd )
{
	u8 status;
	int i;

	if( QTouchSendCommand( cmd ) != 0 )
	{
		i = cmd->index -1;
		if( cmd->reply[i] == 0xF1 )
			printk("DL sent %x reply = %x\n",cmd->list[i], cmd->reply[i]);

		/* still sending */
		return( 0 );
	}

	/* command sent so find what last went */

	status = cmd->reply[36];

	if( status == 0xFE )
	{
		/* ok device ready for download */
		printk("\tDownload Status 0x%02X Successful\n", status);
		qTouchDoneEvent(cmd);
	}
	else if( status == 0xF2 )
	{
		printk("\tDownload Status 0x%02X EEPROM write failed\n",status);
		qTouchErrorEvent(cmd);
	}
	else if( status == 0xF0 )
	{
		printk("\tDownload Status 0x%02X Failed CRC error\n",status);
		qTouchErrorEvent(cmd);
	}
	else
	{
		printk("\tDownload Status 0x%02X Failed\n",status);
		qTouchErrorEvent(cmd);
	}

	return( 0 );
}
/**
* QTouchStartDownloadCmd routine adds CRC check for the SetupBlock (byte 35) message.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\returns (0)
*/
static int QTouchStartDownloadCmd( QTouchDevice_t *cmd )
{
	static u8 setupList[37];
#define BURST_SPACING(x) ((x)<<2)

/*	memcpy (setupList,QTouch_default_SetupBlock, 35 );*/

	memset( &setupList[0],0x93,6);
	memset( &setupList[6],0 ,4);
/*	memset( &setupList[0],0x43,10); */
	memset( &setupList[10],0x75,10);
	memset( &setupList[20],0x01,10);
	setupList[30] = 0x51; /* Pos Thresh */
	setupList[31] = 0x57; /* Pos drift comp + */
	setupList[32] = 0x30; /* FDIL NTM SE DI + */
	setupList[33] = 0x03; /* lower burst limit */
	setupList[34] = 0x61 | BURST_SPACING(5); /* control bits burst spacing 5ms KxxB BBUU*/
/*	setupList[34] = 0x61;  control bits burst spacing 2ms KxxB BBUU*/
/*	setupList[34] = 0x69;  control bits burst spacing 3ms KxxB BBUU*/
/*	setupList[34] = 0x6D;  control bits burst spacing 3.5ms KxxB BBUU*/
/*	setupList[34] = 0x71;  control bits burst spacing 4ms KxxB BBUU*/
/*	setupList[34] = 0x79;  control bits burst spacing 6ms KxxB BBUU*/
/*	setupList[34] = 0x7D;  control bits burst spacing 7ms KxxB BBUU*/
	setupList[35] = QTouchCRC( setupList, 35 );

/**
 *  The NULL command will read the response from the device
 *  to the SetupBlock message.
 */
	setupList[36] = 0x00;	/* NULL command */

	cmd->list = setupList;
	cmd->index = 0;
	cmd->count = sizeof(setupList);

	return( 0 );
}

/**
* Upon receipt of a device ready (CRDY) interrupt from the QTouch device,
* command 0xC9 (Quick Report First Key) is sent to device to get a key
* touch status.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\return (0)
*/
static int QTouchKeyAquisition( QTouchDevice_t *cmd )
{
	u8 key;
	u32 keymask;

	if( QTouchSendCommand( cmd ) != 0 )
	{
		/* still sending */
		return( 0 );
	}

	/* command sent so find what last went */

	key = cmd->reply[1];

	if( key == 0x55 )
	{
		/* he didn't understand us so we try again next time */
		printk("QTouch Cmd Error %x\n",key);
		cmd->notUnderStoodCount++;
	}
	else if( (key & 0x80) != 0  && QTOUCH_ACTON_ERRORS )
	{
		/* we have an error */
		printk("Key Error bit set %x\n",key);
//		printk("Key Error bit set!!!! Key Error bit set!!!! Key Error bit set!!!!\n");
//		printk("Key Error bit set!!!! Key Error bit set!!!! Key Error bit set!!!!\n");
		cmd->IsrIgnorecount = 0;
		cmd->KeyErrorCount++;
		qTouchDoneEvent(cmd);
	}
	else if( (key & 0x20) != 0 )
	{
		/* we have an error */
#ifdef QT_DEBUG
		printk("Multiple keys Event ignored %x\n",key);
#endif
		cmd->MulipleKeyCount++;
	}
	else if( (key & 0x10) != 0 )
	{
		/* we have a key press */
		key &= 0xf;
		keymask = (1 << key );
		if( (cmd->keystate & keymask ) == 0 )
		{
			/* its a new one */
			if( cmd->keystate != 0 )
			{
				int i;
				for( i=0; i < KERNEL_MODE_KEY_RANGE; i++ )
				{
					if( ((1 << i) & cmd->keystate ) != 0 )
					{
#ifdef QT_DEBUG
						printk("Implied Key Release %d\n", i);
#endif
						/* set top bit to indicate release */
						QTouchEventNotify( KERNEL_MODE_KEY + (i | 0x80), KEY_RELEASED );
					}
				}
				cmd->keystate = 0;
			}
#ifdef QT_DEBUG
			printk("Key Press %d code=%02x\n",key,cmd->reply[1]);
#endif
			cmd->keystate |= keymask;
			cmd->TotalKeyCount++;
			cmd->KeyCounts[key]++;
			QTouchEventNotify( KERNEL_MODE_KEY+key, KEY_PRESSED );
		}
	}
	else
	{
		/* no keys pressed now */
		if( cmd->keystate != 0 )
		{
			int i;
			for( i=0; i < KERNEL_MODE_KEY_RANGE; i++ )
			{
				if( ((1 << i) & cmd->keystate ) != 0 )
				{
#ifdef QT_DEBUG
					printk("Key Release %d\n", i);
#endif
					/* set top bit to indicate release */
					key = i | 0x80;
//					writeToKeyDataQueue( cmd, key );
					QTouchEventNotify( KERNEL_MODE_KEY+i, KEY_RELEASED );
				}
			}
			cmd->keystate = 0;
		}
	}

	cmd->IsrIgnorecount = 9;
	return( 0 );
}

/**
* QTouchStartKeyAquisitionCmd Creates a Quick Report Key command
* to get the data from the key that has been touched.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\returns (0)
*/
static int QTouchStartKeyAquisitionCmd( QTouchDevice_t *cmd )
{
	static const u8 keySeq[] = { 0xC9,0 };

	cmd->list = keySeq;
	cmd->index = 0;
	cmd->count = sizeof(keySeq);

	return( 0 );
}

/**
* QTouchCalibrating routine is run after a setups load command (0x1)
* to get the device to calibrate and run as a sensor.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\returns (0)
*/
static int QTouchCalibrating( QTouchDevice_t *cmd )
{
	u8 status;

	if( QTouchSendCommand( cmd ) != 0 )
	{
		/* still sending */
		return( 0 );
	}

	/* command sent so find what last went */

	status = cmd->reply[2];

	if( status == 0xFC )
	{
		/* ok device ready for download */
		printk("0x03\tCalibration reply=0x%02X Successful\n", status);
		qTouchDoneEvent(cmd);
	}
	else
	{
		printk("0x03\tCalibration reply=0x%02X Failed\n",status);
	}

	return( 0 );
}

/**
* QTouchStartCalibrationCmd routine creates a calibration.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\returns (0)
*/

static int QTouchStartCalibrationCmd( QTouchDevice_t *cmd )
{
	static const u8 calSeq[] = { 0x03,0x03, 0 };

	cmd->list = calSeq;
	cmd->index = 0;
	cmd->count = sizeof(calSeq);

	return( 0 );
}

/**
* QTouchFinish routine will terminate QTouch_task.
*\param cmd pointer to Qtouch key & driver state machine data structure
*\returns (0)
*/
static int QTouchFinish( QTouchDevice_t *cmd )
{
	printk("QTouchFinish\n");

	return( 0 );
}





/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct proc_dir_entry *QTouch_proc;
/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/QTouch'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/
static int QTouch_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len;
	int i;
    len = sprintf(buf, "QTouch version 2.0 Stats\n");
	len += sprintf(buf+len,"KeyState               = %08x\n",QTouch_Device.keystate );
	len += sprintf(buf+len,"DevState               = %d\n",QTouch_Device.deviceState );
	len += sprintf(buf+len,"Error Action Count     = %d\n",QTouch_Device.errorActionCount );
	len += sprintf(buf+len,"Not Understood Count   = %d\n",QTouch_Device.notUnderStoodCount );
	len += sprintf(buf+len,"Not Ready Count        = %d\n",QTouch_Device.notReadyCount );
	len += sprintf(buf+len,"Total kernel Job Count = %d\n",QTouch_Device.TotalIsrCount );
	len += sprintf(buf+len,"Late kernel Job Count  = %d\n",QTouch_Device.LateIsrCount );
	len += sprintf(buf+len,"Key Error Count        = %d\n",QTouch_Device.KeyErrorCount );
	len += sprintf(buf+len,"Multiple Key Count     = %d\n",QTouch_Device.MulipleKeyCount );
	len += sprintf(buf+len,"Total Key Count        = %d\n",QTouch_Device.TotalKeyCount );
	
	for( i=0; i < 10; i++ )
	{
		len += sprintf(buf+len,"Key %d Count            = %d\n", i, QTouch_Device.KeyCounts[i] );
	}
	return len;
}
/**
* Function to handle a proc entry write not often used
*/
static int QTouch_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init Qtouch_init(void)
{
	keyRegInterface_t registration;

	/* create PROC entry */
	if( (QTouch_proc = create_proc_entry( "driver/QTouch", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/QTouch!\n");
		return( -ENOENT );
	}
	QTouch_proc->read_proc  = QTouch_read_procmem;
	QTouch_proc->write_proc = QTouch_write_procmem;

	printk(KERN_NOTICE "/proc/driver/QTouch created!\n");

	spin_lock_init(&QTouch_Device.lock);



	QTouch_Device.IsrIgnorecount = 0;
	qtDevPtr = &QTouch_Device;



	QTouch_task.routine = QTouchKernelJob; /*scheduled function*/
	QTouch_task.data = &QTouch_Device; /*argument to function*/

	if( QTouch_isr_init(&QTouch_Device) != 0 )
		printk(KERN_INFO "ERROR: QTouch_isr_init failed\n");

	QTouch_callback = 0;
	registration.keyInputEvent = 0;
	registration.name = "QTouch";
	if( keyInputRegister( &registration ) == 0 )
	{
		QTouch_callback = registration.keyInputEvent;
	}

	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/
static void __exit Qtouch_exit(void)
{
	keyRegInterface_t registration;

	registration.name = "QTouch";
	if( keyInputUnRegister( &registration ) == 0 )
	{
	}
	
	QTouch_callback = 0;

	disable_fpga_irq( FPGA_MM_KEY_DRDY_INT );
	disable_fpga_irq( FPGA_MM_KEY_INT );
	
	free_fpga_irq( FPGA_MM_KEY_DRDY_INT );
	free_fpga_irq( FPGA_MM_KEY_INT );

	remove_proc_entry("driver/QTouch", 0);

	printk(KERN_NOTICE "QTouch module exit\n");
}


#ifdef MODULE
module_init(Qtouch_init);
module_exit(Qtouch_exit);
#endif
/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
