/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* The touch wheel driver interfaces to the QT5C03(QWheel Touch Slider) device
* using the SPI hardware of the INCA-IP2. This device reports a 7 bit number
* that represents the absolute position on the wheel that is touched. This
* into acceleration data by reporting the change in absolute position between
* one absolute position is  converted update and the next. Updated position
* data is reported every 40ms.
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>
#include "fpga.h"
#include "Quantum_drivers.h"
#include <asm/incaip2/tsf.h>
#include "keyinput.h"

//#define QW_DEBUG_CMD 1


/**
* Time out for poll timer is 40mS
*/
#define TIMER_INTERVAL_IN_MILLISECS 40


/** Major ID of QWheel device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID QWheel (default: 0=dynamic)");


/**
* Length of command queue buffer
*/

#define CMD_Q_LEN	(1<<4)

/**
* Mask for command queue buffer
*/

#define CMD_Q_MASK	((1<<4)-1)

#define WHEEL_EVENT_QUEUE_LEN		(1<<5)
#define WHEEL_EVENT_QUEUE_MASK	(WHEEL_EVENT_QUEUE_LEN-1)

#define PTR_INCR(x) x = ((x == WHEEL_EVENT_QUEUE_MASK) ? 0 : x +1)
#define FIFO_FULL(rp,wp) ((((rp) + WHEEL_EVENT_QUEUE_LEN - (wp)) & WHEEL_EVENT_QUEUE_MASK) == 2)



typedef struct {
	u8 value;
}QWheel_event_t;

typedef struct {
	spinlock_t lock;
	int wqPtr;
	int rqPtr;
	QWheel_event_t WheelEventQueue[WHEEL_EVENT_QUEUE_LEN];
} QWheelInputDevice_t;


typedef struct {
	spinlock_t lock;
	int wqPtr;
	int rqPtr;
	unsigned char wQueue[CMD_Q_LEN];
	unsigned char mode;
	unsigned char upDateMode;
} QWheelConfigDevice_t;

typedef struct {
	QWheelConfigDevice_t configDev;
	QWheelInputDevice_t eventNotificationDev;
} QWheelDevice_t;

QWheelDevice_t QWheelDevice;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/

DECLARE_WAIT_QUEUE_HEAD(QWheelInput_wait_queue);


/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;



/**
* Prototype declartion of the file operations structure for QWheel device which 
* effectively aliases application functions such as read, write etc. to equivalent 
* kernel funtion
*/

ssize_t QWheel_read(struct file *, char *, size_t, loff_t *);
ssize_t QWheel_write(struct file *, const char *, size_t, loff_t *);
int QWheel_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int QWheel_open(struct inode *, struct file *);
int QWheel_close(struct inode *, struct file *);
unsigned int QWheel_poll(struct file *, struct poll_table_struct *);


int QWheelInputEvent( u8 value, unsigned long timestamp )
{
	QWheelInputDevice_t *dev = &QWheelDevice.eventNotificationDev;
	
//	if( dev->wqPtr != dev->rqPtr )
//	{
//		printk(KERN_NOTICE "QWheelInputEvent: backed up\n");
//	}

	if( !FIFO_FULL(dev->rqPtr,dev->wqPtr) )
	{
		dev->WheelEventQueue[dev->wqPtr].value = value;
		PTR_INCR(dev->wqPtr);
		wake_up_interruptible( &QWheelInput_wait_queue );
	}
	else
	{
		printk(KERN_NOTICE "QWheelInputEvent: FIFO full\n");
	}
	return( 0 );
}

int QWheelInputRegister( QWheelRegInterface_t *wheelThing )
{
	wheelThing->QWheelInputEvent = QWheelInputEvent;
	if( wheelThing->name != 0 )
		printk(KERN_NOTICE "QWheelInputRegister: %s\n",wheelThing->name);
	return( 0 );
}

int QWheelInputUnRegister( QWheelRegInterface_t *wheelThing )
{
	if( wheelThing->name != 0 )
		printk(KERN_NOTICE "QWheelInputUnRegister: %s\n",wheelThing->name);
	return( 0 );
}


int QWheelReadEvent( QWheelInputDevice_t *dev, QWheel_event_t *event )
{
//	printk(KERN_NOTICE "QWheelReadEvent: wr=%d rd=%d\n",dev->wqPtr,dev->rqPtr);
	if( dev->wqPtr != dev->rqPtr )
	{
		event->value = dev->WheelEventQueue[dev->rqPtr].value;
		PTR_INCR(dev->rqPtr);
		return( sizeof( QWheel_event_t ) );
	}
	return( 0 );
}

void initQWheelQueue( QWheelInputDevice_t *dev )
{
	dev->wqPtr = 0;
	dev->rqPtr = 0;
	spin_lock_init( &dev->lock);
}

void initQWheelConfig( QWheelConfigDevice_t *dev )
{
	dev->wqPtr = 0;
	dev->rqPtr = 0;
	dev->mode = 0xF5;
	dev->upDateMode = 0;
	
	spin_lock_init( &dev->lock);
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations QWheel_fops = {
	read:		QWheel_read,	/*read */
	write:		QWheel_write, /* write */
	ioctl:		QWheel_ioctl, /* ioctl */
	poll:		QWheel_poll,  /* poll */
	open:		QWheel_open,  /* open */
	release:	QWheel_close, /* release */
};

/**
*Utility function to convert recieved QWheel data to hex for display purposes
*\param text pointer to text string character buffer
*\param buf pointer to input data buffer
*\param size size of input data buffer
*/
static void QWheelHexOutput(const char *text, const char *buf, size_t size)
{
	char khexbuff[61];
	static const char hexChar[] = "0123456789ABCDEF";
	int i,j;

	if( size > 20 )
		size = 20;

	for( i=0,j=0;i<size;i++)
	{
		khexbuff[j++] = hexChar[((buf[i]>>4)& 0xf)];
		khexbuff[j++] = hexChar[(buf[i]& 0xf)];
		khexbuff[j++] = ' ';
	}
	khexbuff[j]=0;
#ifdef QW_DEBUG_CMD
	printk("QWheel %s: %s\n",text,khexbuff);
#endif
}


/**
* Function called by QWheel_read to return acknowlege sequence for commands
* on to command queue to userspace.
*\param tl pointer to kernel device timer structure
*\param buf pointer to userspace buffer
*\param size size of userspace buffer
*\returns The length of the message read
*/

static ssize_t QWheelCommandQRead( QWheelConfigDevice_t* tl, char * buf, size_t size )
{
	char ackBuff[CMD_Q_LEN];
	ssize_t ret;
	unsigned long flags;


	spin_lock_irqsave(&tl->lock,flags);

	for( ret = 0 ; (ret < size) && (tl->rqPtr != tl->wqPtr); ret++ )
	{
		ackBuff[ret] = tl->wQueue[tl->rqPtr];
		tl->rqPtr = (tl->rqPtr + 1) & CMD_Q_MASK;
	}

	spin_unlock_irqrestore(&tl->lock,flags);

	if( ret )
	{
#ifdef QW_DEBUG
		printk("QWheelCommandQRead r=%d,w=%d,s=%d\n", tl->rqPtr, tl->wqPtr, ret);
#endif
		QWheelHexOutput("read",ackBuff,ret);
		if (copy_to_user(buf, ackBuff, ret))
		{
			printk(KERN_ERR "tl_read: copy_to_user failed\n");
			ret = -ERESTARTSYS;
		}
	}
	return ret;
}

/**
 * Aliased read function called from user space
 * Reads availale command acknowledges else during normal operation
 * it gets polled data off the incoming data queue buffer manipulated it and returns the result to usre space.
 *\param *file pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns either number of data items transferred or error state
 */

ssize_t QWheel_read(struct file *file, char *buf, size_t size, loff_t * off)
{
	QWheelDevice_t* dev = (QWheelDevice_t*)file->private_data;
	QWheelConfigDevice_t* configDevPtr = &dev->configDev;
	QWheelInputDevice_t* eventNotificationDevPtr = &dev->eventNotificationDev;


	unsigned long flags;
	ssize_t ret;
	mPacket_t	pkt;
	char zMove;
	QWheel_event_t event;

	ret = QWheelCommandQRead( configDevPtr, buf, size );

	if( ret == 0 )
	{
		spin_lock_irqsave(&configDevPtr->lock,flags);
		if( configDevPtr->upDateMode != 0 )
		{
			configDevPtr->mode = configDevPtr->upDateMode;
			configDevPtr->upDateMode = 0;
		}
		spin_unlock_irqrestore(&configDevPtr->lock,flags);
	}
	else
	{
		return ret;
	}

	while((eventNotificationDevPtr->rqPtr) == (eventNotificationDevPtr->wqPtr))
	{/* nothing to read */

		if (file->f_flags & O_NONBLOCK)
		{
			return -EAGAIN;
		}

		if(wait_event_interruptible(QWheelInput_wait_queue, (eventNotificationDevPtr->rqPtr) != (eventNotificationDevPtr->wqPtr)))
			return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
	}



	QWheelReadEvent( eventNotificationDevPtr, &event );

	zMove = event.value;

//	if (zMove < -2)
//		zMove = -2;
//	else if (zMove > 2)
//		zMove = 2;

//	if( zMove < -1 || zMove > 1 )
//		printk(KERN_ERR "QWheel: zMove %d\n", zMove );

	pkt.btnStatus = MPKT_IS_ONE;
	pkt.xMove = 0;
	pkt.yMove = 0;
	pkt.zMove = -zMove;

//	if ( zMove < 0 )
//		pkt.btnStatus |= 0x20;

	ret = (size < ret) ? size : sizeof(mPacket_t);

	if(copy_to_user(buf, &pkt, ret))
	{
		printk(KERN_ERR "dev_read: copy_to_user failed\n");
		ret = -ERESTARTSYS;
	}

	return( ret );
}
/**
* Function called by QWheelReceiveCmd to put acknowlege sequence for commands
* on to command queue for returning to userspace.
*\param tl pointer to kernel device timer structure
*\param size number of acknowlegements to but place in queue
*\param buf to acknowlegementsequence buffer
*\returns Length of the acknowlegement message written to queue
*/

static ssize_t QWheelCmdAckWrite( QWheelConfigDevice_t* tl, size_t size, char * buf )
{
	ssize_t len;
	unsigned long flags;

	spin_lock_irqsave(&tl->lock,flags);

	/* work out how many fit in in Queue */
	len = CMD_Q_MASK - ((tl->wqPtr - tl->rqPtr) & CMD_Q_MASK);

	if( size > len )
	{
		/* won't fit */
		size = 0;
	}

	for( len = 0 ; (len < size) ; len++ )
	{
		tl->wQueue[tl->wqPtr] = buf[len];
		tl->wqPtr = (tl->wqPtr + 1) & CMD_Q_MASK;
	}
	spin_unlock_irqrestore(&tl->lock,flags);

#ifdef QW_DEBUG

	printk("QWheelCmdAckWrite r=%d,w=%d,s=%d\n", tl->rqPtr, tl->wqPtr, size);

	if( len == 0 )
		printk("QWheelCmdAckWrite Won't fit\n");
	else
		QWheelHexOutput("QWheelCmdAckWrite",buf,len);
#endif
	return len;
}

/**
* Function called by QWheel_write to determine correct acknowledge
* dummy mouse setup commands recieved from userspace
*\param command recieved command
*\param tl pointer to kernel device timer structure
*/

static ssize_t QWheelReceiveCmd( unsigned char command, QWheelConfigDevice_t* tl)
{
	unsigned long flags;
	unsigned char ack[4];
	unsigned char mode;
	ssize_t count;

	ack[0] = 0xFA; /* standard reply */
	count = 1;
	mode = 0;
#ifdef QW_DEBUG
	printk("QWheelReceiveCmd %02X\n", command);
#endif
	switch(command)
	{
		case 0xF2:	ack[1] = 0x03; count++; mode = 0xF4; break;
		case 0xF4:	mode = 0xF4;	break;
		case 0xF5:	mode = 0xF5;	break;
		case 0xFF:	ack[1] = 0xAA; ack[2] = 0x00; count += 2;	break;
	}
	if( QWheelCmdAckWrite( tl, count, ack ) != 0 )
	{
		spin_lock_irqsave(&tl->lock,flags);
		if( mode != 0 && mode != tl->mode )
		{
			tl->upDateMode = mode;
		}
		spin_unlock_irqrestore(&tl->lock,flags);
		return( 1 );
	}
	return( 0 );
}

/**
 * Aliased write function called from user space
 * Currently used to only setup QWheel as pseudo mouse driver
 *\param *fp pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns Number of data items sent or error state
 */

ssize_t QWheel_write(struct file *fp, const char *buf, size_t size, loff_t *off)
{
	char kbuff[20];
	int i;
	QWheelDevice_t* dev = (QWheelDevice_t*)fp->private_data;
	QWheelConfigDevice_t* configDevPtr = &dev->configDev;
	ssize_t ret;

	if (size > sizeof(kbuff))
	{
		printk(KERN_WARNING "QWheel_write: illegal size %d\n", size);
		return 0;
	}

	if (copy_from_user(kbuff, buf, size))
	{
		return( -EFAULT );
	}

	QWheelHexOutput("write",kbuff,size);

	ret = 0;
	for( i=0;i<size;i++)
	{
		ret += QWheelReceiveCmd(kbuff[i], configDevPtr);
	}
	return ret;
}

/**
 * Aliased ioctl function called from user space
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *filp pointer to a kernel file data structure
 *\param cmd ioctl command
 *\param arg associated command arguments
 * Currently not used
 *\returns 0
 */

int QWheel_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err;
	printk(KERN_ERR "QWheel_ioctl: cmd=%08x\n", cmd);
	err = 0;
	return err;
}

/**
 * Aliased poll function called from user space by poll or select
 * Currently used to wait on information being availble to be read from incoming data queue buffer.
 * Woken by the QWheelActOnPoll function.
 *\param *fp pointer to a kernel file data structure
 *\param *pts pointer to a kernel poll table data structure
 *\returns flags to indicate valid data available
 */

unsigned int QWheel_poll(struct file *fp, struct poll_table_struct *pts)
{
	unsigned int mask = 0;
	QWheelInputDevice_t* dev = &((QWheelDevice_t*)fp->private_data)->eventNotificationDev;

	/*
	 * The buffer is circular.
	 */

	poll_wait(fp, &QWheelInput_wait_queue, pts);
	if ((dev->rqPtr) != (dev->wqPtr)) mask |= POLLIN | POLLRDNORM;  /* readable */

	return mask;
}

/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */

int QWheel_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "QWheel_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &QWheelDevice;

	initQWheelConfig( &QWheelDevice.configDev );

	return 0;
}

/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int QWheel_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "QWheel_close:\n");
	err = 0;
	return err;
}





/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *QWheel_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/QWheel'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int QWheel_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "QWheel Device version 1.0\n";
    printk(KERN_INFO "/proc/drivers/QWheel %d\n",count);

    len = sizeof(info);

    if( count < sizeof(info))
    	len = count;
    memcpy(buf,info,len);

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int QWheel_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init Qwheel_init(void)
{
	int res = 0;

	/* create and register device */

	if( (input_devfs_handle = devfs_mk_dir(NULL, "input", NULL)) == 0 )
		return -EBUSY;

	devfs_register(input_devfs_handle, "mice",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&QWheel_fops,0);


	if (major == 0) major = res;

	SET_MODULE_OWNER(&QWheel_fops);

	printk(KERN_NOTICE "\nQWheel mouse device registered.\n");

	/* create PROC entry */
	if( (QWheel_proc = create_proc_entry( "driver/QWheel", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/qwheel!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}
	QWheel_proc->read_proc  = QWheel_read_procmem;
	QWheel_proc->write_proc = QWheel_write_procmem;

	printk(KERN_NOTICE "/proc/driver/QWheel created!\n");

	initQWheelQueue( &QWheelDevice.eventNotificationDev );
	initQWheelConfig( &QWheelDevice.configDev );

	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit Qwheel_exit(void)
{
	remove_proc_entry("driver/QWheel", 0);
	devfs_unregister(input_devfs_handle);
	printk(KERN_NOTICE "QWheel_exit\n");
}


#ifdef MODULE
module_init(Qwheel_init);
module_exit(Qwheel_exit);
EXPORT_SYMBOL(QWheelInputRegister);
EXPORT_SYMBOL(QWheelInputUnRegister);
#endif

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
