/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/** \file QWheelTestIfc.c
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include "keyinput.h"

#define IOCTL_MAGIC                  'k'
#define IOCTL_QWTEST_ON              _IO(IOCTL_MAGIC, 0)
#define IOCTL_QWTEST_OFF             _IO(IOCTL_MAGIC, 1)
#define IOCTL_QW_SENSITIVITY         _IO(IOCTL_MAGIC, 2)

/** Major ID of QWheel device (default: 0=dynamic)
*/
static int major = 0;	// Major ID of QWheelTestIfc device (default: 0=dynamic)

/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");

/**
* Module parameter description of variable major
*/
MODULE_PARM_DESC(major, "Major ID of QWheelTestIfc device (default: 0=dynamic)");

static int QwheelTest = 0;
MODULE_PARM(QwheelTest, "i");
MODULE_PARM_DESC(QwheelTest, "enable/disable wheel test event generation");
/**
* Length of command queue buffer
*/
#define QWHEEL_TEST_EVENT_QUEUE_LEN		(1<<5)
#define QWHEEL_TEST_EVENT_QUEUE_MASK	(QWHEEL_TEST_EVENT_QUEUE_LEN-1)

#define PTR_INCR(x) x = ((x == QWHEEL_TEST_EVENT_QUEUE_MASK) ? 0 : x +1)
#define FIFO_FULL(rp,wp) ((((rp) + QWHEEL_TEST_EVENT_QUEUE_LEN - (wp)) & QWHEEL_TEST_EVENT_QUEUE_MASK) == 2)

/**
* Structure for QWheelTestIfc event & read queue data
*/
typedef struct {
	u8 value;
}QWheel_event_t;

typedef struct {
	spinlock_t lock;
	int wqPtr;
	int rqPtr;
	QWheel_event_t WheelTestIfcEventQueue[QWHEEL_TEST_EVENT_QUEUE_LEN];
} QWheelTestIfcDevice_t;

/**
* An instance of the QWheelTestIfc io & driver state machine data structure
*/

static QWheelTestIfcDevice_t QWheelTestIfc_Device;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/
DECLARE_WAIT_QUEUE_HEAD(QWheelTestIfcInput_wait_queue);

static QWheel_callback_t QWheel_callback = 0;
static QWheelSensitivity_callback_t QWheelTestIfcSetSensitivity;

/**
* Device filesystem handle structure for the creation of device files
*/

static devfs_handle_t input_devfs_handle;

/**
* Prototype declartion of the file operations structure for QWheelTestIfc device which 
* effectively aliases application functions such as read, write etc. to equivalent 
* kernel funtion
*/
ssize_t QWheelTestIfc_read(struct file *, char *, size_t, loff_t *);
ssize_t QWheelTestIfc_write(struct file *, const char *, size_t, loff_t *);

int QWheelTestIfc_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int QWheelTestIfc_open(struct inode *, struct file *);
int QWheelTestIfc_close(struct inode *, struct file *);
unsigned int QWheelTestIfc_poll(struct file *, struct poll_table_struct *);

int QWheelTestIfcInputEvent( u8 value, unsigned long timestamp )
{
	QWheelTestIfcDevice_t *dev = &QWheelTestIfc_Device;
	
	if( QwheelTest == 0 )
		return( 0 );
		
	if( !FIFO_FULL(dev->rqPtr,dev->wqPtr) )
	{
		dev->WheelTestIfcEventQueue[dev->wqPtr].value = value;
		PTR_INCR(dev->wqPtr);
		wake_up_interruptible( &QWheelTestIfcInput_wait_queue );
	}
	else
	{
		printk(KERN_NOTICE "QWheelInputEvent: FIFO full\n");
	}
	return( 0 );
}

int QWheelTestIfcEvent( char wheelEvent )
{
	if( QWheel_callback != 0 )
	{
		return( QWheel_callback( wheelEvent, jiffies ) );
	}
	
	return( 0 );
}

int QWheelTestIfcRegister( QWheelTestIfRegInterface_t *wheelThing )
{
	wheelThing->QWheelInputEvent = QWheelTestIfcInputEvent;
	QWheelTestIfcSetSensitivity = wheelThing->QWheelSetSensitivity;
	if( wheelThing->name != 0 )
		printk("QWheelTestIfcRegister: %s\n",wheelThing->name);
	return( 0 );
}

int QWheelTestIfcUnRegister( QWheelTestIfRegInterface_t *wheelThing )
{
	if( wheelThing->name != 0 )
		printk(KERN_NOTICE ": %s\n",wheelThing->name);
	QWheelTestIfcSetSensitivity = 0;
	return( 0 );
}

int QWheelTestIfcReadEvent( QWheelTestIfcDevice_t *dev, QWheel_event_t *event )
{
	if( dev->wqPtr != dev->rqPtr )
	{
		event->value = dev->WheelTestIfcEventQueue[dev->rqPtr].value;
		PTR_INCR(dev->rqPtr);
		return( sizeof( QWheel_event_t ) );
	}
	return( 0 );
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/
struct file_operations QWheelTestIfc_fops = {
	write:		QWheelTestIfc_write,	/* write */
	read:		QWheelTestIfc_read,	/* read */
	ioctl:		QWheelTestIfc_ioctl, /* ioctl */
	poll:		QWheelTestIfc_poll,  /* poll */
	open:		QWheelTestIfc_open,  /* open */
	release:	QWheelTestIfc_close, /* release */
};

/**
* Aliased read function called from user space
* Waits for the device to put data in the wheel data queue
*\param *file pointer to a kernel file data structure
*\param *buf pointer to a user space data buffer
*\param size of the user space data buffer
*\param *off pointer to file position
*\returns either number of data items transferred or error state
*/
ssize_t QWheelTestIfc_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	QWheelTestIfcDevice_t* dev = (QWheelTestIfcDevice_t*)fp->private_data;
	ssize_t ret;
	QWheel_event_t event;
	
	if( size < sizeof(QWheel_event_t))
	{
		printk(KERN_ERR "QWheelTestIfc_read: bad size %d\n", size );
		return( 0 );
	}
		
	while((dev->rqPtr) == (dev->wqPtr))
	{/* nothing to read */

        if (fp->f_flags & O_NONBLOCK)
        {
            return -EAGAIN;
		}

        if(wait_event_interruptible(QWheelTestIfcInput_wait_queue, (dev->rqPtr) != (dev->wqPtr)))
        {
			printk(KERN_ERR "QWheelTestIfc_read: restart sys\n" );
            return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
        }
	}

	ret = 0;
	while( (size >= sizeof(QWheel_event_t)) && (QWheelTestIfcReadEvent( dev, &event ) != 0) )
	{
		if(copy_to_user(buf, &event, sizeof( QWheel_event_t) ))
		{
			printk(KERN_ERR "QWheelTestIfc_read: copy_to_user failed\n");
			return -ERESTARTSYS;
		}
		ret += sizeof( QWheel_event_t );
		buf += sizeof( QWheel_event_t );
		size -= sizeof( QWheel_event_t );
	}

	return( ret );
}


ssize_t QWheelTestIfc_write(struct file *fp, const char *buf, size_t len, loff_t * off)
{
	ssize_t	numbytes;
	char event;
//	QWheelTestIfcDevice_t* dev = (QWheelTestIfcDevice_t*)fp->private_data;

	if (len != 1)
	{
		printk("QWheelTestIfc_write: illegal size %d\n", len);
		return 0;
	}

	if (copy_from_user(&event, buf, len))
	{
		printk("QWheelTestIfc_write: failed copy_from_user");
		return( -EFAULT );
	}

	numbytes = 0;
	if( QWheelTestIfcEvent( event ) == 0 )
	{
		return( len );
	}

	return numbytes;
}

/**
 * Aliased ioctl function called from user space
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *filp pointer to a kernel file data structure
 *\param cmd ioctl command
 *\param arg associated command arguments
 * Currently not used
 *\returns 0
 */
int QWheelTestIfc_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int result = -ENOIOCTLCMD;
	switch( cmd )
	{
		case IOCTL_QWTEST_ON:
		{
			QwheelTest = 1;
			result = 0; 
//			printk(KERN_NOTICE "QWheelTestIfc_ioctl: cmd=%08x (IOCTL_QWTEST_ON)\n", cmd);
		}
		break;
		
		case IOCTL_QWTEST_OFF:
		{
			QwheelTest = 0;
			result = 0;
//			printk(KERN_NOTICE "QWheelTestIfc_ioctl: cmd=%08x (IOCTL_QWTEST_OFF)\n", cmd);
		}
		break;
		
		case IOCTL_QW_SENSITIVITY:
		{
			int sensitivity;
//			printk(KERN_NOTICE "QWheelTestIfc_ioctl: cmd=%08x (IOCTL_QW_SENSITIVITY)\n", cmd);

			if( (result = get_user(sensitivity,(int*)arg) ) != 0 )
				return( result );

			if( QWheelTestIfcSetSensitivity != 0 )
			{
				result = QWheelTestIfcSetSensitivity( sensitivity );
			}
		}
		break;
		
		default:
		{
			printk(KERN_ERR "QWheelTestIfc_ioctl: cmd=%08x\n", cmd);
		} 
		break;
	}
	return result;
}

/**
* Aliased poll function called from user space by poll or select
* Currently used to wait on information being availble to be read from incoming data queue buffer.
* Woken by the QWheelTestIfcDrdyEvent function.
*\param *fp pointer to a kernel file data structure
*\param *pts pointer to a kernel poll table data structure
*\returns flags to indicate valid data available
*/
unsigned int QWheelTestIfc_poll(struct file *fp, struct poll_table_struct *pts)
{
	unsigned int mask = 0;
	QWheelTestIfcDevice_t* dev = (QWheelTestIfcDevice_t*)fp->private_data;

	/*
	 * The buffer is circular.
	 */

	poll_wait(fp, &QWheelTestIfcInput_wait_queue, pts);
	if ((dev->rqPtr) != (dev->wqPtr)) mask |= POLLIN | POLLRDNORM;  /* readable */

    return mask;
}

/**
* Aliased open function called from user space by open
*\param *inode pointer to the inode structure for the device this contains such information such
*as the major and minor numbers.
*\param *fp pointer to a kernel file data structure
*\returns 0 (in turn application returns a file descriptor)
*/
int QWheelTestIfc_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "QWheelTestIfc_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &QWheelTestIfc_Device;

	return 0;
}
/**
* Aliased close function called from user space by close
*\param *inode pointer to the inode structure for the device this contains such information such
*as the major and minor numbers.
*\param *fp pointer to a kernel file data structure
*\returns 0
*/
int QWheelTestIfc_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "QWheelTestIfc_close:\n");
	err = 0;
	return err;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct proc_dir_entry *QWheelTestIfc_proc;
/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/QWheelTestIfc'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/
static int QWheelTestIfc_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "QWheelTestIfc Device version 1.0\n";
    printk(KERN_INFO "/proc/drivers/QWheelTestIfc %d\n",count);

    len = sizeof(info);

    if( count < sizeof(info))
    	len = count;
    memcpy(buf,info,len);

	return len;
}
/**
* Function to handle a proc entry write not often used
*/
static int QWheelTestIfc_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}


/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init QWheelTestIfc_init(void)
{
	int res = 0;

	QWheelRegInterface_t wheelRegistration;

	if( (input_devfs_handle = devfs_register(0, "input/QWheelTestIfc",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&QWheelTestIfc_fops,0)) == 0 )
		return -EBUSY;


	if (major == 0) major = res;

	SET_MODULE_OWNER(&QWheelTestIfc_fops);

	/* create PROC entry */
	if( (QWheelTestIfc_proc = create_proc_entry( "driver/QWheelTestIfc", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/QWheelTestIfc!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}
	QWheelTestIfc_proc->read_proc  = QWheelTestIfc_read_procmem;
	QWheelTestIfc_proc->write_proc = QWheelTestIfc_write_procmem;

	printk(KERN_NOTICE "/proc/driver/QWheelTestIfc created!\n");

	spin_lock_init(&QWheelTestIfc_Device.lock);

/**********************/
	wheelRegistration.QWheelInputEvent = 0;
	wheelRegistration.name = "QWheelTestIfc";

	if( QWheelInputRegister( &wheelRegistration ) == 0 )
	{
		QWheel_callback = wheelRegistration.QWheelInputEvent;
	}
/***************************/

	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/
static void __exit QWheelTestIfc_exit(void)
{
	QWheelRegInterface_t wheelRegistration;

	wheelRegistration.name = "QWheelTestIfc";
	if( QWheelInputUnRegister( &wheelRegistration ) == 0 )
	{
		printk(KERN_NOTICE "incaip2_wheelTestIfc_unregister_callback failed:\n");
	}
	
	QWheel_callback = 0;

	remove_proc_entry("driver/QWheelTestIfc", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_NOTICE "QWheelTestIfc module exit\n");
}


module_init(QWheelTestIfc_init);
module_exit(QWheelTestIfc_exit);
EXPORT_SYMBOL(QWheelTestIfcRegister);
EXPORT_SYMBOL(QWheelTestIfcUnRegister);

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
