/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/irq.h>
#include "fpga_defs.h"
#include "fpga.h"

#define FPGA_INT_NUM INT_NUM_IM5_IRL6
#define FPGA_INT_NR 6

typedef enum
{
	AUTO,
	MANUAL,
}eFpgaIntType;

typedef enum
{
	DISABLED,
	ENABLED,
}eFpgaIntStatus;

typedef struct {
		int vector;
		int unhandled_count;
		void * data;
		const char * text;
		fpgaIntHander handler;
		eFpgaIntStatus status;
		eFpgaIntType type;
} fpgaIrqDesc_t;

static fpgaIrqDesc_t irq_desc[FPGA_SPARE_15_INT+1];

static const char * fpgaInterruptText[] = {
	"Hook switch",
	"PS 2",
	"dhsg",
	"spare 3",
	"v24",
	"drdy",
	"mmi key drdy",
	"spare 7",
	"Lan",
	"spare 9",
	"display",
	"MMI key",
	"DM key",
	"midi",
	"usb",
	"spare 15"
};

int fpgaDefaultIntHander( eFpgaInterruptNum vector, void *data )
{
	*FPGA_INT_MASK |= (1<<vector);
	*FPGA_INT_EVENT = (1<<vector);
	if( (irq_desc[vector].unhandled_count & 0x3f) == 0 )
	{
		printk(KERN_NOTICE "oops fpgaDefaultIntHander vector %d %s\n",vector,irq_desc[vector].text);
		if( (*FPGA_INT_MASK & ~(1<<vector)) != 0 )
			printk(KERN_NOTICE "failed mask INT %04x bit %04x\n",*FPGA_INT_MASK, (1<<vector));
		if( (*FPGA_INT_EVENT & (1<<vector)) != 0 )
			printk(KERN_NOTICE "failed to clear event %04x bit %04x\n",*FPGA_INT_EVENT, (1<<vector));
		printk(KERN_NOTICE "EBU CS3 Control reg = %08x\n",*INCA_IP2_EBU_CON_3);
	}
	
	irq_desc[vector].unhandled_count++;	
	return( 0 );
}

void initFpgaIntDescriptors( void )
{
int i;
	for(i=0;i<16;i++)
	{
		irq_desc[i].vector = i;
		irq_desc[i].unhandled_count = 0;
		irq_desc[i].status = DISABLED;
		irq_desc[i].type = MANUAL;
		irq_desc[i].data = 0;
		irq_desc[i].handler = fpgaDefaultIntHander;
		irq_desc[i].text = fpgaInterruptText[i];
	}
	irq_desc[FPGA_HOOKSWITCH_INT].type = AUTO;
	irq_desc[FPGA_DRDY_INT].type = AUTO;
	irq_desc[FPGA_MM_KEY_DRDY_INT].type = AUTO;
	irq_desc[FPGA_MM_KEY_INT].type = AUTO;
	irq_desc[FPGA_DM_KEY_INT].type = AUTO;

}

int request_fpga_irq( eFpgaInterruptNum number, fpgaIntHander hander, void *data )
{
	int retVal = 0;
	unsigned long flags;
	spinlock_t lock;

	spin_lock_init(&lock);
	spin_lock_irqsave(lock,flags);


	if( irq_desc[number].handler == fpgaDefaultIntHander )
	{
		irq_desc[number].handler = hander;
		irq_desc[number].data = data;
		irq_desc[number].status = DISABLED;
		retVal = 1;
	}

	spin_unlock_irqrestore(lock,flags);

	return( retVal );
}

int free_fpga_irq( eFpgaInterruptNum number )
{
	int retVal = 0;
	unsigned long flags;
	spinlock_t lock;

	spin_lock_init(&lock);
	spin_lock_irqsave(lock,flags);

	irq_desc[number].status = DISABLED;
	irq_desc[number].handler = fpgaDefaultIntHander;
	irq_desc[number].data = 0;

	spin_unlock_irqrestore(lock,flags);

	return( retVal );
}

int enable_fpga_irq( eFpgaInterruptNum vector )
{
	int retVal = 0;
	unsigned long flags;
	spinlock_t lock;

	spin_lock_init(&lock);
	spin_lock_irqsave(lock,flags);

	
	irq_desc[vector].status = ENABLED;
	*FPGA_INT_MASK &= ~(1<<vector);

	spin_unlock_irqrestore(lock,flags);
	
	return( retVal );
}

int disable_fpga_irq( eFpgaInterruptNum vector )
{
	int retVal = 0;
	unsigned long flags;
	spinlock_t lock;

	spin_lock_init(&lock);
	spin_lock_irqsave(lock,flags);



	*FPGA_INT_MASK |= (1<<vector);
	irq_desc[vector].status = DISABLED;

	spin_unlock_irqrestore(lock,flags);
	
	return( retVal );
}


void fpga_show_interrupt_pending(void)
{
	u16 int_pend_reg = *FPGA_INT_PEN;
	u16 int_event_reg = *FPGA_INT_EVENT;
	
	if( int_pend_reg != 0 )
	{
		printk(KERN_NOTICE "FPGA_INT_PEN %04x\n",int_pend_reg);
	}
	if( int_event_reg != 0 )
	{
		printk(KERN_NOTICE "FPGA_INT_EVENT %04x\n",int_event_reg);
	}
}

u8 fpga_reg_read( int reg )
{
	u8 value = *((volatile u8*)((SIEMENS_FPGA) + (reg)));
	return( value );
}

int fpga_reg_write( int reg, u8 value )
{
	*((volatile u8*)((SIEMENS_FPGA) + (reg))) = value;
	return( 0 );
}

void fpgaIsrHandle( void )
{
	int vector;
	u16 unHandledEvents = *FPGA_INT_EVENT;
	u16 handledEvents = 0;
	
	while( unHandledEvents )
	{
		for(vector=0;vector<16;vector++)
		{
			if( (unHandledEvents & (1 << vector )))
			{
				if( irq_desc[vector].type == AUTO )
				{
					*FPGA_INT_EVENT = (1<<vector);
				}
				(irq_desc[vector].handler)(vector,irq_desc[vector].data);
			}
		}
		handledEvents |= unHandledEvents;
#if 0
		if( (*FPGA_INT_EVENT & ~handledEvents) != 0 )
		{
			printk(KERN_NOTICE "fpgaIsrHandle unHandledEvents %04x handledEvents  %04x EV %04x\n",unHandledEvents,handledEvents,*FPGA_INT_EVENT);
			unHandledEvents = 0;
		}
		else
			unHandledEvents = *FPGA_INT_EVENT & ~handledEvents;
#endif
		/* ack all the events we have cleared so far */
		*INCA_IP2_ICU_IM5_ISR = INCA_IP2_ICU_IM5_IR(FPGA_INT_NR);
		unHandledEvents = *FPGA_INT_EVENT & ~handledEvents;
	}
}

void fpga_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	
	fpgaIsrHandle();
//	*FPGA_INT_EVENT = 0xffff; /* clear all events */
//	ack_incaip2_irq(FPGA_INT_NUM);
}

int fpga_isr_init(void)
{
	int err;
	unsigned long flags;
	spinlock_t lock;
	u16 fpgaEventReg;

	printk(KERN_NOTICE "\nTouch Key Int Polarity %04x\n",*FPGA_INT_POL );		
	*FPGA_INT_POL = 0;

	spin_lock_init(&lock);

	spin_lock_irqsave(lock,flags);
	/* disable the interrupt */

	initFpgaIntDescriptors();
	/* disable EXIN0 interrupt */
	incaip2_disable_external_int(0);


	*FPGA_INT_EVENT = 0xffff; /* clear all events */
	fpgaEventReg = *FPGA_INT_EVENT;


	spin_unlock_irqrestore(lock,flags);

	if( (err = request_irq(FPGA_INT_NUM, fpga_isr, SA_INTERRUPT,"Fpga", NULL)) == 0 )
	{
		spin_lock_irqsave(lock,flags);

		/* EXIN0 for both edges */
		incaip2_enable_external_int(0,6);
		*FPGA_INT_MASK = 0xffff; /* 0x1840 Enable all events except DM FPK touch*/

		spin_unlock_irqrestore(lock,flags);
//		printk(KERN_NOTICE "\nFPGA ISR Init OK\n");
	}
	else
	{
		printk(KERN_NOTICE "\nFPGA ISR Init Failed\n");
	}
//	printk(KERN_NOTICE "FPGA Event Register %04x\n",*FPGA_INT_EVENT);
//	printk(KERN_NOTICE "FPGA Mask  Register %04x\n",*FPGA_INT_MASK);
//	printk(KERN_NOTICE "FPGA Pend  Register %04x\n",*FPGA_INT_PEN);
	return( err );
}

int fpga_isr_release(void)
{
	int err = 0;


	*FPGA_INT_EVENT = 0xffff; /* clear all events */
	*FPGA_INT_MASK = 0xffff; /* clear all events */

	/* disable EXIN0 interrupt */
	incaip2_disable_external_int(0);

	free_irq(FPGA_INT_NUM, 0);
	printk(KERN_NOTICE "\nFPGA ISR Release  Done\n");

	printk(KERN_NOTICE "FPGA Event Register %04x\n",*FPGA_INT_EVENT);
	printk(KERN_NOTICE "FPGA Mask  Register %04x\n",*FPGA_INT_MASK);
	printk(KERN_NOTICE "FPGA Pend  Register %04x\n",*FPGA_INT_PEN);
	return( err );
}

int spi_ss_set( spi_ss_t device )
{
	static int good;
	static int bad;
	int access = 0;
	u8 reg;
	u8 newreg;
	
	if( device > SS_MM_WHEEL )
	{
		printk(KERN_NOTICE "spi_ss_set: bad device %d\n",device);
		return( access );
	}


	reg = *FPGA_CS_REG;

	if( reg & 0x7 )
	{
		printk(KERN_NOTICE "spi_ss_set: request for %d bus in use by %d \n",device, reg & 0x7);
		return( access );
	}

	switch( device )
	{
		case SS_DM_KEY: 	access = (reg & DM_KEY_RDY_BIT); break;
		case SS_MM_KEY:		access = (reg & MM_KEY_RDY_BIT); break;
		case SS_MM_SLIDER:	access = (reg & MM_SLIDER_RDY_BIT ); break;
		case SS_MM_WHEEL:	access = (reg & MM_WHEEL_RDY_BIT ); break;
		default:			access = 1; break;
	}
	
	if( access != 0 )
	{
		*FPGA_CS_REG = device;
		newreg = *FPGA_CS_REG; /* make sure write has been propaged */
		if( (newreg & 0x8) != 0 )
		{
			printk("spi_ss_set:modified no drdy %02x\n", newreg );
		}
		if( (reg & 0xf0) != (newreg & 0xf0 ) )
		{
//			printk("spi_ss_set:modifyed rdy bits %02x %02x\n",reg, newreg );
		}
		good++;
	}
	else
	{		
		bad++;
//		fpga_show_interrupt_pending();
//		printk(KERN_NOTICE "spi_ss_set bad access %d %02X g=%d b=%d\n",device,reg,good,bad);
	}
	return( access );
}

int spi_ss_clear( spi_ss_t device )
{
	int spiTransferDone;
	u8 reg = *FPGA_CS_REG;
/*			
	switch( device )
	{
		case SS_DM_KEY: 	spiTransferDone = (reg & DM_KEY_RDY_BIT); break;
		case SS_MM_KEY:		spiTransferDone = (reg & MM_KEY_RDY_BIT); break;
		case SS_MM_SLIDER:	spiTransferDone = (reg & MM_SLIDER_RDY_BIT ); break;
		case SS_MM_WHEEL:	spiTransferDone = (reg & MM_WHEEL_RDY_BIT ); break;
		default:			spiTransferDone = 0; break;
	}
	
	if( spiTransferDone != 0 ){
		printk(KERN_NOTICE "spi_ss_clear %d SPI Transfer not Finished %02X\n",device,reg);
	}
*/		

	if( (reg & 0x7) != device )
	{
		printk("spi_ss_clear:wrong %02x dev=%02x\n", reg, device );
	}
	*FPGA_CS_REG = 0;
	reg = *FPGA_CS_REG; /* make sure write has been propaged */
	if( ((reg & 0x8) == 0) && (SS_DM_LED != device) && (SS_MM_LED != device))
	{
		printk("spi_ss_clear:modified no drdy %02x dev=%02x\n", reg, device );
	}
	return( 0 );
}

int spi_ss_sync( spi_ss_t device )
{
	switch( device )
	{
		case SS_DM_KEY:
		case SS_MM_KEY:
		case SS_MM_SLIDER:
		case SS_MM_WHEEL:
		{
			*FPGA_CS_REG = 0;
			udelay(15);
			*FPGA_CS_REG = device;
			udelay(15);
			*FPGA_CS_REG = 0;
		}
		default: break;
	}
	return( 0 );
}

int spi_drdy( spi_ss_t device )
{
	u8 reg = *FPGA_CS_REG;
	switch( device )
	{
		case SS_DM_KEY: 	return(reg & DM_KEY_RDY_BIT);
		case SS_MM_KEY:		return(reg & MM_KEY_RDY_BIT);
		case SS_MM_SLIDER:	return(reg & MM_SLIDER_RDY_BIT);
		case SS_MM_WHEEL:	return(reg & MM_WHEEL_RDY_BIT);
		default:			return( 1 );
	}
	return( 1 );
}

void set_chassis_spk_amp_on(void)
{
	*FPGA_SPK_EN |= FPGA_SPK_EN_SPEAKER_EN_BIT;
	return;
}

void set_chassis_spk_amp_off(void)
{
	*FPGA_SPK_EN &= ~FPGA_SPK_EN_SPEAKER_EN_BIT;
	return;
}

int is_chassis_spk_amp_on(void)
{
	return((*FPGA_SPK_EN) & FPGA_SPK_EN_SPEAKER_EN_BIT);
}

static int __init fpga_init(void)
{
	u32 reg;
	printk(KERN_NOTICE "Siemens fpga version %d\n",*FPGA_VER_REG);

	*FPGA_TEST_REG1 = 0;
	printk(KERN_NOTICE "\t%s %s\n","Lan power",			((*FPGA_LAN_RESET) & (1<<0)) ? "Off":"On");		
	printk(KERN_NOTICE "\t%s %s\n","Display power",		((*FPGA_DISP_RESET) & (1<<0)) ? "Off":"On");		
	printk(KERN_NOTICE "\t%s %s\n","BlueTooth",			((*FPGA_BT_RES) & (1<<0)) ? "Off":"On");		
	printk(KERN_NOTICE "\t%s %s\n","Hook switch Unit",	((*FPGA_GU_EN) & (1<<0)) ? "On":"Off");		
	printk(KERN_NOTICE "\t%s %s\n","USB power",			((*FPGA_USB_EN) & (1<<0)) ? "On":"Off");		
	printk(KERN_NOTICE "\t%s %s\n","Speaker power",		((*FPGA_SPK_EN) & (1<<0)) ? "On":"Off");		
	printk(KERN_NOTICE "\t%s %s\n","Sidecar power",		((*FPGA_SC_EN) & (1<<0)) ? "On":"Off");		

	reg = *FPGA_INT_MASK;
	printk(KERN_NOTICE "\nFPGA Interrupt mask     %04x\n",reg);
	reg = *FPGA_INT_PEN;
	printk(KERN_NOTICE "FPGA Interrupts pending %04x\n",reg);
	reg = *FPGA_INT_EVENT;
	printk(KERN_NOTICE "FPGA Interrupts events %04x\n",reg);


	fpga_isr_init();

	printk(KERN_NOTICE "\nEnabling Hook Switchs\n");		
	*FPGA_GU_EN = 1;
	printk(KERN_NOTICE "\t%s %s\n","Hook switch",		((*FPGA_GU_OFFHOOK) & (1<<0)) ? "Off":"On");		
	
	printk(KERN_NOTICE "\nDisabling Chassis Speaker\n");
	*FPGA_SPK_EN = 0;
	
	printk(KERN_NOTICE "\nEnabling LED's\n");		
	*FPGA_MMI_PWM = 60;

	printk(KERN_NOTICE "\nEnabling USB\n");		
	*FPGA_USB_EN = 1;

	printk(KERN_NOTICE "%s %s\n\n","NAND Flash is ",((*FPGA_NAND_WP) & (1<<0)) ? "write-protected":"writable");		
	return 0; 
}
							   
static void __exit fpga_exit(void)
{
	fpga_isr_release();
	printk(KERN_NOTICE "fpga exit\n");
}


/* sidecar section  ****************************************************/
int spi_ss_sidecar_set(){
    
    u8 reg;
    u8 newreg = 0;
    
    reg = 0x00;
    //set the value of signal to zero
    *FPGA_SC_CS = reg;
    newreg = *FPGA_SC_CS;
    //printk(KERN_INFO "spi_ss_sidecar_set: FPGA_SC_CS:%X \n",newreg);
    return 0;
    
}

int spi_ss_sidecar_clear(){
    
    u8 reg;
    u8 newreg = 0;
    
    reg = 0x01;
    //set the value of signal to one
    *FPGA_SC_CS = reg;
    newreg = *FPGA_SC_CS;
    //printk(KERN_INFO "spi_ss_sidecar_clear: FPGA_SC_CS:%X \n",newreg);
    return 0;
    
}

void sidecar_power( eSidecarPower state )
{
	
	switch( state )
	{
		 case SIDECAR_POWER_OFF: *FPGA_SC_EN = 0; break;
		 case SIDECAR_POWER_ON: *FPGA_SC_EN = 1; break;
		 default: break;
	}
}

/***********************************************************************/

module_init(fpga_init);
module_exit(fpga_exit);

EXPORT_SYMBOL(fpga_reg_read);
EXPORT_SYMBOL(fpga_reg_write);
EXPORT_SYMBOL(spi_ss_set);
EXPORT_SYMBOL(spi_ss_clear);
EXPORT_SYMBOL(spi_ss_sync);
EXPORT_SYMBOL(spi_drdy);
EXPORT_SYMBOL(set_chassis_spk_amp_on);
EXPORT_SYMBOL(set_chassis_spk_amp_off);
EXPORT_SYMBOL(is_chassis_spk_amp_on);
EXPORT_SYMBOL(request_fpga_irq);
EXPORT_SYMBOL(free_fpga_irq);
EXPORT_SYMBOL(enable_fpga_irq);
EXPORT_SYMBOL(disable_fpga_irq);

EXPORT_SYMBOL(spi_ss_sidecar_set);
EXPORT_SYMBOL(spi_ss_sidecar_clear);
EXPORT_SYMBOL(sidecar_power);

