/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* Opera FPGA Serial driver
* Used ifx_asc.c as the initial template.
*/

#if 1
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/circ_buf.h>
#include <linux/serial.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/proc_fs.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/serial_core.h>

#include <linux/delay.h>

#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/irq.h>

#else

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/port.h>


#include <linux/tqueue.h> /* task queues */
#include <linux/time.h>

#endif

#include "fpga_defs.h"
#include "fpga.h"

#define MY_NAME	"fpgaSerial"

#define dbg(fmt, arg...)						\
	do {								\
		if (debug)						\
			printk ("%s: %s: " fmt "\n",		\
				MY_NAME , __FUNCTION__ , ## arg);	\
	} while (0)
#define err(format, arg...) printk(KERN_ERR "%s: " format "\n" , MY_NAME , ## arg)
#define info(format, arg...) printk(KERN_INFO "%s: " format "\n" , MY_NAME , ## arg)
#define warn(format, arg...) printk(KERN_WARNING "%s: " format "\n" , MY_NAME , ## arg)

static int debug = 0;
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug, "Debugging mode enabled or not");

#define FPGA_SERIAL_PORT_COUNT	1
#define FPGA_SERIAL_DEFAULT_BAUDRATE B115200

/** Major ID of Backlight device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID Opera FPGA Serial (default: 0=dynamic)");


/* for debug */
#define DEB_BUF_SIZE (1<<8)
volatile int rx_count = 0;
volatile int tx_count = 0;
int flipCountOffset = 0;
volatile unsigned char rxBuf[DEB_BUF_SIZE];
volatile unsigned char txBuf[DEB_BUF_SIZE];
unsigned short int flipCountBuf[DEB_BUF_SIZE];


static int fpgaSerial_structs_initialized = 0; 
static int rx_full_count[FPGA_SERIAL_PORT_COUNT]={};
static int rx_nearly_count[FPGA_SERIAL_PORT_COUNT]={};
static struct uart_driver fpgaSerial_drv;
static struct tty_driver normal, callout;
static struct tty_struct *fpgaSerial_table[FPGA_SERIAL_PORT_COUNT];
static struct termios *fpgaSerial_termios[FPGA_SERIAL_PORT_COUNT];
static struct termios *fpgaSerial_termios_locked[FPGA_SERIAL_PORT_COUNT];
static struct uart_port fpgaSerial_ports[FPGA_SERIAL_PORT_COUNT];
static struct uart_info *fpgaSerial_info[FPGA_SERIAL_PORT_COUNT];

/* FPGA UART does not have enables for rx and tx */
static volatile int txEnable = 0;
static volatile int rxEnable = 0;

/* Timer for delayed processing - EG we want to recall isr routine in the future */
static struct timer_list timer;

/**
* Device filesystem handle structure for the creation of device files
*/
/*static devfs_handle_t input_devfs_handle;*/

static u_int fpgaSerial_tx_empty(struct uart_port *port)
{
    int status;
    dbg();
/*
	if( txEnable == 0 )
    	return TIOCSER_TEMT;
    else
    	return 0;
*/	
    /*
     * if tx not busy transmitter is empty.
     */
    status = (*FPGA_V24_STATCON) & FPGA_V24_TXBUSY;
    return status ? 0 : TIOCSER_TEMT;
}

static void fpgaSerial_set_mctrl(struct uart_port *port, u_int mctrl)
{
    dbg();
    /*
    printk("\nSet Modem:%08x\n",mctrl);
    if( TIOCM_RTS & mctrl )
    {
    	printk("RTS set\n");
    }
    else
    {
    	printk("RTS clear\n");
    }
    if( TIOCM_DTR & mctrl )
    {
    	printk("DTR set\n");
    }
    else
    {
    	printk("DTR clear\n");
    }
    */
    /* no modem control - just return */
    return;
}

static u_int fpgaSerial_get_mctrl(struct uart_port *port)
{
	uint status;
	u_int mctrl = 0;
    dbg();
    
	status = *FPGA_V24_STATCON;

	if( (status & FPGA_V24_CTS) != 0 )
	{
		mctrl |= TIOCM_CTS;
	}

	if( (status & FPGA_V24_RTS) != 0 )
	{
		mctrl |= TIOCM_RTS;
	}
/*
    printk("Get Modem:%08x\n",mctrl);
*/
    /* no modem control signals - the readme says to pretend all are set */
    return TIOCM_CTS|TIOCM_CAR|TIOCM_DSR;
}

/**
 * Stop transmission.
 * This function stops transmission by disabling the Tx interrupt.
 *
 * \param port - not used
 * \param from_tty - not used
 */
static void fpgaSerial_stop_tx(struct uart_port *port, u_int from_tty)
{
    /* dbg(); */

    /* We have no seperate rx and tx interrupts; disable tx int at SW layer */

    txEnable = 0;
    printk(KERN_INFO "fpga serial stop tx\n");
}

/**
 * Start transmission.
 * This function starts transmission by enabling the Tx interrupt. If the info
 * structure has already been set, buffered chars will be sent out if necessary.
 *
 * \param port - not used
 * \param nonempty - not used
 * \param from_tty - not used
 */
static void fpgaSerial_start_tx(struct uart_port *port, u_int nonempty, u_int from_tty)
{
	struct uart_info *info;

	uint status;
	unsigned char tx_char;
	info = fpgaSerial_info[port->line];
//    dbg();
    if (info)
    {
    	if( txEnable != 0 )
    	{
//			printk("fpgaSerial: still in tx %d %d\n",nonempty,from_tty);
			return;
    	}

    	status = *FPGA_V24_STATCON;
    	if( (status & FPGA_V24_CTS) == 0 )
    	{
			printk(KERN_ERR "fpgaSerial: tx no cts\n");
			return;
	    }

		if( (status & FPGA_V24_TXBUSY) == 0 )
		{
			tx_char = info->xmit.buf[info->xmit.tail];
    		info->xmit.tail = (info->xmit.tail + 1) & (UART_XMIT_SIZE - 1);
    		port->icount.tx++;
    		txEnable = 1;
			txBuf[tx_count++ & (DEB_BUF_SIZE -1)] = tx_char;
			asm("sync");
			*FPGA_V24_TXDAT = tx_char;
		}
		else
		{
			printk(KERN_ERR "fpgaSerial: tx busy\n");
		}
    }
    else
    {
		printk(KERN_ERR "fpgaSerial: bad line\n");
    }
}

/**
 * Stop reception.
 * This function stops the receiver.
 *
 * \param port - not used
 */
static void fpgaSerial_stop_rx(struct uart_port *port)
{
    dbg();

    /* We have no rx enable in HW do this in SW */

    rxEnable = 0;
    printk(KERN_INFO "fpga serial stop rx\n");
}

/**
 * Enable modem signals.
 * This function should enable modem signals, but this is not supported.
 *
 * \param port - not used
 */
static void fpgaSerial_enable_ms(struct uart_port *port)
{
    dbg();
    /* no modem signals */
    return;
}



/**
 * The serial interface has a 32 byte receive fifo and has a fixed baud rate
 * of 115200 1 start and stop bit (total 10 bits per byte).
 * 
 * At this speed a character will be recived every 86us it follows then that
 * the fifo takes 2.7 ms to fill. The flip buffer size is 512 bytes. When a 
 * buffer is pushed the flip before is scheduled for processing on the timer
 * queue. This queue should be serviced every 10ms. It takes 44ms to fill the
 * flip buffer so overruns should not occur.
 * 
 * If flip buffer scheduled every 10ms and we have a continuous stream we should
 * get a maximum of 116 chars per tick.
 */
 

static void fpgaSerial_fill_tty( struct uart_info *info )
{
    struct tty_struct *tty = info->tty;
    unsigned char ch;


	do
	{
		ch = *FPGA_V24_RXDAT;
		rxBuf[rx_count++ & (DEB_BUF_SIZE -1)] = ch;
		tty_insert_flip_char( tty,ch,TTY_NORMAL);
	} while ((*FPGA_V24_STATCON & FPGA_V24_RXNOTEMPTY) != 0);

	flipCountBuf[flipCountOffset++ & (DEB_BUF_SIZE-1)] = tty->flip.count;

	tty_flip_buffer_push(tty);
//	tty_schedule_flip(tty);

}




static int tx_byte(struct uart_info *info)
{
    struct uart_port *port = info->port;
    int count;
    unsigned char ch;
 
    if( info->xmit.head == info->xmit.tail )
    {
    	/* nothing to send */
    	txEnable = 0;
		return( 0 );
    }

    if (port->x_char)
    {
		printk(KERN_INFO "fpga serial tx x_char %x \n", port->x_char );
    }

	ch = info->xmit.buf[info->xmit.tail];
    info->xmit.tail = (info->xmit.tail + 1) & (UART_XMIT_SIZE - 1);
    port->icount.tx++;

    *FPGA_V24_TXDAT = ch;
	txBuf[tx_count++ & (DEB_BUF_SIZE -1)] = ch;

    
    if ( (count = CIRC_CNT(info->xmit.head, info->xmit.tail, UART_XMIT_SIZE)) == 0 )
    {
    	txEnable = 0;
        uart_event(info, EVT_WRITE_WAKEUP);
    }

	return( count );
}

int fpgaSerial_rxtx_int( eFpgaInterruptNum vector, void *data )
{
	uint status;
	struct uart_info *info;
	int tx_bytes = 1;
	int loop_count = 0;

	info = (struct uart_info *)data;

	status = *FPGA_V24_STATCON;
	
	do
	{
		
		if( status & FPGA_V24_RXFULL )
		{
			*FPGA_V24_STATCON = 0;
    		rx_full_count[info->port->line]++;
			printk(KERN_ERR "fpga serial full fifo %02x %d\n", status, loop_count);
		}

		if( status & FPGA_V24_RXALLMOSTFULL )
		{
			*FPGA_V24_STATCON = 0;
    		rx_nearly_count[info->port->line]++;
			printk(KERN_ERR "fpga serial nearly full fifo %02x %d\n", status, loop_count);
		}

		/* while fifo not empty and more chars to tx */
		
		if( (status & FPGA_V24_RXNOTEMPTY) != 0 )
		{
			fpgaSerial_fill_tty( info );
		}
		
		*FPGA_INT_EVENT = (1<<FPGA_V24_INT);

		tx_bytes = 0;
		status = *FPGA_V24_STATCON;
		if( (status & FPGA_V24_CTS) == 0 )
		{
			printk(KERN_ERR "fpga serial NO cts\n" );
			status = *FPGA_V24_STATCON;
		}

		if((status & (FPGA_V24_TXBUSY | FPGA_V24_CTS)) == FPGA_V24_CTS )
		{
			/* Transmitter empty */
			tx_bytes = tx_byte( info );
		}
		status = *FPGA_V24_STATCON;
		loop_count++;
		
	} while( ( status & FPGA_V24_RXNOTEMPTY) || tx_bytes );

	if( (status & FPGA_V24_RTS) == 0 )
	{
		*FPGA_V24_STATCON = FPGA_V24_RTS;
	}
	
//	ack_incaip2_irq(INT_NUM_IM5_IRL6);
//	*INCA_IP2_ICU_IM5_ISR = INCA_IP2_ICU_IM5_IR(INT_NUM_IM5_IRL6 - 160);
	
	return( 0 );
}


static void fpgaSerial_cts_timeout( unsigned long data )
{
//	struct uart_info *info = (struct uart_info *)data;

	if( (*FPGA_V24_STATCON & FPGA_V24_CTS) != 0 )
	{
		timer.expires = jiffies+1;
		add_timer( &timer );
	}
	else
	{
		printk(KERN_ERR "fpga serial restarting transmit after cts\n");
	}
}

static void fpgaSerial_break_ctl(struct uart_port *port, int break_state)
{
    dbg();
    /* no way to send a break */
    return;
}


static int fpgaSerial_startup(struct uart_port *port, struct uart_info *info)
{
    unsigned long flags;
    int i;
    unsigned int status;

    dbg();

    /* release the UART FPGA and Blutooth reset */
    *FPGA_BT_RES = 0x01;
    asm("sync");
    udelay( 100 );
    *FPGA_BT_WAKUP = 0x01;
    asm("sync");
    *FPGA_BT_RES = 0x00;
    asm("sync");

    /* Poll Bluemoon expecting CTS to be driven */
    /* Poll 1000 * 250us */

    i = 1000;
    while( i-- ){
    	status = (*FPGA_V24_STATCON);
	if( status & FPGA_V24_CTS ){
		break;
	}
	udelay( 250 );
    }

    if( (status & FPGA_V24_CTS) == 0 ){
	/* Bluemoon never came ready */
	printk(KERN_ERR "fpgaSerial_startup - Bluemoon not ready\n");
	return( 1 );
    }

//    printk("fpgaSerial_startup Bluemoon ready after %d uS\n", (1000-i)*250 );

    fpgaSerial_ports[port->line].uartclk = 1843200;
    fpgaSerial_info[port->line] = info;

    init_timer( &timer );
    timer.data = (int)info;
    timer.function = fpgaSerial_cts_timeout;

    /* block the IRQs */
    local_irq_save(flags);

    /* Install the tx/rx interrupt handler */

    /* flush receive FIFO and enable receiver */
    (*FPGA_V24_STATCON) = FPGA_V24_RXFLUSH;

    request_fpga_irq( FPGA_V24_INT, fpgaSerial_rxtx_int, info );

    rxEnable = 1;

    /* unblock the IRQs */
    local_irq_restore(flags);

    enable_fpga_irq( FPGA_V24_INT );

    /* Generate RTS to inform partner we are ready to recieve */
    (*FPGA_V24_STATCON) = FPGA_V24_RTS;
    
    return 0;
}

static void fpgaSerial_shutdown(struct uart_port *port, struct uart_info *info)
{
    unsigned long flags;

    dbg();

    /* block the IRQs */
    local_irq_save(flags);

    txEnable = 0;
    rxEnable = 0;

    del_timer_sync( &timer );
    disable_fpga_irq( FPGA_V24_INT );
    free_fpga_irq( FPGA_V24_INT );

    /* unblock the IRQs */
    local_irq_restore(flags);
}

static void fpgaSerial_change_speed(struct uart_port *port, u_int cflag, u_int iflag, u_int quot)
{
    dbg();
    /* FPGA UART only supports 115200 Baud, 1 startbit, 8 databit, 1 stopbit and RTS/CTS */
    return;
}

static const char *fpgaSerial_type(struct uart_port *port)
{
    dbg();
    return "FPGA_SERIAL";
}

/*
 * Release the memory region(s) being used by 'port'
 */
static void fpgaSerial_release_port(struct uart_port *port)
{
    dbg();
    return;
}

/*
 * Request the memory region(s) being used by 'port'
 */
static int fpgaSerial_request_port(struct uart_port *port)
{
    dbg();
    return 0;
}

/*
 * Configure/autoconfigure the port.
 */
static void fpgaSerial_config_port(struct uart_port *port, int flags)
{
    dbg();
    if (flags & UART_CONFIG_TYPE) {
        port->type = PORT_SIE_FPGA;
        fpgaSerial_request_port(port);
    }
}

/*
 * verify the new serial_struct (for TIOCSSERIAL).
 */
static int fpgaSerial_verify_port(struct uart_port *port, struct serial_struct *ser)
{
    int ret = 0;
    dbg();
#if 0
    if (ser->type != PORT_UNKNOWN && ser->type != PORT_IFX_ASC)
        ret = -EINVAL;
    if (ser->irq < 0 || ser->irq >= NR_IRQS)
        ret = -EINVAL;
    if (ser->baud_base < 9600)
        ret = -EINVAL;
#endif
    return ret;
}

    
static struct uart_ops fpgaSerial_pops = {
	tx_empty: fpgaSerial_tx_empty,
	set_mctrl: fpgaSerial_set_mctrl,
	get_mctrl: fpgaSerial_get_mctrl,
	stop_tx: fpgaSerial_stop_tx,
	start_tx: fpgaSerial_start_tx,
	stop_rx: fpgaSerial_stop_rx,
	enable_ms: fpgaSerial_enable_ms,
	break_ctl: fpgaSerial_break_ctl,
	startup: fpgaSerial_startup,
	shutdown: fpgaSerial_shutdown,
	change_speed: fpgaSerial_change_speed,
	type: fpgaSerial_type,
	release_port: fpgaSerial_release_port,
	request_port: fpgaSerial_request_port,
	config_port: fpgaSerial_config_port,
	verify_port: fpgaSerial_verify_port,
};

static void fpgaSerial_init_hardware( void )
{
}

static void fpgaSerial_init_structs( void )
{
    int t;
    
    for (t=0;t<FPGA_SERIAL_PORT_COUNT;t++) {

        
/* THIS NEEDS FIXING */
        fpgaSerial_ports[t].membase= (void *)SIEMENS_FPGA;
        fpgaSerial_ports[t].mapbase=(u_long) SIEMENS_FPGA;
        fpgaSerial_ports[t].irq=0;
        fpgaSerial_ports[t].unused[0]=0;
        fpgaSerial_ports[t].unused[1]=0;
        fpgaSerial_ports[t].unused[2]=0;
        fpgaSerial_ports[t].uartclk=0; /* filled in dynamically */
        fpgaSerial_ports[t].fifosize=32;

        fpgaSerial_ports[t].iotype=SERIAL_IO_MEM;
        fpgaSerial_ports[t].type=PORT_SIE_FPGA;
        fpgaSerial_ports[t].ops=&fpgaSerial_pops;
        fpgaSerial_ports[t].flags=ASYNC_BOOT_AUTOCONF;
        fpgaSerial_ports[t].line=t;
    }
    
    fpgaSerial_drv.owner=NULL;
    /* fpgaSerial_drv.normal_major=major; */
    fpgaSerial_drv.normal_major=200;
#ifdef CONFIG_DEVFS_FS
    fpgaSerial_drv.normal_name="bttty%d";
    fpgaSerial_drv.callout_name="btcua%d";
#else
    fpgaSerial_drv.normal_name="bttty";
    fpgaSerial_drv.callout_name="btcua";
#endif
    fpgaSerial_drv.normal_driver=&normal;
    /*fpgaSerial_drv.callout_major=CALLOUT_IFX_ASC_MAJOR;*/
    fpgaSerial_drv.callout_major=201;
    fpgaSerial_drv.callout_driver=&callout;
    fpgaSerial_drv.table=fpgaSerial_table;
    fpgaSerial_drv.termios=fpgaSerial_termios;
    fpgaSerial_drv.termios_locked=fpgaSerial_termios_locked;
    fpgaSerial_drv.minor=1;
    fpgaSerial_drv.nr=FPGA_SERIAL_PORT_COUNT;
    fpgaSerial_drv.port=fpgaSerial_ports;
    /* fpgaSerial_drv.cons=IFX_ASC_CONSOLE; */

    fpgaSerial_structs_initialized = 1;

//    printk("fpgaSerial_init_structs ops %x ops.startup %x\n", (int)&fpgaSerial_pops, (int)fpgaSerial_pops.startup );
}


int fpgaSerial_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    long    len = 0;
    int t=0, i;
	int rx_buffer_count;
	int tx_buffer_count;
	int flip_buffer_count;
	
//	printk("buffer size=%d\n",count);
    len = sprintf(buf,"\nSiemens FpgaSerial Debug %d\n", count);


    for (t=0;t<FPGA_SERIAL_PORT_COUNT;t++) {

        len += sprintf(buf+len,"Port FPGASERIAL%d\n",t);
        len += sprintf(buf+len,"rx_full_count=%d\n",rx_full_count[t]);
        len += sprintf(buf+len,"rx_nearly_count=%d\n",rx_nearly_count[t]);
        len += sprintf(buf+len,"tx_count=%d\n",tx_count);
        len += sprintf(buf+len,"rx_count=%d\n",rx_count);
        len += sprintf(buf+len,"flipCountOffset=%d\n",flipCountOffset);
        len += sprintf(buf+len,"status=%02x txEnable=%d\n",*FPGA_V24_STATCON,txEnable);
        len += sprintf(buf+len,"tty->low_latency = %d\n",fpgaSerial_table[t]->low_latency);
        len += sprintf(buf+len,"tty->max_flip_cnt = %d\n",fpgaSerial_table[t]->max_flip_cnt);


    
		len += sprintf(buf+len,"txBuf:\n");
	
		tx_buffer_count = (tx_count > DEB_BUF_SIZE) ? DEB_BUF_SIZE: tx_count;
	
		for( i=0; i<tx_buffer_count; i++ ){
			len += sprintf(buf+len,"%2.2x ", txBuf[i]);
		}
	
		len += sprintf(buf+len,"\n");
	
		len += sprintf(buf+len,"rxBuf:\n");
	
		rx_buffer_count = (rx_count > DEB_BUF_SIZE) ? DEB_BUF_SIZE: rx_count;
	
		for( i=0; i<rx_buffer_count; i++ ){
			len += sprintf(buf+len,"%2.2x ", rxBuf[i]);
		}
	
	    len += sprintf(buf+len,"\n");
	
		len += sprintf(buf+len,"flipCountBuf:\n");
		flip_buffer_count = (flipCountOffset > DEB_BUF_SIZE) ? DEB_BUF_SIZE: flipCountOffset;

		for( i=0; i<flip_buffer_count; i++ ){
			len += sprintf(buf+len,"%d ", flipCountBuf[i]);
		}
        len += sprintf(buf+len,"\nlen=%ld\n",len);
    }
    
    *eof = 1;   // No more data available
    return len;
}


static int __init fpgaSerial_init(void)
{
    int ret=0;

    dbg();

    if (fpgaSerial_structs_initialized == 0) {
        fpgaSerial_init_structs();
        fpgaSerial_init_hardware();
    }
    
    ret = uart_register_driver(&fpgaSerial_drv);

//    info("uart_register_driver sts %d\n", ret );
    

    /* Overwrite init_termios speed value - console will do some prints
       with UART default baudrate (115200) otherwise... */
    normal.init_termios.c_cflag = FPGA_SERIAL_DEFAULT_BAUDRATE | CS8 | CREAD | HUPCL | CLOCAL;

    /* Create proc file */ 
    create_proc_read_entry("driver/fpgaSerial", 0, NULL, fpgaSerial_read_procmem, NULL);

    return ret;
}

static void __exit fpgaSerial_exit(void)
{
    dbg();
    uart_unregister_driver(&fpgaSerial_drv);
}


#ifdef MODULE
module_init(fpgaSerial_init);
module_exit(fpgaSerial_exit);
#endif

EXPORT_NO_SYMBOLS;

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_AUTHOR("Owain Phillips");
MODULE_DESCRIPTION("Siemens FPGA Serial driver");
MODULE_LICENSE("GPL");
