/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
*	This module accepts the key presses from the FPK key pad (via keypadDevice ) and passes them
*	to the QT driver
*/


#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>
#include "Quantum_drivers.h"
#include "fpga.h"


/**
* Controls poll timer period
*/

/**
* Time out for poll timer is 40mS
*/
#define TIMER_INTERVAL_IN_MILLISECS 40


/** Major ID of QFPKey device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID QFPKey (default: 0=dynamic)");


/**
* Length of command queue buffer
*/

#define CMD_Q_LEN	(1<<4)

/**
* Mask for command queue buffer
*/

#define CMD_Q_MASK	((1<<4)-1)

/**
* Structure for kernel timer device used for the QFPKey poll 'tick' 
*/

typedef struct  {
	struct timer_list timer;
	struct tq_struct *task;
	int RunAquisition;
	spinlock_t lock;
	int wqPtr;
	int rqPtr;
	poll_data_t lastPoll;
	unsigned char wQueue[100];
} QFPKey_device_t;

static QFPKey_device_t QFPKey_data;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/

DECLARE_WAIT_QUEUE_HEAD(qFPK_wait_queue);

/**
* Declartion of a task queue structure
*/
struct tq_struct QFPKey_task; /* global: initialized to zero */


/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;

void QFPKeyServiceTask(void *ptr);


ssize_t QFPKey_read(struct file *file, char *buf, size_t size, loff_t * off);
unsigned int QFPKey_poll(struct file *fp, struct poll_table_struct *pts);
int QFPKey_open(struct inode *inode, struct file *fp);
int QFPKey_close(struct inode *inode, struct file *fp);


/**
* Poll timer time out function which schedules the QFPKey hardware data poll
* and reschedules its self.
*/
static void QFPKey_timeout(unsigned long data)
{
	QFPKey_device_t *dev = (QFPKey_device_t*)data;
	static int lost_flag = 0;
	unsigned long flags;

	/* Protect the following critical region */

	spin_lock_irqsave(&dev->lock,flags);

/**
* Diagnostic: print a warning message if the QFPKey_task failed to run
* when it was queued last time.
*/
	if (dev->RunAquisition == 0)
	{
		lost_flag = 0;
		dev->RunAquisition = 1;
	}
	else
	{
		if (!lost_flag )
		{
			printk(KERN_WARNING "Run Aquisition will be lost!\n");
			lost_flag = 1;
		}
	}

	/* Critical region ends here */
	spin_unlock_irqrestore(&dev->lock,flags);

	/* Wake up the reading process */

	schedule_task(&QFPKey_task);

	/**
	 * Start a timer that expires in
	 * TIMER_INTERVAL_IN_MILLISECS ms and that
	 * calls tl_timeout on expiration. Hint: tl->timer
	 */

	dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ)/1000;
	add_timer(&dev->timer);

}



/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations QFPKey_fops = {
	read:		QFPKey_read,	/*read */
	poll:		QFPKey_poll,  /* poll */
	open:		QFPKey_open,  /* open */
	release:	QFPKey_close, /* release */
};


/**
 * Function called by sheduled task queue by the timer timeout task.
 * It controls the physical hardware via inca_ip API function calls to poll the
 * hardware for data.
 *\param ptr pointer to kernel device timer structure
 *\returns pollData a byte of data read from the QFPKey device.
 */

static poll_data_t QFPKeyPoll(struct QFPKey_device_t* ptr)
{
	static poll_data_t nullPoll;
	poll_data_t pollData;


	nullPoll.poll = 0;
	pollData.poll = 0x7F; /* indicate error */

	if( spi_ss_set(SS_DM_KEY) == 0 )
	{
		return( pollData );
	}

	/* give time for chip to get its act together */
	udelay(15); /* Q511 this should be <= 13us */

	/* write poll to ssc driver*/
	quantum_write_spi(&nullPoll,sizeof(poll_data_t));


	/* read response back from scc driver */
	quantum_read_spi(&pollData,sizeof(poll_data_t));

	/* at 32Khz it takes 250us to transmit 8 bits */
	udelay(250);

	/* raise /SS */
	spi_ss_clear(SS_DM_KEY);
	udelay(15);  /* device holds onto SPI bus <= 12 us */

	return pollData;
}


/**
 * Aliased read function called from user space
 * Reads availale command acknowledges else during normal operation
 * it gets polled data off the incoming data queue buffer manipulated it and returns the result to usre space.
 *\param *file pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns either number of data items transferred or error state
 */

ssize_t QFPKey_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	QFPKey_device_t* dev = (QFPKey_device_t*)fp->private_data;

	int i;
	char keys[100];

	while((dev->rqPtr) == (dev->wqPtr))
	{/* nothing to read */

		if (fp->f_flags & O_NONBLOCK)
		{
			return -EAGAIN;
		}

		if(wait_event_interruptible(qFPK_wait_queue, (dev->rqPtr) != (dev->wqPtr)))
			return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
	}

	for( i = 0; i < size; i++ )
	{
		if( (dev->rqPtr) == (dev->wqPtr) )
			break;
		
		keys[i] = dev->wQueue[dev->rqPtr];
		if ((dev->rqPtr) < 100)
		{
			dev->rqPtr = (dev->rqPtr + 1);
		}
		else
		{
			dev->rqPtr = 0;	
		}
		printk("FPKeypad_read: %02x\n",(u8)keys[i]);
	}

	printk("FPKeypad_read: Total keys=%d Requested keys=%d\n",i, size);

	if(copy_to_user(buf, keys, i))
	{
		printk(KERN_ERR "FPKeypad_read: copy_to_user failed\n");
		return( -ERESTARTSYS );
	}

	return( i );
}



/**
 * Aliased poll function called from user space by poll or select
 * Currently used to wait on information being availble to be read from incoming data queue buffer.
 * Woken by the write_polldata function.
 *\param *fp pointer to a kernel file data structure
 *\param *pts pointer to a kernel poll table data structure
 *\returns flags to indicate valid data available
 */

unsigned int QFPKey_poll(struct file *fp, struct poll_table_struct *pts)
{
	unsigned int mask = 0;
	QFPKey_device_t* dev = (QFPKey_device_t*)fp->private_data;

	/*
	 * The buffer is circular.
	 */

	poll_wait(fp, &qFPK_wait_queue, pts);
	if ((dev->rqPtr) != (dev->wqPtr)) mask |= POLLIN | POLLRDNORM;  /* readable */

	return mask;
}

/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */

int QFPKey_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;
	QFPKey_device_t* dev;
	unsigned long flags;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "QFPKey_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &QFPKey_data;
	dev = (QFPKey_device_t*)fp->private_data;

	spin_lock_irqsave(&dev->lock,flags);
	QFPKey_data.lastPoll.poll = 0;

	spin_unlock_irqrestore(&tl->lock,flags);

	return 0;
}

/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int QFPKey_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "QFPKey_close:\n");
	err = 0;
	return err;
}



/**
 * Function placed on the sheduled task queue by the timer timeout task.
 * It polls the hardware for data and puts it into the incoming data queue buffer.
 */
void QFPKeyServiceTask(void *ptr)
{
	unsigned long flags;
	poll_data_t pollData;
	QFPKey_device_t* dev = (QFPKey_device_t*) ptr;
	char keyID;
	spin_lock_irqsave(&dev->lock,flags);

	dev->RunAquisition = 0;

	/* Critical Region ends here */
	spin_unlock_irqrestore(&dev->lock,flags);


	pollData = QFPKeyPoll(ptr);

	if( pollData.poll == 0x7f )
	{
		spi_ss_sync(SS_DM_KEY);
		printk(KERN_NOTICE "FPKey: Error not in SYNC %02X %d\n",pollData.poll, pollData.poll);
	}
	else if( pollData.poll != 0 )
	{
		if( (pollData.poll & 0x80 ) == 0x80 )
		{

			printk(KERN_NOTICE "FPKey: Touch %d : key=%d\n",(pollData.poll & 0x7f), ((pollData.poll & 0x7f)/14 ));
//#if 0
			/* Not sure about the mapping yet!! */
			keyID = (pollData.poll & 0x7f);
			if (keyID>=0 && keyID<15)
//				keyID = PHYSICAL_PHONE_FPK_0; 
				keyID = 0; 
			else if (keyID>=15 && keyID<29)
//				keyID = PHYSICAL_PHONE_FPK_1;
				keyID = 1; 
			else if (keyID>=29 && keyID<43)
//				keyID = PHYSICAL_PHONE_FPK_2;
				keyID = 2; 
			else if (keyID>=43 && keyID<57)
//				keyID = PHYSICAL_PHONE_FPK_3;
				keyID = 3; 
			else if (keyID>=57 && keyID<71)
//				keyID = PHYSICAL_PHONE_FPK_4;
				keyID = 4; 
			else if (keyID>=71 && keyID<85)
//				keyID = PHYSICAL_PHONE_FPK_5;
				keyID = 5; 
			else if (keyID>=85 && keyID<99)
//				keyID = PHYSICAL_PHONE_FPK_6;
				keyID = 6; 
			else if (keyID>=99 && keyID<113)
//				keyID = PHYSICAL_PHONE_FPK_7;
				keyID = 7; 
			else if (keyID>=113 && keyID<127)
//				keyID = PHYSICAL_PHONE_FPK_8;
				keyID = 8; 
			else (printk(KERN_NOTICE "We should not have got here: invalid key!!\n"));

			printk(KERN_NOTICE "FPKey: key=%d\n",(keyID ));
//#endif	
//			keyID = ((pollData.poll & 0x7f)/14);
//printk(KERN_NOTICE "FPKey: Write the keyID %d to wQueue buffer dev->wqPtr =%d dev->wrPtr =%d\n",keyID,dev->wqPtr,dev->rqPtr);
//			dev->wQueue[dev->wqPtr] = ((pollData.poll & 0x7f)/14 );
			dev->wQueue[dev->wqPtr] = ( keyID );

			/* should protect against pointer curruption here I think */
			if ((dev->wqPtr) < 100)
			{
				dev->wqPtr = (dev->wqPtr + 1);
			}
			else
			{
				dev->wqPtr = 0;
			}
			wake_up_interruptible( &qFPK_wait_queue );
		}
		else
			printk(KERN_NOTICE "FPKey: Error %02X %d\n",pollData.poll, pollData.poll);
	}

}




/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *QFPKey_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/QFPKey'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int QFPKey_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "QFPKey Device version 1.0\n";
    printk(KERN_INFO "/proc/drivers/QFPKey %d\n",count);

    len = sizeof(info);

    if( count < sizeof(info))
    	len = count;
    memcpy(buf,info,len);

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int QFPKey_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;
}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init QFPKey_init(void)
{
	int res = 0;

	/* create and register device */

	if( (input_devfs_handle = devfs_mk_dir(NULL, "input", NULL)) == 0 )
		return -EBUSY;

	devfs_register(input_devfs_handle, "QFPKey",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&QFPKey_fops,0);


	if (major == 0) major = res;

	SET_MODULE_OWNER(&QFPKey_fops);

	printk(KERN_NOTICE "\nQFPKey device registered.\n");

	/* create PROC entry */
	if( (QFPKey_proc = create_proc_entry( "driver/QFPKey", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/QFPKey!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	QFPKey_proc->read_proc  = QFPKey_read_procmem;
	QFPKey_proc->write_proc = QFPKey_write_procmem;

	printk(KERN_NOTICE "/proc/driver/QFPKey created!\n");

	spin_lock_init(&QFPKey_data.lock);

	printk(KERN_INFO "initializing timer for the first time %d ticks\n",(TIMER_INTERVAL_IN_MILLISECS * HZ)/1000);

	init_timer(&QFPKey_data.timer);

	QFPKey_data.RunAquisition = 0;
	QFPKey_data.lastPoll.poll = 0;
	QFPKey_data.wqPtr = 0;
	QFPKey_data.rqPtr = 0;
	QFPKey_data.timer.data = (int)&QFPKey_data;
	QFPKey_data.timer.function = QFPKey_timeout;

	QFPKey_timeout((unsigned long)&QFPKey_data);

	QFPKey_task.routine = QFPKeyServiceTask; /*scheduled function*/
	QFPKey_task.data = &QFPKey_data; /*argument to function*/

	
	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit QFPKey_exit(void)
{


	remove_proc_entry("driver/QFPKey", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_INFO "stopping timer\n");
	/* Delete the timer */
	del_timer_sync( &QFPKey_data.timer );

	printk(KERN_NOTICE "QFPKey_exit\n");
}


#ifdef MODULE
module_init(QFPKey_init);
module_exit(QFPKey_exit);
#endif

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
