/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* This driver operates for both Workpoints 'low end' (WP1 and WP 2) and for     
* 'high end' (WP3 and WP4). WP_HI variants use the FPGA driver to power and
* monitor the optical hookswitch by requesting an FPGA interrupt and 
* sevicing any interrupt (in hookSwitchIntHander); WP_LO variants drive the 
* optical hookswitch directly by pulsing the GU_EN high for no more than 2 Ms
* every 50 Ms and servicing any interrupt from the GU_IRQ line (via external 
* interrupt mechanism)
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>
#include <asm/incaip2/irq.h> /* to enable external interrupts (EXIN1) */
#include <asm/incaip2/incaip2.h> /* for hardware timers GPTC */

#include "fpga.h"
#include "fpga_defs.h"

#define	WP12_GU_EN      	( 1 << 1 )		/* Hookswitch power ON */
#define	WP12_SPEAKER_EN     ( 1 << 2 )		/* Speaker power ON */
#define HOOK_SWITCH_INT_NUM INT_NUM_IM3_IRL29 	/* GU_IRQ interrupt line */
#define HOOK_SWITCH_TIMER_DELAY 50 * HZ / 1000	/* with 100Hz kernel speed this gives us a 50 Ms timer */

#define OFF_HOOK  		0x01
#define ON_HOOK 		0x00
#define HOOK_SWITCH_INIT 	0xF
int offHookPreviousStatus 	= OFF_HOOK;
int offHookDetected 		= OFF_HOOK;
int offHookCount 			= 0;

static int wp2_is_T40;

typedef struct {
	u8 event;
} hookEvent_t;

/** Major ID of HookSw device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID HookSwitch (default: 0=dynamic)");


/**
* Length of command queue buffer
*/

#define CMD_Q_LEN	(1<<4)

/**
* Mask for command queue buffer
*/

#define CMD_Q_MASK	((1<<4)-1)

/**
* Structure for kernel timer device used for the HookSw poll 'tick' and mouse emulation
*/

typedef struct  {
	spinlock_t lock;
	int wqPtr;
	int rqPtr;
	u8 hookState;
	unsigned char wQueue[CMD_Q_LEN];
} HookSw_device_t;

static HookSw_device_t HookSw_data;
static struct timer_list HookSwitch_timer;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/

DECLARE_WAIT_QUEUE_HEAD(HookSw_wait_queue);


/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;

ssize_t HookSw_read(struct file *file, char *buf, size_t size, loff_t * off);
unsigned int HookSw_poll(struct file *fp, struct poll_table_struct *pts);
int HookSw_open(struct inode *inode, struct file *fp);
int HookSw_close(struct inode *inode, struct file *fp);
void hookSwitchLampTimer_isr(int irq, void *dev_id, struct pt_regs *regs);
int HookSw_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg);


								  

static ssize_t readQueue( HookSw_device_t* tl, char * buf, size_t size )
{
	ssize_t ret;
	unsigned long flags;
	spinlock_t lock;

	char temp[CMD_Q_LEN];

	spin_lock_init(&lock);

//	printk(KERN_ERR "HookSwReadQueue: asked for %d\n",size);

	spin_lock_irqsave(lock,flags);

	for( ret = 0 ; (ret < size) && (tl->rqPtr != tl->wqPtr); ret++ )
	{
		temp[ret] = tl->wQueue[tl->rqPtr];
		tl->rqPtr = (tl->rqPtr + 1) & CMD_Q_MASK;
	}

	spin_unlock_irqrestore(lock,flags);

//	printk(KERN_ERR "HookSwReadQueue: found %d\n",ret);

	if( ret )
	{
		if (copy_to_user(buf, temp, ret))
		{
			printk(KERN_ERR "HookSwReadQueue: copy_to_user failed\n");
			ret = -ERESTARTSYS;
		}
		
	}
	return ret;
}

static ssize_t writeQueue( HookSw_device_t* tl, size_t size, char * buf )
{
	
	

	ssize_t len;
	unsigned long flags;
	spinlock_t lock;

	spin_lock_init(&lock);

	spin_lock_irqsave(lock,flags);

	/* work out how many fit in in Queue */
	len = CMD_Q_MASK - ((tl->wqPtr - tl->rqPtr) & CMD_Q_MASK);

	if( size > len )
	{
		/* won't fit */
		printk(KERN_ERR "HookSwWriteQueue: won't fit size = %d, len = %d\n", size,len);
		size = 0;
	}

	for( len = 0 ; (len < size) ; len++ )
	{
		tl->wQueue[tl->wqPtr] = buf[len];
		tl->wqPtr = (tl->wqPtr + 1) & CMD_Q_MASK;
	}
	spin_unlock_irqrestore(lock,flags);

	return len;
}

int hookSwitchIntHander( eFpgaInterruptNum vector, void *data )
{
	HookSw_device_t *dev = (HookSw_device_t *) data;
	u8 hookState;
	hookState = *FPGA_GU_OFFHOOK;
	if( dev->hookState	!= hookState )
	{
		dev->hookState = hookState;
		/* Don't do this in ISR's as it blocks other interrupt processing
		 * printk(KERN_NOTICE "hookSwitchIntHander vector %d Hook State=%02x\n",vector,hookState);
		 */
		if( writeQueue(dev, 1, &hookState) != 0 )
			wake_up_interruptible( &HookSw_wait_queue );
	}
	return( 0 );
}

void start_GPTC_timer(void)
{
	//  re-load and run the timer
	*INCA_IP2_GPTC_RUN_2A  		=   0x00000005;
}

void stop_GPTC_timer(void)
{
	//stop the timer running
	*INCA_IP2_GPTC_RUN_2A  		=   0x00000002;
}


static void HookSwitch_timerHandler(unsigned long data)
{
	/* 50 Ms has expired */
	/* Hook Switch POWER ON */
   	*INCA_IP2_GPIO_POUT |= WP12_GU_EN;

//	 printk(KERN_NOTICE "HookSwitch_timerHandler (checking)\n");
	/* and a 2 Ms delay interrupted to hookSwitchLampTimer_isr */
	start_GPTC_timer();  	

	/* and re-instate the 50 Ms timer */
	HookSwitch_timer.expires = jiffies + HOOK_SWITCH_TIMER_DELAY;
	add_timer(&HookSwitch_timer);
	return;
}

void hookSwitch_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	offHookDetected = OFF_HOOK;
	return;
}

void setup_GPTC_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	int err;

	if( (err = request_irq(INT_NUM_IM3_IRL24,hookSwitchLampTimer_isr, SA_INTERRUPT,"hookSwitchLampTimer", NULL)) != 0 )
	{
		printk(KERN_NOTICE "\nHOOK SWITCH PULSE ISR Init Failed\n");
	}

	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	//Power supply for GPTC module
	*INCA_IP2_PMS_PMS_GEN		|=	0x00000400;
	// set up the timer run at the same speed as the 8 kHz signal
	//set up Clock Control the same as mps_device.c, for voice firmware
	*INCA_IP2_GPTC_CLC 		=   0x00010114;
	//timer 1 mS
	*INCA_IP2_GPTC_CON_2A  		=   0x000001ca;
	// don't think that this value represents the 1 mS
	*INCA_IP2_GPTC_RELOAD_2A	=   0x00000001;
	*INCA_IP2_GPTC_COUNT_2A		=   0x00000001;
	//enable interrupt
	*INCA_IP2_GPTC_IRNEN   	   	|=  0x00000004;
	/* stop the timer running */
	*INCA_IP2_GPTC_RUN_2A  		=   0x00000002;

	spin_unlock_irqrestore(i_lock, i_flags);
}
void destroy_GPTC_timer(void)
{
	unsigned long i_flags;
	spinlock_t i_lock;

	free_irq(INT_NUM_IM3_IRL24,0);
	//timer 1 mS
	*INCA_IP2_GPTC_CON_2A  		=   0x00000000;
	*INCA_IP2_GPTC_RELOAD_2A	=   0x00000000;
	*INCA_IP2_GPTC_COUNT_2A		=   0x00000000;

	//disable interrupt
	spin_lock_init(&i_lock);
	spin_lock_irqsave(i_lock, i_flags);

	*INCA_IP2_GPTC_IRNEN   		&= ~0x00000004;

	spin_unlock_irqrestore(i_lock, i_flags);

	/* stop the timer running */
	*INCA_IP2_GPTC_RUN_2A  		=   0x00000002;

}

void hookSwitchLampTimer_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	//1 mS pulse of GU_EN  has completed, see description at head 
	u8 hookState = offHookPreviousStatus;
	stop_GPTC_timer();
	
//	printk(KERN_NOTICE "hookSwitchLampTimer_isr (checking)\n");
	if( wp2_is_T40 )
	{
		offHookDetected = (*INCA_IP2_GPIO_PIN & (1 << 3 )) ? OFF_HOOK : ON_HOOK;
	}
	
	*INCA_IP2_GPIO_POUT &= ~WP12_GU_EN;  /* HOOKswitch unit power off */
	/* set up the status after startup */
	if(offHookPreviousStatus != offHookDetected )
	{
		offHookPreviousStatus = offHookDetected;
		hookState = offHookDetected;
		if( hookState == OFF_HOOK )
		{
			offHookCount++;
		}
	}

	offHookDetected = ON_HOOK;

	/* pass the information to the GPAL interface */
	if( HookSw_data.hookState	!= hookState )
	{
		HookSw_data.hookState = hookState;
//		printk(KERN_NOTICE "hookSwitchLampTimer_isr - Hook State = %02x\n",hookState);
		if( writeQueue(&HookSw_data, 1, &hookState) != 0 )
		{
			wake_up_interruptible( &HookSw_wait_queue );
		}
	}
	return;
}





/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations HookSw_fops = {
	read:		HookSw_read,	/*read */
	poll:		HookSw_poll,  /* poll */
	open:		HookSw_open,  /* open */
	release:	HookSw_close, /* release * */
	ioctl:		HookSw_ioctl, /* ioctl * */
};




/**
 * Aliased read function called from user space
 * Reads availale command acknowlnowledges else during normal operation
 * it gets polled data off the incoming data queue buffer manipulated it and returns the result to usre space.
 *\param *file pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns either number of data items transferred or error state
 */

ssize_t HookSw_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	HookSw_device_t* dev = (HookSw_device_t*)fp->private_data;
	ssize_t ret=sizeof(hookEvent_t);


	while((dev->rqPtr) == (dev->wqPtr))
	{/* nothing to read */

		if (fp->f_flags & O_NONBLOCK)
		{
			return -EAGAIN;
		}

		if(wait_event_interruptible(HookSw_wait_queue, (dev->rqPtr) != (dev->wqPtr)))
			return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
	}

	ret = readQueue( dev, buf, size );

	return( ret );
}



/**
 * Aliased poll function called from user space by poll or select
 * Currently used to wait on information being availble to be read from incoming data queue buffer.
 * Woken by the write_polldata function.
 *\param *fp pointer to a kernel file data structure
 *\param *pts pointer to a kernel poll table data structure
 *\returns flags to indicate valid data available
 */

unsigned int HookSw_poll(struct file *fp, struct poll_table_struct *pts)
{
	unsigned int mask = 0;
	HookSw_device_t* dev = (HookSw_device_t*)fp->private_data;

	/*
	 * The buffer is circular.
	 */

	poll_wait(fp, &HookSw_wait_queue, pts);
	if ((dev->rqPtr) != (dev->wqPtr)) mask |= POLLIN | POLLRDNORM;  /* readable */

	return mask;
}

/**
 * Aliased ioctl function called from user space
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *filp pointer to a kernel file data structure
 *\param cmd ioctl command
 *\param arg associated command arguments
 * Currently not used
 *\returns 0 (success) or -EINVAL/-EFAULT (error)
 */


int HookSw_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
	int retVal = -1;
	HookSw_device_t* dev = (HookSw_device_t*)fp->private_data;

	int wpType = get_WP_type();
	switch(wpType)
	{
		case 1:
		case 2:
		{
			/* WP_LO gets it status from device data since 
			as device is brought up it initialises hookState 
			from the current state */
			/*
                                        printk(KERN_INFO "HookSw_ioctl: Workpoint 1 or 2, hookstate is %d\n",dev->hookState);
                                */
		        retVal = dev->hookState;	
			break;
		}
		case 3:		// WP3 and WP4 cases are the same
		case 4:
		{
			/* WP_HI gets its status from FPGA register GU_OFFHOOK, 
			   hookState only gets updated on change of status */
		        retVal = *FPGA_GU_OFFHOOK;	
			break;
		}
		default:	// Invalid phone type
		{
			printk(KERN_INFO "HookSw_ioctl: Workpoint not detected\n");
			retVal = -EINVAL;
		}
	}
	return retVal;
}


/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */



int HookSw_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;
	HookSw_device_t* dev;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "HookSw_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &HookSw_data;
	dev = (HookSw_device_t*)fp->private_data;
	return 0;
}

/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int HookSw_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "HookSw_close:\n");
	err = 0;
	return err;
}






/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *HookSw_proc;


/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/HookSw'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int HookSw_read_procmem(char *buf, char *start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "HookSw Device version 1.0\n";
    	printk(KERN_INFO "/proc/drivers/HookSw %d\n",count);

    	len = sizeof(info);

    	if( count < sizeof(info))
    		len = count;
    	memcpy(buf,info,len);
	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int HookSw_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;
}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init HookSw_init(void)
{
	int res = 0;
	int err;
	/* create and register device */

	if( (input_devfs_handle = devfs_register(0, "input/HookSw",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&HookSw_fops,0)) == 0 )
		return -EBUSY;

	if (major == 0) major = res;

	SET_MODULE_OWNER(&HookSw_fops);

	printk(KERN_NOTICE "\nHookSw device registered.\n");

	/* create PROC entry */
	if( (HookSw_proc = create_proc_entry( "driver/HookSw", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/HookSw!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	HookSw_proc->read_proc  = HookSw_read_procmem;
	HookSw_proc->write_proc = HookSw_write_procmem;

	printk(KERN_NOTICE "/proc/driver/HookSw created!\n");

	spin_lock_init(&HookSw_data.lock);

	HookSw_data.wqPtr = 0;
	HookSw_data.rqPtr = 0;
	HookSw_data.hookState = ON_HOOK;
	
	/* WP HI / LO detection */
	int wpType = get_WP_type();
	if(wpType > 2)
	{
	  printk(KERN_NOTICE "HookSw_init (WP %d detected) - Enabling fpga hook switch interrupts\n",wpType);
	  request_fpga_irq( FPGA_HOOKSWITCH_INT, hookSwitchIntHander, &HookSw_data );
 	  enable_fpga_irq( FPGA_HOOKSWITCH_INT );
	}else
	{
	  printk(KERN_NOTICE "HookSw_init (WP %d detected) - Enabling Speaker (no FPGA)\n",wpType);
	  *INCA_IP2_GPIO_POUT |= WP12_SPEAKER_EN;

      wp2_is_T40 = (incaip2_get_cpu_ver() < 3) ? 0 : 1;

	  if( wpType == 2 && wp2_is_T40 )
	  {
	  	printk(KERN_NOTICE "HookSw_init (WP %d T40 detected)\n");
	  }
	  else
	  {
		  if( (err = request_irq(HOOK_SWITCH_INT_NUM,hookSwitch_isr, SA_INTERRUPT,"hookSwitch", NULL)) == 0 )
		  {
	  		/* enable EXIN1 for rising edge - see interrupt.c */
			incaip2_enable_external_int(1,1);				
		  }
		  else
		  {
			printk(KERN_NOTICE "HookSw_init - failed to create Off Hook interrupt\n");
		  }
	  }
	  
	  offHookPreviousStatus = HOOK_SWITCH_INIT;
	  setup_GPTC_timer();
	  init_timer(&HookSwitch_timer);
	  HookSwitch_timer.function = &HookSwitch_timerHandler;
	  HookSwitch_timer.expires = jiffies + HOOK_SWITCH_TIMER_DELAY;
	  add_timer(&HookSwitch_timer);
	   
	}
	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit HookSw_exit(void)
{
	/* WP HI / LO detection */
	int wpType = get_WP_type();
	if(wpType > 2)
	{
	  disable_fpga_irq( FPGA_HOOKSWITCH_INT );
	  free_fpga_irq( FPGA_HOOKSWITCH_INT );
	}else
	{
    	  destroy_GPTC_timer();
	  /* disable EXIN1 interrupt and stop the kernel timer */
	  if( ! wp2_is_T40 )
	  {
		  incaip2_disable_external_int(1);
		  free_irq(HOOK_SWITCH_INT_NUM,0);
	  }
	  del_timer_sync(&HookSwitch_timer);
	}

	/* common */
	remove_proc_entry("driver/HookSw", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_NOTICE "HookSw_exit\n");
}


#ifdef MODULE
module_init(HookSw_init);
module_exit(HookSw_exit);
#endif

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
