/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
* The DMM led driver interfaces to the 595 shift-registers
* data is updated every 40ms.
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/port.h>

//#include "/vobs/Opera_Platform_Linux/Linux_BSP/ifx-com-linux/source/kernel/ifx/bsp/include/asm-mips/incaip2/port.h"

#include <linux/tqueue.h> /* task queues */
#include <linux/time.h>
#include "Quantum_drivers.h"


#define INCA_IP2_PORT_BASE	(KSEG1+0x1f102600)

#define PORT_OUT_REG		((volatile u32*)(INCA_IP2_PORT_BASE + 0x0000))
#define PORT_IN_REG			((volatile u32*)(INCA_IP2_PORT_BASE + 0x0004))
#define PORT_DIR_REG		((volatile u32*)(INCA_IP2_PORT_BASE + 0x0008))
#define PORT_ALTSEL1_REG	((volatile u32*)(INCA_IP2_PORT_BASE + 0x000C))
#define PORT_ALTSEL2_REG	((volatile u32*)(INCA_IP2_PORT_BASE + 0x0010))
#define PORT_STOFF_REG		((volatile u32*)(INCA_IP2_PORT_BASE + 0x0014))
#define PORT_OD_REG			((volatile u32*)(INCA_IP2_PORT_BASE + 0x0018))
#define PORT_PUDEN_REG		((volatile u32*)(INCA_IP2_PORT_BASE + 0x001C))
#define PORT_PUDSEL_REG		((volatile u32*)(INCA_IP2_PORT_BASE + 0x0020))






/** Major ID of hwVersion device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID hwVersion (default: 0=dynamic)");



/**

*/

unsigned int Version_number;




/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;


ssize_t hwVersion_read(struct file *file, char *buf, size_t size, loff_t * off);
int hwVersion_open(struct inode *inode, struct file *fp);

/*
The first column is as documented
in the schematic. The second column
is as read from the GPIO port. This
is just a byte swop of the first.

  			ttvv    vvtt
OS20 		0000	0000
OS20G 		0010	1000
OS40 		0100	0001
OS40G 		0110	1001
OS60 		1010	1010
OS60G 		1000	0010
OS80 		1100	0011
OS80G 		1110	1011
OS15  		0001	0100
os40 CORE 	0101	0101
*/
static void phoneVersion(void)
{
	u32 Phonetype = ((*PORT_IN_REG) >> 12) & 0xF;
	char *p = "(null)";
	
	switch( Phonetype )
	{
	case 0x0: p = "OS20"; break;
	case 0x1: p = "OS40"; break;
	case 0x2: p = "OS60"; break;
	case 0x3: p = "OS80"; break;
	case 0x4: p = "OS15"; break;
	case 0x5: p = "OS40 CORE"; break;
	case 0x8: p = "OS20G"; break;
	case 0x9: p = "OS40G"; break;	
	case 0xA: p = "OS60G"; break;	
	case 0xB: p = "OS80G"; break;	
	}
	printk(KERN_NOTICE "Phone Type %s (%02X)\n",p,Phonetype);
}




/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/

struct file_operations hwVersion_fops = {
	open:		hwVersion_open,  /* open */
	read:		hwVersion_read,	/*read */


};

/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */
int hwVersion_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	//printk(KERN_NOTICE "hwVersion_open: major = %d minor = %d\n", omajor, ominor);

	return 0;
}

/**
 * Function called from user space (by aliased read function below) and from kernel space.
 * It reads WP version and Gigabit capability data from Port Data Input Register.
 *\returns WP version/Gigabit data.
 */

int get_WP_version(void)
{
    return ( (*PORT_IN_REG >> 12) & 0x0000000F );
}

/**
 * Function called from kernel space.
 * It calls get_WP_version() above to get WP version and Gigabit capability data from Port Data Input Register.
 *\returns WP type.
 */

int get_WP_type(void)
{
	return ( (get_WP_version() & 3) + 1 );
}

/**
 * Aliased read function called from user space
 * Reads availale command acknowledges else during normal operation
 * it gets polled data off the incoming data queue buffer manipulated it and returns the result to usre space.
 *\param *file pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns either number of data items transferred or error state
 */

ssize_t hwVersion_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	ssize_t ret;

	struct sysinfo i;
	hwStats_t hwStats;

	if (size > sizeof(hwStats))
	{
		printk(KERN_WARNING "hwVersion_read: illegal size %d\n", size);
		return 0;
	}

	/* Get the hw version */
	hwStats.hwVersion = get_WP_version();

	/* Get the Memory Stats */
	si_meminfo(&i);
	hwStats.freeRAM = ((page_cache_size - i.bufferram) + i.freeram) * PAGE_SIZE;

	ret=sizeof(hwStats);

	if(copy_to_user(buf, &hwStats, ret))
	{
		printk(KERN_ERR "hwVersion_read: copy_to_user failed\n");
		ret = -ERESTARTSYS;
	}

	return( ret );

}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *hwVersion_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/hwVersion'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int hwVersion_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	int read_WP_type;
	
	len=sizeof(int);

	/**
	 * OS15 has to be recognized against OS20 here.
	 * Originally, get_WP_type() = ((get_WP_version() & 3) + 1) is returned.
	 *
	 * As get_WP_version() is 0 for OS20 and 4 for OS15, this was returning same value (1) for both OS20 and OS15.
	 */
	if (get_WP_version() == 4)
	{
		// OS15 here
		read_WP_type = 48;
	}
	else
	{
		read_WP_type = (get_WP_type() + 48);	// convert it to character "1" or "2" etc
	}

	if( count < len )
    	len = count;
    memcpy(buf, &read_WP_type,len);
	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int hwVersion_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init hwVersion_init(void)
{
	int res = 0;
	unsigned long flags;
	spinlock_t lock;

	spin_lock_init(&lock);
	spin_lock_irqsave(&lock,flags);

	/* set the hardware */
	*PORT_DIR_REG &= ~0xf000;

	spin_unlock_irqrestore(&lock,flags);

	/* create and register device */

	if( (input_devfs_handle = devfs_mk_dir(NULL, "input", NULL)) == 0 )
		return -EBUSY;

	devfs_register(input_devfs_handle, "hwVersion",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&hwVersion_fops,0);


	if (major == 0) major = res;

	SET_MODULE_OWNER(&hwVersion_fops);

	printk(KERN_NOTICE "\nhwVersion device registered.\n");

	/* create PROC entry */
	if( (hwVersion_proc = create_proc_entry( "driver/hwVersion", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/hwVersion!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}

	hwVersion_proc->read_proc  = hwVersion_read_procmem;
	hwVersion_proc->write_proc = hwVersion_write_procmem;

	phoneVersion();
	
#ifdef CONFIG_INCAIP2_WATCHDOG
	incaip2_wdt_init();
	incaip2_wdt_monitor_start();
	printk(KERN_NOTICE "Kernel watchdog monitor running\n");
#endif /* CONFIG_INCAIP2_WATCHDOG */

	return(0);
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit hwVersion_exit(void)
{

	remove_proc_entry("driver/hwVersion", 0);

	devfs_unregister(input_devfs_handle);


	printk(KERN_NOTICE "hwVersion_exit\n");
}


//#ifdef MODULE
module_init(hwVersion_init);
module_exit(hwVersion_exit);
//#endif

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");

EXPORT_SYMBOL(get_WP_version);
EXPORT_SYMBOL(get_WP_type);
