/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/** \file keyInput.c
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include "keyinput.h"


/**
 * WP Version being used. Set in the initialisation of the module
 */
typedef enum wpType {
	WP_LO,
	WP_HI
} wp_t;

static wp_t wp_type;
static int wp2_is_T40;
static int wp;  // wp type
static int wp_version; // wp version (distinct value for all WPs)

/** 
 * Array for holding the current state of the keypad keys
 */
static const int number_of_keypad_keys = 12;
static int keypad_current_state[12];

/**
 * Table to enable easier reference between keypad keys
 * and the keypad_current_state array
 */
typedef struct {
	unsigned long kernelKeyCode;
}keypad_mapping;

static keypad_mapping keypadMappingTable[] =
{
	{ KERNEL_KEYPAD_0 },
	{ KERNEL_KEYPAD_1 },
	{ KERNEL_KEYPAD_2 },
	{ KERNEL_KEYPAD_3 },
	{ KERNEL_KEYPAD_4 },
	{ KERNEL_KEYPAD_5 },
	{ KERNEL_KEYPAD_6 },
	{ KERNEL_KEYPAD_7 },
	{ KERNEL_KEYPAD_8 },
	{ KERNEL_KEYPAD_9 },
	{ KERNEL_KEYPAD_AST  },
	{ KERNEL_KEYPAD_HASH }
};


/** Major ID of QWheel device (default: 0=dynamic)
*/
static int major = 0;	// Major ID of keyInput device (default: 0=dynamic)

/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");

/**
* Module parameter description of variable major
*/
MODULE_PARM_DESC(major, "Major ID of KeyInput device (default: 0=dynamic)");


/**
* Length of command queue buffer
*/

#define KEY_EVENT_QUEUE_LEN		(1<<6)
#define KEY_EVENT_QUEUE_MASK	(KEY_EVENT_QUEUE_LEN-1)

#define PTR_INCR(x) x = ((x == KEY_EVENT_QUEUE_MASK) ? 0 : x +1)
#define FIFO_FULL(rp,wp) ((((rp) + KEY_EVENT_QUEUE_LEN - (wp)) & KEY_EVENT_QUEUE_MASK) == 2)

/**
* Structure for keyInput key & state machine data
*/
typedef struct {
	spinlock_t lock;
	int wqPtr;
	int rqPtr;
	key_event_t_ul keyEventQueue[KEY_EVENT_QUEUE_LEN];
} keyInputDevice_t;

/**
* An instance of the keyInput key & driver state machine data structure
*/
static keyInputDevice_t keyInput_Device;

/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* in our case for data to be returned from the hardware device.
*/
DECLARE_WAIT_QUEUE_HEAD(keyInput_wait_queue);


/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;

extern int  ifx_key_register_callback   (keypad_callback_t fp);
extern int  ifx_key_unregister_callback (void);

/**
* Prototype declaration of the file operations structure for keyInput device which
* effectively aliases application functions such as read, write etc. to equivalent
* kernel funtion
*/
ssize_t keyInput_read(struct file *, char *, size_t, loff_t *);
ssize_t keyInput_write(struct file *, const char *, size_t, loff_t *);

int keyInput_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int keyInput_open(struct inode *, struct file *);
int keyInput_close(struct inode *, struct file *);
unsigned int keyInput_poll(struct file *, struct poll_table_struct *);


// Local Definitions for Kernel<->Physical key translation

typedef struct {
	unsigned long hardware;
	unsigned long physical;
}keyDevice_mapping_t;

#define XLATE_ERROR	((unsigned long)(-1))

static keyDevice_mapping_t keyXlateTable[] =
{
	{ KERNEL_KEYPAD_0,		PHYSICAL_KEYPAD_0 },
	{ KERNEL_KEYPAD_1,		PHYSICAL_KEYPAD_1 },
	{ KERNEL_KEYPAD_2,		PHYSICAL_KEYPAD_2 },
	{ KERNEL_KEYPAD_3,		PHYSICAL_KEYPAD_3 },
	{ KERNEL_KEYPAD_4,		PHYSICAL_KEYPAD_4 },
	{ KERNEL_KEYPAD_5,		PHYSICAL_KEYPAD_5 },
	{ KERNEL_KEYPAD_6,		PHYSICAL_KEYPAD_6 },
	{ KERNEL_KEYPAD_7,		PHYSICAL_KEYPAD_7 },
	{ KERNEL_KEYPAD_8,		PHYSICAL_KEYPAD_8 },
	{ KERNEL_KEYPAD_9,		PHYSICAL_KEYPAD_9 },
	{ KERNEL_KEYPAD_AST,	PHYSICAL_KEYPAD_AST },
	{ KERNEL_KEYPAD_HASH,	PHYSICAL_KEYPAD_HASH },
	{ KERNEL_NAVI_KEY_NORTH,	PHYSICAL_NAVI_KEY_NORTH },
	{ KERNEL_NAVI_KEY_SOUTH,	PHYSICAL_NAVI_KEY_SOUTH },
	{ KERNEL_NAVI_KEY_WEST,		PHYSICAL_NAVI_KEY_WEST },
	{ KERNEL_NAVI_KEY_EAST,		PHYSICAL_NAVI_KEY_EAST },
	{ KERNEL_NAVI_KEY_CENTRAL,	PHYSICAL_NAVI_KEY_CENTRAL },
	{ KERNEL_WHEEL_KEY,		PHYSICAL_NAVI_TOUCH },
	{ KERNEL_HEADSET_KEY,	PHYSICAL_SOCKETED_HEADSET_KEY },
	{ KERNEL_HEADSET_HOOKSW, PHYSICAL_HEADSET_HOOK },
	{ KERNEL_LANSTATUS_KEY,	PHYSICAL_LANSTATUS_KEY },

	{ KERNEL_CONTROL_KEY_0,	PHYSICAL_CONTROL_KEY_0 },
	{ KERNEL_CONTROL_KEY_1,	PHYSICAL_CONTROL_KEY_1 },
	{ KERNEL_CONTROL_KEY_2,	PHYSICAL_CONTROL_KEY_2 },
	{ KERNEL_CONTROL_KEY_3,	PHYSICAL_CONTROL_KEY_3 },
	{ KERNEL_CONTROL_KEY_4,	PHYSICAL_CONTROL_KEY_4 },
	{ KERNEL_CONTROL_KEY_5,	PHYSICAL_CONTROL_KEY_5 },
	{ KERNEL_CONTROL_KEY_6,	PHYSICAL_CONTROL_KEY_6 },
	{ KERNEL_CONTROL_KEY_7,	PHYSICAL_CONTROL_KEY_7 },
	{ KERNEL_CONTROL_KEY_8,	PHYSICAL_CONTROL_KEY_8 },
	// Add any new control key values here

	{ KERNEL_MODE_KEY+0,	PHYSICAL_MODE_KEY_0 },
	{ KERNEL_MODE_KEY+1,	PHYSICAL_MODE_KEY_1 },
	{ KERNEL_MODE_KEY+2,	PHYSICAL_MODE_KEY_2 },
	{ KERNEL_MODE_KEY+3,	PHYSICAL_MODE_KEY_3 },
	{ KERNEL_MODE_KEY+4,	PHYSICAL_MODE_KEY_4 },
	{ KERNEL_MODE_KEY+5,	PHYSICAL_MODE_KEY_5 },
	// Add any new mode key values here

	{ KERNEL_FPK_KEY+0,		PHYSICAL_PHONE_FPK_0 },
	{ KERNEL_FPK_KEY+1,		PHYSICAL_PHONE_FPK_1 },
	{ KERNEL_FPK_KEY+2,		PHYSICAL_PHONE_FPK_2 },
	{ KERNEL_FPK_KEY+3,		PHYSICAL_PHONE_FPK_3 },
	{ KERNEL_FPK_KEY+4,		PHYSICAL_PHONE_FPK_4 },
	{ KERNEL_FPK_KEY+5,		PHYSICAL_PHONE_FPK_5 },
	{ KERNEL_FPK_KEY+6,		PHYSICAL_PHONE_FPK_6 },
	{ KERNEL_FPK_KEY+7,		PHYSICAL_PHONE_FPK_7 },
	{ KERNEL_FPK_KEY+8,		PHYSICAL_PHONE_FPK_8 },
	// Add any new FPK key values here

	{ KERNEL_VOLUME_KEY+0,	PHYSICAL_SLIDER_POSITION_0 },
	{ KERNEL_VOLUME_KEY+1,	PHYSICAL_SLIDER_POSITION_1 },
	{ KERNEL_VOLUME_KEY+2,	PHYSICAL_SLIDER_POSITION_2 },
	{ KERNEL_VOLUME_KEY+3,	PHYSICAL_SLIDER_POSITION_3 },
	{ KERNEL_VOLUME_KEY+4,	PHYSICAL_SLIDER_POSITION_4 },
	{ KERNEL_VOLUME_KEY+5,	PHYSICAL_SLIDER_POSITION_5 },
	{ KERNEL_VOLUME_KEY+6,	PHYSICAL_SLIDER_POSITION_6 },
	{ KERNEL_VOLUME_KEY+7,	PHYSICAL_SLIDER_POSITION_7 },
	{ KERNEL_VOLUME_KEY+8,	PHYSICAL_SLIDER_POSITION_8 },
	{ KERNEL_VOLUME_KEY+9,	PHYSICAL_SLIDER_POSITION_9 },
	{ KERNEL_VOLUME_KEY+10,	PHYSICAL_SLIDER_POSITION_10 },
	{ KERNEL_VOLUME_KEY+11,	PHYSICAL_SLIDER_POSITION_11 },
	{ KERNEL_VOLUME_KEY+12,	PHYSICAL_SLIDER_POSITION_12 },
	{ KERNEL_VOLUME_KEY+13,	PHYSICAL_SLIDER_POSITION_13 },
	{ KERNEL_VOLUME_KEY+14,	PHYSICAL_SLIDER_POSITION_14 },
	{ KERNEL_VOLUME_KEY+15,	PHYSICAL_SLIDER_POSITION_15 },
	{ KERNEL_SLIDER_KEY,	PHYSICAL_SLIDER_TOUCH },

	{ KERNEL_HOODED_CLAW_147,	 PHYSICAL_HOODED_CLAW_147 },
	{ KERNEL_HOODED_CLAW_35HASH, PHYSICAL_HOODED_CLAW_35HASH },
	{ KERNEL_HOODED_CLAW_289,    PHYSICAL_HOODED_CLAW_289 },
	{ KERNEL_HOODED_CLAW_013,    PHYSICAL_HOODED_CLAW_013 }
};

/** 
 * Number of WPs (currently 2 namely WP_LO and WP_HI)
 */ 
#define MAX_WPS 5 

/** 
 * Maximum number of keys on a WP
 */ 
#define MAX_KEYS 28

/**
 * Mapping from hardware values to Kernel key codes
 */
static keyDevice_mapping_t matrixKeyXlateTable[MAX_WPS][MAX_KEYS] =
{
		   /** WP_LO Workpoint 2 (T28 Hardware only) **/
	  { { 13, KERNEL_KEYPAD_0 	  },
      { 20, KERNEL_KEYPAD_1 	  },
	  { 16, KERNEL_KEYPAD_2 	  },
  	  { 11, KERNEL_KEYPAD_3 	  },
	  {  1, KERNEL_KEYPAD_4 	  },
	  { 15, KERNEL_KEYPAD_5 	  },
	  { 10, KERNEL_KEYPAD_6 	  },
	  {  0, KERNEL_KEYPAD_7 	  },
	  { 14, KERNEL_KEYPAD_8 	  },
	  {  9, KERNEL_KEYPAD_9 	  },
	  { 18, KERNEL_KEYPAD_AST 	  },
	  {  8, KERNEL_KEYPAD_HASH 	  },
	  { 21, KERNEL_NAVI_KEY_NORTH },
	  { 17, KERNEL_NAVI_KEY_SOUTH },
	  {  6, KERNEL_NAVI_KEY_WEST  },
	  { 12, KERNEL_NAVI_KEY_EAST  },
	  { 19, KERNEL_NAVI_KEY_CENTRAL },
  	  { 23, KERNEL_CONTROL_KEY_0  },	/* Release 			*/	
	  { 24, KERNEL_CONTROL_KEY_8  },	/* Redial 			*/    
	  {  7, KERNEL_CONTROL_KEY_1  },	/* Call Forwarding 	*/   
	  {  4, KERNEL_CONTROL_KEY_4  },	/* Headset 			*/
	  {  5, KERNEL_CONTROL_KEY_5  },	/* Mute 			*/		
	  { 22, KERNEL_MODE_KEY+3 	  },	/* Messages 		*/				
	  {  2, KERNEL_MODE_KEY+4	  },	/* Menu 			*/				
	  { 25, KERNEL_CONTROL_KEY_6  },	/* Minus 			*/		
	  {  3, KERNEL_CONTROL_KEY_3  }, 	/* Handsfree 		*/	
	  { 26, KERNEL_CONTROL_KEY_7  }		/* Plus 			*/		
	},
	
		/** WP_HI **/
	  {
	  { 11, KERNEL_KEYPAD_0       },
      { 21, KERNEL_KEYPAD_1       },
	  { 17, KERNEL_KEYPAD_2       },
  	  {  4, KERNEL_KEYPAD_3       },
	  { 16, KERNEL_KEYPAD_4       },
	  { 14, KERNEL_KEYPAD_5       },
	  {  9, KERNEL_KEYPAD_6       },
	  { 10, KERNEL_KEYPAD_7       },
	  { 20, KERNEL_KEYPAD_8       },
	  {  3, KERNEL_KEYPAD_9       },
	  { 18, KERNEL_KEYPAD_AST     },
	  { 12, KERNEL_KEYPAD_HASH    },
	  {  8, KERNEL_NAVI_KEY_NORTH },
	  {  1, KERNEL_NAVI_KEY_SOUTH },
	  {  7, KERNEL_NAVI_KEY_WEST  },
	  {  2, KERNEL_NAVI_KEY_EAST  },
	  {  0, KERNEL_NAVI_KEY_CENTRAL },
	  { 23, KERNEL_CONTROL_KEY_0  },	/* Release */
	  { 13, KERNEL_CONTROL_KEY_1  },	/* Forward */
	  { 19, KERNEL_CONTROL_KEY_2 },	/* Voice Recognition */
	  {  6, KERNEL_CONTROL_KEY_3  },	/* Handsfree */
	  { 15, KERNEL_CONTROL_KEY_4  }, 	/* Headset */
	  {  5, KERNEL_CONTROL_KEY_5  }	/* Mute */
	},   
		/** WP_LO (T40 Hardware only) - the same as WP_HI 
		    but Key Code 19 drives Menu and not VR        **/
	  {
	  { 11, KERNEL_KEYPAD_0       },
      { 21, KERNEL_KEYPAD_1       },
	  { 17, KERNEL_KEYPAD_2       },
  	  {  4, KERNEL_KEYPAD_3       },
	  { 16, KERNEL_KEYPAD_4       },
	  { 14, KERNEL_KEYPAD_5       },
	  {  9, KERNEL_KEYPAD_6       },
	  { 10, KERNEL_KEYPAD_7       },
	  { 20, KERNEL_KEYPAD_8       },
	  {  3, KERNEL_KEYPAD_9       },
	  { 18, KERNEL_KEYPAD_AST     },
	  { 12, KERNEL_KEYPAD_HASH    },
	  {  8, KERNEL_NAVI_KEY_NORTH },
	  {  1, KERNEL_NAVI_KEY_SOUTH },
	  {  7, KERNEL_NAVI_KEY_WEST  },
	  {  2, KERNEL_NAVI_KEY_EAST  },
	  {  0, KERNEL_NAVI_KEY_CENTRAL },
	  { 23, KERNEL_CONTROL_KEY_0  },	/* Release */
	  { 13, KERNEL_CONTROL_KEY_1  },	/* Forward */
	  {  6, KERNEL_CONTROL_KEY_3  },	/* Handsfree */
	  { 15, KERNEL_CONTROL_KEY_4  }, 	/* Headset */
	  {  5, KERNEL_CONTROL_KEY_5  },	/* Mute */

	  { 25, KERNEL_CONTROL_KEY_6  },	/* Minus */		
	  { 26, KERNEL_CONTROL_KEY_7  },	/* Plus */		
	  { 24, KERNEL_CONTROL_KEY_8  },	/* Redial */    
	  { 22, KERNEL_MODE_KEY+3     },    /* Messages */				
	  { 19, KERNEL_MODE_KEY+4     } 	/* Menu */				
	},	   
      /** WP_LO Workpoint 1 (T28 Hardware only) **/
	  { { 13, KERNEL_KEYPAD_0 	  },
      { 20, KERNEL_KEYPAD_1 	  },
	  { 16, KERNEL_KEYPAD_2 	  },
  	  { 11, KERNEL_KEYPAD_3 	  },
	  {  1, KERNEL_KEYPAD_4 	  },
	  { 15, KERNEL_KEYPAD_5 	  },
	  { 10, KERNEL_KEYPAD_6 	  },
	  {  0, KERNEL_KEYPAD_7 	  },
	  { 14, KERNEL_KEYPAD_8 	  },
	  {  9, KERNEL_KEYPAD_9 	  },
	  { 18, KERNEL_KEYPAD_AST 	  },
	  {  8, KERNEL_KEYPAD_HASH 	  },
	  { 21, KERNEL_NAVI_KEY_NORTH },
	  { 17, KERNEL_NAVI_KEY_SOUTH },
	  {  6, KERNEL_NAVI_KEY_WEST  },
	  { 12, KERNEL_NAVI_KEY_EAST  },
	  { 19, KERNEL_NAVI_KEY_CENTRAL },
  	  { 24, KERNEL_CONTROL_KEY_0  },	/* Release 			*/	
	  { 7,  KERNEL_CONTROL_KEY_8  },	/* Redial 			*/    
	  {  4, KERNEL_CONTROL_KEY_1  },	/* Call Forwarding 	*/   
	  {  5, KERNEL_CONTROL_KEY_5  },	/* Mute 			*/		
	  { 22, KERNEL_MODE_KEY+3 	  },	/* Messages 		*/				
	  {  2, KERNEL_MODE_KEY+4	  },	/* Menu 			*/				
	  { 25, KERNEL_CONTROL_KEY_6  },	/* Minus 			*/		
	  {  3, KERNEL_CONTROL_KEY_3  }, 	/* Handsfree 		*/	
	  { 26, KERNEL_CONTROL_KEY_7  }		/* Plus 			*/		
	},

    // OS15, same as WP_LO (T40), except that OS15 has 8 fpks, which are
    // re-using keycodes from keys not present on OS15, e.g. NAVIKEY_WEST/EAST
    {
        { 11, KERNEL_KEYPAD_0       },
        { 21, KERNEL_KEYPAD_1       },
        { 17, KERNEL_KEYPAD_2       },
        {  4, KERNEL_KEYPAD_3       },
        { 16, KERNEL_KEYPAD_4       },
        { 14, KERNEL_KEYPAD_5       },
        {  9, KERNEL_KEYPAD_6       },
        { 10, KERNEL_KEYPAD_7       },
        { 20, KERNEL_KEYPAD_8       },
        {  3, KERNEL_KEYPAD_9       },
        { 18, KERNEL_KEYPAD_AST     },
        { 12, KERNEL_KEYPAD_HASH    },
        {  8, KERNEL_NAVI_KEY_NORTH },
        {  1, KERNEL_NAVI_KEY_SOUTH },
        {  0, KERNEL_NAVI_KEY_CENTRAL },
        {  6, KERNEL_CONTROL_KEY_3  },	/* Handsfree */
        { 15, KERNEL_FPK_KEY }, 
        { 27, KERNEL_FPK_KEY+1 },        
        { 24, KERNEL_FPK_KEY+2 },
        { 23, KERNEL_FPK_KEY+3 },	
        {  5, KERNEL_FPK_KEY+4 },
        { 13, KERNEL_FPK_KEY+5 },
        {  7, KERNEL_FPK_KEY+6 },
        {  2, KERNEL_FPK_KEY+7 },

        { 25, KERNEL_CONTROL_KEY_6  },	/* Minus */		
        { 26, KERNEL_CONTROL_KEY_7  },	/* Plus */		
        { 22, KERNEL_MODE_KEY+3     },  /* Messages */				
        { 19, KERNEL_MODE_KEY+4     } 	/* Menu */				
    }	   
       
};

static unsigned long keyXlateToHardware( unsigned long keyNumber )
{
	int i;
	for(i=0; i < sizeof(keyXlateTable)/sizeof(keyDevice_mapping_t); i++ )
	{
		if( keyNumber == keyXlateTable[i].physical )
			return( keyXlateTable[i].hardware );
	}
	return(XLATE_ERROR);
}

static unsigned long keyXlateToPhysical( unsigned long keyNumber )
{
	int i;
	for(i=0; i < sizeof(keyXlateTable)/sizeof(keyDevice_mapping_t); i++ )
	{
		if( keyNumber == keyXlateTable[i].hardware )
		{
			// Found kernel key number - now see what the corresponding physical key value is
			if( keyXlateTable[i].physical == 0 )
				return(XLATE_ERROR);				// Physical key is unimplemented
			else
				return(keyXlateTable[i].physical);	// Physical key number OK
		}
	}
	return(XLATE_ERROR);
}


static unsigned long matrixKeyXlateToKernelKeyCode( unsigned long keyNumber )
{
	int i;
	
	/*  wp_type 		== 0 indicates WP_LO 
	 *  wp_type 		== 1 indicates WP_HI 
	 *  wp2_is_T40		== 1 T40 hardware 
	 *  wp              == WP type 1 ,2, 3, 4
	 */
	 int xlateIx=1; // default to WP_HI

	 if(wp == 2 && wp2_is_T40 == 0 ){
	   //WP_LO (T28 WP2 only)
	   xlateIx = 0;							 
	 }									      
	 else if (wp_type == 1)
	 {										  
	   // translation for all WP_HI variants
	   xlateIx = 1;							  
	 }										  
	 else if (wp_type == 0 && wp2_is_T40 == 1)
	 {										     
	     // WP_LO T40 variants (WP1 and WP2)	    
             
             // if this is OS15
             if (wp_version == 4)
             {
                xlateIx = 4;
             }
             else
             {
                xlateIx = 2;
             }
	 }										     
	 else if (wp_type == 0 && wp == 1){		     
	   //WP_LO (T28 WP1)                             
	   xlateIx = 3;
	 }

	for (i=0; i<sizeof(matrixKeyXlateTable[xlateIx])/sizeof(keyDevice_mapping_t); i++)
	{
		if (keyNumber == matrixKeyXlateTable[xlateIx][i].hardware)
		{
			// Found hardware key number - now find the corresponding kernel key vlaue			
			
			//printk(KERN_NOTICE "MatrixKeyXlateToKernelKeyCode: Hardware %d\n", matrixKeyXlateTable[wp_type][i].hardware);
			//printk(KERN_NOTICE "MatrixKeyXlateToKernelKeyCode: Kernel %d\n", matrixKeyXlateTable[wp_type][i].physical);
			return (matrixKeyXlateTable[xlateIx][i].physical); // Key number OK
		}
	}

       
	return (XLATE_ERROR);
}


int keyInputEvent( unsigned long key, int state, unsigned long timestamp )
{
	keyInputDevice_t *dev = &keyInput_Device;

	//printk(KERN_NOTICE "keyInputEvent: key=%d state=%d\n",key,state);

	if( !FIFO_FULL(dev->rqPtr,dev->wqPtr) )
	{
		dev->keyEventQueue[dev->wqPtr].key_code = key;
		dev->keyEventQueue[dev->wqPtr].key_state = state;
		dev->keyEventQueue[dev->wqPtr].time_stamp = timestamp;
		PTR_INCR(dev->wqPtr);
		wake_up_interruptible( &keyInput_wait_queue );
	}
	else
	{
		printk(KERN_NOTICE "keyInputEvent: FIFO full\n");
	}
	return( 0 );
}

int keyInputRegister( keyRegInterface_t *keyThing )
{
	keyThing->keyInputEvent = keyInputEvent;
	if( keyThing->name != 0 )
		printk(KERN_NOTICE "keyInputRegister: %s\n",keyThing->name);
	return( 0 );
}

int keyInputUnRegister( keyRegInterface_t *keyThing )
{
	if( keyThing->name != 0 )
		printk(KERN_NOTICE "keyInputUnRegister: %s\n",keyThing->name);
	return( 0 );
}


int keyReadEvent( keyInputDevice_t *dev, key_event_t_ul *event )
{
	//printk(KERN_NOTICE "keyReadEvent: wr=%d rd=%d\n", dev->wqPtr, dev->rqPtr);

	if( dev->wqPtr != dev->rqPtr )
	{
		event->key_code = dev->keyEventQueue[dev->rqPtr].key_code;
		event->key_state = dev->keyEventQueue[dev->rqPtr].key_state;
		event->time_stamp = dev->keyEventQueue[dev->rqPtr].time_stamp;
		PTR_INCR(dev->rqPtr);
		return( sizeof( key_event_t_ul ) );
	}
	return( 0 );
}

int three_key_combo_pressed(void)
{
	int ret_val = -1;

	if(	keypad_current_state[1] == KEY_PRESSED &&
		keypad_current_state[4] == KEY_PRESSED &&
		keypad_current_state[7] == KEY_PRESSED)
	{
		ret_val = KERNEL_HOODED_CLAW_147;
		// printk(KERN_NOTICE "Hooded Claw Combination 1-4-7\n");
	}
	else if(keypad_current_state[3] == KEY_PRESSED &&
		keypad_current_state[5] == KEY_PRESSED &&
		keypad_current_state[11] == KEY_PRESSED)
	{
		ret_val = KERNEL_HOODED_CLAW_35HASH;
		// printk(KERN_NOTICE "Hooded Claw Combination 3-5-#\n");
	}
	else if(keypad_current_state[2] == KEY_PRESSED &&
		keypad_current_state[8] == KEY_PRESSED &&
		keypad_current_state[9] == KEY_PRESSED)
	{
		ret_val = KERNEL_HOODED_CLAW_289;
		// printk(KERN_NOTICE "Hooded Claw Combination 2-8-9\n");
	}
	else if(keypad_current_state[0] == KEY_PRESSED &&
		keypad_current_state[1] == KEY_PRESSED &&
		keypad_current_state[3] == KEY_PRESSED)
	{
		ret_val = KERNEL_HOODED_CLAW_013;
		// printk(KERN_NOTICE "Hooded Claw Combination 0-1-3\n");
	}

	return(ret_val);
}

int number_of_keys_pressed(int* index_ptr)
{
	int i;
	int key_count = 0;

	for(i=0;i<number_of_keypad_keys; i++)
	{
		if (keypad_current_state[i] == KEY_PRESSED)
		{
			key_count++;

			if(index_ptr != NULL)
			{
				*index_ptr = i;
			}
		}
	}

	return(key_count);
}

int keyMatrixInputEvent( char key, int state, unsigned long timestamp )
{
	enum keypadPressState {
		key_released, 			// Key released (idle state)
		keypress_not_sent, 		// Key pressed but key press event not yet sent
		keypress_sent 			// Key pressed, keypress event sent
	};

	int key_index = 0;
	int translatedKey; 			// Used to hold the kernel key code of the key

	int keys_pressed_previously;
	int keys_pressed_now;

	int single_key_previously;
	int single_key_now;
	int three_key_combo_previously;
	int three_key_combo_now;

	//printk(KERN_NOTICE "\nKeyInput: key %d, state %d\n", (int)key, state);

	// Translate key from hardware value to kernel key code as hardware value
	// is different between WP_LO and WP_HI
	translatedKey = matrixKeyXlateToKernelKeyCode(key);

	//printk(KERN_NOTICE "KeyInput: translated key %d\n", translatedKey);

	// The resulting value of i can be used as the index for the keypad_current_state array
	while (key_index<number_of_keypad_keys+1 && keypadMappingTable[key_index].kernelKeyCode != translatedKey)
	{
		key_index++;
	}

	if (key_index == number_of_keypad_keys+1)
	{
		// a key other than 0-9, * or # has been pressed
		return (keyInputEvent((unsigned long)translatedKey, state, timestamp));
	}

	keys_pressed_previously = number_of_keys_pressed(&single_key_previously);
	three_key_combo_previously = three_key_combo_pressed();

	keypad_current_state[key_index] = state;

	keys_pressed_now = number_of_keys_pressed(&single_key_now);
	three_key_combo_now = three_key_combo_pressed();


	//look for single key pressed transitions

	if(	(keys_pressed_previously != 1)	&&
		(keys_pressed_now == 1)	)
	{
		return (keyInputEvent((unsigned long)keypadMappingTable[single_key_now].kernelKeyCode, KEY_PRESSED, timestamp));
	}

	if(	(keys_pressed_previously == 1)	&&
		(keys_pressed_now != 1)	)
	{
		return (keyInputEvent((unsigned long)keypadMappingTable[single_key_previously].kernelKeyCode, KEY_RELEASED, timestamp));
	}


	//look for three keys pressed transitions

	if(	(keys_pressed_previously != 3)	&&
		(keys_pressed_now == 3)	)
	{
		if(three_key_combo_now != -1)
		{
			return (keyInputEvent((unsigned long)three_key_combo_now, KEY_PRESSED, timestamp));
		}
	}

	if(	(keys_pressed_previously == 3)	&&
		(keys_pressed_now != 3)	)
	{
		if(three_key_combo_previously != -1)
		{
			return (keyInputEvent((unsigned long)three_key_combo_previously, KEY_RELEASED, timestamp));
		}
	}

	return(0);
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
/**
* Declaration of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/
struct file_operations keyInput_fops = {
	write:		keyInput_write,	/* write */
	read:		keyInput_read,	/* read */
	ioctl:		keyInput_ioctl, /* ioctl */
	poll:		keyInput_poll,  /* poll */
	open:		keyInput_open,  /* open */
	release:	keyInput_close, /* release */
};

/**
* Aliased read function called from user space
* Waits for the device to put data in the key data queue
*\param *file pointer to a kernel file data structure
*\param *buf pointer to a user space data buffer
*\param size of the user space data buffer
*\param *off pointer to file position
*\returns either number of data items transferred or error state
*/
ssize_t keyInput_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	keyInputDevice_t* dev = (keyInputDevice_t*)fp->private_data;
	ssize_t ret;
	key_event_t_ul event;
	unsigned long newKeyCode = 0;

	if( size < sizeof(key_event_t_ul))
	{
		printk(KERN_ERR "keyReadEvent: bad size %d\n", size );
		return( 0 );
	}

	while((dev->rqPtr) == (dev->wqPtr))
	{/* nothing to read */

        if (fp->f_flags & O_NONBLOCK)
        {
			//printk("Non blocking read");

            return -EAGAIN;
		}

        if(wait_event_interruptible(keyInput_wait_queue, (dev->rqPtr) != (dev->wqPtr)))
        {
			printk(KERN_ERR "keyReadEvent: restart sys\n" );
            return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
        }
	}

	ret = 0;
//	printk(KERN_NOTICE "keyInput_read: Reading event size=%d\n",size);

	while( (size >= sizeof(key_event_t_ul)) && (keyReadEvent( dev, &event ) != 0) )
	{
		// Copy and reformat buffer before passing it upwards to user
		// (We have a keypad key now we have to translate it).

		newKeyCode = keyXlateToPhysical( event.key_code );
		if( newKeyCode == XLATE_ERROR)
		{
			// Don't know what the event was, so pass it on without translating it
			newKeyCode = event.key_code;

			// Debug messages
			if (event.key_state == KEY_RELEASED)
			{
//				printk("keyinput::keyInput_read unknown key release 0x%02lx (hardware). Not translated.\n", event.key_code );
			}
			else if (event.key_state == KEY_PRESSED)
			{
//				printk("keyinput::keyInput_read unknown keypress 0x%02lx (hardware). Not translated.\n", event.key_code );
			}
		}
		else	// Key translated OK
		{
			// Debug messages
			if (event.key_state == KEY_RELEASED)
			{
//				printk("keyinput::keyInput_read key release 0x%02lx (hardware) translated to 0x%02lx (physical)\n", event.key_code, newKeyCode );
			}
			else if (event.key_state == KEY_PRESSED)
			{
//				printk("keyinput::keyInput_read keypress 0x%02lx (hardware) translated to 0x%02lx (physical)\n", event.key_code, newKeyCode );
			}
			else
			{
				printk("keyinput::keyInput_read KEY_????? event..\n");
			}
		}

		// Pass the keypress (or key release) up to the user

		event.key_code = newKeyCode;
//		printk(KERN_NOTICE "keyInput_read: event\n");
		if(copy_to_user(buf, &event, sizeof( key_event_t_ul) ))
		{
			printk(KERN_ERR "keyReadEvent: copy_to_user failed\n");
			return -ERESTARTSYS;
		}
		ret += sizeof( key_event_t_ul );
		buf += sizeof( key_event_t_ul );
		size -= sizeof( key_event_t_ul );
	}

	return( ret );
}

// Function to handle writes from user space
// Buffers written will appear as keypresses - this is used for testing.
ssize_t keyInput_write(struct file *fp, const char *buf, size_t len, loff_t * off)
{
	ssize_t	numbytes;
	key_event_t_ul event;
	unsigned long newKeyCode;
//	keyInputDevice_t* dev = (keyInputDevice_t*)fp->private_data;

//	printk(KERN_ERR "keyInput_write: len=%d\n", len);

	if (len != sizeof(key_event_t_ul))
	{
		printk(KERN_WARNING "keyInput_write: illegal size %d\n", len);
		return 0;
	}

	if (copy_from_user(&event, buf, sizeof(key_event_t_ul)))
	{
		return( -EFAULT );
	}

	// Get hardware key value from physical key number
	newKeyCode = keyXlateToHardware(event.key_code);

	if( newKeyCode == XLATE_ERROR )
	{
		// Don't know what the event was, so pass it on without translating it
		newKeyCode = event.key_code;

//		printk("keyinput::keyInput_write: unsupported key number 0x%02lx\n",event.key_code );
//		return( -EFAULT );
	}

	// Physical->Hardware translation worked OK - continue
//	printk("keyinput::keyInput_write: physical key number 0x%02lx is hardware 0x%02lx\n",event.key_code, newKeyCode );
	event.key_code = newKeyCode;

	numbytes = 0;
	if( keyInputEvent( event.key_code, event.key_state, event.time_stamp ) == 0 )
	{
		return( sizeof(key_event_t_ul) );
	}

	return numbytes;
}

/**
 * Aliased ioctl function called from user space
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *filp pointer to a kernel file data structure
 *\param cmd ioctl command
 *\param arg associated command arguments
 * Currently not used
 *\returns 0
 */
int keyInput_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err;
	printk(KERN_ERR "keyInput_ioctl: cmd=%08x\n", cmd);
	err = 0;
	return err;
}

/**
* Aliased poll function called from user space by poll or select
* Currently used to wait on information being availble to be read from incoming data queue buffer.
* Woken by the keyInputDrdyEvent function.
*\param *fp pointer to a kernel file data structure
*\param *pts pointer to a kernel poll table data structure
*\returns flags to indicate valid data available
*/
unsigned int keyInput_poll(struct file *fp, struct poll_table_struct *pts)
{
	unsigned int mask = 0;
	keyInputDevice_t* dev = (keyInputDevice_t*)fp->private_data;

	/*
	 * The buffer is circular.
	 */

	poll_wait(fp, &keyInput_wait_queue, pts);
	if ((dev->rqPtr) != (dev->wqPtr)) mask |= POLLIN | POLLRDNORM;  /* readable */

    return mask;
}

/**
* Aliased open function called from user space by open
*\param *inode pointer to the inode structure for the device this contains such information such
*as the major and minor numbers.
*\param *fp pointer to a kernel file data structure
*\returns 0 (in turn application returns a file descriptor)
*/
int keyInput_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	printk(KERN_NOTICE "keyInput_open: major = %d minor = %d\n", omajor, ominor);

	fp->private_data = &keyInput_Device;

	return 0;
}
/**
* Aliased close function called from user space by close
*\param *inode pointer to the inode structure for the device this contains such information such
*as the major and minor numbers.
*\param *fp pointer to a kernel file data structure
*\returns 0
*/
int keyInput_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "keyInput_close:\n");
	err = 0;
	return err;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/
static struct proc_dir_entry *keyInput_proc;
/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/keyInput'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/
static int keyInput_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len = 0;
	const char info[] = "keyInput Device version 1.0\n";
	printk(KERN_INFO "/proc/drivers/keyInput %d\n",count);

	len = sizeof(info);

	if( count < sizeof(info))
	len = count;
	memcpy(buf,info,len);

	return len;
}
/**
* Function to handle a proc entry write not often used
*/
static int keyInput_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	char buff[50];
	int len;

	unsigned long key_code;
	int key_state;
	unsigned long time_stamp = jiffies;

	len = ( count < sizeof(buff) )? count : sizeof(buff);

	memcpy(buff, buffer,len);
	buff[len-1] = 0;

	sscanf(buff, "c=%lu s=%d", &key_code, &key_state);
	keyInputEvent(key_code, key_state, time_stamp);

	printk(KERN_INFO "key input write proc: c=%lu s=%d (t=%lu)\n", key_code, key_state, time_stamp);
	return count;
}

/**
* Kernel initialisation function run when the module is installed via insmod
*/
static int __init keyInput_init(void)
{
	int res = 0;

	if( (input_devfs_handle = devfs_mk_dir(NULL, "input", NULL)) == 0 )
		return -EBUSY;

	devfs_register(input_devfs_handle, "keyInput",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&keyInput_fops,0);


	if (major == 0) major = res;

	SET_MODULE_OWNER(&keyInput_fops);

	printk(KERN_NOTICE "\nkeyInput device registered.\n");

	/* create PROC entry */
	if( (keyInput_proc = create_proc_entry( "driver/keyInput", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/keyInput!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}
	keyInput_proc->read_proc  = keyInput_read_procmem;
	keyInput_proc->write_proc = keyInput_write_procmem;

	printk(KERN_NOTICE "/proc/driver/keyInput created!\n");

	spin_lock_init(&keyInput_Device.lock);



	keyInput_Device.wqPtr = 0;
	keyInput_Device.rqPtr = 0;

	/* Register the key matrix input handler with infineon */
	if( ifx_key_register_callback( keyMatrixInputEvent ) )
	{
		printk(KERN_NOTICE "incaip2_key_register_callback (key matrix) failed:\n");
	}


	/* Determine the WP type */
	wp = get_WP_type();
	
	if ( wp == 1 || wp == 2 )
	{
		wp_type = WP_LO;	
		printk(KERN_NOTICE "keyInput: WP_LO\n");
	} 
	else if ( wp == 3 || wp == 4 )
	{
		wp_type = WP_HI;
		printk(KERN_NOTICE "keyInput: WP_HI\n");
	}
	else // Unknown Version - set to WP_HI as a default
	{
		wp_type = WP_HI;
		printk(KERN_NOTICE "WP type %d unknown. Defaulting to WP_HI.\n", wp);
	}
	
        wp2_is_T40 = (incaip2_get_cpu_ver() < 3) ? 0 : 1;

        /* Determine WP version */
        wp_version = get_WP_version();
	
	/* Initalise the array for holding the current state of the keypad keys */
	int i;
	for (i=0; i<number_of_keypad_keys; i++) {
		keypad_current_state[i] = KEY_RELEASED;
	}

	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/
static void __exit keyInput_exit(void)
{

	if( ifx_key_unregister_callback() != 0 )
	{
		printk(KERN_NOTICE "incaip2_key_unregister_callback failed:\n");
	}

	remove_proc_entry("driver/keyInput", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_NOTICE "keyInput module exit\n");
}


module_init(keyInput_init);
module_exit(keyInput_exit);
EXPORT_SYMBOL(keyInputRegister);
EXPORT_SYMBOL(keyInputUnRegister);

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
