/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <asm/incaip2/tsf.h>
#include <asm/incaip2/mps.h>
#include <linux/tqueue.h> /* task queues */
#include "mps_event.h"

void ifx_mps_bufman_free(const void* ptr);
static int default_headset_callback (int state);

static mps_headset_callback_t headset_callback = default_headset_callback;

static MPS_Voice_Event_Count_t		mps_voice_evcount[4];
static MPS_Command_Event_Count_t	mps_cmd_evcount;

static u32	cmd_error_code[3];

struct tq_struct mps_job; /* global: initialized to zero */

// SYSTEM COMMANDS

static const u32 ifxDefaultSystemCmdErrorAckMsg[] =
{
	0x0600E000		// C1d
};
static const u32 ifxDefaultSystemDataErrorAckMsg[] =
{
	0x0600E500		// C1d
};

static const u32 ifxDefaultSystemStatusErrorAckMsg[] =
{
	0x0600E104,		// C1d
	0x00400000		// C1d
};

static const u32 ifxDefaultSystemFirmwareVersionMsg[] =
{
	0x8600E604		// C1d
};

static const u32 ifxDefaultSystemGetLastCmdErrorMsg[] =
{
	0x8600E408		// C1d NOTE: Spec is WRONG about command number!
};

#define NUM_ERROR_TEXTS 39

char *error_short_text[NUM_ERROR_TEXTS] = {
	"0 CMD_RC_OK",
	"1 CMD_ERR_CTRL_BAD_BC",
	"2 CMD_ERR_CTRL_BAD_CH",
	"3 CMD_ERR_CTRL_BAD_LEN",
	"4 CMD_ERR_CTRL_BC_ON_R",
	"5 CMD_ERR_CTRL_CH_OFF",
	"6 CMD_ERR_CTRL_BAD_RES_NUM",
	"7 CMD_ERR_CTRL_RES_USED",
	"8 CMD_ERR_CTRL_BIT_CHANGE",
	"9 CMD_ERR_CTRL_BAD_RES_NUM_DEACT",
	"10 CMD_ERR_CTRL_ALG_ON",
	"11 CMD_ERR_CTRL_BAD_VALUE",
	"12 CMD_ERR_CTRL_BAD_TASK",
	"13 CMD_ERR_CTRL_BAD_RW",
	"16 CMD_ERR_RES_BAD_BC",
	"17 CMD_ERR_RES_BAD_RES_NUM",
	"18 CMD_ERR_RES_BAD_RES_LEN",
	"19 CMD_ERR_RES_BC_ON_R",
	"20 CMD_ERR_RES_BAD_WRITE",
	"21 CMD_ERR_RES_BAD_VALUE",
	"22 CMD_ERR_RES_BAD_RW",
	"33 CMD_ERR_NOT_SUPPORTED",
	"34 CMD_ERR_SC_ON",
	"35 CMD_ERR_BAD_CMD",
	"36 CMD_ERR_BAD_MOD",
	"37 CMD_ERR_BAD_ECMD",
	"38 CMD_ERR_BAD_LEN"
};

char *error_text[NUM_ERROR_TEXTS] = {
	"No command error",
	"Broadcast bit value is not supported by the command at all (neither for read nor for write).",
	"Wrong channel number",
	"Wrong command length",
	"BC bit set in read command",
	"Unable to successfully process command while channel is not active.",
	"Wrong resource number",
	"Resource already in use",
	"Change of a bit or bit field, which must not be changed while the module is active",
	"Attempt to deactivate a resource with a different resource number than the one used for activation",
	"Attempt to deactivate a channel with active algorithms",
	"Field values or field value combinations are not correct.",
	"Unable to create (in case of activation) or remove (in case of deactivation) OS task.",
	"The read or write access is not supported by this command.",
	"Broadcast bit value is not supported by resource command at all (neither for read nor for write).",
	"Wrong resource number in resource command.",
	"Wrong command length in resource command",
	"Broadcast bit is set in a read command",
	"Not allowed attempt to write to a coefficient command while the algorithm is active",
	"Field values or field value combinations are not correct.",
	"The read or write access is not supported by this resource command",
	"The feature requested is not available in this version of the firmware",
	"bit is set in the command while it is not allowed.",
	"The CMD bits not correct",
	"The MOD (Module) command field value is incorrect",
	"The ECMD (EDSP Command) command field value is incorrect.",
	"The length of the command is greater than mailbox size"
};

static char *mps_short_text( u32 code )
{
	if( code < NUM_ERROR_TEXTS)
		return( error_short_text[code] );
	return( "Unknown" );
}

static char *mps_main_text( u32 code )
{
	if( code < NUM_ERROR_TEXTS)
		return( error_text[code] );
	return( "None" );
}

static int mps_schedule_job( void (*fn)(void *ptr), void *arg )
{
	mps_job.routine = fn; /*scheduled function*/
	mps_job.data = &arg; /*argument to function*/
	schedule_task(&mps_job);
	return( 0 );
}


static int default_headset_callback (int state)
{
	if( state == MPS_HEADSET_DISCONNECTED )
	{
		printk("Headset unplugged\n");
	}
	else
	{
		printk("Headset plugged in\n");
	}

	return( 0 );
}

static int mps_event_send_cmd( const u32 *cmd, int len )
{
    mps_message msg;
    
    msg.nDataBytes = len;
    msg.pData = (u8 *)cmd;
    msg.RTP_PaylOffset = 0;
    msg.cmd_type       = CMD_EOP_PACKET;

    return( ifx_mps_write_mailbox( command, &msg ) );
}

static int mps_event_get_cmd( mps_message *msg )
{
	int retval;
	char *dest;
	char *src;
	int	bufsize;
	
	bufsize = msg->nDataBytes;
	dest = msg->pData;
	
    while( (retval=ifx_mps_read_mailbox( command, msg )) == -ENODATA )
    {
    	schedule();
    }

	if( retval == 0 )
	{
		if( msg->nDataBytes > bufsize )
		{
			printk( KERN_ERR "(%s) bad size %d > %d\n",__FUNCTION__,msg->nDataBytes , bufsize);
			msg->nDataBytes = bufsize;
		}
		src = msg->pData;
		memcpy(dest,src,msg->nDataBytes);
		msg->pData = dest;
		
		ifx_mps_bufman_free( src );
	}
	return( retval );
}


static int mps_event_handle_cmd_error( void )
{
	int retval;
	if( (retval = mps_event_send_cmd( ifxDefaultSystemCmdErrorAckMsg, sizeof(ifxDefaultSystemCmdErrorAckMsg))) != 0 )
	{
		printk(KERN_ERR "(%s) failed to clear commnad error %d\n",__FUNCTION__, retval);
	}
	else
	{
//		printk(KERN_NOTICE "(%s) OK\n",__FUNCTION__);
	}
	return( retval );
}

static int mps_event_get_cmd_error( void )
{
	int retval;
	if( (retval = mps_event_send_cmd( ifxDefaultSystemGetLastCmdErrorMsg, sizeof(ifxDefaultSystemGetLastCmdErrorMsg))) != 0 )
	{
		printk(KERN_ERR "(%s) failed to clear commnad error %d\n",__FUNCTION__, retval);
	}
	else
	{
//		printk(KERN_NOTICE "(%s) OK\n",__FUNCTION__);
	}
	return( retval );
}

void mps_retrieve_last_error( void *p )
{
    mps_message msg;
//    u32 buffer[3];
	
	msg.pData          = (u8*)cmd_error_code;
	msg.nDataBytes     = sizeof(cmd_error_code);
	msg.RTP_PaylOffset = 0;
	msg.cmd_type       = CMD_EOP_PACKET; //should get this from data not hardcoded

	ifx_mps_open((struct inode *)command, 0);
	mps_event_handle_cmd_error( );
	mps_event_get_cmd_error();
	
	if( mps_event_get_cmd( &msg ) == 0 )
	{
//		printk( KERN_ERR "\nMPS_CMD Error = %08x\n",cmd_error_code[0]);
//		printk( KERN_ERR "Error code    = %08x %s\n",cmd_error_code[1],mps_short_text(cmd_error_code[1]));
//		printk( KERN_ERR "Faulting CMD  = %08x\n",cmd_error_code[2]);
//		printk( KERN_ERR "Notes :- %s\n",mps_main_text(cmd_error_code[1]));
	}
	ifx_mps_close((struct inode *)command, 0);
}

void mps_get_last_error( void )
{
	mps_schedule_job( mps_retrieve_last_error, 0 );
}



static void ifxAudioEventEventCallback( MbxEventRegs_s* events )
{
    MPS_Ad0Reg_u MPS_Ad0StatusReg;
    MPS_Ad1Reg_u MPS_Ad1StatusReg;

	MPS_Ad0StatusReg.val = events->MPS_Ad0Reg.val;
	MPS_Ad1StatusReg.val = events->MPS_Ad1Reg.val;
	
//	printk(KERN_NOTICE "mps event command channel = MPS_Ad0Reg=%08x MPS_Ad1Reg=%08x\n",MPS_Ad0StatusReg.val,MPS_Ad1StatusReg.val);
	if( MPS_Ad0StatusReg.fld.data_err )
	{
		printk("\noops......MPS data_err\n");
		mps_cmd_evcount.data_err++;
	}
	if( MPS_Ad0StatusReg.fld.cmd_err )
	{
//		printk("\noops......MPS cmd_err\n");
		mps_cmd_evcount.cmd_err++;
//		mps_get_last_error();
//		events->MPS_Ad0Reg.fld.cmd_err = 0;
	}
	if( MPS_Ad0StatusReg.fld.pcm_crash )
	{
		printk("\noops......MPS pcm_crash\n");
		mps_cmd_evcount.pcm_crash++;
	}
	if( MPS_Ad0StatusReg.fld.mips_ol )
	{
		printk("\noops......MPS mips_ol\n");
		mps_cmd_evcount.mips_ol++;
	}

	if( MPS_Ad0StatusReg.fld.rcv_ov )
	{
		mps_cmd_evcount.rcv_ov++;
//		mps_event_send_cmd( ifxDefaultSystemStatusErrorAckMsg, sizeof(ifxDefaultSystemStatusErrorAckMsg));
		printk("\nOOPS......MPS rcv_ov\n");
	}

	if( MPS_Ad1StatusReg.fld.hs_con == 1)
	{
		// Headset connected event
		// Pass it to the keyboard input handler
		events->MPS_Ad1Reg.fld.hs_con = 0;
		mps_cmd_evcount.hs_con++;
		headset_callback( MPS_HEADSET_CONNECTED );
//		printk(KERN_NOTICE "mps reports Headset connected event\n");
	}

	if( MPS_Ad1StatusReg.fld.hs_dis == 1)
	{
		// Headset disconnected event
		// Pass it to the keyboard input handler
		events->MPS_Ad1Reg.fld.hs_dis = 0;
		mps_cmd_evcount.hs_dis++;
		headset_callback( MPS_HEADSET_DISCONNECTED );
//		printk(KERN_NOTICE "mps reports Headset disconnected event\n");
	}

	if( MPS_Ad1StatusReg.fld.utga == 1)
	{
//		printk("\noops......MPS utga\n");
		mps_cmd_evcount.utga++;
	}

	if( MPS_Ad1StatusReg.fld.sr_err == 1)
	{
		printk("\noops......MPS sr_err\n");
		mps_cmd_evcount.sr_err++;
	}

	if( MPS_Ad1StatusReg.fld.sr_ivw == 1)
	{
		printk("\noops......MPS sr_ivw\n");
		mps_cmd_evcount.sr_ivw++;
	}

	if( MPS_Ad1StatusReg.fld.sr_req == 1)
	{
		printk("\noops......MPS sr_req\n");
		mps_cmd_evcount.sr_req++;
	}
}



/**
* Callback function called by MPS for headset events (i.e. plugged in, unplugged).
*/
static void ifxVoiceEventEventCallback( int chan, MbxEventRegs_s* events )
{

    MPS_VCStatReg_u MPS_VCStatusReg;

    MPS_VCStatusReg.val       = events->MPS_VCStatReg[chan].val;
    
	if( MPS_VCStatusReg.fld.fdp_err )
	{
		mps_voice_evcount[chan].fdp_err++;
		printk("\nEvent......MPS fax data pump error chan=%d\n", chan);
	}

	if( MPS_VCStatusReg.fld.vpou_jbl )
	{
		mps_voice_evcount[chan].vpou_jbl++;
		printk("\nEvent......MPS recieve voice vpou_jbl chan=%d\n", chan);
	}

	if( MPS_VCStatusReg.fld.vpou_jbh )
	{
		mps_voice_evcount[chan].vpou_jbh++;
		printk("\nEvent......MPS recieve voice vpou_jbh chan=%d\n", chan);
	}

	if( MPS_VCStatusReg.fld.epq_st )
	{
		mps_voice_evcount[chan].epq_st++;
		printk("\nEvent......MPS recieve voice epq_st chan=%d\n", chan);
	}

	if( MPS_VCStatusReg.fld.vpou_st )
	{
		mps_voice_evcount[chan].vpou_st++;
		printk("\nEvent......MPS recieve voice vpou_st chan=%d\n", chan);
	}

	if( MPS_VCStatusReg.fld.lin_req )
	{
		mps_voice_evcount[chan].lin_req++;
		printk("\nEvent......MPS recieve voice lin_req chan=%d\n", chan);
	}

	if( MPS_VCStatusReg.fld.dec_chg )
	{
		mps_voice_evcount[chan].dec_chg++;
		printk("\nEvent......MPS recieve voice dec_chg chan=%d\n", chan);
	}
}

static void ifxVoice0EventEventCallback( MbxEventRegs_s* events )
{
//	printk(KERN_NOTICE "mps event voice channel 0\n");
	ifxVoiceEventEventCallback( 0, events );
}

static void ifxVoice1EventEventCallback( MbxEventRegs_s* events )
{
//	printk(KERN_NOTICE "mps event voice channel 1\n");
	ifxVoiceEventEventCallback( 1, events );
}

static void ifxVoice2EventEventCallback( MbxEventRegs_s* events )
{
//	printk(KERN_NOTICE "mps event voice channel 2\n");
	ifxVoiceEventEventCallback( 2, events );
}

static void ifxVoice3EventEventCallback( MbxEventRegs_s* events )
{
//	printk(KERN_NOTICE "mps event voice channel 3\n");
	ifxVoiceEventEventCallback( 3, events );
}

s32 mps_register_headset_callback( mps_headset_callback_t callback )
{
	headset_callback = callback;
	return( 0 );
}

s32 mps_unregister_headset_callback(void)
{
	headset_callback = default_headset_callback;
	return( 0 );
}



/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *ifxAudioEvent_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/Headset'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/
static int ifxAudioEvent_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int len, i;

    len = sprintf(buf, "MPS Event Stats\n");

	len += sprintf(buf+len,"cmd_err   = %d\n",mps_cmd_evcount.cmd_err );
	len += sprintf(buf+len,"mips_ol   = %d\n",mps_cmd_evcount.mips_ol );
	len += sprintf(buf+len,"pcm_crash = %d\n",mps_cmd_evcount.pcm_crash );
	len += sprintf(buf+len,"data_err  = %d\n",mps_cmd_evcount.data_err );
	len += sprintf(buf+len,"hs_con    = %d\n",mps_cmd_evcount.hs_con );
	len += sprintf(buf+len,"hs_dis    = %d\n",mps_cmd_evcount.hs_dis );
	len += sprintf(buf+len,"utga      = %d\n",mps_cmd_evcount.utga );
	len += sprintf(buf+len,"sr_err    = %d\n",mps_cmd_evcount.sr_err );
	len += sprintf(buf+len,"sr_ivw    = %d\n",mps_cmd_evcount.sr_ivw );
	len += sprintf(buf+len,"sr_req    = %d\n",mps_cmd_evcount.sr_req );
	len += sprintf(buf+len,"rcv_ov    = %d\n",mps_cmd_evcount.rcv_ov );
	
	for( i=0 ; i < 4; i++ )
	{
		len += sprintf(buf+len,"\nVoice %d\n", i );
		len += sprintf(buf+len,"lin_req  = %d\n",mps_voice_evcount[i].lin_req );
		len += sprintf(buf+len,"epq_st   = %d\n",mps_voice_evcount[i].epq_st );
		len += sprintf(buf+len,"vpou_st  = %d\n",mps_voice_evcount[i].vpou_st );
		len += sprintf(buf+len,"vpou_jbl = %d\n",mps_voice_evcount[i].vpou_jbl );
		len += sprintf(buf+len,"vpou_jbh = %d\n",mps_voice_evcount[i].vpou_jbh );
		len += sprintf(buf+len,"dec_chg  = %d\n",mps_voice_evcount[i].dec_chg );
		len += sprintf(buf+len,"fdp_err  = %d\n",mps_voice_evcount[i].fdp_err );
	}

	len += sprintf(buf+len,"\nMPS_CMD Error = %08x\n",cmd_error_code[0]);
	len += sprintf(buf+len,"Error code    = %08x %s\n",cmd_error_code[1],mps_short_text(cmd_error_code[1]));
	len += sprintf(buf+len,"Faulting CMD  = %08x\n",cmd_error_code[2]);
	len += sprintf(buf+len,"Notes :- %s\n",mps_main_text(cmd_error_code[1]));

	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int ifxAudioEvent_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "ifxAudioEvent Write procmem\n");
	return count;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>> Init/Shutdown Section <<<<<<<<<<<<<<<<<<<<<<<<<<<*/

// This function carries out the required actions when the headset
// device is installed, either on startup, or via insmod. Currently,
// this means registering the required callback functions.

static void ifxAudioEvent_startup(void)
{

	MbxEventRegs_s mask;
    MPS_VCStatReg_u MPS_VCStatusReg;
	/***************************/
	// Register for headset event callback
	// (We register with mps_linux.c and supply a callback function ourselves.
	// mps calls our function when it has an event to pass to us.)

	// Set up the mask:
	// zero everything ...
	mask.MPS_Ad0Reg.val = 0;
	mask.MPS_Ad1Reg.val = 0;
	mask.MPS_VCStatReg[0].val = 0;
	mask.MPS_VCStatReg[1].val = 0;
	mask.MPS_VCStatReg[2].val = 0;
	mask.MPS_VCStatReg[3].val = 0;

	// ... except the required fields
	mask.MPS_Ad0Reg.fld.data_err = 1;	// See document "INCA-IP2 Firmware Description",
	mask.MPS_Ad0Reg.fld.cmd_err = 1;	//  section 2.4.1 "Status Register Description"
	mask.MPS_Ad0Reg.fld.pcm_crash = 1;	//  section 2.4.1 "Status Register Description"
	mask.MPS_Ad0Reg.fld.mips_ol = 1;	//  section 2.4.1 "Status Register Description"
	mask.MPS_Ad0Reg.fld.rcv_ov = 1;		//  7.x firmware onwards 

	// ... except the required fields
	mask.MPS_Ad1Reg.fld.hs_con = 1;	// See document "INCA-IP2 Firmware Description",
	mask.MPS_Ad1Reg.fld.hs_dis = 1;	//  section 2.4.1 "Status Register Description"

	//temporary hack: only enable interrupts for headset...
	//once proper headset detection mechinism is implemented
	//(headset module, gpal audio, and dvd) the other events can be commented back in.
//	mask.MPS_Ad1Reg.fld.utga = 1;
//	mask.MPS_Ad1Reg.fld.sr_err = 1;
//	mask.MPS_Ad1Reg.fld.sr_ivw = 1;
//	mask.MPS_Ad1Reg.fld.sr_req = 1;


	ifx_mps_register_event_callback( command, &mask, ifxAudioEventEventCallback );
	ifx_mps_event_activation(command, &mask);

	/***************************/
	mask.MPS_Ad0Reg.val = 0;
	mask.MPS_Ad1Reg.val = 0;

	MPS_VCStatusReg.val = 0;
    MPS_VCStatusReg.fld.fdp_err = 1;
    MPS_VCStatusReg.fld.lin_req = 1;
    MPS_VCStatusReg.fld.dec_chg = 1;
    MPS_VCStatusReg.fld.epq_st = 1;
    MPS_VCStatusReg.fld.vpou_jbl = 1;
    MPS_VCStatusReg.fld.vpou_jbh = 1;
    MPS_VCStatusReg.fld.vpou_st = 1;

	mask.MPS_VCStatReg[0].val = MPS_VCStatusReg.val;
	ifx_mps_register_event_callback( voice0, &mask, ifxVoice0EventEventCallback );
	ifx_mps_event_activation(voice0, &mask);
	mask.MPS_VCStatReg[0].val = 0;

	mask.MPS_VCStatReg[1].val = MPS_VCStatusReg.val;
	ifx_mps_register_event_callback( voice1, &mask, ifxVoice1EventEventCallback );
	ifx_mps_event_activation(voice1, &mask);
	mask.MPS_VCStatReg[1].val = 0;

	mask.MPS_VCStatReg[2].val = MPS_VCStatusReg.val;
	ifx_mps_register_event_callback( voice2, &mask, ifxVoice2EventEventCallback );
	ifx_mps_event_activation(voice2, &mask);
	mask.MPS_VCStatReg[0].val = 0;

	mask.MPS_VCStatReg[3].val = MPS_VCStatusReg.val;
	ifx_mps_register_event_callback( voice3, &mask, ifxVoice3EventEventCallback );
	ifx_mps_event_activation(voice3, &mask);
	mask.MPS_VCStatReg[3].val = 0;

}


// This function carries out the required actions when the headset
// device is uninstalled, either on shutdown, or via rmmod. Currently,
// this means unregistering the required callback functions.

static void ifxAudioEvent_shutdown(void)
{
	// Unregister for headset event callback
	ifx_mps_unregister_event_callback( command );
	ifx_mps_unregister_event_callback( voice0 );
	ifx_mps_unregister_event_callback( voice1 );
	ifx_mps_unregister_event_callback( voice2 );
	ifx_mps_unregister_event_callback( voice3 );
}


/**
* Kernel initialisation function run when the module is installed via insmod
*/
static int __init ifxAudioEvent_init(void)
{

	/* create PROC entry */
	if( (ifxAudioEvent_proc = create_proc_entry( "driver/mps_event", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/mps_event!\n");
		return( -ENOENT );
	}

	printk(KERN_ERR "mps_event handler installed\n");
	ifxAudioEvent_proc->read_proc  = ifxAudioEvent_read_procmem;
	ifxAudioEvent_proc->write_proc = ifxAudioEvent_write_procmem;

	ifxAudioEvent_startup();

	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/
static void __exit ifxAudioEvent_exit(void)
{
	ifxAudioEvent_shutdown();
	remove_proc_entry("driver/ifxAudioEvent", 0);
	printk(KERN_ERR "mps_event handler removed\n");
}


module_init(ifxAudioEvent_init);
module_exit(ifxAudioEvent_exit);

EXPORT_SYMBOL(mps_register_headset_callback);
EXPORT_SYMBOL(mps_unregister_headset_callback);

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
