/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
 * The Sidecar driver interfaces to Sidecar devices
 * using the SPI hardware of the INCA-IP2 (sidecar bus).
 * The driver accepts sidecar commands (=encoded L3 messages
 * and internal commands) from the user space, converts them
 * to L2 messages and sends them to sidecar bus.
 * It also accepts L2 messages from the sidecar bus, encodes
 * them to sidecar commands and sends them to user space.
 * Data are polled every 40ms.
 */

#include "sidecar.h" /* include all stuff we need */
#include "sidecar_hapi.c"

#include <linux/circ_buf.h>

/* circular buffer functions for accessing data and incrementing pointers */
#define SC_CIRC_OFF(BUF, RW, HT, OFFSET) \
  (BUF->buf[(BUF->HT + (OFFSET)) % RW##_BUFFER_SIZE])

#define SC_CIRC_INC(BUF, RW, HT, OFFSET) \
  (BUF->HT = (BUF->HT + (OFFSET)) % RW##_BUFFER_SIZE)

#define RD_CIRC_HEAD_OFF(OFFSET)	SC_CIRC_OFF(sidecar_data.read,  READ,  head, OFFSET)
#define WR_CIRC_HEAD_OFF(OFFSET)	SC_CIRC_OFF(sidecar_data.write, WRITE, head, OFFSET)
#define RD_CIRC_TAIL_OFF(OFFSET)	SC_CIRC_OFF(sidecar_data.read,  READ,  tail, OFFSET)
#define WR_CIRC_TAIL_OFF(OFFSET)	SC_CIRC_OFF(sidecar_data.write, WRITE, tail, OFFSET)

#define RD_CIRC_HEAD_OFF_P(BUF, OFFSET)	SC_CIRC_OFF(BUF, READ, head, OFFSET)
#define RD_CIRC_TAIL_OFF_P(BUF, OFFSET)	SC_CIRC_OFF(BUF, READ, tail, OFFSET)

#define RD_CIRC_HEAD_INC(OFFSET)	SC_CIRC_INC(sidecar_data.read,  READ,  head, OFFSET);
#define WR_CIRC_HEAD_INC(OFFSET)	SC_CIRC_INC(sidecar_data.write, WRITE, head, OFFSET);
#define RD_CIRC_TAIL_INC(OFFSET)	SC_CIRC_INC(sidecar_data.read,  READ,  tail, OFFSET);
#define WR_CIRC_TAIL_INC(OFFSET)	SC_CIRC_INC(sidecar_data.write, WRITE, tail, OFFSET);

#define RD_CIRC_HEAD_INC_P(BUF, OFFSET)	SC_CIRC_INC(BUF, READ, head, OFFSET);
#define RD_CIRC_TAIL_INC_P(BUF, OFFSET)	SC_CIRC_INC(BUF, READ, tail, OFFSET);

/* circular buffer functions from include/linux/circ_buf.h */
#define SC_CIRC_FUNC(BUF, RW, FUNC) \
  CIRC_##FUNC(BUF->head, BUF->tail, RW##_BUFFER_SIZE)

#define RD_CIRC_CNT			SC_CIRC_FUNC(sidecar_data.read,  READ,  CNT)
#define WR_CIRC_CNT			SC_CIRC_FUNC(sidecar_data.write, WRITE, CNT)
#define RD_CIRC_SPACE			SC_CIRC_FUNC(sidecar_data.read,  READ,  SPACE)
#define WR_CIRC_SPACE			SC_CIRC_FUNC(sidecar_data.write, WRITE, SPACE)
#define RD_CIRC_CNT_TO_END		SC_CIRC_FUNC(sidecar_data.read,  READ,  CNT_TO_END)
#define WR_CIRC_CNT_TO_END		SC_CIRC_FUNC(sidecar_data.write, WRITE, CNT_TO_END)
#define RD_CIRC_SPACE_TO_END		SC_CIRC_FUNC(sidecar_data.read,  READ,  SPACE_TO_END)
#define WR_CIRC_SPACE_TO_END		SC_CIRC_FUNC(sidecar_data.write, WRITE, SPACE_TO_END)

#define RD_CIRC_CNT_P(BUF)		SC_CIRC_FUNC(BUF, READ, CNT)
#define RD_CIRC_SPACE_P(BUF)		SC_CIRC_FUNC(BUF, READ, SPACE)
#define RD_CIRC_CNT_TO_END_P(BUF)	SC_CIRC_FUNC(BUF, READ, CNT_TO_END)
#define RD_CIRC_SPACE_TO_END_P(BUF)	SC_CIRC_FUNC(BUF, READ, SPACE_TO_END)

/* circular buffer functions for flushing and flipping */
#define SC_CIRC_FLUSH(RW) \
  (sidecar_data.RW->tail = sidecar_data.RW->head)

#define RD_CIRC_FLUSH			SC_CIRC_FLUSH(read)
#define WR_CIRC_FLUSH			SC_CIRC_FLUSH(write)

#define RD_CIRC_CURRENT						\
  ({								\
    sidecar_cbuf_t *__p;					\
    spin_lock_bh(&sidecar_data.read_flip_lock);			\
    __p = sidecar_data.read;					\
    spin_unlock_bh(&sidecar_data.read_flip_lock);		\
    __p;							\
  })

#define RD_CIRC_CNT_SAFE					\
  ({								\
    int __ret;							\
    spin_lock_bh(&sidecar_data.read_flip_lock);			\
    __ret = RD_CIRC_CNT;					\
    spin_unlock_bh(&sidecar_data.read_flip_lock);		\
    __ret;							\
  })

#define RD_CIRC_FLUSH_SAFE					\
  do {								\
    sidecar_cbuf_t *p;						\
    spin_lock_bh(&sidecar_data.read_flip_lock);			\
    if (RD_CIRC_CNT) {						\
      sidecar_data.read_flip->head = 0;				\
      sidecar_data.read_flip->tail = 0;				\
      p = sidecar_data.read;					\
      sidecar_data.read = sidecar_data.read_flip;		\
      sidecar_data.read_flip = p;				\
      printk(KERN_INFO "sidecar: read buffers flipped\n");	\
    }								\
    spin_unlock_bh(&sidecar_data.read_flip_lock);		\
  } while (0)

#define MOD1	0	/* 1st sidecar module */
#define MOD2	1	/* 2nd sidecar module */
#define MOD3	2	/* 3rd sidecar module - only initialized, never polled */

#define RESTART_ALL_NO_HAPI_RESET    0
#define RESTART_ALL_DO_HAPI_RESET    1

/**
 * whenever the sidecar device is opened from the user space,
 * the sidecar_open function is invoked
 */
int sidecar_open(struct inode *inode, struct file *filp) {

#ifdef SIDECAR_DEBUG
  int omajor;
  int ominor;

  /* get and print the driver device's minor and major number */
  ominor = MINOR(inode->i_rdev);
  omajor = MAJOR(inode->i_rdev);

  PDEBUG2("sidecar_open: major=%d, minor=%d\n", omajor, ominor);
#endif /* SIDECAR_DEBUG */

  /* keep in filp->private_data reference to sidecar_data, so other
   * functions can access it */
  filp->private_data = &sidecar_data;

  return 0;

}



/**
 * whenever the sidecar device is beeing closed,
 * the sidecar_close function is invoked
 */
int sidecar_close(struct inode *inode, struct file *filp) {

  PDEBUG2("sidecar_close: the sidecar device is beeing closed\n");

  return 0;

}



/**
 * read function called from the user space
 * copies data from kernel space to user space
 *
 * *filp is a pointer to the kernel file data structure
 * *buf is a pointer to the user space data buffer
 * count is a count of the user space data buffer
 * *f_pos is a pointer to file position
 * returns either number of data items transferred or error state
 */
ssize_t sidecar_read(struct file *filp, char *buf,
                     size_t count, loff_t * f_pos) {

  sidecar_device_t *dev = (sidecar_device_t*) filp->private_data;
  sidecar_cbuf_t *p;

  /* return value; says, how many bytes were copied to user space */
  ssize_t ret = 0;

  PDEBUG2("sidecar_read: someone is trying to read some data "
         "from the sidecar kernel module\n");

  /* check read buffer and sleep when empty */
  if (!(RD_CIRC_CNT_SAFE)) {

    if (filp->f_flags & O_NONBLOCK)
      return -EAGAIN;

    if (wait_event_interruptible(wait_queue_sidecar, RD_CIRC_CNT_SAFE))
      return -ERESTARTSYS;

  }

  /* copy pointer to current read buffer
   *
   * From this point on we will only use this local copy inside the
   * read() function. When the soft IRQ needs to flush the read buffer
   * it can do this be simply flipping the two read buffers. The 
   * ongoing read will not be influence be this as it still uses the
   * local copy and thereby the "flushed" buffer.
   */
  p = RD_CIRC_CURRENT;

  if (down_interruptible(&dev->rd_sem))
    return -ERESTARTSYS;

  /* check, if whole data in the buffer are in contiuous memory area --
   * we have a cyclical buffer, so it may happen, that start of data is
   * at the end of the buffer and the rest of data is at the beginning
   * of the buffer; if so, transfer only data from the end of the buffer
   */
  ret = RD_CIRC_CNT_TO_END_P(p);

  /* check, if user has enough space to get all data; if not, send
   * only a part of it */
  if (count < ret) {
    ret = count;
  }

  /* copy data to user space */
  if (copy_to_user(buf, &RD_CIRC_TAIL_OFF_P(p, 0), ret)) {
    printk(KERN_ERR "sidecar: copy_to_user failed\n");
    ret = -EFAULT;
    goto out;
  }

  /* increment tail pointer */
  RD_CIRC_TAIL_INC_P(p, ret);

out:
  /* return the count of bytes sent to user space */
  up(&dev->rd_sem);
  return ret;
}



/**
 * write function called from the user space
 * it copies data from user space to kernel space
 *
 * *filp is a pointer to the kernel file data structure
 * *buf is a pointer to the user space data buffer
 * count is a count of the user space data buffer
 * *f_pos is a pointer to file position
 * returns either number of data items transferred or error state
*/
ssize_t sidecar_write(struct file *filp, const char *buf,
                      size_t count, loff_t * f_pos) {

  sidecar_device_t *dev = (sidecar_device_t*) filp->private_data;

  /* return value; says, how many bytes were copied to user space */
  ssize_t ret;

  PDEBUG2("sidecar_write: someone is trying to write some data "
         "to the sidecar kernel module\n");

  if (down_interruptible(&dev->wr_sem))
    return -ERESTARTSYS;

  /* check, if whole data in the buffer are in contiuous memory area --
   * we have a cyclical buffer, so it may happen, that start of data is
   * at the end of the buffer and the rest of data is at the beginning
   * of the buffer; if so, we have available the whole data area
   * remaining in the buffer; if not, we can write only from end of
   * already stored data till the physical end of the buffer
   */
  ret = WR_CIRC_SPACE_TO_END;

  /* if user wants to write less data than we have available, decrease ret */
  if (count < ret) {
    ret = count;
  }

  /* copy data from user space */
  if (copy_from_user(&WR_CIRC_HEAD_OFF(0), buf, ret)) {
    printk(KERN_ERR "sidecar: copy_from_user failed\n");
    ret = -EFAULT;
    goto out;
  }

  /* increase the data size stored in the buffer */
  WR_CIRC_HEAD_INC(ret);

out:
  /* return the count of bytes sent to user space */
  up(&dev->wr_sem);
  return ret;

}



/* this function is invoked by the 'select' or 'poll' command from userspace;
 * it waits, until there are some output data available in the module
 * and then announces this fact
 */
unsigned int sidecar_poll(struct file *fp, struct poll_table_struct *pts) {

  unsigned int mask = 0;

  poll_wait(fp, &wait_queue_sidecar, pts);

  if (RD_CIRC_CNT_SAFE)
    mask |= POLLIN | POLLRDNORM;  /* readable */

  return mask;
}




/* flush all kernel module's buffers,
 * reset and initialize all sidecar devices
 */
void restart_all(int do_hapi_reset) {

  int i;

  /* flush all buffers */
  WR_CIRC_FLUSH;
  RD_CIRC_FLUSH_SAFE;

  for (i = MAX_OPTIONS - 1; i >= 0; --i) {

    sidecar_data.mod[i].downlink_len = 0;
    sidecar_data.mod[i].uplink_len   = 0;

    if (RD_CIRC_SPACE < 2)
      printk(KERN_ERR "sidecar: restart_all: read buffer full :-(\n");

    if ((sidecar_data.mod[i].device_presence > 0) &&
	(RD_CIRC_SPACE >= 2)) {

      /* send the DISCONNECT sidecar command uplink */
      RD_CIRC_HEAD_OFF(0) = CMD_TYPE_INTERNAL | CMD_DEVICE(i) | CMD_INT_DISCONNECT;
      RD_CIRC_HEAD_OFF(1) = 0;
      RD_CIRC_HEAD_INC(2);

      /* now we have some data for userspace */
      wake_up_interruptible(&wait_queue_sidecar);

    }

    /* no sidecar device present at the beginning */
    sidecar_data.mod[i].device_presence = 0;

  }

  /* (re)initialize sidecar bus */
  if ((do_hapi_reset == RESTART_ALL_DO_HAPI_RESET) &&
      (SidecarHAPI_Reset() != 0)) {
    printk(KERN_ERR "sidecar: SidecarHAPI_Reset failed\n");
  }


}




/* generates from the obtained uplink L2 message sidecar command
 * and stores it into cmd_buf; cmd_buf's available length
 * must be MAX_CMD_LEN; l2_len must be <= MAX_L2_LEN.
 * If there is something wrong with the obtained L2 message,
 * the length of the returned command in cmd_len will be 0
 * the sidecar bus and the whole module should be restarted
 * afterwards
 */
void generate_cmd_from_l2(unsigned char *l2_buf, int l2_len,
                          unsigned char *cmd_buf, int *cmd_len) {

  int int_len; /* length of the L3 message contained in the L2 message */
  int option;

  int i;


#ifdef SIDECAR_DEBUG
  PDEBUG1("generate_cmd_from_l2: obtained l2 message: ");
  for (i=0; i<l2_len; i++) {
    PDEBUG1("%c", "0123456789ABCDEF"[((l2_buf[i] >> 4) & 0xF)]);
    PDEBUG1("%c", "0123456789ABCDEF"[((l2_buf[i] >> 0) & 0xF)]);
    PDEBUG1(" ");
  }
  PDEBUG1("\n");
#endif /* SIDECAR_DEBUG */


  /* each L2 message must contain at least
   * status byte, length/type byte and at least
   * one message select byte and message type from the L3 message */
  if (l2_len < 4) {
    *cmd_len = 0;
    return;
  }

  PDEBUG2("l2_len: %i\n", l2_len);


  /* check, if the length of the message corresponds to the length
   * written in the "type and length" byte; if not or if the
   * length is > 28 or == 0, return with empty command
   * (which means that the calling function should
   * restart the sidecar bus and the module)
   */
   int_len = (int) (l2_buf[1] & L3_LENGTH_MASK);

   PDEBUG2("int_len: %i\n", int_len);

   if ((int_len == 0) || (int_len > 28) ||
       (int_len != l2_len - 2)) {
    *cmd_len = 0;
    return;
  }


  /* from the message select L3 byte find out, from which device
   * this message comes
   */

  option = (int) ((l2_buf[2] & L3_MS_INDEX_MASK) >> 4);

  PDEBUG2("option: %i\n", option);

  /* we support only 2 devices */
  if ((option != 0) && (option != 1)) {
    *cmd_len = 0;
    return;
  }



  /* write the Command byte and the length of the L3 message to the
   * cmd_buf buffer */
  cmd_buf[0] = (option == 0) ? CMD_DEVICE_1 : CMD_DEVICE_2 | CMD_TYPE_L3;
  if ((l2_buf[1] & L2_GLOB_LOC_MASK) == L2_LOCAL)
    cmd_buf[0] |= CMD_LOCAL;
  cmd_buf[1] = int_len;

  /* read the L3 message in the L2 message and write it as a command
   * to the cmd_buf
   */

  for (i=0; i<int_len; i++) {
    cmd_buf[i+2] = l2_buf[i+2];
  }

  *cmd_len = int_len + 2;


}





/* generates from the obtained command the L2 message
 * and stores it into buf; buf's available length must be MAX_L2_LEN */
void generate_l2_from_cmd(unsigned char cmd, unsigned char *buf, int *len) {

  unsigned char ctrl_1;

  /* set the Control-1 byte */
  ctrl_1 = CTRL1_DOWNLINK | CTRL1_DW_1 | CTRL1_D7_I_0;
  if ((cmd & CMD_DEVICE_MASK) == CMD_DEVICE_1) {
    /* message for the 1st device */
    ctrl_1 |= CTRL1_DEVICE_1;
  } else if ((cmd & CMD_DEVICE_MASK) == CMD_DEVICE_2) {
    /* message for the 2nd device */
    ctrl_1 |= CTRL1_DEVICE_2;
  } else {
    /* broadcast message */
    ctrl_1 |= CTRL1_BROADCAST;
  }

  /* set the 1st Control-1 byte */
  buf[0] = ctrl_1;


  switch (cmd & CMD_CONTENT_MASK) {

    default:
      /* wrong (unkwnown) command, return empty L2 message */
      *len = 0;
    break;

  }

}



/* generates from the obtained L3 msg the L2 msg
 * and stores it into buf; buf's available length must be MAX_L2_LEN */
void generate_l2_from_l3(unsigned char cmd, unsigned char *l3_buf, int l3_len,
                         unsigned char *buf, int *len) {

  int i;
  unsigned char ctrl_1;


#ifdef SIDECAR_DEBUG
  PDEBUG2("generate_l2_from_l3: obtained l3 message: ");
  for (i=0; i<l3_len; i++) {
    PDEBUG2("%c", "0123456789ABCDEF"[((l3_buf[i] >> 4) & 0xF)]);
    PDEBUG2("%c", "0123456789ABCDEF"[((l3_buf[i] >> 0) & 0xF)]);
    PDEBUG2(" ");
  }
  PDEBUG2("\n");
#endif /* SIDECAR_DEBUG */


  /* set the Control-1 byte */
  ctrl_1 = CTRL1_DOWNLINK | CTRL1_DW_1 | CTRL1_D7_I_0;
  if ((cmd & CMD_DEVICE_MASK) == CMD_DEVICE_1) {
    /* message for the 1st device */
    ctrl_1 |= CTRL1_DEVICE_1;
  } else if ((cmd & CMD_DEVICE_MASK) == CMD_DEVICE_2) {
    /* message for the 1st device */
    ctrl_1 |= CTRL1_DEVICE_2;
  } else {
    /* broadcast message */
    ctrl_1 |= CTRL1_BROADCAST;
  }

  /* set the 1st Control-1 byte */
  buf[0] = ctrl_1;

  /* set the length/type byte: Message Type is 0 (non-local msg),
   * Length Extension is 0 (not supported); the only non-zero value
   * is the length of the L3 message (it can't be longer than 31 bytes) */
  buf[1] = (unsigned char) l3_len & 0x1F;

  /* set the local/global bit in the length/type byte
   * it should be used here "buf[1] |= L2_LOCAL"
   * but L2_LOCAL is the 7th bit of the ctrl2 byte,
   * which is always beeing transfered in D7 bit
   * of the ctrl1 byte */
  if ((cmd & CMD_GLOB_LOC_MASK) == CMD_LOCAL)
    buf[0] |= CTRL1_D7_I_1;

  /* convert the L3 message to the L2 message */
  for (i=0; i<l3_len; i++) {
    /* insert the Control-1 byte;
     * the 4th bit of the Control-1 byte contains the 7th bit of the
     * following L3 data byte */
    buf[(i*2)+2] = ctrl_1 | (l3_buf[i] & 0x80) >> 3;
    /* from the data byte delete the 7th bit (it has to be 0) */
    buf[(i*2)+3] = l3_buf[i] & 0x7F;
  }


  *len = 2 + (l3_len * 2);


#ifdef SIDECAR_DEBUG
  PDEBUG1("generate_l2_from_l3: generated l2 message: ");
  for (i=0; i<*len; i++) {
    PDEBUG1("%c", "0123456789ABCDEF"[((buf[i] >> 4) & 0xF)]);
    PDEBUG1("%c", "0123456789ABCDEF"[((buf[i] >> 0) & 0xF)]);
    PDEBUG1(" ");
  }
  PDEBUG1("\n");
#endif /* SIDECAR_DEBUG */


}



/**
 * reads one command from the write buffer (buffer to store data from the user),
 * extracts the contained L3 message (if present), converts it to the L2
 * downlink message and puts it into the proper downlink_buf_n buffer,
 * put it's length to downlink_len_n and removes the command from the
 * write buffer; the function supposes all used buffers locked
 */
int process_cmd(unsigned char *buf, int *len) {

  unsigned char l3_buf[MAX_L3_LEN]; /* temporary buffer for actual L3 message */
  int l3_len; /* length of the L3 message contained in the command */
  unsigned char cmd;
  int i;


  /* empty or too small write buffer, no command to return
   * MAX_CMD_LEN - MAX_L3_LEN means the length of the aditional
   * data in the commmand excluding L3 message
   */
  if (WR_CIRC_CNT < (MAX_CMD_LEN - MAX_L3_LEN)) {
    *len = 0;
    return 0;
  }

  /* get the length of the L3 message contained in the command */
  l3_len = WR_CIRC_TAIL_OFF(1);

  /* command is too long -- unrecovable error, flush the write buffer */
  if (l3_len > MAX_L3_LEN) {
    WR_CIRC_FLUSH;
    *len = 0;
    return 1;
  }

  /* too small write buffer (user will hopefully add the rest
   * of the command later, currently no command to return
   * MAX_CMD_LEN - MAX_L3_LEN means the length of the aditional
   * data in the commmand excluding L3 message
   */
  if (l3_len + (MAX_CMD_LEN - MAX_L3_LEN) > WR_CIRC_CNT) {
    *len = 0;
    return 0;
  }

  /* read the command */
  cmd = WR_CIRC_TAIL_OFF(0);

  /* copy the L3 message to the internal buffer */
  for (i=0; i<l3_len; i++)
    l3_buf[i] = WR_CIRC_TAIL_OFF(i + MAX_CMD_LEN - MAX_L3_LEN);

  /* remove the actual command from the write buffer */
  WR_CIRC_TAIL_INC(l3_len + MAX_CMD_LEN - MAX_L3_LEN);

  /* create the L2 downlink message from the internal command or
   * from the L3 message
   */
  if ((cmd & CMD_TYPE_MASK) == CMD_TYPE_INTERNAL) {
    generate_l2_from_cmd(cmd, buf, len);
  } else {
    generate_l2_from_l3(cmd, l3_buf, l3_len, buf, len);
  }

#ifdef SIDECAR_DEBUG
  for (i=0; i<*len; i++) {
    PDEBUG2("%c", "0123456789ABCDEF"[((buf[i] >> 4) & 0xF)]);
    PDEBUG2("%c", "0123456789ABCDEF"[((buf[i] >> 0) & 0xF)]);
    PDEBUG2(" ");
  }
  PDEBUG2("\n");
#endif /* SIDECAR_DEBUG */

 return 0;


}



/* tries to initialize non-initialized sidecar devices;
 * returns the count of initialized devices
 * (old and newly together)
 */
int init_devices(sidecar_device_t *dev) {

  unsigned char ctrl[2]; /* small buffer used to write L2 ctrl1 and ctrl2 bytes */
  unsigned char ctrl_up = 0; /* used to read L2 uplink bytes */

  int ret = 0;
  int i, j;

  for (i = 0; i < MAX_OPTIONS; ++i) {
    
    if (RD_CIRC_SPACE < 2)
      printk(KERN_ERR "sidecar: init_devices: read buffer full :-(\n");

    /* if the i-th device is not yet present, and there is a space available
     * in the read buffer to put there the CONNECT command, try to initialize it */
    if ((dev->mod[i].device_presence == 0) && (RD_CIRC_SPACE >= 2)) {

      PDEBUG2("Trying to init the %d-th device\n", i + 1);

      /* try to initialize the i-th option */
      ctrl[0] = L2_CTRL1_INIT_DEV_1 | (i & 0x07); /* send ctrl 1 init byte */
      ctrl[1] = L2_CTRL2_INIT_DEV_1_TERM;

      if (SidecarHAPI_Write(2, ctrl) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
        if (SidecarHAPI_Flush(0) != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Flush read buffer failed\n");
        }
        goto init_end;
      }


      if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
      }

      /* device replied the correct init response */
      if (ctrl_up == (L2_INIT_RESP_DEV_1 | (i & 0x07))) {

        dev->mod[i].device_presence = 1;

        /* send the CONNECT sidecar command uplink */
        RD_CIRC_HEAD_OFF(0) = CMD_TYPE_INTERNAL | CMD_DEVICE(i) | CMD_INT_CONNECT;
        RD_CIRC_HEAD_OFF(1) = 0;
        RD_CIRC_HEAD_INC(2);

        PDEBUG1("Sidecar timeout -- the %d-th option has been initialized\n", i + 1);

        /* now we have some data for userspace */
        wake_up_interruptible(&wait_queue_sidecar);

      } else {
        /* when the i-th device is not initiated, the following ones can't be also */
        for (j = (i + 1); j < 3; ++j)
          dev->mod[j].device_presence = 0;
        ret = i;
        goto init_end;
      }

    } else {

      /* all devices up to i-th are initialized */
      ret = i + 1;

    }

  }


init_end:

  return ret;

}

/* releases the n-th device and sends a 'disconnect' sidecar command upwards */
void drop_device(int mod) {

  int i;

  if (mod == 0) {
    printk (KERN_INFO "sidecar: reset sidecar bus");
    for (i = 0; i < MAX_OPTIONS; ++i) {
      printk(" SC%d=%d", i + 1, sidecar_data.mod[i].device_presence);
    }
    printk("\n");
  }

  /* flush buffers */
  sidecar_data.mod[mod].downlink_len = 0;
  sidecar_data.mod[mod].uplink_len = 0;

  /* send the 'disconnect' sidecar command, if there is a space
   * in the read buffer */

  if (RD_CIRC_SPACE < 2)
    printk(KERN_ERR "sidecar: drop_device(%d): read buffer full :-(\n", mod);

  if ((sidecar_data.mod[mod].device_presence > 0) &&
      (RD_CIRC_SPACE >= 2)) {

    /* send the DISCONNECT sidecar command uplink */
    RD_CIRC_HEAD_OFF(0) = CMD_TYPE_INTERNAL | CMD_DEVICE(mod) | CMD_INT_DISCONNECT;
    RD_CIRC_HEAD_OFF(1) = 0;
    RD_CIRC_HEAD_INC(2);

    /* now we have some data for userspace */
    wake_up_interruptible(&wait_queue_sidecar);

  }

  /* delete the presence flag */
  sidecar_data.mod[mod].device_presence = 0;

  /* drop next device as well */
  if ((mod + 1) < MAX_OPTIONS)
    drop_device(mod + 1);

  /* reset all sidecar device options */
  if (mod == 0) {
    if (SidecarHAPI_Reset() != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Reset failed\n");
    }
  }

  return;

}

/* read one sidecar command from the user's input (write buffer),
 * process it to L2 downlink message and store it to mod[MOD1].downlink_buf or
 * mod[MOD2].downlink_buf buffer; if the next command in the write buffer
 * is dedicated to the other device than the 1st, process
 * it also
 * return 0 for no error, 1 when there was performed the restart of
 * all devices and the driver
 */
int process_write_buffer(sidecar_device_t *dev) {

  unsigned char cmd_msg[MAX_CMD_LEN];
  int cmd_len;

  if (WR_CIRC_CNT > 0) {

    /* internal command ordering to reset:
     * reset all devices and the kernel driver */
    if (((WR_CIRC_TAIL_OFF(0) & CMD_TYPE_MASK   ) == CMD_TYPE_INTERNAL) &&
        ((WR_CIRC_TAIL_OFF(0) & CMD_CONTENT_MASK) == CMD_INT_L2_RESET)) {

      if (((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_1) ||
	  ((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_2) ||
	  ((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_3) ||
	  ((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_BROADCAST))
      {

        /* clear all buffers, restart the sidecar bus */
        restart_all(RESTART_ALL_DO_HAPI_RESET);

      }

      PDEBUG1("Sidecar timeout -- CMD_INT_L2_RESET processed\n");

      return 1;

    }


    /* broadcast command (all devices) */
    if (((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_BROADCAST) &&
	(dev->mod[MOD1].device_presence == 1)) {

      PDEBUG1("Sidecar timeout -- broadcast command\n");

      /* read one command from the write buffer (from userspace),
       * convert it to the L2 message and write it to the downlink_buf buffer */
      if (dev->mod[MOD1].downlink_len == 0) {
        process_cmd(dev->mod[MOD1].downlink_buf, &(dev->mod[MOD1].downlink_len));
      }

      return 0;

    }


    /* command for the 1st device */
    if (((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_1) &&
	(dev->mod[MOD1].device_presence == 1)) {

      PDEBUG1("Sidecar timeout -- command for the 1st device\n");

      /* read one command from the write buffer (from userspace),
       * convert it to the L2 message and write it to the downlink_buf buffer */
      if (dev->mod[MOD1].downlink_len == 0) {
        process_cmd(dev->mod[MOD1].downlink_buf, &(dev->mod[MOD1].downlink_len));
      }

      /* now let's take a look, if the next message in the buffer
       * is for the 2nd device; if yes, process it */
      if (WR_CIRC_CNT &&
          ((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_2) &&
	  (dev->mod[MOD2].device_presence == 1)) {

        PDEBUG1("Sidecar timeout -- command for the 2nd device\n");

        /* read one command from the write buffer (from userspace),
         * convert it to the L2 message and write it to the
	 * downlink_buf buffer */
        if (dev->mod[MOD2].downlink_len == 0) {
          process_cmd(dev->mod[MOD2].downlink_buf, &(dev->mod[MOD2].downlink_len));
        }

      }


    /* command for the 2nd device */
    } else if (((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_2) &&
	       (dev->mod[MOD2].device_presence == 1)) {

      PDEBUG1("Sidecar timeout -- command for the 2nd device\n");

      /* read one command from the write buffer (from userspace),
       * convert it to the L2 message and write it to the downlink_buf buffer */
      if (dev->mod[MOD2].downlink_len == 0) {
        process_cmd(dev->mod[MOD2].downlink_buf, &(dev->mod[MOD2].downlink_len));
      }

      /* now let's take a look, if the next message in the buffer
       * is for the 1st device; if yes, process it */
      if (WR_CIRC_CNT &&
          ((WR_CIRC_TAIL_OFF(0) & CMD_DEVICE_MASK) == CMD_DEVICE_1) &&
	  (dev->mod[MOD1].device_presence == 1)) {

        PDEBUG1("Sidecar timeout -- command for the 1st device\n");

        /* read one command from the write buffer (from userspace),
         * convert it to the L2 message and write it to the
	 * downlink_buf buffer */
        if (dev->mod[MOD1].downlink_len == 0) {
          process_cmd(dev->mod[MOD1].downlink_buf, &(dev->mod[MOD1].downlink_len));
        }

      }


    /* unknown device, or command for non-existing device,
     * drop the command */
    } else {

      PDEBUG1("Sidecar timeout -- wrong command or device\n");
      PDEBUG1("Sidecar timeout -- CMD_DEVICE_MASK: %i\n", CMD_DEVICE_MASK);
      PDEBUG1("Sidecar timeout -- CMD_DEVICE_1: %i\n", CMD_DEVICE_1);
      PDEBUG1("Sidecar timeout -- CMD_DEVICE_2: %i\n", CMD_DEVICE_2);
      PDEBUG1("Sidecar timeout -- COMMAND: %i\n", WR_CIRC_TAIL_OFF(0));

      /* drop one command from the write buffer */
      process_cmd(cmd_msg, &cmd_len);

    }
  }


  return 0;

}



/* poll for data from initialized devices;
 * if there are some data available, read from
 * each device one L2 message, convert it to
 * sidecar command and put it to the read buffer;
 * return value is the count of devices, from
 * which we've obtained some data
 */
int read_uplink_data(sidecar_device_t *dev) {

  int transfer_up = 0; /* signalizes, if there were some uplink data transfered */

  /* temporary buffer for sidecar commands */
  unsigned char cmd_msg[MAX_CMD_LEN];
  int cmd_len;
  int i;

  unsigned char ctrl[2]; /* small buffer used to write L2 ctrl1 and ctrl2 bytes */
  unsigned char ctrl_up = 0; /* used to read L2 uplink bytes */
  int msg_len; /* length of the L2 msg to transfer from device to phone */

  ssize_t avail; /* lenght of data available in the read buffer for write */

  /* poll for data, only if we have space for them */
  if ((dev->mod[MOD1].uplink_len == 0) && (dev->mod[MOD1].device_presence == 1)) {

    PDEBUG2("sidecar: we have uplink_1 space, poll for data...\n");

    /* poll for data from the 1st device */
    ctrl[0] = L2_CTRL1_POLL_DEV_1; /* send the poll byte for the 1st device */
    ctrl[1] = 0; /* send empty ctrl 2 byte */
    if (SidecarHAPI_Write(2, ctrl) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
      return 0;
    }


    if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
    }


    /* if the status byte is 0xFF, it means, that the sidecar device has been
     * disconnected */
    if (ctrl_up == 0xFF) {
      PDEBUG1("1st sidecar device disconnected...\n");
      drop_device(0);
      return 0;
    }

    /* if sidecar device has some message waiting, read it */
    if ((ctrl_up == L2_MSG_WAITING_DEV_1) || (ctrl_up == L2_MSG_WAITING_DB_DEV_1)) {

      PDEBUG2("sidecar: there is a message from the 1st device, read it's len...\n");

      dev->mod[MOD1].uplink_buf[0] = ctrl_up;

      ctrl[0] = L2_READ_PACING; /* send the read pacing byte to the 1st device */
      ctrl[1] = 0; /* send empty ctrl 2 byte */
      if (SidecarHAPI_Write(2, ctrl) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
        if (SidecarHAPI_Flush(0) != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Flush read buffer failed\n");
        }
        return 0;
      }

      if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
      }

      /* get the length of the message from the 1st device */
      msg_len = (int) (ctrl_up & L2_UPLINK_LEN_MASK);

      /* if the length is ==0 or >28, restart the sidecar bus */
      if ((msg_len == 0) || (msg_len > 28)) {
        PDEBUG1("Wrong L2 length, resetting sidecar bus\n");
        if (SidecarHAPI_Reset() != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Reset failed\n");
        }
        return 0;
      }

      dev->mod[MOD1].uplink_buf[1] = ctrl_up;

      /* send read pacing so many times, so the whole uplink message
       * will be trasfered */

      PDEBUG2("sidecar: read message from the 1st device (len %i)...\n", msg_len);

      for (i=0; i<msg_len; i++) {
        ctrl[0] = L2_READ_PACING; /* send the read pacing byte to the 1st device */
        ctrl[1] = 0; /* send empty ctrl 2 byte */
        if (SidecarHAPI_Write(2, ctrl) != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
          if (SidecarHAPI_Flush(0) != 0) {
            printk(KERN_ERR "sidecar: SidecarHAPI_Flush read buffer failed\n");
          }
          return 0;
        }
      }

      if (SidecarHAPI_Read(msg_len, &(dev->mod[MOD1].uplink_buf[2])) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
      } else {

        dev->mod[MOD1].uplink_len = msg_len + 2;
        transfer_up++; /* mark, that the uplink transfer was processed */

#ifdef SIDECAR_DEBUG
        PDEBUG2("DEV 1 Message obtained: ");
        for (i=0; i<dev->mod[MOD1].uplink_len; i++) {
          PDEBUG2("%c", "0123456789ABCDEF"[((dev->mod[MOD1].uplink_buf[i] >> 4) & 0xF)]);
          PDEBUG2("%c", "0123456789ABCDEF"[((dev->mod[MOD1].uplink_buf[i] >> 0) & 0xF)]);
          PDEBUG2(" ");
        }
        PDEBUG2("\n");
#endif /* SIDECAR_DEBUG */


      }

    }

  }


  /* convert obtained L2 message to sidecar command and put it
   * to the output buffer */
  if (dev->mod[MOD1].uplink_len > 0) {

    /* find out, how much space do we have available in the buffer,
     * we will write into
    */
    avail = RD_CIRC_SPACE;

    /* we can't write more data than we have space available
     * SIDECAR_CMD_LEN = L2_UPLINK_LEN */
    if (dev->mod[MOD1].uplink_len <= avail) {

      generate_cmd_from_l2(dev->mod[MOD1].uplink_buf, dev->mod[MOD1].uplink_len,
                           cmd_msg, &cmd_len);


#ifdef SIDECAR_DEBUG
        PDEBUG2("Generated command: ");
        for (i=0; i<cmd_len; i++) {
          PDEBUG2("%c", "0123456789ABCDEF"[((cmd_msg[i] >> 4) & 0xF)]);
          PDEBUG2("%c", "0123456789ABCDEF"[((cmd_msg[i] >> 0) & 0xF)]);
          PDEBUG2(" ");
        }
        PDEBUG2("\n");
#endif /* SIDECAR_DEBUG */


      /* write the generated command to the output buffer */
      for (i=0; i<cmd_len; i++)
	RD_CIRC_HEAD_OFF(i) = cmd_msg[i];

      /* increase the size of the buffer */
      RD_CIRC_HEAD_INC(cmd_len);

      /* now we have some data for userspace */
      wake_up_interruptible(&wait_queue_sidecar);

      /* flush data from the uplink_buf buffer */
      dev->mod[MOD1].uplink_len = 0;

    } else {

      printk(KERN_ERR "sidecar: read_uplink_data: read buffer full :-(\n");

    }

  }


  /* poll for data, only if we have space for them */
  if ((dev->mod[MOD2].uplink_len == 0) && (dev->mod[MOD2].device_presence == 1)) {
    /* poll for data from the 1st device */
    ctrl[0] = L2_CTRL1_POLL_DEV_2; /* send the poll byte for the 1st device */
    ctrl[1] = 0; /* send empty ctrl 2 byte */
    if (SidecarHAPI_Write(2, ctrl) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
      if (SidecarHAPI_Flush(0) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Flush read buffer failed\n");
      }
      return transfer_up;
    }


    if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
    }


    /* if the status byte is 0xFF, it means, that the sidecar device has been
     * disconnected */
    if (ctrl_up == 0xFF) {
      PDEBUG1("2nd sidecar device disconnected...\n");
      drop_device(1);
      return transfer_up;
    }


    /* it sidecar device has some message waiting, read it */
    if ((ctrl_up == L2_MSG_WAITING_DEV_2) || (ctrl_up == L2_MSG_WAITING_DB_DEV_2)) {

      dev->mod[MOD2].uplink_buf[0] = ctrl_up;

      ctrl[0] = L2_READ_PACING; /* send the read pacing byte to the 1st device */
      ctrl[1] = 0; /* send empty ctrl 2 byte */
      if (SidecarHAPI_Write(2, ctrl) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
        if (SidecarHAPI_Flush(0) != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Flush read buffer failed\n");
        }
        return transfer_up;
      }

      if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
      }

      /* get the length of the message from the 1st device */
      msg_len = (int) (ctrl_up & L2_UPLINK_LEN_MASK);


      /* if the length is ==0 or >28, restart the sidecar bus */
      if ((msg_len == 0) || (msg_len > 28)) {
        PDEBUG1("Wrong L2 length, resetting sidecar bus\n");
        if (SidecarHAPI_Reset() != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Reset failed\n");
        }
        return 0;
      }


      dev->mod[MOD2].uplink_buf[1] = ctrl_up;

      /* send read pacing so many times, so the whole uplink message
       * will be trasfered */

      for (i=0; i<msg_len; i++) {
        ctrl[0] = L2_READ_PACING; /* send the read pacing byte to the 1st device */
        ctrl[1] = 0; /* send empty ctrl 2 byte */
        if (SidecarHAPI_Write(2, ctrl) != 0) {
          printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
          if (SidecarHAPI_Flush(0) != 0) {
            printk(KERN_ERR "sidecar: SidecarHAPI_Flush read buffer failed\n");
          }
          return transfer_up;
        }
      }


      if (SidecarHAPI_Read(msg_len, &(dev->mod[MOD2].uplink_buf[2])) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
      } else {

        dev->mod[MOD2].uplink_len = msg_len + 2;
        transfer_up++; /* mark, that the uplink transfer was processed */

#ifdef SIDECAR_DEBUG
        PDEBUG2("DEV 2 Message obtained: ");
        for (i=0; i<dev->mod[MOD2].uplink_len; i++) {
          PDEBUG2("%c", "0123456789ABCDEF"[((dev->mod[MOD2].uplink_buf[i] >> 4) & 0xF)]);
          PDEBUG2("%c", "0123456789ABCDEF"[((dev->mod[MOD2].uplink_buf[i] >> 0) & 0xF)]);
          PDEBUG2(" ");
        }
        PDEBUG2("\n");
#endif /* SIDECAR_DEBUG */

      }

    }

  }


  /* convert obtained L2 message to sidecar command and put it
   * to the output buffer */
  if (dev->mod[MOD2].uplink_len > 0) {

    /* find out, how much space do we have available in the buffer,
     * we will write into
    */
    avail = RD_CIRC_SPACE;

    /* we can't write more data than we have space available
     * SIDECAR_CMD_LEN = L2_UPLINK_LEN */
    if (dev->mod[MOD2].uplink_len <= avail) {

      generate_cmd_from_l2(dev->mod[MOD2].uplink_buf, dev->mod[MOD2].uplink_len,
                           cmd_msg, &cmd_len);

      /* write the generated command to the output buffer */
      for (i=0; i<cmd_len; i++)
	RD_CIRC_HEAD_OFF(i) = cmd_msg[i];

      /* increase the size of the buffer */
      RD_CIRC_HEAD_INC(cmd_len);

      /* now we have some data for userspace */
      wake_up_interruptible(&wait_queue_sidecar);

      /* flush data from the uplink_buf buffer */
      dev->mod[MOD2].uplink_len = 0;

    } else {
      printk(KERN_ERR "sidecar: read_uplink_data: read buffer full :-(\n");
    }

  }

  /* poll for data, only if we have space for them */
  if ((dev->mod[MOD3].uplink_len == 0) && (dev->mod[MOD3].device_presence == 1)) {

    PDEBUG2("sidecar: we have uplink_1 space, poll for data...\n");

    /* poll for data from the 1st device */
    ctrl[0] = L2_CTRL1_POLL_DEV_3; /* send the poll byte for the 1st device */
    ctrl[1] = 0; /* send empty ctrl 2 byte */
    if (SidecarHAPI_Write(2, ctrl) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
      return 0;
    }


    if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
    }


    /* if the status byte is 0xFF, it means, that the sidecar device has been
     * disconnected */
    if (ctrl_up == 0xFF) {
      PDEBUG1("3rd sidecar device disconnected...\n");
      drop_device(2);
      return 0;
    }

  }

  PDEBUG2("Transfer_up is %i\n", transfer_up);

  return transfer_up;

}





/* writes data prepared in downlink buffers to the
 * sidecar bus
 */
int write_downlink_data(sidecar_device_t *dev) {

  int transfer = 0;

  /* no device present, nothing to send... */
  if (dev->mod[MOD1].device_presence == 0)
    return transfer;

  /* find out, if there are some data to trasfer */
  if (dev->mod[MOD1].downlink_len > 0) {

    if (SidecarHAPI_Write(dev->mod[MOD1].downlink_len, dev->mod[MOD1].downlink_buf) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
    }

    if (SidecarHAPI_Flush(0) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Flush failed\n");
    }

    dev->mod[MOD1].downlink_len = 0;

    transfer++;

  }



  /* 2nd device not present, nothing to send... */
  if (dev->mod[MOD2].device_presence == 0)
    return transfer;

  /* find out, if there are some data to trasfer */
  if (dev->mod[MOD2].downlink_len > 0) {

    if (SidecarHAPI_Write(dev->mod[MOD2].downlink_len, dev->mod[MOD2].downlink_buf) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
    }

    if (SidecarHAPI_Flush(0) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Flush failed\n");
    }

    dev->mod[MOD2].downlink_len = 0;

    transfer++;

  }

  return transfer;

}



/* process fastmode transfer for given device */
int process_fastmode(int arg) {

  unsigned char ctrl[2]; /* small buffer used to write L2 ctrl1 and ctrl2 bytes */
  unsigned char ctrl_up = 0; /* used to read L2 uplink bytes */
  int device;

  PDEBUG1("Sidecar fastmode invoked...\n");

  /* get from the userspace argument no of sidecar device */
  device = ((arg & FASTMODE_DEV_2) ? 1 : 0);

  /* fastmode data for the 1st device */
  if (((device == 0) && (sidecar_data.mod[MOD1].device_presence)) ||
      ((device == 1) && (sidecar_data.mod[MOD2].device_presence))) {

    /* send the 'fastmode request' L2 message to the device */
    ctrl[0] = FASTMODE_REQUEST | ((device == 0) ? CTRL1_DEVICE_1 : CTRL1_DEVICE_2);
    ctrl[1] = FASTMODE_CTRL2;
    if (SidecarHAPI_Write(2, ctrl) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
      return -1;
    }

    /* read answer from the device */
    if (SidecarHAPI_Read(1, &ctrl_up) != 0) {
      printk(KERN_ERR "sidecar: SidecarHAPI_Read failed\n");
      return -1;
    }

    /* fastmode ack -> wait for device to be prepared and process the fastmode transfer */
    if (ctrl_up == (FASTMODE_ACK | ((device == 0) ? CTRL1_DEVICE_1 : CTRL1_DEVICE_2))) {

      if (SidecarHAPI_Fastmode_Write(
          (arg & (~(FASTMODE_DEV_1 | FASTMODE_DEV_2))) ? FASTMODE_DATA_LABEL_SIZE
	                                               : FASTMODE_DATA_WHOLE_SIZE,
          ((device == 0) ? sidecar_data.mod[MOD1].fastmode_buf
	                 : sidecar_data.mod[MOD2].fastmode_buf)) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Fastmode_Write failed\n");
        return -1;
      }

    }

    PDEBUG1("Sidecar fastmode successfully finished...\n");
    return 0;

  } else {

    PDEBUG1("Sidecar fastmode not even started (no device)...\n");
    return -1;
  }

  return -1;

}




/**
 * the sidecar timeout function; it is periodically invoked
 * by the kernel timeout mechanism;
 * in this function are performed actions and operations,
 * which are not invoked from the user space
 * (processing of messages in buffers etc.)
 */
static void sidecar_timeout(unsigned long data) {

  sidecar_device_t *dev = (sidecar_device_t*) data;

  /* if there is some other transfer ongoing (probably fastmode) exit */
  if (down_trylock(&dev->hapi_sem)) {
    printk (KERN_INFO "sidecar: could not get hapi_sem in %s"
            " - wait for next timer interval\n", __FUNCTION__);
    goto out;
  }

  /* try to init all uninitialized devices;
   * if there is no sidecar device, exit */
  if (init_devices(dev) == 0) {
    up(&dev->hapi_sem);
    goto out;
  }

  /* process first one or two sidecar commands to L2 downlink messages;
     if there was the restart performed, exit */
  if (process_write_buffer(dev) == 1) {
    up(&dev->hapi_sem);
    goto out;
  }

  /* each DOWN_BEFORE_UP_CYCLE-times make write before read */
  if (dev->down_up < DOWN_BEFORE_UP_CYCLE) {

    /* if there was no uplink data transfer, make downlink transfers */
    PDEBUG2("Read uplink data...\n");
    if (read_uplink_data(dev) == 0) {
      PDEBUG2("No uplink message, try to write downlink data...\n");
      write_downlink_data(dev);
    }

    dev->down_up++;

  } else {

    /* if there was no downlink data transfer, make uplink transfer */
    PDEBUG2("Write downlink data...\n");
    if (write_downlink_data(dev) == 0) {
      PDEBUG2("No downlink message, try to read uplink data...\n");
      read_uplink_data(dev);
    }

    dev->down_up = 0;

  }

  up(&dev->hapi_sem);

out:
  /* put again this function to the kernel timer queue */
  spin_lock_bh(&sidecar_data.timer_active_lock);
  if (sidecar_data.timer_active) {
    dev->timer.expires = jiffies + (TIMER_INTERVAL_IN_MILLISECS * HZ) / 1000;
    add_timer(&dev->timer);
    sidecar_data.timer_active = 1;
  }
  spin_unlock_bh(&sidecar_data.timer_active_lock);
}



/** mmap function invoked when there is called the mmap function
 * from userspace; here is the set-up of basic parameters, the
 * data processing is done in sidecar_vma_nopage function
 */
int sidecar_mmap(struct file *fp, struct vm_area_struct *vma) {

  vma->vm_ops = &sidecar_vm_ops;
  vma->vm_flags |= VM_RESERVED;
  vma->vm_private_data = &sidecar_data;

  return 0;

}



/** implementation of the mmap nopage function; returns reference
 * to proper memory page requested by userspace
 */
struct page *sidecar_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write_access) {

  struct page *page;
  sidecar_device_t *dev;
  unsigned long offset;
  
  /* compute the offset value */
  offset = (address - vma->vm_start) + (vma->vm_pgoff << PAGE_SHIFT);

  /* get the private info of the device */
  dev = vma->vm_private_data;

  /* for each device there is one page allocated */
  if (offset == 0) {
    /* page allocated for the 1st device */
    page = virt_to_page(dev->mod[MOD1].fastmode_buf);
  } else {
    /* page allocated for the 2nd device */
    page = virt_to_page(dev->mod[MOD2].fastmode_buf);
  }

  /* increase usage info of the page */
  get_page(page);


  return page;

}



/** ioctl function invoked when there is called the ioctl function
 * from userspace; it is used to send asynchronously the
 * LED's synchronisation command to sidecar devices
 * and to initiate the fastmode data transfer to sidecars
 */
int sidecar_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg) {

  sidecar_device_t *dev = (sidecar_device_t*) filp->private_data;
  unsigned char ctrl[2];
  int ret = 0;
  int status;

  switch (cmd) {

    case IOCTL_SYNC: /* LED's sync */

      /* wait until HAPI becomes available */
      if (down_interruptible(&dev->hapi_sem))
	return -ERESTARTSYS;

      /* send the LED's sync L2 message to sidecar bus */
      ctrl[0] = L2_SYNC_CTRL1;
      ctrl[1] = L2_SYNC_CTRL2;

      status = SidecarHAPI_Write(2, ctrl);
      if (status == NOT_OPENED_ERROR) {
        ret = -EBUSY;
      } else if (status != 0) {
        printk(KERN_ERR "sidecar: LED's sync: SidecarHAPI_Write failed\n");
        ret = (status < 0) ? status : -EFAULT;
      }

      status = SidecarHAPI_Flush(0);
      if (status == NOT_OPENED_ERROR) {
        ret = -EBUSY;
      } else if (status != 0) {
	printk(KERN_ERR "sidecar: LED's sync: SidecarHAPI_Flush failed\n");
        ret = (status < 0) ? status : -EFAULT;
      }

      up(&dev->hapi_sem);

      break;

    case IOCTL_FASTMODE: /* fastmode transfer */

      /* wait until HAPI becomes available */
      if (down_interruptible(&dev->hapi_sem))
	return -ERESTARTSYS;

      ret = process_fastmode(arg);
      up(&dev->hapi_sem);

      break;

    case IOCTL_START: /* startup of the sidecar bus and polling timer */

      /* wait until HAPI becomes available */
      if (down_interruptible(&dev->hapi_sem))
	return -ERESTARTSYS;

      /* initiate the HAPI interface */
      status = SidecarHAPI_Open();
      if (status == ALREADY_OPENED_BUS_ERROR) {
        printk (KERN_INFO "sidecar: IOCTL_START - trying to open already opened device\n");
        up(&dev->hapi_sem);
        ret = -EBUSY;
        break;
      } else if (status != 0) {
        printk (KERN_ERR "sidecar: IOCTL_START failed %d\n", status);
        up(&dev->hapi_sem);
        ret = (status < 0) ? status : -EFAULT;
        break;
      }

      /* reset all sidecar device options */
      ctrl[0] = L2_CTRL1_RESET;
      ctrl[1] = L2_CTRL2_RESET;
      if (SidecarHAPI_Write(2, ctrl) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Write failed\n");
      }

      /* flush obtained data from sidecar device */
      if (SidecarHAPI_Flush(2) != 0) {
        printk(KERN_ERR "sidecar: SidecarHAPI_Flush failed\n");
      }

      up(&dev->hapi_sem);

      /* reset the uplink / downlink data transfer priority flag */
      sidecar_data.down_up = 0;

      /* startup of the sidecar timer function */
      spin_lock_bh(&sidecar_data.timer_active_lock);
      if (!sidecar_data.timer_active) {
        sidecar_data.timer.expires = jiffies +
          (TIMER_INTERVAL_IN_MILLISECS * HZ) / 1000;
        add_timer(&sidecar_data.timer);
        sidecar_data.timer_active = 1;
      }
      spin_unlock_bh(&sidecar_data.timer_active_lock);

      break;

    case IOCTL_STOP:

      /* stop the timer */
      spin_lock_bh(&sidecar_data.timer_active_lock);
      sidecar_data.timer_active = 0;
      spin_unlock_bh(&sidecar_data.timer_active_lock);
      del_timer_sync(&sidecar_data.timer);

      /* wait until HAPI becomes available */
      if (down_interruptible(&dev->hapi_sem))
	return -ERESTARTSYS;

      /* close HAPI */
      status = SidecarHAPI_Close();
      if (status == NOT_OPENED_ERROR)
        printk (KERN_INFO "sidecar: IOCTL_STOP - trying to close already closed device\n");
      else if (status != 0)
        printk (KERN_ERR "sidecar: IOCTL_STOP failed %d\n", status);

      /* flush buffers */
      restart_all(RESTART_ALL_NO_HAPI_RESET);

      up(&dev->hapi_sem);

      break;

    case IOCTL_L2_RESET:

      /* wait until HAPI becomes available */
      if (down_interruptible(&dev->hapi_sem))
	return -ERESTARTSYS;

      restart_all(RESTART_ALL_DO_HAPI_RESET);

      up(&dev->hapi_sem);

      break;

    default:
      return -ENOTTY;

  }

  return ret;

}




#ifdef SIDECAR_DEBUG
/* function which is called whenever user tries to read the
 * /proc/drivers/sidecar file; used only for debugging purposes
 */
static int sidecar_read_procmem(char *buf, char **start, off_t offset,
                               int count, int *eof, void *data) {

  int len = 0;
  int i;

  char info[800];


  len += sprintf(info+len, "read_buffer_len    : ");
  len += sprintf(info+len, "%i\n", RD_CIRC_CNT);

  len += sprintf(info+len, "write_buffer_len   : ");
  len += sprintf(info+len, "%i\n", WR_CIRC_CNT);


  len += sprintf(info+len, "read_buffer        : ");

  for (i = 0; i < (RD_CIRC_CNT < 21 ? RD_CIRC_CNT : 21); i++) {

    len += sprintf(info+len, "%c",
		   "0123456789ABCDEF"[(RD_CIRC_TAIL_OFF(i) >> 4) & 0xF]);
    len += sprintf(info+len, "%c ",
		   "0123456789ABCDEF"[(RD_CIRC_TAIL_OFF(i) >> 0) & 0xF]);

  }

  len += sprintf(info+len, "\n");


  len += sprintf(info+len, "write_buffer       : ");

  for (i = 0; i < (WR_CIRC_CNT < 21 ? WR_CIRC_CNT : 21); i++) {

    len += sprintf(info+len, "%c",
		   "0123456789ABCDEF"[(WR_CIRC_TAIL_OFF(i) >> 4) & 0xF]);
    len += sprintf(info+len, "%c ",
		   "0123456789ABCDEF"[(WR_CIRC_TAIL_OFF(i) >> 0) & 0xF]);

  }

  len += sprintf(info+len, "\n\n");

  len += sprintf(info+len, "downlink_buffer_1  : ");

  for (i = 0;
       i < (sidecar_data.mod[MOD1].downlink_len < MAX_L2_LEN ?
       sidecar_data.mod[MOD1].downlink_len :
       MAX_L2_LEN); i++) {

    len += sprintf(info+len, "%c",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD1].downlink_buf[i]) >> 4) & 0xF]);
    len += sprintf(info+len, "%c ",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD1].downlink_buf[i]) >> 0) & 0xF]);

  }

  len += sprintf(info+len, "\n");

  len += sprintf(info+len, "downlink_buffer_2  : ");

  for (i = 0;
       i < (sidecar_data.mod[MOD2].downlink_len < MAX_L2_LEN ?
       sidecar_data.mod[MOD2].downlink_len :
       MAX_L2_LEN); i++) {

    len += sprintf(info+len, "%c",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD2].downlink_buf[i]) >> 4) & 0xF]);
    len += sprintf(info+len, "%c ",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD2].downlink_buf[i]) >> 0) & 0xF]);
  }

  len += sprintf(info+len, "\n");

  len += sprintf(info+len, "uplink_buffer_1    : ");

  for (i = 0;
       i < (sidecar_data.mod[MOD1].uplink_len < MAX_L2_LEN ?
       sidecar_data.mod[MOD1].uplink_len :
       MAX_L2_LEN); i++) {

    len += sprintf(info+len, "%c",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD1].uplink_buf[i]) >> 4) & 0xF]);
    len += sprintf(info+len, "%c ",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD1].uplink_buf[i]) >> 0) & 0xF]);
  }

  len += sprintf(info+len, "\n");

  len += sprintf(info+len, "uplink_buffer_2    : ");

  for (i = 0;
       i < (sidecar_data.mod[MOD2].uplink_len < MAX_L2_LEN ?
       sidecar_data.mod[MOD2].uplink_len :
       MAX_L2_LEN); i++) {

    len += sprintf(info+len, "%c",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD2].uplink_buf[i]) >> 4) & 0xF]);
    len += sprintf(info+len, "%c ",
      "0123456789ABCDEF"[((sidecar_data.mod[MOD2].uplink_buf[i]) >> 0) & 0xF]);
  }

  len += sprintf(info+len, "\n");


  if( len > count) {
    len = count;
  }

  memcpy(buf, info, len);

  return len;

}

#endif /* SIDECAR_DEBUG */




#ifdef SIDECAR_DEBUG
/* function which is called whenever user tries to write something
 * to the /proc/drivers/sidecar file; not used actually
 */
static int sidecar_write_procmem(struct file *file, const char *buffer,
                                unsigned long count, void *data) {


  return count;

}
#endif /* SIDECAR_DEBUG */





/**
 * init function, which is called when the kernel module is
 * loaded into the memory
 */
static int __init sidecar_init(void) {

  int result;
  int i;

  /* create the /dev/sidecar devfs entry
   * parameters:
   *   NULL -- the sidecar device entry will be directly in /dev,
   *           not in some subdirectory in /dev
   *   "sidecar" -- name of the sidecar device file
   *   DEVFS_FL_AUTO_DEVNUM -- automatically allocate a device number
   *   0, 0 -- major and minor numbers are allocated automatically
   *   S_IFCHR | S_IRUGO | S_IWUGO -- access mode of the device:
   *     S_IFCHR = character special file
   *     S_IRUGO = allow to read for everyone
   *     S_IWUGO = allow to write for everyone
   */

 sidecar_devfs_handle = devfs_register(NULL, "sidecar", DEVFS_FL_AUTO_DEVNUM,
    0, 0, S_IFCHR | S_IRUGO | S_IWUGO, &sidecar_fops, 0);

  /* we need to set the 'owner' field of the 'file_operations' structure */
  SET_MODULE_OWNER(&sidecar_fops);

  printk(KERN_NOTICE "\nsidecar device registered.\n");

  /* create the sidecar proc entry */
  if ((sidecar_proc = create_proc_entry("driver/sidecar", 0, 0)) == 0) {
    printk(KERN_ERR "sidecar: could not create /proc/driver/sidecar entry\n");
    result = -ENOENT;
    goto fail_proc;
  }

#ifdef SIDECAR_DEBUG
  sidecar_proc->read_proc  = sidecar_read_procmem;
  sidecar_proc->write_proc = sidecar_write_procmem;

  PDEBUG1("sidecar_init: /proc/driver/sidecar created...\n");
#endif /* SIDECAR_DEBUG */

  /* initialize the timer structure */
  init_timer(&sidecar_data.timer);

  /* initialize all used semaphores */
  init_MUTEX(&sidecar_data.rd_sem);
  init_MUTEX(&sidecar_data.wr_sem);
  init_MUTEX(&sidecar_data.hapi_sem);
  spin_lock_init(&sidecar_data.read_flip_lock);
  spin_lock_init(&sidecar_data.timer_active_lock);

  /* initialize the read and write buffers */
  /* no sidecar device present at the beginning */
  for (i = 0; i < MAX_OPTIONS; ++i) {
    sidecar_data.mod[i].downlink_len    = 0;
    sidecar_data.mod[i].uplink_len      = 0;
    sidecar_data.mod[i].device_presence = 0;
  }

  /* allocate buffers */
  sidecar_data.read = kmalloc(sizeof(sidecar_cbuf_t) + READ_BUFFER_SIZE, GFP_KERNEL);
  if (!sidecar_data.read) {
    result = -ENOMEM;
    goto fail_malloc1;
  }

  sidecar_data.read_flip = kmalloc(sizeof(sidecar_cbuf_t) + READ_BUFFER_SIZE, GFP_KERNEL);
  if (!sidecar_data.read_flip) {
    result = -ENOMEM;
    goto fail_malloc2;
  }

  sidecar_data.write = kmalloc(sizeof(sidecar_cbuf_t) + WRITE_BUFFER_SIZE, GFP_KERNEL);
  if (!sidecar_data.write) {
    result = -ENOMEM;
    goto fail_malloc3;
  }

  memset((void *)sidecar_data.read,      0, sizeof(sidecar_cbuf_t) + READ_BUFFER_SIZE );
  memset((void *)sidecar_data.read_flip, 0, sizeof(sidecar_cbuf_t) + READ_BUFFER_SIZE );
  memset((void *)sidecar_data.write,     0, sizeof(sidecar_cbuf_t) + WRITE_BUFFER_SIZE);

  /* allocate fastmode buffers */
  for (i = 0; i < MAX_OPTIONS; ++i)
    sidecar_data.mod[i].fastmode_buf = NULL;

  for (i = 0; i < MAX_OPTIONS; ++i) {
    sidecar_data.mod[i].fastmode_buf = (unsigned char *)__get_free_page(__GFP_DMA);
    if (!sidecar_data.mod[i].fastmode_buf) {
      result = -ENOMEM;
      goto fail_malloc4;
    }
  }

  /* initialize timer */
  sidecar_data.timer_active = 0;
  sidecar_data.timer.data = (int) &sidecar_data;
  sidecar_data.timer.function = sidecar_timeout;

  printk(KERN_INFO "sidecar: the sidecar kernel module has been loaded\n");

  return 0;

fail_malloc4:
  for (i = 0; i < MAX_OPTIONS; ++i) {
    if (sidecar_data.mod[i].fastmode_buf != NULL)
      free_page((unsigned long) sidecar_data.mod[i].fastmode_buf);
  }
  kfree(sidecar_data.write);
fail_malloc3:
  kfree(sidecar_data.read);
fail_malloc2:
  kfree(sidecar_data.read);
fail_malloc1:
  remove_proc_entry("driver/sidecar", 0);
fail_proc:
  devfs_unregister(sidecar_devfs_handle);

  printk(KERN_ERR "sidecar: module loading failed error %d\n", result);
  return result;
}



/**
 * exit function, which is called when the kernel module is
 * unloaded from the memory
*/
static void __exit sidecar_exit(void) {

  int status;
  int i;

  /* stop the timer */
  spin_lock_bh(&sidecar_data.timer_active_lock);
  sidecar_data.timer_active = 0;
  spin_unlock_bh(&sidecar_data.timer_active_lock);
  del_timer_sync(&sidecar_data.timer);

  /* close HAPI */
  status = SidecarHAPI_Close();
  if (status == ALREADY_OPENED_BUS_ERROR)
    printk (KERN_INFO "sidecar: IOCTL_STOP - trying to close already closed device\n");
  else if (status != 0)
    printk (KERN_ERR "sidecar: IOCTL_STOP failed %d\n", status);

  /* release memory */
  for (i = 0; i < MAX_OPTIONS; ++i)
    free_page((unsigned long) sidecar_data.mod[i].fastmode_buf);

  kfree(sidecar_data.write);
  kfree(sidecar_data.read);
  kfree(sidecar_data.read);

  /* release other ressources */
  remove_proc_entry("driver/sidecar", 0);
  devfs_unregister(sidecar_devfs_handle);

  printk(KERN_NOTICE "sidecar module exit\n");

}



/* if the kernel driver is compiled as a module, it has to have
 * the init and exit functions */
#ifdef MODULE
module_init(sidecar_init);
module_exit(sidecar_exit);
#endif


/* sidecar module's licence; with non-GPL licences, the kernel complains
 * about 'tainting' when the module is beeing loaded */
MODULE_LICENSE("GPL");




