/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
 * The Sidecar driver interfaces to Sidecar devices
 * using the SPI hardware of the INCA-IP2 (sidecar bus).
 * The driver accepts sidecar commands (=encoded L3 messages
 * and internal commands) from the user space, converts them
 * to L2 messages and sends them to sidecar bus.
 * It also accepts L2 messages from the sidecar bus, encodes
 * them to sidecar commands and sends them to user space.
 * Data are polled every 40ms.
 */

#include <linux/module.h> /* necessary for all kernel modules */
#include <linux/kernel.h> /* because of printk */
#include <linux/init.h> /* module_init and module_exit definition */
#include <linux/fs.h> /* many of the kernel API functions are defined here */
#include <linux/devfs_fs_kernel.h> /* functions for devfs operations */
#include <linux/sched.h> /* scheduler functions -- sleep etc. */
#include <linux/proc_fs.h> /* the /proc functions */
#include <asm/uaccess.h> /* copy_from_user and copy_to_user functions */
#include <linux/poll.h> /* to enable 'select' function from userspace */
#include <linux/mm.h> /* mmap functionality */
#include <linux/delay.h> /* udelay function */


/* if defined, the sidecar module will print informational
 * debugging information (printk) and it will also
 * create the /proc/drivers/sidecar entry to display
 * debugging data */
#define SIDECAR_DEBUG


#ifdef SIDECAR_DEBUG
/* ipmortant debugging messages */
//#  define PDEBUG1(fmt, args...)   printk(KERN_INFO fmt, ## args)
#  define PDEBUG1(fmt, args...)
/* messy debugging messages, not used actually */
//#  define PDEBUG2(fmt, args...)   printk(KERN_INFO fmt, ## args)
#  define PDEBUG2(fmt, args...)
#else
#  define PDEBUG1(fmt, args...)
#  define PDEBUG2(fmt, args...)
#endif /* SIDECAR_DEBUG */



/* timer interval for the sidecar_timeout function */
#define TIMER_INTERVAL_IN_MILLISECS 40


/* size of the read buffer (buffer kernel --> user) */
#define READ_BUFFER_SIZE 512

/* size of the write buffer (buffer user --> kernel) */
#define WRITE_BUFFER_SIZE 1024



/* maximun supported length of L3 messages */
#define MAX_L3_LEN 28

/* maximun supported length of L2 messages (status byte +
 * length/type byte + L3 message */
#define MAX_L2_LEN ((MAX_L3_LEN * 2) + 4)

/* maximun length of the commad obtained from the user or passed to the user */
#define MAX_CMD_LEN (MAX_L3_LEN + 2)

/* maximum number of options */
#define MAX_OPTIONS 3


/* the command byte is separated into 3 parts like this:
 * D D T C C C C C
 * where D part is the device number (00 - 1st device, 01 - 2nd device, 11 - broadcast)
 *       T part is the command type (0 - L3 command, 1 - internal command)
 *       C part is the content - internal command's value or non-defined
 *       (for L3)
 *
 * the CMD_DEVICE_MASK says, in which bits of the command is the device
 * address stored
 * the CMD_DEVICE_n is the bitmask of the device number
 */
#define CMD_DEVICE_MASK            0xC0
#define CMD_TYPE_MASK              0x20
#define CMD_CONTENT_MASK           0x1F
#define CMD_GLOB_LOC_MASK          0x10

#define CMD_DEVICE_1               0x00
#define CMD_DEVICE_2               0x40
#define CMD_DEVICE_3               0x80
#define CMD_BROADCAST              0xC0
#define CMD_DEVICE(i)              ((i & 0x03) << 6)

/* internal or L3 command */
#define CMD_TYPE_INTERNAL          0x20
#define CMD_TYPE_L3                0x00

/* global or local commands (L2 messages) */
#define CMD_GLOBAL                 0x00
#define CMD_LOCAL                  0x10

/* internal commands */
#define CMD_INT_L2_RESET           0x00
#define CMD_INT_FASTMODE_REQ       0x01
#define CMD_INT_CONNECT            0x02
#define CMD_INT_DISCONNECT         0x03
#define CMD_INT_FASTMODE_ERR       0x04


/* L3 downlink messages */
#define L3_OPTION_INIT_IND         0x00
#define L3_OPTION_INFO_REQ         0x01
#define L3_OPTION_RESET            0x02
#define L3_OPTION_INIT_REJ         0x03
#define L3_DSS_LED                 0x11
#define L3_DSS_KEY_LAYER           0x13
#define L3_DSS_LABEL_CONTRAST      0x14
#define L3_DSS_BACKLIGHT           0x40 /* must be local */
#define L3_DSS_LED_TEMP_OFF        0x41 /* must be local */


/* L3 uplink messages */
#define L3_DSS_KEY                 0x11
#define L3_OPTION_INFO_IND         0x01
#define L3_OPTION_L3_ERROR         0x02
#define L3_OPTION_FASTMODE_ACK     0x40 /* must be local */


/* mask for the length value in the length/type byte
 * of the L3 message */
#define L3_LENGTH_MASK             0x1F


/* mask for the 'option index' value in the message select byte */
#define L3_MS_INDEX_MASK           0x30



/* L2 Control 1 byte for downlink data messages
 *
 *  |==================================================================|
 * || 1 | Read Pacing | Data Write | Data Bit 7/Init | Poll | Device # ||
 * || ------------------------ CTRL1_DOWNLINK ------------------------ ||
 * || 1 |      0      |      1     |         0       |   0  |   000    ||
 * || ------------------------ CTRL1_DEVICE_1 ------------------------ ||
 * || 0 |      0      |      0     |         0       |   0  |   000    ||
 * || ------------------------ CTRL1_DEVICE_2 ------------------------ ||
 * || 0 |      0      |      0     |         0       |   0  |   001    ||
 * || ------------------------- CTRL1_DW_1 --------------------------- ||
 * || 0 |      0      |      1     |         0       |   0  |   000    ||
 * || ------------------------- CTRL1_DW_1 --------------------------- ||
 * || 0 |      0      |      0     |         0       |   0  |   000    ||
 * || ------------------------ CTRL1_D7_I_1 -------------------------- ||
 * || 0 |      0      |      0     |         1       |   0  |   000    ||
 * || ------------------------ CTRL1_D7_I_0 -------------------------- ||
 * || 0 |      0      |      0     |         0       |   0  |   000    ||
 *  |==================================================================|
 *
 */
#define CTRL1_DOWNLINK             0xA0
#define CTRL1_DEVICE_1             0x00
#define CTRL1_DEVICE_2             0x01
#define CTRL1_BROADCAST            0x06
#define CTRL1_DW_1                 0x20
#define CTRL1_DW_0                 0x00
#define CTRL1_D7_I_1               0x10
#define CTRL1_D7_I_0               0x00



/* initialisation L2 control bytes sent from the phone to devices */
#define L2_CTRL1_INIT_DEV_1        0x90
#define L2_CTRL1_INIT_DEV_2        0x91
#define L2_CTRL1_INIT_DEV_3        0x92
#define L2_CTRL2_INIT_DEV_1_TERM   0x00 /* init dev as a Terminator */
#define L2_CTRL2_INIT_DEV_2_TERM   0x00 /* init dev as a Terminator */
#define L2_CTRL2_INIT_DEV_3_TERM   0x00 /* init dev as a Terminator */
#define L2_CTRL2_INIT_DEV_1_REP    0x20 /* init dev as a Repeater */
#define L2_CTRL2_INIT_DEV_2_REP    0x20 /* init dev as a Repeater */
#define L2_CTRL2_INIT_DEV_3_REP    0x20 /* init dev as a Repeater */


/* devices' responses to L2 initialisation from the phone */
#define L2_INIT_RESP_DEV_1         0x40
#define L2_INIT_RESP_DEV_2         0x41
#define L2_INIT_RESP_DEV_3         0x42

/* control 1 byte to poll for data from devices */
#define L2_CTRL1_POLL_DEV_1        0x88
#define L2_CTRL1_POLL_DEV_2        0x89
#define L2_CTRL1_POLL_DEV_3        0x8A

/* control 1 and 2 bytes used to restart the sidecar bus */
#define L2_CTRL1_RESET             0x86
#define L2_CTRL2_RESET             0x08

/* control 1 and 2 bytes used to restart one sidecar device only */
#define L2_CTRL1_RESET_ONE_DEVICE  0x80
#define L2_CTRL2_RESET_ONE_DEVICE  0x08

/* device can send simple 'message waiting' status byte, or
 * message waiting byte with the 'display bgrd light on' set */
#define L2_MSG_WAITING_DEV_1       0x20
#define L2_MSG_WAITING_DEV_2       0x21
#define L2_MSG_WAITING_DEV_3       0x22
#define L2_MSG_WAITING_DB_DEV_1    0x28
#define L2_MSG_WAITING_DB_DEV_2    0x29
#define L2_MSG_WAITING_DB_DEV_3    0x2A

/* the L2 downlink 'read pacing' byte */
#define L2_READ_PACING             0xC7

/* mask to get length of the message from the 'type and length' uplink
 * byte from the device */
#define L2_UPLINK_LEN_MASK         0x1F

/* in the MessageType/Length byte of L2 message determines
 * whether the L2 message is local or global */
#define L2_GLOB_LOC_MASK           0x80
#define L2_GLOBAL                  0x00
#define L2_LOCAL                   0x80

/* normally, uplink data are processed (sent) to the sidecar bus
 * with higher priority than downlink data; but when there is an
 * neverending stream of uplink data, downlink data are not
 * beeing processed; this define says, that after each
 * DOWN_BEFORE_UP_CYCLE cycles, one-times the downlink data have the
 * advantage
 */
#define DOWN_BEFORE_UP_CYCLE       5

/* used for the fastmode argument sent from userspace;
 * it says for which sidecar device should be the fastmode performed */
#define FASTMODE_DEV_1             0x00
#define FASTMODE_DEV_2             0x80

/* L2 CTRL1 byte for initiating of fastmode; it is ORed with the
 * device number */
#define FASTMODE_REQUEST           0x80
/* L2 CTRL2 byte for initiating of fastmode */
#define FASTMODE_CTRL2             0x00

/* reply from sidecar device: fastmode acknowledgement */
#define FASTMODE_ACK               0x30

/* size of transfered data */
#define FASTMODE_DATA_WHOLE_SIZE   2042
#define FASTMODE_DATA_LABEL_SIZE   194


/* delay between obtaining fastmode ACK and starting the fastmode
 transfer (in micro secs) */
#define FASTMODE_UDELAY            180


#define IOCTL_MAGIC                'k'
#define IOCTL_SYNC                 _IO(IOCTL_MAGIC, 0)
#define IOCTL_FASTMODE             _IO(IOCTL_MAGIC, 1)
#define IOCTL_START                _IO(IOCTL_MAGIC, 2)
#define IOCTL_STOP                 _IO(IOCTL_MAGIC, 3)
#define IOCTL_L2_RESET             _IO(IOCTL_MAGIC, 4)

#define L2_SYNC_CTRL1              CTRL1_BROADCAST | 0x80
#define L2_SYNC_CTRL2              0x10 /* sync device as Terminator */


typedef struct {

  /* it says, if the particular sidecar device is present at the sidecar
   * bus (which means, the device replied to the initialisation
   * command from the phone */
  int device_presence;


  /* The downlink buffer of the sidecar device
   * it is used to store L2 message which is to be send
   * to the sidecar device
   * The L2 message is stored in the form as it will be sent
   * to the sidecar bus:
   * |CTRL 1|byte 1|CTRL 1|byte 2|...|CTRL 1|byte n|
   *
   * H52940: buffer needs no locking - it is only accessed from inside
   * the sidecar timer soft IRQ. Soft IRQs are serialized: timeout function
   * is added after all processing is done.
   */
  unsigned char downlink_buf[MAX_L2_LEN];
  int downlink_len;


  /* the uplink buffer of the sidecar device
   * it is used to store L2 message obtained from the device
   * this message will be converted to sidecar command and
   * written to the output buffer (read buffer)
   * the L2 message is stored in the form as it is obtained from
   * HAPI, without 0xFF bytes:
   * |byte 1|byte 2|...|byte n|
   *
   * H52940: buffer needs no locking - it is only accessed from inside
   * the sidecar timer soft IRQ. Soft IRQs are serialized: timeout function
   * is added after all processing is done.
   */
  unsigned char uplink_buf[MAX_L2_LEN];
  int uplink_len;


  /* buffers for fastmode transfer to sidecar device;
   * these buffers are also shared to userspace using the mmap
   * functionality; buffers are allocated as memory pages */
  unsigned char* fastmode_buf;

} sidecar_module_t;

/* cyclic buffer structure */
typedef struct {

  int head;
  int tail;
  unsigned char buf[];

} sidecar_cbuf_t;

/* structure used to keep the driver's data */
typedef struct {

  /* userspace semaphores */
  struct semaphore rd_sem;
  struct semaphore wr_sem;

  /* semaphore to protect HAPI access */
  struct semaphore hapi_sem;

  /* timer used to run periodically the timer function */
  struct timer_list timer;
  spinlock_t timer_active_lock;
  int timer_active;

  /* the read buffer; a cyclical (flip) buffer */
  spinlock_t read_flip_lock;
  sidecar_cbuf_t *read;
  sidecar_cbuf_t *read_flip;

  /* the write buffer; a cyclical buffer:
   * data start at write_start position of the buffer and are write_len
   * bytes long
   */
  sidecar_cbuf_t *write;

  /* variable used to select, whether the priority have the
   * uplink or downlink sidecar bus operations */
  int down_up;

  /* Sidecar module specific data */
  sidecar_module_t mod[MAX_OPTIONS];

} sidecar_device_t;

static sidecar_device_t sidecar_data;


/* sidecar's driver /proc entry used for informational
 * and debugging output
 */
#ifdef SIDECAR_DEBUG
static struct proc_dir_entry *sidecar_proc;
#endif /* SIDECAR_DEBUG */

/*
 * kernel 'wait queue' on which a process can be queued waiting an event
 */
DECLARE_WAIT_QUEUE_HEAD(wait_queue_sidecar);


/* declarations of the 'open', 'release', 'read' and 'write' functions */
int sidecar_open(struct inode *inode, struct file *filp);
int sidecar_close(struct inode *inode, struct file *filp);
ssize_t sidecar_read(struct file *fp, char *buf, size_t size, loff_t * off);
ssize_t sidecar_write(struct file *filp, const char *buf,
                      size_t count, loff_t * f_pos);
unsigned int sidecar_poll(struct file *fp, struct poll_table_struct *pt);
int sidecar_mmap(struct file *fp, struct vm_area_struct *vma);
int sidecar_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);

/* declaration of the file operations structure, which effectively aliases
 * application functions such as read, write etc. to equivalent
 * kernel function */
struct file_operations sidecar_fops = {
        read:           sidecar_read,  /*read */
        write:          sidecar_write, /* write */
        poll:           sidecar_poll, /* poll */
        open:           sidecar_open,  /* open */
        release:        sidecar_close, /* release */
        mmap:           sidecar_mmap, /* mmap */
        ioctl:          sidecar_ioctl /* ioctl */
};


/* mmap functionality to trasfer bitmap data for fastmode from userspace
 * to kernelspace buffer */
struct page *sidecar_vma_nopage(struct vm_area_struct *vma, unsigned long address, int write_access);

struct vm_operations_struct sidecar_vm_ops = {
        nopage:         sidecar_vma_nopage
};



/* device filesystem handle structure for the creation of device files */
static devfs_handle_t sidecar_devfs_handle;



