/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/*
Sidecar kernel module hardware API
2006

this HAPI uses SSC driver, included file ifx_ssc.h
uses also a fpga.h file
*/

#ifndef _SIDECAR_HAPI_C_
#define _SIDECAR_HAPI_C_

//includes
#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/time.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/tqueue.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <asm-mips/incaip2/ifx_types.h>
#include <asm-mips/irq.h>
#include <asm-mips/incaip2/mux.h>
#include <asm-mips/incaip2/port.h>
#include <asm-mips/incaip2/ifx_ssc.h>
#include "fpga.h"

//only for debugging
//#define SIDECAR_HAPI_DEBUG

#ifdef	SIDECAR_HAPI_DEBUG
#define PD(fmt, args...) printk(KERN_INFO fmt, ## args)
#else
#define	PD(fmt, args...)
#endif

//number of ssc module, which is used for sidecar 
#define SIDECAR_MODULE_NUM	(1)
#define SIDECAR_MODULE_INODE	((struct inode *)SIDECAR_MODULE_NUM)

//control of bus status
#define BUS_NOT_OPENED	0
#define BUS_OPENED	1
unsigned int bus_status = BUS_NOT_OPENED;

//data width for ioctl
#define NORMAL_MODE_DATA_WIDTH	8
#define FAST_MODE_DATA_WIDTH	8

//define baudrate
#define SIDECAR_BAUDRATE_NORMAL_MODE	485000
#define SIDECAR_BAUDRATE_FAST_MODE	1500000

//delay for waiting between controls C1 and C2
//value is in micro seconds	
#define SENDING_DELAY	(((int)(((double)((double)1/(double)(SIDECAR_BAUDRATE_NORMAL_MODE))*1000000)*8))+1)
#define CONTROL_DELAY	182
//delay for SS signal at the init
//value is in micro seconds
#define SS_DELAY	182

/*read buffer*/
#define READ_BUFFER_LEN	256
unsigned char readBuf[READ_BUFFER_LEN]; //Length of buffer have to consult with coleagues.
unsigned int write_index = 0;
//have to remember which byte was written, if it was C1 or C2/D
#define WRITTEN_C1 	0
#define WRITTEN_C2_D	1
unsigned int written_control = WRITTEN_C1;
/*************/

//define bytes which tells sidecars that they have to reset
//first control byte, broadcast to all sidecars
#define C1_BROADCAST	0x86
//reset
#define C2_RESET	0x08
//length of command
#define CMD_LENGTH	2
/**************************************/



/**************************************
Fastmode indication */

unsigned int fastmode_in_progress = 0;

#define FASTMODE_REQUEST_MASK	0xf8
//#define FASTMODE_REQUEST_MASK_2	0x78
#define FASTMODE_REQUEST	0x80
//#define FASTMODE_REQUEST_2	0x00

#define FASTMODE_ACK_MASK	0xf8
#define FASTMODE_ACK		0x30
/**************************************/


//no error
#define OK	0

//define returning error codes
//can't read not opened bus
#define NOT_OPENED_ERROR	5001
//error while writing data to bus
#define IFX_WRITE_ERROR	5002
//error while reading data from bus
#define IFX_READ_ERROR	5003
//error while opening opened bus
#define ALREADY_OPENED_BUS_ERROR	5004
//error while exceeding readBuf length
#define READ_LEN_ERROR	5005
//error while reading more than we have in buffer
#define READ_MORE_WE_HAVE_ERROR	5006
//null pointer
#define BUFFER_NULL_ERROR	5007
//choosing the flush mode error
#define WRONG_MODE_ERROR	5008
//call of write when fastmode in progress
#define FAST_ERROR		5009



//function prototypes
unsigned int SidecarHAPI_Open(void);
unsigned int SidecarHAPI_Close(void);
unsigned int SidecarHAPI_Flush(unsigned int mode);
unsigned int SidecarHAPI_Reset(void);
unsigned int SidecarHAPI_Read(unsigned int length, unsigned char * pRxData);
unsigned int SidecarHAPI_Write(unsigned int length, unsigned char * pTxData);

unsigned int SidecarHAPI_Fastmode_Write(unsigned int length, unsigned char * pTxData);
unsigned char *my_write_buf;

//int ifx_ssc_open(struct inode *, struct file *);
//int ifx_ssc_close(struct inode *, struct file *);
//int ifx_ssc_ioctl(struct inode *, struct file *, unsigned int, unsigned long);



//fatal error mesages
void printFatalError(const char *func,int err);


//debugging functions prototypes
void printHwOpts(struct ifx_ssc_hwopts hwopts);
void printFrmOpts(struct ifx_ssc_frm_opts frm_opts);
void printBaudrate(u32 baud);
//u32 is unsigned int


/******************************************************************************************/
//function declarations
//
/*******************************************
Initialization of The Bus, only after success of this function is bus available for writing and reading
*/
unsigned int SidecarHAPI_Open(void){

    struct ifx_ssc_hwopts hwopts;    
    int err;
    unsigned long baud;
//    unsigned long flags;
//    unsigned int i;

    PD("SidecarHAPI_Open function\n");

#ifdef SIDEACAR_HAPI_DEBUG
    for(i=0;i<1024;i++){
	debug_buffer[i] = 0;
    }
#endif
    
    if(bus_status==BUS_OPENED){
	PD("Sidecar bus error: opening of already opened bus\n");
	return ALREADY_OPENED_BUS_ERROR;    
    }


#if 0
    PD("SidecarHAPI_Open: INITIAL HARDWARE SETUP\n");
    //get the hardware options
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_GET,(unsigned long)&hwopts))){
	printFatalError("SidecarHAPI_Open:1",err);
	return -err;
    }
    //print them, if debug
    printHwOpts(hwopts);
    
    //get the baudrate
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_BAUD_GET,(unsigned long)&baud))){
	printFatalError("SidecarHAPI_Open:2",err);
	return -err;
    }
    //print it, if debug
    printBaudrate(baud);
    
#endif
        
    //open the ssc module
    if((err =  ifx_ssc_open(SIDECAR_MODULE_INODE,NULL))){
	printFatalError("SidecarHAPI_Open:3",err);
	return -err;
    }

    
    //setup the baudrate
    baud = SIDECAR_BAUDRATE_NORMAL_MODE;
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_BAUD_SET,(unsigned long)&baud))){
	printFatalError("SidecarHAPI_Open:4",err);
	return -err;
    }

#if 0    
    //get the baudrate
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_BAUD_GET,(unsigned long)&baud))){
	printFatalError("SidecarHAPI_Open:5",err);
	return -err;
    }
    //print it, if debug
    printBaudrate(baud);
#endif

    //setup normal mode, disable DMA
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_DMA_DISABLE,(unsigned long)NULL))){
	printFatalError("SidecarHAPI_Open:6",err);
	return -err;
    }

//setup initial values
    //first setup a strucure with hw options
    hwopts.AbortErrDetect = 0;
    hwopts.rxOvErrDetect = 0;
    hwopts.rxUndErrDetect = 0;
    hwopts.txOvErrDetect = 0;
    hwopts.txUndErrDetect = 0;
    hwopts.echoMode = 0;
    hwopts.loopBack = 0;
    hwopts.idleValue = 1;
    hwopts.clockPolarity = 1;
    hwopts.clockPhase = 0;
    hwopts.headingControl = IFX_SSC_MSB_FIRST;
    hwopts.dataWidth = NORMAL_MODE_DATA_WIDTH;
    hwopts.masterSelect = IFX_SSC_MASTER_MODE;
    hwopts.modeRxTx = IFX_SSC_MODE_RXTX;	//normaly set to RXTX, this is only for test    
    //setup default values     
    hwopts.gpoCs = IFX_SSC_DEF_GPO_CS;
    hwopts.gpoInv = IFX_SSC_DEF_GPO_INV;
    //disable dma in rxtx mode
    hwopts.dmamode = 0;
    hwopts.slaveCs = 0;

    //then call ioctl
    
//     PD("SidecarHAPI_Open function, hwopts set \n");
    if((err =  ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_SET,(unsigned long)&hwopts))){
	printFatalError("SidecarHAPI_Open:7",err);
	return -err;
    }

#if 0
    hwopts.AbortErrDetect = 0;
    hwopts.rxOvErrDetect = 0;
    hwopts.rxUndErrDetect = 0;
    hwopts.txOvErrDetect = 0;
    hwopts.txUndErrDetect = 0;
    hwopts.echoMode = 0;
    hwopts.loopBack = 0;
    hwopts.idleValue = 0;
    hwopts.clockPolarity = 0;
    hwopts.clockPhase = 0;
    hwopts.headingControl = 0;
    hwopts.dataWidth = 0;
    hwopts.masterSelect = 0;
    hwopts.modeRxTx = 0;
    hwopts.gpoCs = 0;//set SSC1_CS0 as chip select, is it correct?
    hwopts.gpoInv = 0;//no invert
    
    PD("SidecarHAPI_Open: MY HARDWARE SETUP\n");
    //get the hardware options
    
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_GET,(unsigned long)&hwopts))){
	printFatalError("SidecarHAPI_Open:8",err);
	return -err;
    }
    //print them, if debug
    printHwOpts(hwopts);    
#endif

    //initial delay    
    //have to set SS to low,than hi, after this, options starts to listen with next SS low

//    spi_ss_sidecar_set();
    //now we should wait at least 180 usec
//    udelay(SS_DELAY);

    
    //power disable and enable
    sidecar_power( SIDECAR_POWER_OFF );
    mdelay(50);
    sidecar_power( SIDECAR_POWER_ON );

    spi_ss_sidecar_clear();        
    PD("SS up\n");
    udelay(SS_DELAY);

    //now we finished the initialization of the bus
    bus_status = BUS_OPENED;
    
    
    //no error
//    PD("SidcarHAPI_Open: returns OK,56\n");
            
    return OK;
}



/*****************************************************
Close the bus, after closing it is no more available for any action
*/
unsigned int SidecarHAPI_Close(void){

    int err;

    PD("SidecarHAPI_Close function\n");

    /* turn sidecar power off */
    sidecar_power( SIDECAR_POWER_OFF );

    if(bus_status==BUS_OPENED){
	
	bus_status = BUS_NOT_OPENED;
	
	//free the interrupt
//	free_irq(154,NULL);
	
	//call the ifx_ssc_close to do it's job
	if((err = ifx_ssc_close(SIDECAR_MODULE_INODE,NULL))){
	    printFatalError("SidecarHAPI_Close",err);
	    return -err;
	}
    }
    else{
	PD("Sidecar bus error: closing of not opened bus\n");
	return NOT_OPENED_ERROR;
    }
    return OK;
}


/******************************************************
Flush internal buffers and fifos
*/
unsigned int SidecarHAPI_Flush(unsigned int mode){

    int err;
    
    PD("HAPIFlush:%d\n",mode);
    
    if(bus_status==BUS_NOT_OPENED){
	PD("Sidecar bus error: flush of not opened bus\n");
	return NOT_OPENED_ERROR;
    }
    
    switch(mode){
    case 0:
	//HAPI buffer
	//flush read buffer
    	write_index = 0;
	
	//ifx driver buffer
	
	//ssc module RX fifo
	
    break;
    case 1://flush write buffer
	//this is question, what we should do? writing is done in one byte bursts, at first write one byte
	//and immediately after write, read one byte that means we don't have any write buffer, and also we read
	//received byte each time, only buffer for writing is that we obtain from caller, but we don't care about it
		//HAPI buffer
	//no buffer
	
	//ifx write buffer
	
	//ssc module TX fifo
	//problem will be fastmode
	written_control = WRITTEN_C1;
	
    break;
    case 2://flush both buffers
	//HAPI buffers
	//first read buffer
	write_index = 0;
	//write buffer
	written_control = WRITTEN_C1;
	//ifx driver buffers
	
	//ssc module fifos
	//both fifos at ssc module
	if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_FIFO_FLUSH,(unsigned long)NULL))){
	    printFatalError("SidecarHAPI_Flush",err);
	    return -err;
	}
    break;
    default:
	PD("SidecarHAPI_Flush function: error wrong mode in parameter\n");
	return WRONG_MODE_ERROR;
    break;
    }

    return OK;
}


/***************************************************
Reset HAPI internal buffers and varialbles
*/
unsigned int SidecarHAPI_Reset(void){

    int err;
    struct ifx_ssc_hwopts hwopts;    
    unsigned long baud;
    
    //prepare a buffer for sending a reset
    unsigned char cmd[2];


//    PD("HAPIReset\n");
    if(bus_status==BUS_NOT_OPENED){
	PD("Sidecar bus error: reset of not opened bus\n");
	return NOT_OPENED_ERROR;
    }
    
    /*some reset operations*/

    //abort the current transmission
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_ABORT,(unsigned long)NULL))){
	printFatalError("SidecarHAPI_Reset",err);
	return -err;
    }

    //we should Flush all of buffers
    if(SidecarHAPI_Flush(2)){
	return 1;
    }

    //point to the beginning of readBuf
    write_index = 0;

    //wait for C1
    written_control = WRITTEN_C1;
    
    //setup normal mode, disable DMA
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_DMA_DISABLE,(unsigned long)NULL))){
	printFatalError("SidecarHAPI_Reset",err);
	return -err;
    }
    
    //first setup a strucure with hw options
    hwopts.AbortErrDetect = 0;
    hwopts.rxOvErrDetect = 0;
    hwopts.rxUndErrDetect = 0;
    hwopts.txOvErrDetect = 0;
    hwopts.txUndErrDetect = 0;
    hwopts.echoMode = 0;
    hwopts.loopBack = 0;
    hwopts.idleValue = 1;
    hwopts.clockPolarity = 1;
    hwopts.clockPhase = 0;
    hwopts.headingControl = IFX_SSC_MSB_FIRST;
    hwopts.dataWidth = NORMAL_MODE_DATA_WIDTH;
    hwopts.masterSelect = IFX_SSC_MASTER_MODE;
    hwopts.modeRxTx = IFX_SSC_MODE_RXTX;
    //setup default values     
    hwopts.gpoCs = IFX_SSC_DEF_GPO_CS;
    hwopts.gpoInv = IFX_SSC_DEF_GPO_INV;
    //disable dma in rxtx mode
    hwopts.dmamode = 0;
    hwopts.slaveCs = 0;

    //then call ioctl
//     PD("SidecarHAPI_Reset function, hwopts set\n");
    
    if((err =  ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_SET,(unsigned long)&hwopts))){
	printFatalError("SidecarHAPI_Reset",err);
	return -err;
    }
    
    //setup the baudrate
    baud = SIDECAR_BAUDRATE_NORMAL_MODE;
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_BAUD_SET,(unsigned long)&baud))){
	printFatalError("SidecarHAPI_Reset",err);
	return -err;
    }
/*
    //before start sending data, have to do one low to high transition 
    //sidecars starts to listen only after that transition
    spi_ss_sidecar_set();
    //wait for some us
    udelay(SS_DELAY);
    spi_ss_sidecar_clear();
    udelay(SS_DELAY);
*/    

    /*send the reset command to sidecars*/    
    //send to sidecars information of reset
    cmd[0] = C1_BROADCAST;
    cmd[1] = C2_RESET;
    //write reset command
    if((err = SidecarHAPI_Write(CMD_LENGTH,cmd))){
	PD("SidecarHAPI_Reset: error while sending reset command to sidecars\n");
	return err;
    }
    
    //and flush the buffers
    if((err = SidecarHAPI_Flush(2))){
	PD("SidecarHAPI_Reset: error while flushing buffers\n");
	return err;
    }
    
    return OK;
}


/*********************************************************
Reading from HAPI internal read buffer
*/
unsigned int SidecarHAPI_Read(unsigned int length, unsigned char * pRxData){

    int index;

//    PD("HAPIRead\n");
    //check if the bus is opened
    if(bus_status==BUS_NOT_OPENED){
	PD("Sidecar bus error: read of not opened bus\n");
	return NOT_OPENED_ERROR;
    }
    //check if the pointer to buffer is not NULL
    if(pRxData==NULL){
	PD("SidecarHAPI_Read function, error NULL pointer to buffer in second parameter\n");
	return BUFFER_NULL_ERROR;
    }
    //check the length of read data
    if(length>READ_BUFFER_LEN){
	PD("SidecarHAPI_Read function, error length of read data exceeds internal buffer size \n");
	return READ_LEN_ERROR;
    }
    //check if the reader reads more than we have in bufer
    if(write_index<length){
	PD("SidecarHAPI_Read function, error length is more than we have in buffer\n");
	return READ_MORE_WE_HAVE_ERROR;
    }    
    //only thing we do is copy from internal buffer to given buffer
    for(index=0;index<length;index++){
	pRxData[index] = readBuf[index];
    }
    //release the space
    if(length<write_index){
	//copy the remaining data in buffer to the beginig of readBuf
	for(index=0;index<(write_index-length);index++){
	    readBuf[index] = readBuf[index+length];
	}
	//corect the write_index in readBuf
	write_index = write_index - length;
    }
    else
    {
	// all the data was read, write_index points to beginnig of readBuf
	write_index = 0;
    }
//    PD("i:%d\n",write_index);
    return OK;
}


/************************************************************************************
This function writes bytes to the Tx buffer. We also have to read bytes from sidecar to read buffer.
*/
unsigned int SidecarHAPI_Write(unsigned int length, unsigned char * pTxData){

    ssize_t written;
    ssize_t read;
    unsigned int index;
    struct timeval tv1,tv2;
    long rwdelay;
    unsigned int fastmode_request = 0;
    unsigned char written_value;


    if(bus_status==BUS_NOT_OPENED){
	PD("Sidecar bus error: writing onto not opened bus\n");
	return NOT_OPENED_ERROR;
    }
    
    if(pTxData==NULL){
	PD("SidecarHAPI_Write function: error NULL pointer to buffer in second parameter \n");
	return BUFFER_NULL_ERROR;
    }

    //test if the fastmode in progress
    if(fastmode_in_progress){
	PD("SidecarHAPI_Write: conflict write, fastmode in progress\n");
	return FAST_ERROR;
    }
        
    //make an own buffer for data
    unsigned char *myWriteBuf;
    unsigned char myReadBuf[1];
    
    //don't know who is calling this function, so write the data at one byte blocks
    //and also read one byte after each written byte
    for(index=0;index<length;index++){
	
	//one byte buffer for write,have to alloc, because kwite free's it
	if((myWriteBuf = (unsigned char*)kmalloc((sizeof(unsigned char)),GFP_ATOMIC))==NULL){
	    PD("SidecarHAPI_Write: error kmalloc returns NULL pointer\n");
    	    return BUFFER_NULL_ERROR;
	}
	
	//copy one byte
	memcpy((void*)myWriteBuf,(void*)((unsigned char*)(pTxData+index)),(sizeof(unsigned char)));
	written_value = myWriteBuf[0];
	
	//here we have to lower the SS* signal, set to active
	spi_ss_sidecar_set();

	do_gettimeofday(&tv1);

	/********     Write/Read section   ********/
	//do the write
	if((written = ifx_ssc_kwrite(SIDECAR_MODULE_NUM,myWriteBuf,(ssize_t)(sizeof(unsigned char))))!=(sizeof(unsigned char))){
	    if(written>=0){
		PD("SidecarHAPI_Write function: kwrite returns different number of written bytes: %d\n",written);
		return IFX_WRITE_ERROR;
	    }
	    if(written<0){
		//when writing, the port should be busy
		printFatalError("SidecarHAPI_Write",written);
		return -written;
	    }
	}
	
	//sending delay

	/*  it waits for data, so the SidecarHAPI_Write doesn't continue	*/
	//after write, we have to read one received byte
	if((read = ifx_ssc_kread(SIDECAR_MODULE_NUM,myReadBuf,(ssize_t)(sizeof(unsigned char))))!=(sizeof(unsigned char))){
	    if(read>=0){
		PD("SidecarHAPI_Write function: kread returns different number of read bytes: %d\n",read);
		return IFX_READ_ERROR;
	    }
	    if(read<0){
		printFatalError("SidecarHAPI_Write",read);
		return -read;
	    }
	}
	
//	PD("SidecarHAPI_Write: data read\n");
	/******************************************/

	do_gettimeofday(&tv2);
	
	if((rwdelay=(tv2.tv_usec-tv1.tv_usec))<SENDING_DELAY){
	    udelay(SENDING_DELAY-rwdelay);
	}

	//special case is fastmode, if not fastmode raise the SS
	if(!fastmode_request){
	    spi_ss_sidecar_clear();
//	    PD("SS up\n");
	}

	//check the write buffer, if it was released
//	if(myWriteBuf!=NULL){
//	    kfree((void*)myWriteBuf);
//	}
		
	//make a decision if the received byte store or not
	if(written_control==WRITTEN_C1){
	    //don't store received byte, only change state
	    written_control = WRITTEN_C2_D;

	    //check if this sent byte was fastmode request	    
	    if(((written_value & FASTMODE_REQUEST_MASK)==(FASTMODE_REQUEST))){
		fastmode_request = 1;
//		PD("fm req\n");
	    }
	    
	    //transmit and receive is complete, so wait for delay between two bytes
	    udelay(CONTROL_DELAY);
	}
	else
	{
	    //change state
	    written_control = WRITTEN_C1;
	    
	    //check if the write_index is not exceeding the length of buffer
	    if(write_index<READ_BUFFER_LEN){
		//store
		readBuf[write_index++] = myReadBuf[0];
//		PD("i:%d\n",write_index);
	    }
	    else{
		PD("SidecarHAPI_Write function: writing index exceeds the length of buffer\n");
	    }
	    
	    //check the returned data byte
	    //if the fastmode was rejected, wait for 180us,than set SS wait 180, clear SS
	    if(fastmode_request){
		
		if((myReadBuf[0] & FASTMODE_ACK_MASK)!=FASTMODE_ACK){
		    //wait
		    udelay(CONTROL_DELAY);
		    //raise SS
		    spi_ss_sidecar_clear();

		    //clear fastmode request flag
		    fastmode_request = 0;
//			PD("fm rej, SS up\n");
		    udelay(CONTROL_DELAY);
		}
		else{
		    //fastmode was acknowledged, wait for 180us
//		    if(fastmode_request) PD("fm ack\n");
		    udelay(CONTROL_DELAY);
		}
	    }
	    else{
		spi_ss_sidecar_clear();
		udelay(CONTROL_DELAY);
	    }
	}
    }
    
    return OK;
}



/*******************************************************
Fastmode write
*/
unsigned int SidecarHAPI_Fastmode_Write(unsigned int length, unsigned char * pTxData){

                                                                                                                                              
    int err;
    unsigned long baud;
    struct ifx_ssc_hwopts hwopts;
    ssize_t written;
                                                                                                                                          
//    PD("SidecarHAPI_Fastmode_Write function\n");
                                                                                                                                          
//    printk("X");
                                                                                                                                          
    if(bus_status==BUS_NOT_OPENED){
        PD("SidecarHAPI_Fastmode_Write: error writing to not opened bus\n");
        return NOT_OPENED_ERROR;
    }
                                                                                                                                          
    //check the length, we have 500ms to finish the transmission
//!!!!!!!!!!!!
                                                                                                                                          
    if(pTxData==NULL){
        PD("SidecarHAPI_Fastmode_Write: error NULL pointer to buffer in second parameter\n");
        return BUFFER_NULL_ERROR;
    }
                                                                                                                                          
    //switch to fastmode
    if(fastmode_in_progress){
        PD("SidecarHAPI_Fastmode_Write: fastmode in progress\n");
        return 1;
    }

    //set fastmode to "in progress"
    fastmode_in_progress = 1;
                                                                                                                                          
//    PD("SidecarHAPI_Fastmode_Write function, before kmalloc\n");
                                                                                                                                          
    if((my_write_buf = (unsigned char*)kmalloc(length,__GFP_DMA|GFP_ATOMIC))==NULL){
        PD("SidecarHAPI_Fastmode_Write: error kmalloc returns NULL pointer\n");
        return BUFFER_NULL_ERROR;
    }
//    PD("SidecarHAPI_Fastmode_Write function, after kmalloc\n");
                                                                                                                                          
                                                                                                                                          
    memcpy((void*)my_write_buf,(void*)((unsigned char*)(pTxData)),(length));
                                                                                                                                          
                                                                                                                                          
//    PD("SidecarHAPI_Fastmode_Write function, after memcpy\n");
                                                                                                                                          
    //setup the baudrate
    baud = SIDECAR_BAUDRATE_FAST_MODE;
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_BAUD_SET,(unsigned long)&baud))){
        printFatalError("Fastmode_Write baud 1",err);
        return -err;
    }
//    PD("SidecarHAPI_Fastmode_Write function, baudrate set\n");
                                                                                                                                          
                                                                                                                                              
    //set tx mode,data width, dma
    //setup a strucure with hw options
    hwopts.AbortErrDetect = 0;
    hwopts.rxOvErrDetect = 0;
    hwopts.rxUndErrDetect = 0;
    hwopts.txOvErrDetect = 0;
    hwopts.txUndErrDetect = 0;
    hwopts.echoMode = 0;
    hwopts.loopBack = 0;
    hwopts.idleValue = 1;
    hwopts.clockPolarity = 1;
    hwopts.clockPhase = 0;
    hwopts.headingControl = IFX_SSC_MSB_FIRST;
    hwopts.dataWidth = FAST_MODE_DATA_WIDTH;
    hwopts.masterSelect = IFX_SSC_MASTER_MODE;
    hwopts.modeRxTx = IFX_SSC_MODE_TX;
    //setup default values
    hwopts.gpoCs = IFX_SSC_DEF_GPO_CS;
    hwopts.gpoInv = IFX_SSC_DEF_GPO_INV;
    //enable dma in for only tx mode
    hwopts.dmamode = 1;
    hwopts.slaveCs = 0;
                                                                                                                                          
    //then call ioctl
                                                                                                                                          
//     PD("SidecarHAPI_Fastmode_Write function, hwopts set 1\n");
    if((err =  ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_SET,(unsigned long)&hwopts))){
        printFatalError("Fastmode_Write hwopts 1",err);
        return -err;
    }
//    PD("SidecarHAPI_Fastmode_Write function, hwopts set\n");
                                                                                                                                          
    //enable dma
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_DMA_ENABLE,(unsigned long)NULL))){
        printFatalError("Fastmode_Write dma enable",err);
        return -err;
    }
//    PD("SidecarHAPI_Fastmode_Write function, dma enabled\n");
	                                                                                                                                          
    //set slave select
    //    spi_ss_sidecar_set();
    //    udelay(CONTROL_DELAY);
	    //call kwrite
    if((written = ifx_ssc_kwrite(SIDECAR_MODULE_NUM,(const char*)my_write_buf,(ssize_t)(length)))!=(length)){
        if(written>=0){
            PD("SidecarHAPI_Fastmode_Write: kwrite returns different number of written bytes: %d\n",written);
            return IFX_WRITE_ERROR;
        }
        if(written<0){
            //when writing, the port should be busy
            printFatalError("Fastmode_Write kwrite",written);
            return -written;
        }
    }
                                                                                                                                          
//    PD("SHFW,kwrite\n");
                                                                                                                                          
    //now go to sleep, wake after the end of transmission awakes
    sidecar_ifx_wait_sleep();
//    mdelay(12);
    udelay(50);
                                                                                                                                          
    //set back ssc1 to normal mode
    //test if the fastmode was started
    if(!fastmode_in_progress){
        return 1;
    }
    //fastmode is not in progres, now
    fastmode_in_progress = 0;
                                                                                                                                          
                                                                                                                                          
    //disable DMA
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_DMA_DISABLE,(unsigned long)NULL))){
        printFatalError("Fastmode_Write dma disable",err);
        return -err;
    }
                                                                                                                                          
                                                                                                                                              
    //setup a strucure with hw options
    hwopts.AbortErrDetect = 0;
    hwopts.rxOvErrDetect = 0;
    hwopts.rxUndErrDetect = 0;
    hwopts.txOvErrDetect = 0;
    hwopts.txUndErrDetect = 0;
    hwopts.echoMode = 0;
    hwopts.loopBack = 0;
    hwopts.idleValue = 1;
    hwopts.clockPolarity = 1;
    hwopts.clockPhase = 0;
    hwopts.headingControl = IFX_SSC_MSB_FIRST;
    hwopts.dataWidth = NORMAL_MODE_DATA_WIDTH;
    hwopts.masterSelect = IFX_SSC_MASTER_MODE;
    hwopts.modeRxTx = IFX_SSC_MODE_RXTX;
    //setup default values
    hwopts.gpoCs = IFX_SSC_DEF_GPO_CS;
    hwopts.gpoInv = IFX_SSC_DEF_GPO_INV;
    //disable dma in rxtx mode
    hwopts.dmamode = 0;
    hwopts.slaveCs = 0;
                                                                                                                                          
#if 0
    //then call ioctl
    if((err =  ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_SET,(unsigned long)&hwopts))){
        printFatalError("Fastmode_Write hwopts 2",err);
        return -err;
    }
#endif
//    PD("SidecarHAPI_Fastmode_Write function, hwopts set 2 \n");
                                                                                                                                          
    while(ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_HWOPTS_SET,(unsigned long)&hwopts)){
        udelay(200);
        PD("SidecarHAPI_Fastmode_Write, hwopts try\n");
    }
                                                                                                                                          
                                                                                                                                          
    //setup the baudrate
    baud = SIDECAR_BAUDRATE_NORMAL_MODE;
    if((err = ifx_ssc_ioctl(SIDECAR_MODULE_INODE,NULL,IFX_SSC_BAUD_SET,(unsigned long)&baud))){
        printFatalError("Fastmode_Write baud 2",err);
        return -err;
    }
                                                                                                                                          
                                                                                                                                          
    //and clear the slave select
    spi_ss_sidecar_clear();
//    PD("SS up\n");
                                                                                                                                          

    //check if the write bufer was released
//    if(my_write_buf!=NULL){
//	kfree((void*)my_write_buf);
//    }
                                                                                                                                          
    //have to wait for at least 180us
    udelay(CONTROL_DELAY);
                                                                                                                                          
//    PD("SHFW,finished\n");
#ifdef SIDECAR_HAPI_DEBUG
    PD("readBuf: ");
    for(i=0;i<write_index;i++){
        PD("%x ",readBuf[i]);
    }
    PD("\n");
#endif
//    printk("X\n");
                                                                                                                                          
    return OK;
        		                                                                                                                                             			    
}


/**************************************************************************************/
/**************************************************************************************/

//fatal error messages
void printFatalError(const char *func, int err){
    
    switch(err){
	case -ENOIOCTLCMD:
	    printk(KERN_ERR "ERROR: %s,  error: ENOIOCTLCMD\n",func);
	break;
	case -ENXIO:
	    printk(KERN_ERR "ERROR: %s,  error: ENXIO\n",func);
	break;
	case -EBUSY:
	    printk(KERN_ERR "ERROR: %s,  error: EBUSY\n",func);
	break;
	case -EFAULT:
	    printk(KERN_ERR "ERROR: %s,  error: EFAULT\n",func);
	break;
	case -EIO:
	    printk(KERN_ERR "ERROR: %s,  error: EIO\n",func);
	break;
	case -EINVAL:
	    printk(KERN_ERR "ERROR: %s,  error: EINVAL\n",func);
	break;
	default:
	    printk(KERN_ERR "ERROR: %s,  error: Unknown error\n",func);
	break;
    }
}

//debugging functions
void printHwOpts(struct ifx_ssc_hwopts hwopts){

    PD("Hardware options\n");
    PD("hwopts.AbortErrDetect: %x\n",hwopts.AbortErrDetect);
    PD("hwopts.rxOvErrDetect: %x\n",hwopts.rxOvErrDetect);
    PD("hwopts.rxUndErrDetect: %x\n",hwopts.rxUndErrDetect);
    PD("hwopts.txUndErrDetect: %x\n",hwopts.txOvErrDetect);
    PD("hwopts.txUndErrDetect: %x\n",hwopts.txUndErrDetect);
    PD("hwopts.echoMode: %x\n",hwopts.echoMode);
    PD("hwopts.loopBack: %x\n",hwopts.loopBack);
    PD("hwopts.idleValue: %x\n",hwopts.idleValue);
    PD("hwopts.clockPolarity: %x\n",hwopts.clockPolarity);
    PD("hwopts.clockPhase: %x\n",hwopts.clockPhase);
    PD("hwopts.headingControl: %x\n",hwopts.headingControl);
    PD("hwopts.dataWidth: %x\n",hwopts.dataWidth);
    PD("hwopts.masterSelect: %x\n",hwopts.masterSelect);
    PD("hwopts.modeRxTx: %x\n",hwopts.modeRxTx);
    PD("hwopts.gpoCs: %x\n",hwopts.gpoCs);
    PD("hwopts.gpoInv: %x\n",hwopts.gpoInv);
    PD("hwopts.dmamode: %x\n",hwopts.dmamode);
    PD("hwopts.slaveCs: %x\n",hwopts.slaveCs);
    PD("\n");

}

void printFrmOpts(struct ifx_ssc_frm_opts frm_opts){
    
    PD("Framing options\n");
    PD("frm_opts.FrameEnable: %x\n",frm_opts.FrameEnable);
    PD("frm_opts.IntBefore: %x\n",frm_opts.IntBefore);
    PD("frm_opts.IntAfter: %x\n",frm_opts.IntAfter);
    PD("frm_opts.DataLength: %x\n",frm_opts.DataLength);
    PD("frm_opts.PauseLength: %x\n",frm_opts.PauseLength);
    PD("frm_opts.IdleData: %x\n",frm_opts.IdleData);
    PD("frm_opts.StopAfterPause: %x\n",frm_opts.StopAfterPause);
    PD("\n");
}

void printBaudrate(u32 baud){

    PD("Baudrate\n");
    PD("baudrate: %d\n",baud);
    PD("\n");
}

#endif
