/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>

#include <linux/pci.h>
#include <video/fbcon.h> 
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb32.h>

#include "st7529fb.h"
#include "st7529fb_ioctl.h"
#include "fpga.h"


//----------------------------------------------------------------------------- 
// Include C functions for worker thread for the update of video memory
//----------------------------------------------------------------------------- 
#ifdef ST7529FB_USE_UPDATE_THREAD
#include "st7529Thread.c"
#endif

#define GRAY_2B3P 0x1
#define GRAY_3B3P 0x2
#define DATSDR_PPB GRAY_2B3P

//----------------------------------------------------------------------------- 
//
// Sanity checks
//
//----------------------------------------------------------------------------- 

// Check if at least one supported colordepth was configured

#if !defined (FBCON_HAS_MFB)
    #error Kernel does not support Mononchrome
#endif



//-----------------------------------------------------------------------------
//
// Local Definitions
// 
//---------------------------------------------------------------------------


typedef struct{
	struct fb_info_gen gen;
	unsigned int vidMemSize;	/* Video mem size in bytes */
	unsigned char *virtVidMem;	/* A virtual copy of the phys mem (Page Aligned) */
	unsigned char *virtVidMemAllocation;	/* A virtual copy of the phys mem */
	unsigned char *physVidMemAllocation;	/* Copy of what is in real video memory */
	unsigned char *physVidCopy;	/* Copy of what is in real video memory */
	volatile unsigned char *regBase;	/* Physical register base */
	unsigned int wpType;		/* Is workpoint 1 as opposed to 2 or ??? */

	/* Default limits for WP columns and lines */
	unsigned int startColumn;	/* default CASET Column Address set info */ 	
	unsigned int endColumn;
	unsigned int startLine;		/* defualt LASET Line Address set info */
	unsigned int endLine;
}FB_INFO_st7529;


//-----------------------------------------------------------------------------
//
// Forward Definitions
// 
//-----------------------------------------------------------------------------

static struct fb_ops st7529fb_ops;
static struct fbgen_hwswitch st7529_hwswitch;
static int isWp1 = 0;;

//-----------------------------------------------------------------------------
//
// Function Prototypes 
// 
//-----------------------------------------------------------------------------
extern int st7529ThreadCreate( void );
void st7529_SyncVirtToPhys( void *ptr );

//----------------------------------------------------------------------------- 
//
// Local globals
// 
//----------------------------------------------------------------------------- 


static FB_INFO_st7529 fb_info;
static struct display disp;  
static char __initdata default_fontname[40] = { 0 }; 



/* When/if we have our own ioctls. */
static int st7529_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	      unsigned long arg, int con, struct fb_info *info)
{
	unsigned int upper,lower;

	switch (cmd) {
		case ST7529FB_SET_CONTRAST:

	       //printk( KERN_INFO "st7529fb_ioctl: called with arg %d\n",arg);
	       
	       upper = (arg & 0x000001C0) >> 6;
	       lower = arg & 0x0000003F;
	       
	       //printk(KERN_INFO "lower 0x%x upper 0x%x\n",lower,upper);
	       
      
			WR_CMD( VOLCTRL );	/* Electronic Volume Control */
			WR_DAT( lower );	
			WR_DAT( upper );

			return 0;
			break;

		case ST7529FB_SYNC_SCREEN:
	       	//printk( KERN_INFO "st7529fb_ioctl: SYNC_SCREEN\n");
			st7529_SyncVirtToPhys( 0 );
			return 0 ;
			break;

		default:
			return -ENOIOCTLCMD;
			break;


	}
}


/* Structure to be dynamically initialised depending on workpoint type */
static struct 
{
	unsigned int width;
	unsigned int height;
	unsigned int scan_lines;
	unsigned int panelSize;
}displaySize;


static int st7529_mmap( struct fb_info *fb, struct file *file, struct vm_area_struct *vma )
{
	vma->vm_flags |= VM_RESERVED;
	if( remap_page_range( vma->vm_start, virt_to_phys( (void *)fb_info.virtVidMem ),
				vma->vm_end - vma->vm_start,vma->vm_page_prot ) ){
		return( -EINVAL );
	}
	return( 0 );
}


#if DATSDR_PPB == GRAY_2B3P

//#error "2B3P"

static const unsigned char pixToByte[8][2] = {
		{0x00,0x00}, /* 000 */
		{0x00,0x1f}, /* 001 */
		{0x07,0xe0}, /* 010 */
		{0x07,0xff}, /* 011 */
		{0xf8,0x00}, /* 100 */
		{0xf8,0x1f}, /* 101 */
		{0xff,0xe0}, /* 110 */
		{0xff,0xff} /* 111 */
};



void st7529_SyncVirtToPhys2B3P( int full_refresh )
{
	/* Sync data from our virtual framebuffer to the real HW framebuffer */
	/* Called from a worker thread to syncronise virual and real framebuffer */

	/* Make the virtual framebuffer linear addresed; perform any transformations */
	/* on copy to physical buffer */

	unsigned int line;
	unsigned char *vp, pixbuff;
	int i;
	int displayColumn;
	int skipping;
	int bytesPerLine;
	unsigned char *vc = fb_info.physVidCopy;
	unsigned long flags;
	u32 dirtyLine;
	
	vp = fb_info.virtVidMem;
	
	WR_CMD( EXTIN );	/* Select Command Table 0 */

	WR_CMD( CASET );	/* Column Address */

	WR_DAT( fb_info.startColumn );
	WR_DAT( fb_info.endColumn );

	skipping = 1;

	bytesPerLine = displaySize.width >> 3;

//	local_irq_save(flags);
	
	for( line = 0; line < displaySize.scan_lines; line++ )
	{
		int pixcount = 0;
		pixbuff = 0;

		if( (dirtyLine = memcmp( vc, vp, bytesPerLine )) != 0 )
		{
			memcpy( vc, vp, bytesPerLine );
		}			

		dirtyLine |= full_refresh;

		if( dirtyLine )
		{
			if( skipping != 0 )
			{
				WR_CMD( LASET );	/* Line Address */
				WR_DAT( fb_info.startLine+line );
				WR_DAT( fb_info.endLine );
				skipping = 0;
				WR_CMD( RAMWR );
			}

			for( displayColumn = 0; displayColumn < displaySize.width/3; displayColumn++ )
			{
				int pix3 = 0;
	
				/* read three pixels */
				for( i=0;i<3;i++)
				{
					pix3 <<= 1;
					if( (pixcount & 0x7) == 0 )
					{
						pixbuff = *vp++;
					}
					pix3 |= (pixbuff & 0x80) >> 7;
					pixbuff <<= 1;
					pixcount++;
				}
				
				WR_DAT( pixToByte[pix3][0] );
				WR_DAT( pixToByte[pix3][1] );
			}
		}
		else
		{
			skipping++;
			vp += bytesPerLine;
		}
		vc += bytesPerLine;
	} 
//    local_irq_restore(flags);
}

void st7529_SyncVirtToPhys( void *ptr )
{
	static u32 lastTime;
	u32 fullRefresh = 0;
	if( (jiffies - lastTime) > (HZ * 10) )
	{
		lastTime = jiffies;
		fullRefresh = 1;
	}
	st7529_SyncVirtToPhys2B3P( fullRefresh );
}

#else

#error "3B3P"

void st7529_SyncVirtToPhys3B3P( )
{
	/* Sync data from our virtual framebuffer to the real HW framebuffer */
	/* Called from a worker thread to syncronise virual and real framebuffer */

	/* Make the virtual framebuffer linear addresed; perform any transformations */
	/* on copy to physical buffer */

	unsigned int line, currLine, currColumn;
	unsigned int pixGrpIndx, pix, line_len_bytes;
	unsigned char *vp, pixCol, pixGrp;
	unsigned int linePix;

	currLine = 1000;
	currColumn = 1000;

	/* Start simply blindly copy from Virtual to Video */
	/* Don't bother about mapping initially.....*/
	WR_CMD( EXTIN );	/* Select Command Table 0 */

	WR_CMD( LASET );	/* Line Address */

	WR_DAT( fb_info.startLine );
	WR_DAT( fb_info.endLine );

	WR_CMD( CASET );	/* Column Address */

	WR_DAT( fb_info.startColumn );
	WR_DAT( fb_info.endColumn );

	vp = fb_info.virtVidMem;
	WR_CMD( RAMWR );


	line_len_bytes = displaySize.width/8;	

	
	for( line = 0; line < displaySize.scan_lines; line++ ){
		linePix = 0;
		if( currLine != line ){

			/* Set line Address...*/
		}
		for( pixGrpIndx = 0; pixGrpIndx < line_len_bytes; pixGrpIndx++ ){
			pixGrp = *vp++;	/* Get next 8 pixels */
			pix = 8;	/* Do 8 pixels */
			while( pix-- ){
				linePix++;			/* Count pix index in line */
				if( isWp1 && (linePix == 207) ){
					/* WP line length is not integral no of bytes wide */
					linePix = 0;
					continue;
				}
				pixCol = (pixGrp & 0x80) ? 0xff: 0x00;
			 	WR_DAT(  pixCol ); 
				pixGrp = pixGrp << 1;
			}
		}
	} 
}

void st7529_SyncVirtToPhys( void *ptr )
{
	st7529_SyncVirtToPhys3B3P();
}

#endif


//----------------------------------------------------------------------------- 
// 
// This function should detect the current video mode settings and store  
// it as the default video mode 
//  
//----------------------------------------------------------------------------- 
static void st7529_detect(void) 
{ 
        //printk( KERN_INFO "st7529fb_detect: called\n");
} 


//----------------------------------------------------------------------------- 
// 
// This function should fill in the 'fix' structure based on the values  
// in the `par' structure.
//  
//----------------------------------------------------------------------------- 

static int st7529_encode_fix(struct fb_fix_screeninfo *fix, const void *par, 
                struct fb_info_gen *info) 
{
    strcpy(fix->id, fb_info.gen.info.modename);

    fix->type = FB_TYPE_PACKED_PIXELS;  /* see FB_TYPE_* in fb.h    */       
    fix->type_aux = 0;                  /* Interleave for interleaved Planes */

    /* set fix->line_length to the length of the line in bytes */
    fix->line_length = displaySize.width/8; /* 1BPP */     

    fix->visual = FB_VISUAL_MONO10;

    /* Start and length of Memory Mapped I/O */
    fix->mmio_start = (u32)fb_info.regBase;;
    fix->mmio_len   = ST7529_PHYSICAL_REG_SIZE;

    /* Start and length of frame buffer mem */
    fix->smem_start = (u32)virt_to_phys((void *)fb_info.virtVidMem);
    fix->smem_len   = fb_info.vidMemSize;

    fix->xpanstep = fix->ypanstep = fix->ywrapstep = 0;
    fix->accel = FB_ACCEL_NONE;

    return 0;
} 
 
//----------------------------------------------------------------------------- 
// 
//  Get the video params out of 'var'. If a value doesn't fit, round it up,  
//  if it's too big, return -EINVAL.
//  
//  Suggestion: Round up in the following order: bits_per_pixel, xres,  
//  yres, xres_virtual, yres_virtual, xoffset, yoffset, grayscale,  
//  bitfields, horizontal timing, vertical timing.  
//  
//----------------------------------------------------------------------------- 
static int st7529_decode_var(const struct fb_var_screeninfo *var, void *fb_par, 
                struct fb_info_gen *info) 
{ 
    return -EINVAL;		/* We don't support any programmable config */ 
} 

//-----------------------------------------------------------------------------
//
// Fill the 'var' structure based on the values in 'par' and maybe other 
// values read out of the hardware.
// 
//----------------------------------------------------------------------------- 
static int st7529_encode_var(struct fb_var_screeninfo *var, const void *fb_par,
                struct fb_info_gen *info)
{
    u32 size = displaySize.panelSize;
    unsigned char *vp;
    unsigned char shade;


    vp = fb_info.virtVidMem;
    shade = 0x00;
    size = fb_info.vidMemSize;

    do {
	*vp++ = shade;
    }while(--size);



    st7529_SyncVirtToPhys( 0 );


    var->xres =  displaySize.width;     //visible resolution
    var->yres =  displaySize.height;

    var->xres_virtual = var->xres;
    var->yres_virtual = var->yres;
    var->xoffset = var->yoffset = 0;
    var->bits_per_pixel = 1;
    var->grayscale = 0;
    var->nonstd = 0;                 /* != 0 Non standard pixel format */
    var->activate = FB_ACTIVATE_NOW; /* see FB_ACTIVATE_*              */
    var->height = -1;                /* height of picture in mm        */
    var->width  = -1;                /* width of picture in mm         */
    var->accel_flags = 0;            /* acceleration flags (hints      */
    /* var->pixclock = S1D_DISPLAY_PCLK; */
    var->right_margin = 0;
    var->lower_margin = 0;
    var->hsync_len = 0;
    var->vsync_len = 0;
    var->left_margin = 0;
    var->upper_margin = 0;
    var->sync = FB_SYNC_ON_GREEN;
    var->vmode = FB_VMODE_NONINTERLACED;

    var->red.offset = var->green.offset = var->blue.offset = 0;
    var->red.length = var->blue.length = 0;
    var->green.length = 1;

    var->red.msb_right = var->green.msb_right = var->blue.msb_right = 0;
    var->transp.offset = var->transp.length = var->transp.msb_right = 0;

    return 0; 
} 
 
//----------------------------------------------------------------------------- 
//
// Set the 'par' according to hardware
//
//----------------------------------------------------------------------------- 
static void st7529_get_par(void *fb_par, struct fb_info_gen *info) 
{ 
} 
 
//----------------------------------------------------------------------------- 
//
// Set the hardware according to 'par'. 
//
//----------------------------------------------------------------------------- 
static void st7529_set_par(const void *fb_par, struct fb_info_gen *info)
{ 
} 
 
//----------------------------------------------------------------------------- 
// 
// Read a single color register and split it into colors/transparent.
// The return values must have a 16 bit magnitude.  
// Return != 0 for invalid regno.  
// 
//----------------------------------------------------------------------------- 
static int st7529_getcolreg(unsigned regno, unsigned *red, unsigned *green, 
                unsigned *blue, unsigned *transp, struct fb_info *info) 
{ 
    return 0;
} 
 
//---------------------------------------------------------------------------- 
// 
// Set a single color register. The values supplied have a 16 bit  
// magnitude.
// Return != 0 for invalid regno.  
//
// We get called even if we specified that we don't have a programmable palette
// or in direct/true color modes!
//---------------------------------------------------------------------------- 
static int st7529_setcolreg(unsigned regno, unsigned red, unsigned green,
                unsigned blue, unsigned transp, struct fb_info *info) 
{ 
    return 0; 
} 
 
//-----------------------------------------------------------------------------
// 
// Pan (or wrap, depending on the `vmode' field) the display using the 
// `xoffset' and `yoffset' fields of the `var' structure. 
// If the values don't fit, return -EINVAL. 
// 
//-----------------------------------------------------------------------------
static int st7529_pan_display(const struct fb_var_screeninfo *var, 
                struct fb_info_gen *info) 
{ 
    return 0; 
} 
 
//-----------------------------------------------------------------------------
//     
// Blank the screen if blank_mode != 0, else unblank. If blank == NULL  
// then the caller blanks by setting the CLUT (Color Look Up Table) to all  
// black. Return 0 if blanking succeeded, != 0 if un-/blanking failed due  
// to e.g. a video mode which doesn't support it. Implements VESA suspend  
// and powerdown modes on hardware that supports disabling hsync/vsync:  
//     
// 0 = unblank 
// 1 = blank 
// 2 = suspend vsync 
// 3 = suspend hsync 
// 4 = off 
//
//----------------------------------------------------------------------------- 
static int st7529_blank(int blank_mode, struct fb_info_gen *info)
{ 
    return 0; 
} 
 
//----------------------------------------------------------------------------- 
//
// Fill in a pointer with the virtual address of the mapped frame buffer.  
// Fill in a pointer to appropriate low level text console operations (and  
// optionally a pointer to help data) for the video mode `par' of your  
// video hardware. These can be generic software routines, or hardware  
// accelerated routines specifically tailored for your hardware.  
// If you don't have any appropriate operations, you must fill in a  
// pointer to dummy operations, and there will be no text output.  
//
//----------------------------------------------------------------------------- 
static void st7529_set_disp(const void *par, struct display *disp,
                struct fb_info_gen *info)
{
    disp->screen_base = fb_info.virtVidMem;
    disp->dispsw = &fbcon_mfb;
    disp->scrollmode = SCROLL_YREDRAW;
} 



void initWp1Registers( void )
{
	/* Intialisation values as per uboot/board/opera/LCD_Test.c */

	WR_CMD( EXTIN );	/* Select Command table 0 */

	WR_CMD( SLPOUT );	/* Sleep Out */

	WR_CMD( OSCON );	/* Oscillator On */

	WR_CMD( PWRCTRL );	/* Power Control */
	WR_DAT( 0x03 );		/* ref gen on, volt follow on, Boos off */

	udelay( 1000 );	/* 1mS delay */

	WR_CMD( VOLCTRL );	/* Electronic Volume Control */
	WR_DAT( 0x04 );		/* Vop=14.0v  */
	WR_DAT( 0x04 );

	WR_CMD( DISCTRL );	/* Display Control */
	WR_DAT( 0x04 );		/* 2 divisions */
	WR_DAT( 0x19 );
	WR_DAT( 0x00 );

	WR_CMD( DISNOR );	/* Normal Display */

	WR_CMD( COMSCN );	/* Scan direction */
	WR_DAT( 0x01 );		/* 0->79, 159->80 */

	WR_CMD( DATSDR );	/* Data Scan Direction */
	WR_DAT( 0x00 );		/* li normal, ci normal, address col direction */
	WR_DAT( 0x00 );		/* Seg output */
	WR_DAT( DATSDR_PPB );		/* 32 Grayscale 3 byte 3 pix mode */

	WR_CMD( LASET );	/* Line Address */
	WR_DAT( WP1_START_LINE );		/* 128 lines, start line 60, end line 100 */
	WR_DAT( WP1_END_LINE );

	WR_CMD( CASET );	/* Column Address */
	WR_DAT( WP1_START_COLUMN );		/* 70 columns * 3 pix = 210 Pix */
	WR_DAT( WP1_END_COLUMN );		/* Start column 0, end column 69 */

	WR_CMD( EXTOUT );	/* Select Command Table 1 */

	WR_CMD( ANASET );	/* Analog Circuit Set */
	WR_DAT( 0x04 );		/* OSC Freq adj 0-7 */
	WR_DAT( 0x01 );		/* Booster eff 0-3 */
	WR_DAT( 0x03 );		/* LCD bias ratio 0-7, 3=1/11 */

	WR_CMD( SWINT );

	WR_CMD( EXTIN );	/* Select Command table 0 */

	WR_CMD( EPINT );	/* EEPROM Initial Code */
	WR_DAT( 0x19 );

	WR_CMD( EXTOUT );	/* Select Command Table 1 */

	WR_CMD( EPCTIN );	/* Control EEPROM */
	WR_DAT( 0x00 ); 	/* enable read */
	mdelay(50);

	WR_CMD( EPMRD );    /* read data from the EEPROM */
	mdelay(50);

	WR_CMD( EPCOUT );   /* Cancel EEPROM command */

	WR_CMD( EXTIN );	/* Select Command Table 0 */

	WR_CMD( DISON );	/* Display On */

	udelay( 10 );
}

void initWp2Registers( void )
{
	/* Intialisation values as per uboot/board/opera/LCD_Test.c */

	WR_CMD( EXTIN );	/* Select Command table 0 */

	WR_CMD( SLPOUT );	/* Sleep Out */

	WR_CMD( OSCON );	/* Oscillator On */

	WR_CMD( PWRCTRL );	/* Power Control */
	WR_DAT( 0x03 );		/* ref gen on, volt follow on, Boos off */

	udelay( 1000 );	/* 1mS delay */

	WR_CMD( VOLCTRL );	/* Electronic Volume Control */
	WR_DAT( 0x06 );		/* Vop=14.08v  */
	WR_DAT( 0x04 );

	WR_CMD( DISCTRL );	/* Display Control */
	WR_DAT( 0x04 );		/* divide by 2 */
	WR_DAT( 0x23 );		/* duty = 144 */
	WR_DAT( 0x10 );		/* 17 Line cycles */

	WR_CMD( DISNOR );	/* Normal Display */

	WR_CMD( COMSCN );	/* Scan direction */
	WR_DAT( 0x01 );		/* 0->79, 159->80 */

	WR_CMD( DATSDR );	/* Data Scan Direction */
	WR_DAT( 0x02 );		/* Pix3 2 1 arrangement, 3 Byte 3 Pix dither mode 1 */
	WR_DAT( 0x01 );
	WR_DAT( DATSDR_PPB );

	WR_CMD( LASET );	/* Line Address */
	WR_DAT( WP2_START_LINE );		/* 128 lines, start line 16, end line 143 */
	WR_DAT( WP2_END_LINE );

	WR_CMD( CASET );	/* Column Address */
	WR_DAT( WP2_START_COLUMN );		/* 80 columns * 3 pix = 240 Pix */
	WR_DAT( WP2_END_COLUMN );		/* Start column 5, end column 84 */

	WR_CMD( EXTOUT );	/* Select Command Table 1 */

	WR_CMD( ANASET );	/* Analog Circuit Set */
	WR_DAT( 0x02 );		/* OSC freq adjust */
	WR_DAT( 0x01 );		/* booster efficiency */
	WR_DAT( 0x03 );		/* lcd bias ratio */

	WR_CMD( SWINT );

	WR_CMD( EXTIN );	/* Select Command Table 0 */

	WR_CMD( DISON );	/* Display On */

	udelay( 10 );
}

void initWp2RegistersT40( void )
{
	/* Intialisation values as per uboot/board/opera/LCD_Test.c */
printk( KERN_INFO "st7529fb.initWp2RegistersT40: initialising display\n");


	WR_CMD( EXTIN );	/* Select Command table 0 */

	WR_CMD( SLPOUT );	/* Sleep Out */

	WR_CMD( OSCON );	/* Oscillator On */

	WR_CMD( PWRCTRL );	/* Power Control */
	WR_DAT( 0x03 );		/* ref gen on, volt follow on, Boos off */

	udelay( 1000 );	/* 1mS delay */

	WR_CMD( VOLCTRL );	/* Electronic Volume Control */
	WR_DAT( 0x0B );		/* Vop=14.08v  */
	WR_DAT( 0x04 );

	WR_CMD( DISCTRL );	/* Display Control */
	WR_DAT( 0x00 );		/* divide by 2 */
	WR_DAT( 0x23 );		/* duty = 144 */
	WR_DAT( 0x10 );		/* 17 Line cycles */

	WR_CMD( DISNOR );	/* Normal Display */

	WR_CMD( COMSCN );	/* Scan direction */
	WR_DAT( 0x01 );		/* 0->79, 159->80 */

	WR_CMD( DATSDR );	/* Data Scan Direction */
	WR_DAT( 0x02 );		/* Pix3 2 1 arrangement, 3 Byte 3 Pix dither mode 1 */
	WR_DAT( 0x01 );
	WR_DAT( DATSDR_PPB );

	WR_CMD( LASET );	/* Line Address */
	WR_DAT( WP2_START_LINE );		/* 128 lines, start line 16, end line 143 */
	WR_DAT( WP2_END_LINE );

	WR_CMD( CASET );	/* Column Address */
	WR_DAT( WP2_START_COLUMN );		/* 80 columns * 3 pix = 240 Pix */
	WR_DAT( WP2_END_COLUMN );		/* Start column 5, end column 84 */

	WR_CMD( EXTOUT );	/* Select Command Table 1 */

	WR_CMD( ANASET );	/* Analog Circuit Set */
	WR_DAT( 0x01 );		/* OSC freq adjust */
	WR_DAT( 0x01 );		/* booster efficiency */
	WR_DAT( 0x03 );		/* lcd bias ratio */

	WR_CMD( EXTIN );	/* Select Command table 0 */
	
	WR_CMD( EPINT );	/* EEPROM Initial Code */
	WR_DAT( 0x19 );	
	
	WR_CMD( EXTOUT );	/* Select Command Table 1 */

    WR_CMD( EPCTIN );	/* Control EEPROM */		
	WR_DAT( 0x00 ); 	/* enable read */
	mdelay(50);

	WR_CMD( EPMRD );    /* read data from the EEPROM */
	mdelay(50);

	WR_CMD( EPCOUT );   /* Cancel EEPROM command */
	WR_CMD( EXTIN );	/* Select Command Table 0 */

	WR_CMD( DISON );	/* Display On */

}



//-----------------------------------------------------------------------------
//
// Initialize the chip and the frame buffer driver.
//
//-----------------------------------------------------------------------------
int __init
st7529fb_init(char *dummy)
{

	unsigned char *addr;
	/* initialise semaphore */



	//printk( KERN_INFO "st7529fb_init: called\n");

	/* Switch off reset of Display */
	ST7529_RESET();

	/* detect here the type of Workpoint and initialise structures */
	switch( fb_info.wpType = get_WP_type() ){
		case 1:
			/* Really only 205 pixels, but whats that in bytes->round up */
			isWp1 = 1;
			displaySize.width = 208; 
			displaySize.height = 41;
			initWp1Registers();
			fb_info.startLine = WP1_START_LINE;
			fb_info.endLine = WP1_END_LINE;
			fb_info.startColumn = WP1_START_COLUMN;
			fb_info.endColumn = WP1_END_COLUMN;
			break;
		case 2:
			isWp1 = 0;
			displaySize.width = 240;
			displaySize.height = 128;
			//different hardware identified by the cpu version
			if(incaip2_get_cpu_ver() < 3)
				initWp2Registers();
			else
				initWp2RegistersT40();
			fb_info.startLine = WP2_START_LINE;
			fb_info.endLine = WP2_END_LINE;
			fb_info.startColumn = WP2_START_COLUMN;
			fb_info.endColumn = WP2_END_COLUMN;
			break;
		default:
			isWp1 = 0;
			displaySize.width = 0;
			displaySize.height = 0;
        		printk( KERN_ERR "st7529fb_init: not workpoint 1 or 2\n");
        		return -EINVAL;
			break;
	}

        //printk( KERN_INFO "st7529fb_init: HW inialised\n");

	/*  accordingly */
	/* Assume 5BPP is a byte per pixel */
	/* But our virtual buffer is 1BPP */
	displaySize.scan_lines = displaySize.height;
	displaySize.panelSize = (displaySize.height * displaySize.width);

	fb_info.vidMemSize = displaySize.width * displaySize.height / 8;


	/* Get windows to physical registers and video memory */
    	fb_info.regBase  = (unsigned char*) ioremap(ST7529_PHYSICAL_REG_ADDR,
							ST7529_PHYSICAL_REG_SIZE);

    	if ( !fb_info.regBase) {
        	printk(KERN_ERR "st7529fb_init: ioremap() returned NULL\n");
        	return -EINVAL;
        }

	/* Allocate our Virtual display buffer and dirty pixel map */

	/*fb_info.virtVidMem = kmalloc( fb_info.vidMemSize, GFP_ATOMIC );*/ 
	fb_info.virtVidMemAllocation = kmalloc( fb_info.vidMemSize+2*PAGE_SIZE, GFP_ATOMIC ); 

    	if ( !fb_info.virtVidMemAllocation ) {
		iounmap( (void *)fb_info.regBase );
        	printk(KERN_ERR "st7529fb_init: kmalloc() returned NULL\n");
        	return -EINVAL;
	}
	fb_info.virtVidMem = (unsigned char*)(((unsigned long)fb_info.virtVidMemAllocation + PAGE_SIZE - 1) & PAGE_MASK); 

	for( addr = fb_info.virtVidMem; addr < (fb_info.virtVidMem+fb_info.vidMemSize); addr+=PAGE_SIZE ){
		mem_map_reserve( virt_to_page( addr ) );
	}

	
	fb_info.physVidMemAllocation = kmalloc( fb_info.vidMemSize+(2*PAGE_SIZE), GFP_ATOMIC ); 
	fb_info.physVidCopy = (unsigned char*)(((unsigned long)fb_info.physVidMemAllocation + PAGE_SIZE - 1) & PAGE_MASK); 

	for( addr = fb_info.physVidCopy; addr < (fb_info.physVidCopy+fb_info.vidMemSize); addr+=PAGE_SIZE ){
		mem_map_reserve( virt_to_page( addr ) );
	}


	strcpy(fb_info.gen.info.modename, "st7529");
	fb_info.gen.fbhw = &st7529_hwswitch;
	fb_info.gen.fbhw->detect();
	fb_info.gen.info.changevar = NULL;
	fb_info.gen.info.node = -1;
	fb_info.gen.info.fbops = &st7529fb_ops;
	fb_info.gen.info.disp = &disp;

	fb_info.gen.info.switch_con = &fbgen_switch;
	fb_info.gen.info.updatevar = &fbgen_update_var;
	fb_info.gen.info.blank = &fbgen_blank;

        //printk( KERN_INFO "st7529fb_init: setting font name\n");
	strcpy(fb_info.gen.info.fontname, default_fontname);
	fb_info.gen.parsize = 0;
	fb_info.gen.info.flags = FBINFO_FLAG_DEFAULT;

        //printk( KERN_INFO "st7529fb_init: fbgen_get_var\n");
	fbgen_get_var(&disp.var, -1, &fb_info.gen.info);
	disp.var.activate = FB_ACTIVATE_NOW;
        //printk( KERN_INFO "st7529fb_init: fbgen_do_set_var\n");
	fbgen_do_set_var(&disp.var, 1, &fb_info.gen);
        //printk( KERN_INFO "st7529fb_init: fbgen_set_disp\n");
	fbgen_set_disp(-1, &fb_info.gen);
        //printk( KERN_INFO "st7529fb_init: fbgen_instll_cmap\n");
	fbgen_install_cmap(0, &fb_info.gen);

#ifdef ST7529_USE_UPDATE_THREAD
	st7529ThreadCreate();
#endif
	//printk( KERN_INFO "st7529fb_init: just before register_framebuffer\n");
	if (register_framebuffer(&fb_info.gen.info) < 0)
		return -EINVAL;

        //printk( KERN_INFO "st7529fb_init: OK\n");
	return 0;
}


//-----------------------------------------------------------------------------
//
// Parse user speficied options (`video=st7529fb:')
//
//-----------------------------------------------------------------------------
int __init
st7529fb_setup(char *options, int *ints)
{
	return 0;
}

/* ------------------------------------------------------------------------- */

static struct fb_ops st7529fb_ops = {
	owner:       THIS_MODULE,
	fb_get_fix:  fbgen_get_fix,
	fb_get_var:  fbgen_get_var,
	fb_set_var:  fbgen_set_var,
	fb_get_cmap: NULL,
	fb_set_cmap: NULL,
	fb_ioctl:	 st7529_ioctl,
	fb_mmap:     st7529_mmap,
};

/*
 *  In most cases the `generic' routines (fbgen_*) should be satisfactory.
 *  However, you're free to fill in your own replacements.
 */
static struct fbgen_hwswitch st7529_hwswitch =  { 
	st7529_detect, 
	st7529_encode_fix, 
	st7529_decode_var, 
	st7529_encode_var,
	st7529_get_par, 
	st7529_set_par, 
	st7529_getcolreg,
   	st7529_setcolreg, 
	st7529_pan_display,
	st7529_blank,
	st7529_set_disp
};

/* 
 *  Modularization
 */
#ifdef MODULE
int init_module(void)
{
    	st7529fb_init( NULL );
    	return 0;
}

void cleanup_module(void)
{
#ifdef ST7529_USE_UPDATE_THREAD
    	st7529ThreadDestroy();
#endif
    	unregister_framebuffer(&fb_info.gen.info);
    	/* kfree( fb_info.dirtyMap  ); */
    	kfree( fb_info.virtVidMemAllocation );
    	kfree( fb_info.physVidMemAllocation );
	
	iounmap( (void *)fb_info.regBase );
} 
#endif 
