/****************************************************************************************
	Copyright (c) Siemens Enterprise Communications Limited

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
****************************************************************************************/

/**\file
*	This module accepts the key presses from the wheel key pad (via keypadDevice ) and passes them
*	to the QT driver
*/

#include <linux/module.h>
#include <asm/types.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include <asm/uaccess.h>
#include <linux/devfs_fs_kernel.h>

#include <linux/tqueue.h> /* task queues */

#include <linux/time.h>

#include "keyinput.h"


/** Major ID of ButtonPipe device (default: 0=dynamic)
*/
static int major = 0;
/**
* Module parameter declaration of variable major as type integer
*/
MODULE_PARM(major, "i");
/**
* Module parameter description of variable major
*/

MODULE_PARM_DESC(major, "Major ID ButtonPipe (default: 0=dynamic)");


/**
* Declaring a kernel 'wait queue' on which a process can be queued waiting an event
* 
*/

DECLARE_WAIT_QUEUE_HEAD(hwButtonWaitQueue);


/**
* Structure device used for the ButtonPipe 
*/

#define BUTTON_EVENT_QUEUE_LEN		(1<<6)
#define BUTTON_EVENT_QUEUE_MASK	(BUTTON_EVENT_QUEUE_LEN-1)

#define NUM_BUTTONS	18

#define PTR_INCR(x) x = ((x == BUTTON_EVENT_QUEUE_MASK) ? 0 : x +1)
#define FIFO_FULL(rp,wp) ((((rp) + BUTTON_EVENT_QUEUE_LEN - (wp)) & BUTTON_EVENT_QUEUE_MASK) == 2)

typedef u8 buttonID_t;

typedef struct  {
	spinlock_t lock;
	volatile int wqPtr;
	volatile int rqPtr;
	volatile buttonID_t keyEventQueue[BUTTON_EVENT_QUEUE_LEN];
	buttonID_t lastButton;
	int historyPtr;
} ButtonPipe_t;

typedef struct {
	char *name;
	char *inStateText;
	char *outStateText;
	int inPressCount;
	int inReleaseCount;
	int outPressCount;
	int outReleaseCount;
	int inErrorCount;
} buttonStat_t;

static buttonStat_t buttonStatTable[NUM_BUTTONS];
/**
* An instance of the key pad driver data structure
*/
static ButtonPipe_t ButtonPipe_Device;

#define BUTTON_HISTORY_SIZE (1<<6)

static buttonID_t buttonHistory[BUTTON_HISTORY_SIZE];

/**
* Device filesystem handle structure for the creation of device files
*/
static devfs_handle_t input_devfs_handle;

static void initButtonStatTable( void )
{
	int i;
	memset(buttonStatTable,0,sizeof(buttonStatTable));
	buttonStatTable[0].name  = "digit 0      ";
	buttonStatTable[1].name  = "digit 1      ";
	buttonStatTable[2].name  = "digit 2      ";
	buttonStatTable[3].name  = "digit 3      ";
	buttonStatTable[4].name  = "digit 4      ";
	buttonStatTable[5].name  = "digit 5      ";
	buttonStatTable[6].name  = "digit 6      ";
	buttonStatTable[7].name  = "digit 7      ";
	buttonStatTable[8].name  = "digit 8      ";
	buttonStatTable[9].name  = "digit 9      ";
	buttonStatTable[10].name = "digit *      ";
	buttonStatTable[11].name = "digit #      ";
	buttonStatTable[12].name = "wheel up     ";
	buttonStatTable[13].name = "wheel down   ";
	buttonStatTable[14].name = "wheel left   ";
	buttonStatTable[15].name = "wheel right  ";
	buttonStatTable[16].name = "wheel ok     ";
	buttonStatTable[17].name = "wheel touched";
	for(i=0; i < NUM_BUTTONS; i++ )
	{
		buttonStatTable[i].inStateText = "released";
		buttonStatTable[i].outStateText = "released";
	}
}

static void updateButtonStatTableIn( buttonID_t lastButton, buttonID_t buttonID )
{
	int state = (buttonID & 0x80)? KEY_RELEASED: KEY_PRESSED;
	buttonID &= 0x7f;
	if( buttonID < NUM_BUTTONS )
	{
		if( state == KEY_PRESSED )
		{
			buttonStatTable[buttonID].inStateText = "pressed";
			buttonStatTable[buttonID].inPressCount++;
			if( (lastButton & 0x80) != 0x80 )
			{
				/* got a key press but last is not released */
				buttonStatTable[lastButton & 0x7f].inErrorCount++;
//				printk(KERN_ERR "ButtonPipe_write: ERROR button %02x not released before %02x pressed\n",lastButton, buttonID);
			}
		}
		else
		{
			buttonStatTable[buttonID].inStateText = "released";
			buttonStatTable[buttonID].inReleaseCount++;
			if( lastButton != buttonID )
			{
				/* got a key release but not had press */
				buttonStatTable[buttonID].inErrorCount++;
//				printk(KERN_ERR "ButtonPipe_write: ERROR button %02x released before pressed\n",buttonID);
			}
		}
	}
	else
	{
			printk(KERN_ERR "updateButtonStatTable: bad button=%d state=%d\n",buttonID,state);
	}
}

static void updateButtonStatTableOut( buttonID_t buttonID )
{
	int state = (buttonID & 0x80)? KEY_RELEASED: KEY_PRESSED;
	buttonID &= 0x7f;
	if( buttonID < NUM_BUTTONS )
	{
		if( state == KEY_PRESSED )
		{
			buttonStatTable[buttonID].outStateText = "pressed";
			buttonStatTable[buttonID].outPressCount++;
		}
		else
		{
			buttonStatTable[buttonID].outStateText = "released";
			buttonStatTable[buttonID].outReleaseCount++;
		}
	}
	else
	{
			printk(KERN_ERR "updateButtonStatTableOut: bad button=%d state=%d\n",buttonID,state);
	}
}

static int ButtonInputEvent( ButtonPipe_t *dev, key_event_t_ul *event )
{
	buttonID_t buttonID;
	int ret = 0;

	//printk(KERN_NOTICE "ButtonInputEvent: key=%d state=%d\n",key,state);

	switch( event->key_code )
	{
		/* We have number keys, translate them */
		case PHYSICAL_KEYPAD_0:		buttonID =  0;  break; /* 0 */
		case PHYSICAL_KEYPAD_1:		buttonID =  1;  break; /* 1 */
		case PHYSICAL_KEYPAD_2:		buttonID =  2;  break; /* 2 */
		case PHYSICAL_KEYPAD_3:		buttonID =  3;  break; /* 3 */
		case PHYSICAL_KEYPAD_4:		buttonID =  4;  break; /* 4 */
		case PHYSICAL_KEYPAD_5:		buttonID =  5;  break; /* 5 */
		case PHYSICAL_KEYPAD_6:		buttonID =  6;  break; /* 6 */
		case PHYSICAL_KEYPAD_7:		buttonID =  7;  break; /* 7 */
		case PHYSICAL_KEYPAD_8:		buttonID =  8;  break; /* 8 */
		case PHYSICAL_KEYPAD_9:		buttonID =  9;  break; /* 9 */
		case PHYSICAL_KEYPAD_AST:	buttonID = 10;  break; /* * */
		case PHYSICAL_KEYPAD_HASH:	buttonID = 11;  break; /* # */

		/* We have a wheel keypad key now we have to translate it */
		case PHYSICAL_NAVI_KEY_NORTH:	buttonID = 12; 	break; /* UP */
		case PHYSICAL_NAVI_KEY_SOUTH:	buttonID = 13; 	break; /* DOWN */
		case PHYSICAL_NAVI_KEY_WEST:	buttonID = 14; 	break; /* LEFT */
		case PHYSICAL_NAVI_KEY_EAST:	buttonID = 15; 	break; /* RIGHT */
		case PHYSICAL_NAVI_KEY_CENTRAL:	buttonID = 16; 	break; /* OK */
		case PHYSICAL_NAVI_TOUCH:		buttonID = 17;	break; /* TOUCH */

		/* Don't send any other keys to the QT */
		default:
		{
			return( -EFAULT );
		}
		break;
	}

	/* add in the button State */
	if (event->key_state == KEY_RELEASED) 
	{
		buttonID |= 0x80;
	}
	else
	{
		buttonID &= 0x7f;
	}

	buttonHistory[dev->historyPtr & (BUTTON_HISTORY_SIZE-1)] = buttonID;
	dev->historyPtr++;
	updateButtonStatTableIn( dev->lastButton, buttonID );
	dev->lastButton = buttonID;
	
	if( !FIFO_FULL(dev->rqPtr,dev->wqPtr) )
	{
		dev->keyEventQueue[dev->wqPtr] = buttonID;
		PTR_INCR(dev->wqPtr);
		wake_up_interruptible( &hwButtonWaitQueue );
		ret = sizeof( key_event_t_ul );
	}
	else
	{
		printk(KERN_WARNING "ButtonInputEvent: FIFO full %d %d\n", dev->rqPtr,dev->wqPtr);
	}
//	printk(KERN_NOTICE "ButtonInputEvent: %d %d\n",dev->rqPtr,dev->wqPtr);
	return( ret );
}

static int ButtonReadEvent( ButtonPipe_t *dev, buttonID_t * buttonID )
{
	//printk(KERN_NOTICE "ButtonReadEvent: wr=%d rd=%d\n", dev->wqPtr, dev->rqPtr);

	if( dev->wqPtr != dev->rqPtr )
	{
		*buttonID = dev->keyEventQueue[dev->rqPtr];
		PTR_INCR(dev->rqPtr);
		updateButtonStatTableOut( *buttonID );
		return( sizeof( buttonID_t ) );
	}
	return( 0 );
}

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> IOCTL Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* Declartion of the file operations structure which effectively aliases application
* functions such as read, write etc. to equivalent kernel funtion
*/
/**
* Prototype declartion of the file operations structure for Wheel keypad which 
* effectively aliases application functions such as read, write etc. to equivalent 
* kernel funtion
*/
ssize_t ButtonPipe_read(struct file *, char *, size_t, loff_t *);
ssize_t ButtonPipe_write(struct file *, const char *, size_t, loff_t *);
int ButtonPipe_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
int ButtonPipe_open(struct inode *, struct file *);
int ButtonPipe_close(struct inode *, struct file *);
unsigned int ButtonPipe_poll(struct file *, struct poll_table_struct *);

struct file_operations ButtonPipe_fops = {
	read:		ButtonPipe_read,	/*read */
	write:		ButtonPipe_write, /* write */
	ioctl:		ButtonPipe_ioctl, /* ioctl */
	poll:		ButtonPipe_poll,  /* poll */
	open:		ButtonPipe_open,  /* open */
	release:	ButtonPipe_close, /* release */
};



/**
 * Aliased read function called from user space (QT driver)
 * Reads available key press and sends it to user space.
 *\param *file pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns either number of data items transferred or error state
 */

ssize_t ButtonPipe_read(struct file *fp, char *buf, size_t size, loff_t * off)
{
	buttonID_t buttonID;
	ssize_t ret = 0;
	ButtonPipe_t* dev = (ButtonPipe_t*)fp->private_data;

	while((dev->rqPtr) == (dev->wqPtr))
	{/* nothing to read */
		if (fp->f_flags & O_NONBLOCK)
		{
			//printk("ButtonPipe_read: Non blocking read");

			return -EAGAIN;
		}
		if(wait_event_interruptible(hwButtonWaitQueue, (dev->rqPtr) != (dev->wqPtr)))
			return -ERESTARTSYS; /* signal: tell the fs layer to handle it */
	}
	
//	printk(KERN_NOTICE "ButtonPipeRead: request size %d\n", size);
	/* OK we have at least one key in the buffer ring */

	while( (size >= sizeof(buttonID_t)) && (ButtonReadEvent( dev, &buttonID ) != 0) )
	{
		// Pass the keypress (or key release) up to the user

		if(copy_to_user(buf, &buttonID, sizeof( buttonID_t ) ))
		{
			printk(KERN_ERR "ButtonPipeRead: copy_to_user failed\n");
			return -ERESTARTSYS;
		}
		ret += sizeof( buttonID_t );
		buf += sizeof( buttonID_t );
		size -= sizeof( buttonID_t );
//		printk(KERN_NOTICE "ButtonPipeRead: %02x\n", buttonID);
	}

	return( ret );
}

/**
 * Aliased write function called from user space (keypadDevice)
 *\param *fp pointer to a kernel file data structure
 *\param *buf pointer to a user space data buffer
 *\param size of the user space data buffer
 *\param *off pointer to file position
 *\returns Number of data items sent or error state
 */

ssize_t ButtonPipe_write(struct file *fp, const char *buf, size_t size, loff_t *off)
{
	ButtonPipe_t* dev = (ButtonPipe_t*)fp->private_data;
	
	key_event_t_ul event;

	if (size != sizeof(key_event_t_ul))
	{
		printk(KERN_WARNING "ButtonPipe_write: illegal size %d\n", size);
		return 0;
	}

	if (copy_from_user(&event, buf, sizeof(key_event_t_ul)))
	{
		return( -EFAULT );
	}

	if( (size = ButtonInputEvent( dev, &event )) != sizeof( key_event_t_ul ) )
	{
//		printk(KERN_WARNING "ButtonPipe_write: failed size %d\n", size);
	}

	return( size );
}

/**
 * Aliased ioctl function called from user space
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *filp pointer to a kernel file data structure
 *\param cmd ioctl command
 *\param arg associated command arguments
 * Currently not used
 *\returns 0
 */

int ButtonPipe_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int err;
	printk(KERN_ERR "ButtonPipe_ioctl: cmd=%08x\n", cmd);
	err = 0;
	return err;
}

/**
 * Aliased poll function called from user space by poll or select
 * 
 * 
 *\param *fp pointer to a kernel file data structure
 *\param *pts pointer to a kernel poll table data structure
 *\returns flags to indicate valid data available
 */

unsigned int ButtonPipe_poll(struct file *fp, struct poll_table_struct *pts)
{
	unsigned int mask = 0;
	ButtonPipe_t* dev = (ButtonPipe_t*)fp->private_data;

	poll_wait(fp, &hwButtonWaitQueue, pts);

    if ((dev->rqPtr) != (dev->wqPtr)) mask |= POLLIN | POLLRDNORM;  /* readable */

	return mask;
}

/**
 * Aliased open function called from user space by open
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0 for success (in turn the kernel returns a file descriptor)
 */

int ButtonPipe_open(struct inode *inode, struct file *fp)
{
	int omajor;
	int ominor;

	fp->private_data = &ButtonPipe_Device;

	ominor = MINOR(inode->i_rdev);
	omajor = MAJOR(inode->i_rdev);
	if (fp->f_flags & O_NONBLOCK)
	
	printk(KERN_NOTICE "ButtonPipe_open: %d:%d %s\n", omajor, ominor, (fp->f_flags & O_NONBLOCK)?"Blocking":"Non Blocking");

	return 0;
}

/**
 * Aliased close function called from user space by close
 *\param *inode pointer to the inode structure for the device this contains such information such
 *as the major and minor numbers.
 *\param *fp pointer to a kernel file data structure
 *\returns 0
 */

int ButtonPipe_close(struct inode *inode, struct file *fp)
{
	int err;
	printk(KERN_NOTICE "ButtonPipe_close:\n");
	err = 0;
	return err;
}


/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>> PROC Section <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/

/**
* A pointer to a proc directory entry structure used to create a proc entry for a device
*/
static struct proc_dir_entry *ButtonPipe_proc;

/**
* Function to handle a proc entry read i.e. 'cat /proc/drivers/ButtonPipe'
*\param *buf pointer to buffer where data is to be written
*\param **start used by the function to point to a specific page of data
*\param offset offset into file structure
*\param count number of data items returned
*\param *eof pointer to an interger set by the driver to signal no more data is to be sent
*\param *data general purpose data pointer
*\returns number of bytes of data returned
*/

static int ButtonPipe_read_procmem(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
	int len, i;
	int historySize;
	int historyIndex;

    len = sprintf(buf, "ButtonPipe Device version 1.0\n");
	for( i=0; i < NUM_BUTTONS; i++ )
	{
		len += sprintf(buf+len,"%s\n",buttonStatTable[i].name );
		len += sprintf(buf+len,"\tstate    in %s - out %s\n",buttonStatTable[i].inStateText, buttonStatTable[i].outStateText );
		len += sprintf(buf+len,"\tpressed  in %4d - out %4d\n",buttonStatTable[i].inPressCount, buttonStatTable[i].outPressCount );
		len += sprintf(buf+len,"\treleased in %4d - out %4d\n",buttonStatTable[i].inReleaseCount, buttonStatTable[i].outReleaseCount );
		len += sprintf(buf+len,"\terrors   in %4d\n\n",buttonStatTable[i].inErrorCount );
	}

	historyIndex = 0;
	historySize = ButtonPipe_Device.historyPtr;
	if( historySize > BUTTON_HISTORY_SIZE )
	{
		historySize = BUTTON_HISTORY_SIZE;
		historyIndex = ButtonPipe_Device.historyPtr;
		/* list starting at oldest */
	}

	len += sprintf(buf+len,"Input history count %d\n",ButtonPipe_Device.historyPtr );
	for( i=0; i < historySize; i++ )
	{
		u8 codechar[NUM_BUTTONS] = "0123456789*#^v<>KT";
		u8 code = buttonHistory[historyIndex & (BUTTON_HISTORY_SIZE-1)];
		u8 sign = ( code & 0x80 ) ? '-':'+';
		code = codechar[code & 0x7f];
		len += sprintf(buf+len,"%c%c,",sign,code );
		historyIndex++;
	}
	
	len += sprintf(buf+len,"\n\n" );
	return len;
}

/**
* Function to handle a proc entry write not used
*/

static int ButtonPipe_write_procmem(struct file *file, const char *buffer,unsigned long count, void *data)
{
	printk(KERN_INFO "Write procmem\n");
	return count;

}

/**
* Kernel initialisation function run when the module is installed via insmod
*/

static int __init ButtonPipe_init(void)
{
	int res = 0;

	/* create and register device */

	if( (input_devfs_handle = devfs_register(0, "input/keyboards",DEVFS_FL_AUTO_DEVNUM,0, 0, S_IFCHR | S_IRUGO | S_IWUGO,&ButtonPipe_fops,0)) == 0 )
		return -EBUSY;


	if (major == 0) major = res;

	SET_MODULE_OWNER(&ButtonPipe_fops);

	/* create PROC entry */
	if( (ButtonPipe_proc = create_proc_entry( "driver/buttonPipe", 0, 0)) == 0 )
	{
		printk(KERN_ERR "could not create /proc/driver/buttonPipe!\n");
		devfs_unregister(input_devfs_handle);

		return( -ENOENT );
	}
	ButtonPipe_proc->read_proc  = ButtonPipe_read_procmem;
	ButtonPipe_proc->write_proc = ButtonPipe_write_procmem;
	
	ButtonPipe_Device.wqPtr = 0;
	ButtonPipe_Device.rqPtr = 0;
	ButtonPipe_Device.lastButton = 0x80;
	ButtonPipe_Device.historyPtr = 0;
	initButtonStatTable();

	printk(KERN_NOTICE "/proc/driver/ButtonPipe created!\n");

	return( 0 );
}

/**
* Kernel clean-up function run when the module is removed via rmmod
*/

static void __exit ButtonPipe_exit(void)
{


	remove_proc_entry("driver/ButtonPipe", 0);

	devfs_unregister(input_devfs_handle);

	printk(KERN_NOTICE "ButtonPipe_exit\n");
}


#ifdef MODULE
module_init(ButtonPipe_init);
module_exit(ButtonPipe_exit);
#endif

/**
* Macro to tell the kernel that the module bears a free license (stops the kernel complaining!)
*/
MODULE_LICENSE("GPL");
