/*
 * BRIEF MODULE DESCRIPTION
 *      Non-PCI or "builtin" OHCI support.
 *
 * Author: Steve Longerbeam <stevel@mvista.com, or source@mvista.com>
 *
 * 2001 (c) MontaVista Software, Inc. This file is licensed under
 * the terms of the GNU General Public License version 2. This program
 * is licensed "as is" without any warranty of any kind, whether express
 * or implied.
 *
 * Modified by winder.sung@infineon.com
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/usb.h>
#include <linux/pci.h>

#include <asm/irq.h>
#include <asm/io.h>

#include "usb-ohci.h"

#if defined(CONFIG_INCAIP2)       /* Some platform related parameter */
#include <asm/incaip2/incaip2.h>
#include <asm/incaip2/irq.h>
#define IFX_INCAIP2_USB_REG__BASE 0x14103400
#define IFX_INCAIP2_USB_REG__LEN 0xC00
#define IFX_INCAIP2_USB_IRQ INT_NUM_IM4_IRL21 
#endif

#define IFX_USB_REG_BASE IFX_INCAIP2_USB_REG__BASE
#define IFX_USB_REG_LEN IFX_INCAIP2_USB_REG__LEN
#define IFX_USB_IRQ IFX_INCAIP2_USB_IRQ

/********** External Functions **********/
int __devinit
hc_add_ohci(struct pci_dev *dev, int irq, void *membase, unsigned long flags,
	    ohci_t **ohci, const char *name, const char *slot_name);
extern void hc_remove_ohci(ohci_t *ohci);
/****************************************/

/*********** Static Variable ************/
static ohci_t *ifx_ohci;
static int ohci_base=0, ohci_len=0;
static int ohci_irq=-1;
// bogus pci_dev
static struct pci_dev bogus_pcidev;
// ioremapped ohci_base
static void *mem_base;
/****************************************/

#define INCA_IP2_PMS_PMS_GEN_USB   (1 << 17)
static int __init ifx_ohci_init(void)
{
	int ret;

	switch ( INCA_IP2_MPS_CHIPID_VERSION_GET(*INCA_IP2_MPS_CHIPID) ) {
	case 1: /* V1.1 */
	case 2:	/* V1.2 */
		err("USB runs on INCA-IP2 v1.3 and higher!");
		return -ENODEV;
	default:
		/* Enable HW module */
		*INCA_IP2_PMS_PMS_GEN |= INCA_IP2_PMS_PMS_GEN_USB;
	}

	/* Setup base address, register length and irq. */
	ohci_base = IFX_USB_REG_BASE;
	ohci_len = IFX_USB_REG_LEN;
	ohci_irq = IFX_USB_IRQ;

	/*
	 * Request memory resources.
	 */
	if (!request_mem_region(ohci_base, ohci_len, "usb-ohci"))
		return -EBUSY;

	mem_base = ioremap_nocache (ohci_base, ohci_len);
	if (!mem_base) {
		err("Error mapping OHCI memory");
		return -EFAULT;
	}

	/*
	 * Fill in some fields of the bogus pci_dev.
	 */
	memset(&bogus_pcidev, 0, sizeof(struct pci_dev));
	strcpy(bogus_pcidev.name, "IFX OHCI");
	strcpy(bogus_pcidev.slot_name, "builtin");
	bogus_pcidev.resource[0].name = "OHCI Operational Registers";
	bogus_pcidev.resource[0].start = ohci_base;
	bogus_pcidev.resource[0].end = ohci_base + ohci_len;
	bogus_pcidev.resource[0].flags = 0;
	bogus_pcidev.irq = ohci_irq;

	/*
	 * Initialise the generic OHCI driver.
	 */
	ret = hc_add_ohci(&bogus_pcidev, ohci_irq, mem_base, 0, 
			   &ifx_ohci, "usb-ohci", "builtin");

	if (ret) {
		iounmap(mem_base);
		release_mem_region(ohci_base, ohci_len);
	}

	return ret;
}

static void __exit ifx_ohci_exit(void)
{
	hc_remove_ohci(ifx_ohci);

	/*
	 * Release memory resources.
	 */
	release_mem_region(ohci_base,ohci_len);

}

module_init(ifx_ohci_init);
module_exit(ifx_ohci_exit);

