//-----------------------------------------------------------------------------
//
// linux/drivers/video/epson/s1d13xxxfb.c -- frame buffer driver for Epson
// S1D13xxxseries of LCD/CRT/TV controllers.
//
// Copyright(c) 2000, 2005 Epson Research and Development, Inc.
// All rights reserved.
//
// This file is subject to the terms and conditions of the GNU General Public
// License. See the file COPYING in the main directory of this archive for
// more details.
//
//----------------------------------------------------------------------------
//
// WARNING: This file uses a chip specific include file. The include file
// contains all hardware information needed to generate proper timings, color
// depth, screen resolution etc. If the include file was generated by a Windows
// based chip configuration utility, make sure the generated file was converted
// to Linux native text file format, i.e. process the file with dos2unmem_map_reserveix utility
// or similar.
//
// THIS DRIVER WILL NOT BUILD CORRECTLY IF THE INCLUDE FILE IS IN DOS
// TEXT FILE FORMAT.
//
// This driver supports one of the following Epson CRT/TV/LCD controllers:
//
// S1D13504
// S1D13505
// S1D13506
// S1D13704
// S1D13705
// S1D13706
// S1D13714
// S1D13715
// S1D13732
// S1D13733
// S1D13806
// S1D13A03
// S1D13A04
// S1D13A05
//
// The intended platforms for this driver are embedded systems, where memory
// resources are very limited, hence there are virtually no run-time checks
// present in the driver code. The objective is to keep the footprint as small
// as possible. For example, we do not keep a mirror image of palette. The con-
// sequence being that writing a palette entry and then reading it back may
// result in different value, as some of the supported chips do not support
// full 8 bits per primary colors (red, green, blue).
//
// During the driver compilation, the following hash-defines are being used to
// identify the target Epson controller:
//
// CONFIG_FBCON_EPSON_S1D13504
// CONFIG_FBCON_EPSON_S1D13505
// CONFIG_FBCON_EPSON_S1D13506
// CONFIG_FBCON_EPSON_S1D13704
// CONFIG_FBCON_EPSON_S1D13705
// CONFIG_FBCON_EPSON_S1D13706
// CONFIG_FBCON_EPSON_S1D13714
// CONFIG_FBCON_EPSON_S1D13715
// CONFIG_FBCON_EPSON_S1D13732
// CONFIG_FBCON_EPSON_S1D13733
// CONFIG_FBCON_EPSON_S1D13806
// CONFIG_FBCON_EPSON_S1D13A03
// CONFIG_FBCON_EPSON_S1D13A04
// CONFIG_FBCON_EPSON_S1D13A05
//
// All of the above chips were tested using Epson evaluation boards. The boards
// come basically in two flavors: ISA bus adapters or PCI adapters. To support
// the PCI adapters, make sure to specify PCI bridge adapter during configuration.
// This will generate the following hash-define:
//
// CONFIG_FBCON_EPSON_PCI
//
// The driver will be build for one colordepth only. It must be one of the
// following: 1bpp, 2bpp, 4Bpp, 8Bpp, 16Bpp or 32Bpp. This is selected during
// the configuration phase and generates one of the hash-defines:
//
// FBCON_HAS_MFB
// FBCON_HAS_CFB2
// FBCON_HAS_CFB4
// FBCON_HAS_CFB8
// FBCON_HAS_CFB16
// FBCON_HAS_CFB32
//
// Additionally,there are some hash-defines specifying the display geometry
// and LUT register accesses. These are chip dependent and are located in
// corresponding S1D13XXX header files.
//
//-----------------------------------------------------------------------------

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/wrapper.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>

#include <linux/pci.h>
#include <video/fbcon.h>
#include <video/fbcon-mfb.h>
#include <video/fbcon-cfb2.h>
#include <video/fbcon-cfb4.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb32.h>


//-----------------------------------------------------------------------------
//
// Sanity checks
//
//-----------------------------------------------------------------------------

// Check if at least one supported colordepth was configured

#if !defined (FBCON_HAS_MFB) && !defined (FBCON_HAS_CFB2) && \
    !defined (FBCON_HAS_CFB4) && !defined (FBCON_HAS_CFB8) && \
    !defined (FBCON_HAS_CFB16) && !defined(FBCON_HAS_CFB32)
    #error No colordepth specified
#endif

// Check at most one colordepth was configured

#if defined (FBCON_HAS_MFB)
    #if defined (FBCON_HAS_CFB2) || defined (FBCON_HAS_CFB4) || \
        defined (FBCON_HAS_CFB8) || defined (FBCON_HAS_CFB16)|| \
        defined (FBCON_HAS_CFB32)
        #error More than one colordepth specified
    #endif
#endif

#if defined (FBCON_HAS_CFB2)
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB4) || \
        defined (FBCON_HAS_CFB8) || defined (FBCON_HAS_CFB16) || \
        defined (FBCON_HAS_CFB32)
        #error More than one colordepth specified
    #endif
#endif

#if defined (FBCON_HAS_CFB4)
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || \
        defined (FBCON_HAS_CFB8) || defined (FBCON_HAS_CFB16) ||\
        defined (FBCON_HAS_CFB32)
        #error More than one colordepth specified
    #endif
#endif

#if defined (FBCON_HAS_CFB8)
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || \
        defined (FBCON_HAS_CFB4) || defined (FBCON_HAS_CFB16)|| \
        defined (FBCON_HAS_CFB32)
        #error More than one colordepth specified
    #endif
#endif

#if defined (FBCON_HAS_CFB16)
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || \
        defined (FBCON_HAS_CFB4) || defined (FBCON_HAS_CFB8)|| \
        defined (FBCON_HAS_CFB32)
        #error More than one colordepth specified
    #endif
#endif

#if defined (FBCON_HAS_CFB32)
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || \
        defined (FBCON_HAS_CFB4) || defined (FBCON_HAS_CFB8)|| \
        defined (FBCON_HAS_CFB16)
        #error More than one colordepth specified
    #endif
#endif

// S1D13504 supports only 1bpp, 2bpp, 4Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13504
    #if defined (FBCON_HAS_CFB8) || !defined(FBCON_HAS_CFB16)
        #error Invalid configuration!
    #endif
#include "s1d13504.h"
#endif

// S1D13505 supports 1bpp, 2bpp, 4bpp, 8Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13505
#include "s1d13505.h"
#endif

// S1D13506 supports 4Bpp, 8Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13506
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2)
        #error Invalid configuration!
    #endif
#include "s1d13506.h"
#endif

// S1D13704 supports 1bpp, 2bpp, 4bpp
#ifdef CONFIG_FBCON_EPSON_S1D13704
    #if defined (FBCON_HAS_CFB16) || defined (FBCON_HAS_CFB8) || defined (FBCON_HAS_CFB32)
        #error Invalid configuration!
    #endif
#include "s1d13704.h"
#endif

// S1D13705 supports 1bpp, 2bpp, 4Bpp, 8Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13705
    #if defined (FBCON_HAS_CFB16)|| defined (FBCON_HAS_CFB32)
        #error Invalid configuration!
    #endif
#include "s1d13705.h"
#endif

// S1D13706 supports 1bpp, 2bpp, 4Bpp, 8Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13706
#include "s1d13706.h"
#endif

// S1D13714 supports 8Bpp, 16Bpp, 32Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13714
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || defined(FBCON_HAS_CFB4)
        #error Invalid configuration!
    #endif
#include "s1d13714.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

// S1D13715 supports 8Bpp, 16Bpp, 32Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13715
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || defined(FBCON_HAS_CFB4)
        #error Invalid configuration!
    #endif
#include "s1d13715.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

// S1D13732 supports 8Bpp, 16Bpp, 32Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13732
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || defined(FBCON_HAS_CFB4)
        #error Invalid configuration!
    #endif
#include "s1d13732.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

// S1D13733 supports 8Bpp, 16Bpp, 32Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13733
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || defined(FBCON_HAS_CFB4)
        #error Invalid configuration!
    #endif
#include "s1d13733.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

// S1D13806 supports 4Bpp,8Bpp,16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13806
    #if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || defined (FBCON_HAS_CFB32)
        #error Invalid configuration!
    #endif
#include "s1d13806.h"
#endif

// S1D13A03 supports 1bpp, 2bpp, 4Bpp, 8Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13A03
    #if defined (FBCON_HAS_CFB32)
        #error Invalid configuration!
    #endif
#include "s1d13a0x.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

// S1D13A04 supports 1bpp, 2bpp, 4Bpp, 8Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13A04
    #if defined (FBCON_HAS_CFB32)
        #error Invalid configuration!
    #endif
#include "s1d13a0x.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

// S1D13A05 supports 1bpp, 2bpp, 4Bpp, 8Bpp, 16Bpp
#ifdef CONFIG_FBCON_EPSON_S1D13A05
    #if defined (FBCON_HAS_CFB32)
        #error Invalid configuration!
    #endif
#include "s1d13a05.h"
S1D_INSTANTIATE_REGISTERS(static,aS1DRegs);
#endif

#ifdef CONFIG_FBCON_EPSON_RAM_BASED_LCD
#include "s1d13xxxlcd.h"
LCD_INSTANTIATE_REGISTERS(static,aLCDRegs);
#endif


// Now that the include file has been successfully read in, make sure
// it's for the right colordepth

#if (defined (FBCON_HAS_MFB) && (S1D_DISPLAY_BPP != 1)) || \
    (!defined(FBCON_HAS_MFB) && (S1D_DISPLAY_BPP == 1))
    #error Include file colordepth mismatch!
#endif

#if (defined (FBCON_HAS_CFB2) && (S1D_DISPLAY_BPP != 2)) || \
    (!defined(FBCON_HAS_CFB2) && (S1D_DISPLAY_BPP == 2))
    #error Include file colordepth mismatch!
#endif

#if (defined (FBCON_HAS_CFB4) && (S1D_DISPLAY_BPP != 4)) || \
    (!defined(FBCON_HAS_CFB4) && (S1D_DISPLAY_BPP == 4))
    #error Include file colordepth mismatch!
#endif

#if (defined (FBCON_HAS_CFB8) && (S1D_DISPLAY_BPP != 8)) || \
    (!defined(FBCON_HAS_CFB8) && (S1D_DISPLAY_BPP == 8))
    #error Include file colordepth mismatch!
#endif

#if (defined (FBCON_HAS_CFB16) && (S1D_DISPLAY_BPP != 16)) || \
    (!defined(FBCON_HAS_CFB16) && (S1D_DISPLAY_BPP == 16))
    #error Include file colordepth mismatch!
#endif

#if (defined (FBCON_HAS_CFB32) && (S1D_DISPLAY_BPP != 32)) || \
    (!defined(FBCON_HAS_CFB32) && (S1D_DISPLAY_BPP == 32))
    #error Include file colordepth mismatch!
#endif

//-----------------------------------------------------------------------------
//
// Local Definitions
//
//---------------------------------------------------------------------------
#if (defined (CONFIG_FBCON_EPSON_S1D13714) || (CONFIG_FBCON_EPSON_S1D13715) || \
    (CONFIG_FBCON_EPSON_S1D13732) || (CONFIG_FBCON_EPSON_S1D13733))
    #define REG0022_INDIRECT_INTERFACE_MEMORY_ADDRESS_0     0x0022
    #define REG0024_INDIRECT_INTERFACE_MEMORY_ADDRESS_1     0x0024
    #define REG0028_INDIRECT_INTERFACE_MEMORY_DATA_PORT     0x0028
    #define REG0034_LCD_INTERFACE_COMMAND                   0x0034
    #define REG0036_LCD_INTERFACE_PARAMETER                 0x0036
    #define REG0038_LCD_INTERFACE_STATUS                    0x0038
    #define REG003A_LCD_INTERFACE_FRAME_TRANSFER            0x003A
    #define MAIN_WINDOW_OFFSET                              0x0   // see main windows display start address registers
#endif

#define SIZE_REQUIRED                                       displaySize.display_width*displaySize.display_height*S1D_DISPLAY_BPP/8
#define PANEL_SIZE                                          displaySize.display_width*displaySize.display_height
#define LOOP_LIMIT                                          1000     // prevent endless loops

typedef struct
    {
    struct fb_info_gen gen;
    unsigned char *VmemAddr;
    volatile unsigned char *RegAddr;
#ifdef CONFIG_FBCON_EPSON_INDIRECT
    u32 VirtualVmemAddr;
#endif
#ifdef CONFIG_FBCON_EPSON_PCI
    u32 PhysAddr;
#endif
#ifdef FBCON_HAS_CFB16
    u16 cfb16[16];
#endif
#ifdef FBCON_HAS_CFB32
    u32 cfb32[16];
#endif
    }FB_INFO_S1D13xxx;

//-----------------------------------------------------------------------------
//
// Function Prototypes
//
//-----------------------------------------------------------------------------

int __init s1d13xxxfb_setup(char *options, int *ints) ;
int __init s1d13xxxfb_init(char*) ;

static void s1d13xxx_detect(void);
static int  s1d13xxx_encode_fix(struct fb_fix_screeninfo *fix, const void *par,
                struct fb_info_gen *info);
static int  s1d13xxx_decode_var(const struct fb_var_screeninfo *var, void *par,
                struct fb_info_gen *info);
static int  s1d13xxx_encode_var(struct fb_var_screeninfo *var, const void *par,
                struct fb_info_gen *info);
static void s1d13xxx_get_par(void *par, struct fb_info_gen *info);
static void s1d13xxx_set_par(const void *par, struct fb_info_gen *info);
static int  s1d13xxx_getcolreg(unsigned regno, unsigned *red, unsigned *green,
                unsigned *blue, unsigned *transp, struct fb_info *info);
static int  s1d13xxx_setcolreg(unsigned regno, unsigned red, unsigned green,
                unsigned blue, unsigned transp, struct fb_info *info);
static int  s1d13xxx_pan_display(const struct fb_var_screeninfo *var,
                struct fb_info_gen *info);
static int  s1d13xxx_blank(int blank_mode, struct fb_info_gen *info);
static void s1d13xxx_set_disp(const void *par, struct display *disp,
                struct fb_info_gen *info);
static u16 readReg16(u16 nReg);
static void writeReg16(u16 reg, u16 val);
static void writeDisplay16(u32 addr, u32 val);

int get_WP_type(void);

int thread_init(void);
#ifdef CONFIG_FBCON_EPSON_INDIRECT
void indirect_init( u32 RegAddr );
void indirect_write_reg( u16 Index, u16 Value );
u16  indirect_read_reg( u16 Index );
void indirect_write_display(u32 Offset, u32 Value);
static int s1d13xxx_mmap( struct fb_info *fb, struct file *file, struct vm_area_struct *vma );
static int s1d13xxx_set_virtual_vmem(void);
#endif
//-----------------------------------------------------------------------------
//
// Local globals
//
//-----------------------------------------------------------------------------
#if (defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2))

static int s1d13xxx_set_cmap(struct fb_cmap *cmap, int kspc, int con,struct fb_info *info);

static struct fb_ops s1d13xxxfb_ops =
    {
    owner:       THIS_MODULE,
    fb_get_fix:  fbgen_get_fix,
    fb_get_var:  fbgen_get_var,
    fb_set_var:  fbgen_set_var,
    fb_get_cmap: fbgen_get_cmap,
    fb_set_cmap: s1d13xxx_set_cmap,
#ifdef CONFIG_FBCON_EPSON_INDIRECT
    fb_mmap:     s1d13xxx_mmap,
#endif
    };

static int s1d13xxx_set_cmap(struct fb_cmap *cmap, int kspc, int con,
           struct fb_info *info)
    {
    return 0;
    }

#else
static struct fb_ops s1d13xxxfb_ops =
    {
    owner:       THIS_MODULE,
    fb_get_fix:  fbgen_get_fix,
    fb_get_var:  fbgen_get_var,
    fb_set_var:  fbgen_set_var,
    fb_get_cmap: fbgen_get_cmap,
    fb_set_cmap: fbgen_set_cmap,
#ifdef CONFIG_FBCON_EPSON_INDIRECT
    fb_mmap:     s1d13xxx_mmap,
#endif
    };
#endif

/*
 *  In most cases the `generic' routines (fbgen_*) should be satisfactory.
 *  However, you're free to fill in your own replacements.
 */
static struct fbgen_hwswitch s1d13xxx_hwswitch =
    {
    s1d13xxx_detect,
    s1d13xxx_encode_fix,
    s1d13xxx_decode_var,
    s1d13xxx_encode_var,
    s1d13xxx_get_par,
    s1d13xxx_set_par,
    s1d13xxx_getcolreg,
    s1d13xxx_setcolreg,
    s1d13xxx_pan_display,
    s1d13xxx_blank,
    s1d13xxx_set_disp
    };

static FB_INFO_S1D13xxx fb_info;
static struct display disp;
static char __initdata default_fontname[40] = { 0 };

static void writeReg16(u16 reg, u16 val)
{
    #ifdef CONFIG_FBCON_EPSON_INDIRECT
        indirect_write_reg( reg, val );
    #else
	*(u16 *)(fb_info.RegAddr + (reg & ~0x1)) = val;
    #endif
}

static u16 readReg16(u16 reg)
{
    #ifdef CONFIG_FBCON_EPSON_INDIRECT
        return indirect_read_reg(reg);
    #else
        return *(u16 *)(fb_info.RegAddr + (reg & ~0x1));
    #endif
}


static void writeDisplay16(u32 addr, u32 val)
{
    #ifdef CONFIG_FBCON_EPSON_INDIRECT
        indirect_write_display( addr, val );
    #else
	*(u16 *)(fb_info.VmemAddr + (addr & ~0x1)) = val;
    #endif
}

static struct
{
	unsigned int display_width;
	unsigned int display_height;
	unsigned int scan_lines;
}displaySize = {640, 480, 1280};

#ifdef CONFIG_FBCON_EPSON_INDIRECT
void copy_from_virtual_to_real( void )
{
    u32   size=PANEL_SIZE, timeout = LOOP_LIMIT;
    u32   indirect_base_addr = (u32)fb_info.RegAddr;
    u16*  pSource = (u16*) fb_info.VirtualVmemAddr;
    u16*  pIndex  = (u16*) indirect_base_addr;
    u16*  pData   = (u16*) (indirect_base_addr+4);
    u16*  pStatus = (u16*) (indirect_base_addr+2);

    lock_kernel();

    // Set the memory address.
    *pIndex = REG0024_INDIRECT_INTERFACE_MEMORY_ADDRESS_1;
    *pData  = 0;

    //this line is a workaround to fix an address shifted problem.
    while( ((*pStatus)&1) && timeout-- );// check status

    *pIndex = REG0022_INDIRECT_INTERFACE_MEMORY_ADDRESS_0;
    *pData  = 0;

     // Point to the indirect memory port
    while( ((*pStatus)&1) && timeout-- );// check status

    *pIndex = REG0028_INDIRECT_INTERFACE_MEMORY_DATA_PORT;

    do
    {
        *pData = *pSource++;
    }
    while(--size);

    unlock_kernel();
}


static int s1d13xxx_mmap( struct fb_info *fb, struct file *file, struct vm_area_struct *vma )
    {

    vma->vm_flags |= VM_RESERVED;
    if( remap_page_range( vma->vm_start, virt_to_phys((void*)fb_info.VirtualVmemAddr),
      vma->vm_end - vma->vm_start, vma->vm_page_prot ) )
        return -EINVAL;

    return 0;
    }


static int s1d13xxx_set_virtual_vmem(void)
    {
    u32 order = 0;
    u32 size = SIZE_REQUIRED;
    u32 addr;

    while ( size > (PAGE_SIZE * (1 << order)))
        order++;

    fb_info.VirtualVmemAddr = __get_free_pages( GFP_KERNEL, order );

    if( fb_info.VirtualVmemAddr==0 )
            return 1;

    for( addr=fb_info.VirtualVmemAddr; addr<(fb_info.VirtualVmemAddr+size); addr+=PAGE_SIZE )
        mem_map_reserve( virt_to_page( addr ) );

    return 0;
    }

#endif



void do_frame_transfer( void )
{
        if( ( readReg16(REG0038_LCD_INTERFACE_STATUS) & 0x0001 ) == 0)
            writeReg16(REG003A_LCD_INTERFACE_FRAME_TRANSFER, 0x0001); // transfer one frame
}

//-----------------------------------------------------------------------------
//
// This function should detect the current video mode settings and store
// it as the default video mode
//
//-----------------------------------------------------------------------------
static void s1d13xxx_detect(void)
    {
    }

//-----------------------------------------------------------------------------
//
// Header File that includes a BMP image used during startup of the kernel
//
//-----------------------------------------------------------------------------

#include "wp4Data.h"
#include "wp3Data.h"

//-----------------------------------------------------------------------------
//
// This function should fill in the 'fix' structure based on the values
// in the `par' structure.
//
//-----------------------------------------------------------------------------
static int operaFirstGo = 0;

static int s1d13xxx_encode_fix(struct fb_fix_screeninfo *fix, const void *par,
                struct fb_info_gen *info)
    {
    u32 addr = 0;

	u32 row, col, offset = 0;

	unsigned char size =1;
	int type = 0;

	type = panelReadType();

	if( type == 2 )
	{
		size = 2;
		
		u32 startRow = (displaySize.display_height - (wp3Data.height/size)) / 2;
		u32 endRow = startRow + (wp3Data.height/size);
		
		u32 startCol = (displaySize.display_width - (wp3Data.width/size)) / 2;
		u32 endCol = startCol + (wp3Data.width/size);
		
		if( !operaFirstGo )
		{
			for(row = 0; row < displaySize.display_height; row++)
			{
				if(row > startRow && row <= endRow) 
				{
					offset += (wp3Data.width * (size - 1));
				}
				
				for(col = 0; col < displaySize.display_width; col++) 
				{
					if(row < startRow || row >= endRow) 
					{
						writeDisplay16(addr, 0xdf1c);
					}
					else 
					{
						if(col >= startCol && col < endCol) 
						{
							writeDisplay16(addr, wp3Data.pixel_data[offset+=size]);
						}
						else 
						{
							writeDisplay16(addr, 0xdf1c);
						}
					}
					addr += (S1D_DISPLAY_BPP/8);
				}
			}
			operaFirstGo = 1;
		}
	}
	else
	{
		u32 startRow = (displaySize.display_height - (wp4Data.height/size)) / 2;
		u32 endRow = startRow + (wp4Data.height/size);
		
		u32 startCol = (displaySize.display_width - (wp4Data.width/size)) / 2;
		u32 endCol = startCol + (wp4Data.width/size);
		
		if( !operaFirstGo )
		{
			for(row = 0; row < displaySize.display_height; row++)
			{
				if(row > startRow && row <= endRow) 
				{
					offset += (wp4Data.width * (size - 1));
				}
				
				for(col = 0; col < displaySize.display_width; col++) 
				{
					if(row < startRow || row >= endRow) 
					{
						writeDisplay16(addr, 0xdf1c);
					}
					else 
					{
						if(col >= startCol && col < endCol) 
						{
							writeDisplay16(addr, wp4Data.pixel_data[offset+=size]);
						}
						else 
						{
							writeDisplay16(addr, 0xdf1c);
						}
					}
					addr += (S1D_DISPLAY_BPP/8);
				}
			}
			operaFirstGo = 1;
		}
	}
	
    strcpy(fix->id, fb_info.gen.info.modename);

    fix->type = FB_TYPE_PACKED_PIXELS;  /* see FB_TYPE_* in fb.h    */
    fix->type_aux = 0;                  /* Interleave for interleaved Planes */

    /* set fix->line_length to the length of the line in bytes */
    fix->line_length = displaySize.scan_lines;

#ifdef FBCON_HAS_MFB
    fix->visual = FB_VISUAL_MONO10;
#endif

#ifdef FBCON_HAS_CFB2
    fix->visual = FB_VISUAL_PSEUDOCOLOR;
#endif

#ifdef FBCON_HAS_CFB4
    fix->visual = FB_VISUAL_PSEUDOCOLOR;
#endif

#ifdef FBCON_HAS_CFB8
    fix->visual = FB_VISUAL_PSEUDOCOLOR;
#endif

#ifdef FBCON_HAS_CFB16
    fix->visual = FB_VISUAL_TRUECOLOR;
#endif

#ifdef FBCON_HAS_CFB32
    fix->visual = FB_VISUAL_TRUECOLOR;
#endif

#ifdef CONFIG_FBCON_EPSON_PCI
    #if (defined (CONFIG_FBCON_EPSON_S1D13705) || (CONFIG_FBCON_EPSON_S1D13704))

    /* Start and length of frame buffer mem */
    fix->smem_start = fb_info.PhysAddr;
    fix->smem_len   = S1D_PHYSICAL_VMEM_SIZE;

        /* Start and length of Memory Mapped I/O */
        fix->mmio_len   = S1D_PHYSICAL_REG_SIZE;

        #ifdef CONFIG_FBCON_EPSON_S1D13705
            fix->mmio_start = (fb_info.PhysAddr+0x1FFE0);
        #endif

        #ifdef CONFIG_FBCON_EPSON_S1D13704
            fix->mmio_start = (fb_info.PhysAddr+0x0FFE0);
        #endif

    #else
        #ifdef CONFIG_FBCON_EPSON_INDIRECT
            /* Start and length of Memory Mapped I/O */
            fix->mmio_start = (u32)fb_info.PhysAddr;;
            fix->mmio_len   = S1D_PHYSICAL_REG_SIZE;

            /* Start and length of frame buffer mem */
            fix->smem_start = (u32)virt_to_phys((void *)fb_info.VirtualVmemAddr);
            fix->smem_len   = S1D_PHYSICAL_VMEM_REQUIRED;
        #else
            /* Start and length of Memory Mapped I/O */
            fix->mmio_start = fb_info.PhysAddr;
            fix->mmio_len   = S1D_PHYSICAL_REG_SIZE;
           /* Start and length of frame buffer mem */
            fix->smem_start = (fb_info.PhysAddr+S1D_PHYSICAL_VMEM_ADDR);
            fix->smem_len   = S1D_PHYSICAL_VMEM_SIZE;
       #endif
    #endif

#else
    /* Start and length of Memory Mapped I/O */
    fix->mmio_start = S1D_PHYSICAL_REG_ADDR;
    fix->mmio_len   = S1D_PHYSICAL_REG_SIZE;

    /* Start and length of frame buffer mem */
    fix->smem_start = S1D_PHYSICAL_VMEM_ADDR;
    fix->smem_len   = S1D_PHYSICAL_VMEM_SIZE;
#endif

    fix->xpanstep = fix->ypanstep = fix->ywrapstep = 0;
    fix->accel = FB_ACCEL_NONE;

    return 0;
    }

//-----------------------------------------------------------------------------
//
//  Get the video params out of 'var'. If a value doesn't fit, round it up,
//  if it's too big, return -EINVAL.
//
//  Suggestion: Round up in the following order: bits_per_pixel, xres,
//  yres, xres_virtual, yres_virtual, xoffset, yoffset, grayscale,
//  bitfields, horizontal timing, vertical timing.
//
//-----------------------------------------------------------------------------
static int s1d13xxx_decode_var(const struct fb_var_screeninfo *var, void *fb_par,
                struct fb_info_gen *info)
    {
    return 0;
    }

//-----------------------------------------------------------------------------
//
// Fill the 'var' structure based on the values in 'par' and maybe other
// values read out of the hardware.
//
//-----------------------------------------------------------------------------
static int s1d13xxx_encode_var(struct fb_var_screeninfo *var, const void *fb_par,
                struct fb_info_gen *info)
{
	u32 size = PANEL_SIZE, addr = 0;

	u32 row, col, offset = 0;

	/*u32 startRow = (displaySize.display_height - siemens_splash.height) / 2;
	u32 endRow = startRow + siemens_splash.height;

	u32 startCol = (displaySize.display_width - siemens_splash.width) / 2;
	u32 endCol = startCol + siemens_splash.width;

	for(row = 0; row < displaySize.display_height; row++) {
		for(col = 0; col < displaySize.display_width; col++) {
			writeDisplay16(addr++, 0xdf1c);
		}
	}*/

    /*do
	{
		writeDisplay16(addr++, 0xd1fc);
	}while(--size);*/

    var->xres =  displaySize.display_width;     //visible resolution
    var->yres =  displaySize.display_height;

    var->xres_virtual = var->xres;
    var->yres_virtual = var->yres;
    var->xoffset = var->yoffset = 0;
    var->bits_per_pixel = S1D_DISPLAY_BPP;
    var->grayscale = 0;
    var->nonstd = 0;                 /* != 0 Non standard pixel format */
    var->activate = FB_ACTIVATE_NOW; /* see FB_ACTIVATE_*              */
    var->height = -1;                /* height of picture in mm        */
    var->width  = -1;                /* width of picture in mm         */
    var->accel_flags = 0;            /* acceleration flags (hints      */
    var->pixclock = S1D_DISPLAY_PCLK;
    var->right_margin = 0;
    var->lower_margin = 0;
    var->hsync_len = 0;
    var->vsync_len = 0;
    var->left_margin = 0;
    var->upper_margin = 0;
    var->sync = 0;
    var->vmode = FB_VMODE_NONINTERLACED;

#if defined (FBCON_HAS_MFB) || defined (FBCON_HAS_CFB2) || \
    defined (FBCON_HAS_CFB4) || defined (FBCON_HAS_CFB8)
    var->red.offset = var->green.offset = var->blue.offset = 0;
    var->red.length = var->green.length = var->blue.length = S1D_DISPLAY_BPP;
#endif

#ifdef FBCON_HAS_CFB16  // RGB 565
    var->red.offset = 11;
    var->red.length = 5;
    var->green.offset = 5;
    var->green.length = 6;
    var->blue.offset = 0;
    var->blue.length = 5;
#endif

#ifdef FBCON_HAS_CFB32
    var->red.length = var->green.length = var->blue.length = 8;
    var->blue.offset = 0;
    var->green.offset = 8;
    var->red.offset = 16;
#endif

    var->red.msb_right = var->green.msb_right = var->blue.msb_right = 0;
    var->transp.offset = var->transp.length = var->transp.msb_right = 0;

    return 0;
    }

//-----------------------------------------------------------------------------
//
// Set the 'par' according to hardware
//
//-----------------------------------------------------------------------------
static void s1d13xxx_get_par(void *fb_par, struct fb_info_gen *info)
    {
    }

//-----------------------------------------------------------------------------
//
// Set the hardware according to 'par'.
//
//-----------------------------------------------------------------------------
static void s1d13xxx_set_par(const void *fb_par, struct fb_info_gen *info)
    {
    }

//-----------------------------------------------------------------------------
//
// Read a single color register and split it into colors/transparent.
// The return values must have a 16 bit magnitude.
// Return != 0 for invalid regno.
//
//-----------------------------------------------------------------------------
static int s1d13xxx_getcolreg(unsigned regno, unsigned *red, unsigned *green,
                unsigned *blue, unsigned *transp, struct fb_info *info)
    {
    u32 r,g,b;
#ifdef CONFIG_FBCON_EPSON_INDIRECT
    S1D_VALUE reg1, reg2;
#endif

    if (regno >= S1D_PALETTE_SIZE)
        return 1;

#ifdef CONFIG_FBCON_EPSON_INDIRECT
    reg1=readReg16( 0x400+regno);
    reg2=readReg16( 0x402+regno);
    g = (reg1>>8)&0xFF;
    r = reg1&0xFF;
    b = reg2&0xFF;
#else
    S1D_READ_PALETTE(fb_info.RegAddr,regno,r,g,b);
#endif

    *red    = (r << 8) | r;
    *green  = (g << 8) | g;
    *blue   = (b << 8) | b;
    *transp = 0;

    return 0;
    }

//----------------------------------------------------------------------------
//
// Set a single color register. The values supplied have a 16 bit
// magnitude.
// Return != 0 for invalid regno.
//
// We get called even if we specified that we don't have a programmable palette
// or in direct/true color modes!
//----------------------------------------------------------------------------
static int s1d13xxx_setcolreg(unsigned regno, unsigned red, unsigned green,
                unsigned blue, unsigned transp, struct fb_info *info)
    {
//  printk("S1D13XXX: (setcolreg) i:%d r=%x g=%x b=%x\n", regno,red,green,blue);

    if (regno >= S1D_PALETTE_SIZE)
        return 1;

#ifdef CONFIG_FBCON_EPSON_INDIRECT
    writeReg16( 0x400+regno, (green & 0xff00) | red>>8 );
    writeReg16( 0x402+regno, blue>>8 );
#else
    S1D_WRITE_PALETTE(fb_info.RegAddr,regno,red>>8,green>>8,blue>>8);
#endif

#if defined(FBCON_HAS_CFB16)
    // Make the first 16 colors of the palette available to fbcon

    if (regno < 16)
        {
        fb_info.cfb16[regno] =
            (red & 0xf800)|((green & 0xfc00) >> 5)|((blue & 0xf800) >> 11);
        }
#endif

#ifdef FBCON_HAS_CFB32
    // Make the first 16 colors of the palette available to fbcon

    if (regno < 16)
        {
        red >>= 8;
        green >>= 8;
        blue >>= 8;
        fb_info.cfb32[regno] = (red << 16) | (green << 8) | (blue);
        }
#endif
    return 0;
    }

//-----------------------------------------------------------------------------
//
// Pan (or wrap, depending on the `vmode' field) the display using the
// `xoffset' and `yoffset' fields of the `var' structure.
// If the values don't fit, return -EINVAL.
//
//-----------------------------------------------------------------------------
static int s1d13xxx_pan_display(const struct fb_var_screeninfo *var,
                struct fb_info_gen *info)
    {
    return 0;
    }

//-----------------------------------------------------------------------------
//
// Blank the screen if blank_mode != 0, else unblank. If blank == NULL
// then the caller blanks by setting the CLUT (Color Look Up Table) to all
// black. Return 0 if blanking succeeded, != 0 if un-/blanking failed due
// to e.g. a video mode which doesn't support it. Implements VESA suspend
// and powerdown modes on hardware that supports disabling hsync/vsync:
//
// 0 = unblank
// 1 = blank
// 2 = suspend vsync
// 3 = suspend hsync
// 4 = off
//
//-----------------------------------------------------------------------------
static int s1d13xxx_blank(int blank_mode, struct fb_info_gen *info)
    {
    return 0;
    }

//-----------------------------------------------------------------------------
//
// Fill in a pointer with the virtual address of the mapped frame buffer.
// Fill in a pointer to appropriate low level text console operations (and
// optionally a pointer to help data) for the video mode `par' of your
// video hardware. These can be generic software routines, or hardware
// accelerated routines specifically tailored for your hardware.
// If you don't have any appropriate operations, you must fill in a
// pointer to dummy operations, and there will be no text output.
//
//-----------------------------------------------------------------------------
static void s1d13xxx_set_disp(const void *par, struct display *disp,
                struct fb_info_gen *info)
    {
#ifdef CONFIG_FBCON_EPSON_INDIRECT
    disp->screen_base = (u8 *)fb_info.VirtualVmemAddr;
#else
    disp->screen_base = fb_info.VmemAddr;
#endif

#ifdef FBCON_HAS_MFB
    disp->dispsw = &fbcon_mfb;
#endif

#ifdef FBCON_HAS_CFB2
    disp->dispsw = &fbcon_cfb2;
#endif

#ifdef FBCON_HAS_CFB4
    disp->dispsw = &fbcon_cfb4;
#endif

#ifdef FBCON_HAS_CFB8
    disp->dispsw = &fbcon_cfb8;
#endif

#ifdef FBCON_HAS_CFB16
    disp->dispsw = &fbcon_cfb16;
    disp->dispsw_data = fb_info.cfb16;  /* console palette */
#endif

#ifdef FBCON_HAS_CFB32
    disp->dispsw = &fbcon_cfb32;
    disp->dispsw_data = fb_info.cfb32;  /* console palette */
#endif

    disp->scrollmode = SCROLL_YREDRAW;
    }

//-----------------------------------------------------------------------------
//
// Detect our PCI bridge adapter card, get the BAR[0]. Card is assumed to be
// enabled.
//
//-----------------------------------------------------------------------------
#ifdef CONFIG_FBCON_EPSON_PCI
static int __init
s1d13xxx_IsOurID(unsigned vID, unsigned dID)
    {
    if (vID == 0x10F4 && dID == 0x1300)
        return 1;

    if (vID == 0x14EB)
        {
        if (dID >= 0x0000 && dID < 0x002A)
            return 1;
        if (dID > 0x002A && dID <= 0x002F)
            return 1;
        }
    return 0;
    }

unsigned long __init
s1d13xxx_get_physical_address(void)
    {
    struct pci_dev *pdev;

    pdev = NULL;

    pci_for_each_dev(pdev)
        {
        if (s1d13xxx_IsOurID(pdev->vendor,pdev->device))
            return (pci_resource_start(pdev, 0) & PCI_BASE_ADDRESS_MEM_MASK);
        }

    return 0;
    }
#endif

#ifdef CONFIG_FBCON_EPSON_S1D13733

#define SIEMENS_FPGA		(KSEG1+0x1c000000)
#define FPGA_DISP_RESET		((volatile u8*)((SIEMENS_FPGA) + 0x0014))
#define FPGA_PAN_CNTRL		((volatile u8*)((SIEMENS_FPGA) + 0x001E))

#define FPGA_PANEL_POWER_EN_BIT (1<<0)
#define FPGA_PANEL_SPI_EN_BIT (1<<1)
#define FPGA_PANEL_SPI_CLK_BIT (1<<2)
#define FPGA_PANEL_SPI_DATA_BIT (1<<3)
#define FPGA_PANEL_SPI_DIR_BIT (1<<4)
#define FPGA_PANEL_DEFAULT_WRITE_BITS ( FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_EN_BIT | FPGA_PANEL_SPI_DIR_BIT)
#define FPGA_PANEL_DEFAULT_READ_BITS ( FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_EN_BIT )

static void panelSPIDelay( u8 data )
{
	udelay( data );
}

static void panelSPIPanelWrite( u8 output )
{
//	printk("mw.b bc00001e %02X\n",output);
	*FPGA_PAN_CNTRL = output;
	*FPGA_PAN_CNTRL = output;
}

static u8 panelSPIPanelRead( void )
{
	u8 output = *FPGA_PAN_CNTRL;
//	printk("mr.b bc00001e %02X\n",output);
	return( output );
}

static u8 panelSPIPanelReadBit( void )
{
	u8 bit = (panelSPIPanelRead() & FPGA_PANEL_SPI_DATA_BIT) >> 3;
//	printk("read bit %02X\n",bit);
	return( bit );
}

static void panelSPI_CS_set( void )
{
	/* Idle SPI clock high */
	u8 output = FPGA_PANEL_SPI_EN_BIT | FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_CLK_BIT;
	panelSPIPanelWrite( output );
}

static void panelSPI_CS_clear( void )
{
	/* Idle SPI clock high */
	u8 output = FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_CLK_BIT;
	panelSPIPanelWrite( output );
}

static void panelSPIBitWrite( u8 data )
{
	int i;
	u8 output;
	u8 base = FPGA_PANEL_DEFAULT_WRITE_BITS;
	for( i=0; i < 8; i++ )
	{
		/*MSB first*/
		output = base | (((data<<i) & 0x80) >> 4);
		panelSPIPanelWrite( output );
		output |= FPGA_PANEL_SPI_CLK_BIT;
		panelSPIPanelWrite( output );
	}
}

static u8 panelSPIBitRead( void )
{
	int i;
	u8 output;
	u8 base = FPGA_PANEL_DEFAULT_READ_BITS;
	output = 0;
	for( i=0; i < 8; i++ )
	{
		/*MSB first*/
		output <<= 1;
		panelSPIPanelWrite( base );
		panelSPIPanelWrite( base | FPGA_PANEL_SPI_CLK_BIT );
		output |= panelSPIPanelReadBit();
	}
	return( output );
}

static void panelSPIBitWriteReg( u8 reg )
{
	/* When wrtting register set bit 2, the write bit */
	panelSPIBitWrite( (reg << 2) | 0x2 );
}

static void panelSPIBitReadReg( u8 reg )
{
	/* When wrtting register set bit 2, the write bit */
	panelSPIBitWrite( (reg << 2) );
}

static void panelSPISetDir( void )
{
	u8 output = FPGA_PANEL_POWER_EN_BIT | FPGA_PANEL_SPI_DIR_BIT | FPGA_PANEL_SPI_CLK_BIT;
	panelSPIPanelWrite( output );
//	printk("echo\n");
}

static void panelWriteReg( u8 reg, u8 value)
{
//	printk("echo %02X %02X\n",reg, value);
	panelSPISetDir();
	panelSPIBitWriteReg( reg );
	panelSPIBitWrite( value );
	panelSPISetDir();
	panelSPIDelay(2);
}


static u8 panelReadReg( u8 reg )
{
	u8 value;
//	printk("echo %02X\n",reg);
	panelSPISetDir();
	panelSPIBitReadReg( reg );
	value = panelSPIBitRead( );
	panelSPISetDir();
	panelSPIDelay(2);
	return( value );
}

static int panelProgramWintech( void )
{
//	printk("echo Program Display\n");
//	printk("echo Master\n");
//	printk("\r\n Display Parameters: 0x5->0x10, 0x9->0x40\r\n");
	panelWriteReg(0x00, 0x03);
	panelWriteReg(0x01, 0x82);
	panelWriteReg(0x02, 0x12);
	panelWriteReg(0x03, 0xcc);
	panelWriteReg(0x04, 0x27);
	panelWriteReg(0x05, 0x10);
	panelWriteReg(0x06, 0x00);
	panelWriteReg(0x07, 0x30);
	panelWriteReg(0x08, 0x08);
	panelWriteReg(0x09, 0x40);
	panelWriteReg(0x0a, 0x88);
	panelWriteReg(0x0b, 0x88);
	panelWriteReg(0x0c, 0x20);
	panelWriteReg(0x0d, 0x20);
//	printk("echo SLAVE\n");
	panelWriteReg(0x20, 0x03);
	panelWriteReg(0x21, 0x82);
	panelWriteReg(0x22, 0x12);
	panelWriteReg(0x23, 0xcc);
	panelWriteReg(0x24, 0x27);
	panelWriteReg(0x25, 0x10);
	panelWriteReg(0x26, 0x00);
	panelWriteReg(0x27, 0x30);
	panelWriteReg(0x28, 0x08);
	panelWriteReg(0x29, 0x40);
	panelWriteReg(0x2a, 0x88);
	panelWriteReg(0x2b, 0x88);
	panelWriteReg(0x2c, 0x20);
	panelWriteReg(0x2d, 0x20);
	mdelay(200);
	return( 0 );
}

static int panelProgramReadWintech( void )
{
	printk("echo Program Display\n");
	printk("echo Master\n");
	printk("reg %02x = %02x\n",0x00,panelReadReg(0x00));
	printk("reg %02x = %02x\n",0x01,panelReadReg(0x01));
	printk("reg %02x = %02x\n",0x02,panelReadReg(0x02));
	printk("reg %02x = %02x\n",0x03,panelReadReg(0x03));
	printk("reg %02x = %02x\n",0x04,panelReadReg(0x04));
	printk("reg %02x = %02x\n",0x05,panelReadReg(0x05));
	printk("reg %02x = %02x\n",0x06,panelReadReg(0x06));
	printk("reg %02x = %02x\n",0x07,panelReadReg(0x07));
	printk("reg %02x = %02x\n",0x08,panelReadReg(0x08));
	printk("reg %02x = %02x\n",0x09,panelReadReg(0x09));
	printk("reg %02x = %02x\n",0x0a,panelReadReg(0x0a));
	printk("reg %02x = %02x\n",0x0b,panelReadReg(0x0b));
	printk("reg %02x = %02x\n",0x0c,panelReadReg(0x0c));
	printk("reg %02x = %02x\n",0x0d,panelReadReg(0x0d));
	printk("echo SLAVE\n");
	printk("reg %02x = %02x\n",0x20,panelReadReg(0x20));
	printk("reg %02x = %02x\n",0x21,panelReadReg(0x21));
	printk("reg %02x = %02x\n",0x22,panelReadReg(0x22));
	printk("reg %02x = %02x\n",0x23,panelReadReg(0x23));
	printk("reg %02x = %02x\n",0x24,panelReadReg(0x24));
	printk("reg %02x = %02x\n",0x25,panelReadReg(0x25));
	printk("reg %02x = %02x\n",0x26,panelReadReg(0x26));
	printk("reg %02x = %02x\n",0x27,panelReadReg(0x27));
	printk("reg %02x = %02x\n",0x28,panelReadReg(0x28));
	printk("reg %02x = %02x\n",0x29,panelReadReg(0x29));
	printk("reg %02x = %02x\n",0x2a,panelReadReg(0x2a));
	printk("reg %02x = %02x\n",0x2b,panelReadReg(0x2b));
	printk("reg %02x = %02x\n",0x2c,panelReadReg(0x2c));
	printk("reg %02x = %02x\n",0x2d,panelReadReg(0x2d));
	return( 0 );
}

static int videoPanelType;

int panelReadType( void )
{
	return( videoPanelType );
}

static int panelType( void )
{
	u8 value;

	/* Check for hardware version */
	if(get_WP_type() == 3)
	{
		/*This is a WP3 display*/

		*FPGA_PAN_CNTRL = 0;
		displaySize.display_width = 320;
		displaySize.display_height = 240;
		displaySize.scan_lines = 640;

    	videoPanelType = 2;
	}
	else
	{
	//    printk("Panel Control %02x\n",*FPGA_PAN_CNTRL);
		*FPGA_PAN_CNTRL = FPGA_PANEL_POWER_EN_BIT; /* turn on the panel */
	//    printk("Panel Read %02x\n",*FPGA_PAN_CNTRL);
	    panelSPI_CS_set();
	    value = panelReadReg(0);
	    printk("Panel Read reg 0 %02x\n",value);

		if( value != 0xff )
		{
	    	videoPanelType = 1;
	    	/* something is there assume that it is WINTECH panel */
	    	panelProgramReadWintech();
	    	return( videoPanelType );
		}

	    /* nothing there so disable it. Maybe it is a Philips panel on WP4
	     * */
		*FPGA_PAN_CNTRL = 0;

	  	videoPanelType = 0;
  	}

   	return( videoPanelType );
}

static void epsom_power_on_seq(void)
{
    *FPGA_DISP_RESET = 0;         /* display controller reset */
    udelay (1000);              /* 1ms reset time - reset will be */
                                /* in DCTRL_register_test() */

    /* Power on sequence - PLL set */

	/* PLL Power on sequence */
    writeReg16(0x06,0x0003);     /* reset control */
    /* set clock configuration registers */
    writeReg16(0x14,0x0411);
    writeReg16(0x12,0x0001);
    writeReg16(0x0C,0x1027);
    writeReg16(0x0E,0x7390);
    writeReg16(0x10,0x0103);
    writeReg16(0x18,0x0000);
    writeReg16(0x12,0x0000);
    udelay( 10000 );        /* wait 10ms */
    writeReg16(0x14,0x0410);
}

static void setupForPhilipsPanel( void )
{
    u32 i, timeout = LOOP_LIMIT;
    S1D_INDEX s1dReg;
    S1D_VALUE s1dValue;
	printk("Configuring for Philips Panel\n");
    for (i = 0; i < sizeof(aS1DRegs)/sizeof(aS1DRegs[0]); i++)
    {
    s1dReg   = aS1DRegs[i].Index;
    s1dValue = aS1DRegs[i].Value;

    if (s1dReg == REG0034_LCD_INTERFACE_COMMAND || s1dReg == REG0036_LCD_INTERFACE_PARAMETER)
        while ((( readReg16(REG0038_LCD_INTERFACE_STATUS) & 0x0001) != 0) && timeout--)
            continue;

    if (s1dReg == S1D_REGDELAYOFF || s1dReg == S1D_REGDELAYON || s1dReg == S1D_REGDELAYPLL)
        mdelay((int)s1dValue);
    else
        writeReg16( s1dReg, s1dValue);
    }

}

static void setupForPhilipsPanel_01( void )
{
    u32 i;
	S1D_INDEX s1dReg;
    S1D_VALUE s1dValue;

static S1D_REGS aS1DRegs[] =
{
#include "LCD_WP4_regconf.h"
};
	printk("Configuring for Philips Panel\n");

	epsom_power_on_seq();
    udelay(1000);
	for (i = 0; i < sizeof(aS1DRegs)/sizeof(aS1DRegs[0]); i++)
	{
        s1dReg   = aS1DRegs[i].Index;
        s1dValue = aS1DRegs[i].Value;
    	writeReg16( aS1DRegs[i].Index, aS1DRegs[i].Value);
	}
    writeReg16( 0x32, 0x31);
    writeReg16( 0x46, 0xdf);
    writeReg16( 0x50, 0x87);
    udelay(1000);
}

static void setupForWinTechPanel( void )
{
    u32 i;
	S1D_INDEX s1dReg;
    S1D_VALUE s1dValue;

static S1D_REGS aS1DRegs[] =
{
#include "LCD_WP4_regconf.h"
};
    printk("Configuring for Wintech Panel\n");

	epsom_power_on_seq();
	panelSPISetDir();
    udelay(1000);
	for (i = 0; i < sizeof(aS1DRegs)/sizeof(aS1DRegs[0]); i++)
	{
        s1dReg   = aS1DRegs[i].Index;
        s1dValue = aS1DRegs[i].Value;
    	writeReg16( aS1DRegs[i].Index, aS1DRegs[i].Value);
	}


    udelay(1000);
	panelProgramWintech();
    udelay(1000);
//	panelProgramReadWintech();
}

static void setupForOptrixPanel( void )
{
    u32 i;
	S1D_INDEX s1dReg;
    S1D_VALUE s1dValue;

static S1D_REGS aS1DRegs[] =
{
#include "LCD_WP3_regconf.h"
};
	printk("Configuring for Optrex Panel\n");

	epsom_power_on_seq();
    udelay(1000);
	for (i = 0; i < sizeof(aS1DRegs)/sizeof(aS1DRegs[0]); i++)
	{
        s1dReg   = aS1DRegs[i].Index;
        s1dValue = aS1DRegs[i].Value;
    	writeReg16( aS1DRegs[i].Index, aS1DRegs[i].Value);
	}

    udelay(1000);

}

static void modifyPanelRegisters( void )
{
	switch( panelType() )
	{
		case 0: setupForPhilipsPanel(); break;
		case 1: setupForWinTechPanel(); break;
		case 2: setupForOptrixPanel(); break;
	}
}

#endif

//-----------------------------------------------------------------------------
//
// Initialize the chip and the frame buffer driver.
//
//-----------------------------------------------------------------------------
int __init
s1d13xxxfb_init(char *dummy)
    {
#ifndef CONFIG_FBCON_EPSON_S1D13733
    u32 i, timeout = LOOP_LIMIT;
    S1D_INDEX s1dReg;
    S1D_VALUE s1dValue;
#endif

#ifdef CONFIG_FBCON_EPSON_PCI

    unsigned long phAddr = s1d13xxx_get_physical_address();

    #ifdef CONFIG_FBCON_EPSON_S1D13705
       fb_info.VmemAddr = (unsigned char*) ioremap(phAddr,0x1FFE0+S1D_PHYSICAL_REG_SIZE);
       fb_info.RegAddr  = fb_info.VmemAddr + 0x1FFE0;
    #else
       #ifdef CONFIG_FBCON_EPSON_S1D13704
          fb_info.VmemAddr = (unsigned char*) ioremap(phAddr,0xDFFE0+S1D_PHYSICAL_REG_SIZE);
          fb_info.RegAddr  = fb_info.VmemAddr + 0xDFFE0;
       #else
          fb_info.RegAddr  = (unsigned char*) ioremap(phAddr,S1D_PHYSICAL_REG_SIZE);
          fb_info.VmemAddr = (unsigned char*) ioremap_nocache(phAddr+S1D_PHYSICAL_VMEM_ADDR,S1D_PHYSICAL_VMEM_REQUIRED);
       #endif
    #endif

    fb_info.PhysAddr = phAddr;

#else
    fb_info.VmemAddr = (unsigned char*) ioremap_nocache(S1D_PHYSICAL_VMEM_ADDR,S1D_PHYSICAL_VMEM_SIZE);
    fb_info.RegAddr  = (unsigned char*) ioremap(S1D_PHYSICAL_REG_ADDR,S1D_PHYSICAL_REG_SIZE);
#endif

#if (defined (CONFIG_FBCON_EPSON_S1D13714) || (CONFIG_FBCON_EPSON_S1D13715) || \
    (CONFIG_FBCON_EPSON_S1D13732) || (CONFIG_FBCON_EPSON_S1D13733))
    fb_info.VmemAddr += MAIN_WINDOW_OFFSET;
#endif

    if (!fb_info.VmemAddr || !fb_info.RegAddr)
        {
        printk("S1D13xxx_init: ioremap() returned NULL\n");
        return -EINVAL;
        }

#ifdef CONFIG_FBCON_EPSON_INDIRECT
    indirect_init( (u32)fb_info.RegAddr );

    if(s1d13xxx_set_virtual_vmem())
        {
        printk("s1d13xxxfb_init: _get_free_pages() failed.\n");
        return -EINVAL;
        }
#endif


#ifdef CONFIG_FBCON_EPSON_S1D13733
	*FPGA_DISP_RESET = 0; /* turn on the display */

	modifyPanelRegisters();
#else

    for (i = 0; i < sizeof(aS1DRegs)/sizeof(aS1DRegs[0]); i++)
        {
        s1dReg   = aS1DRegs[i].Index;
        s1dValue = aS1DRegs[i].Value;

        if (s1dReg == REG0034_LCD_INTERFACE_COMMAND || s1dReg == REG0036_LCD_INTERFACE_PARAMETER)
            while ((( readReg16(REG0038_LCD_INTERFACE_STATUS) & 0x0001) != 0) && timeout--)
                continue;

        if (s1dReg == S1D_REGDELAYOFF || s1dReg == S1D_REGDELAYON || s1dReg == S1D_REGDELAYPLL)
            mdelay((int)s1dValue);
        else
            writeReg16( s1dReg, s1dValue);
        }
#endif

#ifdef CONFIG_FBCON_EPSON_RAM_BASED_LCD
    for (i = 0; i < sizeof(aLCDRegs)/sizeof(aLCDRegs[0]); i++)
        {
        s1dReg   = aLCDRegs[i].Index;
        s1dValue = aLCDRegs[i].Value;

         if (s1dReg == REG0034_LCD_INTERFACE_COMMAND || s1dReg == REG0036_LCD_INTERFACE_PARAMETER)
            {
                 while ((( readReg16(REG0038_LCD_INTERFACE_STATUS) & 0x0001) != 0) && timeout--)
                 continue;
            }

        if (s1dReg == S1D_REGDELAYOFF || s1dReg == S1D_REGDELAYON || s1dReg == S1D_REGDELAYPLL )
            mdelay((int)s1dValue);
        else
            writeReg16(s1dReg,s1dValue);
        }
#endif

    strcpy(fb_info.gen.info.modename, "s1d13xxx");
    fb_info.gen.info.changevar = NULL;
    fb_info.gen.info.node = -1;
    fb_info.gen.info.fbops = &s1d13xxxfb_ops;
    fb_info.gen.info.disp = &disp;

    fb_info.gen.info.switch_con = &fbgen_switch;
    fb_info.gen.info.updatevar = &fbgen_update_var;
    fb_info.gen.info.blank = &fbgen_blank;

    strcpy(fb_info.gen.info.fontname, default_fontname);
    fb_info.gen.parsize = 0;
    fb_info.gen.info.flags = FBINFO_FLAG_DEFAULT;
    fb_info.gen.fbhw = &s1d13xxx_hwswitch;
    fb_info.gen.fbhw->detect();

    fbgen_get_var(&disp.var, -1, &fb_info.gen.info);
    disp.var.activate = FB_ACTIVATE_NOW;
    fbgen_do_set_var(&disp.var, 1, &fb_info.gen);
    fbgen_set_disp(-1, &fb_info.gen);
    fbgen_install_cmap(0, &fb_info.gen);
    thread_init();
    if (register_framebuffer(&fb_info.gen.info) < 0)
        return -EINVAL;

    return 0;
    }


//-----------------------------------------------------------------------------
//
// Parse user speficied options (`video=s1d13xxxfb:')
//
//-----------------------------------------------------------------------------
int __init
s1d13xxxfb_setup(char *options, int *ints)
    {
    return 0;
    }

/* ------------------------------------------------------------------------- */

/*
 *  Modularization
 */
#ifdef MODULE
int init_module(void)
    {
    s1d13xxxfb_init();
    thread_init();
    return 0;
    }

void cleanup_module(void)
    {
    unregister_framebuffer(&fb_info.gen.info);
    thread_cleanup();
    }
#endif
