/************************************************************************
 *
 * Copyright (c) 2005
 * Infineon Technologies AG
 * St. Martin Strasse 53; 81669 Muenchen; Germany
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 ************************************************************************/

/* Group definitions for Doxygen */
/** \addtogroup IOCTL Defined IOCTLs */

#ifndef __IFX_SSC_H
#define __IFX_SSC_H

#define PORT_CNT        1

/**
 * SSC Statistics.
 * Symbolic constants to be used in SSC routines 
 */
struct ifx_ssc_statistics {
    unsigned int  abortErr;   /**< abort error */       
    unsigned int  modeErr;    /**< master/slave mode error */       
    unsigned int  txOvErr;    /**< TX Overflow error */         
    unsigned int  txUnErr;    /**< TX Underrun error */
    unsigned int  rxOvErr;    /**< RX Overflow error */
    unsigned int  rxUnErr;    /**< RX Underrun error */
    unsigned int  rxBytes;    /**< Received bytes */
    unsigned int  txBytes;    /**< Transmitted bytes */
    unsigned int  unknownErr;  /**< Unknown error cause */
};

/**
 * SSC Hardware Options.
 */
struct ifx_ssc_hwopts {
    unsigned int  AbortErrDetect :1;   /**< Abort Error detection (in slave mode) */
    unsigned int  rxOvErrDetect  :1;   /**< Receive Overflow Error detection */
    unsigned int  rxUndErrDetect :1;   /**< Receive Underflow Error detection */
    unsigned int  txOvErrDetect  :1;   /**< Transmit Overflow Error detection */
    unsigned int  txUndErrDetect :1;   /**< Transmit Underflow Error detection */
    unsigned int  echoMode       :1;   /**< Echo mode */
    unsigned int  loopBack       :1;   /**< Loopback mode */
    unsigned int  idleValue      :1;   /**< Idle value */
    unsigned int  clockPolarity  :1;   /**< Idle clock is high or low */
    unsigned int  clockPhase     :1;   /**< Tx on trailing or leading edge*/
    unsigned int  headingControl :1;   /**< LSB first or MSB first */
    unsigned int  dataWidth      :6;   /**< Date width from 2 up to 32 bits */
    unsigned int  masterSelect   :1;   /**< Master or Slave mode */
    unsigned int  modeRxTx       :2;   /**< Rx/Tx mode */
    unsigned int  gpoCs          :8;   /**< Choose outputs to use for chip select */
    unsigned int  gpoInv         :8;   /**< Invert GPO outputs */
    unsigned int  dmamode        :1;   /**< Use DMA for Rx and Tx */
    unsigned int  slaveCs        :1;   /**< Use chip select in slave mode */
};

/**
 * SSC Framing Options.
 */
struct ifx_ssc_frm_opts {
    unsigned int    FrameEnable;    /**< See description of SFCON.SFEN in datasheet */
    unsigned int    IntBefore;      /**< See description of SFCON.IBEN in datasheet */
    unsigned int    IntAfter;       /**< See description of SFCON.IAEN in datasheet */
    unsigned int    DataLength;     /**< See description of SFCON.DLEN in datasheet */
    unsigned int    PauseLength;    /**< See description of SFCON.PLEN in datasheet */
    unsigned int    IdleData;       /**< See description of SFCON.IDAT in datasheet */
    unsigned int    IdleClock;      /**< See description of SFCON.ICLK in datasheet */
    unsigned int    StopAfterPause; /**< See description of SFCON.STOP in datasheet */
};
/**
 * SSC Framing Status.
 */
struct ifx_ssc_frm_status {
    unsigned int    DataBusy;       /**< See description of SFSTAT.DBSY in datasheet */
    unsigned int    PauseBusy;      /**< See description of SFSTAT.PBSY in datasheet */
    unsigned int    DataCount;      /**< See description of SFSTAT.DCNT in datasheet */
    unsigned int    PauseCount;     /**< See description of SFSTAT.PCNT in datasheet */
    unsigned int    EnIntAfterData; /**< See description of SFCON.IBEN in datasheet */
    unsigned int    EnIntAfterPause;/**< See description of SFCON.IAEN in datasheet */
};
/**
 * SSC Framing Callback.
 */
struct ifx_ssc_frm_callback {
    void (*IntCallback)(int);     /**< When Frame Interrupt is enabled this function is called if provided */
};

typedef struct {
    char    *buf;
    size_t  len;
} ifx_ssc_buf_item_t;

#ifdef __KERNEL__
#define IFX_SSC_IS_MASTER(p) ((p)->opts.masterSelect == SSC_MASTER_MODE)
/**
 * SSC Port Structure.
 * Data and configuration structure describing the respective port.
 */
struct ifx_ssc_port {
    unsigned long               mapbase;
    struct ifx_ssc_hwopts       opts;                         /**< Port hardware settings */
    struct ifx_ssc_statistics   stats;                        /**< Port statistics */
    struct ifx_ssc_frm_status   frm_status;                   /**< Port framing status */
    struct ifx_ssc_frm_opts     frm_opts;                     /**< Port framing options */
    struct ifx_ssc_frm_callback frm_callback;                 /**< Port framing callback */
    wait_queue_head_t           rwait, pwait;                 /**< Wait queues for ifx_ssc_read() */
    int                         port_nr;                      /**< Port index */
    char                        port_is_available;            /**< Port availability (hardware multiplexing) */
    char                        port_is_open;                 /**< Exclusive open */
    char * volatile             rxbuf;                        /**< Buffer for RX */
    char * volatile             rxbuf_end;                    /**< Buffer end pointer for RX */
    char * volatile             rxbuf_ptr;                    /**< Buffer write pointer for RX */
    char * volatile             txbuf;                        /**< Buffer for TX */
    char * volatile             txbuf_end;                    /**< Buffer end pointer for TX */
    char * volatile             txbuf_ptr;                    /**< Buffer read pointer for TX */
    unsigned int                baud;                         /**< Current baudrate */
    unsigned int                txirq, rxirq, errirq, frmirq; /**< Channel interrupts */
    char                        irqname[4][20];               /**< Interrupt names */
    struct dma_device_info      *dma_dev;                     /**< DMA device structure */
    volatile int                frameNotDone;
    struct semaphore            rsem;
};

/* default values for SSC configuration */
/* values of CON */
#define IFX_SSC_DEF_IDLE_DATA       1  /* enable */
#define IFX_SSC_DEF_BYTE_VALID_CTL  1  /* enable */
//#define IFX_SSC_DEF_DATA_WIDTH      32 /* bits */
#define IFX_SSC_DEF_DATA_WIDTH      8 /* bits */
#define IFX_SSC_DEF_ABRT_ERR_DETECT 0  /* disable */
#define IFX_SSC_DEF_RO_ERR_DETECT   1  /* enable */
#define IFX_SSC_DEF_RU_ERR_DETECT   0  /* disable */
#define IFX_SSC_DEF_TO_ERR_DETECT   0  /* disable */
#define IFX_SSC_DEF_TU_ERR_DETECT   0  /* disable */
#define IFX_SSC_DEF_LOOP_BACK       0  /* disable */
#define IFX_SSC_DEF_ECHO_MODE       0  /* disable */
#define IFX_SSC_DEF_CLOCK_POLARITY  0  /* low */
#define IFX_SSC_DEF_CLOCK_PHASE     1  /* 0: shift on leading edge, latch on trailling edge, 1, otherwise */
#define IFX_SSC_DEF_HEADING_CONTROL IFX_SSC_MSB_FIRST
#define IFX_SSC_DEF_MODE_RXTX       IFX_SSC_MODE_RXTX
/* other values */
#define IFX_SSC_DEF_MASTERSLAVE     IFX_SSC_MASTER_MODE /* master */
//#define IFX_SSC_DEF_BAUDRATE        1000000
#define IFX_SSC_DEF_BAUDRATE        250000
#define IFX_SSC_DEF_RMC             0x1

#define IFX_SSC_DEF_TXFIFO_FL       1
#define IFX_SSC_DEF_RXFIFO_FL       2

#if 1 /* TODO: make chip selects selectable */
#define IFX_SSC_DEF_GPO_CS      2   /* no chip select */
#define IFX_SSC_DEF_GPO_INV     0   /* no chip select */
#else
#error "what is your Chip Select???"
#endif
#define IFX_SSC_DEF_SFCON       0   /* no serial framing */
#define IFX_SSC_DEF_IRNEN       IFX_SSC_T_BIT | IFX_SSC_R_BIT | IFX_SSC_E_BIT
#endif /* __KERNEL__ */

/* Macros to configure SSC hardware */
/* headingControl: */
#define IFX_SSC_LSB_FIRST            0
#define IFX_SSC_MSB_FIRST            1
/* dataWidth: */
#define IFX_SSC_MIN_DATA_WIDTH       2
#define IFX_SSC_MAX_DATA_WIDTH       32
/* master/slave mode select */
#define IFX_SSC_MASTER_MODE          1
#define IFX_SSC_SLAVE_MODE           0
/* rx/tx mode */
#define IFX_SSC_MODE_RXTX        0
#define IFX_SSC_MODE_RX          1
#define IFX_SSC_MODE_TX          2
#define IFX_SSC_MODE_OFF         3
#define IFX_SSC_MODE_MASK        IFX_SSC_MODE_RX | IFX_SSC_MODE_TX

/* GPO values */
#define IFX_SSC_MAX_GPO_OUT      7

#define IFX_SSC_RXREQ_BLOCK_SIZE     32768

/***********************/
/* defines for ioctl's */
/***********************/
#define IFX_SSC_IOCTL_MAGIC     'S'
/**
 * Read out the statistics.
 * \param   arg #ifx_ssc_statistics
 * \ingroup IOCTL
 */
#define IFX_SSC_STATS_READ _IOR(IFX_SSC_IOCTL_MAGIC, 1, struct ifx_ssc_statistics)
/**
 * Clear the statistics.
 * \ingroup IOCTL
 */
#define IFX_SSC_STATS_RESET _IO(IFX_SSC_IOCTL_MAGIC, 2)
/**
 * Set the baudrate.
 * \param   arg baudrate
 * \ingroup IOCTL
 */
#define IFX_SSC_BAUD_SET _IOW(IFX_SSC_IOCTL_MAGIC, 3, unsigned int)
/**
 * Get the current baudrate.
 * \param   arg baudrate
 * \ingroup IOCTL
 */
#define IFX_SSC_BAUD_GET _IOR(IFX_SSC_IOCTL_MAGIC, 4, unsigned int)
/**
 * Set hardware options.
 * \param   arg #ifx_ssc_hwopts
 * \ingroup IOCTL
 */
#define IFX_SSC_HWOPTS_SET _IOW(IFX_SSC_IOCTL_MAGIC, 5, struct ifx_ssc_hwopts)
/**
 * Get the current hardware options.
 * \param   arg #ifx_ssc_hwopts
 * \ingroup IOCTL
 */
#define IFX_SSC_HWOPTS_GET _IOR(IFX_SSC_IOCTL_MAGIC, 6, struct ifx_ssc_hwopts)
/**
 * Set transmission mode.
 * \param   arg Transmission mode
 * \ingroup IOCTL
 */
#define IFX_SSC_RXTX_MODE_SET _IOW(IFX_SSC_IOCTL_MAGIC, 7, unsigned int)
/**
 * Get the current transmission mode.
 * \param   arg Transmission mode
 * \ingroup IOCTL
 */
#define IFX_SSC_RXTX_MODE_GET _IOR(IFX_SSC_IOCTL_MAGIC, 8, unsigned int)
/**
 * Abort transmission.
 * \ingroup IOCTL
 */
#define IFX_SSC_ABORT _IO(IFX_SSC_IOCTL_MAGIC, 9)
/**
 * Set general purpose outputs.
 * \ingroup IOCTL
 */
#define IFX_SSC_FIFO_FLUSH _IO(IFX_SSC_IOCTL_MAGIC, 10)
/**
 * Set general purpose outputs.
 * \param   arg Port control
 * \ingroup IOCTL
 */
#define IFX_SSC_GPO_OUT_SET _IOW(IFX_SSC_IOCTL_MAGIC, 11, unsigned int)
/**
 * Clear general purpose outputs.
 * \param   arg Port control
 * \ingroup IOCTL
 */
#define IFX_SSC_GPO_OUT_CLR _IOW(IFX_SSC_IOCTL_MAGIC, 12, unsigned int)
/**
 * Get general purpose outputs.
 * \param   arg Port control
 * \ingroup IOCTL
 */
#define IFX_SSC_GPO_OUT_GET _IOR(IFX_SSC_IOCTL_MAGIC, 13, unsigned int)
/**
 * Get status of serial framing.
 * \param   arg Port control
 * \ingroup IOCTL
 */
#define IFX_SSC_FRM_STATUS_GET _IOR(IFX_SSC_IOCTL_MAGIC, 14, struct ifx_ssc_frm_status)
/**
 * Get counter reload values and control bits.
 * \param   arg #ifx_ssc_frm_status
 * \ingroup IOCTL
 */
#define IFX_SSC_FRM_CONTROL_GET _IOR(IFX_SSC_IOCTL_MAGIC, 15, struct ifx_ssc_frm_opts)
/**
 * Set counter reload values and control bits.
 * \param   arg #ifx_ssc_frm_opts
 * \ingroup IOCTL
 */
#define IFX_SSC_FRM_CONTROL_SET _IOW(IFX_SSC_IOCTL_MAGIC, 16, struct ifx_ssc_frm_opts)
/**
 * Enable DMA for SSC.
 * \param   arg Number of SSC
 * \ingroup IOCTL
 */
#define IFX_SSC_DMA_ENABLE _IO(IFX_SSC_IOCTL_MAGIC, 17)
/**
 * Disable DMA for SSC.
 * \param   arg Number of SSC
 * \ingroup IOCTL
 */
#define IFX_SSC_DMA_DISABLE _IO(IFX_SSC_IOCTL_MAGIC, 18)
/**
 * Free txbuf and rxbuf.
 * \ingroup IOCTL
 */
#define IFX_SSC_RESURRECT _IO(IFX_SSC_IOCTL_MAGIC, 19)

/**
 * Enable freeing of buffers for all SSCs.
 * \param   arg unused
 * \ingroup IOCTL
 */
#define IFX_SSC_BUFFER_FREE_ENABLE _IO(IFX_SSC_IOCTL_MAGIC, 20)
/**
 * Disable freeing of buffers for all SSCs.
 * \param   arg unused
 * \ingroup IOCTL
 */
#define IFX_SSC_BUFFER_FREE_DISABLE _IO(IFX_SSC_IOCTL_MAGIC, 21)


#ifdef __KERNEL__
/* kernel interface for read and write */
ssize_t ifx_ssc_kread(int, char *, size_t);
ssize_t ifx_ssc_kwrite(int, const char *, size_t);
int ifx_ssc_open(struct inode *inode, struct file *filp);
int ifx_ssc_close(struct inode *inode, struct file *filp);
int ifx_ssc_ioctl(struct inode *inode, struct file *filp, unsigned int cmd,
                  unsigned long data);
int ifx_ssc_register_frm_callback(int port, void (*func)(int port));
int ifx_ssc_unregister_frm_callback(int port);
ssize_t ifx_ssc_kframe(int port, char *buffer, int len, int pause );
ssize_t ifx_ssc_kframeInit( int port );

//sidecar
void sidecar_ifx_wait_sleep(void);
//sidecar end

#endif /* __KERNEL__ */
#endif /* __IFX_SSC_H */

